/*
 * $Id: Compressor.java 5239 2021-01-19 18:58:24Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_NODE;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_RELATION;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_WAY;

import java.io.File;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.OsmObject;
import de.cm.osm2po.model.OsmRelation;
import de.cm.osm2po.model.OsmRelation.Member;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamAdapterBuffered;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Reduziert die vom {@link Tiler} generierten Dateien auf
 * die tatsaechlich referenzierten bzw. validen Daten.
 * Untersucht dazu rueckwaerts, also entgegen der Einlese-Richtung,
 * die Relationen, Wege und Knoten erneut, schreibt die betroffenen
 * Dateien neu und aktualisiert die Statistiken des TileManagers.
 *
 * @see TileManager
 * @see Tiler
 * @see Joiner
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class Compressor implements Processor {

    private TileManager tileManager;

    private Log log;
    private File workDir;

    private LongArray relationIdSet;
    private LongArray wayIdSet;
    private LongArray nodeIdSet;

    @Override
    public void init(Config config) {
        try {
            this.log = config.getLog();
            this.workDir = config.getWorkDir();
            this.tileManager = new TileManager(config).readFromFile();

            this.relationIdSet = new LongArray();
            this.wayIdSet = new LongArray();
            this.nodeIdSet = new LongArray();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void run() {
        File relationsFile = new File(this.workDir, Tiler.RAW_RELATIONS_FILENAME);
        this.collectRelations(relationsFile);
        long nR = this.compress(relationsFile, this.relationIdSet);
        this.tileManager.setNumberOfRelations(nR);
        this.relationIdSet.clear();

        File waysFile = new File(this.workDir, Tiler.RAW_WAYS_FILENAME);
        this.collectWays(waysFile);
        long nW = this.compress(waysFile, this.wayIdSet);
        this.tileManager.setNumberOfWays(nW);
        this.wayIdSet.clear();

        this.tileManager.compressTiles(this.nodeIdSet);
        this.nodeIdSet.clear();

        this.tileManager.writeToFile();
    }

    @Override
    public void close() {
    }

    /**
     * Sammelt alle referenzierten Ids fuer Wege und Knoten.
     */
    private void collectWays(File file) {

        System.gc();
        log.info("Collecting Ways from "
                + file.getName() + " - " + MB());

        LongArray newWayIdSet = new LongArray();

        InStream inStream = new InStreamAdapterBuffered(file);

        byte varType = inStream.readByte();
        if (!(VarTypeDesk.typeIdOf(OsmWay.class) == varType))
            throw new RuntimeException("Unexpected VarType " + varType
                    + " in " + file.getAbsolutePath());

        long n = 0, nWays = this.tileManager.getNumberOfWays();
        OsmWay osmWay = new OsmWay();
        while (!inStream.isEof()) {
            osmWay.readFromStream(inStream);

            if (!osmWay.isInvalid() || this.wayIdSet.contains(osmWay.getId())) {
                newWayIdSet.push(osmWay.getId());

                int nRefs = osmWay.getRefsSize();
                for (int i = 0; i < nRefs; i++) {
                    this.nodeIdSet.push(osmWay.getRef(i));
                }
            }

            if (++n % 500000 == 0) log.progress(DF(n) + " of " + DF(nWays)
                    + " ways processed - " + MB());
        }
        log.info(DF(n) + " of " + DF(nWays) + " ways processed - " + MB());

        inStream.close();

        this.wayIdSet.clear();
        this.wayIdSet = newWayIdSet;

        log.progress("Building set of ways");
        this.wayIdSet.shrinkToSet();
        log.progress("Building set of nodes (Takes a while)");
        this.nodeIdSet.shrinkToSet();
    }


    /**
     * Sammelt alle referenzierten Ids fuer Relationen, Wege und Knoten.
     */
    private void collectRelations(File file) {
        // TODO Mehrstufige RelationToRelation-Abhaengigkeiten rekursieren

        System.gc();
        log.info("Collecting relations from "
                + file.getName() + " - " + MB());

        InStream inStream = new InStreamAdapterBuffered(file);

        byte varType = inStream.readByte();
        if (!(VarTypeDesk.typeIdOf(OsmRelation.class) == varType))
            throw new RuntimeException("Unexpected VarType " + varType
                    + " in " + file.getAbsolutePath());

        long n = 0, nRels = this.tileManager.getNumberOfRelations();
        OsmRelation osmRelation = new OsmRelation();
        while (!inStream.isEof()) {
            osmRelation.readFromStream(inStream);

            if (!osmRelation.isInvalid()) {
                this.relationIdSet.push(osmRelation.getId());
                for (Member member : osmRelation.getMembers()) {
                    long idRef = member.getRef();
                    switch (member.getType()) {
                    case MEMBER_TYPE_RELATION:
                        // sinnig, wenn RelationRelation moeglich
                        // this.relationIdSet.push(idRef);
                        break;
                    case MEMBER_TYPE_WAY:
                        this.wayIdSet.push(idRef);
                        break;
                    case MEMBER_TYPE_NODE:
                        this.nodeIdSet.push(idRef);
                        break;
                    }
                }
            }

            if (++n % 500000 == 0) log.progress(DF(n) + " of " + DF(nRels)
                    + " relations processed - " + MB());
        }
        log.info(DF(n) + " of " + DF(nRels) + " relations processed - " + MB());

        inStream.close();

        log.progress("Building set of relations");
        this.relationIdSet.shrinkToSet();
        log.progress("Building set of ways");
        this.wayIdSet.shrinkToSet();
        log.progress("Building set of nodes");
        this.nodeIdSet.shrinkToSet();
    }

    private long compress(File osm2poFile, LongArray idFilterSet) {
        File dir = osm2poFile.getParentFile();
        File tmpFile = new File(dir, osm2poFile.getName() + ".tmp");
        // TODO InStreamDiskDb anstatt InStreamDisk?
        InStream inStream = new InStreamDisk(osm2poFile);
        OutStream outStream = new OutStreamDisk(tmpFile);

        byte typeId = inStream.readByte();
        outStream.writeByte(typeId);
        Var<?> var = VarTypeDesk.newInstanceOf(typeId);

        long nRead = 0, nWritten = 0;

        while (!inStream.isEof()) {
            var.readFromStream(inStream);
            ++nRead;
            OsmObject varOsm = (OsmObject) var;
            if (idFilterSet.contains(varOsm.getId())) {
                var.writeToStream(outStream);
                nWritten++;
            }

            if (nRead % 500000 == 0) log.progress(
                    osm2poFile.getName() + " compressed from "
                    + DF(nRead) + " to " + DF(nWritten) + " elements");
        }

        inStream.close();
        outStream.close();

        osm2poFile.delete();
        tmpFile.renameTo(osm2poFile);

        log.info(osm2poFile.getName() + " compressed from "
                + DF(nRead) + " to " + DF(nWritten) + " elements");

        return nWritten;
    }

}
