/*
 * $Id: GraphBuilder.java 5848 2023-10-19 14:08:07Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.Utils.createTmpFile;
import static de.cm.osm2po.routing.GraphHeader.HEADER_SIZE_DEFAULT;
import static java.lang.Integer.MAX_VALUE;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Random;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.graph.GraphBuildInterceptor;
import de.cm.osm2po.extensions.GraphExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.model.Restriction;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.model.WaySegment;
import de.cm.osm2po.primitives.ByteArray;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.Index;
import de.cm.osm2po.primitives.IndexOpt;
import de.cm.osm2po.primitives.IndexSeqOpt;
import de.cm.osm2po.primitives.IntArray;
import de.cm.osm2po.primitives.IntLongBag;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamAdapterBuffered;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarByte;
import de.cm.osm2po.primitives.VarInt;
import de.cm.osm2po.primitives.VarMergeSorter;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.GraphHeader;

/**
 * Restrukturiert die vom {@link Segmenter} erzeugten Informationen,
 * so dass diese performant und platzsparend von einem Routing-Algo
 * verwendet werden koennen.<br>
 * Das Ergebnis ist eine einzige Routing-Datei. Mandatory-Infos
 * befinden sich in der Binaerdatei vorne, so dass diese mittels
 * {@link InStream} performant geladen werden koennen.<br>
 * Die Reihenfolge ist:
 * <ul>
 * <li>Header {@link GraphHeader} (Version, Anzahl Entities, FilePointer, etc.) </li>
 * <li>EdgeEntries {@link VarInt} Zeiger auf Kreuzungen. SourceVertexId-1=Index </li>
 * <li>Edges {@link Edge}
 *      sortiert nach SourceVertexId, angereichert um zwei
 *      DeadEdges als Sentinels am Anfang und Ende der Liste.
 *      Aufeinanderfolgende Edges mit gleicher SourceVertexId bilden eine
 *      Kreuzung von wegfuehrenden Kanten.
 * </li>
 * <li>VertexClasses {@link VarByte} VertexId-1=Index</li>
 * <li>Geometries {@link IndexOpt} von {@link LatLons} SegmentId-1=Index</li>
 * <li>EdgeNames {@link IndexSeqOpt} von {@link VarString} SegmentId-1=Index</li>
 * <li>Custom Extensions</li>
 * </ul>
 * Gelesene Segmente werden vom Builder in Edges (Hin- und Rueckrichtung),
 * also 2x geschrieben. Ungerade EdgeIDs sind die Originale,
 * wohingegen gerade IDs immer die Rueckrichtung bilden.<br>
 * Die Zaehlung beginnt bei 1. Das heisst die Reverse-Edge hat die Id 2.<br>
 * Ferner wird in das EdgeID-Bit(32) die Information isOneWay() hineinkodiert.<br>
 * Hieraus ergibt sich folgendes Bit-Muster:<br><br>
 * <code>
 * 0x.......1 =&gt; Hinweg, kein OneWay <br>
 * 0x.......0 =&gt; Rueckweg, kein OneWay <br>
 * 0x8......1 =&gt; OneWay ist Hinweg <br>
 * 0x8......0 =&gt; OneWay ist Rueckweg (WrongWay)
 * </code>
 * <br><br>
 * Java-seitig ist dies rueckwirkend folgendermassen ermittelbar <br><br>
 * <code>
 * boolean isWrongWay = (edgeId &amp; 0x80000001) == 0x80000000; <br>
 * boolean isOneWay = (edgeId &amp; 0x80000000) != 0 // oder == 0x80000000; <br>
 * int segmentId = (edgeId &amp; 0x7FFFFFFF) + 1 &gt;&gt;&gt; 1; // Durch 2 <br>
 * </code>
 * <br>
 * Nach Fertigstellung der Graph-Datei wird {@link GraphBuildInterceptor#onClose(File)}
 * gerufen, so dass hier nachgelagert zum Beispiel Extensions mittels
 * {@link #addExtension(File, Var, byte)} hinzugefuegt werden koennen.<br>
 * Alternativ hierzu kann dies auch ueber den letzten Verarbeitungsschritt automatisch
 * konfiguriert werden. Hierzu muessen Extensions entweder als Beans aus
 * {@link Config#getBean(String)} oder als Dateien vorliegen und vom Typ {@link VarBean}
 * oder {@link Var} sein. Bei letzterem muss dem Mechanismus zusaetzlich die
 * <code>typeId</code> bekannt gemacht werden. Config Beispiel:<br><br>
 * <code>graph.build.addex.X=varFilePath|varBeanName,id</code><br><br>
 * wobei X ein Zaehler-Schluessel (1-256) ist.
 *
 * @see Graph#isOneWay(int)
 * @see Graph#isReverse(int)
 * @see Graph#isWrongWay(int)
 * @see Graph#toSegmentId(int)
 * @see GraphBuildInterceptor
 * @see GraphExtension
 *
 * @author (c) 2012/2017 - Carsten Moeller - info@osm2po
 */
public class GraphBuilder implements Processor {
    public static final int COMPAT_FILE_TYPE = 0x1A2B3C4D; // DateiTypKennung.
    public static final int COMPAT_VERSION = 22; // Kompatibilitaets-Nummer

    private Config config;
    private Log log;
    private File workDir;
    private boolean containsWrongWays;
    private GraphBuildInterceptor graphBuildInterceptor;

    private GraphHeader graphHeader;
    private int nSegmentsTotal;
    private int nEdges;
    private int nVertices;

    @Override
    public void init(Config config) {
        try {
            this.config = config;
            this.log = config.getLog();
            this.workDir = config.getWorkDir();

            this.containsWrongWays = !Boolean.valueOf(
                    config.getProperty("graph.build.excludeWrongWays"));

            this.graphBuildInterceptor = config.getGraphBuildInterceptorClass().getDeclaredConstructor().newInstance();

            int graphHeaderSize = Integer.decode(config.getProperty(
                    "graph.build.headerSize", HEADER_SIZE_DEFAULT + ""));
            this.graphHeader = new GraphHeader(graphHeaderSize);

        } catch (Exception e) {throw new RuntimeException(e);}
    }

    @Override
    public void run() {
        File graphFile = this.config.findGraphFile(false);
        OutStreamDisk graphOutStream = new OutStreamDisk(graphFile);

        this.graphBuildInterceptor.onOpen(this.config); // wichtig

        this.graphHeader.setContainsWrongWays(this.containsWrongWays);
        this.graphHeader.writeToStream(graphOutStream); // Header leer fuellen.

        System.gc();
        this.log.info("Transforming Ways - " + MB());
        IndexOpt<LatLons> geomsIndex = new IndexOpt<LatLons>();
        geomsIndex.create(this.workDir, this.log);
        IndexSeqOpt<VarString> edgeNameIndex = new IndexSeqOpt<VarString>();
        edgeNameIndex.create(this.workDir, this.log);

        VarMergeSorter<Edge> edges = new VarMergeSorter<Edge>(
                Edge.class, 0x400000/*4M*/, this.workDir);
        this.processWays(edges, edgeNameIndex, geomsIndex);
        this.graphHeader.setNumberOfSegments(this.nSegmentsTotal);
        this.graphHeader.setNumberOfEdges(this.nEdges);

        this.log.info("Transforming Vertices - " + MB());

        File vertexCoordsFile = createTmpFile(this.workDir);
        File vertexClazzesFile = createTmpFile(this.workDir);
        IntLongBag viaToBag = this.processVerticesAndCacheRestrictions(
                vertexCoordsFile, vertexClazzesFile);
        this.graphHeader.setNumberOfVertices(this.nVertices);
        this.log.info("Finding SourceVertex EntryPoints - " + MB());
        int[] edgeIds = new int[this.nEdges];
        int[] edgeSourceIds = new int[this.nEdges];
        int[] edgeTargetIds = new int[this.nEdges];
        int[] edgeEntries = new int[this.nVertices];
        this.cacheVertexEntryPointsAndEdges(
                edges, edgeEntries, edgeIds, edgeSourceIds, edgeTargetIds);

        this.log.info("Creating NoTurnBits - " + MB());
        byte[] noTurnBits = this.resolveNoTurnBits(viaToBag,
                edgeEntries, edgeIds, edgeSourceIds, edgeTargetIds);

        this.log.info("Writing GraphFile " + graphFile.getName() + " - " + MB());
        this.writeGraph(graphOutStream, edges, edgeEntries, noTurnBits,
                vertexCoordsFile, vertexClazzesFile, edgeNameIndex, geomsIndex);

        graphOutStream.close();

        RandomAccess ra = new RandomAccessDisk(graphFile, false);
        this.graphHeader.writeToStream(ra); // Header mit Echtdaten ueberschreiben.
        ra.close();

        // Speicher freigeben.
        edges.clear();
        edges = null;
        viaToBag.clear();
        viaToBag = null;
        edgeEntries = null;

        // Erst hier, damit Interceptor nachgelagert den Graphen erweitern kann
        this.graphBuildInterceptor.onClose(graphFile);

        this.addExtensions(graphFile);
    }

    @Override
    public void close() {
    }

    private void processWays(VarMergeSorter<Edge> edges,
            Index<VarString> edgeNameIndex, Index<LatLons> geomsIndex) {

        // Eingabedatei aller Wege
        final File waysInFile = new File(this.workDir, Segmenter.SEGMENTS_FILENAME);
        final InStream waysInStream = new InStreamDisk(waysInFile);
        byte varType = waysInStream.readByte();
        if (varType != VarTypeDesk.typeIdOf(SegmentedWay.class))
            throw new RuntimeException("Unexpected VarType " + varType);

        // Teile der Ways werden als Edges fuer Weiterverarbeitung gecacht.

        SegmentedWay way = new SegmentedWay(); // Holder
        int segIdGuard = 0; // Ueberwacht fortlaufende IDs.
        int nWays = 0;
        while (!waysInStream.isEof()) {
            nWays++;
            way.readFromStream(waysInStream); // Weg deserialisieren.
            this.graphBuildInterceptor.onWay(way);

            VarString name = way.getName(); // New Instance
            WaySegment[] segments = way.getSegments();
            int nWaySegments = segments.length;
            boolean isOneWay = way.isOneWay();

            for (int i = 0; i < nWaySegments; i++) {
                LatLons coords = way.getSegmentGeometry(i);

                if (this.nSegmentsTotal != edgeNameIndex.push(name))
                    throw new RuntimeException("EdgeName-Index not in sync");
                if (this.nSegmentsTotal != geomsIndex.push(coords))
                    throw new RuntimeException("Geometry-Index not in sync");

                this.nSegmentsTotal++;

                WaySegment segment = segments[i];
                int segmentId = segment.getId();
                int sourceId = segment.getSourceId();
                int targetId = segment.getTargetId();

                float km = (float) coords.calcLengthKm();
                // OSM-Bug: Knoten mit unterschiedlichen IDs aber gleichen Koordinaten!
                if (km <= 1e-6f) { // Fix: Minimale Laenge auf 1 Millimeter setzen
                    km = 1e-6f;
                    if (this.log.isEnabled(Log.LEVEL_BUGFIX))
                        this.log.bugfix("Zero length fixed in " + way);
                }

                if (++segIdGuard != segmentId) throw new RuntimeException(
                        "SegmentId " + segIdGuard + " not in sequence");

                // SegmentId = 2 x EdgeId - Trick: Gerade IDs sind reverse!!!
                int edgeId = ((segmentId -1) << 1) +1;
                if (isOneWay) edgeId |= 0x80000000; // negative Id.

                int flags = 0xFF & way.getClazz();
                int kmh = way.getKmh();
                if (kmh < 1) kmh = 1; // Division by Zero sichern.
                float h = km / kmh;

                {// Source -> Target
                    Edge edge = new Edge(edgeId, sourceId, targetId, km, h, flags);
                    this.graphBuildInterceptor.onEdgeCreate(edge, i, false);
                    edges.add(edge);
                    this.nEdges++;
                }
                if (!isOneWay || this.containsWrongWays) { // Target -> Source
                    edgeId += 1; // Reverse: gerade Id - funktioniert auch mit Bit32
                    // Achtung! Source/Target Rollentausch!
                    Edge edge = new Edge(edgeId, targetId, sourceId, km, h, flags);
                    this.graphBuildInterceptor.onEdgeCreate(edge, i, true);
                    edges.add(edge);
                    this.nEdges++;
                }
            }

            if (nWays % 100000 == 0)
                this.log.progress(DF(nWays) + " Ways read - " + MB());
        }

        // DeadEdges einfuegen
        Edge firstSentinelEdge = new Edge(0, 0, 0, 0, 0, 0);
        edges.add(firstSentinelEdge); // Initial SourceId = 0
        this.nEdges++;
        // Nachfolgend wird nach Source sortiert, daher auf Max setzen
        Edge lastSentinelEdge = new Edge(0, MAX_VALUE, 0, 0, 0, 0);
        edges.add(lastSentinelEdge);
        this.nEdges++;

        this.log.info(DF(nWays) + " Ways read, "
                + DF(this.nEdges) + " Edges cached - " + MB());

        waysInStream.close();
    }

    private IntLongBag processVerticesAndCacheRestrictions(
            File vertexCoordsFile, File vertexClazzesFile) {

        // Kuenstliches Bag fuer Vertices
        ByteArray vertexClazzes = new ByteArray();
        LongArray vertexCoords = new LongArray(vertexClazzes);
        IntArray vertexIds = new IntArray(vertexCoords);

        IntLongBag viaToBag = new IntLongBag();

        File vInFile = new File(this.workDir, Segmenter.VERTICES_FILENAME);
        InStream vInStream = new InStreamDisk(vInFile);
        byte varType = vInStream.readByte();
        if (varType != VarTypeDesk.typeIdOf(Vertex.class))
            throw new RuntimeException("Unexpected VarType " + varType);

        this.nVertices = 0;
        int vertexIdMax = 0;
        Vertex vertex = new Vertex();

        while (!vInStream.isEof()) {
            vertex.readFromStream(vInStream);
            this.graphBuildInterceptor.onVertex(vertex);

            int vertexId = vertex.getId();
            Restriction[] vnrs = vertex.getRestrictions();

            // Kuenstlicher Bag-Eintrag
            vertexIds.push(vertexId);
            vertexCoords.push(vertex.getCoord());
            vertexClazzes.push(vertex.getClazz());

            if (vnrs != null) {
                int nVnrs = vertex.getRestrictions().length;
                for (int i = 0; i < nVnrs; i++) {
                    // ViaNodeRestriction wandeln und cachen.
                    int from = vnrs[i].getFrom();
                    long via = vertexId;
                    long to = vnrs[i].getTo();
                    byte clazz = vnrs[i].getClazz();
                    // Kann passieren: Nur gueltige Seg-Id-Kombis verarbeiten.
                    if (0 == from || 0 == to) {
                        continue;
                    }

                    // Trick: No-Turns (ungerade clazz) als negative IDs schreiben.
                    if ((clazz & 1) != 0) {
                        via *= -1;
                    }
                    long viaTo = (via << 32) | to;

                    viaToBag.add(from, viaTo);
                }
            }

            if (vertexId > vertexIdMax) vertexIdMax = vertexId;
            this.nVertices++;

            if (this.nVertices % 500000 == 0)
                this.log.progress(DF(this.nVertices) + " Vertices read - " + MB());
        }
        if (vertexIdMax != this.nVertices) {
            throw new RuntimeException("Number of Vertices does not match highest VertexId");
        }

        vInStream.close();

        this.log.info(DF(this.nVertices) + " Vertices read, "
                + DF(viaToBag.getSize()) + " Restrictions cached - " + MB());


        // Koordinaten sortieren und temporaer rausschreiben
        vertexIds.sort(); // Cascading: SortCoords, sortClazzes. See above.

        // Koordinaten rausschreiben
        OutStream coordsOutStream = new OutStreamDisk(vertexCoordsFile);
        LatLon latLon = new LatLon();
        int i = 0;
        for (i = 0; i < this.nVertices; i++) {
            long coord = vertexCoords.get(i);
            latLon.setCoord(coord);
            latLon.writeToStream(coordsOutStream);
            if (i % 1000000 == 0)
                this.log.progress(DF(i) + " of " + DF(this.nVertices) + " Coords written");
        }
        this.log.progress(DF(i) + " of " + DF(this.nVertices) + " Coords written");
        coordsOutStream.close();

        // Vertex-Clazzes rausschreiben.
        OutStream clazzesOutStream = new OutStreamDisk(vertexClazzesFile);
        int j = 0;
        for (j = 0; j < this.nVertices; j++) {
            byte clazz = vertexClazzes.get(j);
            clazzesOutStream.writeByte(clazz);
            if (j % 1000000 == 0)
                this.log.progress(DF(j) + " of " + DF(this.nVertices) + " Classes written");
        }
        this.log.progress(DF(j) + " of " + DF(this.nVertices) + " Classes written");
        clazzesOutStream.close();

        // Speicher freigeben
        vertexIds.clear();
        vertexCoords.clear();
        vertexClazzes.clear();

        this.log.progress("Sorting Restrictions");
        viaToBag.sort();
        return viaToBag;
    }

    /**
     * Iteriert ueber die nach SourceVertexId sortierten
     * Edges und erstellt das 0-wertige Vertex-Array.
     * Das heisst, der Vertex mit der Id 1 ist an Position 0
     * und zeigt auf die erste abgehende Kante innerhalb einer
     * Liste von nach Source-Vertex sortierten Kanten.
     * Gleichzeitig werden in diesem Durchlauf die noetigen
     * Edge-Informationen fuer die Weiterverarbeitung extrahiert.
     * @param edges Bereits gefuellter {@link VarMergeSorter} von Edges.
     * @param edgeEntries int[]
     * @param edgeIds int[]
     * @param edgeSourceIds int[]
     * @param edgeTargetIds int[]
     */
    private void cacheVertexEntryPointsAndEdges(VarMergeSorter<Edge> edges,
            int[] edgeEntries, int[] edgeIds, int[] edgeSourceIds, int[] edgeTargetIds) {

        int nTargeOnlyVertex = 0;
        int edgeIdx = 0;
        int lastVertexId = 0;

        for (Edge edge : edges) { // Fuehrt implizit MergeSort durch
            int edgeId = edge.getId();
            edgeIds[edgeIdx] = edgeId;

            this.graphBuildInterceptor.onEdgePosition(edgeIdx, edge);

            edgeSourceIds[edgeIdx] = edge.getSourceId();
            edgeTargetIds[edgeIdx] = edge.getTargetId();

            int vertexId = edge.getSourceId();
            // DeadEdges-VertexIds sind 0 u. MaxValue
            if (vertexId < MAX_VALUE && vertexId != lastVertexId) {
                int deltaId = vertexId - lastVertexId;
                if (deltaId > 1) {
                    this.log.bugfix("TargetOnly-Vertex detected in Range "
                            + (lastVertexId + 1) + "-" + (vertexId -1));
                    nTargeOnlyVertex += deltaId - 1;
                }
                edgeEntries[vertexId -1] = edgeIdx;
                lastVertexId = vertexId;
            }

            edgeIdx++;

            if ((edgeIdx & 0x1FFFFF) == 0x1FFFFF)
                this.log.progress(DF(edgeIdx) + " Edges analyzed");
        }

        this.log.info(DF(edgeIdx) + " Edges analyzed");
        if (nTargeOnlyVertex > 0) this.log.warn(
                DF(nTargeOnlyVertex) + " TargetOnly-Vertices detected!");

        // Sollte ausgerechnet der letzte Vertex ein
        // Target-Only-Vertex sein, so ist er automatisch
        // mit 0 vorbelegt.
    }


    /**
     * Wandelt Restrictions in reine No-Turns und liefert diese
     * als Byte-Array zurueck.
     * Dies ist notwendig, da OSM NoTurns und OnlyTurns kennt.
     * OnlyTurns bedeuten jedoch, dass man die komplette Kreuzung
     * betrachten muss, um auf NoTurns anderer Strassen zu schliessen.
     * @param viaToBag Bereits ermittelte SegId-Paarungen (No and Only).
     * @param edgeEntries int[]
     * @param edgeIds int[]
     * @param edgeSourceIds int[]
     * @param edgeTargetIds int[]
     * @return byte[] NoTurnBits pro Edge.
     */
    private byte[] resolveNoTurnBits(IntLongBag viaToBag,
            int[] edgeEntries, int[] edgeIds, int[] edgeSourceIds, int[] edgeTargetIds) {

        int nCheckedCrossings = 0;
        int nCheckedEdges = 0;

        byte[] noTurnBits = new byte[this.nEdges];

        for (int fromEdgeIdx = 0; fromEdgeIdx < this.nEdges; fromEdgeIdx++) {
            int fromEdgeId = edgeIds[fromEdgeIdx];
            int fromSegmentId = (fromEdgeId & 0x7FFFFFFF) + 1 >>> 1; // SegId rekonstruieren
            int viaVertexId = edgeTargetIds[fromEdgeIdx];

            if (viaVertexId > 0) { // Gibt es ueberhaupt einen Abbieger?
                int toEdgeIdx = edgeEntries[viaVertexId -1];

                // Gibt es Abbiegevorschriften zu abgehenden Segmenten
                long[] viaTos = viaToBag.getAll(fromSegmentId);
                if (viaTos != null) {
                    nCheckedCrossings++;
                    boolean onlyTurnMode = false; // false=noTurn true=onlyTurn
                    int turnBits = 0;
                    int turnBitIdx = -1;

                    // Kreuzung betrachten und alle abgehenden Segmente untersuchen.
                    do {
                        if (++turnBitIdx > 7) {
                            if (turnBits != 0) this.log.bugfix(
                                    "Too many TurnRestriction via Vertex  " + viaVertexId);
                            turnBits = 0;
                            break;
                        }
                        int toEdgeId = edgeIds[toEdgeIdx];
                        int toSegmentId = (toEdgeId & 0x7FFFFFFF) + 1 >>> 1; // Segment-ID rekonstruieren
                        int turnBit = 1 << turnBitIdx;

                        // Fuer Folge-Segmente alle Restriktionen ueberpruefen
                        int nNoOnlySwap = 0;
                        for (int i = 0; i < viaTos.length; i++) {
                            long viaTo = viaTos[i];
                            int restrToId = (int) viaTo; // Trick Bits 0-31 sind toSegId
                            int restrViaId = (int) (viaTo >> 32); // Obere Bits sind ViaVertexId
                            boolean isOnlyTurnEdge = restrViaId > 0; // Positiv = OnlyTurn
                            if (!isOnlyTurnEdge) {
                                restrViaId *= -1; // Id zurueckwandeln
                            }

                            if (restrToId == toSegmentId) { // Moegliche Restriktion!
                                // Kann passieren wenn wenn z.B. Way wie U mit Balken drueber
                                // TODO Ueberpruefen: Hier kommen in letzter Zeit hunderte durch!
                                if (restrViaId != viaVertexId) {
                                    if (this.log.isEnabled(Log.LEVEL_DATA)) {
                                        this.log.data("Vertices not equal viaId:"
                                                + restrViaId + " vertexId:" + viaVertexId);
                                    }
                                } else { // Ristriktion
                                    if (isOnlyTurnEdge) {
                                        if (!onlyTurnMode) {
                                            nNoOnlySwap++;
                                            onlyTurnMode = true;
                                            turnBits = 0; // Bei Wechsel Reset. No und Only Exklusiv!
                                        }
                                    } else {
                                        if (onlyTurnMode) {
                                            nNoOnlySwap++;
                                            onlyTurnMode = false;
                                            turnBits = 0; // Bei Wechsel Reset. No und Only Exklusiv!
                                        }
                                    }
                                    turnBits |= turnBit;
                                }
                            }
                        }
                        if (nNoOnlySwap > 1) {
                            this.log.warn("Mixed NoTurn-OnlyTurn via Vertex " + viaVertexId);
                        }

                    } while (edgeSourceIds[++toEdgeIdx] == viaVertexId);

                    if (turnBits != 0) {
                        if (onlyTurnMode) {
                            turnBits =~turnBits;
                        }
                        noTurnBits[fromEdgeIdx] = (byte) turnBits;
                    }
                }
            }
            if (++nCheckedEdges % 1000000 == 0)
                this.log.progress(DF(nCheckedEdges) + " Edges analyzed");

        }
        this.log.info(DF(nCheckedEdges) + " Edges analyzed, "
                + DF(nCheckedCrossings) + " Crossings checked");

        return noTurnBits;
    }

    private void writeGraph(OutStreamDisk outStreamDisk,
            VarMergeSorter<Edge> edges, int[] edgeEntries, byte[] noTurnBits,
            File vertexCoordsFile, File vertexClazzesFile,
            IndexSeqOpt<VarString> edgeNameIndex, IndexOpt<LatLons> geomsIndex) {

        int graphId = Math.abs(new Random().nextInt());
        this.log.debug("GraphId is " + graphId);
        this.graphHeader.setId(graphId);

        this.graphHeader.setFptrEdgeEntries(outStreamDisk.getWriteIndex());
        for (int i = 0; i < this.nVertices; i++) {
            outStreamDisk.writeInt(edgeEntries[i]);
            if (i % 1000000 == 0) this.log.progress(DF(i)
                    + " of " + DF(this.nVertices) + " vertices written");
        }
        this.log.info(DF(this.nVertices) + " vertices written");

        this.graphHeader.setFptrEdges(outStreamDisk.getWriteIndex());
        int edgeIdx = 0;
        for (Edge edge : edges) {
            edge.setNoTurnBits(noTurnBits[edgeIdx]);
            edge.writeToStream(outStreamDisk);
            if (++edgeIdx % 1000000 == 0) this.log.progress(DF(edgeIdx)
                    + " of " + DF(this.nEdges) + " edges written");
        }
        this.log.info(DF(this.nEdges) + " edges written");

        this.graphHeader.setFptrVertexClazzes(outStreamDisk.getWriteIndex());
        this.log.progress("Concatenating Vertex Classes");
        outStreamDisk.concatFile(vertexClazzesFile, 0);
        vertexClazzesFile.delete();
        this.graphHeader.setFptrVertexCoords(outStreamDisk.getWriteIndex());
        this.log.progress("Concatenating Vertex Coords");
        outStreamDisk.concatFile(vertexCoordsFile, 0);
        vertexCoordsFile.delete();
        this.graphHeader.setFptrGeomIndex(outStreamDisk.getWriteIndex());
        this.log.progress("Concatenating Geometries");
        geomsIndex.build(outStreamDisk);
        this.graphHeader.setFptrEdgeNameIndex(outStreamDisk.getWriteIndex());
        this.log.progress("Concatenating EdgeNames");
        edgeNameIndex.build(outStreamDisk);
        // EOF
        this.graphHeader.setFptrBaseEof(outStreamDisk.getWriteIndex());
        this.graphHeader.setFptrEof(this.graphHeader.getFptrBaseEof());
    }

    /**
     * Erweitert den Graphen bzw. die Graph-Datei nachgelagert
     * um benutzerdefinierte Daten aus einem bereits im Speicher
     * befindlichen Objekt.
     * @param graphFile Pfad zur Graph-Datei
     * @param var {@link Var} Daten-Objekt
     * @param id byte (&gt;0) Id
     * @throws Exception beliebiger Natur
     */
    public static void addExtension(File graphFile, Var<?> var, byte id) throws Exception {
        OutStream os = null;
        try {
            GraphHeader gh = new GraphHeader(graphFile);
            @SuppressWarnings("unchecked")
            Class<Var<?>> varClass = (Class<Var<?>>) var.getClass();

            if (gh.isExtensionExist(varClass, id)) {
                throw new IllegalStateException(
                        "Extension " + varClass.getSimpleName() + "#" + id + " already exists");
            }

            long offset = graphFile.length();

            // OutStream ist schneller als RandomAccess
            OutputStream outputStream = new FileOutputStream(graphFile, true); // true=append
            os = new OutStreamAdapterBuffered(outputStream);

            var.writeToStream(os);
            long size = os.getWriteIndex();
            os.close(); // also closes underlying Stream

            gh.addExtension(varClass, id, offset, size);

            RandomAccess ra = new RandomAccessDisk(graphFile, false);
            gh.setFptrEof(offset + size);
            gh.writeToStream(ra);
            ra.close();

        } finally {
            if (os != null) os.close();
        }
    }

    /**
     * Erweitert den Graphen bzw. die Graph-Datei nachgelagert
     * um benutzerdefinierte Daten. Da hier aus einem Stream gelesen
     * wird, koennen so auch mehrere Extensions aus einer Quelle
     * gelesen werden.
     * @param graphFile Pfad zur Graph-Datei
     * @param varType byte Typ aus {@link VarTypeDesk}
     * @param id byte (&gt;0) Id
     * @param size long Anzahl zu schreibender Bytes
     * @param inStream {@link InStream}
     * @throws Exception beliebiger Natur
     */
    public static void addExtension(File graphFile,
            byte varType, byte id, long size, InStream inStream) throws Exception {
        OutStream os = null;
        try {
            Class<? extends Var<?>> varClass = VarTypeDesk.classOfType(varType);
            GraphHeader gh = new GraphHeader(graphFile);
            if (gh.isExtensionExist(varClass, id)) {
                throw new IllegalStateException(
                        "Extension " + varClass.getSimpleName() + "#" + id + " already exists");
            }

            long offset = graphFile.length();

            // OutStream ist schneller als RandomAccess
            OutputStream outputStream = new FileOutputStream(graphFile, true);
            os = new OutStreamAdapterBuffered(outputStream);

            long nLeft = size;
            byte[] buffer = new byte[0x1000];
            while (nLeft > 0) {
                if (inStream.isEof()) throw new RuntimeException(
                    "InputStream is EOF but there are " + nLeft + " bytes left!");
                if (nLeft < buffer.length)
                    buffer = new byte[(int) nLeft];
                int nRead = inStream.readBytes(buffer);
                os.writeBytes(buffer, nRead);
                nLeft -= nRead;
            }
            os.close(); // also closes underlying Stream

            gh.addExtension(varClass, id, offset, size);

            RandomAccess ra = new RandomAccessDisk(graphFile, false);
            gh.setFptrEof(offset + size);
            gh.writeToStream(ra);
            ra.close();

            long bytesWritten = graphFile.length() - offset;
            if (size != bytesWritten) {
                throw new RuntimeException("Extension size does not match");
            }

        } finally {
            if (os != null) os.close();
        }
    }

    // VarBean oder VarBean-Datei als addex-property in Config.
    private void addExtensions(File graphFile) {
        for (int i = 0; i < 256; i++) {
            String key = "graph.build.addex." + i;
            String[] vals = config.getProperty(key, "").split(",");
            for (int v = 0; v < vals.length; v++) vals[v] = vals[v].trim();
            String bean = vals[0]; // BeanName oder BeanFile
            if ("".equals(bean)) continue;
            byte id = vals.length > 1 ? Byte.valueOf(vals[1]) : 0; // Optionale ID, sonst 0

            VarBean varBean = this.config.getBean(bean); // Bereits global geladen?
            try {
                if (varBean != null) { // VarBean, wenn aus Cache, vorher entpacken
                    Var<?> var = varBean.get();
                    this.log.info("Writing Extension " + var.getClass().getSimpleName() + "#" + id
                            + " from '" + bean + "'"
                            + " to " + graphFile.getName());
                    addExtension(graphFile, var, id);

                } else { // Dann Datei
                    File file = new File(bean);
                    if (file.exists()) {
                        InStream inStream = new InStreamDisk(file);
                        long size = file.length();
                        byte varType = inStream.readByte();
                        size--; // Wichtig! 1 Byte bereits gelesen
                        Class<? extends Var<?>> varClass = VarTypeDesk.classOfType(varType);
                        this.log.info("Writing Extension " + varType
                                + " " + varClass.getSimpleName() + "#" + id
                                + " from '" + bean + "'"
                                + " to " + graphFile.getName());
                        addExtension(graphFile, varType, id, size, inStream);
                        inStream.close();
                    }
                }

            } catch (Exception e) {
                log.error(e.getMessage());
            }
        }
    }

} // EOC
