/*
 * $Id: Joiner.java 5790 2023-04-13 16:17:55Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static java.util.Arrays.copyOf;

import java.io.File;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.joiner.DefaultJoinFilter;
import de.cm.osm2po.converter.joiner.JoinFilter;
import de.cm.osm2po.converter.joiner.RelationCache;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Membership;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmRelation;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.BigLongSet;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Diese Klasse sorgt im Wesentlichen fuer den Join-Lauf und generiert dabei
 * aus den vom {@link Tiler} generierten Dateien (Knoten, Wege, Relationen)
 * neue, ebenfalls nach {@link Tile}s sortierte Dateien fuer Wege. Diese
 * allerdings angereichert um alle noetigen Informationen z.B. Koordinaten,
 * Relationen usw., vergleichbar mit einem Datenbank Join.<br>
 * Da hier allerdings Unmengen von Daten aufkommen, wird auch hier mit Tiles
 * gearbeitet. Das heisst, es muessen Teilmengen (Knoten) gecacht und dann
 * gegen alle Wege aufgeloest werden. Dies passiert solange, bis alle Knoten
 * verarbeitet sind. Hieraus resultieren Wege, die entweder komplett in einem
 * Tile liegen, andere, die komplett ausserhalb sind und auch solche, die
 * mehrere Tiles kreuzen.<br>
 * Letztere werden einem eigenem <strong>Shared</strong>-Pool zugeordnet,
 * d.h. sie werden letztendlich in eine zusaetzliche Datei geschrieben.<br>
 * Der {@link Joiner} arbeitet immer mit einem {@link JoinFilter} zusammen.
 * @see Merger
 * @see Tiler
 * @see Segmenter
 * @see JoinFilter
 * @see DefaultJoinFilter
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class Joiner implements Processor {

    // TODO SHARED-Name hier - TILED-Name im TileManager - wieso?
    public static final String WAY_FILENAME_SHARED = "jw_shared.2po";
    public static final String WAY_FILENAME_ORPHANS = "jw_orphans.2po";

    private static final int WAY_UNRESOLVED = 0;
    private static final int WAY_RESOLVED = 1;
    private static final int WAY_IN_TILE = 2;
    private static final int WAY_IN_BUFFER = 4;

    private TileManager tileManager;
    private RelationCache osmRelationCache;

    private JoinFilter joinFilter;

    private Config config;
    private Log log;
    private File workDir;
    private int tileSizeY;
    private int tileSizeX;
    private boolean singleTileMode;

    private int nWaysSharedTotal;
    private int nWaysSharedNodesTotal;

    @Override
    public void init(Config config) {
        try {
            this.config = config;
            this.log = config.getLog();
            this.workDir = config.getWorkDir();
            this.tileManager = new TileManager(config).readFromFile();
            this.tileSizeY = this.tileManager.getTileSizeY();
            this.tileSizeX = this.tileManager.getTileSizeX();
            this.singleTileMode = this.tileManager.isSingleTileMode();
            this.joinFilter = config.getJoinFilterClass().getDeclaredConstructor().newInstance();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Cached alle Relationen und stellt diese zur verfuegung
     */
    private void cacheOsmRelations() {
        File relationsFile = new File(this.workDir, Tiler.RAW_RELATIONS_FILENAME);
        System.gc();
        this.log.info("Caching relations from "
                + relationsFile.getName() + " - " + MB());
        InStream inStream = new InStreamDisk(relationsFile);
        byte varType = inStream.readByte();
        if (!(VarTypeDesk.typeIdOf(OsmRelation.class) == varType))
            throw new RuntimeException("Unexpected VarType " + varType
                    + " in " + relationsFile.getAbsolutePath());

        long n = 0, nRels = this.tileManager.getNumberOfRelations();
        while (!inStream.isEof()) {
            OsmRelation osmRelation = new OsmRelation();
            osmRelation.readFromStream(inStream);
            if (osmRelation.isInvalid()) { // PostFilter, wenn Prefilter aus
                continue;
            }
            if (this.osmRelationCache.add(osmRelation)) {
                if (++n % 100000 == 0) this.log.progress(DF(n) + " of " + DF(nRels)
                        + " relations cached - " + MB());
            }
        }
        this.log.info(DF(n) + " of " + DF(nRels) + " relations cached - " + MB());

        inStream.close();
        this.osmRelationCache.sort();
    }


    /**
     * Startet den Lauf, in dem Wege, Knoten und Relationen
     * zusammengefuehrt werden.
     * <br>
     * Das heisst, hier wird ueber alle Wege iteriert und die
     * Ref(s) gegen eine vom Speicher begrenzte Sub-Menge an Node-Tiles
     * (Koordinaten u.a.) aufgeloest. Ist der Speicher nicht ausreichend,
     * so werden neue Tiles geladen und die Wege erneut durchlaufen.
     * <br>
     * Dabei werden Wege, deren Knoten komplett in einem Tile liegen,
     * wiederum in Tiles sortiert. Hingegen werden solche Wege, die mehrere
     * Tiles kreuzen, in eine globale Datei geschrieben und vermerkt,
     * so dass diese nicht doppelt verarbeitet werden.
     * <br>
     * Vorher werden bereits alle Relationen geladen.
     * Es gibt nicht allzu viele davon, die fuer ein Routing
     * relevant sind. Daher wird hier auf ein Tiling verzichtet.
     * <br><br>
     * Dieser Lauf koennte theoretisch bereits im
     * Konstruktor gestartet werden, jedoch ist solch ein Konstrukt
     * unschoen und zudem eroeffnet diese Trennung die Moeglichkeit,
     * vorher noch auf die Members, z.B. den {@link TileManager}
     * Einfluss zu nehmen.
     */
    @Override
    public void run() {
        this.joinFilter.open(this.config, null);

        Way wayHolder = new Way();
        File osmWayFile = new File(this.workDir, Tiler.RAW_WAYS_FILENAME);
        long nWaysTotal = this.tileManager.getNumberOfWays();
        long nWaysTiledTotal = 0;
        this.nWaysSharedNodesTotal = 0;
        this.nWaysSharedTotal = 0;

        // Alles fuers Shared Ways Merken und Schreiben.
        File sharedWayFile = new File(this.workDir, WAY_FILENAME_SHARED);
        OutStream sharedWaysOutStream = new OutStreamDisk(sharedWayFile);
        sharedWaysOutStream.writeByte(VarTypeDesk.typeIdOf(Way.class));
        // TODO HashSize optimieren. eu (18 Mio Ways) 20 Tiles a 15x15 nur 8000 shared!
        BigLongSet sharedWaysIdSet = new BigLongSet((int) (nWaysTotal / 100 + 100));
        this.log.debug("HashSize for shared wayIds is "+ DF(sharedWaysIdSet.getHashSize()));

        // Relationen cachen
        this.osmRelationCache = new RelationCache();
        this.cacheOsmRelations();

        // Orphaned Ways - Am Ende moeglicherweise unresolved.
        BigLongSet orphanedWaysIdSet = new BigLongSet(0x400);
        File orphanedWaysFile = new File(this.workDir, WAY_FILENAME_ORPHANS);
        OutStream orphanedWaysOutStream = new OutStreamDisk(orphanedWaysFile);
        orphanedWaysOutStream.writeByte(VarTypeDesk.typeIdOf(OsmWay.class));

        // Iteriere ueber eine Menge von bereits gecachten Tiles.
        // Tricky: Das Caching wird direkt im Iterator umgesetzt.
        for (List<Tile> cachedTileList : this.tileManager) {
            // Oeffne alle TiledWays-Ausgabe-Dateien-Stroeme
            this.openTiledWaysFiles(cachedTileList);

            // Oeffne Datei und iteriere ueber alle Ways.
            InStream osmWaysInStream = new InStreamDisk(osmWayFile);
            osmWaysInStream.readByte(); // OsmWay.VARTYPE-Byte
            long nWaysRead = 0, nWaysTiled = 0, nWaysShared = 0;

            // Iteriere ueber alle Wege
            while (!osmWaysInStream.isEof()) {

                // Weg deserialisieren und dabei upcasten.
                wayHolder.readFromOsmWayStream(osmWaysInStream);

                // Wurde der Weg bereits als Shared verarbeitet?
                if (sharedWaysIdSet.contains(wayHolder.getId())) continue;

                for (Tile tile : cachedTileList) {
                    // Welches Tile (inklusive Puffer) ist ueberhaupt betroffen?
                    // Der Startknoten ist entscheidend!
                    long firstRef = wayHolder.getRef(0);
                    if (!tile.containsNode(firstRef)) continue;

                    // Refs zu KNOTEN aufloesen und RELATIONEN ranpappen
                    int WAY_RESOLVE_STATE = this.resolveWay(wayHolder, tile);

                    // Liegt mindestens ein Knoten im Pufferbereich?
                    boolean isWayInBuffer = (WAY_RESOLVE_STATE & WAY_IN_BUFFER) != 0;
                    // Liegt mindestens ein Knoten echt im Tile?
                    boolean isWayInTile = (WAY_RESOLVE_STATE & WAY_IN_TILE) != 0;
                    // Wurden ueberhaupt alle Knoten aufgeloest?
                    boolean isWayResolved = (WAY_RESOLVE_STATE & WAY_RESOLVED) != 0;

                    if (isWayResolved) { // Alle Knoten des Wegs aufgeloest
                        /*
                         * Weg muss mindestens mit einem Knoten echt in diesem Tile
                         * liegen. Wenn nicht, dann ist er zwar aufgeloest, liegt aber
                         * komplett im Puffer. Das heisst, er wird im weiteren
                         * Verlauf nochmals behandelt und einem anderen Tile zugeordnet.
                         * Worst Case Form: Ein Ring im Pufferbereich eines Tiles.
                         */
                        if (isWayInTile) {

                            boolean mayWrite = this.joinFilter.validate(wayHolder);

                            if (mayWrite) {
                                int nNodes = wayHolder.getRefsSize();
                                if (isWayInBuffer) { // Weg kreuzt weitere Tiles

                                    // Behandlung von moeglichen Orphans:
                                    // isWayInBuffer reduziert die Menge erheblich.
                                    // Unwahrscheinlicher WorstCase hierbei: Weg liegt in
                                    // 2 Tiles aber mit keinem Knoten im Pufferbereich.
                                    orphanedWaysIdSet.remove(wayHolder.getId());

                                    sharedWaysIdSet.add(wayHolder.getId()); // ID merken
                                    wayHolder.writeToStream(sharedWaysOutStream); // Schreiben
                                    nWaysShared++; this.nWaysSharedTotal++;
                                    this.nWaysSharedNodesTotal += nNodes;

                                    this.joinFilter.tiledInto(-1, true);

                                } else { // Weg liegt nur in diesem Tile
                                    nWaysTiled++; nWaysTiledTotal++;
                                    wayHolder.writeToStream(tile.getOutStreamDisk());
                                    tile.incTiledWaysCounter();
                                    tile.incTiledWaysNodesCounter(nNodes);

                                    this.joinFilter.tiledInto(tile.getId(), true);
                                }

                            } else { // Nicht geschrieben
                                int tileId = isWayInBuffer ? -1 : tile.getId();
                                this.joinFilter.tiledInto(tileId, false);
                            }

                            break; // restliche Tiles ignorieren
                        }
                    } else { // nicht aufgeloest und geschrieben
                        if (isWayInTile) {
                            //  kann moeglicherweise im Nachlauf aufgeloest werden.
                            orphanedWaysIdSet.add(wayHolder.getId());
                            wayHolder.writeToOsmWayStream(orphanedWaysOutStream);
                        }
                    }
                }

                if (++nWaysRead % 100000 == 0) {
                    this.log.progress(DF(nWaysRead) + " of " + DF(nWaysTotal) + " ways read, "
                            + DF(nWaysTiled + nWaysShared) + " written");
                }
            }

            this.log.info(DF(nWaysRead) + " of " + DF(nWaysTotal) + " ways read, "
                    + DF(nWaysTiled + nWaysShared) + " written\nTotal "
                    + DF(nWaysTiledTotal) + " tiled, "
                    + DF(this.nWaysSharedTotal) + " shared");

            // Schliesse Eingabe Datei aller Wege
            osmWaysInStream.close();

            // Oeffne alle TiledWays-Ausgabe-Dateien-Stroeme
            this.closeTiledWaysFiles(cachedTileList);
        }


        long nWaysResolved = nWaysTiledTotal + this.nWaysSharedTotal;
        this.log.info(DF(nWaysResolved) + " of " + DF(nWaysTotal) + " ways resolved.");

        /*
         * Nachgelagerte Behandlung von Orphans (nicht aufgeloesten Wegen)
         * Im derzeitigen Planet bei 45x45,1.0 sind das in etwa 10 auf 71 Mio.
         * Diese koennen allerdings auch dadurch bedingt sein, dass aufgrund
         * von tileThreshold=100 einzelne Zielknoten nicht existieren.
         * z.B. Deutschland Faehre Luebeck-Malmoe
         */
        orphanedWaysOutStream.close();
        long nOrphans = orphanedWaysIdSet.getSize();
        if (nOrphans != 0) {
            this.log.warn(DF(nOrphans) + " orphaned ways registered.");
            this.log.debug("WayIds: " + Arrays.toString(orphanedWaysIdSet.getKeys()));
            if (Boolean.valueOf(this.config.getProperty("rescueOrphanedWays", "false"))) {
                this.log.info("Starting RescueMode at " + new Date().toString());
                this.rescueOrphanedWays(
                        orphanedWaysIdSet, orphanedWaysFile, sharedWaysOutStream);
                long nRescued = nOrphans - orphanedWaysIdSet.getSize();
                this.log.info(DF(nRescued) + " of " + DF(nOrphans) + " ways could be rescued");
                for (long wayId : orphanedWaysIdSet.getKeys()) {
                    this.log.debug("Still missing: http://openstreetmap.org/way/" + wayId);
                }
            } else {
                this.log.info("Increase [tileSize] and/or set [rescueOrphanedWays=true]");
            }
        }

        this.joinFilter.close();
        sharedWaysOutStream.close();

        // Letzte Statistiken in den Tilemanager und persistieren.
        this.tileManager.setNumberOfSharedWays(this.nWaysSharedTotal);
        this.tileManager.setNumberOfSharedWaysNodes(this.nWaysSharedNodesTotal);
        this.tileManager.writeToFile();
    }

    /**
     * Loest, wenn moeglich, die Knotenreferenzen eines Weges
     * gegen die gecachten Knoten eines Tiles auf und liefert
     * den Verarbeitungsstatus zurueck.<br>
     * @param way unaufgeloester {@link Way} noch mit Referenzen.
     * @param tile {@link Tile} - gecachte Knoten Tile+Buffer.
     * @return int status (binaer)
     *      z.B. {@link #WAY_IN_TILE} oder {@link #WAY_UNRESOLVED}.
     */
    private int resolveWay(Way way, Tile tile) {

        int WAY_RESOLVE_STATE = WAY_UNRESOLVED;

        // Array (Puffer) fuer aufgeloeste Knoten aufmachen.
        int refsSize = way.getRefsSize();
        Node nodes[] = new Node[refsSize];

        for (int refIdx = 0; refIdx < refsSize; refIdx++) {
            long ref = way.getRef(refIdx);
            Node nd = tile.lookupNode(ref); // Auto-UpCast
            if (null == nd) { // Knoten nich aufloesbar. Naechstes Tile.
                if (this.singleTileMode) this.log.warn(
                        "Node #" + ref + " does not exist in Way #" + way.getId());
                return WAY_RESOLVE_STATE;
            }

            nodes[refIdx] = nd;

            if (this.singleTileMode) {
                WAY_RESOLVE_STATE = WAY_IN_TILE;
            } else {
                // Wo liegt der Knoten?
                int tileId = Tile.getTileId(
                        nd.getLat(), nd.getLon(), this.tileSizeY, this.tileSizeX);
                if (tileId == tile.getId()) WAY_RESOLVE_STATE |= WAY_IN_TILE;
                    else WAY_RESOLVE_STATE |= WAY_IN_BUFFER;
            }
        }

        // Weg ist aufgeloest
        WAY_RESOLVE_STATE |= WAY_RESOLVED;

        // Nodes nachgelagert mit Relationen anreichern
        for (int i = 0; i < nodes.length; i++) {
            Node nd = nodes[i];
            final Membership[] nodeMemberships =
                this.osmRelationCache.lookupNodeMemberships(nd.getId()); // nullable
            nd.setMemberships(nodeMemberships);
        }

        // Way nachgelagert mit Relationen anreichern
        final Membership[] wayMemberships =
            this.osmRelationCache.lookupWayMemberships(way.getId()); // nullable

        // Super-Relation(s) ermitteln und Relations-Members selbst anreichern
        if (wayMemberships != null) for (Membership wayMembership : wayMemberships) {
            // TODO Super-Relations rekursiv ermitteln. Aktuell nur einstufig.
            // TODO Super-Relations ggf. auch fuer Knoten ermitteln. Aktuell kein UseCase.
            wayMembership.setParents(this.osmRelationCache
                    .lookupRelationMemberships(wayMembership.getId()));
        }

        // ### Hier ist endlich der Join ;-)
        way.setNodes(nodes);
        way.setMemberships(wayMemberships);

        return WAY_RESOLVE_STATE;
    }

    /**
     * Startet den Lauf, um verwaiste Wege aufzuloesen.
     * Gerettete Wege werden dem Shared-Tile zugewiesen.
     * @param orphanedWaysIdSet Set von verwaisten WayIds (Ein-/Ausgabe)
     * @param orphanedWaysFile 2po-File von {@link OsmWay}-Kandidaten
     * @param waysOutStream {@link OutStream} fuer aufgeloeste {@link Way}s
     */
    private void rescueOrphanedWays(
            BigLongSet orphanedWaysIdSet, File orphanedWaysFile, OutStream waysOutStream) {

        InStream inStream = new InStreamDisk(orphanedWaysFile);
        byte varType = inStream.readByte();
        if (!(VarTypeDesk.typeIdOf(OsmWay.class) == varType))
            throw new RuntimeException("Unexpected VarType " + varType
                    + " in " + orphanedWaysFile.getAbsolutePath());

        // Orphaned Ways sammeln und NodeIds (Refs)
        // in einem LongArray sammeln.
        // Ways und Nodes koennen durchaus mehrfach vorkommen.
        LongArray ndRefIds = new LongArray();

        this.log.info("Collecting missing NodeRefs");
        OsmWay osmWay = new OsmWay();
        while (!inStream.isEof()) {
            osmWay.readFromStream(inStream);
            if (orphanedWaysIdSet.contains(osmWay.getId())) {
                int nRefs = osmWay.getRefsSize();
                for (int i = 0; i < nRefs; i++) {
                    long ndRef = osmWay.getRef(i);
                    ndRefIds.push(ndRef);
                }
            }
        }
        inStream.close();

        // Kuenstliches Tile fuer die vermissten Knoten erstellen
        Tile tile = this.tileManager.createTileFromNodeIds(ndRefIds);

        // Ways erneut durchlaufen und retten.
        Way way = new Way();
        InStream waysInStream = new InStreamDisk(orphanedWaysFile);
        waysInStream.readByte(); // VarType already checked above

        this.log.info("Resolving missing NodeRefs");
        long[] notInBufs = new long[0x1000];
        int nNotInBufs = 0;
        while (!waysInStream.isEof()) {
            // Der selbe Weg kann hier durchaus mehrfach durchlaufen!
            way.readFromOsmWayStream(waysInStream);

            if (orphanedWaysIdSet.contains(way.getId())) {
                int WAY_RESOLVE_STATE = this.resolveWay(way, tile);
                boolean isWayInBuffer = (WAY_RESOLVE_STATE & WAY_IN_BUFFER) != 0;
                boolean isWayResolved = (WAY_RESOLVE_STATE & WAY_RESOLVED) != 0;

                if (isWayResolved) {
                    if (!isWayInBuffer) { // TODO ist dies gefaehrlich?
                        if (notInBufs.length == nNotInBufs) // ArrayOverflowCheck
                            notInBufs = copyOf(notInBufs, nNotInBufs * 2);
                        notInBufs[nNotInBufs++] = way.getId();
                    }

                    orphanedWaysIdSet.remove(way.getId());

                    if (this.joinFilter.validate(way)) {
                        way.writeToStream(waysOutStream);
                        this.nWaysSharedTotal++;
                        this.nWaysSharedNodesTotal += way.getRefsSize();
                        this.joinFilter.tiledInto(-2, true); // Shared
                    } else {
                        this.joinFilter.tiledInto(-2, false);
                    }
                }
            }
        }
        waysInStream.close();

        if (nNotInBufs > 0) {
            notInBufs = copyOf(notInBufs, nNotInBufs);
            this.log.warn(DF(nNotInBufs) + " rescued ways are not in Buffer");
            this.log.debug("WayIds: " + Arrays.toString(notInBufs));
        }
    }

    /**
     * Oeffnet alle Tiled-Ways-Files fuer den Durchlauf eines
     * Tilesets und resetted die Counter fuer den Joiner.
     * @param tileList {@link List} of {@link Tile}
     */
    private void openTiledWaysFiles(List<Tile> tileList) {
        for (Tile tile : tileList) {
            File tiledWayFile = new File(
                    this.workDir, TileManager.TILED_WAYS_FILENAME(tile.getId()));
            tile.setOutStreamDisk(new OutStreamDisk(tiledWayFile));
            tile.getOutStreamDisk().writeByte(VarTypeDesk.typeIdOf(Way.class));
            tile.resetTiledWaysCounter();
        }
    }

    /**
     * Schliesst alle Tiled-Ways-Files nach der Verarbeitung
     * eines Tilesets.
     * @param tileList {@link List} of {@link Tile}
     */
    private void closeTiledWaysFiles(List<Tile> tileList) {
        for (Tile tile : tileList) {
            if (tile.getOutStreamDisk() != null) {
                tile.getOutStreamDisk().close();
            }
        }
    }

    @Override
    public void close() {
        this.osmRelationCache.clear();
    }

}
