/*
 * $Id: Merger.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.misc.Utils.DF;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.Restriction;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Fuegt das Resultat mehrerer Joining-Ergebnisse
 * zu einem einzigen zusammen. Einzelextrakte, wie z.B.
 * Hamburg und Schleswig-Holstein koennen so miteinander
 * verbunden werden. Auch besteht die Moeglichkeit, verschiedene
 * Strassentypen / Konfigurationen uebereinanderzulegen.<br>
 * Als Parameter werden lediglich Verzeichnisse (Pfade) erwartet,
 * in denen sich die Resultate der {@link Joiner}-Laeufe befinden.
 * Das Merge-Ergebnis wird wie gewohnt im aktuellen
 * {@link Config#getWorkDir()} abgelegt.
 * <br><br><b>Wichtig:</b><br>
 * Intern werden doppelte Vorkommen der Wege aufgrund ihrer
 * IDs erkannt bzw. verworfen. In speziellen Faellen kann also
 * die Merge-Reihenfolge von Bedeutung sein. Dies gilt ebenfalls
 * fuer die Knoten, die durchaus mehrfach vorkommen koennen und
 * nicht zwangslaeufig die gleichen Zusatzinfos wie z.B.
 * {@link Restriction}s beinhalten muessen. Hier gilt das Prinzip
 * Der Erste gewinnt.
 * @author (c) 2012 - Carsten Moeller - info@osm2po.de
 */
public class Merger implements Processor {

    private Config config;
    private File workDir;
    private Log log;
    private List<TileManager> tmSourceList = new ArrayList<TileManager>();
    private Map<Integer, Tile> tmTargetTileMap = new HashMap<Integer, Tile>();
    private TileManager tmTarget;

    private int nSourceWays;

    @Override
    public void init(Config config) {
        this.config = config;
        this.workDir = config.getWorkDir();
        this.log = config.getLog();
    }

    @Override
    public void run() {
        if (!this.setupSources(config.getDirs())) return;
        this.setupTarget();
        this.merge();

        this.tmTarget.setTiles(this.tmTargetTileMap.values());
        this.tmTarget.writeToFile();
    }

    @Override
    public void close() {
        for (Entry<Integer,Tile> tileEntry : this.tmTargetTileMap.entrySet()) {
            tileEntry.getValue().getOutStreamDisk().close();
        }
    }

    /**
     * Vergleicht, ob zwei Pfade oder Dateien gleich sind.
     * Grund: Die Methode getPath() oder getAbsolutePath erkennen
     * nicht die Gleichheit von z.B. "./path" und "path".
     * @param path1 Pfad
     * @param path2 Pfad
     * @return true, wenn gleich.
     */
    private boolean pathsEqual(File path1, File path2) {
        try {
            String p1 = path1.getCanonicalPath();
            String p2 = path2.getCanonicalPath();
            return p1.equals(p2);
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * Untersucht die uebergebenen Verzeichnisse
     * auf Kompatibilitaet und fuellt die interne
     * {@link TileManager}-Liste der zu mergenden Gebiete.
     * @return Parameter verarbeitbar, ja/nein.
     */
    private boolean setupSources(List<File> dirList) {
        for (File dir : dirList) {
            if (!this.pathsEqual(this.workDir, dir)) {
                File tmFile = new File(dir, TileManager.FILENAME);
                if (tmFile.exists()) {
                    log.info("Reading infos from " + tmFile.getPath());
                    TileManager tm = TileManager.createFromFile(tmFile);
                    tm.setWorkDir(dir); // Transient
                    this.tmSourceList.add(tm);
                }
            }
        }
        if (this.tmSourceList.size() < 2) {
            log.warn("Nothing to merge here");
            return false;
        }
        TileManager tm0 = this.tmSourceList.get(0);
        for (int i = 1; i < this.tmSourceList.size(); i++) {
            TileManager tm = this.tmSourceList.get(i);
            if (tm0.getTileSizeX() != tm.getTileSizeX()
                    || tm0.getTileSizeY() != tm.getTileSizeY()
                    || tm0.isSingleTileMode() != tm.isSingleTileMode()) {
                log.error("TileSizes must be equal");
                return false;
            }
            this.nSourceWays += tm.getNumberOfWays();
        }
        return true;
    }

    private void setupTarget() {
        TileManager tmSource0 = this.tmSourceList.get(0);
        this.tmTarget = new TileManager(this.config);
        this.tmTarget.setSingleTileMode(tmSource0.isSingleTileMode());
        this.tmTarget.setTileSizeX(tmSource0.getTileSizeX());
        this.tmTarget.setTileSizeY(tmSource0.getTileSizeY());

        this.workDir.mkdirs();

        for (TileManager tmSource : this.tmSourceList) {
            for (int i = 0; i < tmSource.getNumberOfTiles(); i++) {
                Tile tileSource = tmSource.getTile(i);
                int tileId = tileSource.getId();
                if (!this.tmTargetTileMap.containsKey(tileId)) {
                    Tile tileTarget = new Tile(tileId);
                    this.tmTargetTileMap.put(tileId, tileTarget);
                    File fileTarget = new File(this.workDir,
                            TileManager.TILED_WAYS_FILENAME(tileId));
                    OutStreamDisk outStreamTarget = new OutStreamDisk(fileTarget);
                    outStreamTarget.writeByte(VarTypeDesk.typeIdOf(Way.class));
                    tileTarget.setOutStreamDisk(outStreamTarget);
                }
            }
        }
    }

    private void merge() {
        Way way = new Way(); // Holder
        LongArray targetWayIdSet = new LongArray();

        File sharedFileTarget = new File(this.workDir, Joiner.WAY_FILENAME_SHARED);
        OutStreamDisk sharedOutStream = new OutStreamDisk(sharedFileTarget);
        sharedOutStream.writeByte(VarTypeDesk.typeIdOf(Way.class));

        for (TileManager tmSource : this.tmSourceList) {
            log.info("Merging " + tmSource.getPrefix() + " into " + tmTarget.getPrefix());
            int nSharedWays = 0, nIgnoredSharedWays = 0;
            int nSharedWaysNodes = 0;

            log.debug(Tile.SHARED_TILE_ID);

            File sharedFileSource = new File(tmSource.getWorkDir(), Joiner.WAY_FILENAME_SHARED);
            InStreamDisk sharedInStream = new InStreamDisk(sharedFileSource);
            if (sharedInStream.readByte() != VarTypeDesk.typeIdOf(Way.class))
                throw new RuntimeException("No WayFile : " + sharedFileSource);

            LongArray sharedSourceWayIdSet = new LongArray();
            while (!sharedInStream.isEof()) {
                way.readFromStream(sharedInStream);
                long wayId = way.getId();
                if (targetWayIdSet.searchAny(wayId) == LongArray.NOTFOUND) {
                    nSharedWays++;
                    nSharedWaysNodes += way.getNodes().length;
                    way.writeToStream(sharedOutStream);
                    sharedSourceWayIdSet.push(wayId);
                } else {
                    nIgnoredSharedWays++;
                }
            }
            tmTarget.setNumberOfSharedWays(
                    tmTarget.getNumberOfSharedWays() + nSharedWays);
            tmTarget.setNumberOfSharedWaysNodes(
                    tmTarget.getNumberOfSharedWaysNodes() + nSharedWaysNodes);

            sharedInStream.close();

            log.debug(DF(nIgnoredSharedWays) + " of " + DF(nSharedWays) + " duplicate ways removed");

            for (int i = 0; i < tmSource.getNumberOfTiles(); i++) {
                int nWays = 0, nIgnoredWays = 0;

                Tile tileSource = tmSource.getTile(i);
                int tileId = tileSource.getId();
                log.debug("Tile " + Tile.getTileName(tileId));
                Tile tileTarget = this.tmTargetTileMap.get(tileId);

                File fileSource = new File(tmSource.getWorkDir(),
                        TileManager.TILED_WAYS_FILENAME(tileId));
                InStreamDisk inStream = new InStreamDisk(fileSource);
                if (inStream.readByte() != VarTypeDesk.typeIdOf(Way.class))
                    throw new RuntimeException("No WayFile : " + fileSource);
                OutStreamDisk outStream = tileTarget.getOutStreamDisk();

                LongArray sourceWayIdSet = new LongArray();
                while (!inStream.isEof()) {
                    way.readFromStream(inStream);
                    long wayId = way.getId();
                    if (targetWayIdSet.searchAny(wayId) == LongArray.NOTFOUND) {
                        nWays++;
                        way.writeToStream(outStream);
                        sourceWayIdSet.push(wayId);

                        // TileCounter fuer Segmenter anpassen
                        tileTarget.incTiledWaysCounter();

                        Node[] wayNodes = way.getNodes();
                        int nWayNodes = wayNodes.length;
                        tileTarget.incTiledWaysNodesCounter(nWayNodes);
                        for (int j = 0; j < nWayNodes; j++)
                            tileTarget.incNodeCounter(wayNodes[j].isLatLonOnly());

                    } else {
                        nIgnoredWays++;
                    }
                }

                targetWayIdSet.push(sourceWayIdSet);
                sourceWayIdSet.clear();
                targetWayIdSet.sort();
                inStream.close();

                log.debug(DF(nIgnoredWays) + " of " + DF(nWays) + " duplicate ways removed");
            }
        }
        sharedOutStream.close();
    }

}
