/*
 * $Id: Segmenter.java 5795 2023-04-20 10:02:40Z carsten $
 *
 * Copyright 2011/2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_MASK;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION;
import static de.cm.osm2po.converter.tiler.Tile.SHARED_TILE_ID;
import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_FROM;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_TO;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_VIA;

import java.io.File;
import java.util.Arrays;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.segmenter.VertexHashMap;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.Membership;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.model.Restriction;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.model.WaySegment;
import de.cm.osm2po.postp.PostProcessor;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.IntArray;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.LongIntBag;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarTypeDesk;
/**
 * Diese Klasse ermittelt aus den vom {@link Joiner} generierten Way-Tile-Files
 * die eigentliche Topologie.<br>
 * Intern werden zuerst alle SharedWays gelesen und hieraus saemtliche Knoten
 * gecacht und bei Mehrfachauftreten der VertexCounter erhoeht. Dieser Cache
 * bleibt ueber den ganzen Verarbeitungsprozess erhalten. Danach werden
 * nacheinander die einzelnen Way-Tile-Files eingelesen, auch hier die Knoten
 * gezaehlt, Kreuzungen erkannt, segmentiert und verarbeitet. Dabei werden aber
 * auch die bereits gecachten SharedWays-Knoten mit einbezogen und ggf. diese
 * aktualisiert. Intern wird also im Normalfall immer auf zwei Maps gearbeitet.
 * Sind alle Tiles durchlaufen, so werden abschliessend auch die SharedWays
 * segmentiert und geschrieben. Abschliessend werden dann noch die Vertices mit
 * samt den aufgeloesten Abbiegevorschriften persistiert. Das Resultat besteht
 * folglich aus zwei Dateien. Die eine beinhaltet segmentierte Wege, wobei hier
 * die Ids der Restriktionen noch nicht aufgeloest sind und noch auf die
 * OSM-Ids verweisen, die zweite beinhaltet nur die Vertices, wobei allerdings
 * hier die Restriktionen gegen die echten WaySegment-Ids ausgetauscht wurden.
 * <br><br><b>Wichtig:</b><br>
 * Aufgrund bestimmter Konstellationen, z.B. bei gleichzeitiger Uebersetzung
 * von Boundaries und Strassen-Netzwerk, werden hier nur Wege verarbeitet,
 * die eine Klasse &gt; 0 besitzen oder Mitglied einer Route sind.<br>
 * Ferner werden hier die folgenden Knoten-Eigenschaften umgesetzt:
 * <ul>
 * <li>{@link OsmNode#isBarrier()} setzt Bit 8 der {@link Vertex#getClazz()} </li>
 * <li>{@link OsmNode#isLevelCrossing()} entkoppelt den Knoten wenn {@link OsmWay#isRail()}</li>
 * </ul>
 * Die Entkopplung bewirkt die Korrektur eines sehr ekligen OSM-Fehlers vieler Tagger,
 * die Strasse und Schiene ueber einen gemeinsamen Knoten verbinden. Ist der Knoten ein
 * Bahnuebergang, so kann dies hier korrigiert werden. Config:<br>
 * <pre>
 * decoupleMode = off | simple | strict
 * </pre>
 * Der <code>simple</code>-Mode funktioniert nur bei Kreuzungen, an denen OSM-Wege nicht anfangen
 * oder enden, sondern nur durchfuehren. Das beseitigt viele aber nicht alle Fehler. Intern werden
 * hier lediglich keine Vertices vergeben.
 * Im <code>strict</code>-Mode werden intern die obersten 2 Bits der OSM-Id missbraucht, um
 * auch Start- und Endpunkte entkoppeln zu koennen. Hierfuer muss natuerlich sichergestellt sein,
 * dass Ids nur die unteren 62 Bits belegen. Negative Ids werden dennoch korrekt umgesetzt.
 * Vertices werden hier nicht verworfen, sondern entsprechen denen im Mode <code>off</code>.
 * Default ist <code>simple</code>
 *
 * @see GraphBuilder
 * @see PostProcessor
 * @see Joiner
 * @see Tiler
 * @see SegmentedWay
 * @see WaySegment
 *
 * @author (c) 2012/2020 - Carsten Moeller - info@osm2po
 */
public class Segmenter implements Processor {
    private static long DECOUPLE_XOR = 0x3L<<62; // Two high bits of long
    private static int DECOUPLE_MODE_OFF = 0;
    private static int DECOUPLE_MODE_SIMPLE = 1;
    private static int DECOUPLE_MODE_STRICT = 2;

    private static final int VIAS_CRITICAL_LEN = 16;

    public static final String SEGMENTS_FILENAME = "sw_all.2po";
    public static final String VERTICES_FILENAME = "sv_all.2po";

    private Log log;
    private File workDir;
    private TileManager tileManager;

    private boolean useQuadTileOrder;
    private boolean keepWayMemberships;
    private int decoupleMode;

    private VertexHashMap sharedVertexMap;
    private VertexHashMap tiledVertexMap;

    private WaySegment[] segmentsBuffer = new WaySegment[1000];
    /** Knoten-IDs werden fuer die Segmentierung neu vergeben. */
    private int nextVertexId;
    /** Way-IDs werden fuer die Segmentierung neu vergeben. */
    private int nextSegmentId;

    @Override
    public void init(Config config) {
        this.log = config.getLog();
        this.workDir = config.getWorkDir();

        this.tileManager = new TileManager(config).readFromFile();

        this.useQuadTileOrder = Boolean.valueOf(config.getProperty(
                "useQuadTileOrder", "false"));
        this.keepWayMemberships = Boolean.valueOf(config.getProperty(
                "keepWayMemberships", "false"));

        switch (config.getProperty("levelCrossingFix", "simple").toLowerCase()) {
        case "simple": this.decoupleMode = DECOUPLE_MODE_SIMPLE; break;
        case "strict": this.decoupleMode = DECOUPLE_MODE_STRICT; break;
        default: this.decoupleMode = DECOUPLE_MODE_OFF; }
    }

    @Override
    public void close() {
        this.sharedVertexMap.clear();
        this.sharedVertexMap = null;
        this.tiledVertexMap = null;
        this.segmentsBuffer = null;
    }

    /**
     * Startet den Build-Lauf. Beschreibung siehe {@link Segmenter}
     */
    @Override
    public void run() {

        File segmentsOutFile = new File(this.workDir, SEGMENTS_FILENAME);
        OutStream segmentsOutStream = new OutStreamDisk(segmentsOutFile);
        segmentsOutStream.writeByte(VarTypeDesk.typeIdOf(SegmentedWay.class));
        File verticesOutFile = new File(this.workDir, VERTICES_FILENAME);
        OutStream verticesOutStream = new OutStreamDisk(verticesOutFile);
        verticesOutStream.writeByte(VarTypeDesk.typeIdOf(Vertex.class));
        long nSegments; // Segmente mitzaehlen.

        this.assignVertexIds(null); // null = Shared Ways Vertices cachen

        for (int i = 0; i < this.tileManager.getNumberOfTiles(); i++) {
            Tile tile = this.tileManager.getTile(i);
            this.assignVertexIds(tile); // Vertices aus tile cachen
            nSegments = this.segmentizeAndWrite(tile, segmentsOutStream);
            tile.setNumberOfSegments(nSegments);
            this.writeVerticesToStream(this.tiledVertexMap, verticesOutStream);
            this.tiledVertexMap.clear();
        }

        nSegments = this.segmentizeAndWrite(null, segmentsOutStream);
        this.tileManager.setNumberOfSharedSegments(nSegments);
        this.writeVerticesToStream(this.sharedVertexMap, verticesOutStream);
        this.sharedVertexMap.clear();

        segmentsOutStream.close();
        verticesOutStream.close();

        this.tileManager.writeToFile();

        if (this.useQuadTileOrder) {
            this.orderByQuadTile(verticesOutFile, segmentsOutFile);
        }
    }

    /**
     * Prueft, ob ein Weg fuer das Routing relevant ist und
     * daher segmentiert und geschrieben werden soll.<br>
     * Dazu muss es sich entweder um einen validen bzw. deklarierten
     * Weg oder um einen Weg, der Mitglied einer Route-Relation ist, handeln.
     * Wege, die hier aufgrund anderer Relationen durchkommen,
     * so z.B. Boundaries, werden als <code>false</code> bewertet
     * und im Nachgang verworfen! Wege die fuer {@link OsmWay#isPolygon()}
     * true liefern, so z.B. Buildings, werden ebenfalls ignoriert.
     * @param way {@link Way}
     * @return boolean verarbeiten true/false
     * @see OsmWay#isInvalid()
     * @see OsmWay#isPolygon()
     * @see OsmWay#isRouteMember()
     */
    private boolean postValidate(Way way) {
        // Viel if-then-else hier, aber gut zu debuggen
        if (way.isPolygon()) {
            return false;
        }
        if (way.isInvalid()) {
            if (!way.isRouteMember()) {
                return false;
            }
        }
        return true;
    }


    private void xorNodeId(Node node) {
        long osmId = node.getId();
        osmId ^= DECOUPLE_XOR;
        node.setId(osmId);
    }
    private boolean decoupleNodeId(Way way, Node node) {
        if (DECOUPLE_MODE_STRICT == this.decoupleMode) {
            if (way.isRail() && node.isLevelCrossing()) {
                this.xorNodeId(node);
                return true;
            }
        }
        return false;
    }

    /**
     * Liest die vom {@link Joiner} erzeugten Ways-Dateien und registriert bzw.
     * cached die {@link Node}-Objekte als {@link Vertex}-Objekte.
     * Bei Mehrfachvorkommen wird der interne Knotenzaehler um eins erhoeht.
     * Knoten, bei denen dieser Counter groesser 0 ist, sind Kreuzungen
     * (Vertices). Auch Start- und Endpunkte des Original-Wegs sind Vertices.
     * Ist eine dieser Bedingungen erfuellt, wird eine Vertex-Id vergeben,
     * aufgrund welcher im Folgeschritt
     * ({@link #segmentizeAndWrite(Tile, OutStream)}) segmentiert wird.
     * Zusaetzlich koennen noch andere Bedingungen dazu fuehren, dass eine
     * Vertext-ID vergeben wird, so z.B. aufgrund einer speziellen Klasse
     * (NodeClazz != 0), Beispiel Ampel. <br><br>
     * Zusaetzlich wird hier geprueft, ob Knoten bereits im Shared-Bereich
     * (Tile-uebergreifend) verarbeitet wurden. In diesem Fall wird der Knoten
     * hier aktualisiert. Und wenn fuer {@link Tile} <code>null</code> uebgergeben
     * wird, dann wird nur auf der Shared-Datei/Bereich gearbeitet.
     *
     * @param tile {@link Tile} oder <code>null</code>, wenn shared.
     */
    private void assignVertexIds(Tile tile) {
        System.gc();
        String tileIdName;
        long nNodes, n;
        File waysInFile;

        if (null == tile) {
            tileIdName = SHARED_TILE_ID;
            nNodes = this.tileManager.getNumberOfSharedWaysNodes();
            waysInFile = new File(this.workDir, Joiner.WAY_FILENAME_SHARED);
            // shared map etwas schneller machen, da immer zuerst gefragt
            this.sharedVertexMap = new VertexHashMap((int) (nNodes * 1.5));
        } else {
            tileIdName = Tile.getTileName(tile.getId());
            nNodes = tile.getNumberOfWaysNodes();
            waysInFile = new File(this.workDir, TileManager.TILED_WAYS_FILENAME(tile.getId()));
            this.tiledVertexMap = new VertexHashMap((int) (nNodes));
        }

        InStream waysInStream = new InStreamDisk(waysInFile);
        byte varType = waysInStream.readByte();
        if (varType != VarTypeDesk.typeIdOf(Way.class))
            throw new RuntimeException("Unexpected VarType " + varType);

        n = 0;
        Way way = new Way();
        while (!waysInStream.isEof()) {
            way.readFromStream(waysInStream); // clear() bereits ausgefuehrt!
            if (!postValidate(way)) continue;

            Node[] nodes = way.getNodes();
            int nl = nodes.length;

            // Iteriere ueber alle Knoten des Weges
            for (int i = 0; i < nl; i++) {
                Node node = nodes[i];
                long osmNodeId = node.getId();
                byte osmNodeClazz = node.getClazz();

                boolean isDecoupledNodeId = this.decoupleNodeId(way, node);

                // Sonderlocke fuer graph.createCarBarriersAsNoTurnBits()
                if (node.isBarrier()) { // TODO konfigurierbar machen
                    osmNodeClazz |= 0x80; // Boeser Workaround - funzt auch bei 0, dann -128
                }

                // Ist dieser Knoten bereits in der SharedMap?
                VertexHashMap targetMap = this.sharedVertexMap;
                Vertex vertex = targetMap.get(node.getId());
                if (tile != null && null == vertex) { // nein? dann TileMap, wenn TileMode
                    targetMap = this.tiledVertexMap;
                    vertex = targetMap.get(node.getId());
                }

                if (null == vertex) { // Neuer Knoten / Vertex
                    Membership[] memberships = node.getMemberships();

                    boolean isVertex = (i == 0 || i == nl - 1) // Start oder Ziel
                            || node.isBarrier()
                            || !node.isLatLonOnly(); // z.B. clazz = Ampel gesetzt

                    // TODO VertexId vergeben, wenn Relationen vorhanden?
                    // War frueher. Unnoetig? Analyse: hier nur ONLY_STRAIGHT_ONs.
                    // Achtung: isLatLonOnly() s.o. deckt Relations nicht ab!
                    //isVertex |= memberships != null;

                    try {
                        // VertexId vergeben, wenn noetig.
                        if (isVertex) {
                            // Ohne OsmNodeId in den Vertex uebertragen (dort Kompression).
                            // Landet als Key in Map und wird spaeter rekonstruiert.
                            vertex = new Vertex();
                            vertex.setClazz(osmNodeClazz);
                            vertex.setCoord(node.getCoord());
                            vertex.setOsmName(node.getName());
                            vertex.setCustom(node.getCustom());

                            int vid = ++this.nextVertexId;
                            vertex.setId(isDecoupledNodeId ? -vid : vid); // Tricky. Set negative Flag if decoupled

                            Restriction[] vias = this.filterVias(memberships);
                            if (vias != null) {
                                vertex.setRestrictions(vias);
                                if (vias.length >= VIAS_CRITICAL_LEN) {
                                    log.warn("Node #" + osmNodeId + " (Vertex #"
                                        + this.nextVertexId + ") referenced by "
                                        + vias.length + " Restrictions");
                                }
                            }
                        }

                        // Tricky: Wenn vertex==null, dann werden lediglich
                        // die NodeIds im Hash gespeichert,  bei get() als leere
                        // Huelle zurueckgeliefert und bei erneutem put() ohne
                        // Speicherverlust ggf. ueberschrieben.
                        targetMap.put(node.getId(), vertex);

                        if (++n % 1000000 == 0) log.progress(DF(n) + " of " + DF(nNodes)
                                + " WayNodes cached (" + tileIdName + ") - " + MB());

                    } catch (OutOfMemoryError ome) {
                        log.fatal("Failed to cache Node " + DF(n)
                                + " of " + DF(nNodes) + " - " + MB());
                        throw ome;
                    }

                } else { // Kreuzung oder Abzweiger

                    vertex.incRefCounter();
                    // Wenn noch nicht als Vertex gecacht,
                    // z.B weil kein Start- oder End-Knoten.
                    if (0 == vertex.getId()) {
                        int vid = ++this.nextVertexId;
                        vertex.setId(isDecoupledNodeId ? -vid : vid);

                        vertex.setClazz(osmNodeClazz);
                        vertex.setCoord(node.getCoord());
                        vertex.setOsmName(node.getName());
                        vertex.setCustom(node.getCustom());
                        // Restriction-Batrachtung hier nicht noetig, da fuer
                        // OSM nur fuer Start- oder EndKnoten moeglich / sinnvoll.
                        // Vertex in diesem Fall bereits oben gefuellt
                        targetMap.put(osmNodeId, vertex);
                    } else {
                        targetMap.override(vertex);
                    }
                }
            }
        }
        waysInStream.close();


        log.info(DF(n) + " of " + DF(nNodes) + " WayNodes cached ("
                + tileIdName + ") - " + MB());
        VertexHashMap map = null == tile ? this.sharedVertexMap : this.tiledVertexMap;
        log.debug("Hash Collisions Max : "
                + DF(map.getCollisionsMax()) + " Avg : " + DF2(map.getCollisionsAvg()));
    }


    /**
     * Liest eine Shared- oder Tiled-Way-Datei erneut und veranlasst die
     * Segmentierung an echten Vertices, die aufgrund von
     * {@link #assignVertexIds(Tile)} ermittelt wurden.
     * Abschliessend werden diese segmentierten Wege herausgeschrieben.
     *
     * @param tile {@link Tile} oder null, wenn shared.
     * @param outStream {@link OutStream} Ausgabestrom fuer Ergebnis.
     * @return long Anzahl ermittelter Segmente.
     */
    private long segmentizeAndWrite(Tile tile, OutStream outStream) {
        File waysInFile;
        long nWays = 0, nSegments = 0;
        String tileName;

        if (null == tile) { // Shared
            waysInFile = new File(this.workDir, Joiner.WAY_FILENAME_SHARED);
            nWays = this.tileManager.getNumberOfSharedWays();
            tileName  = SHARED_TILE_ID;
        } else { // Tiled
            waysInFile = new File(this.workDir,
                    TileManager.TILED_WAYS_FILENAME(tile.getId()));
            nWays = tile.getNumberOfWays();
            tileName = Tile.getTileName(tile.getId());
        }

        SegmentedWay wayHolder = new SegmentedWay();
        InStream inStream = new InStreamDisk(waysInFile);
        byte varType = inStream.readByte();
        if (varType != VarTypeDesk.typeIdOf(Way.class))
            throw new RuntimeException("Unexpected VarType " + varType);

        long n = 0;
        while (!inStream.isEof()) {
            wayHolder.readFromWayStream(inStream);
            if (!postValidate(wayHolder)) continue;

            WaySegment[] segments = this.segmentize(wayHolder);
            wayHolder.setSegments(segments);
            if (wayHolder.getMemberships() != null && !this.keepWayMemberships) {
                wayHolder.clearMemberships(); // Dateigroesse reduzieren
            }
            wayHolder.writeToStream(outStream);

            nSegments += segments.length;
            if (++n % 100000 == 0) log.progress(DF(n) + " of "
                    + DF(nWays) + " ways segmented (" + tileName + ") - " + MB());
        }
        inStream.close();
        log.info(DF(nWays) + " ways analyzed, " + DF(nSegments)
                + " segments created (" + tileName + ") - " + MB());

        return nSegments;
    }

    /**
     * Ruft {@link VertexHashMap#get(long)} fuer die aktuell gecachten Vertices
     * aus dem Tile oder, wenn nicht gefunden, aus der shared map. Setzt dabei
     * im Vertex das transiente Attribut {@link Vertex#setShared(boolean)}, um
     * beim Zurueckschreiben entscheiden zu koennen, aus welcher der beiden
     * Maps dieser Vertex urspruenglich entnommen wurde.
     *
     * @param key long Schluessel des Vertex in der Map.
     * @return {@link Vertex} oder null, wenn nicht gefunden = Fehler.
     */
    private Vertex lookupVertex(long key) {
        Vertex vertex = this.tiledVertexMap.get(key);
        if (null == vertex) {
            vertex = this.sharedVertexMap.get(key);
            if (null == vertex) return null; // Fehler!
            vertex.setShared(true);
        }
        return vertex;
    }

    /**
     * Schreibt den Vertex nach der Entnahme durch {@link #lookupVertex(long)}
     * wieder zurueck in die richtige der beiden Maps. Realisiert wird dieser
     * Automatismus durch den Vertex selbst, der ein transientes Attribut
     * {@link Vertex#isShared()} besitzt, das beim Lookup gesetzt wird und in
     * der Folge die Entscheidung ermoeglicht, woher dieser urspruenglich stammt.
     * @param vertex {@link Vertex}
     */
    private void overrideVertex(Vertex vertex) {
        if (vertex.isShared()) {
            this.sharedVertexMap.override(vertex);
        } else {
            this.tiledVertexMap.override(vertex);
        }
    }

    private void moveNodeToVertexIfDifferent(Node node, Vertex vertex) {
        if (node.getCoord() != vertex.getCoord()) {
            log.warn("Coordinates of Node and Vertex do not match.\nMoving " + node
                    + " to (" + vertex.getLat() + " " + vertex.getLon() + ")");
            node.setCoord(vertex.getCoord());
        }
    }

    /**
     * Segmentiert eine Kette von Knoten eines Weges, wenn die
     * korrespondierenden Vertices eine ID ungleich 0 haben, also
     * tatsaechliche Start- oder Endpunkte sind.<br>
     * Obwohl designtechnisch unsauber, werden hier auch die
     * {@link Restriction} Objekte behandelt und dabei die {@link OsmWay}-Ids
     * gegen soeben erzeugte {@link WaySegment}-Ids im {@link Vertex}
     * ausgetauscht. Dieses oberflaechlich betrachtet maessige Design beruht
     * auf der Tatsache, dass hier in einer komplexen Schleife die
     * entscheidenden Daten aufeinandertreffen und eine solche Schleife nicht
     * unbedingt erneut wiederholt werden muss.
     *
     * @param way {@link Way}.
     * @return Array[] von {@link WaySegment} Objekten.
     */
    private WaySegment[] segmentize(Way way) {
        int nSegments = 0;
        int idxFrom = 0;
        Node[] nodes = way.getNodes();
        int idxMax = nodes.length -1;
        Node node = nodes[0];

        boolean isDecoupledNodeId = this.decoupleNodeId(way, node);
        Vertex sourceVertex = this.lookupVertex(node.getId());
        if (isDecoupledNodeId) { // und zurueck zum Original
            sourceVertex.setId(-sourceVertex.getId());
            this.xorNodeId(node);
        }

        Vertex targetVertex = null;
        WaySegment waySegment = null;

        this.moveNodeToVertexIfDifferent(node, sourceVertex);

        for (int idx = 1; idx < idxMax; idx++) {
            node = nodes[idx];

            isDecoupledNodeId = this.decoupleNodeId(way, node);
            targetVertex = this.lookupVertex(node.getId());
            if (isDecoupledNodeId) { // und zurueck zum Original
                targetVertex.setId(-targetVertex.getId());
                this.xorNodeId(node);
            }

            if (targetVertex.getId() != 0) { // Segmentieren!

                if (DECOUPLE_MODE_SIMPLE == this.decoupleMode && node.isLevelCrossing()) {
                    if (way.isRail()) continue; // Loop ensures no source or target
                }

                // Vertex mit Id 0 hat keine Koordinate!
                this.moveNodeToVertexIfDifferent(node, targetVertex);

                int segmentId = ++this.nextSegmentId;
                waySegment = new WaySegment(segmentId,
                        sourceVertex.getId(), targetVertex.getId(), idxFrom, idx);
                this.updateRestrictions(sourceVertex, way, segmentId);
                this.updateRestrictions(targetVertex, way, segmentId);
                this.segmentsBuffer[nSegments++] = waySegment;

                idxFrom = idx;
                sourceVertex = targetVertex;

                // So viele Segmente kann ein Weg eigentlich nicht haben.
                // Aber sicher ist sicher
                if (nSegments > this.segmentsBuffer.length - 10) {
                	this.segmentsBuffer = Arrays.copyOf(
                			this.segmentsBuffer, this.segmentsBuffer.length * 2);
                	log.warn("Increasing SegmentsBuffer to "
                	        + DF(this.segmentsBuffer.length));
                	log.debug("Way: " + way);
                }
            }
        }

        // Letztes WaySegment immer schreiben
        int segmentId = ++this.nextSegmentId;
        node = nodes[idxMax];
        isDecoupledNodeId = this.decoupleNodeId(way, node);
        targetVertex = this.lookupVertex(node.getId());
        if (isDecoupledNodeId) {  // und zurueck zum Original
            targetVertex.setId(-targetVertex.getId());
            this.xorNodeId(node);
        }

        this.moveNodeToVertexIfDifferent(node, targetVertex);

        waySegment = new WaySegment(segmentId,
                sourceVertex.getId(), targetVertex.getId(), idxFrom, idxMax);
        this.updateRestrictions(sourceVertex, way, segmentId);
        this.updateRestrictions(targetVertex, way, segmentId);
        this.segmentsBuffer[nSegments++] = waySegment;

        return Arrays.copyOf(this.segmentsBuffer, nSegments);
    }


    /**
     * Findet alle {@link Membership}en der Rolle Via und liefert diese als
     * initiale {@link Restriction}s zurueck (ohne From und To).
     * @param memberships nullable Array von {@link Membership}
     * @return Array von {@link Restriction} (n > 0) oder null
     */
    private Restriction[] filterVias(Membership[] memberships) {
        if (null == memberships) return null;

        Restriction[] vrs = new Restriction[memberships.length];
        int n = 0;

        for(int i = 0; i < memberships.length; i++) {
            long id = memberships[i].getId(); // OsmId
            byte role = memberships[i].getRole();
            byte clazz = memberships[i].getClazz();

            // Pruefen auf ViaRestriction
            if (role != MEMBER_ROLE_VIA) continue;
            if (CLAZZ_RESTRICTION != (CLAZZ_MASK & clazz)) continue;
            vrs[n++] =  new Restriction(id, clazz, 0, 0); // OK. Ist eine.
        }

        return (0 == n) ? null : Arrays.copyOf(vrs, n);
    }


    /**
     * Setzt die echten Segment-Ids fuer die Way-Rollen from und To.
     * @param vertex {@link Vertex} Vertex mit Restriktionen
     * @param way {@link Way} gegen den die Rollen From und To aufgeloest werden.
     * @param segmentId int zu uebertragende id des Segments
     */
    private void updateRestrictions(Vertex vertex, Way way, int segmentId) {
        Restriction[] vertexRestrictions = vertex.getRestrictions();
        if (null == vertexRestrictions) return;
        Membership[] wayMemberships = way.getMemberships();
        if (null == wayMemberships) return;

        // TODO Schleife umdrehen und Source und Target gleichzeitig verarbeiten
        for (int i = 0; i < vertexRestrictions.length; i++) {
            Restriction restriction = vertexRestrictions[i];
            for (int j = 0; j < wayMemberships.length; j++) {
                long relId = restriction.getId();
                Membership wayRel = wayMemberships[j];
                if (relId == wayRel.getId()) {
                    byte role = wayRel.getRole();
                    if (MEMBER_ROLE_FROM == role) {
                        vertexRestrictions[i].setFrom(segmentId);
                        this.overrideVertex(vertex);
                    } else if (MEMBER_ROLE_TO == role) {
                        vertexRestrictions[i].setTo(segmentId);
                        this.overrideVertex(vertex);
                    }
                }
            }
        }
    }

    /**
     * Schreibt die Vertices aus der Map in einen Stream. Es werden nur echte
     * Vertices geschrieben, welche eine Id ungleich 0 besitzen. Ferner werden
     * hier auch keine Map-Keys, also die urspruenglichen OsmNodeIds
     * geschrieben, sondern lediglich die Vertex-Objekte selbst. Dabei werden
     * moegliche Modifikationen von OsmIds fuer das Decoupling wieder korrigiert.
     * @param map {@link VertexHashMap} von {@link Vertex}-Objekten.
     * @param outStream {@link OutStream} Ausgabe-Strom.
     */
    private void writeVerticesToStream(VertexHashMap map, OutStream outStream) {
        LongArray keys = map.getKeys();
        long nKeys = keys.getSize();
        long n = 0;
        for (long i = 0; i < nKeys; i++) {
            long osmId = keys.get(i);
            if (!map.isNull(osmId)) {
                Vertex vertex = map.get(osmId);
                int vertexId = vertex.getId();
                if (vertexId != 0) {
                    if (vertexId < 0) { // decoupled in strict mode
                        vertex.setId(-vertexId);
                        osmId ^= DECOUPLE_XOR;
                    }
                    vertex.setOsmId(osmId);
                    vertex.writeToStream(outStream);
                    if (++n % 500000 == 0) log.progress(DF(n) + " vertices of "
                            + DF(nKeys) + " nodes written - " + MB());
                } else {
                    log.error("Sth. went terribly wrong here");
                }
            }
        }
        log.info(DF(n) + " vertices of " + DF(nKeys) + " nodes written - " + MB());
    }


    /**
     * Renummeriert Vertices basierend auf ihrer geografischen Lage mittels
     * QuadTiling. <br>Hintergrund:<br>
     * Bis hierher wurden IDs analog der Anordnung und der
     * Verarbeitungsreihenfolge inkrementell vergeben. Fuer ein pures
     * InMemory-Routing ist dies ausreichend. Kommt allerdings eine
     * Festplatte oder eine SD-Card ins Spiel, so ist es besser, die Vertices
     * geografisch zu sortieren, damit spaeter eine hoehere Performance durch
     * ReadAhead-Buffering erreicht werden kann.
     *
     * @param vFile {@link File} Vertices-Datei
     * @param wFile {@link File} Weg-Segmente-Datei
     */
    private void orderByQuadTile(File vFile, File wFile) {
        InStream inStream;
        OutStream outStream;
        File tmpFile;
        int idx, newId;

        Vertex vertex = new Vertex();
        SegmentedWay way = new SegmentedWay();

        LongIntBag vertexBag = new LongIntBag();

        // Vertex-Datei 1x lesen um IDs und Koordinaten zu holen.
        log.info("Calculating QuadTiles for Vertex-Coordinates");
        inStream = new InStreamDisk(vFile);
        inStream.readByte(); // Skip VarType
        while (!inStream.isEof()) {
            vertex.readFromStream(inStream);
            long q = GeoUtils.calcQuadTile(vertex.getCoord());
            vertexBag.add(q, vertex.getId());
        }
        inStream.close();

        // Jetzt Vertices nach Koordinaten/QuadTiles sortieren
        // Und danach analog der Sortierung neu vergeben.
        log.info("Renumbering Vertices by QuadTileOrder");
        vertexBag.sort();
        IntArray oldIds = vertexBag.getValues(); // Sortiert nach QuadTile
        long n = vertexBag.getSize();
        int[] newIds = new int[(int) n]; // TODO Kritisch bei 2Gig
        for (int i = 0; i < n; i++) {
            newIds[oldIds.get(i) -1] = i +1;
        }
        vertexBag.clear();

        // Vertex-Datei kopieren und dabei IDs umschluesseln
        tmpFile = Utils.createTmpFile(vFile.getParentFile());
        outStream = new OutStreamDisk(tmpFile);
        inStream = new InStreamDisk(vFile);
        outStream.writeByte(inStream.readByte()); //VarType
        while (!inStream.isEof()) {
            vertex.readFromStream(inStream);
            idx = vertex.getId() -1;
            newId = newIds[idx];
            vertex.setId(newId);
            vertex.writeToStream(outStream);
        }
        inStream.close();
        outStream.close();
        vFile.delete();
        tmpFile.renameTo(vFile);

        // Way-Datei kopieren und dabei VertexIDs umschluesseln
        log.info("Modifying Vertices in WaySegments");
        tmpFile = Utils.createTmpFile(vFile.getParentFile());
        outStream = new OutStreamDisk(tmpFile);
        inStream = new InStreamDisk(wFile);
        outStream.writeByte(inStream.readByte()); //VarType
        while (!inStream.isEof()) {
            way.readFromStream(inStream);
            WaySegment[] segs = way.getSegments();
            for (int i = 0; i < segs.length; i++) {
                idx = segs[i].getSourceId() -1;
                newId = newIds[idx];
                segs[i].setSourceId(newId);
                idx = segs[i].getTargetId() -1;
                newId = newIds[idx];
                segs[i].setTargetId(newId);
            }
            way.writeToStream(outStream);
        }
        inStream.close();
        outStream.close();
        wFile.delete();
        tmpFile.renameTo(wFile);

        log.info("QuadTiling finished");
    }

}