/*
 * $Id: Tiler.java 5288 2021-05-07 19:04:34Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter;

import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_MASK;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION;
import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.Utils.createTmpFile;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_FROM;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_TO;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_VIA;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_NODE;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_WAY;

import java.io.File;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.parser.OsmParser;
import de.cm.osm2po.converter.parser.OsmParserHandler;
import de.cm.osm2po.converter.tiler.NodeTagResolver;
import de.cm.osm2po.converter.tiler.RelationTagResolver;
import de.cm.osm2po.converter.tiler.TagResolver;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.converter.tiler.WayTagResolver;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.OsmRelation;
import de.cm.osm2po.model.OsmRelation.Member;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Hauptklasse {@link OsmParserHandler} zum Verarbeiten einer OSM-Datei.<br>
 * Wird von einem {@link OsmParser} als Callback verwendet und so ueber den
 * Inhalt der Quelldatei informiert. Ein Parser kann aber ebenso aus anderen
 * Quellen lesen, so dass Inhalte auch aus Datenbanken stammen koennen.
 * Der Tiler sorgt nun dafuer, dass Nodes, Ways und Relations ueber angeschlossene
 * {@link TagResolver} vorgefiltert und zusaetzlich die Knoten mittels
 * {@link TileManager} in Quadranten einsortiert werden, was fuer sehr grosse
 * Datenmengen den Speicherverbrauch reduziert.<br>
 * Folgende Optionen stehen zusaetzlich zur Verfuegung - Defaults in Klammern:
 * <ul>
 * <li><code>tileSize</code>: Groesse eines Tiles (10x10,1)
 *     <code>YxX,B</code>, wobei Y und X Integer in Grad und B der
 *     Ueberlappungs-Puffer als Float sind. Dieser wird benoetigt, um
 *     Tile-uebergreifende Wege aufloesen zu koennen.<br>
 *     <code>tileSize=x</code> deaktiviert das Tiling. Dies ist optimal,
 *     jedoch direkt vom verfuegbaren RAM abhaengig.
 * </li>
 * <li><code>nodeIdOffset</code>: (OSM-)NodeIds um diesen Wert erhoehen (0)</li>
 * <li><code>wayIdOffset</code>: (OSM-)WayIds um diesen Wert erhoehen (0)</li>
 * <li><code>relationIdOffset</code>: (OSM-)RelationIds um diesen Wert erhoehen (0)</li>
 * </ul>
 * Zudem werden nach jedem Durchlauf die hoechsten IDs fuer Knoten, Wege und
 * Relationen als <code>osm2po.maxNodeId</code>, <code>osm2po.maxWayId</code>
 * und <code>osm2po.maxRelationId</code> in die System-Properties geschrieben,
 * was in einer Multi-Netz-Konfiguration mittels
 * <code>nodeIdOffset=[maxNodeId]</code>,
 * <code>wayIdOffset=[maxWayId]</code> und
 * <code>relationIdOffset=[maxRelationId]</code> an den naechsten
 * Verarbeitungsschritt weitergereicht werden kann. Tricky:<br>
 * Initialisiert kann der erste Schritt ueber normale Properties, da diese
 * in der Aufloesungskette der {@link Props} Vorrang haben.
 * @see Joiner
 * @see Segmenter
 * @see OsmParserHandler
 * @see Compressor
 * @author (c) 2012/2017 - Carsten Moeller - info@osm2po
 */
public class Tiler implements OsmParserHandler {
    private final static String SYSPROP_NODE_ID_MAX = "osm2po.maxNodeId";
    private final static String SYSPROP_WAY_ID_MAX = "osm2po.maxWayId";
    private final static String SYSPROP_REL_ID_MAX = "osm2po.maxRelationId";
    private final static String PROP_NODE_ID_OFFSET = "nodeIdOffset";
    private final static String PROP_WAY_ID_OFFSET = "wayIdOffset";
    private final static String PROP_REL_ID_OFFSET = "relationIdOffset";

    public static final String RAW_WAYS_FILENAME = "tw_raw.2po";
    public static final String RAW_RELATIONS_FILENAME = "tr_raw.2po";
    // RAW_NODES will be a tempfile. See code below

    private static final int MAX_OME_ERRORS = 3;

    private File workDir;
    private Log log;
    private LongArray ndRefSet; // Als Set ein Array, da besser als Hash.

    private boolean preFilterDisabled;

    private long nodeIdOffset;
    private long wayIdOffset;
    private long relationIdOffset;

    private int nAddNdRefErrors;

    // Tiles ----------------------------------------------
    private TileManager tileManager;

    // Nodes ----------------------------------------------
    private OsmNode nodeHolder = new OsmNode();
    private long nNodesRead, nNodesWritten;
    private NodeTagResolver nodeTagResolver;
    private File nodeRawFile;
    private OutStreamDisk nodeOutStream;

    // Ways -----------------------------------------------
    private OsmWay wayHolder = new OsmWay();
    private WayTagResolver wayTagResolver;
    private long nWaysRead, nWaysWritten;
    private long nWaysInvalid, nOneWaysReverse, nOneWaysUnknown;
    private File wayRawFile;
    private OutStreamDisk wayOutStream;

    // Relations ------------------------------------------
    private OsmRelation relationHolder = new OsmRelation();
    private RelationTagResolver relationTagResolver;
    private long nRelationsRead, nRelationsWritten;
    private File relationRawFile;
    private OutStreamDisk relationOutStream;

    private boolean nodesProcessingOpen = false;
    private boolean waysProcessingOpen = false;
    private boolean relationsProcessingOpen = false;

    private long maxNodeId;
    private long maxWayId;
    private long maxRelationId;

    private boolean isOpen = false;

    /**
     * Konstruktor.
     * @param config {@link Config}
     */
    @Override
    public void open(Config config) {
        try {
            this.log = config.getLog();
            this.workDir = config.getWorkDir();
            this.nodeTagResolver = config.getNodeTagResolverClass().getDeclaredConstructor().newInstance();
            this.nodeTagResolver.open(config);
            this.wayTagResolver = config.getWayTagResolverClass().getDeclaredConstructor().newInstance();
            this.wayTagResolver.open(config);
            this.relationTagResolver = config.getRelationTagResolverClass().getDeclaredConstructor().newInstance();
            this.relationTagResolver.open(config);

            this.nodeIdOffset = Long.parseLong(config.getProperty(PROP_NODE_ID_OFFSET, "0"));
            this.wayIdOffset = Long.parseLong(config.getProperty(PROP_WAY_ID_OFFSET, "0"));
            this.relationIdOffset = Long.parseLong(config.getProperty(PROP_REL_ID_OFFSET, "0"));

            this.maxNodeId = Long.MIN_VALUE;
            this.maxWayId = Long.MIN_VALUE;
            this.maxRelationId = Long.MIN_VALUE;

            this.tileManager = new TileManager(config);

            this.preFilterDisabled = Boolean.valueOf(
                    config.getProperty("disablePreFilter", "false"));

            this.log.info(this.tileManager.getTileSizeInfoString()
                    + " - PreFilter is " + (this.preFilterDisabled ? "OFF" : "ON"));

            this.isOpen = true;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    // ------------ Node Compress Mode Specials -------

    private void shrinkNdRefSet() {
        if (null == this.ndRefSet) return;
        log.progress("Compressing NdRefSet (May take a while)");
        System.gc();
        String mbBefore = MB();
        long nBefore = this.ndRefSet.getSize();
        this.ndRefSet.shrinkToSet();
        long nAfter = this.ndRefSet.getSize();
        System.gc();
        log.debug("NdRefSet compressed from " + DF(nBefore) + " (" + mbBefore
                + ") to " + DF(nAfter) + " (" + MB() + ")");
    }

    private boolean addNdRef(long ndRef) {
        try {
            this.ndRefSet.push(ndRef);
            return true;
        } catch (OutOfMemoryError ome) {
            return false;
        }
    }

    private boolean registerNdRefAndRetryOnOme(long ndRef) {
        if (this.addNdRef(ndRef)) return true; // Speicher noch frei

        if (++this.nAddNdRefErrors > MAX_OME_ERRORS) return false;
        log.warn("Not enough memory! CleanUpRetry "
                + this.nAddNdRefErrors + " of " + MAX_OME_ERRORS);
        this.shrinkNdRefSet();
        return this.addNdRef(ndRef); // Retry (recursive)
    }

    // ------------ Start Tags -------------

    @Override
    public void onNode(long id, double lat, double lon) {
        id += this.nodeIdOffset;
        if (id > this.maxNodeId) this.maxNodeId = id;

        if (this.nNodesRead == 0) {
            this.nodesProcessingOpen = true;
            this.nodeRawFile = createTmpFile(this.workDir);
            this.nodeOutStream = new OutStreamDisk(this.nodeRawFile);
            this.nodeOutStream.writeByte(
                    VarTypeDesk.typeIdOf(OsmNode.class));
        }

        this.nodeHolder.clear();
        this.nodeHolder.setId(id);
        this.nodeHolder.setLatLon(lat, lon);

        this.nodeTagResolver.onTags(id);
        this.nodeTagResolver.onCoordinate(lat, lon);
    }

    @Override
    public void onWay(long id) {
        id += this.wayIdOffset;
        if (id > this.maxWayId) this.maxWayId = id;

        if (this.nWaysRead == 0) {
            this.waysProcessingOpen = true;
            this.wayRawFile = new File(this.workDir, RAW_WAYS_FILENAME);
            this.wayOutStream = new OutStreamDisk(this.wayRawFile);
            this.wayOutStream.writeByte(VarTypeDesk.typeIdOf(OsmWay.class));

            // Erst hier, da NodeTiling zu Ende und Speicher wieder frei
            if (!this.preFilterDisabled) this.ndRefSet = new LongArray();
        }

        this.wayHolder.clear();
        this.wayHolder.setId(id);

        this.wayTagResolver.onTags(id);
    }

    @Override
    public void onRelation(long id) {
        id += this.relationIdOffset;
        if (id > this.maxRelationId) this.maxRelationId = id;

        if (this.nRelationsRead == 0) {
            this.relationsProcessingOpen = true;
            this.relationRawFile = new File(this.workDir, RAW_RELATIONS_FILENAME);
            this.relationOutStream = new OutStreamDisk(this.relationRawFile);
            this.relationOutStream.writeByte(
                    VarTypeDesk.typeIdOf(OsmRelation.class));
        }

        this.relationHolder.clear();
        this.relationHolder.setId(id);

        this.relationTagResolver.onTags(id);
    }

    // ------------ Complete Tags -------------

    @Override
    public void onNodeComplete() {
        this.nNodesRead++;
        this.nodeTagResolver.onTagsComplete();
        if (this.nodeTagResolver.isValid()) {

            // node.clear() wird vorher ausgefuehrt
            String nodeName = this.nodeTagResolver.getName();
            if (nodeName != null) this.nodeHolder.setName(new VarString(nodeName));
            this.nodeHolder.setClazz(this.nodeTagResolver.getClazz());
            this.nodeHolder.setFlags(this.nodeTagResolver.getFlags());
            this.nodeHolder.setCustom(this.nodeTagResolver.getCustom());
            this.nodeHolder.setLevelCrossing(this.nodeTagResolver.isLevelCrossing());
            this.nodeHolder.setBarrier(this.nodeTagResolver.isBarrier());

            if (this.preFilterDisabled) {
                this.tileManager.tileNode(this.nodeHolder);
            } else {
                this.nodeHolder.writeToStream(this.nodeOutStream);
            }
            this.nNodesWritten++;
        }

        if (this.nNodesRead % 500000 == 0) {
            log.progress(DF(this.nNodesWritten) + " of "
                + DF(this.nNodesRead) + " nodes extracted");
        }
    }

    @Override
    public void onWayComplete() {
        this.nWaysRead++;

        this.wayTagResolver.onTagsComplete();

        boolean isValid = this.wayTagResolver.isValid();

        if (isValid || this.preFilterDisabled) {

            // Diese Pruefung ist unabhaengig vom TagResolver
            if (this.wayHolder.getRefsSize() < 2) {
                ++this.nWaysInvalid;
                return;
            }

            this.wayHolder.setInvalid(!isValid); // Catch invalid Ways in Joining-Phase
            this.wayHolder.setName(new VarString(this.wayTagResolver.getName()));
            this.wayHolder.setCustom(this.wayTagResolver.getCustom());
            this.wayHolder.setClazz(this.wayTagResolver.getClazz());
            this.wayHolder.setFlags(this.wayTagResolver.getFlags());
            this.wayHolder.setKmh(this.wayTagResolver.getSpeedInKmh());
            this.wayHolder.setRoundabout(this.wayTagResolver.isRoundabout());
            this.wayHolder.setRail(this.wayTagResolver.isRail());
            this.wayHolder.setPolygon(this.wayTagResolver.isPolygon());

            // ONEWAY - Zusatz-Handling
            int oneWay = this.wayTagResolver.getOneWayType();
            this.wayHolder.setOneWay(WayTagResolver.ONEWAY_NO != oneWay);
            if (WayTagResolver.ONEWAY_YES_REVERSED == oneWay) {
                this.nOneWaysReverse++;
                this.wayHolder.reverse();
            }
            if (WayTagResolver.ONEWAY_YES_UNKNOWN == oneWay) {
                this.nOneWaysUnknown++;
                this.wayHolder.setKmh(1);
            }

            this.wayHolder.writeToStream(this.wayOutStream);
            this.nWaysWritten++;

            if (this.ndRefSet != null) {
                int refsSize = this.wayHolder.getRefsSize();
                for (int i = 0; i < refsSize; i++) {
                    long ndRef = this.wayHolder.getRef(i);
                    if (!this.registerNdRefAndRetryOnOme(ndRef)) {
                        log.warn("No memory for Way-NodeRefs. "
                                + DF(this.ndRefSet.getSize()) + " cached - "  + MB());
                        this.ndRefSet.clear();
                        this.ndRefSet = null;
                        break;
                    }
                }
            }
        }

        if (this.nWaysRead % 100000 == 0) {
            log.progress(DF(this.nWaysWritten) + " of "
                + DF(this.nWaysRead) + " ways extracted - " + MB());
        }
    }

    @Override
    public void onRelationComplete() {
        this.nRelationsRead++;

        this.relationTagResolver.onTagsComplete();

        boolean isValid = this.relationTagResolver.isValid();

        if (isValid || this.preFilterDisabled) {
            byte clazz = this.relationTagResolver.getClazz();
            int flags = this.relationTagResolver.getFlags();
            String relationName = this.relationTagResolver.getName();

            // Zusaetzlicher PreCheck wenn Restriction
            if (!preFilterDisabled)
                if (CLAZZ_RESTRICTION == (CLAZZ_MASK & clazz))
                    if (!this.validateRestrictionMembers())
                        return;

            this.relationHolder.setInvalid(!isValid);
            this.relationHolder.setClazz(clazz);
            this.relationHolder.setFlags(flags);
            if (relationName != null)
                this.relationHolder.setName(new VarString(relationName));

            this.relationHolder.writeToStream(this.relationOutStream);
            this.nRelationsWritten++;

            // Im PreFilter-Mode auch die Relation-Node-Refs vermerken.
            // Wichtig bei z.B. Rail-Network-Stations, die nicht direkt
            // auf dem Graphen liegen.
            if (this.ndRefSet != null) {
                for (Member member : this.relationHolder.getMembers()) {
                    if (MEMBER_TYPE_NODE == member.getType()) {
                        long ndRef = member.getRef();
                        if (!this.registerNdRefAndRetryOnOme(ndRef)) {
                            log.warn("No memory for Relation-NodeRefs. "
                                    + DF(this.ndRefSet.getSize()) + " cached - "  + MB());
                            this.ndRefSet.clear();
                            this.ndRefSet = null;
                            break;
                        }
                    }
                }
            }
        }

        if (this.nRelationsRead % 100000 == 0) {
            log.progress(DF(this.nRelationsWritten) + " of "
                + DF(this.nRelationsRead) + " relations extracted");
        }
    }

    // ------------ On Tag -------------

    @Override
    public void onNodeTag(String key, String value) {
        this.nodeTagResolver.onTag(key, value);
    }

    @Override
    public void onNodeObj(String key, Object value) {
        this.nodeTagResolver.onObj(key, value);
    }

    @Override
    public void onWayTag(String key, String value) {
        this.wayTagResolver.onTag(key, value);
    }

    @Override
    public void onWayObj(String key, Object value) {
        this.wayTagResolver.onObj(key, value);
    }

    @Override
    public void onRelationTag(String key, String value) {
        this.relationTagResolver.onTag(key, value);
    }

    // ----------- On Specials ------------

    @Override
    public void onNodeAction(String action) {
    }

    @Override
    public void onWayNdRef(long ndRefId) {
        try {
            ndRefId += this.nodeIdOffset;
            this.wayHolder.addRef(ndRefId);

        } catch (IllegalArgumentException e) {
            // usually if previous ref equals current
            log.bugfix(e.getMessage());
        }
    }

    @Override
    public void onRelationMember(String type, long ref, String role) {
        if (this.nodeIdOffset != 0 && "node".equals(type)) ref += this.nodeIdOffset;
        else if (this.wayIdOffset != 0 && "way".equals(type)) ref += this.wayIdOffset;
        else if (this.relationIdOffset != 0 && "relation".equals(type)) ref += this.relationIdOffset;
        this.relationHolder.addMember(type, ref, role);
    }

    // ------------ Close Processings -------------

    @Override
    public void closeNodeProcessing() {
        this.nodesProcessingOpen = false;
        this.nodeTagResolver.close();
        if (this.nodeOutStream != null) this.nodeOutStream.close();
        if (this.preFilterDisabled) this.tileManager.closeNodeTiling();

        log.info(DF(this.nNodesWritten) + " of "
            + DF(this.nNodesRead) + " nodes extracted - " + MB());
    }

    @Override
    public void closeWayProcessing() {
        this.waysProcessingOpen = false;
        this.wayTagResolver.close();
        if (this.wayOutStream != null) this.wayOutStream.close();
        this.tileManager.setNumberOfWays(this.nWaysWritten);

        log.info(DF(this.nWaysWritten) + " of "
            + DF(this.nWaysRead) + " ways extracted - " + MB());
    }

    @Override
    public void closeRelationProcessing() {
        this.relationsProcessingOpen = false;
        this.relationTagResolver.close();
        if (this.relationOutStream != null) this.relationOutStream.close();
        this.tileManager.setNumberOfRelations(this.nRelationsWritten);

        log.info(DF(this.nRelationsWritten) + " of "
            + DF(this.nRelationsRead) + " relations extracted - " + MB());
    }

    /**
     * Prueft grob die Relation-Members auf eine gueltige
     * Zusammenstellung fuer eine Restriction.
     * @return boolean moegliche, gueltige Restriction ja/nein.
     */
    private boolean validateRestrictionMembers() {
        final int expectedFlags = 0x1F;
        int actualFlags = 0x0;
        for (Member member : this.relationHolder.getMembers()) {
            switch (member.getRole()) {
            case MEMBER_ROLE_FROM: actualFlags |= 0x1; break;
            case MEMBER_ROLE_VIA: actualFlags |= 0x2;  break;
            case MEMBER_ROLE_TO: actualFlags |= 0x4;  break;
            }
            switch (member.getType()) {
            case MEMBER_TYPE_NODE: actualFlags |= 0x8; break;
            case MEMBER_TYPE_WAY: actualFlags |= 0x10;  break;
            }
        }
        return (actualFlags == expectedFlags);
    }

    /**
     * Liest im Nachgang alle Knoten und sortiert diese
     * in Quadranten.
     */
    private void tileNodesAsPostprocess() {
        if (ndRefSet != null) { // No OutOfMemory
            log.info("Building set of referenced NodeIds");
            this.shrinkNdRefSet();
            log.info("Postprocessing "
                    + DF(this.ndRefSet.getSize()) + " referenced nodes");
        } else {
            log.info("Postprocessing nodes (without NdRefSet)");
        }

        InStream inStream = new InStreamDisk(this.nodeRawFile);
        inStream.readByte(); // VARTYPE
        long n = 0;
        while (!inStream.isEof()) {
            this.nodeHolder.readFromStream(inStream);
            if (null == this.ndRefSet // OutOfMemory
                    || this.ndRefSet.contains(this.nodeHolder.getId())) { // Filter
                this.tileManager.tileNode(this.nodeHolder);
                if (++n % 100000 == 0) log.progress(DF(n) + " nodes tiled.");
            }
        }
        inStream.close();
        log.info(DF(n) + " nodes tiled.");

        if (this.ndRefSet != null) this.ndRefSet.clear();
        this.tileManager.closeNodeTiling();
    }

    @Override
    public void close() {
        if (!this.isOpen) {
            if (this.log != null) log.warn(this.getClass().getSimpleName()
                    + " already closed or not yet opened");
            return;
        }

        // MaxIds global-galaktisch vermerken
        System.setProperty(SYSPROP_NODE_ID_MAX, this.maxNodeId + "");
        System.setProperty(SYSPROP_WAY_ID_MAX, this.maxWayId + "");
        System.setProperty(SYSPROP_REL_ID_MAX, this.maxRelationId + "");

        if (this.nNodesRead == 0) {
            log.error("OSM-Source does not contain Nodes");
            this.onNode(0, 0, 0);
        }

        if (this.nWaysRead == 0) {
            log.warn("OSM-Source does not contain Ways");
            this.onWay(0);
        }

        // Ohne Knoten und Wege ist's sinnfrei, jedoch
        // ggf. leere Relations-Datei fuer Joiner erzeugen.
        if (this.nRelationsRead == 0) {
            log.note("OSM-Source does not contain Relations");
            this.onRelation(0);
        }

        // Schliessen sicherstellen, wenn Reihenfolge von
        // Nodes, Ways und Relations beliebig oder keine Relations.
        if (this.nodesProcessingOpen) this.closeNodeProcessing();
        if (this.waysProcessingOpen) this.closeWayProcessing();
        if (this.relationsProcessingOpen) this.closeRelationProcessing();

        if (!this.preFilterDisabled) {
            this.tileNodesAsPostprocess();
        }

        this.tileManager.writeToFile();
        this.nodeRawFile.delete();

        this.isOpen = false;
    }

}
