/*
 * $Id: DefaultGraphBuildInterceptor.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.graph;

import java.io.File;

import de.cm.osm2po.Config;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;

/**
 * Default-Implementation von {@link GraphBuildInterceptor}, die
 * lediglich die {@link #onEdgeCreate(Edge, int, boolean)}
 * ueberschreibt.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class DefaultGraphBuildInterceptor implements GraphBuildInterceptor {
    private SegmentedWay way;
    @Override public void onWay(SegmentedWay way) {
        this.way = way;
    }

    /**
     * Modifiziert die Kante lediglich mittels {@link Edge#setFlags(int)}<br>
     * Es werden dabei 4 einzelne Bytes kodiert.
     * Die Reihenfolge (low to high byte) ist:<br>
     * <ol>
     * <li>way.Clazz</li>
     * <li>targetNode.Clazz</li>
     * <li>way.Flags</li>
     * <li>way.internFlags</li>
     * </ol>
     * Somit liefert ein einfacher Cast dieses Ints auf (byte) bereits
     * den Strassentypen (way.Clazz). Aber Achtung:<br>
     * Einige der Attribute sind eigentlich selber ints, was bedeutet,
     * dass tatsaechlich nur ihre untersten 8 Bits geschrieben werden.
     * @param edge {@link Edge}
     * @param segmentIdx int Index des Segments im {@link SegmentedWay}.
     * @param reverse true/false Richtung der {@link Edge}-Erzeugung.
     */
    @Override
    public void onEdgeCreate(Edge edge, int segmentIdx, boolean reverse) {
        Node targetNode = reverse ? this.way.getSourceNode(segmentIdx) : this.way.getTargetNode(segmentIdx);
        int wayClazz = this.way.getClazz() & 0xFF;
        int nodeClazz = targetNode.getClazz() & 0xFF;
        int wayFlags = this.way.getFlags() & 0xFF;
        int wayInternFlags = this.way.getInternFlags() & 0xFF;
        int flags = wayInternFlags << 24 | wayFlags << 16 | nodeClazz << 8 | wayClazz;
        edge.setFlags(flags);
    }

    // not used
    @Override public void onOpen(Config config) {}
    @Override public void onClose(File graphFile) {}
    @Override public void onVertex(Vertex vertex) {}
    @Override public void onEdgePosition(int edgeIdx, Edge edge) {}

}
