/*
 * $Id: GraphBuildInterceptor.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.graph;

import java.io.File;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.GraphBuilder;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.model.WaySegment;
import de.cm.osm2po.routing.Graph;

/**
 * Callback-Schnittstelle, um {@link SegmentedWay}s in mehrere {@link Edge}s
 * zu transformieren bzw. diese fuer den {@link Graph} zu verdichten.
 * Verdichten heisst, dass eine Vielzahl von Merkmalen des Weges
 * platzsparend in einem einzigen Integer ({@link Edge#getFlags()})
 * untergrebracht werden muessen.
 * <br><br><b>Wichtig:</b><br>
 * Der {@link GraphBuilder} errechnet Laenge und Zeit fuer
 * eine Kante direkt aus dem Weg. Als Flags wird lediglich die Clazz,
 * also der Weg-Typ als Byte uebertragen. Ein {@link GraphBuildInterceptor}
 * hat also nun die Moeglichkeit, bestimmte Parameter neu zu berechnen und
 * zu ueberschreiben. Iteriert wird ueber alle Wege, wobei fuer jeden
 * einzelnen folgendes aufgerufen wird:
 * <ol>
 * <li>Pro {@link SegmentedWay} 1x {@link #onWay(SegmentedWay)}</li>
 * <li>Pro {@link WaySegment} 2x {@link #onEdgeCreate(Edge, int, boolean)}</li>
 * </ol>
 * Danach werden die Daten transformiert und sortiert, was ebenfalls fuer die
 * Vertices gilt. Daher werden abschliessend noch folgende Callbacks gerufen.
 * <ol>
 * <li>Pro {@link Vertex} 1x {@link #onVertex(Vertex)} </li>
 * <li>Fuer jede zu schreibende {@link Edge} 1x {@link #onEdgePosition(int, Edge)}</li>
 * </ol>
 * @see DefaultGraphBuildInterceptor
 * @author (c) 2012/2017 - Carsten Moeller - info@osm2po
 */
public interface GraphBuildInterceptor {

    /**
     * Wird einmalig nach der Instanziierung aufgerufen.
     * Dies ist faktisch der Quasi-Konstruktor.
     * @param config {@link Config} Konfiguration.
     */
    void onOpen(Config config);

    /**
     * Bietet die Moeglichkeit, Resourcen aufzuraeumen.
     * Da diese Methode nach dem Schliessen der fertiggestellten Graph-Datei
     * gerufen wird, koennen bereits hier Extensions geschrieben werden, ohne
     * dass ein weiterer Postprozessor angestossen werden muss.
     * @param graphFile Pfad zur fertigen Graph-Datei.
     */
    void onClose(File graphFile);

    /**
     * Wird vor allen Verarbeitungsschritten eines Weges genau 1 Mal aufgerufen.
     * @param way {@link SegmentedWay}
     */
    void onWay(final SegmentedWay way);

	/**
	 * Wird fuer jede Kante 2 Mal, und zwar fuer die
	 * Hin- und Rueckrichtung, aufgerufen.
	 * Erlaubt so das Ueberschreiben von bereits ermittelten
	 * Attributen wie z.B. Zeit oder Strecke.
	 * @param edge {@link Edge} die aus dem {@link WaySegment} abgeleitete Kante.
	 * @param segmentIdx int Index im {@link SegmentedWay}.
	 * @param reverse true/false Richtung der {@link Edge}.
	 * @see #onWay(SegmentedWay)
	 */
	void onEdgeCreate(Edge edge, int segmentIdx, boolean reverse);

	/**
	 * Wird vor dem endgueltigen Schreiben gerufen und informiert
	 * ueber die finale Position der Kante im Graphen.
	 * <br><br><b>Wichtig:</b><br>
	 * Diese Position ist spaeter der wichtigste Parameter fuer den
	 * Zugriff auf diverse Methoden des Graphen, so zum Beispiel
	 * {@link Graph#getEdgeFlags()}.
	 * @param edgeIdx fortlaufender Index der Edge in der Adjazenz-Liste
	 * @param edge {@link Edge} Kante.
	 */
	void onEdgePosition(int edgeIdx, Edge edge);

	/**
	 * Simpler callback, der fuer jeden Vertex 1 Mal ausgeloest wird.
	 * @param vertex {@link Vertex}
	 */
	void onVertex(final Vertex vertex);
}
