/*
 * $Id: DefaultJoinFilter.java 5789 2023-04-11 11:33:11Z carsten $
 *
 * Copyright 2011/2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;
import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Bounding;
import de.cm.osm2po.model.Boundings;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.model.Way;

/**
 * Ein {@link JoinFilter}, der mehrere Bereiche ausschneiden kann.<br>
 * Liefert fuer {@link #validate(Way)} <code>true</code>, sofern ein Start-
 * oder Zielknoten eines Weges innerhalb eines Bereichs liegt.<br>
 * {@link OsmWay#isInvalid()} wird hier nicht validiert.
 * <br><br><b>Wichtig:</b><br>
 * Um eine oder mehrere Regionen ({@link Bounding}) auszuschneiden,
 * muss die property <code>joinFilter.boundings</code> gesetzt werden.
 * Diese erwartet eine alternierende Kette von LatLon-Werten:
 * <code>lat1,lon1,lat2,lon2 ...</code><br>
 * Dabei werden zwei Koordinaten als Rechteck (BoundingBox) und mehr
 * als zwei als Polygon-Ring interpretiert.<br>
 * Mehrere Regionen sind durch Semikolons zu trennen.<br>
 * Sollen die Paarungen als LonLat anstatt LatLon interpretiert werden,
 * so muss <code>boundings.longitudeFirst=true</code> gesetzt sein.
 * <br>Beispiel:
 * <pre>
 * joinFilter.boundings = 53.5,10.0,53.6,10.1,53.5,10.1,53.5,10.0
 * </pre>
 * Die Regionen koennen auch einer Datei entnommen werden.
 * Eine Polyfile-Syntax wird akzeptiert.
 * <br>Beispiel:
 * <pre>
 * myPoly = [path/to/poly.file]
 * joinFilter.boundings.longitudeFirst=true
 * joinFilter.boundings = [myPoly]; [anotherPoly]; ...; 10,0,53.5,...
 * </pre>
 * @author (c) 2012/2023 - Carsten Moeller - info@osm2po
 */
public class BoundingsJoinFilter implements JoinFilter {
    protected String PROP_BOUNDINGS = PROP_PREFIX + "boundings";
    protected String PROP_LONLAT_ORDER = PROP_PREFIX + "boundings.longitudeFirst";

    private Boundings[] boundings; // Support multiple clippings
    private Log log;

    @Override public void tiledInto(int tileId, boolean written) {}

	@Override
	public void open(Config config, String prefix) {
	    this.log = config.getLog();
	    if (null == this.log) this.log = Log.empty();

	    if (prefix != null) {
	        PROP_BOUNDINGS = prefix + "." + PROP_BOUNDINGS;
            PROP_LONLAT_ORDER = prefix + "." + PROP_LONLAT_ORDER;
	    }

        String strBoundings = config.getProperty(PROP_BOUNDINGS);
        if (strBoundings != null) {
            boolean lonFirst = Boolean.valueOf(
                    config.getProperty(PROP_LONLAT_ORDER));
            String[] sbs = strBoundings.split(";");
            log.info("Parsing " + sbs.length
                    + " boundings, longitudeFirst=" + lonFirst);
            this.boundings = new Boundings[sbs.length];
            for (int i = 0; i < sbs.length; i++) {
                Boundings poly = new Boundings(sbs[i], lonFirst);
                this.boundings[i] = poly;
                log.debug(poly.toString());
            }
        }
	}

	@Override
	public void close() {
	}

	@Override
	public boolean validate(Way way) {

	    if (null == this.boundings) {
	        return false;
	    }

	    Node[] nodes = way.getNodes();
	    int nx1 = 0;
	    int nx2 = nodes.length -1;
	    double lat1 = nodes[nx1].getLat();
	    double lon1 = nodes[nx1].getLon();
	    double lat2 = nodes[nx2].getLat();
	    double lon2 = nodes[nx2].getLon();
	    for (int i = 0; i < this.boundings.length; i++) {
	        if (this.boundings[i].contains(lat1, lon1)) return true;
	        if (this.boundings[i].contains(lat2, lon2)) return true;
	    }

	    return false;
	}

}
