/*
 * $Id: ClazzJoinFilter.java 5194 2020-10-31 13:30:03Z carsten $
 *
 * Copyright 2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import de.cm.osm2po.Config;
import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.Way;

/**
 * Ein {@link JoinFilter}, der einen Weg daraufhin validiert, ob er einen bestimmten Typ
 * und/oder Zusatz-Eigenschaten besitzt.
 * <ul>
 * <li><strong>joinFilter.clazzes</strong> : optionales CSV zu filternder Way-Clazzes</li>
 * <li><strong>joinFilter.flags</strong> : optionale Bit-Flags als Int Zusatzfilter</li>
 * <li><strong>joinFilter.validateWay</strong> : Ueberprueft zusaetzlich, ob der Weg valide ist (false)</li>
 * </ul>
 * Der Filter prueft, ob <code>joinFilter.clazzes</code> gleich {@link Way#getClazz()}
 * ist und, wenn gesetzt, ein Flag aus <code>joinFilter.flags</code> mit {@link Way#getFlags()}
 * uebereinstimmt. Gibt es keine Uebereinstimmung, so wird <code>false</code> fuer
 * {@link #validate(Way)} zurueckgeliefert. Werden weder <code>flags</code> noch <code>clazzes</code>
 * gesetzt, dann ist dieser Filter passiv und liefert immer <code>true</code> zurueck.
 * @author (c) 2020 - Carsten Moeller, info@osm2po.de
 * @see FlowJoinFilter
 */
public class ClazzJoinFilter implements JoinFilter {

    private String PROP_CLAZZES = PROP_PREFIX + "clazzes";
    private String PROP_FLAGS = PROP_PREFIX + "flags";
    private String PROP_VALIDATEWAY = PROP_PREFIX + "validateWay";

    private ClazzFlagPredicate wayFilter;
    private boolean validateWay;

    @Override
    public void open(Config config, String prefix) {

        if (prefix != null) {
            PROP_CLAZZES = prefix + "." + PROP_CLAZZES;
            PROP_FLAGS = prefix + "." + PROP_FLAGS;
            PROP_VALIDATEWAY = prefix + "." + PROP_VALIDATEWAY;
        }

        String strFlags = config.getProperty(PROP_FLAGS);
        String strClazzes = config.getProperty(PROP_CLAZZES);
        this.wayFilter = new ClazzFlagPredicate(strFlags, strClazzes);

        this.validateWay = Boolean.valueOf(config.getProperty(PROP_VALIDATEWAY, "false"));
    }

    @Override
    public boolean validate(Way way) {
        if (this.validateWay && way.isInvalid()) return false;
        return this.wayFilter.matches(way);
    }

    // Not used
    @Override public void close() {}
    @Override public void tiledInto(int tileId, boolean written) {}
}
