/*
 * $Id: DefaultJoinFilter.java 5790 2023-04-13 16:17:55Z carsten $
 *
 * Copyright 2011/2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_MASK;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_ROUTE;

import java.util.HashSet;
import java.util.Set;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Membership;
import de.cm.osm2po.model.OsmWay;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.VarString;

/**
 * Default-Implementation von {@link JoinFilter}.<br>
 * Liefert fuer {@link #validate(Way)} <code>true</code>, sofern der Weg fuer
 * {@link OsmWay#isInvalid()} <code>false</code> liefert oder Mitglied einer
 * relevanten Relation ist.
 * <br><br><b>Wichtig:</b><br>
 * Die Weiterverarbeitung kann bei einer Nicht-Mitgliedschaft zu einer
 * Route-Relation mittels der property <code>joinFilter.routeWayClazzes</code>
 * nachtraeglich unterbunden werden. Somit werden auch valide Wege verworfen, wenn deren
 * {@link OsmWay#getClazz()} hier als CSV aufgelistet ist und
 * {@link OsmWay#isRouteMember()} <code>false</code> liefert.
 * Dies ist wichtig, wenn z.B. alle Schienen fuer die Aufloesung einer Train-Relation benoetigt
 * werden, jedoch am Ende nur die wirklich relevanten weiterverarbeitet werden sollen.
 * @author (c) 2012/2017/2023 - Carsten Moeller - info@osm2po
 */
public class DefaultJoinFilter implements JoinFilter {
    protected String PROP_DEFAULT_SPEED = PROP_PREFIX + "defaultSpeed";
    protected String PROP_ROUTEWAYCLAZZES = PROP_PREFIX + "routeWayClazzes";

    private Log log;
    private int defaultSpeed;
    private Set<Byte> routeWayClazzSet;

    @Override public void tiledInto(int tileId, boolean written) {}

	@Override
	public void open(Config config, String prefix) {
	    this.log = config.getLog();
	    if (null == this.log) this.log = Log.empty();

	    if (prefix != null) {
            PROP_DEFAULT_SPEED = prefix + "." + PROP_DEFAULT_SPEED;
            PROP_ROUTEWAYCLAZZES = prefix + "." + PROP_ROUTEWAYCLAZZES;
	    }

	    this.defaultSpeed = Integer.parseInt(
	            config.getProperty(PROP_DEFAULT_SPEED, "30"));

	    String routeWayClazzes = config.getProperty(PROP_ROUTEWAYCLAZZES);
	    if (routeWayClazzes != null) {
	        this.routeWayClazzSet = new HashSet<Byte>();
	        for (String strClazz : routeWayClazzes.split(",")) {
	            this.routeWayClazzSet.add(Byte.valueOf(strClazz.trim()));
	        }
	    }
	}

	@Override
	public void close() {
	}

	@Override
	public boolean validate(Way way) {

	    boolean validByRelation = false;
	    if (way.getMemberships() != null) {

	        boolean wayNameModified = false;

	        for (Membership membership : way.getMemberships()) {
	            byte relationClazz = membership.getClazz();
	            if (CLAZZ_RESTRICTION == (CLAZZ_MASK & relationClazz)) {
	                continue; // hier hat der Strassentyp-Filter vorrang.
	            }

	            validByRelation = true;

	            // Spezialbehandlung, wenn Weg ein Routen-Mitglied
	            if (CLAZZ_ROUTE == (CLAZZ_MASK & relationClazz)) {
                    way.setRouteMember(true); // Fuer Segmenter interessant

                    // RelationName und WayName uebereinander legen
                    VarString relationName = membership.findDistinctParentOrSelf().getName();
                    if (null == relationName) relationName = membership.getName(); // Fallback

                    if (relationName != null) {
                        // Wenn Mitglied von mehr als einer Route, dann Doppel-Benamsung unterbinden
                        if (!wayNameModified) {
                            final VarString wayName = way.getName();
                            String joinedName = relationName.toString();
                            if (wayName != null) joinedName += " (" + wayName.toString() + ")";
                            way.setName(new VarString(joinedName));
                            wayNameModified = true;
                        }
                    }

                    if (way.getKmh() <= 0) {
                        way.setKmh(this.defaultSpeed);
                    }

	            }
	        }
	    }

	    // Logik: Entweder Mitglied einer Relation oder valide.
	    //        Wenn valide aber nicht Mitglied einer Route, dann zusaetzlich
	    //        pruefen, ob die Kombi erlaubt ist. z.B. Schiene ohne Relation Sylt-Shuttle
	    if (!validByRelation) {
	        if (way.isInvalid()) {
	            return false;
	        } else if (!way.isRouteMember()) {
	            if (this.routeWayClazzSet != null
	                    && this.routeWayClazzSet.contains(way.getClazz())) {
	                return false; // z.B. unbekannte Schiene ohne Relation zu etwa ShuttleTrain
	            }
	        }
	    }

        return true;
	}

}
