/*
 * $Id: ExtractorJoinFilter.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Joiner;
import de.cm.osm2po.converter.Merger;
import de.cm.osm2po.converter.Segmenter;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Ein {@link JoinFilter}, der bestimmte Wege waehrend des aktuellen
 * Join-Laufs extrahiert und in ein neues Verzeichnis schreibt.
 * Das Ausgabeformat entspricht wiederum dem des {@link Joiner}s, so dass es von
 * einem {@link Merger} oder {@link Segmenter} weiterverarbeitet werden kann.
 * Die folgenden {@link Props} werden unterstuetzt:
 * <ul>
 * <li><strong>joinFilter.workDir</strong> : Ausgabe workDir (mandatory)</li>
 * <li><strong>joinFilter.prefix</strong> : Ausgabe Prefix (optional)</li>
 * <li><strong>joinFilter.nodeIdOffset</strong> : NodeId um diesen Offset erhoehen. (0)</li>
 * <li><strong>joinFilter.wayIdOffset</strong> : WayId um diesen Offset erhoehen (0)</li>
 * <li><strong>joinFilter.clazzes</strong> : optionales CSV zu filternder Way-Clazzes</li>
 * <li><strong>joinFilter.flags</strong> : optionale Bit-Flags als Int Zusatzfilter</li>
 * </ul>
 * Der Filter verarbeitet nur Wege, bei denen eine der <code>joinFilter.clazzes</code>
 * gleich {@link Way#getClazz()} ist und, wenn gesetzt, ein Flag aus <code>joinFilter.flags</code>
 * mit {@link Way#getFlags()} uebereinstimmt. Gibt es keine Uebereinstimmung, so bleibt
 * dieser Filter passiv und liefert fuer {@link #validate(Way)} <code>true</code> zurueck.
 * Werden weder <code>flags</code> noch <code>clazzes</code> gesetzt, dann wirkt der Filter
 * auf alle Wege, die dann auch extrahiert werden. In diesem Fall wuerde folglich immer
 * <code>false</code> geliefert.
 * <br><br><strong>Achtung:</strong><br>
 * Wurden bereits im {@link Tiler} <code>nodeIdOffset</code> und/oder <code>wayIdOffset</code>
 * gesetzt, dann werden, falls auch hier ungleich <code>0</code>, diese hinzuaddiert.
 * @author (c) 2020 - Carsten Moeller, info@osm2po.de
 * @see FlowJoinFilter
 */
public class ExtractorJoinFilter implements JoinFilter {
    private String PROP_WORK_DIR = PROP_PREFIX + "workDir";
    private String PROP_WORK_PREFIX = PROP_PREFIX + "prefix";
    private String PROP_OFFSET_NODE = PROP_PREFIX + "nodeIdOffset";
    private String PROP_OFFSET_WAY = PROP_PREFIX + "wayIdOffset";
    private String PROP_CLAZZES = PROP_PREFIX + "clazzes";
    private String PROP_FLAGS = PROP_PREFIX + "flags";

    private Log log;
    private File workToDir;
    private Way way;
    private TileManager tileManager;
    private Map<Integer, Tile> tileMap;
    private long nSharedWays;
    private long nSharedWaysNode;
    private OutStreamDisk osShared;
    private long nodeIdOffset;
    private long wayIdOffset;
    private ClazzFlagPredicate filter;

    @Override
    public void open(Config config, String prefix) {
        this.log = config.getLog();

        if (prefix != null) {
            PROP_WORK_DIR = prefix + "." + PROP_WORK_DIR;
            PROP_WORK_PREFIX = prefix + "." + PROP_WORK_PREFIX;
            PROP_OFFSET_NODE = prefix + "." + PROP_OFFSET_NODE;
            PROP_OFFSET_WAY = prefix + "." + PROP_OFFSET_WAY;
            PROP_CLAZZES = prefix + "." + PROP_CLAZZES;
            PROP_FLAGS = prefix + "." + PROP_FLAGS;
        }

        String strFlags = config.getProperty(PROP_FLAGS);
        String strClazzes = config.getProperty(PROP_CLAZZES);
        this.filter = new ClazzFlagPredicate(strFlags, strClazzes);

        this.nodeIdOffset = Long.parseLong(config.getProperty(PROP_OFFSET_NODE, "0"));
        this.wayIdOffset = Long.parseLong(config.getProperty(PROP_OFFSET_WAY, "0"));

        String workPrefix = config.getProperty(PROP_WORK_PREFIX, config.getPrefix());
        String workToDir = config.getProperty(PROP_WORK_DIR);
        if (null == workToDir) {
            throw new RuntimeException("Property '" + PROP_WORK_DIR + "' not set");
        }

        log.note("Extracting '" + workPrefix + "' to '" + workToDir + "'");

        this.workToDir = new File(workToDir);
        this.workToDir.mkdirs();

        this.tileMap = new HashMap<Integer, Tile>(0x100);
        this.tileManager = new TileManager(config);
        this.tileManager.setWorkDir(this.workToDir);
        this.tileManager.setPrefix(workPrefix);

        this.osShared = new OutStreamDisk(
                new File(this.workToDir, Joiner.WAY_FILENAME_SHARED));
        this.osShared.writeByte(VarTypeDesk.typeIdOf(Way.class));
    }

    @Override
    public boolean validate(Way way) {
        if (this.filter.matches(way)) {
            this.way = way; // merken
            return false; // exklusiv.
        } else {
            this.way = null; // Nicht unser Typ
            return true;
        }
    }

    private void handleIdOffsets(Way way) {
        if (this.wayIdOffset != 0) {
            way.setId(way.getId() + this.wayIdOffset);
        }
        if (this.nodeIdOffset != 0) {
            for (Node node : way.getNodes()) {
                long newNodeId = node.getId() + this.nodeIdOffset;
                node.setId(newNodeId);
            }
        }
    }

    @Override
    public void tiledInto(int tileId, boolean written) {
        if (null == this.way) return; // Weg ist nicht unser
        // Wenn hier written==true, dann wurde validate()==false ignoriert
        // und der Weg trotzdem weiterverarbeitet. Das ggf. auch gewollt sein.

        if (tileId < 0) { // shared
            this.nSharedWays++;
            this.nSharedWaysNode += way.getNodes().length;
            this.handleIdOffsets(this.way);
            this.way.writeToStream(this.osShared);
        } else {
            Tile tile = this.tileMap.get(tileId);
            if (null == tile) {
                tile = new Tile(tileId);
                this.tileMap.put(tileId, tile);

                File tiledWayFile = new File(this.workToDir,
                        TileManager.TILED_WAYS_FILENAME(tile.getId()));
                OutStreamDisk osd = new OutStreamDisk(tiledWayFile);
                osd.writeByte(VarTypeDesk.typeIdOf(Way.class));
                tile.setOutStreamDisk(osd);
            }
            handleIdOffsets(this.way);
            this.way.writeToStream(tile.getOutStreamDisk());

            tile.incTiledWaysCounter();
            Node[] wayNodes = way.getNodes();
            int nWayNodes = wayNodes.length;
            tile.incTiledWaysNodesCounter(nWayNodes);
            for (int j = 0; j < nWayNodes; j++)
                tile.incNodeCounter(wayNodes[j].isLatLonOnly());
        }
    }

    @Override
    public void close() {
        this.tileManager.setNumberOfSharedWays(nSharedWays);
        this.tileManager.setNumberOfSharedWaysNodes(nSharedWaysNode);
        this.tileManager.setTiles(this.tileMap.values());
        this.tileManager.writeToFile();

        this.osShared.close();
        for (Tile tile : this.tileMap.values())
            tile.getOutStreamDisk().close();
    }
}
