/*
 * $Id: FlowJoinFilter.java 5288 2021-05-07 19:04:34Z carsten $
 *
 * Copyright 2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.VarBean;

/**
 * Vereinigt mehrere {@link JoinFilter} und steuert deren Kontrollfluss ueber
 * einen simplen Mechanismus.<br>
 * Liefert fuer {@link #validate(Way)} <code>true</code>, wenn alle Aktionen
 * vorher erfolgreich waren und ebenfalls <code>true</code> geliefert haben.
 * Dies ist nicht zwangslaeufig das Ergebnis der einzelnen {@link JoinFilter} selbst,
 * sondern das Resultat der internen Reaktion auf das
 * {@link JoinFilter#validate(Way)}-Ergebnis.<br><br>
 * Die Konfiguration ist denkbar einfach:<br>
 * Anstatt beispielsweise einen {@link DefaultJoinFilter} fuer
 * <code>joinFilter.class</code> zu konfigurieren, wird stattdessen dieser Filter gesetzt.
 * Alle Konfigurations-Eintraege der beteiligten Filter muessen danach mit einem
 * beliebigen Prefix versehen und diesem Filter bekannt gemacht werden.
 * Hier ein Beispiel, mit einem {@link DefaultJoinFilter} der Mittels
 * {@link FlowJoinFilter} verwaltet wird:
 * <pre>
 * joinFilter.class=de.cm.osm2po.converter.joiner.FlowJoinFilter
 * joinFilter.filters=F1,F2,...
 *
 * F1.joinFilter.class=de.cm.osm2po.converter.joiner.DefaultJoinFilter
 * F1.joinFilter.defaultSpeed=50
 *
 * F2.joinFilter.class=...
 * </pre>
 * Normalerweise werden alle Filter in der Deklarations-Reihenfolge durchlaufen.
 * Sobald ein Filter fuer {@link #validate(Way)} <code>false</code> liefert,
 * wird abgebrochen. Dies liegt daran, dass  jeder Filter zusaetzlich die
 * Standard-Reaktion <code>NEXT|EXIT</code> zugewiesen bekommt, die auch wie
 * folgt definiert werden kann:
 * <pre>
 * F1.joinFilter.reaction=NEXT|EXIT
 * </pre>
 * Das erste Kommando wird dabei bei <code>true</code> ausgefuehrt, das zweite
 * logischerweise bei <code>false</code>. Der <code>false</code>-Anteil ist
 * optional und bleibt, falls weggelassen, bei <code>EXIT</code><br>
 * Die folgenden Reaktionen bestimmen den Kontrollfluss:
 * <ul>
 * <li><code>RESUME</code>: Setze das Gesamt-Validationsergebnis auf true zurueck und uebergebe an IGNORE</li>
 * <li><code>IGNORE</code>: Setze das aktuelle Validationsergebnis auf true zurueck und uebergebe an NEXT</li>
 * <li><code>NEXT</code>: Gehe zum naechsten Filter</li>
 * <li><code>EXIT</code>: Liefere sofort das aktuelle Gesamt-Validationsergebnis zurueck</li>
 * <li><code>TRUE</code>: Liefere sofort <code>true</code> zurueck</li>
 * <li><code>FALSE</code>: Liefere sofort <code>false</code> zurueck</li>
 * </ul>
 * Der <cite>naechste</cite> ist immer der deklarierte Nachfolger. Dies kann optional durch Sprungmarken
 * fuer NEXT, IGNORE und RESUME ausgehebelt werden. Beispiele:
 * <ul>
 * <li><code>NEXT:F2</code>: Springe zu Filter mit Prefix F2</li>
 * <li><code>IGNORE:3</code>: Springe zum vierten Filter (Index 3)</li>
 * </ul>
 * Dieser Filter bietet zudem die Option, entgegen {@link Config#loadBeans(String)},
 * externe Bean (-Files) nur temporaer - also nur fuer den Join-Lauf - in den Speicher zu laden.
 * <br><strong>Beispiel:</strong><br>
 * <code>joinFilter.loadbean.myBean=[myBean]</code><br>
 *
 * @author (c) 2020/2021 - Carsten Moeller, info@osm2po.de
 * @see ValidWayJoinFilter
 * @see ClazzJoinFilter
 * @see ProximityJoinFilter
 * @see PoiJoinFilter
 * @see ExtractorJoinFilter
 */
public class FlowJoinFilter implements JoinFilter {

    private String PROP_FILTERS = PROP_PREFIX + "filters";
    private String PROP_BEAN = PROP_PREFIX + "loadbean";

    private static final String PROP_REACTION = "reaction";

    private enum ReactionType {
        RESUME, // Setzt das Gesamtergebnis auf TRUE zurueck und uebergibt an IGNORE.
        IGNORE, // Setzt das aktuelle valid-Ergebnis auf TRUE und uebergibt an NEXT.
        NEXT, // Gehe zum naechsten Filter, ggf. mit Sprung-Label.
        EXIT, // Exit: Liefert das aktuelle True/False-validate-Ergebnis.
        TRUE, // Exit: Liefert sofort True zurueck.
        FALSE, // Exit: Liefert sofort False zurueck.
    }

    private final static class Reaction {
        private final ReactionType type;
        private final String next;
        public Reaction(String typeAndNext, String defaultNext) {
            String[] tan = typeAndNext.split("\\:");
            this.type = ReactionType.valueOf(tan[0].trim().toUpperCase());
            this.next = tan.length > 1 ? tan[1].trim() : defaultNext;
        }
        @Override
        public String toString() {
            return this.type + ":" + this.next;
        }
    }

    private final static class Filter {
        private final JoinFilter joinFilter;
        private final Reaction trueReaction;
        private final Reaction falseReaction;
        private boolean called; // Transient Marker
        private Filter(JoinFilter joinFilter, Reaction trueReaction, Reaction falseReaction) {
            this.joinFilter = joinFilter;
            this.trueReaction = trueReaction;
            this.falseReaction = falseReaction;
        }
        @Override
        public String toString() {
            return joinFilter.getClass().getSimpleName() + " [" + trueReaction + "|" + falseReaction + "]";
        }
    }

    private Config config;
    private Log log;
    private Map<String, Filter> filterMap = new HashMap<String, Filter>();
    private Filter[] filters = new Filter[0x100]; // Fuer simple For-Schleifen
    private List<String> loadedBeanList;

    @Override
    public void open(Config config, String prefix) {
        this.log = config.getLog();
        this.log.debug("Opening " + this.getClass().getSimpleName());
        this.loadedBeanList = new ArrayList<String>();
        this.config = config;

        if (prefix != null) { // Fuer den Fall Forkfilter im Forkfilter
            this.PROP_FILTERS = prefix + "." + this.PROP_FILTERS;
            this.PROP_BEAN = prefix + "." + this.PROP_BEAN;
        }

        Map<String, VarBean> beans = config.loadBeans(PROP_BEAN + ".");
        if (!beans.isEmpty()) {
            for (Entry<String, VarBean> entry : beans.entrySet()) {
                if (config.putBean(entry.getKey(), entry.getValue())) {
                    this.loadedBeanList.add(entry.getKey());
                } else {
                    log.warn(this.getClass().getSimpleName()
                            + ": Bean '" + entry.getKey() + "' already loaded");
                }
            }
        }

        String strFilterPrefixes = config.getProperty(this.PROP_FILTERS);
        if (null == strFilterPrefixes) this.log.error("Missing Property: '" + this.PROP_FILTERS + "'");
        else {
            String[] filterPrefixes = strFilterPrefixes.split(",");
            int nFilters = 0;
            for (int i = 0; i < filterPrefixes.length; i++) {
                String filterPrefix = filterPrefixes[i].trim();
                String classKey = filterPrefix + "." + PROP_PREFIX + "class";
                String className = config.getProperty(classKey);
                try {
                    JoinFilter joinFilter = (JoinFilter) Config.loadClass(className).getDeclaredConstructor().newInstance();
                    this.log.info("Opening '" + filterPrefix+  "' for " + joinFilter.getClass().getSimpleName());
                    joinFilter.open(config, filterPrefix);

                    String[] reactions = (config.getProperty(
                            filterPrefix + "." + PROP_PREFIX + PROP_REACTION, "NEXT|EXIT")
                            + "|EXIT").split("\\|"); // Tricky: if only 1 param set

                    final String thisKey = String.valueOf(i);
                    final String nextKey = String.valueOf(i + 1);
                    Reaction trueReaction = new Reaction(reactions[0], nextKey);
                    Reaction falseReaction = new Reaction(reactions[1], nextKey);

                    Filter filter = new Filter(joinFilter, trueReaction, falseReaction);
                    this.filterMap.put(filterPrefix, filter); // einmal unter Namen ablegen
                    this.filterMap.put(thisKey, filter); // und einmal unter dem Index

                    this.filters[nFilters++] = filter;

                } catch (Exception e) {
                    throw new RuntimeException("Could not create '" + classKey + "' " + className);
                }
            }
            this.filters = Arrays.copyOf(this.filters, nFilters);
        }
    }

    @Override
    public void close() {
        // Nur Beans loeschen, die auch von hier geladen wurden
        for (String bean : this.loadedBeanList) this.config.deleteBean(bean);
        // Wichtig! Schliessen in umgekehrter Reihenfolge.
        for (int i = this.filters.length -1; i >= 0; i--) this.filters[i].joinFilter.close();
    }

    @Override
    public boolean validate(Way way) {
        boolean allValid = true;
        int n = this.filters.length;
        Filter filter = 0 == n ? null : this.filters[0];
        while (filter != null) {
            JoinFilter joinFilter = filter.joinFilter;
            boolean valid = joinFilter.validate(way);
            filter.called = true; // Filter wurde gerufen. Fuer tiledInto() merken
            Reaction reaction = valid ? filter.trueReaction : filter.falseReaction;
            switch (reaction.type) { // Achtung!!! RESUME und IGNORE fallen durch bis NEXT!
            case RESUME: allValid = true; // Sammelergebnis = TRUE, fallthru->IGNORE
            case IGNORE: valid = true; // Letztes Ergebnis = TRUE, fallthru->NEXT
            case NEXT: filter = this.filterMap.get(reaction.next); break;
            case EXIT: filter = null; break;
            case TRUE: return true;
            case FALSE: return false;
            }
            allValid &= valid;
        };
        return allValid;
    }

    @Override
    public void tiledInto(int tileId, boolean written) {
        for (Filter filter : this.filters) {
            if (filter.called) {
                filter.called = false; // reset
                filter.joinFilter.tiledInto(tileId, written); // delegate
            }
        }
    }

}
