/*
 * $Id: JoinedWaysIterator.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import static de.cm.osm2po.misc.Utils.DF;

import java.io.File;
import java.util.Iterator;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Joiner;
import de.cm.osm2po.converter.tiler.Tile;
import de.cm.osm2po.converter.tiler.TileManager;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Tile-uebergreifender Iterator aller vom {@link Joiner}
 * erzeugten Wege bzw. deren Dateien.
 *
 * @author (c) 2014, 2017 - Carsten Moeller - info@osm2po.de
 */
public class JoinedWaysIterator implements Iterator<Way>, Iterable<Way> {

    private File[] tileFiles;
    private Way way;
    private long nWaysExpected;
    private long nWaysRead;
    private int nextTileIdx;
    private InStream is;
    private Log log;

    public JoinedWaysIterator(Config config) {
        File workDir = config.getWorkDir();
        TileManager tm = new TileManager(config).readFromFile();
        int nTiles = tm.getNumberOfTiles();
        this.tileFiles = new File[nTiles +1]; // +1 wegen SharedFile
        this.tileFiles[0] = new File(workDir, Joiner.WAY_FILENAME_SHARED);
        this.nWaysExpected = tm.getNumberOfSharedWays();
        for (int i = 0; i < nTiles; i++) {
            Tile tile = tm.getTile(i);
            this.nWaysExpected += tile.getNumberOfWays();
            this.tileFiles[i+1] = new File(
                    workDir, TileManager.TILED_WAYS_FILENAME(tile.getId()));
        }

        this.nextTileIdx = 0;
        this.is = null; // hasNext()-Criteria
        this.way = new Way();
        this.log = config.getLog();
        this.way = new Way();
        this.gotoNext();
    }

    private boolean gotoNext() {
        // Wenn TileFile offen und noch nicht komplett gelesen
        if (!(null == this.is || this.is.isEof())) return true;

        for (;;) {
            if (this.nextTileIdx > 0) this.is.close(); // close previous
            if (this.nextTileIdx < this.tileFiles.length) {
                File file = this.tileFiles[this.nextTileIdx];
                this.nextTileIdx++;
                log.debug(this.getClass().getSimpleName()
                        + ": Reading from " + file.getName());
                this.is = new InStreamDisk(file);
                byte varType = this.is.readByte();
                if (varType != VarTypeDesk.typeIdOf(Way.class))
                    throw new RuntimeException("Unexpected VarType " + varType);
                if (!this.is.isEof()) { // Tile kann leer sein
                    return true; // naechstes nicht leeres Tile
                }

            } else if (this.is != null) { // Letztes Tile vollstaendig durch
                this.is.close();
                this.is = null;
                if (this.nWaysRead != this.nWaysExpected) {
                    this.log.error(this.getClass().getSimpleName()
                            + ": " + DF(this.nWaysRead) + " ways read but "
                            + DF(this.nWaysExpected) + " expected");
                }
                return false;
            }
        }
    }

    @Override
    public boolean hasNext() {
        return this.is != null;
    }

    @Override
    public Way next() {
        if (null == this.is) return null;
        this.way.readFromStream(this.is); // EOF zuvor in gotoNext() abgesichert
        this.nWaysRead++;
        this.gotoNext();
        return this.way;
    }

    @Override
    public void remove() {
        throw new UnsupportedOperationException(
                "Remove not supported for " + this.getClass().getSimpleName());
    }

    @Override
    public Iterator<Way> iterator() {
        return this;
    }

}
