/*
 * $Id: PoiJoinFilter.java 5490 2021-12-21 21:00:35Z carsten $
 *
 * Copyright 2011/2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.stringToXmlUTF8;

import java.io.File;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Merger;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.extensions.PoiExtension;
import de.cm.osm2po.extensions.PoiExtension.PoiIndexAndDistance;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.model.Bounding;
import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarString;

/**
 * Ein {@link JoinFilter}, der jeweils den dichtesten Knoten aller
 * {@link Way}s zu allen POIs aus einer {@link PoiExtension} herausfiltert.<br>
 * Dieser Filter erstellt daraus kuenstliche Verbindungen und schreibt sie in eine externe
 * osm.xml-Datei, die so von osm2po aber auch <a href="https://josm.openstreetmap.de/">JOSM</a>
 * gelesen werden kann.<br>
 * In einem Folgeschritt koennen die Netzwerke dann mittels {@link Merger}
 * wieder zusammengefuegt werden.<br>
 * Die folgenden {@link Props} werden akzeptiert:
 * <ul>
 * <li><strong>joinFilter.poiBean</strong> : Bean-Name der {@link PoiExtension}.
 *   Muss mittels {@link Config#loadBeans(String)}-<code>loadbean</code>
 *   oder {@link FlowJoinFilter}-<code>loadbean</code> bereitgestellt werden.
 * </li>
 * <li><strong>joinFilter.poiClazzes</strong> : optionales CSV erlaubter Poi-Clazzes</li>
 * <li><strong>joinFilter.xmlFile</strong> : optional abweichender Ausgabedatei-Pfad. (&lt;prefix&gt;_poi.osm)</li>
 * <li><strong>joinFilter.xmlTag</strong> : optional setzbares Tag der Form key/val</li>
 * <li><strong>joinFilter.debug</strong> : Debug-Logging von moeglichen Fehlern (false)</li>
 * <li><strong>joinFilter.clazzes</strong> : optionales CSV zu filternder Way-Clazzes</li>
 * <li><strong>joinFilter.flags</strong> : optionale Bit-Flags als Int Zusatzfilter</li>
 * <li><strong>joinFilter.poiIdOffset</strong> : PoiId um diesen Offset erhoehen. (0)</li>
 * <li><strong>joinFilter.wayIdOffset</strong> : WayId um diesen Offset erhoehen (0), sonst Ids ab 1 fortlaufend vergeben</li>
 * </ul>
 * Der Filter verarbeitet nur Wege, bei denen eine der <code>joinFilter.clazzes</code>
 * gleich {@link Way#getClazz()} ist und, wenn gesetzt, ein Flag aus <code>joinFilter.flags</code>
 * mit {@link Way#getFlags()} uebereinstimmt. Ist nichts gesetzt, so werden alle Wege verarbeitet.
 * <br><br><strong>Wichtig 1:</strong><br>
 * Dieser Filter liefert fuer {@link #validate(Way)} <strong>immer</strong> <code>true</code> zurueck.
 * <br><br><strong>Wichtig 2:</strong><br>
 * Dieser Filter validiert nicht {@link Way#isInvalid()}!<br>
 * Hierfuer sollte eine Kombination aus {@link FlowJoinFilter} und {@link ValidWayJoinFilter}
 * vorgeschaltet werden.
 * <br><br><b>Wichtig 3:</b><br>
 * Die Verwendung von <code>poiIdOffset</code> sollte vermieden werden, da die Original-IDs hier verfremdet werden.
 * Besser: Die PoiIDs bereits in der Tabelle oder geschicktem SQL offsetten. z.B. via <code>SELECT id-1000, ...</code>,
 * um eine Entkopplung der ID-Welten zwischen OSM und externen Sourcen zu gewaehrleisten. Die wayId-Zaehlung beginnt
 * standardmaessig bei 1 und inkrementiert dann pro geschriebenen Weg. Auch hier sollte eine Entkopplung per
 * <code>wayIdOffset</code> sichergestellt sein. Aber Vorsicht: Wird im {@link Tiler} ebenfalls dieser Parameter mit
 * gleichem Namen gesetzt, so addieren sich die IDs. Also Kopf anschalten: Einer von beiden sollte hier wirken.
 * @author (c) 2017/2021 - Carsten Moeller, info@osm2po.de
 * @see FlowJoinFilter
 */
public class PoiJoinFilter implements JoinFilter {
    private final static String V = "version=\"1\""; // required fuer alle OSM-Elemente, >0!

    private String PROP_POIBEAN = PROP_PREFIX + "poiBean";
    private String PROP_POICLAZZES = PROP_PREFIX + "poiClazzes";
    private String PROP_XMLFILE = PROP_PREFIX + "xmlFile";
    private String PROP_XMLTAG = PROP_PREFIX + "xmlTag";
    private String PROP_DEBUG = PROP_PREFIX + "debug";
    private String PROP_CLAZZES = PROP_PREFIX + "clazzes";
    private String PROP_FLAGS = PROP_PREFIX + "flags";
    private String PROP_OFFSET_POI = PROP_PREFIX + "poiIdOffset";
    private String PROP_OFFSET_WAY = PROP_PREFIX + "wayIdOffset";

    private PoiExtension poiExt;
    private Log log;

    private boolean debugErrors;  // mit Error-Report, aber nur innerhalb BBox
    private double lat1, lon1, lat2, lon2;

    private long poiIdOffset;
    private long wayIdOffset;

    private ClazzFlagPredicate poiFilter;
    private ClazzFlagPredicate wayFilter;

    private File toFile;

    private static class NearestNode {OsmNode node; double dist;}
    private Map<Integer, NearestNode> nearestNodeMap;

    private String tagKey, tagValue;

    @Override
    public void open(Config config, String prefix) {

        this.lat1 = 90; // BBOX
        this.lon1 = 180;
        this.lat2 = -90;
        this.lon2 = -180;

        this.log = config.getLog();

        if (prefix != null) {
            PROP_XMLFILE = prefix + "." + PROP_XMLFILE;
            PROP_POIBEAN = prefix + "." + PROP_POIBEAN;
            PROP_XMLTAG = prefix + "." + PROP_XMLTAG;
            PROP_POICLAZZES = prefix + "." + PROP_POICLAZZES;
            PROP_DEBUG = prefix + "." + PROP_DEBUG;
            PROP_CLAZZES = prefix + "." + PROP_CLAZZES;
            PROP_FLAGS = prefix + "." + PROP_FLAGS;
            PROP_OFFSET_POI = prefix + "." + PROP_OFFSET_POI;
            PROP_OFFSET_WAY = prefix + "." + PROP_OFFSET_WAY;
        }

        this.poiIdOffset = Long.parseLong(config.getProperty(PROP_OFFSET_POI, "0"));
        this.wayIdOffset = Long.parseLong(config.getProperty(PROP_OFFSET_WAY, "0"));

        String strFlags = config.getProperty(PROP_FLAGS);
        String strClazzes = config.getProperty(PROP_CLAZZES);
        this.wayFilter = new ClazzFlagPredicate(strFlags, strClazzes);

        String beanName = config.getProperty(this.PROP_POIBEAN);
        VarBean varBean = config.getBean(beanName);
        if (null == varBean) throw new RuntimeException(
                "No Bean set for " + this.getClass().getSimpleName());
        log.debug("Using Bean '" + beanName + "'");
        this.poiExt = (PoiExtension) varBean.get();

        String xmlFilePath = config.getProperty(this.PROP_XMLFILE);
        String osmTag = config.getProperty(this.PROP_XMLTAG);
        this.debugErrors = Boolean.valueOf(config.getProperty(this.PROP_DEBUG));

        if (null == xmlFilePath) {
            this.toFile = new File(config.getWorkDir(), config.getPrefix() + "_poi.osm");
        } else {
            this.toFile = new File(xmlFilePath);
            if (null == this.toFile.getParent()) {
                this.toFile = new File(config.getWorkDir(), this.toFile.getName());
            }
        }

        if (osmTag != null) {
            String[] tagKV = osmTag.split("/");
            if (tagKV.length == 2) {
                this.tagKey = tagKV[0].trim();
                this.tagValue = tagKV[1].trim();
            } else {
                log.warn("Use '/' to separate key and value of '" + tagKV + "'");
            }
        }

        this.poiFilter = new ClazzFlagPredicate(
                null, config.getProperty(this.PROP_POICLAZZES));

        this.nearestNodeMap = new HashMap<Integer, NearestNode>(0x1000);
    }

    @Override
    public void close() {
        this.writeOsmXml();
        this.logStatistics();
    }

    @Override
    public boolean validate(Way way) {
        if (this.wayFilter.matches(way)) {
            for (Node node : way.getNodes()) {

                double lat = node.getLat();
                double lon = node.getLon();

                // BBox of read nodes for final result
                if (lat < this.lat1) this.lat1 = lat;
                if (lat > this.lat2) this.lat2 = lat;
                if (lon < this.lon1) this.lon1 = lon;
                if (lon > this.lon2) this.lon2 = lon;

                PoiIndexAndDistance closestPoi = this.poiExt.findCloserPoiByCoordinate(
                        lat, lon, this.poiFilter, 150, false);

                if (closestPoi != null) {
                    NearestNode nearestNode = this.nearestNodeMap.get(closestPoi.index);
                    if (null == nearestNode) {
                        nearestNode = new NearestNode();
                        nearestNode.node = node;
                        nearestNode.dist = closestPoi.distance;
                        this.nearestNodeMap.put(closestPoi.index, nearestNode);

                    } else if (closestPoi.distance < nearestNode.dist) {
                        nearestNode.dist = closestPoi.distance;
                        nearestNode.node = node;
                    }
                }
            }
        }
        return true;
    }

    private void writeNodeXml(final Writer xml,
            final OsmNode osmNode, boolean isPoi) throws Exception {

        byte clazz = osmNode.getClazz();
        int flags = osmNode.getFlags();
        VarString vName = osmNode.getName();
        String name = null == vName ? null : stringToXmlUTF8(vName.toString());
        long nodeId = osmNode.getId();

        if (isPoi) nodeId += this.poiIdOffset;

        xml.write("\n  <node id=\"" + nodeId + "\" lat=\"" + osmNode.getLat()
                + "\" lon=\"" + osmNode.getLon() + "\" " + V + ">");
        xml.write("\n    <tag k=\"" + "ispoi" + "\" v=\"" + isPoi + "\"/>");
        if (name != null) xml.write("\n    <tag k=\"" + "name" + "\" v=\"" + name + "\"/>");
        if (this.tagKey != null) xml.write(
                "\n    <tag k=\"" + this.tagKey + "\" v=\"" + this.tagValue + "\"/>");
        if (clazz != 0) xml.write("\n    <tag k=\"" + "clazz" + "\" v=\"" + clazz + "\"/>");
        if (flags != 0) xml.write("\n    <tag k=\"" + "flags" + "\" v=\"" + flags + "\"/>");
        xml.write("\n  </node>");
    }

    private void writeOsmXml() {
        try {
            if (this.toFile.getParentFile() != null) this.toFile.getParentFile().mkdirs();
            final Writer xml = new PrintWriter(this.toFile);

            xml.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                    + "\n<osm version=\"0.6\" generator=\"OSM2PO\">");

            for (Entry<Integer, NearestNode> entry : this.nearestNodeMap.entrySet()) {
                int poiIdx = entry.getKey();
                OsmNode poiNode = this.poiExt.findByIndex(poiIdx, new OsmNode());
                OsmNode osmNode = entry.getValue().node;
                this.writeNodeXml(xml, osmNode, false);
                this.writeNodeXml(xml, poiNode, true);
            }

            long wayId = this.wayIdOffset;
            for (Entry<Integer, NearestNode> entry : this.nearestNodeMap.entrySet()) {
                wayId++;
                int poiIdx = entry.getKey();
                OsmNode poiNode = this.poiExt.findByIndex(poiIdx, new OsmNode());
                OsmNode osmNode = entry.getValue().node;

                VarString vPoiName = poiNode.getName();
                String poiName = null == vPoiName ? null : stringToXmlUTF8(vPoiName.toString());
                long poiId = poiNode.getId() + this.poiIdOffset;

                xml.write("\n  <way id=\"" + wayId + "\" " + V + ">");
                xml.write("\n    <nd ref=\"" + osmNode.getId() + "\"/>");
                xml.write("\n    <nd ref=\"" + poiId + "\"/>");
                xml.write("\n    <tag k=\"" + "distance" + "\" v=\"" + entry.getValue().dist + "\"/>");
                if (poiName != null) xml.write("\n    <tag k=\"" + "name" + "\" v=\"" + poiName + "\"/>");
                if (this.tagKey != null) xml.write(
                        "\n    <tag k=\"" + this.tagKey + "\" v=\"" + this.tagValue + "\"/>");
                xml.write("\n  </way>");
            }

            xml.write("\n</osm>");
            xml.close();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private void logStatistics() {
        int nMissing = 0;
        int nPois = 0;
        Bounding bounding = new Bounding(this.lat1, this.lon1, this.lat2, this.lon2);
        log.info("Checking missing connections in "
                + this.toFile.getName() + "\ninside " + bounding);

        for (int x = 0; ; x++) {
            // TODO new OsmNode() als Holder-Ref
            OsmNode poi = this.poiExt.findByIndex(x, new OsmNode());
            if (null == poi) break; // end of list
            if (!this.poiFilter.matches(poi)) continue;
            if (!bounding.contains(poi)) continue;

            if (null == this.nearestNodeMap.get(x)) {
                nMissing++;
                if (this.debugErrors) {
                    log.debug(poi.getName() + " NOT connected: " + poi);
                }
            }
            nPois++;
        }

        if (0 == nMissing) {
            log.info(DF(nPois) + " POIs checked, all connected.");
        } else {
            String msg = DF(nPois) + " POIs checked, "
                + DF(nMissing) + " connections missing.";
            if (this.debugErrors) {
                msg += "\nDetailed Logs on Debug-Level.";
            } else {
                msg += "\nSet " + PROP_DEBUG + "=true to see details.";
            }
            log.warn(msg);
        }
    }

    @Override public void tiledInto(int tileId, boolean written) {}

}
