/*
 * $Id: ProximityJoinFilter.java 5490 2021-12-21 21:00:35Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.PoiExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.primitives.VarBean;

/**
 * Ein {@link JoinFilter}, der Wege daraufhin validiert, ob sich mindestens
 * einer der Knoten in der Nachbarschaft eines aus einer {@link PoiExtension} befindet.
 * Die folgenden {@link Props} werden akzeptiert (Defaults ggf. in Klammern):
 * <ul>
 * <li><strong>joinFilter.poiBean</strong> : Bean-Name der {@link PoiExtension}</li>
 * <li><strong>joinFilter.poiClazzes</strong> : optionales CSV erlaubter Poi-Clazzes</li>
 * <li><strong>joinFilter.poiDistance</strong> : Umkreis in km (5)</li>
 * <li><strong>joinFilter.clazzes</strong> : optionales CSV zu filternder Way-Clazzes</li>
 * <li><strong>joinFilter.flags</strong> : optionale Bit-Flags als Int Zusatzfilter</li>
 * </ul>
 * Der Filter verarbeitet nur Wege, bei denen eine der <code>joinFilter.clazzes</code>
 * gleich {@link Way#getClazz()} ist und, wenn gesetzt, ein Flag aus <code>joinFilter.flags</code>
 * mit {@link Way#getFlags()} uebereinstimmt. Gibt es keine Uebereinstimmung, so wird ohne
 * weitere Pruefung <code>false</code> fuer {@link #validate(Way)} zurueckgeliefert.
 * Werden weder <code>flags</code> noch <code>clazzes</code> gesetzt, dann werden alle
 * Wege auf Umkreisnaehe untersucht.
 * <br><br><strong>Wichtig 1:</strong><br>
 * Die Distanz ist eine Eingrenzung, die allerdings zusaetzlich abhaengig von der bereits
 * in der {@link PoiExtension} gesetzten {@link PoiExtension#getZoneSize()} ist.
 * <br><br><strong>Wichtig 2:</strong><br>
 * Dieser Filter validiert nicht {@link Way#isInvalid()}!
 * @author (c) 2017/2020 - Carsten Moeller, info@osm2po.de
 * @see FlowJoinFilter
 */
public class ProximityJoinFilter implements JoinFilter {

    private String PROP_POIDIST = PROP_PREFIX + "poiDistance";
    private String PROP_POIBEAN = PROP_PREFIX + "poiBean";
    private String PROP_POICLAZZES = PROP_PREFIX + "poiClazzes";
    private String PROP_CLAZZES = PROP_PREFIX + "clazzes";
    private String PROP_FLAGS = PROP_PREFIX + "flags";

    private PoiExtension poiExt;
    private double perimeter;

    private ClazzFlagPredicate poiFilter;
    private ClazzFlagPredicate wayFilter;

    private Log log;

    @Override
    public void open(Config config, String prefix) {
        this.log = config.getLog();

        if (prefix != null) {
            PROP_POIBEAN = prefix + "." + PROP_POIBEAN;
            PROP_POIDIST = prefix + "." + PROP_POIDIST;
            PROP_POICLAZZES = prefix + "." + PROP_POICLAZZES;
            PROP_CLAZZES = prefix + "." + PROP_CLAZZES;
            PROP_FLAGS = prefix + "." + PROP_FLAGS;
        }

        String strFlags = config.getProperty(PROP_FLAGS);
        String strClazzes = config.getProperty(PROP_CLAZZES);
        this.wayFilter = new ClazzFlagPredicate(strFlags, strClazzes);

        this.perimeter = Double.valueOf(config.getProperty(PROP_POIDIST, "5"));

        String beanName = config.getProperty(this.PROP_POIBEAN);
        VarBean varBean = config.getBean(beanName);
        if (null == varBean) throw new RuntimeException(
                "No Bean set for " + this.getClass().getSimpleName());
        log.note("Using Nodes of Bean '" + beanName + "'");
        this.poiExt = (PoiExtension) varBean.get();

        this.poiFilter = new ClazzFlagPredicate(
                null, config.getProperty(this.PROP_POICLAZZES));
    }


    @Override
    public boolean validate(Way way) {
        if (this.wayFilter.matches(way)) {
            Node[] nodes = way.getNodes();
            int nNodes = nodes.length;
            for (int i = 0; i < nNodes; i++) {
                Node node = nodes[i];
                if (this.poiExt.findCloserPoiByCoordinate(node.getLat(), node.getLon(),
                        this.poiFilter, this.perimeter, true) != null)
                    return true;
            }
        }
        return false;
    }

    // Not used
    @Override public void close() {}
    @Override public void tiledInto(int tileId, boolean written) {}
}
