/*
 * $Id: RelationCache.java 4423 2017-04-16 08:26:55Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.joiner;

import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_NODE;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_RELATION;
import static de.cm.osm2po.model.OsmRelation.MEMBER_TYPE_WAY;
import de.cm.osm2po.model.OsmRelation;
import de.cm.osm2po.model.OsmRelation.Member;
import de.cm.osm2po.model.Membership;
import de.cm.osm2po.primitives.LongArrayAsc;
import de.cm.osm2po.primitives.LongLongBag;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessMemory;

/**
 * Diese Klasse bietet einen RelationCache nach dem Prinzip
 * WriteOnce ReadOften.<br>
 * Da wir davon ausgehen koennen, dass OSM-Extrakte
 * aus Datenbanken stammmen und daher die Relationen bereits
 * eindeutig sind, werden hier Bags und keine Maps verwendet.
 * Daher sollte nach dem Fuellen einmal {@link #sort()}
 * aufgerufen werden, damit die folgenden Zugriffe mittels
 * Binaersuche optimiert werden koennen.
 * <br><br><b>Wichtig:</b><br>
 * Maximale Anzahl der Relationen (Elemente) ist 2Gig.
 */
public class RelationCache {
    private static final int REFIDBITS = 62;
    private static final long REFIDMASK = (1L << REFIDBITS) - 1;
    
    // Roles
    private LongLongBag members = new LongLongBag();
    private LongArrayAsc index = new LongArrayAsc();
    private RandomAccess ram = new RandomAccessMemory();

    public void clear() {
        this.members.clear();
        this.index.clear();
        this.ram.clear();
    }
    
    /**
     * Sollte zwingend nach dem Fuellen des Caches aufgerufen werden,
     * damit spaetere lookups performant mittels Binaersuche
     * durchgefuehrt werden koennen.
     */
    public void sort() {
        this.members.sort();
    }
    
    /**
     * Cached eine Relation.
     * @param osmRelation {@link OsmRelation}.
     * @return boolean hinzugefuegt, ja/nein
     */
    public boolean add(OsmRelation osmRelation) {
        long idx = this.index.push(ram.getWriteIndex());
        if (idx > Integer.MAX_VALUE) throw new IllegalStateException(
                "Too many Relations for Cache");
        osmRelation.writeToStreamWithoutMembers(this.ram);
        
        long order = 0;
        for (Member member : osmRelation.getMembers()) {
            long ref = member.getRef(); // Id of Node, Way or Relation
            long type = member.getType() & 0xFF; // Node, Way or Relation
            long role = member.getRole() & 0xFF; // Role e.g. Via, Outer, Forward
            if (ref > REFIDMASK) throw new IllegalStateException("OSM Id " + ref
                    + " exceeds range in Relation " + osmRelation.getId());
            
            long key = (type << REFIDBITS) | (ref & REFIDMASK);
            long value = order++ << 40 | role << 32 | idx;
            
            this.members.add(key, value);
        }
        return true;
    }
    
    public Membership[] lookupNodeMemberships(long nodeId) {
        return lookupMemberships(nodeId, MEMBER_TYPE_NODE);
    }
    
    public Membership[] lookupWayMemberships(long wayId) {
        return lookupMemberships(wayId, MEMBER_TYPE_WAY);
    }

    public Membership[] lookupRelationMemberships(long relationId) {
        return lookupMemberships(relationId, MEMBER_TYPE_RELATION);
    }
    
    private Membership[] lookupMemberships(long ref, byte type) {
        long key = ((long)type << REFIDBITS) | (ref & REFIDMASK);
        long vals[] = this.members.getAll(key);
        if (null == vals) return null;
        Membership[] memberships = new Membership[vals.length];
        OsmRelation osmRelation = new OsmRelation();
        for (int i = 0; i < vals.length; i++) {
            int idx = (int) (vals[i] & 0x7FFFFFFFL);
            byte role = (byte) ((vals[i] >>> 32) & 0xFF);
            int order = (int) (vals[i] >>> 40);
            
            long ptr = this.index.get(idx);
            this.ram.setReadIndex(ptr);
            
            osmRelation.readFromStreamWithoutMembers(this.ram);
            memberships[i] = new Membership(role, order, osmRelation);
        }
        return memberships;
    }

}
