/*
 * $Id: OsmXmlParser.java 4696 2018-03-07 10:53:43Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.parser;

import java.io.InputStream;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Utils;

/**
 * Diese Implementation von {@link OsmParser} liest Plain-XML in
 * Form von Standard *.osm-Dateien und informiert einen angeschlossenen
 * {@link OsmParserHandler} ueber die geparsten Nodes, Ways und Relations.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 * @see OsmParserHandler
 * @see Tiler
 */
public class OsmXmlParser extends DefaultHandler implements OsmParser {

    private SAXParser saxParser;
    private InputStream xmlInputStream;
    private OsmParserHandler osmParserHandler;

    // We do not have deep structures, so there's no need
    // for a stack. Instead I'm using simple boolean flags.
    private boolean isWay = false;
    private boolean isNode = false;
    private boolean isRelation = false;

    private boolean isFirstWay = true;
    private boolean isFirstRelation = true;

    private Log log;
    private long nSkippedTags;

    private boolean arbitrary;

    @Override
    public void open(InputStream xmlInputStream,
            OsmParserHandler osmParserHandler, Config config) {
        this.log = config.getLog();
        try {

            SAXParserFactory saxFactory = SAXParserFactory.newInstance();
            saxFactory.setValidating(false);
            this.saxParser = saxFactory.newSAXParser();
            this.xmlInputStream = xmlInputStream;
            this.osmParserHandler = osmParserHandler;

            this.arbitrary = Boolean.valueOf(
                    config.getProperty("parser.arbitrary"));

        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    @Override
    public void parse() {
        try {
            this.saxParser.parse(this.xmlInputStream, this);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void close() {
        this.saxParser.reset();
        this.saxParser = null;
        this.isFirstWay = true;
        this.isFirstRelation = true;
    }

    @Override
    public void startElement(String uri, String localName,
            String name, Attributes attributes) throws SAXException {

        if (this.isNode) {
            if ("tag".equals(name)) {
                String k = attributes.getValue("k");
                String v = attributes.getValue("v");
                this.osmParserHandler.onNodeTag(k, v);
            }

        } else if (this.isWay) {
            if ("nd".equals(name)) {
                long id = Long.parseLong(attributes.getValue("ref"));
                this.osmParserHandler.onWayNdRef(id);

            } else if ("tag".equals(name)) {
                String k = attributes.getValue("k");
                String v = attributes.getValue("v");
                this.osmParserHandler.onWayTag(k, v);
            }

        } else if ("node".equals(name)) {
            this.isNode = true;
            long id = Long.parseLong(attributes.getValue("id"));
            double lon = Double.parseDouble(attributes.getValue("lon"));
            double lat = Double.parseDouble(attributes.getValue("lat"));

            this.osmParserHandler.onNode(id, lat, lon);
            this.osmParserHandler.onNodeAction(attributes.getValue("action"));

        } else if ("way".equals(name)) {
            if (this.isFirstWay) {
                this.isFirstWay = false;
                if (!this.arbitrary)
                    this.osmParserHandler.closeNodeProcessing();
            }
            this.isWay = true;
            long id = Long.parseLong(attributes.getValue("id"));
            this.osmParserHandler.onWay(id);

        } else if (this.isRelation) {
            if ("member".equals(name)) {
                String type = attributes.getValue("type");
                long ref = Long.parseLong(attributes.getValue("ref"));
                String role = attributes.getValue("role");
                this.osmParserHandler.onRelationMember(type, ref, role);

            } else if ("tag".equals(name)) {
                String key = attributes.getValue("k");
                String value = attributes.getValue("v");
                this.osmParserHandler.onRelationTag(key, value);
            }

        } else if ("relation".equals(name)) {
            if (this.isFirstRelation) {
                if (!this.arbitrary)
                    this.osmParserHandler.closeWayProcessing();
                this.isFirstRelation = false;
            }
            this.isRelation = true;
            long id = Long.parseLong(attributes.getValue("id"));
            this.osmParserHandler.onRelation(id);

        } else {
            this.nSkippedTags++;
            if (this.nSkippedTags % 100000 == 0) {
                this.log.progress(Utils.DF(this.nSkippedTags)
                        + " Tags ignored. (" + name + ")");
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String name) throws SAXException {
        try {
            if ("node".equals(name)) {
                this.isNode = false;
                this.osmParserHandler.onNodeComplete();

            } else if ("way".equals(name)) {
                this.isWay = false;
                this.osmParserHandler.onWayComplete();

            } else if ("relation".equals(name)) {
                this.isRelation = false;
                this.osmParserHandler.onRelationComplete();

            } else if ("osm".equals(name)) {
                if (!this.arbitrary)
                    this.osmParserHandler.closeRelationProcessing();
            }

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }


}
