/*
 * $Id: QuickNet.java 5712 2022-09-23 07:52:26Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.parser;

import static de.cm.osm2po.misc.Utils.stringToXmlUTF8;

import java.io.File;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.tiler.TagResolver;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoJsonBuilder;
import de.cm.osm2po.misc.GeoJsonBuilder.Feature;
import de.cm.osm2po.misc.GeoJsonBuilder.FeatureCollection;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.LongSet;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarString;

/**
 * Ein {@link OsmParserHandler}, der ein On-the-Fly-Netzwerk aus
 * einem Osm-File (nur Knoten und Wege) erzeugt und dieses als
 * Bean(-File) bereitstellt.<br>
 * Diese Klasse ist auch gleichzeitig {@link Var} der Bean selbst.
 * <br><br><b>Wichtig:</b>
 * Hier wird kein {@link TagResolver} aufgerufen. Lediglich fuer Wege
 * wird das Name-Tag direkt im Objekt gespeichert.
 * Alle anderen Tags werden extra vermerkt und koennen per
 * {@link #getTag(int)} via {@link QuickWay#tags} oder {@link QuickNode#tags}
 * abgefragt werden.
 * <ul>
 * <li><strong>parser.handler.beanFile</strong> : Ausgabepfad fuer Bean (quicknet.bean)</li>
 * <li>
 * <strong>parser.handler.bean</strong> : Ruft zusaetzlich
 * {@link Config#putBean(String, VarBean)} mit dem Wert als Name der Bean und
 * <code>new VarBean(this)</code>, so dass im Folgeschritt diese Instanz sofort
 * verwendet werden kann.
 * </li>
 * </ul>
 * <strong>Beispiel</strong>
 * <pre>
 * cmd=t
 * file=path/to/osm.file
 * parser.handler.class=de.cm.osm2po.model.QuickNet
 * parser.handler.beanFile=path/to/bean.file
 * </pre>
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 */
public class QuickNet implements OsmParserHandler, Var<QuickNet> {
    private Log log;
    private File beanFile;

    private LinkedHashMap<Long, QuickNode> nodeMap; // Sichert Einfuegereihenfolge zu
    private List<QuickWay> wayList;
    private HashMap<String, Integer> tagMap;
    private String[] tagArr;

    // Pro-Element-Variablen
    private long id; // Eigentliche, auch abweichende Node-Id
    private String name;
    private float lat;
    private float lon;
    private long[] ndRefs;
    private int nNdRefs;
    private int[] tags;
    private int nTags;

    public QuickNet() {
        this.nodeMap = new LinkedHashMap<Long, QuickNode>();
        this.wayList = new ArrayList<QuickWay>();
        this.tagMap = new HashMap<String, Integer>();
        this.tagArr = new String[]{}; // Fuer UnitTest

        this.tags = new int[0x4000]; // Tags pro Element!
        this.nTags = 0; // Anzahl Tags pro Element!
        this.ndRefs = new long[0x1000]; // Pro Weg
        this.nNdRefs = 0; // Pro Weg
    }

    public final Collection<QuickNode> nodes() {return this.nodeMap.values();}
    public final Collection<QuickWay> ways() {return this.wayList;}
    public final QuickNode node(long ndRef) {return this.nodeMap.get(ndRef);}

    public final static class QuickNode {
        private long id;
        private boolean orphaned; // Transient
        private float lat;
        private float lon;
        private boolean modified;
        public final int[] tags;
        public QuickNode(long id, float lat, float lon, int[] tags) {
            this.id = id;
            this.lat = lat;
            this.lon = lon;
            this.tags = tags;
            this.modified = false;
        }
        public final long getId() {return this.id;}
        public final float getLat() {return this.lat;}
        public final float getLon() {return this.lon;}
        public final void setId(long id) {this.id = id;}

        public final void updatePosition(float lat, float lon) {
            this.lat = lat;
            this.lon = lon;
            this.modified = true;
        }

        @Override public String toString() {return "#" + id + "(" + lat + "|" + lon + ")";}
    }

    public final static class QuickWay {
        private long id;
        public final long[] ndRefs;
        public final int[] tags;
        public final String name;
        public QuickWay(long id,
                long[] ndRefs, int[] tags, String name) {
            this.id = id;
            this.ndRefs = ndRefs;
            this.tags = tags;
            this.name = name;
        }
        public final long getId() {return this.id;}
        public final void setId(long id) {this.id = id;}
        @Override public String toString() {return "#" + id + "[" + ndRefs.length + "]";}
    }

    public final String getTag(int idx) {
        if (null == this.tagArr) return null;
        if (idx < 0 || idx > this.tagArr.length -1)
            throw new ArrayIndexOutOfBoundsException(idx);
        return this.tagArr[idx];
    }

    private final void registerElementTag(String key, String value) {
        String kv = key + "=" + value;
        Integer idx = this.tagMap.get(kv);
        if (null == idx) {
            idx = this.tagMap.size();
            this.tagMap.put(kv, idx);
        }
        this.tags[this.nTags++] = idx;
    }

    @Override
    public void open(Config config) {
        try {
            this.log = config.getLog();

            String beanPath = config.getProperty("parser.handler.beanFile",
                    config.getWorkDir() + "/quicknet.bean");
            this.beanFile = new File(beanPath);
            File dir = this.beanFile.getParentFile();
            if (dir != null && !dir.isDirectory()) {
                dir.mkdirs();
            }

            log.progress("creating " + this.beanFile.getName());

            String bean = config.getProperty("parser.handler.bean");
            if (bean != null) config.putBean(bean, new VarBean(this));

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void close() {
        // tagMap in Array umwandeln
        this.tagArr = new String[this.tagMap.size()];
        for (Entry<String, Integer> tag : this.tagMap.entrySet())
            this.tagArr[tag.getValue()] = tag.getKey(); // In Array wandeln
        this.tagMap.clear(); // Brauchen wir jetzt nicht mehr

        OutStream outStream = new OutStreamDisk(this.beanFile);
        new VarBean(this).writeToStream(outStream);
        outStream.close();
        log.info("Bean written to '" + this.beanFile.getAbsolutePath() +  "'");
    }

    @Override
    public void onNode(long id, double lat, double lon) {
        this.id = id;
        this.lat = (float) lat;
        this.lon = (float) lon;
        this.nTags = 0;
    }

    @Override
    public void onNodeAction(String action) {
    }

    @Override public void onNodeTag(String key, String value) {
        this.registerElementTag(key, value);
    }

    @Override public void onNodeComplete() {
        int[] tags = Arrays.copyOf(this.tags, this.nTags);
        QuickNode qn = new QuickNode(this.id, this.lat, this.lon, tags);
        this.nodeMap.put(this.id, qn);
    }


    @Override
    public void onWay(long id) {
        this.id = id;
        this.nNdRefs = 0;
        this.name = null;
        this.nTags = 0;
    }

    @Override public void onWayTag(String key, String value) {
        if ("name".equals(key)) {
            this.name = value;
        } else {
            this.registerElementTag(key, value);
        }
    }

    @Override
    public void onWayNdRef(long ndRefId) {
        if (!this.nodeMap.containsKey(ndRefId))
            throw new RuntimeException("Node not found #" + ndRefId);
        if (this.ndRefs.length == this.nNdRefs) // ArrayOverflow
            this.ndRefs = Arrays.copyOf(this.ndRefs, nNdRefs * 2);
        this.ndRefs[this.nNdRefs++] = ndRefId;
    }

    @Override public void onWayComplete() {
        long[] ndRefs = Arrays.copyOf(this.ndRefs, this.nNdRefs);
        int[] tags = Arrays.copyOf(this.tags, this.nTags);

        QuickWay qw = new QuickWay(this.id, ndRefs, tags, this.name);
        this.wayList.add(qw);
    }

    @Override public void onNodeObj(String key, Object value) {}
    @Override public void closeNodeProcessing() {}
    @Override public void onWayObj(String key, Object value) {}
    @Override public void closeWayProcessing() {}
    @Override public void onRelation(long id) {}
    @Override public void onRelationMember(String type, long ref, String role) {}
    @Override public void onRelationTag(String key, String value) {}
    @Override public void onRelationComplete() {}
    @Override public void closeRelationProcessing() {}

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.nodeMap.size());
        for (Entry<Long, QuickNode> entry : this.nodeMap.entrySet()) {
            long ndRef = entry.getKey();
            QuickNode qn = entry.getValue();

            outStream.writeLong(ndRef);
            outStream.writeLong(qn.id);
            outStream.writeInt(Float.floatToIntBits(qn.lat));
            outStream.writeInt(Float.floatToIntBits(qn.lon));

            int nTags = null == qn.tags ? 0 : qn.tags.length;
            outStream.writeInt(nTags);
            for (int i = 0; i < nTags; i++) outStream.writeInt(qn.tags[i]);
        }

        outStream.writeInt(this.wayList.size());
        for (QuickWay qw : this.wayList) {
            outStream.writeLong(qw.id);
            new VarString(qw.name).writeToStream(outStream);

            outStream.writeInt(qw.ndRefs.length);
            for (long ndRef : qw.ndRefs) outStream.writeLong(ndRef);

            int nTags = null == qw.tags ? 0 : qw.tags.length;
            outStream.writeInt(nTags);
            for (int i = 0; i < nTags; i++) outStream.writeInt(qw.tags[i]);
        }

        outStream.writeInt(this.tagArr.length);
        for (String tag : this.tagArr) new VarString(tag).writeToStream(outStream);
    }

    @Override
    public QuickNet readFromStream(InStream inStream) {

        int nNodes = inStream.readInt();
        this.nodeMap.clear();
        this.nodeMap = new LinkedHashMap<Long, QuickNode>(nNodes);

        for (int i = 0; i < nNodes; i++) {
            long ndRef = inStream.readLong(); // ID referenziert vom Way
            long id = inStream.readLong(); // Interne, abweichende ID des Nodes
            float lat = Float.intBitsToFloat(inStream.readInt());
            float lon = Float.intBitsToFloat(inStream.readInt());

            int[] tags = new int[inStream.readInt()];
            if (0 == tags.length) tags = null;
            else for (int j = 0; j < tags.length; j++) tags[j] = inStream.readInt();

            this.nodeMap.put(ndRef, new QuickNode(id, lat, lon, tags));
        }

        int nWays = inStream.readInt();
        this.wayList.clear();
        this.wayList = new ArrayList<QuickWay>(nWays);

        for (int i = 0; i < nWays; i++) {
            long id = inStream.readLong();
            String name = new VarString().readFromStream(inStream).toString();
            if (0 == name.length()) name = null;

            long[] refs = new long[inStream.readInt()];
            if (0 == refs.length) refs = null; // Gaga, aber konsistent
            else for (int j = 0; j < refs.length; j++) refs[j] = inStream.readLong();

            int[] tags = new int[inStream.readInt()];
            if (0 == tags.length) tags = null;
            else for (int j = 0; j < tags.length; j++) tags[j] = inStream.readInt();

            QuickWay qw = new QuickWay(id, refs, tags, name);
            this.wayList.add(qw);
        }

        this.tagArr = new String[inStream.readInt()];
        for (int i = 0; i < this.tagArr.length; i++)
            this.tagArr[i] = new VarString().readFromStream(inStream).toString();

        return this;
    }


    public final QuickNet renumberIds() {
        renumberWayIds(1, 1);
        renumberNodeIds(1, 1);
        return this;
    }

    public final QuickNet renumberWayIds(final long firstId, final long step) {
        long nextId = firstId;
        for (QuickWay w : this.wayList) {
            w.id = nextId;
            nextId += step;
        }
        return this;
    }

    public final QuickNet renumberNodeIds(final long firstId, final long step) {
        long nextId = firstId;
        for (QuickNode n : nodes()) {
            n.id = nextId; // Jetzt ID im Knoten selbst neu setzen
            nextId += step;
        }
        for (QuickWay w : wayList) for (int i = 0; i  < w.ndRefs.length; i++) {
            w.ndRefs[i] = node(w.ndRefs[i]).id;
        }
        LinkedHashMap<Long, QuickNode> newNodeMap = new LinkedHashMap<>(this.nodeMap.size());
        for (Entry<Long, QuickNode> entry : this.nodeMap.entrySet()) {
            newNodeMap.put(entry.getValue().id, entry.getValue()); // Neue Lookup-Map schreiben
        }
        this.nodeMap.clear();
        this.nodeMap = newNodeMap; // Lookup-Map neu setzen

        return this;
    }

    public final QuickNet markOrphanedNodes() {
        LongSet nodeIdSet = new LongSet(this.nodes().size());
        for (QuickWay w : ways()) for (long ndRef : w.ndRefs) nodeIdSet.add(ndRef);
        for (QuickNode n : nodes()) n.orphaned = !nodeIdSet.contains(n.id);
        nodeIdSet.clear();
        return this;
    }

    public final String toGeoJson() {
        LatLons latLons = new LatLons();
        Properties tags = new Properties();

        FeatureCollection featureCollection = new GeoJsonBuilder().featureCollection();
        for (QuickWay way : this.ways()) {
            latLons.clear();
            tags.clear();

            // Geometry erstellen
            for (long ref : way.ndRefs) {
                QuickNode node = this.node(ref);
                latLons.append(node.lat, node.lon);
            }
            Feature feature = featureCollection.addLine(latLons);
            feature.with("id", way.id).with("name", way.name).with("refs", way.ndRefs);

            if (way.tags != null && way.tags.length > 0) {
                // Tags als Properties erstellen
                for (int x : way.tags) {
                    String[] kv = this.getTag(x).split("=");
                    tags.setProperty(kv[0], kv[1]);
                }
                feature.with("tags", tags);
            }
        }

        for (QuickNode node : this.nodes()) {
            tags.clear();
            Feature feature = featureCollection.addPoint(new LatLon(node.lat, node.lon));
            feature.with("id", node.id);
            if (node.tags != null && node.tags.length > 0) {
                // Tags als Properties erstellen
                for (int x : node.tags) {
                    String[] kv = this.getTag(x).split("=");
                    tags.setProperty(kv[0], kv[1]);
                }
                feature.with("tags", tags);
                if (node.orphaned) feature.with("orphaned", true);
            }
        }

        final String geoJson = featureCollection.build();
        return geoJson;
    }

    public final void writeToOsmFile(final File osmFile) {
        try {
            // "action='modify' visible='true'" wird von JOSM gesetzt, wenn modifiziert;
            final String V = "version='1'";
            final Writer xml = new PrintWriter(osmFile);

            xml.append("<?xml version='1.0' encoding='UTF-8'?>" // upload=true needed to satisfy JOSM?
                    + "\n<osm version='0.6'" + /*" upload='true'" +*/ " generator='OSM2PO'>");

            for (QuickNode n : this.nodes()) {
                xml.write("\n  <node id='" + n.id + "' " + V + " lat='" + n.lat + "' lon='" + n.lon + "'");
                if (n.modified) xml.write(" action='modify'");
                if (null == n.tags || 0 == n.tags.length) xml.write(" />");
                else {
                    xml.write(">");
                    for (int x : n.tags) {
                        String[] kv = this.getTag(x).split("=");
                        xml.write("\n    <tag k='" + stringToXmlUTF8(kv[0]) + "' v='" + stringToXmlUTF8(kv[1]) + "' />");
                    }
                    xml.write("\n  </node>");
                }
            }

            for (QuickWay w : this.ways()) {
                xml.write("\n  <way id='" + w.id + "' " + V + ">");
                for (long r : w.ndRefs) {
                    xml.write("\n    <nd ref='" + r + "' />");
                }
                if (w.name != null) xml.write("\n    <tag k='name' v='" + stringToXmlUTF8(w.name) + "' />");
                if (w.tags != null) for (int x : w.tags) {
                    String[] kv = this.getTag(x).split("=");
                    xml.write("\n    <tag k='" + stringToXmlUTF8(kv[0]) + "' v='" + stringToXmlUTF8(kv[1]) + "' />");
                }
                xml.write("\n  </way>");
            }

            xml.write("\n</osm>");
            xml.close();

        } catch (Exception e) {
            this.log.error(e);
        }
    }

} // EOC
