/*
 * $Id: SimpleDbNetParser.java 5489 2021-12-18 22:16:40Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.parser;

import static de.cm.osm2po.misc.Utils.hexToBytes;
import static de.cm.osm2po.misc.WkbUtils.fromWkbPoint;

import java.io.InputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.tiler.TagResolver;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.primitives.LongArray;

/**
 * Ein {@link OsmParser}, der ein simples Netz aus einer Knoten- und einer
 * Kanten-Tabelle erstellt.<br>
 * Die Knotenkoordinaten definieren hier implizit die Geometrie, da Kanten hier
 * lediglich auf die Knoten-IDs per Start- und Ziel (source, target) verweisen.<br>
 * Die Datenbank selbst wird per {@link Config#openDbConnection(String)} verbunden.<br>
 * Die Namen der Spalten beider Tabellen sind beliebig, es muessen aber folgende
 * Spalten per SQL der Reihe nach zur Verfuegung gestellt werden:<br><br>
 * <code><strong>Nodes:</strong> id bigint, geom geometry(point, 4326)</code><br>
 * <code><strong>Edges:</strong> id bigint, source bigint, target bigint</code><br>
 * <br><strong>Wichtig:</strong><br>
 * Nachfolgende Spalten werden als Spalten-Name und -Wert an den entsprechenden
 * {@link TagResolver#onTag(String, String)} mit {@link Object#toString()} und
 * zusaetzlich an {@link TagResolver#onObj(String, Object)} weitergereicht.<br>
 * Per Aliasing im SQL-Select lassen sich so leicht Spalten auf Tags projezieren.<br>
 * Parameter (Defaults in Klammern):
 * <ul>
 * <li><strong>parser.sql.nodes</strong> : (SELECT id, geom FROM nodes)</li>
 * <li><strong>parser.sql.edges</strong> : (SELECT id, source, target FROM edges)</li>
 * <li><strong>parser.cartesian</strong> : edge-SQL ignorieren und stattdessen aus Nodes ableiten</li>
 * <li><strong>parser.db</strong> : Prefix fuer {@link Config#openDbConnection(String)} (db)</li>
 * <li><strong>parser.tag.nodes</strong> : CSV von zusaetzlichen Node-Tags der Form k/v</li>
 * <li><strong>parser.tag.edges</strong> : CSV von zusaetzlichen Edge-Tags der Form k/v</li>
 * </ul>
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de *
 */
public class SimpleDbNetParser implements OsmParser {
    public final static String PROP = "parser.";

    private Log log;
    private OsmParserHandler handler;
    private Connection dbCon;
    private String sqlNodes, sqlEdges;

    private boolean cartesian;
    private LongArray nodeIds;

    private Map<String, String> edgeTagsMap;
    private Map<String, String> nodeTagsMap;

    @Override // Parser InputStream will be ignored
    public void open(InputStream inputStream,
            OsmParserHandler osmParserHandler, Config config) {
        this.log = config.getLog();
        this.handler = osmParserHandler;
        this.nodeTagsMap = new HashMap<String, String>();
        this.edgeTagsMap = new HashMap<String, String>();

        this.cartesian = Boolean.valueOf(config.getProperty(PROP + "cartesian", "false"));
        this.nodeIds = new LongArray();

        this.sqlNodes = config.getProperty(PROP + "sql.nodes",
                "SELECT id, geom FROM nodes");
        this.sqlEdges = config.getProperty(PROP + "sql.edges",
                "SELECT id, source, target FROM edges");

        String edgeTags = config.getProperty(PROP + "tag.edges");
        if (edgeTags != null) for (String tag : edgeTags.split(",")) {
            String[] kv = tag.split("/");
            if (kv.length == 2) this.edgeTagsMap.put(kv[0].trim(), kv[1].trim());
            else this.log.warn("Use '/' to separate key and value of '" + kv + "'");
        }
        String nodeTags = config.getProperty(PROP + "tag.nodes");
        if (nodeTags != null) for (String tag : nodeTags.split(",")) {
            String[] kv = tag.split("/");
            if (kv.length == 2) this.nodeTagsMap.put(kv[0].trim(), kv[1].trim());
            else this.log.warn("Use '/' to separate key and value of '" + kv + "'");
        }

        try {
            String dbPrefix = config.getProperty(PROP + "db", "db");
            this.dbCon = config.openDbConnection(dbPrefix);
            this.dbCon.setAutoCommit(false);

        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    @Override // Parser
    public void close() {
        try {this.dbCon.close();} catch (Throwable t) {log.fatal(t);}

    }

    @Override // Parser
    public void parse() {
        try {
            this.parseNodes();
            this.handler.closeNodeProcessing();
            if (this.cartesian) this.buildCartesianEdges(); else this.parseEdges();
            this.handler.closeWayProcessing();

        } catch (Throwable e) {
            throw new RuntimeException(e);
        }
    }

    private void parseNodes() throws Exception {
        Statement stm = this.dbCon.createStatement();
        stm.setFetchSize(0x400);
        ResultSet rst = stm.executeQuery(this.sqlNodes);
        int nCols = rst.getMetaData().getColumnCount();

        log.debug("Parsing '" + this.sqlNodes + "'");

        while (rst.next()) {
            long id = rst.getLong(1);
            String geomHex = rst.getString(2);
            LatLon point = fromWkbPoint(hexToBytes(geomHex));
            this.handler.onNode(id, point.getLat(), point.getLon());

            for (int col = 3; col <= nCols; col++) {
                Object colVal = rst.getObject(col);
                if (colVal != null) {
                    String colName = rst.getMetaData().getColumnName(col);
                    this.handler.onNodeTag(colName, colVal.toString());
                    this.handler.onNodeObj(colName, colVal);
                }
            }

            for (Entry<String, String> tag : this.nodeTagsMap.entrySet())
                handler.onNodeTag(tag.getKey(), tag.getValue());

            this.handler.onNodeComplete();

            if (this.cartesian) this.nodeIds.push(id);

        }
        rst.close();
        stm.close();
    }

    private void buildCartesianEdges() {
        long nNodes = this.nodeIds.getSize();
        log.debug("Building undirected cartesian product of edges for " + nNodes + " nodes.");

        long wayId = 1;
        for (long i = 0; i < nNodes; i++) {
            long fromNodeId = this.nodeIds.get(i);
            for (long j = i + 1; j < nNodes; j++) {
                long toNodeId = this.nodeIds.get(j);

                this.handler.onWay(wayId++);
                this.handler.onWayNdRef(fromNodeId);
                this.handler.onWayNdRef(toNodeId);
                for (Entry<String, String> tag : this.edgeTagsMap.entrySet())
                    handler.onWayTag(tag.getKey(), tag.getValue());
                this.handler.onWayComplete();
            }
        }
    }

    private void parseEdges() throws Exception {
        Statement stm = this.dbCon.createStatement();
        stm.setFetchSize(0x400);
        ResultSet rst = stm.executeQuery(this.sqlEdges);
        int nCols = rst.getMetaData().getColumnCount();

        log.debug("Parsing '" + this.sqlEdges + "'");

        while (rst.next()) {
            this.handler.onWay(rst.getLong(1));
            this.handler.onWayNdRef(rst.getLong(2));
            this.handler.onWayNdRef(rst.getLong(3));

            for (int col = 4; col <= nCols; col++) {
                Object colVal = rst.getObject(col);
                if (colVal != null) {
                    String colName = rst.getMetaData().getColumnName(col);
                    this.handler.onWayTag(colName, colVal.toString());
                    this.handler.onWayObj(colName, colVal);
                }
            }

            for (Entry<String, String> tag : this.edgeTagsMap.entrySet())
                handler.onWayTag(tag.getKey(), tag.getValue());

            this.handler.onWayComplete();
        }
        rst.close();
        stm.close();
    }

} // EOC
