/*
 * $Id: VertexHashMap.java 5451 2021-11-26 20:45:59Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.segmenter;

import static de.cm.osm2po.primitives.BigLongLongHashMap.NULLVALUE;

import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.primitives.BigLongLongHashMap;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessMemory;

/**
 * Dies ist eine extrem speicherfreundliche Hashmap, die
 * fuer das Vertex-Reference-Counting angepasst ist.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */

public class VertexHashMap {
    // TODO NULLOBJECTS aus Speichergruenden in eigenem Set halten
    public static final long NULLOBJECT = BigLongLongHashMap.NULLVALUE +1;

    private RandomAccess cache;
    private BigLongLongHashMap map;

    public VertexHashMap() {this(0x100000);}

    public VertexHashMap(int hashSize) {
        this.cache = new RandomAccessMemory();
        this.map = new BigLongLongHashMap(hashSize);
    }

    public void clear() {
        this.cache.clear();
        this.map.clear();
    }

    public boolean isEmpty() {
        return this.map.getSize() == 0;
    }

    public LongArray getKeys() {
        // RawKeys sind die gleichen, da hier kein remove()
        return this.map.getRawKeys();
    }

    public int getCollisionsMax() {
        return this.map.getCollisionsMax();
    }

    public double getCollisionsAvg() {
        return this.map.getCollisionsAvg();
    }

    // TODO toString() implementieren

    /**
     * Schreibt ein {@link Vertex} in die Map.
     * Zusaetzlich wird {@link Vertex#setReadIndex(long)}
     * gesetzt, so dass ein erneutes Schreiben des selben
     * Objekts mittels {@link #override(Vertex)} durchgefuehrt
     * werden kann und nicht mehr ueber {@link #put(long, Vertex)}
     * laufen muss - schneller.<br>
     * Wird <code>null</code> uebergeben, so wird der Wert
     * {@link #NULLOBJECT} in die interne Map geschrieben. Die Wirkung
     * entspricht damit einem Set ohne Speicherverbrauch, da interne
     * Members nicht geschrieben werden, weil ja initial und rekonstruierbar.
     * @param key long Schluessel.
     * @param vertex {@link Vertex}-Objekt.
     * @see #get(long)
     */
    public void put(long key, Vertex vertex) {
        if (null == vertex) {
            this.map.put(key, NULLOBJECT);
        } else {
            long writeIndex = this.cache.getWriteIndex();
            this.map.put(key, writeIndex);
            vertex.writeToStream(this.cache);
            vertex.setReadIndex(writeIndex);
        }
    }

    /**
     * Liefert ein neues {@link Vertex}-Objekt aus der Map.
     * Beim Lesen wird zusaetzlich der Index im Objekt gespeichert,
     * so dass {@link #override(Vertex)} diesen direkt zurueckschreiben
     * kann.<br>
     * Handelt es sich um ein <code>null</code>-Objekt,
     * das mittels {@link #put(long, Vertex)} eingefuegt
     * wurde, so wird ein initiales Objekt zurueckgeliefert.
     * @param key long Schluessel.
     * @return <code>new</code> {@link Vertex} oder NULL, wenn nicht gefunden.
     * @see #put(long, Vertex)
     *
     */
    public Vertex get(long key) {
        long idx = this.map.get(key);
        Vertex vertex = null;
        if (NULLVALUE != idx) {
            vertex = new Vertex();
            if (idx != NULLOBJECT) {
                this.cache.setReadIndex(idx);
                vertex.readFromStream(this.cache);
                vertex.setReadIndex(idx);
            }
        }
        return vertex;
    }

    public boolean contains(long key) {
        long idx = this.map.get(key);
        return (idx != NULLVALUE);
    }

    public boolean isNull(long key) {
        long idx = this.map.get(key);
        return (idx == NULLOBJECT);
    }

    /**
     * Ueberschreibt einen bereits existierenden {@link Vertex}.
     * Ist bei einfachen Attribut-Aenderungen gegenueber
     * {@link #put(long, Vertex)} schneller und speicherfreundlicher.<br>
     * <strong>Wichtig</strong>: Das Objekt muss zuvor mittels
     * {@link #get(long)} entnommen worden sein. Neu einzufuegende Objekte
     * also immer mittels {@link #put(long, Vertex)}
     * @param vertex {@link Vertex} NOT NULL.
     */
    public void override(Vertex vertex) {
        long savedWriteIndex = this.cache.getWriteIndex();
        this.cache.setWriteIndex(vertex.getReadIndex());
        vertex.writeToStream(this.cache);
        this.cache.setWriteIndex(savedWriteIndex);
    }

}
