/*
 * $Id: DefaultRelationTagResolver.java 5788 2023-04-11 06:18:30Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.cm.osm2po.Config;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.primitives.Var;

/**
 * Registriert eine Menge von OSM-Tags einer Relation.
 * und verdichtet diese auf die fuer ein Routing
 * noetigen Informationen.<br>
 * Restrictions, Boundaries und Routes sind hier hart kodiert.
 * Polygone werden nur grob in clazz kategoriesiert, da es
 * theoretisch unendlich viele Kombinationen gibt. Daher werden
 * fuer selbige zusaetzlich die Flags benutzerdefiniert geschrieben,
 * weshalb die Werte hier aus der Konfiguration stammen muessen.
 * @author (c) 2012/2017 - Carsten Moeller - info@osm2po
 * @see TagResolver
 */
public class DefaultRelationTagResolver implements RelationTagResolver {
    //TODO Neue OSM-Tags wie z.B. restriction:motorcar dynamisch behandeln.
    //TODO Restriktonen-Zusatz-Tag 'except' dynamisch interpretieren

    //TODO zwischen polygon-clazz und -flags unterscheiden
    public final static String PROP_POLYGON_PREFIX = "rtr.support.polygon.type.";

    private final static byte TYPE_NULL = 0;
    private final static byte TYPE_ROUTE = 1;
    private final static byte TYPE_ROUTE_MASTER = 2;
    private final static byte TYPE_RESTRICTION = 3;
    private final static byte TYPE_BOUNDARY = 4;
    private final static byte TYPE_POLYGON = 5;

    // Subtypen koennen mehrfach vorkommen, daher binaer 1, 2 ,4, ...
    private final static byte SUBTYPE_BOUNDARY = 1; // z.B. HH mit Wattenmeer
    private final static byte SUBTYPE_LANDAREA = 2; // z.B. HH ohne Wattenmeer
    private final static byte SUBTYPE_POSTAL = 4; // Postleitzahl

    private final static int VALUE_NULL = 0;

    private byte relType;
    private byte subType;
    private byte clazz;
    private int flags;
    private int value;
    private String name;
    private String code;

    private boolean excepted;

    private Set<Byte> excludeClazzes = new HashSet<Byte>();
    private String[] langNameKeys = null;

    private Map<String, Byte> mainTypeMap, routeClazzMap, restrictionClazzMap;
    private Map<String, Integer> polygonTagMap;

    private int[] polygonTags;
    private int nPolygonTags;

    private int nextBetterNameKeyIndex;
    private boolean supportPolygons;
    private boolean supportBoundaries;
    private boolean supportRestrictions;
    private boolean supportRoutes;

    @Override public void onObj(String key, Object value) {}
    @Override public void close() {}
    @Override public String getName() {return this.name;}
    @Override public int getFlags() {return this.flags;}
    @Override public byte getClazz() {return this.clazz;}

    @Override
    public void open(Config config) {
        this.supportRestrictions = Boolean.valueOf(
                config.getProperty("rtr.support.restrictions", "true"));
        this.supportBoundaries = Boolean.valueOf(
                config.getProperty("rtr.support.boundaries", "false"));

        String[] routeTypes = config.getProperty(
                "rtr.support.route.types", "").split(",");
        for (int i = 0; i < routeTypes.length; i++)
            routeTypes[i] = routeTypes[i].toLowerCase().trim();
        Set<String> routeTypesSet = new HashSet<String>();
        for (String routeType : routeTypes) routeTypesSet.add(routeType);

        this.polygonTagMap = new HashMap<String, Integer>();
        this.polygonTags = new int[0x100];
        this.initCustomPolygonTags(config.getProps());
        this.supportPolygons = !this.polygonTagMap.isEmpty();

        this.mainTypeMap = new HashMap<String, Byte>();
        this.mainTypeMap.put("route", TYPE_ROUTE);
        this.mainTypeMap.put("route_master", TYPE_ROUTE_MASTER);
        this.mainTypeMap.put("restriction", TYPE_RESTRICTION);
        this.mainTypeMap.put("restriction:motorcar", TYPE_RESTRICTION);
        this.mainTypeMap.put("boundary", TYPE_BOUNDARY);
        this.mainTypeMap.put("land_area", TYPE_BOUNDARY);
        this.mainTypeMap.put("multipolygon", TYPE_POLYGON);

        this.routeClazzMap = new HashMap<String, Byte>();
        if (routeTypesSet.contains("road"))
            this.routeClazzMap.put("road", CLAZZ_ROUTE_ROAD);
        if (routeTypesSet.contains("hiking"))
            this.routeClazzMap.put("hiking", CLAZZ_ROUTE_HIKING);
        if (routeTypesSet.contains("bicycle"))
            this.routeClazzMap.put("bicycle", CLAZZ_ROUTE_BICYCLE);
        if (routeTypesSet.contains("mtb"))
            this.routeClazzMap.put("mtb", CLAZZ_ROUTE_MTB);
        if (routeTypesSet.contains("ski"))
            this.routeClazzMap.put("ski", CLAZZ_ROUTE_SKI);
        if (routeTypesSet.contains("horse"))
            this.routeClazzMap.put("horse", CLAZZ_ROUTE_HORSE);
        if (routeTypesSet.contains("bus"))
            this.routeClazzMap.put("bus", CLAZZ_ROUTE_BUS);
        if (routeTypesSet.contains("train"))
            this.routeClazzMap.put("train", CLAZZ_ROUTE_TRAIN);
        if (routeTypesSet.contains("railway"))
            this.routeClazzMap.put("railway", CLAZZ_ROUTE_RAILWAY);
        if (routeTypesSet.contains("subway"))
            this.routeClazzMap.put("subway", CLAZZ_ROUTE_SUBWAY);
        if (routeTypesSet.contains("tram"))
            this.routeClazzMap.put("tram", CLAZZ_ROUTE_TRAM);
        if (routeTypesSet.contains("ferry"))
            this.routeClazzMap.put("ferry", CLAZZ_ROUTE_FERRY);
        if (routeTypesSet.contains("shuttle_train"))
            this.routeClazzMap.put("shuttle_train", CLAZZ_ROUTE_SHUTTLETRAIN);
        this.supportRoutes = !this.routeClazzMap.isEmpty();

        this.restrictionClazzMap = new HashMap<String, Byte>();
        if (this.supportRestrictions) {
            this.restrictionClazzMap.put("no_left_turn", CLAZZ_RESTRICTION_NO_LEFT_TURN);
            this.restrictionClazzMap.put("no_right_turn", CLAZZ_RESTRICTION_NO_RIGHT_TURN);
            this.restrictionClazzMap.put("no_straight_on", CLAZZ_RESTRICTION_NO_STRAIGHT_ON);
            this.restrictionClazzMap.put("no_u_turn", CLAZZ_RESTRICTION_NO_U_TURN);
            this.restrictionClazzMap.put("only_left_turn", CLAZZ_RESTRICTION_ONLY_LEFT_TURN);
            this.restrictionClazzMap.put("only_right_turn", CLAZZ_RESTRICTION_ONLY_RIGHT_TURN);
            this.restrictionClazzMap.put("only_straight_on", CLAZZ_RESTRICTION_ONLY_STRAIGHT_ON);
            this.restrictionClazzMap.put("no_entry", CLAZZ_RESTRICTION_NO_ENTRY);
            this.restrictionClazzMap.put("no_exit", CLAZZ_RESTRICTION_NO_EXIT);
        }

        String strExcludeClazzes = config.getProperty("rtr.exclude.clazzes");
        if (strExcludeClazzes != null) {
            String clazzes[] = strExcludeClazzes.split(",");
            for (int i = 0; i < clazzes.length; i++) {
                this.excludeClazzes.add(Byte.valueOf(clazzes[i].trim()));
            }
        }

        String strLangCodes = config.getProperty("rtr.preferredLangCodes");
        int nLangNameKeys = 0;
        this.langNameKeys = new String[100];
        this.langNameKeys[nLangNameKeys++] = "name"; // Default
        if (strLangCodes != null) for (String code : strLangCodes.split(",")) {
            code = code.trim().toLowerCase();
            if ("" != code) this.langNameKeys[nLangNameKeys++] = "name:" + code;
        }
        this.langNameKeys = Arrays.copyOf(this.langNameKeys, nLangNameKeys);

    }

    @Override
    public void onTags(long entityId) {
        this.clazz = CLAZZ_NULL;
        this.flags = FLAGS_NULL;
        this.relType = TYPE_NULL;
        this.subType = TYPE_NULL;
        this.value = VALUE_NULL;
        this.name = null;
        this.nextBetterNameKeyIndex = 0;
        this.code = null;
        this.excepted = false;
        this.nPolygonTags = 0;
    }

    @Override
    public void onTag(String key, String value) {
        if (key.startsWith("name")) {
            for (int i = this.nextBetterNameKeyIndex; i < this.langNameKeys.length; i++) {
                if (key.equals(this.langNameKeys[i])) {
                    this.name = value;
                    this.nextBetterNameKeyIndex = i + 1;
                    break;
                }
            }
            return;
        }

        if ("type".equals(key)) { // Main-Relation-Type
            Byte relType = this.mainTypeMap.get(value);
            if (relType != null) this.relType = relType;
            return;
        }

        if (this.supportRoutes) {
            if ("route".equals(key)) {
                Byte routeType = this.routeClazzMap.get(value);
                if (routeType != null) this.clazz = routeType;
                return;
            }

            // Super-Relation, e.g.:
            // ref = Sylt Shuttle, type = route_master, route_master = shuttle_train, shuttle = yes
            if ("route_master".equals(key)) {
                Byte routeType = this.routeClazzMap.get(value);
                if (routeType != null) {
                    this.subType = routeType;
                }
                return;
            }
        }

        if (this.supportRestrictions) {
            if ("restriction".equals(key)) {
                Byte restrictionType = this.restrictionClazzMap.get(value);
                if (restrictionType != null) this.clazz = restrictionType;
                return;
            }
            if ("except".equals(key)) {
                this.excepted = value.contains("motorcar");
                return;
            }
        }

        if (this.supportBoundaries) {
            if ("boundary".equals(key)) {
                if ("administrative".equals(value)) {
                    this.subType |= SUBTYPE_BOUNDARY;
                } else if ("postal_code".equals(value)) {
                    this.subType |= SUBTYPE_POSTAL;
                }
            } else if ("land_area".equals(key)) {
                if ("administrative".equals(value)) {
                    this.subType |= SUBTYPE_LANDAREA;
                }
            } else if ("admin_level".equals(key)) {
                this.value = this.parseInt(value);
            } else if ("postal_code_level".equals(key)) {
                this.value = this.parseInt(value);
            } else if ("postal_code".equals(key)) {
                this.code = value;
            }
        }

        if (this.supportPolygons) {
            Integer itag = this.polygonTagMap.get(key + "/" + value);
            if (itag != null) {
                this.polygonTags[this.nPolygonTags++] = itag;
            }
        }

    }

    private int parseInt(String s) {
        try {return Integer.parseInt(s);}
        catch (Exception e) {return 0;}
    }

    @Override
    public void onTagsComplete() {
        if (TYPE_RESTRICTION == this.relType) {
            // Benamste Restriktion ist unsinnig. Speicher sparen
            this.name = null;

        } else if (TYPE_ROUTE_MASTER == this.relType) {
            if (this.subType != TYPE_NULL) {
                this.clazz = subType;
            }

        } else if (TYPE_BOUNDARY == this.relType || TYPE_POLYGON == this.relType) {
            // Komplexe Boundary-Logik im Nachgang interpretieren
            // Achtung!
            // Aeltere Boundaries wurden frueher als MultiPolygone getaggt.

            // http://wiki.openstreetmap.org/wiki/DE:Grenze_zeichnen
            // http://wiki.openstreetmap.org/wiki/Tag:boundary%3Dpostal_code

            if (SUBTYPE_BOUNDARY == this.subType) {
                switch (this.value) {
                case 10: this.clazz = CLAZZ_BOUNDARY_ADMIN_SUBURB; break;
                case 8: this.clazz = CLAZZ_BOUNDARY_ADMIN_TOWN; break;
                case 6: this.clazz = CLAZZ_BOUNDARY_ADMIN_METROPOLE; break;
                case 5: this.clazz = CLAZZ_BOUNDARY_ADMIN_DISTRICT; break;
                case 4: this.clazz = CLAZZ_BOUNDARY_ADMIN_STATE; break;
                case 2: this.clazz = CLAZZ_BOUNDARY_ADMIN_COUNTRY; break;
                }
            } else if (SUBTYPE_LANDAREA == this.subType) {
                switch (this.value) {
                case 4: this.clazz = CLAZZ_BOUNDARY_LANDAREA_STATE; break;
                case 2: this.clazz = CLAZZ_BOUNDARY_LANDAREA_COUNTRY; break;
                }
            } else if ((SUBTYPE_BOUNDARY | SUBTYPE_LANDAREA) == this.subType) {
                switch (this.value) {
                case 4: this.clazz = CLAZZ_BOUNDARY_ADMIN_STATE_LANDAREA; break;
                case 2: this.clazz = CLAZZ_BOUNDARY_ADMIN_COUNTRY_LANDAREA; break;
                }
            } else if (SUBTYPE_POSTAL == this.subType) {
                if (VALUE_NULL == this.value || 8 == this.value) {
                    this.clazz = CLAZZ_BOUNDARY_POSTAL;
                    this.name = this.code;
                }
            } else if ((SUBTYPE_POSTAL | SUBTYPE_BOUNDARY) == this.subType) {
                if (VALUE_NULL == this.value || 8 == this.value) {
                    this.clazz = CLAZZ_BOUNDARY_ADMIN_TOWN_POSTAL;
                    this.name = this.code + " " + this.name;
                }

            // Wenn kein Boundary-Polygon, dann echtes Polygon
            } else if (TYPE_POLYGON == this.relType) {
                if (this.nPolygonTags > 0) {
                    int[] arr = Arrays.copyOf(this.polygonTags, this.nPolygonTags);
                    String key = this.makeKeySet(arr);
                    Integer val = this.polygonTagMap.get(key);
                    if (val != null) {
                        this.clazz = CLAZZ_POLYGON;
                        this.flags = val;
                    }
                }
            }
        }

        // Klassen per Config explizit ausgeschlossen
        if (this.excludeClazzes != null
                && this.excludeClazzes.contains(this.clazz))
            this.clazz = CLAZZ_NULL;
    }

    @Override
    public boolean isValid() {
        byte clazzType = (byte) (CLAZZ_MASK & this.clazz);
        switch (clazzType) {
        case CLAZZ_RESTRICTION: return TYPE_RESTRICTION == this.relType & !this.excepted;
        case CLAZZ_ROUTE: return TYPE_ROUTE == this.relType || TYPE_ROUTE_MASTER == this.relType;
        case CLAZZ_BOUNDARY: return TYPE_BOUNDARY == this.relType || TYPE_POLYGON == this.relType;
        case CLAZZ_POLYGON: return TYPE_POLYGON == this.relType;
        default: return false;
        }
    }

    @Override
    public Var<?> getCustom() {
        return null; // Hopefully never used
    }

    /**
     * Baut die Informationen fuer Polygon-Tags der Form
     * natural=wetland, wetland=tidalflat.
     * @param props {@link Props}
     */
    private void initCustomPolygonTags(Props props) {
        // TODO Modifiziert analog SimpleNodeTagResoler (tagSet mittel ;) - ungetestet!
        int[] idxs = new int[0x100];
        int nextIdx = 1;
        for (String propKey : props) {
            if (propKey.startsWith(PROP_POLYGON_PREFIX)) {
                int clazz = Integer.valueOf(propKey.substring(PROP_POLYGON_PREFIX.length()));
                String propValue = props.getProperty(propKey);
                String[] tagSets = propValue.split(";");
                for (String tagSet : tagSets) {
                    String[] tags = tagSet.split(",");
                    int nTags = tags.length;
                    for (int i = 0; i < nTags; i++) {
                        String key = tags[i].trim();
                        Integer idx = this.polygonTagMap.get(key);
                        if (null == idx) {
                            idx = nextIdx++;
                            this.polygonTagMap.put(key, idx);
                        }
                        idxs[i] = idx;
                    }
                    String key = this.makeKeySet(Arrays.copyOf(idxs, nTags));
                    this.polygonTagMap.put(key, clazz);
                }
            }
        }
    }

    private String makeKeySet(int[] arr) {
        String s = "";
        arr = intArrayDistinct(arr);
        for (int i = 0; i < arr.length; i++)  s += "@" + arr[i];
        return s;
    }

}
