/*
 * $Id: DefaultWayTagResolver.java 5230 2020-12-21 16:55:23Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import java.util.HashMap;
import java.util.Map;

import de.cm.osm2po.Config;
import de.cm.osm2po.primitives.Var;

/**
 * Registriert eine Menge von OSM-Tags eines Ways und verdichtet sie
 * auf die fuer ein Routing noetigen Informationen.<br>
 * Verdichten bedeutet, dass die Daten in eine flache Struktur gepresst
 * werden, die im Kern den Typ <strong>clazz</strong>, die Zusatzinfos
 * <strong>flags</strong> und weitere Infos wie z.B. die Geschwindigkeit
 * umfassen.<br>
 * Folgende Parameter werden u.a. unterstuetzt (Defaults in Klammern):
 * <ul>
 * <li><strong>wtr.trustValid</strong>:
 *      Alle Wege als valide betrachten (false)<br>
 *      Siehe auch {@link #isValid()}<br>
 * </li>
 * <li><strong>wtr.flagList</strong>: Deklaration der Bit-Bezeichner (binaer),
 *      z.B. <code>car,bike,foot</code>, hier also car=1, bike=2, foot=4, usw.<br>
 *      Intern ist dies ein Integer, d.h. es koennen maximal 32 Bit-Bezeichner
 *      frei vergeben werden.<br>
 *      Alternativ kann der resultierender Binaerwert auch direkt, also ohne
 *      wtr.flagList gesetzt werden. Intern wird {@link Integer#decode(String)}
 *      gerufen, was auch eine 0x-Notation (hex) zulaesst.
 * </li>
 * <li><strong>wtr.finalMask</strong>:
 *      Valide, wenn eine der flag-Bit-Masken ubereinstimmt. z.B. car,bike.
 *      Hier also wenn ein Weg fuer Auto oder Fahrrad zulaessig ist. (optional)<br>
 *      Siehe auch {@link #isValid()}
 * </li>
 * <li><strong>wtr.polyMask</strong>:
 *      Bitmaske als Indikator fuer Wege, die Polygone, also Flaechen,
 *      Gebaeude und so weiter, darstellen und vom Strassennetz
 *      entkoppelt werden muessen. (optional)
 * </li>
 * <li><strong>wtr.maxSpeedTags</strong>:
 *      Wenn gesetzt, werden die Standard-Werte fuer die Geschwindigkeit ueberschrieben.
 *      Normalerweise reicht es hier <code>wtr.maxSpeedTags=maxspeed</code> zu setzen.
 *      In bestimmten Faellen kann es aber sinnvoll sein, weitere Optionen zu
 *      ergaenzen - z.B. <code>maxspeed, maxspeed:hgv</code> etc.
 * </li>
 * <li><strong>wtr.tag.key.value</strong>:
 *      z.B. <code>wtr.tag.highway.secondary=1,5,50,car</code><br>
 *      Setzt Standard-Werte, wobei die Parameter der Reihe nach folgende Bedeutung haben:
 *      <ol>
 *      <li>Prioritaet, bei mehreren zutreffenden Tags.</li>
 *      <li>Byte 1-127 zu vergebene Weg-Klasse bzw. Typ.
 *      Negative Werte oder 0 sind reserviert.
 *      </li>
 *      <li>Standardgeschwindigkeit in km/h.</li>
 *      <li>Erlaubte Transport-Typen, z.B. <code>car,bike</code>.
 *          Dieser Wert landet spaeter in den <strong>flags</strong>
 *      </li>
 *      </ol>
 * </li>
 * <li><strong>wtr.avoid.key.value (seit Version 5.3.3)</strong><br>
 *      z.B. <code>wtr.avoid.access.no=car</code><br>
 *      Sperrt den Zugang fuer <code>car</code>.
 * </li>
 * <li><strong>wtr.allow.key.value</strong>:
 *      z.B. <code>wtr.deny.motorcar.yes=car</code><br>
 *      Setzt das <code>car</code>-flag wenn <code>motorcar</code>
 *      erlaubt.
 * </li>
 * <li><strong>wtr.deny.key.value</strong>:
 *      z.B. <code>wtr.deny.tracktype.grade5=bike</code><br>
 *      Entfernt das <code>bike</code>-flag wenn <code>grade5</code>
 * </li>
 * </ul>
 * <code>tag, avoid, allow, deny</code> bilden eine Prioritaets-Reihenfolge.
 * <code>tag</code> liefert den Standardwert, also fuer den Fall, dass keine weiteren
 * Informationen gesetzt sind. <code>avoid</code> loescht ggf. die Standards ab.
 * <code>allow</code> setzt wieder und <code>deny</code> setzt wieder zurueck.
 * Diese Reihenfolge laesst - sofern geschickt eingesetzt - fast alle denkbaren
 * Tag-Kombinationen auf die Konvertierungs-Absichten abbilden.
 *
 * @author (c) 2012/2018/2020 - Carsten Moeller - info@osm2po
 * @see TagResolver
 */
public class DefaultWayTagResolver implements WayTagResolver {

    private static final String PREFIX = "wtr.";
    private static final int PREFIX_LEN = PREFIX.length();

    // CFX: ConfigPrefix
    private static final String CFX_AVOID = "avoid.";
    private static final String CFX_DENY = "deny.";
    private static final String CFX_ALLOW = "allow.";
    private static final String CFX_MAINTAG = "tag.";
    // MKX: MapKeyPrefix
    private static final String MKX_AVOID = "~";
    private static final String MKX_DENY = "-";
    private static final String MKX_ALLOW = "+";
    private static final String MKX_TAG = "*";

    private Map<String, Integer> flags = new HashMap<String, Integer>();
    private Map<String, Object> tagDesk  = new HashMap<String, Object>();

    private int accessDefaultBits;
    private int accessAvoidBits;
    private int accessAllowedBits;
    private int accessDeniedBits;

    private int lastMainTypeLevel = Integer.MAX_VALUE;

    private int[] finalMasks;
    private int nFinalMasks;
    private int polyMask;

    private boolean isRail;

    private String[] maxSpeedTags;
    private int lastMaxSpeedTagsIndex;
    private boolean speedSetByTag;

    private boolean trustValid;

    // Resultierende Properties fuer Ways
    private String name, ref; // ref nur temporaer
    private boolean isRefLeadingName; // z.B. bei Autobahnen. OSM-typisch: ref="A 7"
    private byte clazz;
    private int oneWay;
    private boolean isRoundabout;
    private int kmh;
    private Var<?> custom;

    @Override
    public void open(Config config) {

        String flagList = config.getProperty(PREFIX + "flagList");
        String strFlags[] = null == flagList ? new String[]{} : flagList.split(",");
        for (int i = 0; i < strFlags.length; i++)
            this.flags.put(strFlags[i].trim(), Integer.valueOf(1 << i));

        this.nFinalMasks = 0;
        String strFinalMask = config.getProperty(PREFIX + "finalMask");
        if (strFinalMask != null) {
            if (strFinalMask.indexOf('&') != -1) {
                throw new RuntimeException("format of finalMask has changed!"
                        + "\nUse '|' instead of '&' and ',' instead of '|'");
            }

            String[] masks = strFinalMask.split(",");
            this.nFinalMasks = masks.length;
            this.finalMasks = new int[this.nFinalMasks];
            for (int i = 0; i < this.nFinalMasks; i++) {
                this.finalMasks[i] = this.flags2int(masks[i]);
            }
        }

        this.trustValid = Boolean.valueOf(config.getProperty(PREFIX + "trustValid"));

        String strPolyFlags = config.getProperty(PREFIX + "polyMask");
        if (strPolyFlags != null)
            this.polyMask = this.flags2int(strPolyFlags);

        String strMaxSpeedTags = config.getProperty(PREFIX + "maxSpeedTags");
        if (strMaxSpeedTags != null) {
            String[] tags = strMaxSpeedTags.split(",");
            this.maxSpeedTags = new String[tags.length];
            for (int i = 0; i < tags.length; i++) this.maxSpeedTags[i] = tags[i].trim();
        }

        for (String key : config) {
            if (!key.startsWith(PREFIX)) continue;
            String value = config.getProperty(key);
            if ("".equals(value) || null == value) continue;
            String keyWihoutPrefix = key.substring(PREFIX_LEN); // z.B. 'wtr.'
            if (keyWihoutPrefix.startsWith(CFX_MAINTAG)) {
                String mapKey = keyWihoutPrefix.substring(CFX_MAINTAG.length());
                this.tagDesk.put(MKX_TAG + mapKey, this.mainTagParams2objArr(value));
            } else if (keyWihoutPrefix.startsWith(CFX_AVOID)) {
                String mapKey = keyWihoutPrefix.substring(CFX_AVOID.length());
                this.tagDesk.put(MKX_AVOID + mapKey, Integer.valueOf(this.flags2int(value)));
            } else if (keyWihoutPrefix.startsWith(CFX_ALLOW)) {
                String mapKey = keyWihoutPrefix.substring(CFX_ALLOW.length());
                this.tagDesk.put(MKX_ALLOW + mapKey, Integer.valueOf(this.flags2int(value)));
            } else if (keyWihoutPrefix.startsWith(CFX_DENY)) {
                String mapKey = keyWihoutPrefix.substring(CFX_DENY.length());
                this.tagDesk.put(MKX_DENY + mapKey, Integer.valueOf(this.flags2int(value)));
            }
        }
    }


    private int[] mainTagParams2objArr(String paramsCsv) {
        String[] params = paramsCsv.split(",");
        int paramsCount = params.length;
        int[] intVals = new int[paramsCount];
        for (int i = 0; i < paramsCount; i++) {
            // MainTagLevel, Type, Speed, DefaultAllowedFlags
            int intVal = 0;
            if (i == 3) {
                intVal = this.flags2int(params[i]);
            } else {
                intVal = Integer.valueOf(params[i].trim());
            }
            intVals[i] = intVal;
        }
        return intVals;
    }


    private int flags2int(String flags) {
        try {
            String flagsArr[] = flags.split("\\|");
            int flagsValue = 0;
            for (int i = 0; i < flagsArr.length; i++) {
                String flag = flagsArr[i].trim();
                Integer flagListVal = this.flags.get(flag);
                int bits = (null == flagListVal) ? Integer.decode(flag) : flagListVal;
                flagsValue |= bits;
            }
            return flagsValue;

        } catch (Throwable t) {
            throw new IllegalArgumentException(
                    "Flag(s) '" + flags + "' not declared");
        }
    }

    @Override
    public void onTags(long entityId) {
        this.name = null;
        this.ref = null;
        this.isRefLeadingName = false;

        this.clazz = CLAZZ_NULL; // 0
        this.oneWay = ONEWAY_NO;
        this.kmh = 0;
        this.isRail = false;

        this.isRoundabout = false;

        this.custom = null;

        // priority in this order:
        this.accessDefaultBits = 0; // Set
        this.accessDeniedBits = 0; // UnSet
        this.accessAllowedBits = 0; // OverSet
        this.accessAvoidBits = 0; // FinalUnSet

        this.lastMainTypeLevel = Integer.MAX_VALUE;
        this.lastMaxSpeedTagsIndex = 0;
        this.speedSetByTag = false;
    }

    @Override
    public void onTag(String key, String value) {
        if (null == key) key = ""; // Sicher ist sicher

        if ("name".equals(key)) {
            this.name = value;
        } else if ("ref".equals(key)) {
            this.ref = value;
        } else if ("oneway".equals(key)) {
            if ("yes".equals(value) || "1".equals(value) || Boolean.valueOf(value)) {
                this.oneWay = ONEWAY_YES;
            } else if ("-1".equals(value) || "reverse".equals(value)) {
                this.oneWay = ONEWAY_YES_REVERSED;
            } else if (value.startsWith("yes") || value.startsWith("true")) {
                this.oneWay = ONEWAY_YES_UNKNOWN;
            }
        } else if ("junction".equals(key) && "roundabout".equals(value)) {
            this.isRoundabout = true;
        } else if ("highway".equals(key) && "motorway".equals(value)) {
            this.isRefLeadingName = true;
        } else if ("railway".equals(key) && "rail".equals(value)) {
            this.isRail = true;

        } else if (this.maxSpeedTags != null) {
            for (int i = this.lastMaxSpeedTagsIndex; i < this.maxSpeedTags.length; i++) {
                if (this.maxSpeedTags[i].equals(key)) {
                    try {
                        String v = value.trim().toLowerCase();
                        boolean isMph = false;
                        if (v.endsWith("mph")) {
                            v = v.substring(0, v.length() -3).trim();
                            isMph = true;
                        } else if (v.endsWith("kmh")) {
                            v = v.substring(0, v.length() -3).trim();
                        }
                        this.kmh = Integer.parseInt(v);
                        if (isMph) this.kmh *= 1.609344;
                        this.lastMaxSpeedTagsIndex = i +1;
                        this.speedSetByTag = true;
                        break;
                    } catch (Exception ignore) {}
                }
            }
        }

        // Explizites Matching
        int[] mainParams = (int[]) this.tagDesk.get(MKX_TAG + key + '.' + value);
        Integer avoidFlags = (Integer) this.tagDesk.get(MKX_AVOID + key + '.' + value);
        Integer allowFlags = (Integer) this.tagDesk.get(MKX_ALLOW + key + '.' + value);
        Integer denyFlags = (Integer) this.tagDesk.get(MKX_DENY + key + '.' + value);

        if (mainParams != null) this.onMainTag(mainParams);
        if (avoidFlags != null) this.accessAvoidBits |= avoidFlags.intValue();
        if (allowFlags != null) this.accessAllowedBits |= allowFlags.intValue();
        if (denyFlags != null) this.accessDeniedBits |= denyFlags.intValue();
    }

    @Override
    public void onObj(String key, Object value) {
        if (TAG_CUSTOM.equals(key) && value instanceof Var<?>) {
            this.custom = (Var<?>) value;
        }
    }

    private void onMainTag(int[] params) {
        for (int i = 0; i < params.length; i++) {
            // Mittels Schleifen-Abgriff auch optionale Werte moeglich
            int intVal = params[i];
            switch (i) {
                case 0:
                    if (intVal >= this.lastMainTypeLevel) {
                        i = 2; // Optionalen wtr.allow case 3! springen
                        continue; // Doppelte MainTypeDef. z.B. railway+highway
                    }
                    this.lastMainTypeLevel = intVal;
                    break;
                case 1:
                    if (intVal < 1 || intVal > 127) throw new RuntimeException(
                            "Clazz must be a positive byte value between 1 and 127");
                    this.clazz = (byte) intVal;
                    break;
                case 2:
                    if (!this.speedSetByTag) { // Reihenfolge! MaxSpeed kann vor MainTag kommen!
                        this.kmh = intVal; // Highest Priority will win - see skip in case 0
                    }
                    break;
                case 3:
                    this.accessDefaultBits |= intVal;
                    break;
            }
        }
    }

    @Override
    public String getName() {
        if (null == this.name) {
            return this.ref; // wenn null, dann auch richtig
        } else if (this.ref != null && this.isRefLeadingName) {
            return this.ref + " - " + this.name; // OSM-MultiRef-Trenner ist ";"
        } else {
            return this.name;
        }
    }

    @Override
    public Var<?> getCustom() {
        return this.custom;
    }

    @Override
    public int getFlags() {
        int flags = this.accessDefaultBits;
        flags &= ~this.accessAvoidBits;
        flags |= this.accessAllowedBits;
        flags &= ~this.accessDeniedBits;
        return flags;
    }

    @Override
    public byte getClazz() {
        return this.clazz;
    }

    @Override
    public int getOneWayType() {
        if (ONEWAY_NO == this.oneWay && this.isRoundabout) {
            // OneWay nicht explizit getagged.
            // Achtung! Kann auch reverse sein, daher Aufloesung erst hier.
            this.oneWay = ONEWAY_YES_ROUNDABOUT;
        }
        return this.oneWay;
    }

    /**
     * Liefert die finale Entscheidung, ob ein Weg verarbeitet
     * oder verworfen werden soll.<br>
     * Um als valide zu gelten, muss eine der folgenden
     * Bedingungen erfuellt sein:
     * <ul>
     * <li><code>wtr.trustValid=true</code> (konvertiert alle Wege).
     * <li><code>clazz</code> und <code>flags</code> sind groesser 0.</li>
     * <li>
     *   Wenn <code>wtr.finalMask</code> gesetzt ist, dann muessen
     *   die <code>flags</code> mit einer der gesetzten Masken
     *   1zu1 uebereinstimmen. Mehrere Maskenwerte sind durch Kommata
     *   getrennt aufzulisten.
     * </li>
     * </ul>
     * @return boolean - gueltig ja/nein.
     * @see DefaultWayTagResolver
     */
    @Override
    public boolean isValid() {
        if (this.trustValid) return true;
        // Klasse MUSS gesetzt sein
        if (CLAZZ_NULL == this.clazz) return false;

        // Wenn keine Maske vorhanden, dann MUSS
        // mindestenst ein Flag gesetzt sein
        int flags = this.getFlags();
        if (0 == nFinalMasks && flags != FLAGS_NULL) {
            return true;
        }

        // Wenn Maske(n) vorhanden, dann im Detail pruefen
        for (int i = 0; i < this.nFinalMasks; i++) {
            int finalMask = this.finalMasks[i];
            if ((finalMask & flags) == finalMask) return true;
        }

        return false;
    }

    @Override
    public int getSpeedInKmh() {
        return this.kmh;
    }

    @Override
    public boolean isRail() {
        return this.isRail;
    }

    @Override
    public boolean isPolygon() {
        return this.polyMask != 0
            && ((this.getFlags() & this.polyMask) == this.polyMask);
    }

    @Override
    public boolean isRoundabout() {
        return this.isRoundabout;
    }

    @Override
    public void onTagsComplete() {
    }

    @Override
    public void close() {
    }

}
