/*
 * $Id: OsmNodeCache.java 4423 2017-04-16 08:26:55Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.primitives.LongLongBag;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessMemory;

/**
 * Diese Klasse bietet einen NodeCache nach dem Prinzip
 * WriteOnce ReadOften.<br>
 * Da wir davon ausgehen koennen, dass OSM-Extrakte
 * aus Datenbanken stammmen und daher die Knoten bereits
 * eindeutig sind, werden hier Bags und keine Maps verwendet.
 * Daher sollte nach dem Fuellen einmal {@link #sort()}
 * aufgerufen werden, damit die folgenden Zugriffe mittels
 * Binaersuche optimiert werden koennen.<br>
 * Um den Speicherverbrauch gering zu halten, wird die Tatsache
 * ausgenutzt, dass die meisten OSM-Knoten reine Koordinaten sind
 * und diese sich in einem einzigen Long (7 Nachkommastellen)
 * unterbringen lassen. Deshalb werden hier intern zwei Bags
 * verwaltet und Knoten aufgrund von {@link OsmNode#isLatLonOnly()}
 * entsprechend einsortiert. {@link #getAsNode(long)} muss folglich
 * in beiden Bags suchen, was die Geschichte ein wenig verlangsamt.
 * Dieser Nachteil ist aber vernachlaessigbar, da der geringe
 * Speicherverbrauch hier entscheidend ist.<br>
 * 
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 *
 */
public class OsmNodeCache {
    private LongLongBag latLonBag = new LongLongBag();
    private LongLongBag nodeBag = new LongLongBag();
    private RandomAccess cache = new RandomAccessMemory();

    public long getUsedCache() {
        // TODO bei Zeiten nodeBag auf AscIndexMemory umstellen
        return cache.getWriteIndex();
    }

    public void clear() {
        this.latLonBag.clear();
        this.nodeBag.clear();
        this.cache.clear();
    }

    /** Sortiert fuer schnellere Suche. */
    public void sort() {
        this.latLonBag.sort();
        this.nodeBag.sort();
    }

    /**
     * Fuegt eine Knoten in den Cache ein.
     * Als Schluessel wird {@link OsmNode#getId()} verwendet.
     * <br><b>Wichtig:</b><br>
     * Hier sollten nur {@link OsmNode} Objekte und keine
     * abgeleiteten Klassen uebergeben werden.
     * @param osmNode {@link OsmNode}
     */
    public void add(OsmNode osmNode) {
        if (osmNode.isLatLonOnly()) {
            this.latLonBag.add(osmNode.getId(), osmNode.getCoord());
        } else {
            long idx = this.cache.getWriteIndex();
            this.nodeBag.add(osmNode.getId(), idx);
            osmNode.setId(0); // Tricky Compression
            osmNode.writeToStream(this.cache);
        }
    }

    /**
     * Liefert eine neue Knoteninstanz aus dem Cache.
     * Gleichzeitig wird ein Upcast von OsmNode zu Node durchgefuehrt.
     * @param key long Schluessel des Knoten.
     * @return new {@link Node} oder NULL.
     */
    public Node getAsNode(long key) {
        Node node = null;
        long value = this.latLonBag.get(key);
        if (value != LongLongBag.NULLVALUE) {
            node = new Node();
            node.setId(key); // Retten.
            node.setCoord(value);
        } else {
            long index = this.nodeBag.get(key);
            if (index != LongLongBag.NULLVALUE) {
                this.cache.setReadIndex(index);
                node = new Node();
                node.readFromOsmNodeStream(this.cache);
                node.setId(key); // Retten
            }
        }
        return node;
    }

    /**
     * Pure Contains-Pruefung, ob Knotenschluessel bekannt ist.
     * @param key long KnotenRef-Schluessel.
     * @return im Cache ja/nein.
     */
    public boolean contains(long key) {
        long value = this.latLonBag.get(key);
        if (value == LongLongBag.NULLVALUE) {
            long index = this.nodeBag.get(key);
            if (index == LongLongBag.NULLVALUE) return false;
        }
        return true;
    }

}
