/*
 * $Id: SimpleNodeTagResolver.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import de.cm.osm2po.Config;
import de.cm.osm2po.misc.Props;

/**
 * Erbt von {@link DefaultNodeTagResolver} und kann simpel konfigurierte
 * Tags fuer {@link #getClazz()} uebersetzen.
 * <br><br><b>Beispiel</b> Fuer christliche Kirchen die clazz 77 setzen:<br>
   <code>
     nodeTagResolver.class=de.cm.osm2po.converter.tiler.SimpleNodeTagResolver
     ntr.clazz.77=amenity/place_of_worship,religion/christian
   </code><br>
 * Zusaetzlich wird hier auch das <code>name</code>-tag beruecksichtigt und
 * zudem die <strong>osm2po</strong>-Tags <code>clazz</code>
 * und <code>flags</code> interpretiert. Bei Konflikten gewinnen letztere.
 *
 * @author (c) 2017 - Carsten Moeller - info@osm2po
 * @see Poi2poWriterTagResolver
 */
public class SimpleNodeTagResolver extends DefaultNodeTagResolver {
    public final static String PROP_PREFIX_CLAZZ = "ntr.clazz.";

    private Map<String, Integer> tagMap;
    private int[] tags;
    private int nTags;

    private String name;
    private byte clazz;
    private int flags;

    @Override
    public void open(Config config) {
        super.open(config);
        this.tagMap = new HashMap<String, Integer>();
        this.tags = new int[0x100];
        this.initTags(config.getProps());
    }

    @Override
    public void onTags(long entityId) {
        super.onTags(entityId);
        this.nTags = 0;
        this.clazz = CLAZZ_NULL;
        this.flags = FLAGS_NULL;
        this.name = null;
    }

    @Override
    public void onTagsComplete() {
        super.onTagsComplete();
        if (CLAZZ_NULL == this.clazz && this.nTags > 0) {
            int[] arr = Arrays.copyOf(this.tags, this.nTags);
            String key = this.makeKeySet(arr);
            Integer val = this.tagMap.get(key);
            if (val != null) {
                this.clazz = val.byteValue();
            }
        }
    }

    @Override
    public void onTag(String key, String value) {
        super.onTag(key, value);
        Integer itag = this.tagMap.get(key + "/" + value);
        if (itag != null) {
            this.tags[this.nTags++] = itag;
        } else if ("name".equals(key)) {
            this.name = value;
        } else if ("clazz".equals(key)) {
            this.clazz = (byte) Integer.parseInt(value);
        } else if ("flags".equals(key)) {
            this.flags = (int) Long.parseLong(value);
        }
    }

    @Override
    public byte getClazz() {
        return this.clazz;
    }

    @Override
    public int getFlags() {
        return this.flags;
    }

    @Override
    public String getName() {
        String nam = super.getName();
        if (null == nam) nam = this.name;
        return nam;
    }

    /**
     * Baut die Informationen fuer Node-Tags der Form
     * natural=wetland, wetland=tidalflat.
     * @param props {@link Props}
     */
    private void initTags(Props props) {
        int[] idxs = new int[0x100];
        int nextIdx = 1;
        for (String propKey : props) {
            if (propKey.startsWith(PROP_PREFIX_CLAZZ)) {
                int clazz = Integer.valueOf(propKey.substring(PROP_PREFIX_CLAZZ.length()));
                String propValue = props.getProperty(propKey);
                String[] tagSets = propValue.split(";");
                for (String tagSet : tagSets) {
                    String[] tags = tagSet.split(",");
                    int nTags = tags.length;
                    for (int i = 0; i < nTags; i++) {
                        String key = tags[i].trim();
                        Integer idx = this.tagMap.get(key);
                        if (null == idx) {
                            idx = nextIdx++;
                            this.tagMap.put(key, idx);
                        }
                        idxs[i] = idx;
                    }
                    String key = this.makeKeySet(Arrays.copyOf(idxs, nTags));
                    this.tagMap.put(key, clazz);
                }
            }
        }
    }

    private String makeKeySet(int[] arr) {
        String s = "";
        arr = intArrayDistinct(arr);
        for (int i = 0; i < arr.length; i++)  s += "@" + arr[i];
        return s;
    }

}
