/*
 * $Id: TagResolver.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.converter.parser.OsmParserHandler;
import de.cm.osm2po.primitives.Var;

/**
 * Interface fuer Callback-Klassen, die eine Sammlung von OSM-Tags waehrend
 * der Parsing-Phase auf eindeutige Aussagen verdichten. Wird normalerweise
 * von einem {@link OsmParserHandler}, z.B. {@link Tiler} aufgerufen.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 * @see DefaultNodeTagResolver
 * @see DefaultWayTagResolver
 * @see DefaultRelationTagResolver
 */
public interface TagResolver {

    public static final byte CLAZZ_NULL = 0;
    public static final int FLAGS_NULL = 0;
    public static final String TAG_CUSTOM = "custom";

    /**
     * Wird einmalig nach der Instanziierung aufgerufen.
     * Dies ist faktisch der Quasi-Konstruktor.
     * @param config {@link Config} Konfiguration.
     */
    void open(Config config);

    /**
     * Schliesst und ermoeglicht so das Aufraeumen
     * etwaiger Resourcen.
     */
    void close();

    /**
     * Wird vom {@link OsmParserHandler} aufgerufen, um
     * eine Instanz dieser Klasse darueber zu informieren,
     * dass interne Variablen zu reinitialisieren sind.
     * Also vor dem Lesen einer neuen Tag-Sammlung.
     * @param entityId Id - Node, Way, Relation.
     */
    void onTags(long entityId);

    /**
     * Wird vom {@link OsmParserHandler} fuer jedes gelesene Tag aufgerufen.
     * @param key {@link String} Schluessel.
     * @param value {@link String} Wert.
     */
    void onTag(String key, String value);

    /**
     * Kann von einem {@link OsmParserHandler} gerufen werden,
     * um benutzerdefinierte Daten zu uebertragen.<br>
     * Eine Implementation sollte {@link #TAG_CUSTOM} bereitstellen
     * und Objekte vom Typ {@link Var} als {@link #getCustom()} zurueckliefern.
     * @param key {@link String} Schluessel.
     * @param value {@link Object} beliebiger Wert.
     * @see #getCustom()
     */
    void onObj(String key, Object value);

    /**
     * Wird nach dem Lesen aller Tags
     * aufgerufen und bietet somit die letzte Chance
     * die bisher ermittelten Werte zu interpretieren.
     */
    void onTagsComplete();

    /**
     * Liefert die finale Entscheidung, ob eine Tag-Sammlung
     * verarbeitet oder verworfen werden soll.
     * @return boolean - gueltig ja/nein.
     */
    public boolean isValid();

    /**
     * Sollte ein BitFeld (int) liefern, welches Auskunft
     * ueber Bestimmte Eigenschaften liefert.
     * Was hier tatsaechlich zurueckgeliefert wird, bestimmt
     * die Implementation dieses Interfaces.
     * Es spricht auch nichts dagegen, diesen <code>int</code>
     * zweckzuentfremden.
     * @return int (BitFeld)
     */
    int getFlags();

    /**
     * Liefert eine Typ-Klassen-Id.
     * <br><strong> Werte &lt;=0 sind reserviert </strong>
     * @return Klassen-Id &gt; 0
     */
    byte getClazz();

    /**
     * Liefert den Namen einer OSM-Entity.
     * @return {@link String} Name.
     */
    public String getName();

    /**
     * Liefert frei definierbare Zusatzinformationen.
     * @return custom {@link Var}.
     * @see #onObj(String, Object)
     */
    Var<?> getCustom();

}
