/*
 * $Id: Tile.java 5272 2021-04-11 17:15:28Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;

import java.io.File;
import java.util.Arrays;

import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Node;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Haelt Informationen genau eines Tiles aus dem
 * {@link TileManager} und bietet zusaetzlich Methoden zur
 * Ermittlung, Namensvergabe und Sortierung aufgrund der geografischen
 * Lage. Zusaetzlich werden noch transiente Tile-Infos gehalten,
 * die fuer das NodeTiling (Sortieren von Nodes in Tiles) und den
 * entsprechenden Dateioperationen benoetigt werden.<br>
 * Zudem liefert diese Klasse auch alle Mechaniken zum Cachen
 * von Knoten.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class Tile implements Var<Tile>, Comparable<Tile> {
    public final static String SHARED_TILE_ID = "SHARED";

    private int id;
    private long nComplexNodes, nFormNodes; // Tiler
    private long nTiledWays, nTiledWaysNodes; // Joiner
    private long nSegments; // Segmenter

    private transient OsmNodeCache osmNodeCache;
    private transient OutStreamDisk outStreamDisk;
    private transient boolean isRelevantTile;

    public Tile() {}
    public Tile(int id) {
        this.id = id;
    }

    @Override
    public int compareTo(Tile that) {
        // Tiles mit den meisten Knoten sind oben/vorne.
        long thatNodes = that.nFormNodes + that.nComplexNodes;
        long thisNodes = this.nFormNodes + this.nComplexNodes;
        if (thatNodes > thisNodes) return 1;
        else if (thatNodes < thisNodes) return -1;
        else return 0;
    }

    public void cacheOsmNode(OsmNode osmNode) {
        if (null == this.osmNodeCache)
            this.osmNodeCache = new OsmNodeCache();
        this.osmNodeCache.add(osmNode);
        this.incNodeCounter(osmNode.isLatLonOnly());
    }

    /**
     * Cached alle Knoten dieses Tiles aus der uebergebenen Datei.
     * @param osmNodesFile {@link File}
     * @param log {@link Log} (Logger)
     * @return erfolgreich true/false, false bei {@link OutOfMemoryError}
     */
    public boolean cacheOsmNodes(File osmNodesFile, Log log) {
        int nNodes = 0;
        InStream inStream = null;

        try {
            this.osmNodeCache = new OsmNodeCache();
            inStream = new InStreamDisk(osmNodesFile);
            if (VarTypeDesk.typeIdOf(OsmNode.class) != inStream.readByte())
                throw new RuntimeException(
                        "Wrong Node File " + osmNodesFile.getAbsolutePath());

            OsmNode osmNode = new OsmNode();
            while (!inStream.isEof()) {
                osmNode.readFromStream(inStream);
                this.osmNodeCache.add(osmNode);

                if (++nNodes % 5000000 == 0) {
                    log.progress(DF(nNodes) + " of " + DF(this.getNumberOfNodes()) + " nodes cached ("
                            + getTileName(this.id) + ") - " + MB());
                }
            }

            log.progress("Sorting");
            this.osmNodeCache.sort();

        } catch (OutOfMemoryError ome) {
            // Seit OpenJDK9 gibt's hier ein Problem. Saemtliche calls ausser osmNodeCache.clear()
            // verursachen selbst wieder einen OutOfMemory, was die komplette Logik aushebelt.
            // FIX 2021-03-31: Erst clear(), dann alles andere. Aber komisch: JDK15 doch wieder ok?
            this.osmNodeCache.clear(); // Verbrauchten Speicher wieder freiraeumen
        	log.warn("Wrong estimation of free memory - " + MB());
        	log.debug("NodeCache used for Complex: " + DF(this.osmNodeCache.getUsedCache()));
        	return false;

        } catch (Throwable t) {
            throw new RuntimeException(t);

        } finally {
        	log.info(DF(nNodes) + " of " + DF(this.getNumberOfNodes()) + " nodes cached ("
        			+ getTileName(this.id) + ") - " + MB());
        	if (inStream != null) inStream.close();
        }
        return true;
    }


    /**
     * Liefert einen Knoten aus dem TileCache.
     * @param key long Schluessel
     * @return {@link Node} oder <code>null</code>, wenn nicht gefunden.
     */
    public Node lookupNode(long key) {
        return this.osmNodeCache.getAsNode(key);
    }

    /**
     * Pure Contains-Pruefung, ob Knotenschluessel im Cache bekannt ist.
     * @param key long KnotenRef-Schluessel.
     * @return im Cache ja/nein.
     */
    public boolean containsNode(long key) {
    	return this.osmNodeCache.contains(key);
    }

    /**
     * Reinitialisiert und raeumt Speicher frei.
     */
    public void clearCache() {
        if (this.osmNodeCache != null) {
            this.osmNodeCache.clear();
            this.osmNodeCache = null;
        }
    }

    @Override
    public String toString() {
        String s = ""
            + "Id:" + this.id + " (" + getTileName(this.id) + ")"
            + " Nodes: " + DF(this.nFormNodes) + " plus "
            + DF(this.nComplexNodes) + " complex";
        if (this.nTiledWays > 0) {
            s += "\n"
                + "Ways:" + DF(this.nTiledWays)
                + " WayNodes:" + DF(this.nTiledWaysNodes);
        }
        if (this.nSegments > 0) {
            s += " Segments:" + DF(this.nSegments);
        }
        return s;
    }

    public int getId() {return id;}
    public void setId(int id) {this.id = id;}


    /**
     * Wird vom {@link Tiler} verwendet um zwischen Tiles
     * zu unterscheiden, die Knoten nur aufgrund ihres
     * Ueberschneidungs-Puffers (tileBuffer) oder "echt"
     * beinhalten, also "relevant" sind.
     * @param isRelevantTile Tile ist relevant ja/nein.
     */
    public void setRelevantTile(boolean isRelevantTile) {
        this.isRelevantTile = isRelevantTile;
    }
    /**
     * Merkt sich hier, ob das Tile ueberhaupt benoetigt wird oder
     * ob es aufgrund der Puffer-Einstellung nur vorsorglich angelegt wurde.
     * Wird von {@link TileManager#closeNodeTiling()} gerufen.
     * @return Tile ist relevant ja/nein.
     * @see Tile#setRelevantTile(boolean)
     */
    public boolean isRelevantTile() {
        return isRelevantTile;
    }

    public void resetNodeCounter() {
        this.nFormNodes = 0;
        this.nComplexNodes = 0;
    }

    /**
     * Inkrementiert den internen Knoten-Zaehler.
     * @param isLatLonOnly
     *      Art von Knoten (Koordinate oder Complex)
     */
    public void incNodeCounter(boolean isLatLonOnly) {
        if (isLatLonOnly)
            this.nFormNodes++;
        else
            this.nComplexNodes++;
    }

    public void resetTiledWaysCounter() {
        this.nTiledWays = 0;
        this.nTiledWaysNodes = 0;
    }

    public void incTiledWaysCounter() {
        this.nTiledWays++;
    }

    public void incTiledWaysNodesCounter(int addNum) {
        this.nTiledWaysNodes += addNum;
    }

    public void setNumberOfSegments(long nSegments) {
        this.nSegments = nSegments;
    }

    /**
     * Liefert die Anzahl aller Knoten, egal ob
     * Form- (Koordinate) oder Complex-Node.
     * @return Anzahl aller Knoten in diesem Tile.
     */
    public long getNumberOfNodes() {
        return this.nFormNodes + this.nComplexNodes;
    }

    public long getNumberOfWays() {
        return this.nTiledWays;
    }

    public long getNumberOfSegments() {
        return this.nSegments;
    }

    public long getNumberOfWaysNodes() {
        return this.nTiledWaysNodes;
    }

    /**
     * Zentrale Abschaetzmethode fuer das Node-Caching.
     * @return Speicherverbrauch aller Knoten dieses Tiles im RAM.
     */
    protected long estimateRamUsage() {
        return this.nComplexNodes * 24 + this.nFormNodes * 16;
    }

    /**
     * Transienter Datei-Ausgabestrom.
     * @param outStreamDisk {@link OutStreamDisk}
     */
    public void setOutStreamDisk(
            OutStreamDisk outStreamDisk) {
        this.outStreamDisk = outStreamDisk;
    }

    /**
     * Transienter Datei-Ausgabestrom.
     * @return {@link OutStreamDisk}
     */
    public OutStreamDisk getOutStreamDisk() {
        return outStreamDisk;
    }

    /**
     * Liefert intern die eindeutige TileId aufgrund
     * von y- und x-Koordinate.<br>
     * <strong>Achtung!</strong> Die erste Id hat den Wert 0.
     * @param y int y-Koordinate
     * @param x int x-Koordinate
     * @return int TileId (Kodiert aus y und x)
     */
    private static int getTileId(int y, int x) {
        int tileY = y + 90;
        int tileX = x + 180;
        int id = (tileY << 16) | tileX;
        return id;
    }

    /**
     * Statische Methode, welche Lat und Lon aus einer
     * TileId ermittelt und einen fuer Schluessel oder Dateinamen
     * nuetzlichen String zurueckliefert.
     * @param tileId Id des Tiles bzw. LatLon.
     * @return nuetzlicher LatLon-Name.
     */
    public static String getTileName(int tileId) {
        int y = (tileId >>> 16) - 90;
        int x = (tileId & 0xFFFF) - 180;
        String lon000 = String.valueOf(StrictMath.abs(x) + 1000);
        lon000 = lon000.substring(lon000.length() -3);
        String lat000 = String.valueOf(StrictMath.abs(y) + 1000);
        lat000 = lat000.substring(lat000.length() -3);
        return (y == 0 ? "Z" : y < 0 ? "S" : "N") + lat000
            + (x == 0 ? "Z" : x < 0 ? "W" : "E") + lon000;
    }

    /**
     * Liefert eine eindeutige TileId auf Grundlage der Koordinate
     * und der Groesse eines Tiles. Berechnet also den Quadranten,
     * in dem sich ein Punkt befindet, wobei die Quadranten ueber
     * ihre linke untere Ecke adressiert werden.
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     * @param tileSizeY int Hoehe eines Tiles.
     * @param tileSizeX int Breite eines Tiles.
     * @return int Eindeutige TileId, in der eine Koordinate liegt.
     */
    public static int getTileId(
            double lat, double lon, int tileSizeY, int tileSizeX) {

        int y = snapDownToY(lat, tileSizeY);
        int x = snapDownToX(lon, tileSizeX);
        return getTileId(y, x);
    }

    /**
     * Prueft, ob eine Koordinate exklusiv nur in (s)einem Tile liegt
     * oder aufgrund von Ueberschneidung anderer Tiles (tileBuffer)
     * auch noch anderen Tiles zuzuordnen ist.
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     * @param tileSizeY int Hoehe eines Tiles.
     * @param tileSizeX int Breite eines Tiles.
     * @param tileBuffer double Ueberschneidung aneinander liegender Tiles.
     * @return Koordinate liegt nur in (s)einem Tile. ja/nein.
     */
    public static boolean isInTileExlusive (
            double lat, double lon, int tileSizeY, int tileSizeX, double tileBuffer) {

        int y1 = snapDownToY(lat - tileBuffer, tileSizeY);
        int x1 = snapDownToX(lon - tileBuffer, tileSizeX);
        int y2 = snapDownToY(lat + tileBuffer, tileSizeY);
        int x2 = snapDownToX(lon + tileBuffer, tileSizeX);

        return y1 == y2 && x1 == x2;
    }

    /**
     * Liefert eine Menge (Array) von TileIds, in denen sich eine
     * Koordinate aufgrund von eventuellen Ueberschneidungen mit
     * benachbarten Tiles befindet.
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     * @param tileSizeY int Hoehe eines Tiles.
     * @param tileSizeX int Breite eines Tiles.
     * @param tileBuffer double Ueberschneidung aneinander liegender Tiles.
     * @return int[] Array von TileIDs.
     */
    public static int[] getTileIds(
            double lat, double lon, int tileSizeY, int tileSizeX,
            double tileBuffer /*, boolean spherical */) {

        double tileBufferY = tileBuffer;
        double tileBufferX = tileBuffer;

        final int maxTiles = 16;
        int[] tileIds = new int[maxTiles];

// TODO SPHERICAL wieder einrechnen. Deaktiviert, da bei tileSize=90x180,1 (seatiles) buggy!
// Zu den Polen hin liegen die Laengengrade dichter zusammen.
// Eine Strecke in X-Richtung ist bei gleicher Grad-Entfernung
// auf dem 60. Breitengrad nur noch halb so lang wie am Aequator.
// Zusaetzlich meiden wir +/- 90 Grad. (durch 0 fehler)
// * @param spherical
// *      x = f(cos(y)) fuer Erdkruemmung an den Polen einrechnen
// *      und entsprechend hier den tileBuffer in X-Richtung vergroessern. Ja/Nein.
//
//        if (spherical) {
//        double y = lat < 0 ? lat + tileSizeY : lat - tileSizeY;
//        double xFactor = 1 / StrictMath.cos(StrictMath.toRadians(y));
//        tileBufferX *= xFactor; // Puffer in X-Richtung erweitern.
//        // Der Puffer sollte nicht groesser als das Tile selbst sein!
//        if (tileBufferX > tileSizeX) tileBufferX = tileSizeX;
//        }

        int y1 = snapDownToY(lat - tileBufferY, tileSizeY);
        int x1 = snapDownToX(lon - tileBufferX, tileSizeX);
        int y2 = snapDownToY(lat + tileBufferY, tileSizeY);
        int x2 = snapDownToX(lon + tileBufferX, tileSizeX);

        int count = 0;
        for (int y = y1; ; incY(y, tileSizeY)) {
            for (int x = x1;;) {
                int tileId = getTileId(y, x);
                if (count >= maxTiles) throw new IllegalStateException(
                        "More than " + maxTiles + " Tiles touched by Coordinate " + lat + " " + lon
                        + "\nPlease check whether TileSize and TileBuffer make sense");
                tileIds[count++] = tileId;
                if (x == x2) break;
                x = incX(x, tileSizeX);
            }
            if (y == y2) break;
            y = incY(y, tileSizeY);
        }

        return Arrays.copyOf(tileIds, count);
    }


    private static int snapDownToY(double lat, int tileSizeY) {
        double y = StrictMath.floor((lat + 180) / tileSizeY);
        y = (y * tileSizeY) % 180;
        if (y >= 90) y -= 180;
        return (int) y;
    }

    private static int snapDownToX(double lon, int tileSizeX) {
        double x = StrictMath.floor((lon + 360) / tileSizeX);
        x = (x * tileSizeX) % 360;
        if (x >= 180) x -= 360;
        return (int) x;
    }

    private static int incY(int y, int dy) {
        y = (y + dy + 180) % 180;
        if (y >= 90) y -= 180;
        return y;
    }

    private static int incX(int x, int dx) {
        x = (x + dx + 360) % 360;
        if (x >= 180) x -= 360;
        return x;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.id);
        outStream.writeLong(this.nComplexNodes);
        outStream.writeLong(this.nFormNodes);
        outStream.writeLong(this.nTiledWays);
        outStream.writeLong(this.nTiledWaysNodes);
        outStream.writeLong(this.nSegments);
    }

    @Override
    public Tile readFromStream(InStream inStream) {
        this.id = inStream.readInt();
        this.nComplexNodes = inStream.readLong();
        this.nFormNodes = inStream.readLong();
        this.nTiledWays = inStream.readLong();
        this.nTiledWaysNodes = inStream.readLong();
        this.nSegments = inStream.readLong();
        return this;
    }

}

