/*
 * $Id: TileManager.java 5463 2021-12-01 18:17:54Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.converter.tiler;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.Joiner;
import de.cm.osm2po.converter.Tiler;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.IntTHashMap;
import de.cm.osm2po.primitives.KeyVal.IntV;
import de.cm.osm2po.primitives.LongArray;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.SortableTArray;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Haelt Informationen aller {@link Tile}<code>s</code>, die waehrend der
 * Konvertierung vom {@link Tiler} verarbeitet werden. Zudem werden hier auch
 * die Caches aufgebaut, die fuer eine Zusammenfuehrung von Knoten,
 * Wegen und Relationen im {@link Joiner} in der Folge benoetigt werden.
 * Bildlich gesprochen ist diese Klasse eine persistente Uebergangsphase
 * zwischen {@link Tiler} und {@link Joiner} inklusiver aller statistischen
 * Werte, die waehrend dieser beiden Verarbeitungsschritte ermittelt werden.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class TileManager implements Var<TileManager>, Iterable<List<Tile>> {

    public static final int MAX_EXPECTED_TILES = 0x400/*1k*/;

    public static final String FILENAME = "tm_info.2po";
    public static String TILED_NODES_FILENAME(int tileId) {
        return "tn_" + Tile.getTileName(tileId) + ".2po";}
    public static String TILED_WAYS_FILENAME(int tileId) {
        return "jw_" + Tile.getTileName(tileId) + ".2po";}

    private String prefix;

    private boolean singleTileMode;
    private Tile singleTile;

    private int tileSizeY;
    private int tileSizeX;
    private int tileThreshold;
    private double tileBuffer;
    private long nWays, nRelations; // Tiler
    private long nSharedWays, nSharedWaysNodes; // Joiner
    private long nSharedSegments; // Segmenter

    private transient long reservedXmx;
    private transient Log log;
    private transient File workDir;
    private transient int maxTilesPerLoop;
    private transient IntTHashMap<Tile> tileMap;
    private transient SortableTArray<Tile> tileArray;

    public TileManager() {
        // max. 360*180(Grad)=64.800 (Tiles) fuer die Welt
        this.tileMap = new IntTHashMap<Tile>(MAX_EXPECTED_TILES);
    }

    public TileManager(Config config) {
        this.tileMap = new IntTHashMap<Tile>(MAX_EXPECTED_TILES);

        this.log = config.getLog();
        this.workDir = config.getWorkDir();
        this.prefix = config.getPrefix();
        this.reservedXmx = config.getReservedXmx();

        this.maxTilesPerLoop = Integer.parseInt(
                config.getProperty("maxTilesPerLoop", "1024"));
        this.tileThreshold = Integer.parseInt(
                config.getProperty("tileThreshold", "0"));

        this.tileSizeY = 10; // Defaults
        this.tileSizeX = 10;
        this.tileBuffer = 1;
        this.singleTileMode = false;

        String strTileMode = config.getProperty("tileSize");
        if (strTileMode != null) {
            String[] params = strTileMode.split(",");
            for (int i = 0; i < params.length; i++) {
                params[i] = params[i].trim();
                if (params[i].indexOf("x") != -1) { // tileSize
                    String[] yx = params[i].split("x");
                    if (yx.length == 2) {
                        this.tileSizeY = Integer.parseInt(yx[0]);
                        this.tileSizeX = Integer.parseInt(yx[1]);
                        this.singleTileMode = false;
                    } else  this.singleTileMode = true;
                } else {
                    try {
                        this.tileBuffer = Double.parseDouble(params[i]);
                    } catch (NumberFormatException ignore) {
                        if (params[i].indexOf("c") != -1) { // Legacy
                            this.log.warn("tileSize switch 'c' is deprecated");
                        }
                    }
                }
            }
        }
    }

    public String getPrefix() {return this.prefix;}
    public void setPrefix(String prefix) {this.prefix = prefix;}
    public int getTileSizeY() {return this.tileSizeY;}
    public void setTileSizeY(int tileSizeY) {this.tileSizeY = tileSizeY;}
    public int getTileSizeX() {return this.tileSizeX;}
    public void setTileSizeX(int tileSizeX) {this.tileSizeX = tileSizeX;}
    public double getTileBuffer() {return this.tileBuffer;}
    public void setTileBuffer(double tileBuffer) {this.tileBuffer = tileBuffer;}
    public boolean isSingleTileMode() {return this.singleTileMode;}
    public void setSingleTileMode(boolean singleTileMode) {
        this.singleTileMode = singleTileMode;}
    public File getWorkDir() {return this.workDir;}
    public void setWorkDir(File workDir) {this.workDir = workDir;}

    // Counter
    public long getNumberOfWays() {return this.nWays;}
    public void setNumberOfWays(long wayCounter) {
        this.nWays = wayCounter;}
    public long getNumberOfRelations() {return this.nRelations;}
    public void setNumberOfRelations(long relationCounter) {
        this.nRelations = relationCounter;}
    public long getNumberOfSharedWays() {return this.nSharedWays;}
    public void setNumberOfSharedWays(long nSharedWays) {
        this.nSharedWays = nSharedWays;}
    public long getNumberOfSharedWaysNodes() {return this.nSharedWaysNodes;}
    public void setNumberOfSharedWaysNodes(long nSharedWaysNodes) {
        this.nSharedWaysNodes = nSharedWaysNodes;}
    public long getNumberOfSharedSegments() {return this.nSharedSegments;}
    public void setNumberOfSharedSegments(long nSharedSegments) {
        this.nSharedSegments = nSharedSegments;}


    /**
     * Ruft intern {@link #tileNode(OsmNode)} auf
     * oder schreibt direkt raus.
     * @param osmNode {@link OsmNode}
     */
    public void tileNode(OsmNode osmNode) {
    	if (this.singleTileMode) {
            if (null == this.singleTile) {
            	this.singleTile = new Tile(0); // -180 / -90
            	this.singleTile.setRelevantTile(true);
                File tileFile = new File(this.workDir, TILED_NODES_FILENAME(0));
                this.singleTile.setOutStreamDisk(new OutStreamDisk(tileFile));
                this.singleTile.getOutStreamDisk().writeByte(
                        VarTypeDesk.typeIdOf(OsmNode.class));
                this.tileMap.put(0, this.singleTile);
            }

    		this.singleTile.incNodeCounter(osmNode.isLatLonOnly());
    		osmNode.writeToStream(this.singleTile.getOutStreamDisk());
    	} else {
    		this.tileNodeIntern(osmNode);
    	}
    }

    /**
     * Schreibt bzw. sortiert einen Knoten in eine oder mehrere
     * betroffene Node-Tile-Dateien.
     * Welche Tile-Dateien betroffen sind, haengt von
     * tileSize und tileBuffer ab.
     *
     * @param osmNode {@link OsmNode}
     */
    private void tileNodeIntern(OsmNode osmNode) {
        double lat = osmNode.getLat();
        double lon = osmNode.getLon();

        // In diesem Tile liegt der Knoten
        int tileId = Tile.getTileId(lat, lon,
                this.tileSizeY, this.tileSizeX);

        // Aufgrund des Puffers auch diese Tiles beruecksichtigen.
        int[] tileIds = Tile.getTileIds(lat, lon,
                this.tileSizeY, this.tileSizeX, this.tileBuffer);

        for (int i = 0; i < tileIds.length; i++) {
            Tile tile = this.tileMap.get(tileIds[i]);

            if (null == tile) {
                tile = new Tile(tileIds[i]);

                // Datei oeffnen.
                File tileFile = new File(
                        this.workDir, TILED_NODES_FILENAME(tile.getId()));
                OutStreamDisk sod = new OutStreamDisk(tileFile);
                sod.writeByte(VarTypeDesk.typeIdOf(OsmNode.class));

                // TileInfos sammeln und mappen
                tile.setOutStreamDisk(sod);
                this.tileMap.put(tileIds[i], tile);
            }

            tile.incNodeCounter(osmNode.isLatLonOnly());

            // Knoten liegt echt im Tile und nicht nur im Pufferbereich
            if (tileId == tileIds[i]) tile.setRelevantTile(true);

            osmNode.writeToStream(tile.getOutStreamDisk());
        }
    }

    /**
     * Wird aufgerufen, wenn alle Knoten einer OSM-Datei gelesen wurden.
     * Schliesst Dateien, vermerkt relevante Informationen fuer eine
     * nachfolgende Serialisierung und loescht ggf. Tiles, welche lediglich
     * Knoten im Pufferbereich beinhalten, also ueberfluessig sind.
     * Abschliessend werden relevante Tiles in ein Array ueberfuehrt und
     * dieses nach Tilegroesse (das mit den meisten Knoten ist oben) sortiert.
     */
    public void closeNodeTiling() {
        int idx = 0;
        int nDeletedTiles = 0;
        this.tileArray = new SortableTArray<Tile>();
        for (IntV<Tile> kvTile : this.tileMap) {
            Tile tileInfo = kvTile.val;

            if (tileInfo.getNumberOfNodes() < this.tileThreshold) {
                this.log.debug("Setting Tile " + Tile.getTileName(tileInfo.getId())
                        + " to not relevant. Number of nodes " +  tileInfo.getNumberOfNodes()
                        + " < " + this.tileThreshold );
                tileInfo.setRelevantTile(false);
            }

            // Datei schliessen und loeschen, falls nicht relevant.
            tileInfo.getOutStreamDisk().close();
            if (tileInfo.isRelevantTile()) {
                // Gueltige Tiles in Array kopieren
                this.tileArray.set(idx++, tileInfo);
            } else {
                // Unwichtige Tiles erst jetzt bekannt.
                // Selbige beinhalten nur tileBuffer-Nodes.
                File file = new File(
                        this.workDir, TILED_NODES_FILENAME(tileInfo.getId()));
                this.log.debug("Deleting file " + file.getName());
                file.delete();
                nDeletedTiles++;
            }
        }
        if (nDeletedTiles > 0) this.log.debug(
                nDeletedTiles + " empty Tiles removed.");

        this.tileMap.clear();
        this.tileArray.sort(); // Die mit den meisten Knoten sind vorne.
    }

    /**
     * Liefert die Anzahl der Tiles.
     * @return
     *      Anzahl Tiles oder -1, wenn {@link #closeNodeTiling()}
     *      noch nicht ausgefuehrt wurde.
     */
    public int getNumberOfTiles() {
        if (null == this.tileArray) return -1;
        return (int) this.tileArray.getSize();
    }

    public Tile getTile(int idx) {
        return this.tileArray.get(idx);
    }

    public final String getTileSizeInfoString() {
        String s = "";
        if (!this.singleTileMode) s = "TileSize is "
                + this.tileSizeY + "x" + this.tileSizeX
                + " with a buffer of " + this.tileBuffer;
        else s = "TileSize disabled, using SingleTileMode";
        return s;
    }

    @Override
    public final String toString() {
        int countTiles = 0;
        if (this.tileArray != null)
            countTiles = (int) this.tileArray.getSize();
        String s = "" // nur Tiler
            + "Prefix:" + this.prefix
            + " Ways:" + DF(this.nWays)
            + " Relations:" + DF(this.nRelations);
        if (this.nSharedWays > 0) { // inkl. Joiner
            s += "\n"
                + "SharedWays:" + DF(this.nSharedWays)
                + " SharedWaysNodes:" + DF(this.nSharedWaysNodes);
        }
        if (this.nSharedWays > 0) { // inkl. Segmenter
            s += "\nSharedSegments:" + DF(this.nSharedSegments);
        }
        if (this.singleTileMode) {
        	s += "\nTiles[1] SingleTileMode";
        } else {
	        s += "\n"
	            + "TileSizeYX:" + this.tileSizeY + "x" + this.tileSizeX
	            + " TileBuffer:" + this.tileBuffer
	            + " Tiles[" + countTiles + "]:";
        }
        for (int i = 0; i < countTiles; i++) {
            Tile tile = this.tileArray.get(i);
            s += "\n[" + i + "] " + tile.toString();
        }
        return s;
    }

    /**
     * Analog {@link #readFromFile()}, jedoch wird hier
     * kontextfrei, also ohne Config, ein TileManager gelesen
     * und auch gleich erzeugt.
     * @param file {@link File}
     * @return new {@link TileManager}
     */
    public static TileManager createFromFile(File file) {
        TileManager tm = new TileManager();
        tm.log = new Log();
        tm.workDir = file.getParentFile();
        tm.readFromFile();
        return tm;
    }

    /**
     * Liest die Infos aus einer bereits serialisierten
     * {@link TileManager} Datei.
     * @return this
     */
    public TileManager readFromFile() {
        // TileManager laden.
        File tmFile = new File(this.workDir, FILENAME);
        this.log.debug("Reading TileFile " + tmFile.getName());

        InStream inStream = new InStreamDisk(tmFile);
        byte varType = inStream.readByte();
        if (varType != VarTypeDesk.typeIdOf(TileManager.class))
            throw new RuntimeException("Unexpected VarType " + varType
                    + " in " + tmFile.getAbsolutePath());
        this.readFromStream(inStream);

        inStream.close();
        this.log.debug(this.toString());
        return this;
    }

    /**
     * Schreibt den {@link TileManager} zurueck als Datei.
     * @return this
     */
    public TileManager writeToFile() {
        File tmFile = new File(this.workDir, FILENAME);
        log.debug("Writing TileInfos for '" + this.prefix + "' to " + tmFile.getName());

        OutStream outStream = new OutStreamDisk(tmFile);
        outStream.writeByte(VarTypeDesk.typeIdOf(TileManager.class));
        this.writeToStream(outStream);
        outStream.close();
        return this;
    }

    public void setTiles(Collection<Tile> tiles) {
        this.tileArray = new SortableTArray<Tile>();
        for (Tile tile : tiles) {
            this.tileArray.push(tile);
        }
        // Sortiere nach Anzahl Knoten im Tile.
        // Das Tile mit den meisten Knoten ist vorne.
        this.tileArray.sort();
    }

    @Override
    public void writeToStream(OutStream outStream) {
        new VarString(this.prefix).writeToStream(outStream);
    	outStream.writeInt(this.singleTileMode ? 1 : 0);
        outStream.writeInt(this.tileSizeY);
        outStream.writeInt(this.tileSizeX);
        outStream.writeLong(Double.doubleToRawLongBits(this.tileBuffer));
        outStream.writeLong(this.nWays);
        outStream.writeLong(this.nRelations);
        outStream.writeLong(this.nSharedWays);
        outStream.writeLong(this.nSharedWaysNodes);
        outStream.writeLong(this.nSharedSegments);
        int countTiles = null == this.tileArray ? 0 : (int) this.tileArray.getSize();
        outStream.writeInt(countTiles);
        for (int i = 0; i < countTiles; i++) {
            this.tileArray.get(i).writeToStream(outStream);
        }
    }

    @Override
    public TileManager readFromStream(InStream inStream) {
        this.prefix = new VarString().readFromStream(inStream).toString();
    	this.singleTileMode = inStream.readInt() == 1;
        this.tileSizeY = inStream.readInt();
        this.tileSizeX = inStream.readInt();
        this.tileBuffer = Double.longBitsToDouble(inStream.readLong());
        this.nWays = inStream.readLong();
        this.nRelations = inStream.readLong();
        this.nSharedWays = inStream.readLong();
        this.nSharedWaysNodes = inStream.readLong();
        this.nSharedSegments = inStream.readLong();
        this.tileArray = new SortableTArray<Tile>();
        int countTiles = inStream.readInt();
        for (int i = 0; i < countTiles; i++) {
            this.tileArray.set(i, new Tile().readFromStream(inStream));
        }

        return this;
    }

    /**
     * Wichtige Schaetzfunktion fuer das derzeit
     * verfuegbare Memory.
     * @return long RAM (Heap).
     */
    private long estimateRemainingRam() {
        long ram = Utils.getMemoryAvailable();
        // default (Config) geschaetzte 32MB SystemResourcen
        ram -= this.reservedXmx;
        if (ram < 0) ram = 0;
        return ram;
    }

    /**
     * Reduziert alle Tiles auf eine Menge referenzierter Knoten.
     * @param ndRefs {@link LongArray} Ids von Knoten.
     */
    public void compressTiles(LongArray ndRefs) {
        int nTiles = this.getNumberOfTiles();
        for (int i = 0; i < nTiles; i++) {
            Tile tile = this.getTile(i);
            tile.resetNodeCounter();
            File tileFile = new File(
                    this.workDir, TILED_NODES_FILENAME(tile.getId()));
            this.log.progress("Compressing "
                    + tileFile.getName() + " (" + (i+1) + " of " + nTiles + ")");
            File dir = tileFile.getParentFile();
            File tmpFile = new File(dir, tileFile.getName() + ".tmp");
            InStream inStream = new InStreamDisk(tileFile);
            OutStream outStream = new OutStreamDisk(tmpFile);

            byte typeId = inStream.readByte();
            outStream.writeByte(typeId);

            long nRead = 0, nWritten = 0;

            OsmNode osmNode = new OsmNode();
            while (!inStream.isEof()) {
                nRead++;
                osmNode.readFromStream(inStream);
                if (ndRefs.contains(osmNode.getId())) {
                    nWritten++;
                    tile.incNodeCounter(osmNode.isLatLonOnly());
                    osmNode.writeToStream(outStream);
                }
                if (nRead % 1000000 == 0) this.log.progress(
                        tileFile.getName() + " compressed from "
                        + DF(nRead) + " to " + DF(nWritten) + " elements");
            }

            inStream.close();
            outStream.close();

            tileFile.delete();
            tmpFile.renameTo(tileFile);

            this.log.info(tileFile.getName() + " compressed from "
                    + DF(nRead) + " to " + DF(nWritten) + " elements");

        }
    }

    /**
     * Durchlaeuft alle Tiles (Dateien), filtert die Knoten,
     * deren Ids an diese Methode uebergeben wurden
     * und erstellt daraus ein kuenstliches Tile.<br>
     * <br><b>Wichtig:</b><br>
     * Zur Sicherheit wird hier {@link LongArray#shrinkToSet()}
     * fuer das uebergebene Arrays aufgerufen.
     * @param ndRefs {@link LongArray} IDs der Knoten (Filter).
     * @return {@link Tile}
     */
    public Tile createTileFromNodeIds(LongArray ndRefs) {
        Tile rescueTile = new Tile();
        this.log.info("Building set of " + DF(ndRefs.getSize()) + " NodeRefs");
        ndRefs.shrinkToSet(); // Implizit Sort
        int nTiles = this.getNumberOfTiles();
        this.log.info("Creating pseudo tile for " + DF(ndRefs.getSize()) + " NodeRefs");
        for (int i = 0; i < nTiles; i++) {
            Tile tile = this.getTile(i);
            File nodeFile = new File(
                    this.workDir, TILED_NODES_FILENAME(tile.getId()));
            this.log.progress("Filtering "
                    + nodeFile.getName() + " (" + (i+1) + " of " + nTiles + ")");
            InStream nodesInStream = new InStreamDisk(nodeFile);
            nodesInStream.readByte(); // Type
            OsmNode osmNode = new OsmNode();
            while (!nodesInStream.isEof()) {
                osmNode.readFromStream(nodesInStream);
                if (ndRefs.contains(osmNode.getId())) {
                    rescueTile.cacheOsmNode(osmNode);
                }
            }
            nodesInStream.close();
        }
        return rescueTile;
    }

    /**
     * Sehr spezieller Iterator, der den Speicherverbrauch aller Knoten
     * eines Tiles abschaetzt und entsprechend pro Durchlauf eine
     * maximale Submenge an Tiles cached, die so eben noch in das RAM passen
     * und so lange iteriert, bis alle Tiles zurueckgeliefert wurden.<br>
     * Man iteriert hier also ueber eine bereits gecachte Teilmenge
     * von Knoten, die in einem oder mehreren Tiles liegen koennen.
     *
     * @see java.lang.Iterable#iterator()
     */
    @Override
    public Iterator<List<Tile>> iterator() {
        return new Iterator<List<Tile>>() {
            private List<Tile> remainingList = tileArray.asList();
            private List<Tile> nextList = new ArrayList<Tile>();

            @Override
            public boolean hasNext() {
                for (Tile tile : this.nextList) tile.clearCache();
                this.nextList.clear();
                return !this.remainingList.isEmpty();
            }

            @Override
            public List<Tile> next() {
                log.debug("Caching nodes. " + this.remainingList.size() + " Tiles remaining");

                //  mindestens ein Tile cachen, unabhaengig vom RAM.
                Tile tile = this.remainingList.get(0);
                this.nextList.add(tile);
                if (!tile.cacheOsmNodes(new File(workDir,
                        TileManager.TILED_NODES_FILENAME(tile.getId())),
                        log)) {
                	throw new OutOfMemoryError("Increase Java Heap Space: -Xmx");
                }

                // Keine 1000 Tiles auf einmal cachen ggf. verhindern
                int maxTilesPerLoop = TileManager.this.maxTilesPerLoop;

                // weitere Tiles cachen, wenn vorhanden und genuegend RAM
                System.gc(); // GC gezielt rufen.
                for (int i = 1; i < this.remainingList.size(); i++) {
                    tile = this.remainingList.get(i);
                    long ramNeeded = tile.estimateRamUsage();
                    long ramRemaining = estimateRemainingRam();
                    if (ramNeeded < ramRemaining) {
                    	boolean isCached = tile.cacheOsmNodes(
                    			new File(workDir,
                    					TileManager.TILED_NODES_FILENAME(tile.getId())),
                    					log);
                    	System.gc(); // GC gezielt rufen.
                    	if (!isCached) {
                    		// Aktuellen RAM-Stand merken (hier isser abgeschmiert)
                    		long ram = Utils.getMemoryAvailable();
                    		// SystemRAM neu kalibrieren
                    		ram *= 1.1; // Sicher ist sicher.
                    		log.warn("Increasing Reserved Xmx to " + DF(ram));
                    		reservedXmx = ram;
                    	} else {
                    		this.nextList.add(tile);
                    	}

                        if (maxTilesPerLoop <= this.nextList.size()) {
                            log.debug("Maximum number of tiles " + maxTilesPerLoop
                                    + " reached. Breaking loop. - " + MB());
                            break;
                        }
                    }
                }

                this.remainingList.removeAll(this.nextList);
                return this.nextList;
            }

            @Override
            public void remove() {
                throw new RuntimeException(
                        this.getClass().getName() + ".remove() not supported");
            }
        };
    }


}
