/*
 * $Id: DbExtension.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.extensions;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.postp.DbExtensionBuilder;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.LongLongBag;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.RandomRead;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.Graph;

/**
 * Sammlung von Datenbank-Tabellen und -Sichten als {@link GraphExtension}.
 * <br><br><b>Wichtig:</b><br>
 * Lediglich der Index wird in das RAM geladen, die Daten verbleiben auf
 * Platte ({@link Graph}-Datei). Eine Verwendung mit
 * {@link Graph#Graph(java.io.InputStream)} oder {@link Config#loadBeans(String)}
 * ist daher derzeit <strong>nicht moeglich</strong>.
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 * @see DbExtensionBuilder
 */
public class DbExtension extends GraphExtension<DbExtension> {

    private RandomRead rr;
    private long filePos;
    private Log log;

    private Map<String, DbView> dbViewMap = new HashMap<String, DbView>();

    private static class DbView {
        String key;
        Class<?> varClass;
        LongLongBag index;
        long dataSize;
        long dataPosition;
    }

    @Override
    public DbExtension readFromStream(InStream inStream) {
        // Muss auch ohne connect() aufrufbar sein. Achtung: Log
        if (null == this.log) this.log = Log.empty();

        // wenn kein connect(), dann fileOffset=0
        long readIndex = inStream.getReadIndex();

        List<DbView> dbViewList = new ArrayList<DbView>();

        int nViews = inStream.readInt();
        for (int v = 0; v < nViews; v++) {
            try {
                DbView dbView = new DbView();
                String className = new VarString().readFromStream(inStream).toString();
                dbView.varClass = Config.loadClass(className);
                dbView.key = new VarString().readFromStream(inStream).toString();
                dbView.dataSize = inStream.readLong();
                dbView.index = new LongLongBag().readFromStream(inStream);
                dbViewList.add(dbView);

            } catch (ClassNotFoundException e) {
                log.error(e);
            }
        }

        long headerSize = inStream.getReadIndex() - readIndex;
        long dataPosition = headerSize + this.filePos;
        for (DbView dbView : dbViewList) {
            dbView.dataPosition = dataPosition;
            dataPosition += dbView.dataSize;
            this.dbViewMap.put(dbView.key, dbView);
        }

        if (null == this.rr) throw new IllegalStateException(
                this.getClass().getSimpleName() + ": RAM-Mode NOT supported");

        return this;
    }

    @Override
    public void connect(File file, long offset, Log log, Properties props) {
        this.rr = new RandomAccessDisk(file, true);
        this.filePos = offset;
        this.log = log;
    }

    @Override
    public void disconnect() {
        if (this.rr != null) this.rr.close();
    }

    public final synchronized Var<?>[] get(String viewKey, long index) {
        try {
            DbView dbView = this.dbViewMap.get(viewKey);
            if (null == dbView) return null;

            long[] positions = dbView.index.getAll(index);
            if (null == positions) return null;

            Var<?>[] vars = new Var[positions.length];
            for (int i = 0; i < positions.length; i++) {
                long position = positions[i] + dbView.dataPosition;
                this.rr.setReadIndex(position);
                vars[i] = (Var<?>) dbView.varClass.getDeclaredConstructor().newInstance();
                vars[i].readFromStream(this.rr);
            }

            return vars;

        } catch (Exception e) {
            this.log.error(e);
            return null;
        }
    }

}
