/*
 * $Id: GeoExtensionBase.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.extensions;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.ZoneUtils.createZoneId;
import static de.cm.osm2po.postp.GeoExtensionBuilderBase.COMPAT_VERSION;

import java.io.File;
import java.util.Properties;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.OsmObject;
import de.cm.osm2po.postp.GeoExtensionBuilderBase;
import de.cm.osm2po.primitives.Dictionary;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.IndexRaw;
import de.cm.osm2po.primitives.IntIntBag;
import de.cm.osm2po.primitives.LongIntBag;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.RandomAccessMemory;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarByteArray;

/**
 * Abstrakte Basisklasse fuer {@link GraphExtension}s die vom
 * {@link GeoExtensionBuilderBase} erzeugt wurden.
 * Bietet komplexe Strukturen und Zugriffe fuer eine Sammlung
 * von beliebigen {@link OsmObject}s.<br><br><b>Wichtig:</b><br>
 * Zur Laufzeit versucht diese Implementation alle Daten in den Speicher
 * zu laden. Tritt dabei ein OutOfMemoryError auf, so wird eine direkte
 * Verbindung zur uebergebenen Bean- oder Graph-Datei aufgebaut.
 * Die Parameter in {@link #connect(File, long, Log, Properties)}
 * sind jedoch alle optional. Ist File <code>null</code>, so wurde der Graph nicht
 * aus einer lokalen Datei eingelesen, sondern aus einem Stream, so dass
 * ein Fallback nicht moeglich ist.
 *
 * @author (c) 2017 - Carsten Moeller - info@osm2po.de
 * @see GraphExtension
 * @see GeoExtensionBuilderBase
 * @see #findByCoordinate(double, double, ClazzFlagPredicate)
 * @see #findByName(String, boolean)
 * @see #findByIndex(int, Var)
 * @param <T> T extends {@link OsmObject}
 */
public abstract class GeoExtensionBase<T extends Var<?> & OsmObject>
extends GraphExtension<GeoExtensionBase<T>> {

    private IndexRaw<T> index;
    private Dictionary dic;
    private IntIntBag objZoneBag;
    private File file;
    private RandomAccess ra;
    private Log log;
    private boolean fileMode;
    private byte zoneSize;
    private LongIntBag id2IdxBag;
    private byte[] idxClazzes;

    /**
     * Liefert ein Objekt-Array zu einer Koordinate.<br>
     * Die Implementation haengt vom Geometrie-Typ ab, daher abstract.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @param filter {@link ClazzFlagPredicate} NULLABLE.
     * @return Gefundene {@link OsmObject}s oder null.
     */
    public abstract T[] findByCoordinate(double lat, double lon, ClazzFlagPredicate filter);

    protected final Log getLog() {return this.log;}
    public final int getZoneSize() {return this.zoneSize;}

    @Override
    public final void connect(File file, long offset, Log log, Properties props) {
        this.file = file;
        this.log = null == log ? Log.empty() : log;
        // TODO Extension theoretisch mehrfach moeglich. fileMode aber global
        this.fileMode = file != null && props != null && Boolean.valueOf(
                props.getProperty(this.getClass().getSimpleName() + ".fileMode"));
    }

    @Override
    public final void disconnect() {
        this.ra.close();
    }

    /**
     * Liefert die Groesse des internen Index, also
     * die Anzahl der gespeicherten Elemente.
     * @return int Anzahl der Eintraege
     */
    public final int getIndexLength() {
        return this.index.getIndexSize();
    }

    /**
     * Liefert die {@link OsmObject#getClazz()} des Elements
     * an einem uebergebenen Index. Dies jedoch ohne das komplette
     * Objekt vorher zu deserialisieren.
     * @param index int NOT CHECKED
     * @return byte clazz
     */
    public final byte findClazzByIndex(int index) {
        return this.idxClazzes[index];
    }

    /**
     * Findet den Index zu einer Id.
     * <br><br><b>Wichtig:</b><br>
     * Haben mehrere Objekt die gleiche ID, so liefert diese
     * Methode nur den zuerst gefundenen.
     * @param id long
     * @return int Index oder -1, wenn nicht gefunden
     */
    public final int findIndexForId(long id) {
        int idx = this.id2IdxBag.get(id);
        if (LongIntBag.NULLVALUE == idx) {
            return -1;
        }
        return idx;
    }

    /**
     * Findet ein Object aufgrund seines Index.
     * @param index int Index / Position
     * @param osmObject {@link OsmObject} in das gelesen werden soll. NOT NULL.
     * @return {@link OsmObject} NULL, wenn index out of range
     */
    public final T findByIndex(int index, T osmObject) {
        T obj = this.index.fetch(index, osmObject);
        if (null == obj) return null;
        obj.setIndex(index);
        return obj;
    }

    /**
     * Findet ein Objekt aufgrund seiner Id.
     * <br><br><b>Wichtig:</b><br>
     * Haben mehrere Objekte die gleiche ID, so liefert diese
     * Methode nur die zuerst gefundene.
     * @param id long
     * @param osmObject {@link OsmObject} in das gelesen werden soll. NOT NULL.
     * @return {@link OsmObject} NULL, wenn index out of range
     */
    public final T findById(long id, T osmObject) {
        int idx = this.findIndexForId(id);
        return findByIndex(idx, osmObject);
    }


    /**
     * Liefert die Indexe aller Objekte aus einer Zone.
     * @param zoneId int
     * @return Indizes der Objekte oder null.
     */
    public final int[] findIndexesForZone(int zoneId) {
        return this.objZoneBag.getAll(zoneId);
    }

    /**
     * Liefert die Indexe aller Objekte aus einer Zone, die
     * eine Koordinate enthaelt.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @return Indizes der Objekte oder null.
     */
    public final int[] findIndexesForZone(double lat, double lon) {
        int zoneId = createZoneId(lat, lon, this.zoneSize);
        return this.findIndexesForZone(zoneId);
    }

    /**
     * Simple Pruefung, ob eine Koordinate in den Zones enthalten ist.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @return boolean
     */
    public final boolean isNearby(double lat, double lon) {
        int zoneId = createZoneId(lat, lon, this.zoneSize);
        int nearZone = this.objZoneBag.get(zoneId);
        boolean nearby = nearZone != IntIntBag.NULLVALUE;
        return nearby;
    }

    /**
     * Findet Objekt-Indizes aufgrund ihres Namens.
     * Die Objekte selbst koennen danach mittels
     * {@link #findByIndex(int, Var)} gelesen werden.
     * @param name String Name NULLABLE
     * @param exact boolean false: startsWith()-Suche
     * @return int Array von assoziierten Ids oder null
     */
    public final int[] findByName(final String name, boolean exact) {
        int[] foundIds = null;
        if (name != null) {
            if (exact) foundIds = this.dic.get(name.toLowerCase());
            else foundIds = this.dic.findByStartsWith(name.toLowerCase());
        }
        return foundIds;
    }

    @Override
    public final GeoExtensionBase<T> readFromStream(InStream inStream) {
        // Muss auch ohne connect() aufrufbar sein. Achtung: Log
        if (null == this.log) this.log = Log.empty();

        // FIXME funktioniert nicht bei isolierten graph.loadExtension()
        long ptrOffset = inStream.getReadIndex();

        int compatVersion = inStream.readInt();
        if (compatVersion != COMPAT_VERSION) throw new RuntimeException(
                "Incompatible Extension Version " + compatVersion
                + ". Version " + COMPAT_VERSION + " required!");

        // Dateiposition innerhalb des Extension-Blocks. Hier Vorsicht!
        this.zoneSize = inStream.readByte();
        long ptrIndex = inStream.readLong();
        long ptrDic = inStream.readLong();
        long ptrEof = inStream.readLong();

        // Schnellfilter auf clazz
        this.idxClazzes = new VarByteArray().readFromStream(inStream).getBytes();

        // Schnellfilter fuer Mapping von ID zu Idx
        this.id2IdxBag = new LongIntBag().readFromStream(inStream);

        // ZoneBag immer cachen
        this.objZoneBag = new IntIntBag().readFromStream(inStream);
        long objZoneBagSize = inStream.getReadIndex() - ptrOffset;
        this.log.debug(DF(this.objZoneBag.getSize()) + " zones cached " + " ["
                + DF2((double)objZoneBagSize / 0x100000) + "M]");

        long sizeWithoutPreCache = ptrEof - ptrIndex;
        if (ptrOffset + ptrIndex != inStream.getReadIndex())
            throw new RuntimeException("Corrupt File Position of Index");

        System.gc();

        if (this.fileMode) {
            this.log.info(this.getClass().getSimpleName() + " uses fileMode");
            this.ra = new RandomAccessDisk(this.file, true);
        } else {
            long nLeft = sizeWithoutPreCache;
            try {
                this.ra = new RandomAccessMemory();
                byte buf[] = new byte[0x1000];

                long loop = 0;
                while (nLeft > 0L) {
                    if (nLeft < buf.length) {
                        buf = new byte[(int) nLeft];
                    }
                    int nRead = inStream.readBytes(buf);
                    nLeft -= nRead;
                    this.ra.writeBytes(buf, nRead); // Hier OME moeglich!
                    // Test: throw new OutOfMemoryError();

                    long nReadTotal = sizeWithoutPreCache - nLeft;
                    if (loop++ % 0x1000 == 0) this.log.progress(DF(nReadTotal) + " of "
                            + DF(sizeWithoutPreCache) + " bytes cached - " + MB());
                }
                this.log.info(DF(sizeWithoutPreCache) + " bytes cached - " + MB());

                // Wenn im RAM, dann Pointer anpassen!
                ptrDic -= ptrIndex;
                ptrIndex = 0;
                ptrOffset = 0; // Wichtig!

            } catch (OutOfMemoryError ome) { // FallBack
                this.ra.clear();
                if (null == this.file) throw ome;
                this.log.error("Out of memory. Falling back to isolated Disk Access");
                this.ra = new RandomAccessDisk(this.file, true);
                inStream.skip(nLeft); // Daten im Eingabestrom ueberspringen
            }
        }

        this.ra.setReadIndex(ptrIndex + ptrOffset);
        this.index = new IndexRaw<T>();
        this.index.connect(this.ra);

        this.ra.setReadIndex(ptrDic + ptrOffset);
        this.dic = new Dictionary();
        this.dic.connect(this.ra);

        return this;
    }

}