/*
 * $Id: GraphExtension.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.extensions;

import java.io.File;
import java.util.Properties;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.routing.Graph;

/**
 * Interface fuer komplexere {@link Graph}-Erweiterungen, bei denen eine
 * einfache (De-)Serialisierung mit {@link Var} zur Laufzeit nicht ausreicht.
 * <br><br><b>Wichtig:</b><br>
 * Der Standardmechanismus ist bereits mit einfachen {@link Var}s
 * als Erweiterung zufrieden. Diese werden vom Graphen sofort oder
 * nachgelagert komplett in den Speicher eingelesen. In komplexen Szenarien,
 * besonders dann, wenn keine simplen Objekte mehr transportiert werden,
 * muss es moeglich sein, die Steuerung an die Erweiterung selbst zu
 * uebertragen.
 *
 * @see BndExtension
 * @see SndExtension
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 * @param <T> Class of T
 */
public abstract class GraphExtension<T> implements Var<GraphExtension<T>> {
    /**
     * Wird zur Laufzeit direkt vor
     * {@link #readFromStream(de.cm.osm2po.primitives.InStream)}
     * aufgerufen.
     * @param file File, z.B. Graph-Datei fuer Fallback/DiskMode. NULLABLE
     * @param offset long Dateiposition der Quelle
     * @param log {@link Log} optionaler Logger, NULLABLE
     * @param props zusaetzliche {@link Properties} NULLABLE.
     */
    public abstract void connect(
            File file, long offset, Log log, Properties props);

    /**
     * Wird zur Laufzeit von {@link Graph#close()} aufgerufen,
     * um hier etwaige Resourcen wieder freizugeben.
     */
    public abstract void disconnect();

    @Override
    public final void writeToStream(OutStream outStream) {
        throw new UnsupportedOperationException(
                this.getClass().getSimpleName() + "s do not support writeToStream()");
    }
}
