/*
 * $Id: PoiExtension.java 5490 2021-12-21 21:00:35Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.extensions;
import static de.cm.osm2po.misc.GeoUtils.calcDistance;
import static de.cm.osm2po.misc.ZoneUtils.createZoneId;

import de.cm.osm2po.model.ClazzFlagPredicate;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.postp.GeoExtensionBuilderBase;
import de.cm.osm2po.postp.PoiExtensionBuilder2po;
import de.cm.osm2po.postp.PoiExtensionBuilderDb;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.VarNum;

/**
 * Bietet komplexe Strukturen und Zugriffe fuer eine Sammlung
 * von {@link OsmNode}s.
 * @author (c) 2017 - Carsten Moeller - info@osm2po.de
 * @see #findClosestPoiByCoordinate(double, double, ClazzFlagPredicate)
 * @see #findCloserPoiByCoordinate(double, double, ClazzFlagPredicate, double, boolean)
 * @see PoiExtensionBuilderDb
 * @see PoiExtensionBuilder2po
 * @see GeoExtensionBase
 * @see GeoExtensionBuilderBase
 */
public class PoiExtension extends GeoExtensionBase<OsmNode> {

    /** Kapselt einen Index und eine Distanz */
    public static final class PoiIndexAndDistance {
        public int index;
        public double distance;
    }

    private static final OsmNode essentialNode = new OsmNode() {
        @Override
        public void writeToStream(OutStream outStream) {
            throw new RuntimeException("writeToStream not supported: " + this.getClass().getSimpleName());
        }

        @Override
        public OsmNode readFromStream(InStream inStream) {
            this.clear();

            this.coord = inStream.readLong();
            int flags = inStream.readByte() & 0xFF;

            // De-Kompression
            if ((flags & INTERN_FLAG_HAS_ID) != 0)
                this.setId(VarNum.readUnsigned(inStream));
            if ((flags & INTERN_FLAG_HAS_CLAZZ) != 0)
                this.setClazz(inStream.readByte());
            if ((flags & INTERN_FLAG_HAS_FLAGS) != 0)
                this.setFlags((int) VarNum.readUnsigned(inStream));

            // SpeedUp: Name and Custom omitted
            return this;
        }
    };

    @Override
    public OsmNode[] findByCoordinate(double lat, double lon,
            ClazzFlagPredicate filter) {
        OsmNode osmNode = this.findClosestPoiByCoordinate(lat, lon, filter);
        return (null == osmNode) ? null : new OsmNode[]{osmNode};
    }

    /**
     * Findet den dichtesten POI zu einer Koordinate.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @param filter {@link ClazzFlagPredicate}
     * @return {@link OsmNode} oder null.
     */
    public OsmNode findClosestPoiByCoordinate(
            double lat, double lon, ClazzFlagPredicate filter) {
        OsmNode closestPoi = null;
        PoiIndexAndDistance closestIndex = findCloserPoiByCoordinate(
                lat, lon, filter, Double.MAX_VALUE, false);
        if (closestIndex != null) { // Vollstaendig nachladen
            closestPoi = this.findByIndex(closestIndex.index, new OsmNode());
        }
        return closestPoi;
    }

    /**
     * Findet den dichtesten POI ausgehend von einer Maximal-Distanz.<br>
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @param filter {@link ClazzFlagPredicate}
     * @param currentDist double Maximale Distanz
     * @param returnFirstFound boolean true: liefert einen beliebigen ersten Treffer
     * @return {@link PoiIndexAndDistance}, also PoiIndex und die Distanz oder <code>null</code>
     */
    synchronized public final PoiIndexAndDistance findCloserPoiByCoordinate(double lat, double lon,
            ClazzFlagPredicate filter, double currentDist, boolean returnFirstFound) {
        int zoneId = createZoneId(lat, lon, this.getZoneSize());
        int[] poiIdxs = this.findIndexesForZone(zoneId);
        PoiIndexAndDistance closestPoi = null;
        if (poiIdxs != null) for (int i = 0; i < poiIdxs.length; i++) {
            int idx = poiIdxs[i];
            OsmNode poi = this.findByIndex(idx, essentialNode); // poiHolder must be synchronized
            if (null == filter || filter.matches(poi)) {
                double dist = calcDistance(lat, lon, poi.getLat(), poi.getLon());
                if (dist < currentDist) {
                    if (null == closestPoi) closestPoi = new PoiIndexAndDistance();
                    currentDist = dist;
                    closestPoi.index = idx;
                    closestPoi.distance = dist;
                    if (returnFirstFound) break;
                }
            }
        }
        return closestPoi; // null if not found
    }

}