/*
 * $Id: SndExtension.java 5243 2021-02-09 21:03:31Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.extensions;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.postp.SndExtensionBuilder.COMPAT_VERSION;

import java.io.File;
import java.util.Properties;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.Dictionary;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.RandomAccessMemory;
import de.cm.osm2po.routing.Graph;

/**
 * Dictionary-Extension fuer alle Strassen-Namen.
 * <br><br><b>Wichtig:</b><br>
 * Zur Laufzeit versucht diese Implementation alle Daten in den Speicher
 * zu laden. Tritt dabei ein OutOfMemoryError auf, so wird eine direkte
 * Verbindung zur uebergebenen Graph-Datei aufgebaut.
 * Die Parameter in {@link #connect(File, long, Log, Properties)}
 * sind jedoch optional. Ist File <code>null</code>, so wurde der Graph nicht
 * aus einer lokalen Datei eingelesen und ein Fallback ist nicht moeglich.
 *
 * @author (c) 2017 - Carsten Moeller - info@osm2po.de
 */
public class SndExtension extends GraphExtension<SndExtension> {

    private File file;
    private RandomAccess ra;
    private Log log;
    private boolean fileMode;

    private Dictionary dic;

    @Override
    public void connect(File file, long offset, Log log, Properties props) {
        this.file = file;
        this.log = null == log ? Log.empty() : log;
        this.fileMode = file != null && props != null && Boolean.valueOf(
                props.getProperty(this.getClass().getSimpleName() + ".fileMode"));
    }

    @Override
    public void disconnect() {
        this.ra.close();
    }

    @Override
    public SndExtension readFromStream(InStream inStream) {
        // Muss auch ohne connect() aufrufbar sein. Achtung: Log
        if (null == this.log) this.log = Log.empty();

        // FIXME funktioniert nicht bei isolierten graph.loadExtension()
        long ptrOffset = inStream.getReadIndex();

        int compatVersion = inStream.readInt();
        if (compatVersion != COMPAT_VERSION) throw new RuntimeException(
                "Incompatible Extension Version " + compatVersion
                + ". Version " + COMPAT_VERSION + " required!");

        long ptrDic = inStream.readLong();
        long ptrEof = inStream.readLong();

        long sizeWithoutHeader = ptrEof - ptrDic;

        System.gc();

        if (this.fileMode) {
            log.info(this.getClass().getSimpleName() + " uses fileMode");
            this.ra = new RandomAccessDisk(this.file, true);

        } else {
            long nLeft = sizeWithoutHeader;
            try {
                this.ra = new RandomAccessMemory();
                byte buf[] = new byte[0x1000];

                long loop = 0;
                while (nLeft > 0L) {
                    if (nLeft < buf.length) {
                        buf = new byte[(int) nLeft];
                    }
                    int nRead = inStream.readBytes(buf);
                    nLeft -= nRead;
                    this.ra.writeBytes(buf, nRead); // Hier OME moeglich!
                    // Test: throw new OutOfMemoryError();

                    long nReadTotal = sizeWithoutHeader - nLeft;
                    if (loop++ % 0x1000 == 0) log.progress(DF(nReadTotal) + " of "
                            + DF(sizeWithoutHeader) + " bytes cached - " + MB());
                }
                log.info(DF(sizeWithoutHeader) + " bytes cached - " + MB());

                // Wenn im RAM, dann Pointer anpassen!
                ptrDic = 0;
                ptrOffset = 0; // Wichtig

            } catch (OutOfMemoryError ome) { // FallBack
                this.ra.clear();
                if (null == this.file) throw ome;
                log.error("Out of memory. Falling back to isolated Disk Access");
                this.ra = new RandomAccessDisk(this.file, true);
                inStream.skip(nLeft); // Daten im Eingabestrom ueberspringen
            }
        }

        ra.setReadIndex(ptrDic + ptrOffset);
        this.dic = new Dictionary();
        this.dic.connect(this.ra);

        return this;
    }

    /**
     * Findet Strassen-Indizes aufgrund ihres Namens.
     * Der Name selbst kann danach mittels
     * {@link Graph#lookupEdgeName(int)} gelesen werden.
     * @param name String Name NULLABLE
     * @param exact boolean false: startsWith()-Suche
     * @return int Array von assoziierten EdgeIdxs oder null
     */
    public final int[] findByName(final String name, boolean exact) {
        int[] foundIds = null;
        if (name != null) {
            if (exact) foundIds = this.dic.get(name.toLowerCase());
            else foundIds = this.dic.findByStartsWith(name.toLowerCase());
        }
        return foundIds;
    }


}