/*
 * $Id: ConcaveHull.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.hull;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.DF2;
import static java.lang.Integer.MAX_VALUE;
import static java.lang.StrictMath.hypot;
import static java.lang.System.currentTimeMillis;
import static java.util.Arrays.copyOf;

import java.util.Arrays;
import java.util.PriorityQueue;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.VarBooleanArray;
import de.cm.osm2po.routing.Graph;

/**
 * Bildet die konkave Huelle um eine Menge von Punkten.
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 */
public class ConcaveHull {
    private Log log;
    private float[] lats;
    private float[] lons;
    private ConvexHull convexHull;
    private HullRaster hullRaster;
    private PriorityQueue<HullSegment> pq;
    private VarBooleanArray vidSet;
    private float alpha;

    /**
     * CTOR.
     * @param graph {@link Graph} NOT NULL
     * @param log {@link Log} NULLABLE
     */
    public ConcaveHull(Graph graph, Log log) {
        this.convexHull = new ConvexHull(graph, log);
        this.hullRaster = new HullRaster(graph, log);
        this.pq = new PriorityQueue<HullSegment>();
        this.lats = graph.getLats();
        this.lons = graph.getLons();
        this.log = null == log ? Log.empty() : log;
    }

    /**
     * Liefert das verwendete Raster.
     * @return {@link HullRaster}
     */
    public HullRaster getHullRaster() {
        return this.hullRaster;
    }

    /**
     * Baut die Huelle.<br>
     * Ruft {@link #build(int[], float, int)} mit max=0, d.h.
     * der Build-Prozess wird nicht ab einer gewissen Anzahl
     * an Schritten abgewuergt.
     * @param vids int[] Array von VertexIds
     * @param alpha float minimale Laenge eines Huellensegments
     * @return int[] Array von VertexIds der Huelle - linkslaeufig.
     */
    public int[] build(int[] vids, float alpha) {
        return this.build(vids, alpha, 0);
    }

    /**
     * Baut die Huelle.<br>
     * @param vids int[] Array von VertexIds
     * @param alpha float minimale Laenge eines Huellensegments
     * @param maxSteps Begrenzung der Schritte, 0 = keine Begrenzung.
     * @return int[] Array von VertexIds der Huelle - linkslaeufig.
	 */
    public int[] build(int[] vids, float alpha, int maxSteps) {
	    if (maxSteps <= 0) maxSteps = MAX_VALUE;
	    final float gridSize = 0.01f; // ca. 1 km

	    log.note("Building concave hull of " + DF(vids.length) + " vertices");
	    log.debug("with alpha " + alpha + " and gridSize " + gridSize);
	    long ts = currentTimeMillis();

	    log.progress("Building initial hull");
	    int[] convexHull = this.convexHull.build(vids);
	    if (0f == alpha) return convexHull;

	    log.progress("Building grid");
	    this.hullRaster.build(vids, gridSize);
	    this.vidSet = new VarBooleanArray(this.hullRaster.getMaxVertexId() + 1);
	    this.alpha = alpha;

	    log.progress("Building initial convex ring of segments");
	    HullSegment hs = null;
	    for (int i = 0; i < convexHull.length; i++) {
	        int vid = convexHull[i];
	        this.vidSet.set(vid, true); // Vertex bereits Teil der Huelle
	        hs = null == hs ? new HullSegment(vid) : hs.insertNext(vid);
	    }

	    log.progress("Calculating initial hull priority");
	    if (hs != null) for (HullSegment h : hs) {
	        int vs = h.getVertexId(), vt = h.getNext().getVertexId();
	        h.setPriority(this.calcLength(vs, vt));
	        this.pq.add(h);
	    }

	    log.progress("Building concave hull");
	    for (int i = 0; i < maxSteps; i++) if (!this.processNext()) break;

	    ts = currentTimeMillis() - ts;
	    log.note("Concave Hull built in " + DF2(ts / 1e3) + "s");

	    log.progress("Builing Vertex-List of concave hull");
	    int[] concaveHull = new int[0x100];
	    int n = 0; // Analog convexHull: Keine Ring-Bildung!
	    if (hs != null) for (HullSegment h : hs) {
	        if (concaveHull.length == n) concaveHull = copyOf(concaveHull, n * 2);
	        concaveHull[n++] = h.getVertexId();
	    }

	    return Arrays.copyOf(concaveHull, n);
	}

	/**
	 * Liefert das naechste zu verarbeitende HullSegment in der PQ.
	 * @return {@link HullSegment} oder null
	 */
	public HullSegment peekNext() {
	    if (this.pq.isEmpty()) return null;
	    return this.pq.peek();
	}

    /**
     * Verarbeitet das naechste HullSegment nach Prioritaet.
     * Dabei entstehen entweder zwei neue Segmente oder
     * das aktuelle wird als Teil der finalen Huelle
     * akzeptiert.
     * @return false: Keine weiteren Segmente in der PQ.
     */
    private boolean processNext() {
        if (this.pq.isEmpty()) return false;
        HullSegment hs = this.pq.remove();
        if (hs.getPriority() < this.alpha) return true;
        int v = this.fincClosestVertexId(hs);
        if (v <= 0) return true;
        if (probeHullIntersection(hs, v)) return true;
        this.vidSet.set(v, true); // neuen Punkt akzeptieren

        // Segment ueber diesen Vertex fuehren
        int v1 = hs.getVertexId();
        int v2 = hs.getNext().getVertexId();
        hs.setPriority(this.calcLength(v1, v));
        this.pq.add(hs);
        hs = hs.insertNext(v);
        hs.setPriority(this.calcLength(v, v2));
        this.pq.add(hs);

        return true;
    }

    /**
     * Findet den naechsten Vertex an einem HullSegment.
     * @param hs {@link HullSegment}
     * @return VertexId oder 0, wenn nicht gefunden.
     */
    private int fincClosestVertexId(HullSegment hs) {
        int sourceId = hs.getVertexId();
        int targetId = hs.getNext().getVertexId();
        double y1 = this.lats[sourceId], x1 = this.lons[sourceId];
        double y2 = this.lats[targetId], x2 = this.lons[targetId];

        float maxDepth = hs.getPriority() * 3; // Priority is SegmentLength
        int closestVid = 0;
        double minDist = MAX_VALUE;
        for (float depth = 0.01f; 0 == closestVid && depth < maxDepth; depth *= 2f) {
            int[] searchAreaCells = this.hullRaster.rasterLine(y1, x1, y2, x2, depth);
            for (int i = 0; i< searchAreaCells.length; i++) {
                int cellId = searchAreaCells[i];
                int[] vids = this.hullRaster.getCell(cellId);
                if (null == vids) continue;

                for (int j = 0; j < vids.length; j++) {
                    int vid = vids[j];
                    if (this.vidSet.get(vid)) continue;
                    double y3 = this.lats[vid], x3 = this.lons[vid];
                    // Vektoren normalisieren
                    double ay = y2 - y1, ax = x2 - x1; // a = p2-p1
                    double by = y3 - y1, bx = x3 - x1; // b = p3-p1

                    double ccw = ax * by - ay * bx; // ccw - Abbiegerichtung
                    if (ccw <= 0) continue; // Wichtig! Nur Linksknicks erlauben

                    // Lotfusspunkt berechnen
                    double aa = ax * ax + ay * ay; // Skalarprodukt a * a = |a|^2
                    double ab = ax * bx + ay * by; // Skalarprodukt a * b
                    double e = ab / aa; // Stauchungsfaktor des Vektors p1-p2

                    // Innerhalb der Strecke, sonst Gerade.
                    if (e >= 0 && e <= 1) { // NaN-Sicher!
                        // p4 ist der gesuchte Punkt auf der Strecke
                        double y4 = y1 + ay * e, x4 = x1 + ax * e;
                        double dist = hypot(x3 - x4, y3 - y4);
                        if (dist < depth && dist < minDist) {
                            minDist = dist;
                            closestVid = vid;
                        }
                    }
                }
            }
        }
        return closestVid;
    }

    private float calcLength(int v1, int v2) {
        double y1 = this.lats[v1], x1 = this.lons[v1];
        double y2 = this.lats[v2], x2 = this.lons[v2];
        return (float) hypot(x2 - x1, y2 - y1);
    }

    /**
     * Prueft ob eine anstehende Triangulation zu einer
     * Ueberschneidung mit der aktuellen Huelle fuehren wuerde.
     * @param hs {@link HullSegment} Kanditat
     * @param v int Id des dichtesten Vertex am Segment
     * @return true: Ueberschneidung.
     */
    private boolean probeHullIntersection(HullSegment hs, int v) {
        int vs1 = hs.getVertexId(), vt1 = hs.getNext().getVertexId();
        double ys1 = this.lats[vs1], xs1 = this.lons[vs1];
        double yv1 = this.lats[v], xv1 = this.lons[v];
        double yt1 = this.lats[vt1], xt1 = this.lons[vt1];
        for (HullSegment h : hs) { // Ueber Huelle iterieren
            int vs2 = h.getVertexId(), vt2 = h.getNext().getVertexId();
            double ys2 = this.lats[vs2], xs2 = this.lons[vs2];
            double yt2 = this.lats[vt2], xt2 = this.lons[vt2];
            if (intersects(xs1, ys1, xv1, yv1, xs2, ys2, xt2, yt2)) return true;
            if (intersects(xv1, yv1, xt1, yt1, xs2, ys2, xt2, yt2)) return true;
        }
        return false;
    }

    private static boolean intersects(double l1x1, double l1y1, double l1x2, double l1y2,
            double l2x1, double l2y1, double l2x2, double l2y2) {
        double denom = ((l2y2 - l2y1) * (l1x2 - l1x1)) - ((l2x2 - l2x1) * (l1y2 - l1y1));
        if (denom == 0.0f) return false; // Lines are parallel
        double ua = (((l2x2 - l2x1) * (l1y1 - l2y1)) - ((l2y2 - l2y1) * (l1x1 - l2x1))) / denom;
        double ub = (((l1x2 - l1x1) * (l1y1 - l2y1)) - ((l1y2 - l1y1) * (l1x1 - l2x1))) / denom;

        // Das Original: ((ua >= 0.0d) && (ua <= 1.0d) && (ub >= 0.0d) && (ub <= 1.0d));
        // liefert leider auch true fuer gemeinsamen Start- Zielpunkt, daher:
        return ((ua > 0.0d) && (ua < 1.0d) && (ub > 0.0d) && (ub < 1.0d));
    }

}
