/*
 * $Id: ConvexHull.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.hull;

import static de.cm.osm2po.misc.GeoUtils.ccw;
import static de.cm.osm2po.misc.Utils.DF;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.routing.Graph;

/**
 * Bildet die konvexe Huelle um eine Menge von Punkten.
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 */
public class ConvexHull {

    private Log log;
    private float[] lats;
    private float[] lons;

    /**
     * CTOR.
     * @param graph {@link Graph}
     * @param log {@link Log} NULLABLE
     */
    public ConvexHull(Graph graph, Log log) {
        this.lats = graph.getLats();
        this.lons = graph.getLons();
        this.log = null == log ? Log.empty() : log;
    }

	/**
	 * Bildet die Huelle.
	 * @param vids int[] VertexIds, implizit Indizes auf KoordinatenArray.
	 * @return Huellpunkte (VertexIds) gegen den Uhrzeigersinn.
	 */
	public int[] build(int[] vids) {
		return buildConvexHull(vids);
	}

    /**
     * Berechnet die konvexe Huelle (Graham Scan).<br>
     * @param vids int[] VertexIds, implizit Indizes auf KoordinatenArray.
     * @return int[] Array von Index-Zeiger (VertexIds) auf Koordinaten-Array.
     */
    private int[] buildConvexHull(int[] vids) {
        int[] hull = Arrays.copyOf(vids, vids.length);
        int n = hull.length;
        log.note("Building convex hull of " + DF(n) + " vertices");

        // Finde zuerst die Koordinate, die am weitesten unten oder, wenn
        // gleich, am weitesten RECHTS! liegt und kopiere sie an die erste Position.
        // Urfassung war LINKS. Rechts beseitigt jedoch das Kolliniaritaetsproblem.
        log.progress("Looking for pivot point");
        int x = 0;
        for (int i = 1; i < n; i++) if (this.lats[hull[i]] < this.lats[hull[x]] ||
            (this.lats[hull[i]] == this.lats[hull[x]] &&
                    this.lons[hull[i]] > this.lons[hull[x]])) x = i;
        int min = hull[x]; hull[x] = hull[0]; hull[0] = min;
        float minLat = this.lats[min], minLon = this.lons[min];

        // Berechne die Winkel der Punkte [1-n] zu Punkt [0]
        // im oberen Halbkreis, wobei ganz rechts 0 und ganz links PI ist.
        log.progress("Calculating thetas");
        double[] thetas = new double[n]; // Winkel nur einmal berechnen!
        for (int i = 1; i < n; i++) {
            double lat = this.lats[hull[i]];
            double lon = this.lons[hull[i]];
            thetas[i] = StrictMath.atan2(lat - minLat, lon - minLon);
        }

        // Sortiere nach Winkel zu Punkt[0]
        log.progress("Sorting thetas");
        quicksortHull(1, n - 1, hull, thetas);

        // Fahre die Punkte im oberen Halbkreis gegen den Uhrzeigersinn ab
        // und schaue, dass nur Linksknicks folgen. Folgt ein Rechtsknick,
        // so ist die vorige Koordinate nicht Teil der Huelle.
        log.progress("Applying Graham Scan");
        int m = 1;
        for (int i = 2; i < n; i++) {
            for(;;) {
                int h1 = hull[m-1], h2 = hull[m], h3 = hull[i];
                double y1 = this.lats[h1], x1 = this.lons[h1];
                double y2 = this.lats[h2], x2 = this.lons[h2];
                double y3 = this.lats[h3], x3 = this.lons[h3];

                if (ccw(y1, x1, y2, x2, y3, x3) <= 0) { // Rechtsknick oder Geradeaus.
                    if (m == 1) {
                        int h = hull[m]; hull[m] = hull[i]; hull[i] = h; // Swap
                        if (++i >= n) break;
                    } else m--;
                }
                else break;
            }
            m++;
            int h = hull[m]; hull[m] = hull[i]; hull[i] = h; // Swap
        }

        return Arrays.copyOf(hull, m + 1);
    }

    /**
     * Sortiert nach Winkel und Entfernung zum Punkt[0]
     * und tauscht dabei gleichzeitig die Index-Zeiger des Hull-Arrays.
     * @param low Index links.
     * @param high Index rechts.
     * @param hull int[] synchron zu sortierendes Index-Zeiger-Array.
     * @param thetas double[] zu sortierendes Array von Theta-Winkeln.
     */
    private void quicksortHull(int low, int high,
            int[] hull, double[] thetas) {
        int i = low, j = high;
        double pt = thetas[low + (high-low)/2];
        while (i <= j) {
            while (thetas[i] < pt) i++;
            while (thetas[j] > pt) j--;
            if (i <= j) {
                double t = thetas[i]; thetas[i] = thetas[j]; thetas[j] = t;
                int h = hull[i]; hull[i] = hull[j]; hull[j] = h;
                i++; j--;
            }
        }
        if (low < j) quicksortHull(low, j, hull, thetas);
        if (i < high) quicksortHull(i, high, hull, thetas);
    }
}
