/*
 * $Id: HullRaster.java 4528 2017-10-01 18:47:29Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.hull;

import static de.cm.osm2po.misc.Utils.roundE5;
import static java.lang.Math.abs;
import static java.lang.Math.signum;
import static java.util.Arrays.copyOf;
import static java.util.Arrays.copyOfRange;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.routing.Graph;

/**
 * Erzeugt ein Raster von Vertices fuer die Huellbildung.
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 */
public class HullRaster {
    
    private Log log;
    private int nY, nX;
    private float[] lats;
    private float[] lons;

    private int maxVertexId;
    private double minLat, minLon, maxLat, maxLon; // Bounding
    private double cellSize, cellSizeHalf;
    
    private int[][][] raster;
    
    public HullRaster(Graph graph, Log log) {
        this.lats = graph.getLats();
        this.lons = graph.getLons();
        this.log = null == log ? Log.empty() : log;
    }
    
    public int getMaxVertexId() {return this.maxVertexId;}
    
    /**
     * Liefert eine Zelle, d.h. eine Menge von VertexIds,
     * die in ihr liegen fuer eine Koordinate.
     * @param cellId int Id der Zelle
     * @return int[] Array von VertexIds in der Zelle.
     */
    public int[] getCell(int cellId) {
        int y = cellId / this.nX;
        int x = cellId % this.nX;
        if (y >= this.nY || x >= this.nX) return null;
        int[] cell = this.raster[y][x];
        return cell;
    }
    
    /**
     * Liefert den Ankerpunkt der Zelle.
     * @param cellId int Id der Zelle.
     * @return {@link LatLon} Koordinate des Ankerpunkts.
     */
    public LatLon getAnchor(int cellId) {
        int y = cellId / this.nX;
        int x = cellId % this.nX;
        double lat = y * this.cellSize + this.minLat;
        double lon = x * this.cellSize + this.minLon;
        return new LatLon(lat, lon);
    }
    
    public LatLons toRect(int cellId) {
        LatLon a = this.getAnchor(cellId);
        LatLon b = new LatLon(
                a.getLat() + this.cellSize, a.getLon() + this.cellSize);
        return new LatLons()
            .append(a.getLat(), a.getLon())
            .append(a.getLat(), b.getLon())
            .append(b.getLat(), b.getLon())
            .append(b.getLat(), a.getLon())
            .append(a.getLat(), a.getLon());
    }
    
    public LatLons toPoints(int cellId) {
        LatLons points = new LatLons();
        int[] ids = this.getCell(cellId);
        if (ids != null) for (int i = 0; i < ids.length; i++) {
            points.append(this.lats[ids[i]], this.lons[ids[i]]);
        }
        return points;
    }
    
    /**
     * Liefert die Id der Zelle aufgrund der uebergebenen Koordinaten.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @return int Id der Zelle (implizit x-y-Koordinate des Rasters) oder -1
     */
    public int cellIdOf(double lat, double lon) {
        int y = (int) ((lat - this.minLat) / this.cellSize);
        int x = (int) ((lon - this.minLon) / this.cellSize);
        if (y >= this.nY || x >= this.nX) return -1;
        int id = y * this.nX + x;
        return id;
    }
    
    public int[] getCells(int sourceId, int targetId, float depth) {
        return this.rasterLine(sourceId, targetId, depth);
    }

    public int[] getCells(LatLon c1, LatLon c2, float depth) {
        return this.rasterLine(c1.getLat(), c1.getLon(), c2.getLat(), c2.getLon(), depth);
    }
    
    public int[] rasterLine(int sourceId, int targetId, float depth) {
        return this.rasterLine(this.lats[sourceId], this.lons[sourceId],
                this.lats[targetId], this.lons[targetId], depth);
    }
    
    public int[] rasterLine(double y1, double x1, double y2, double x2, float depth) {
        int[] cellIds = new int[0x400];
        int nCellIds = 0;
        
        double wh = this.cellSize; // Breite=Hoehe als double
        
        // Vektor normalisiert und Betraege bestimmen
        double vx = x2 - x1, dxAbs = abs(vx); 
        double vy = y2 - y1, dyAbs = abs(vy);
        
        // Vektor-Betrag (Laenge) in Zelle einpassen
        int n = 0;
        if (dxAbs > dyAbs) {
            n = (int)(dxAbs / wh) + 2;
            vy *= wh / dxAbs;
            vx = wh * signum(vx);
        } else {
            n = (int)(dyAbs / wh) + 2;
            vx *= wh / dyAbs;
            vy = wh * signum(vy);
        }
        
        // TODO Diskrete Formel verwenden
        // derzeit Fortpflanzung von numerischem Fehler durch
        // Addition kleiner Teilstuecke. 
        int d = (int) (depth / this.cellSize + 2); 
        for (int j = -1; j < d;  j++) {
            // Startpunkt orthogonal verschieben (Parallel-Vector)
            double x = x1 + j * -vy;
            double y = y1 + j * vx;
            for (int i = 0; i < n; i++) {
                // log.debug("x= " + x + "  y= " + y);
                int cellId = this.cellIdOf(y, x);
                if (cellId >= 0) {
                    if (cellIds.length == nCellIds) cellIds = copyOf(cellIds, nCellIds * 2);
                    cellIds[nCellIds++] = cellId;
                }
                x += vx; y += vy;
            }
        }

        return copyOf(cellIds, nCellIds);
    }
    
    /**
     * Erzeugt das Raster fuer eine Menge von VertexIds.
     * @param vids int[] Array von VertexIds, implizit Zeiger auf Koordinaten.
     * @param gridSize Groesse einer Raster-Zelle x und y.
     */
    public void build(int[] vids, float gridSize) {
        this.cellSize = roundE5(gridSize);
        this.cellSizeHalf = this.cellSize / 2;
        
        log.progress("Calculating BoundingBox");
        // Achtung: Float.MIN_VALUE ist 0 !!!
        this.minLat = 90f; this.minLon = 180f;
        this.maxLat = -90f; this.maxLon = -180f; 
        for (int i = 0; i < vids.length; i++) {
            int vid = vids[i];
            if (vid > this.maxVertexId) this.maxVertexId = vid;
            double lat = roundE5(this.lats[vid]);
            double lon = roundE5(this.lons[vid]);
            if (lat < this.minLat) this.minLat = lat;
            if (lon < this.minLon) this.minLon = lon;
            if (lat > this.maxLat) this.maxLat = lat;
            if (lon > this.maxLon) this.maxLon = lon;
        }
        
        // Raster halb nach unten links verschieben und oben rechts erweitern
        this.minLat -= this.cellSizeHalf;
        this.minLon -= this.cellSizeHalf;
        this.maxLat += this.cellSizeHalf;
        this.maxLon += this.cellSizeHalf;
        
        log.progress("Creating raster with cells of size " + cellSize);
        double h = this.maxLat - this.minLat; // Hoehe
        double w = this.maxLon - this.minLon; // Breite
        this.nY = (int) (h / this.cellSize + 1);
        this.nX = (int) (w / this.cellSize + 1);
        this.raster = new int[this.nY][][];
        for (int y = 0; y < this.nY; y++) this.raster[y] =  new int[this.nX][];
        
        log.progress("Rastering vertices");
        for (int i = 0; i < vids.length; i++) {
            int vid = vids[i];
            double lat = this.lats[vid] - this.minLat;
            double lon = this.lons[vid] - this.minLon;
            int y = (int) (lat / this.cellSize);
            int x = (int) (lon / this.cellSize);
            int[] cell = this.raster[y][x];
            if (null == cell) {
                cell = new int[8];
                this.raster[y][x] = cell;
            }
            cell[0]++; // Tricky: Counter an Index 0
            if (cell.length == cell[0]) {
                cell = copyOf(cell, cell.length * 2);
                this.raster[y][x] = cell;
            }
            cell[cell[0]] = vid;
        }
        
        log.progress("Compressing to standard arrays");
        for (int y = 0; y < this.nY; y++) {
            for (int x = 0; x < this.nX; x++) {
                int[] cell = this.raster[y][x];
                if (cell != null) {
                    cell = copyOfRange(cell, 1, cell[0] +1);
                    this.raster[y][x] = cell;
                }
            }
        }
        
	}
	
}
