/*
 * $Id: HullSegment.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.hull;

import java.util.Iterator;

import de.cm.osm2po.model.LatLons;

/**
 * Kapselt ein Segment einer Huelle.<br>
 * Das Segment ist gleichzeitig ein Knoten mit Vorgaenger und
 * Nachfolger, so dass beim Anfuegen neuer Segmente automatisch
 * ein Ring entsteht.
 * <br><br><b>Wichtig:</b><br>
 * Das interne {@link Comparable} sortiert nach highest value first! 
 * 
 * @author (c) 2015 - Carsten Moeller, info@osm2po.de
 */
public class HullSegment implements Iterable<HullSegment>, Comparable<HullSegment> {
    private HullSegment prev;
    private HullSegment next;
    private int vertexId;
    private float priority;
    
    private HullSegment() {}

    /**
     * Konstruktor, der das erste Segment einer Huelle erzeugt.
     * @param vertexId int Id des ersten Punktes (SourceId).
     */
    public HullSegment(int vertexId) {
        this.vertexId = vertexId;
        this.next = this; // Ring-Anordnung
        this.prev = this;
    }

    public HullSegment getPrev() {return this.prev;}
    public HullSegment getNext() {return this.next;}
    public int getVertexId() {return this.vertexId;}
    public float getPriority() {return this.priority;}
    
    public void setPriority(float priority) {this.priority = priority;}

    /**
     * Kettet ein neues Segment hinter dieses und erweitert so den Ring.
     * @param vertexId int Id des ersten Punktes.
     * @return neu angefuegtes, new {@link HullSegment}
     */
    public HullSegment insertNext(int vertexId) {
        HullSegment that = new HullSegment();
        that.vertexId = vertexId;
        that.next = this.next;
        that.prev = this;
        this.next = that;
        if (this.prev == this) { // bei 2 Elementen
            this.prev = that; 
        }
        return that;
    }
    
    public LatLons toLine(float[] lats, float[] lons) {
        LatLons geometry = new LatLons();
        geometry.append(lats[this.vertexId], lons[this.vertexId]);
        geometry.append(lats[this.next.vertexId], lons[this.next.vertexId]);
        return geometry;
    }
    
    public String toRingString() {
        String s = null;
        HullSegment hs = this;
        do {
            s = null == s ? "" : s + "\n";
            s += hs;
            hs = hs.next;
        } while (hs != this);
        return s;
    }
    
    @Override
    public String toString() {
        return "#" + this.vertexId +" $" + this.priority;
    }

    @Override
    public Iterator<HullSegment> iterator() {
        return new Iterator<HullSegment>() {
            private boolean hasNext = true;
            HullSegment hs = HullSegment.this;
            @Override public boolean hasNext() {return this.hasNext;}
            @Override public HullSegment next() {
                HullSegment hsRet = this.hs;
                this.hs = this.hs.next;
                this.hasNext = this.hs != HullSegment.this;
                return hsRet;    
            }
            @Override
            public void remove() {throw new UnsupportedOperationException();}
        };
    }

    @Override
    public int compareTo(HullSegment that) {
        if (this.priority < that.priority) return 1;
        if (this.priority > that.priority) return -1;
        return 0;
    }

}
