/*
 * $Id: Log.java 5763 2023-03-04 14:55:25Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;
import static de.cm.osm2po.logging.LogConsoleWriter.ERR;
import static de.cm.osm2po.logging.LogConsoleWriter.OUT;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * Ein kleiner Logger. Simpel aber gut.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public final class Log implements Iterable<LogWriter> {

    /** Undefiniert, intern verwendet. */
    public final static int LEVEL_NULL = 0;
    /** Ausgabe von Daten und Dumps. */
    public final static int LEVEL_DATA = 1;
    /** Korrigiete oder gerettete Daten. */
    public final static int LEVEL_BUGFIX = 2;
    /** Detaillierte Entwicklungs-Infos. */
    public final static int LEVEL_DEBUG = 3;
    /** Verarbeitungsfortschritt anzeigen. */
    public final static int LEVEL_PROGRESS = 4;
    /** Info Zusatz z.B. fuer Ergebnisanzeigen. */
    public final static int LEVEL_NOTE = 5;
    /** Warn Info unkritischer Natur. */
    public final static int LEVEL_WARN = 6;
    /** Wichtige Programminformation. */
    public final static int LEVEL_INFO = 7;
    /** Kritische Programminfo, ohne Crash. */
    public final static int LEVEL_ERROR = 8;
    /** Sehr wichtige Info, z.B. fuer externe Systeme oder Mail-Trigger */
    public final static int LEVEL_MESSAGE = 9;
    /** Alarm Info, z.B. fuer externe Systeme oder Mail-Trigger */
    public final static int LEVEL_ALERT = 10;
    /** Sehr kritische Programminfo, wahrscheinlich mit Crash. */
    public final static int LEVEL_FATAL = 11;

    public final static String[] LEVEL_PREFIX_ARR = {
        "      ",
        "DATA  ",
        "FIX   ",
        "DEBUG ",
        "...   ",
        "NOTE  ",
        "WARN  ",
        "INFO  ",
        "ERROR ",
        "INFO  ",
        "ALERT ",
        "FATAL ",
    };

    public static int translateLogLevel(String strLogLevel) {
        if (strLogLevel != null) {
            strLogLevel = strLogLevel.trim().toLowerCase();
            if ("data".equals(strLogLevel))
                return LEVEL_DATA;
            else if ("bugfix".equals(strLogLevel))
                return Log.LEVEL_BUGFIX;
            else if ("debug".equals(strLogLevel))
                return LEVEL_DEBUG;
            else if ("progress".equals(strLogLevel))
                return Log.LEVEL_PROGRESS;
            else if ("note".equals(strLogLevel))
                return LEVEL_NOTE;
            else if ("warn".equals(strLogLevel))
                return LEVEL_WARN;
            else if ("info".equals(strLogLevel))
                return LEVEL_INFO;
            else if ("error".equals(strLogLevel))
                return LEVEL_ERROR;
            else if ("message".equals(strLogLevel))
                return LEVEL_MESSAGE;
            else if ("alert".equals(strLogLevel))
                return LEVEL_ALERT;
            else if ("fatal".equals(strLogLevel))
                return LEVEL_FATAL; // muss letztes Element sein
        }
        return -1;
    }

    private int logLevel;
    private boolean logThread;
    private volatile boolean stashMode;

    private List<LogWriter> logWriterList = new ArrayList<LogWriter>();
    private List<String> stashList = new ArrayList<String>();

    private long[] logCounters = new long[LEVEL_FATAL +1];

    /** @return long Array of number of calls per level */
    synchronized public final long[] getCounters() {
        return Arrays.copyOf(this.logCounters, this.logCounters.length);
    }

    /** resets the counter of log calls */
    synchronized public final void resetCounters() {
        Arrays.fill(this.logCounters, 0L);
    }

    /** @return Log ohne {@link LogWriter}. */
    public static Log empty() {
        return new Log();
    }

    /** @return Log mit stdout-Writer Level Warn. */
    public static Log stdout() {
        return stdout(LEVEL_WARN);
    }

    /** @return Log mit stderr-Writer Level Error  */
    public static Log stderr() {
        return stderr(LEVEL_ERROR);
    }

    public static Log stdout(int LEVEL) {
        return new Log(LEVEL).addLogWriter(
                new LogConsoleWriter().setPrintTo(OUT).setLogLevel(LEVEL));
    }

    public static Log stderr(int LEVEL) {
        return new Log(LEVEL).addLogWriter(
                new LogConsoleWriter().setPrintTo(ERR).setLogLevel(LEVEL));
    }

    public Log() {
        this.logLevel = LEVEL_INFO;
    }

    public Log(int logLevel) {
        this.logLevel = logLevel;
    }

    public final Log setLogLevel(int logLevel) {
    	this.logLevel = logLevel;
    	return this;
    }

    public final int getLogLevel() {
    	return this.logLevel;
    }

    public final Log setLogThread(boolean logThread) {
        this.logThread = logThread;
        return this;
    }

    public final boolean isLogThread() {
        return this.logThread;
    }

    /**
     * Fuegt einen neuen Logger hinzu.
     * <br><b>Wichtig:</b><br>
     * Steht dieser initial auf {@link Log#LEVEL_NULL}, so
     * erbt er das Level seines Parents.
     * @param logWriter {@link LogWriter}
     * @return this
     */
    public final Log addLogWriter(LogWriter logWriter) {
        this.logWriterList.add(logWriter);
        return this;
    }

    /**
     * Schliesst alle {@link LogWriter}, entfernt sie aus der
     * internen Liste und reinitialisiert intern zu einer leeren Liste.
     * @return {@link List} der entfernten {@link LogWriter}
     */
    public final List<LogWriter> closeAllWriters() {
    	if (this.logWriterList != null)
    	    for (LogWriter logWriter : this.logWriterList) {
    	        try {logWriter.close();} catch (Throwable ignore) {};
    	    }
    	List<LogWriter> wl = this.logWriterList;
    	this.logWriterList = new ArrayList<LogWriter>();
    	return wl;
    }

    @Override
    public final Iterator<LogWriter> iterator() {
        return this.logWriterList.iterator();
    }

    public final void data(String msg) {this.log(msg, LEVEL_DATA, null);}
    public final void bugfix(String msg) {this.log(msg, LEVEL_BUGFIX, null);}
    public final void debug(String msg) {this.log(msg, LEVEL_DEBUG, null);}
    public final void note(String msg) {this.log(msg, LEVEL_NOTE, null);}
    public final void warn(String msg) {this.log(msg, LEVEL_WARN, null);}
    public final void info(String msg) {this.log(msg, LEVEL_INFO, null);}
    public final void error(String msg) {this.log(msg, LEVEL_ERROR, null);}
    public final void fatal(String msg) {this.log(msg, LEVEL_FATAL, null);}
    public final void message(String msg) {this.log(msg, LEVEL_MESSAGE, null);}
    public final void alert(String msg) {this.log(msg, LEVEL_ALERT, null);}
    public final void progress(String msg) {this.log(msg, LEVEL_PROGRESS, null);}

    public final void error(Throwable t) {this.error(translateError(t, 10));}
    public final void fatal(Throwable t) {this.fatal(translateError(t, 10));}

    public final boolean isEnabled(int logLevel) {
        return logLevel >= this.logLevel;
    }

    public final synchronized Log stash(int level, final String msg) {
        this.stashList.add(level + "|" + msg);
        return this;
    }

    public final synchronized Log stashMode(boolean on) {
        this.stashMode = on;
        return this;
    }

    public final synchronized Log stashFlush() {
        boolean currentStashMode = this.stashMode;
        this.stashMode = false;
        for (String msg : this.stashList) {
            int idx = msg.indexOf('|');
            String level = msg.substring(0, idx);
            msg = msg.substring(idx +1);
            this.log(msg, Integer.parseInt(level), null);
        }
        this.stashList.clear();
        this.stashMode = currentStashMode;
        return this;
    }

    public final synchronized void log(String message,
            int logLevel, LogWriter sender) {

        if (null == this.logWriterList) return;

        if (logLevel < 0) logLevel = 0;
        if (logLevel >= LEVEL_PREFIX_ARR.length)
            logLevel = LEVEL_PREFIX_ARR.length -1;

        if (this.stashMode) { // Wichtig: Immer ohne LevelFilter
            this.stash(logLevel, message); // Merken fuer verzoegerte Ausgabe

        } else {
            this.logCounters[logLevel]++;
            if (logLevel >= this.logLevel) {
                if (this.logThread) {
                    Thread ct = Thread.currentThread();
                    if (LEVEL_PROGRESS == logLevel) {
                        message = "T" + ct.getId() + " " + message;
                    } else {
                        message = "T" + ct.getId() + "-"
                            + ct.getName().toUpperCase() + " " + message;
                    }
                }
            }

            List<LogWriter> brokenWriterList = null;
            for (LogWriter lw : this.logWriterList) {
                try {
                    if (lw != sender) lw.log(message, logLevel);

                } catch (Throwable t) {
                    if (null == brokenWriterList) brokenWriterList = new ArrayList<LogWriter>();
                    try {lw.close();} catch (Throwable ignore) {}

                    this.stash(LEVEL_FATAL, translateError(t, 10));
                    this.stash(LEVEL_WARN, lw.getClass().getSimpleName() + " closed and removed");
                    brokenWriterList.add(lw);
                }
            }
            if (brokenWriterList != null) {
                this.logWriterList.removeAll(brokenWriterList);
            }
        }
    }

    public static String translateError(Throwable t, int stackTraceLen) {
        String msg = null;
        String clazz = null;
        String errString = "Unknown Error";

        if (t instanceof InvocationTargetException) { // reflected error
            t = ((InvocationTargetException) t).getTargetException();
        }

        if (t != null) {

            if (t.getClass() != null) {
                clazz = t.getClass().getName();
            }
            msg = t.getMessage();
            if (null == msg && t.getCause() != null) {
                msg = t.getCause().getMessage();
            }

            errString = "Exception at " + new Date().toString()
                + "\nMessage: " + (null == msg ? "Unknown" : msg)
                + "\nClass: " + (null == clazz ? "Unknown" : clazz)
                + "\nStacktrace:";

            if (null == t.getStackTrace()) {
                errString += "null";
            } else {
                int lines = t.getStackTrace().length;
                if (0 == lines) {
                    errString += " is empty!";
                } else {
                    int more = lines;
                    for (int i = 0; i < stackTraceLen && i < lines; i++, more--) {
                        errString += "\n" + t.getStackTrace()[i].toString();
                    }
                    if (more > 0) errString += "\n" + more + " more...";
                }
            }
        }

        return errString;
    }

    public final void logWarnCounters() {
        this.logWarnCounters(this.logCounters);
    }

    public final void logWarnCounters(long[] counters) {
        String lwc = null;;
        for (int level : new int[]{LEVEL_WARN, LEVEL_ERROR, LEVEL_ALERT, LEVEL_FATAL}) {
            if (counters[level] > 0) {
                if (null == lwc) lwc = "Log statistics:";
                lwc += ' ' + Log.LEVEL_PREFIX_ARR[level].trim() + "s:" + counters[level];
            }
        }
        if (lwc != null) {
            // Hier nur Warn senden, da LogMailer und Co auf errors reagieren
            this.log(lwc, LEVEL_WARN, null);
            this.logCounters[LEVEL_WARN]--; // sich selbst nicht mitzaehlen
        }
    }

    /**
     * Erzeugt ein Log, das sich nicht selbst aufruft.<br>
     * Sinnvoll, wenn ein Logger, z.B. ein {@link LogMailWriter}
     * Fehler-Logs generiert, die allerdings nicht in den eigenen
     * Log-Kanal geschickt werden sollen.
     * @param sender {@link LogWriter} Verursacher
     * @return neues {@link Log}
     */
    public final Log createNonSelfCallingWriterLog(final LogWriter sender) {
        return new Log(this.logLevel).addLogWriter(new LogWriterAdapter() {
            @Override public void log(String message, int logLevel) {
                Log.this.log(message, logLevel, sender);
            }
        });
    }

}
