/*
 * $Id: LogSlackWriter.java 5857 2024-02-24 11:58:07Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.logging;

import static de.cm.osm2po.misc.Utils.TS;
import static de.cm.osm2po.service.ServiceHandler.MIME_JSON;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.HttpsURLConnection;

import de.cm.osm2po.Config;
import de.cm.osm2po.reflection.JsonBuilder;

/**
 * Ein einfacher Logger, der aufgelaufene Meldungen an einen
 * <a href="http://slack.com">Slack</a>-Channel sendet.
 * Ausgeloest wird der Versandt entweder durch die Ueberschreitung einer
 * maximalen Anzahl Zeichen oder nach Erhalt einer {@link Log#LEVEL_MESSAGE}
 * Meldung oder hoeher oder aufgrund des Zeit-Intervall-Triggers.
 * <br>Das eigentliche Default-LogLevel ist {@link Log#LEVEL_WARN}.
 * <br>Config-Beispiel:<br>
 * <pre>
 * log.3.class=de.cm.osm2po.logging.LogSlackWriter
 * log.3.level=warn (default)
 * log.3.channel=/T9XXXXXXX/B9XXXXXXX/111YYYYYYYYYYYYYYYYYYabc
 * log.3.interval=60 (default, AutoSendInterval in Minuten)
 * log.3.maxchars=1024 (default Anzahl max Zeichen in Kilo)
 * </pre>
 * <b>Wichtig:</b><br>
 * Dem Channel wird die URL <code>https://hooks.slack.com/services</code>
 * vorangestellt.
 * <br><br><b>Empfehlung:</b><br>
 * Optimalerweise solllte dieser Logger an Index 0, also noch vor einem
 * etwaigen {@link LogConsoleWriter} oder {@link LogFileWriter}
 * konfiguriert werden, damit bei {@link #close()} die Aufraeum-Sequenz
 * zumindest noch in den oben genannten Logs erscheint.
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 * @see LogMailWriter
 */
public class LogSlackWriter implements LogWriter {
    private volatile boolean closed;

    private Log log;
    private int logLevel;
    private StringBuilder logHistory;
    private int maxHistoryLength;
    private URL channelUrl;
    private ExecutorService senderThreads;
    private Thread intervalSenderThread;
    private long interval;

    @Override
    public LogWriter setLogLevel(int logLevel) {
        this.logLevel = logLevel;
        return this;
    }

    @Override
    public int getLogLevel() {
        return this.logLevel;
    }

    @Override
    public LogWriter open(Config config, int index) {
        this.log = config.getLog().createNonSelfCallingWriterLog(this);

        final String prefix = "log." + index + ".";

        String channel = config.getProperty(prefix + "channel");
        if (null == channel) throw new IllegalArgumentException(
                prefix + "channel not set");
        if (!channel.startsWith("/")) channel = "/" + channel;
        try {
            this.channelUrl = new URI(
                    "https://hooks.slack.com/services" + channel).toURL();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        this.closed = false;

        this.setLogLevel(Log.translateLogLevel(
                config.getProperty(prefix + "level", "warn")));

        this.logHistory = new StringBuilder();
        this.maxHistoryLength = 1024 * Integer.parseInt(
                config.getProperty(prefix + "maxchars", "1024"));
        this.interval = 60000 * Integer.parseInt(
                config.getProperty(prefix + "interval", "60"));

        this.senderThreads = Executors.newFixedThreadPool(5);

        this.intervalSenderThread = new Thread(new IntervalSender());
//        this.intervalSenderThread.setName(
//                LogSlackWriter.class.getSimpleName() + "IntervalThread");
        this.intervalSenderThread.start();

        return this;
    }

    @Override
    public void log(String msg, int logLevel) {
        if (this.closed) return;
        if (logLevel < this.logLevel) return;
        if (logLevel == Log.LEVEL_PROGRESS) return;

        synchronized(this) {
            String levelStr = Log.LEVEL_PREFIX_ARR[logLevel];
            this.logHistory.append(levelStr + msg).append('\n');
        }

        if (logLevel >= Log.LEVEL_MESSAGE
                || this.logHistory.length() >= this.maxHistoryLength)
            this.sendHistory();
    }

    private synchronized void sendHistory() {
        if (this.logHistory.length() > 0) {
            String message = this.logHistory.toString(); // tatsaechlich eine Kopie
            this.logHistory.setLength(0);
            this.senderThreads.execute(new Sender(
                    this.log, this.channelUrl, message));
        }
    }

    @Override
    public void close() {
        if (this.closed) return;
        try {

            this.log.info("SLACK starting close sequence");

            this.closed = true;
            this.intervalSenderThread.interrupt();
            this.sendHistory();
            this.senderThreads.shutdown();

            if (this.senderThreads.awaitTermination(5, TimeUnit.SECONDS)) {
                this.log.info("SLACK closed");
            } else {
                this.log.warn("SLACK closed but not all messages sent");
            }

        } catch (Throwable t) {
            this.log.error(t);
        }
    }

    private class IntervalSender implements Runnable {
        @Override
        public void run() {
            final LogSlackWriter lm = LogSlackWriter.this;
            try {
                while (!lm.closed) {
                    Thread.sleep(lm.interval);
                    if (lm.closed) return; // Doppelt haelt besser

                    if (lm.logHistory.length() > 0) {
                        lm.log.info("SLACK " + lm.getClass().getSimpleName()
                                + "-Interval reached. Flushing - " + TS());
                        lm.sendHistory();
                    } else {
                        lm.log.info("SLACK LogHistory is empty. Nothing sent - " + TS());
                    }
                }

            } catch (InterruptedException ie) {
                lm.log.info("SLACK " + lm.getClass().getSimpleName()
                        + this.getClass().getSimpleName() + " interrupted");
            } catch (Throwable t) {
                lm.log.error(t);
            }
        }
    }

    private static class Sender implements Runnable {
        private final Log log;
        private final String message;
        private URL channel;

        public Sender(Log log, URL channel, String message) {
            this.log = log;
            this.message = message;
            this.channel = channel;
        }

        @Override
        public void run() {
            try {
                this.log.progress("SLACK Log to " + this.channel);

                String json = JsonBuilder.build(this.message, "text");
                byte[] payload = json.getBytes();

                HttpsURLConnection conn = (HttpsURLConnection) this.channel.openConnection();
                conn.setRequestMethod("POST");
                conn.setRequestProperty("Content-Type", MIME_JSON);
                conn.setRequestProperty("Content-Length", "\"" + payload.length + "\"");
                conn.setDoOutput(true);
                OutputStream os = conn.getOutputStream();
                os.write(payload);
                os.close();
                InputStream is = conn.getInputStream(); // actually triggers the request
                is.close();
                // int responseCode = conn.getResponseCode();

            } catch (Exception e) {
                this.log.error(e);
            }
        }

    }

}
