/*
 * $Id: Utils.java 3272 2014-09-29 17:09:42Z cmindividual $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static java.lang.Byte.parseByte;
import static java.lang.Integer.parseInt;
import static java.lang.Long.parseLong;
import static java.lang.System.arraycopy;
import static java.util.Arrays.copyOf;
import static java.util.Arrays.sort;

import java.util.Arrays;
import java.util.List;

public class ArrayUtils {
    private ArrayUtils() {/*static*/}

    /**
     * Tauscht zwei Positionen im Array.
     * @param array int[]
     * @param x int index x
     * @param y int index y
     */
    public static void intArraySwap(int[] array, int x, int y) {
        int z = array[x]; array[x] = array[y]; array[y] = z;
    }

    /**
     * Tauscht die partielle Reihenfolge im Array zwischen
     * zwei Positionen (inklusive), auch Flip genannt.<br>
     * Reverse([0,1,2,3,4,5], 2, 4) =&gt; [0,1,4,3,2,5]
     * @param array int[]
     * @param x int index x
     * @param y int index y
     */
    public static void intArrayReverse(int[] array, int x, int y) {
        for (int i = x, j = y; i < j; i++, j--) {
            int z = array[i]; array[i] = array[j]; array[j] = z;
        }
    }

    /**
     * Verschiebt im Array N Elemente von links nach rechts,
     * aber auch von rechts nach links. Die vordersten bzw
     * hintersten Elemente landen danach je nach Richtung wieder
     * hinten bzw vorne.
     * @param arr int[] Das Array
     * @param x int Start-Index (inklusive)
     * @param y int End-Index (inklusive)
     * @param d Verschiebe-Distanz: 1=1x nach rechts, aber auch -2=2x nach links.
     * <br><br><b>Wichtig:</b><br>
     * Die Grenzen des Arrays werden hier nicht plausibilisiert.
     */
    public static void intArrayRotate(int[] arr, int x, int y, int d) {
        while (d != 0) {
            if (d > 0) { // nach rechts, letztes Element danach vorne
                int tmp = arr[y];
                for (int i = y; i > x; i--) arr[i] = arr[i - 1];
                arr[x] = tmp;
                --d;

            } else if (d < 0) { // nach links, erstes Element danach hinten
                int tmp = arr[x];
                for (int i = x; i < y; i++) arr[i] = arr[i + 1];
                arr[y] = tmp;
                ++d;
            }
        }
    }

    /**
     * Verschiebt im Array N Elemente von Index x nach Index y.
     * Ist x kleiner y, so rutscht y eine Position runter.
     * Ist x groesser y, so schiebt sich y eine Position vor.
     * Dies ist kein Ringtausch oder Swap!
     * @param array int[]
     * @param x int Zu verschiebene Position (Index)
     * @param y int Zielposition (Index)
     * @param n int Anzahl der Elemente
     * <br><br><b>Wichtig:</b><br>
     * Die Grenzen des Arrays werden hier nicht plausibilisiert.
     */
    public static void intArrayMove(int[] array, int x, int y, int n) {
        if (x < y) for (int offset = n -1; offset >= 0; offset--) {
            int nSwaps = y - x;
            for (int i = x + offset, j = 0; j < nSwaps; i++, j++)
                intArraySwap(array, i, i +1);
        } else  for (int offset = 0; offset < n; offset++) {
            int nSwaps = x - y;
            for (int i = x + offset, j = 0; j < nSwaps; i--, j++)
                intArraySwap(array, i, i -1);
        }
    }

    /**
     * Tauscht im Array N Elemente zwischen Index x und Index y.
     * @param array int[]
     * @param x int Erster zu tauschender StartIndex
     * @param y int Erster ZielIndex
     * @param n int Anzahl der Elemente
     * @param mirrored boolean true: Die Elemente werden gespiegelt vertauscht
     * <br><br><b>Wichtig:</b><br>
     * Die Grenzen des Arrays werden hier nicht plausibilisiert.
     */
    public static void intArraySwap(int[] array, int x, int y, int n, boolean mirrored) {
        if (mirrored) {
            int z = y + n -1;
            for (int offset = 0; offset < n; offset++) {
                intArraySwap(array, x + offset, z - offset);
            }
        } else {
            for (int offset = 0; offset < n; offset++) {
                intArraySwap(array, x + offset, y + offset);
            }
        }
    }

    /**
     * Rotiert im Array Elemente zwischen zwei Positionen
     * um eine Index-Position rueckwaerts. Das erste
     * Element steht nach der ersten Iteration hinten.
     * @param array int[]
     * @param x int Erster Index
     * @param y int Zweiter Index
     * <br><br><b>Wichtig:</b><br>
     * Die Grenzen des Arrays werden hier nicht plausibilisiert.
     */
    public static void intArrayRotate(int[] array, int x, int y) {
        int v0 = array[x];
        for (int i = x + 1; i <= y; i++) array[i-1] = array[i];
        array[y] = v0;
    }

    /**
     * Kettet beliebig viele int-Arrays hintereinander und liefert
     * diese als ein Array zurueck.
     * @param arrays int[]... beliebig viele Eingabe-Arrays, auch nullable.
     * @return int[] Array. NOT NULL.
     * @see #intArrayListToArray2D(List)
     */
    public static int[] intArrayConcat(int[]... arrays) {
        int n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) n += arrays[i].length;
        }
        int[] arr = new int[n];
        n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) {
                arraycopy(arrays[i], 0, arr, n, arrays[i].length);
                n += arrays[i].length;
            }
        }
        return arr;
    }

    /**
     * Erzeugt aus einer Liste von int-Arrays ein zweidimensionales
     * int-Array.
     * @param list List of int[]
     * @return int[][]
     * @see #intArrayConcat(int[][])
     */
    public static int[][] intArrayListToArray2D(List<int[]> list) {
        if (null == list || list.isEmpty()) return null;
        int[][] arr2D = new int[list.size()][];
        int i = 0;
        for (int[] arr1D : list) arr2D[i++] = arr1D;
        return arr2D;
    }

    /**
     * Eliminiert Duplikate aus einem int-Array und bildet
     * so faktisch ein int-Array-Set.
     * @param array NOT NULL int[] Array.
     * @return new int[] Array ohne Duplikate, sortiert.
     */
    public static int[] intArrayDistinct(int[] array) {
        int length = array.length;
        if (length < 2) return Arrays.copyOf(array, length);
        int[] arr = copyOf(array, length);
        sort(arr);
        int i = 0;
        for (int j = 1; j < length; j++) {
            if (arr[j] == arr[i]) continue;
            arr[++i] = arr[j];
        }
        return copyOf(arr, i + 1);
    }

    /**
     * Liefert ein neues int[]-Array, in dem ein neuer
     * int Wert an Position x hinzugefuegt wird.
     * Dies ist eine sehr teure Operation, da hier
     * das komplette Array kopiert und verschoben wird.
     * @param array int[] aktuelles Array
     * @param val int hinzuzufuegender Wert
     * @param x Einfuege-Index 0 &gt;= x &lt;= array.length
     * @return neues int[]-Array mit neuem Element und length +1.
     */
    public static int[] intArrayInsert(final int[] array, int val, int x) {
        int[] arr = new int[array.length +1];
        arraycopy(array, 0, arr, 0, x);
        arr[x] = val;
        arraycopy(array, x, arr, x + 1, array.length - x);
        return arr;
    }

    /**
     * Sucht sequentiell im Array nach einem Wert und liefert
     * den gefundenen Index.
     * @param array int[] Such-Array.
     * @param val Suchwert
     * @return int Index im Array oder -1;
     */
    public static int intArrayFind(final int[] array, int val) {
        for (int i = 0; i < array.length; i++) {
            if (val == array[i]) return i;
        }
        return -1;
    }

    /**
     * Erstellt aus einem csv aus Zahlen ein int[] Array.
     * @param csv z.B. '1,2,3' NOT NULL
     * @return int[] Array(csv)
     */
    public static int[] csv2IntArray(String csv) {
        String[] strInts = csv.split(",");
        int[] ints = new int[strInts.length];
        for (int i = 0; i < strInts.length; i++) {
            int id = parseInt(strInts[i].trim());
            ints[i] = id;
        }
        return ints;
    }

    /**
     * Erstellt aus einem Array von ints ein CSV.
     * @param arr int[] NULLABLE
     * @return CSV NOT NULL;
     */
    public static String intArray2Csv(int[] arr) {
        String s = "";
        if (arr != null) for (int i = 0; i < arr.length; i++) {
            if (i > 0) s += ','; s += arr[i];
        }
        return s;
    }

    /**
     * Erstellt aus einem Array von longs ein CSV.
     * @param arr long[] NULLABLE
     * @return CSV NOT NULL;
     */
    public static String longArray2Csv(long[] arr) {
        String s = "";
        if (arr != null) for (int i = 0; i < arr.length; i++) {
            if (i > 0) s += ','; s += arr[i];
        }
        return s;
    }

    /**
     * Erstellt aus einem csv aus Zahlen ein long[] Array.
     * @param csv z.B. '1,2,3' NOT NULL
     * @return long[] Array(csv)
     */
    public static long[] csv2LongArray(String csv) {
        String[] strLongs = csv.split(",");
        long[] longs = new long[strLongs.length];
        for (int i = 0; i < strLongs.length; i++) {
            long id = parseLong(strLongs[i].trim());
            longs[i] = id;
        }
        return longs;
    }

    /**
     * Erstellt aus einem csv aus Zahlen ein byte[] Array.
     * @param csv z.B. '1,2,3'
     * @return byte[] Array(csv)
     */
    public static byte[] csv2ByteArray(String csv) {
        String[] strBytes = csv.split(",");
        byte[] bytes = new byte[strBytes.length];
        for (int i = 0; i < strBytes.length; i++) {
            byte id = parseByte(strBytes[i].trim());
            bytes[i] = id;
        }
        return bytes;
    }

    /**
     * Wandelt ein CSV von Bytes in ein Boolean-Array.
     * Bereichsangaben der Form 1-3,7-8,99 sind zulaessig.
     * @param csv String
     * @return boolean[256] - byte -128 an Index 0 oder null, wenn csv null.
     */
    public static boolean[] csvBytesToBooleanArray(String csv) {
        if (null == csv) return null;
        boolean[] barr = new boolean[256];
        for (String range : csv.split(",")) {
            String fromTo[] = range.split("-");
            int from = Integer.parseInt(fromTo[0].trim());
            int to = from;
            if (fromTo.length > 1) {
                to = Integer.parseInt(fromTo[1].trim());
            }
            for (int bval = from; bval <= to; bval++) {
                int index = bval + 128;
                barr[index] = true;
            }
        }
        return barr;
    }

    /**
     * Eliminiert Duplikate aus einem long-Array und bildet
     * so faktisch ein long-Array-Set.<br>
     * @param array NOT NULL long[] Array.
     * @return new long[] Array ohne Duplikate, sortiert.
     */
    public static long[] longArrayDistinct(long[] array) {
        int length = array.length;
        if (length < 2) return Arrays.copyOf(array, length);
        long[] arr = copyOf(array, length);
        sort(arr);
        int i = 0;
        for (int j = 1; j < length; j++) {
            if (arr[j] == arr[i]) continue;
            arr[++i] = arr[j];
        }
        return copyOf(arr, i + 1);
    }

    /**
     * Kettet beliebig viele long-Arrays hintereinander und liefert
     * diese als ein Array zurueck.
     * @param arrays long[]... beliebig viele Eingabe-Arrays, auch nullable.
     * @return long[] Array. NOT NULL.
     */
    public static long[] longArrayConcat(long[]... arrays) {
        int n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) n += arrays[i].length;
        }
        long[] arr = new long[n];
        n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) {
                arraycopy(arrays[i], 0, arr, n, arrays[i].length);
                n += arrays[i].length;
            }
        }
        return arr;
    }

    /**
     * Kettet beliebig viele String-Arrays hintereinander und liefert
     * diese als ein Array zurueck.
     * @param arrays String[]... beliebig viele Eingabe-Arrays, auch nullable.
     * @return String[] Array. NOT NULL.
     */
    public static String[] stringArrayConcat(String[]... arrays) {
        int n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) n += arrays[i].length;
        }
        String[] arr = new String[n];
        n = 0;
        for (int i = 0; i < arrays.length; i++) {
            if (arrays[i] != null) {
                arraycopy(arrays[i], 0, arr, n, arrays[i].length);
                n += arrays[i].length;
            }
        }
        return arr;
    }

    public static long calcSum(final long[] values) {
        long sum = 0;
        for (int i = 0; i < values.length; i++) sum += values[i];
        return sum;
    }

    public static long calcSum(final int[] values) {
        long sum = 0;
        for (int i = 0; i < values.length; i++) sum += values[i];
        return sum;
    }

} // EOC
