/*
 * $Id: Base64.java 5816 2023-06-15 07:54:25Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import java.util.Arrays;

/**
 * Ein Base64 Encoder / Decoder.
 * Base64 kann mehrere Darstellungen annnehmen.
 * Dieser Coder arbeitet nach RFC 2045 (MIME), also
 * CRLF nach 76 Zeichen mit Padding-Zeichen =, aber auch
 * ohne CRLF, also RFC 3548 oder RFC 4648
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 * @see #encode(byte[])
 * @see #decode(String)
 */
public class Base64 {
    private Base64() {/*static*/}

    // https://en.wikipedia.org/wiki/Base64

    private static final char[] ENCDEF = {
        'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
        'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
        'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
        'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/',
    };
    private static char[] ENCURL;
    private static byte[] DECDEF;
    private static byte[] DECURL;

    static {
        ENCURL = ENCDEF.clone(); // Spezialformat Base64Url
        ENCURL[62] = '-'; // + wird zu -, da URL-Sonderzeichen
        ENCURL[63] = '_'; // / wird zu _, da ebenfalls in URL nicht erlaubt
        DECDEF = new byte[256];
        DECURL = new byte[256];
        for (byte i = 0; i < ENCDEF.length; i++) {
            DECDEF[ENCDEF[i]] = i;
            DECURL[ENCURL[i]] = i;
        }
    }


    /**
     * Base64-Encoding mit Sonderbehandlung von + und / (ohne CRLF, ohne Padding).
     * @param bytes Array of bytes, the payload
     * @return String Base64
     * @see #decodeUrl(String)
     */
    public static String encodeUrl(byte[] bytes) {
        return encode(bytes, 0, false, true);
    }

    /**
     * Base64-Encoding nach RFC 3548 / RFC 4648 (ohne CRLF).
     * @param bytes Array of bytes, the payload
     * @return String Base64
     * @see #encodeMime(byte[])
     * @see #decode(String)
     */
    public static String encode(byte[] bytes) {
        return encode(bytes, 0, true, false);
    }

    /**
     * Base64-Encoding nach RFC 2045 (MIME - CRLF nach Zeichen 76).
     * @param bytes Array of bytes, the payload
     * @return String Base64
     * @see #encodePem(byte[])
     * @see #decode(String)
     */
    public static String encodeMime(byte[] bytes) {
        return encode(bytes, 76, true, false);
    }

    /**
     * Base64-Encoding nach RFC 1421 (PEM - CRLF nach Zeichen 64).
     * <br><br><b>Wichtig:</b><br>
     * Neuere PEMs arbeiten ebenfalls nach RFC 2045, also
     * <br>{@link #encodeMime(byte[])}.
     * @param bytes Array of bytes, the payload
     * @return String Base64
     * @see #decode(String)
     */
    public static String encodePem(byte[] bytes) {
        return encode(bytes, 64, true, false);
    }

    private static String encode(byte[] bytes, int crlfPos, boolean withPadding, boolean url) {
        final char[] encTable = url ? ENCURL : ENCDEF;
        char[] buf = new char[bytes.length * 3 +2];
        int rx = 0, wx = 0, cp = 0; // Read-, WriteIndex, char position per line
        while (rx < bytes.length) {
            int nPads = 0, buf32 = 0;
            for (int shift = 16; shift >= 0; shift -= 8) {
                if (bytes.length == rx) nPads++;
                else buf32 |= (bytes[rx++] & 0xFF) << shift;
            }
            for (int shift = 18; shift >= 0; shift -= 6) {
                int idx = (buf32 >> shift) & 0x3F;
                buf[wx++] = encTable[idx];

                if (++cp == crlfPos && nPads == 0) {
                    cp = 0;
                    buf[wx++] = '\r'; // CR
                    buf[wx++] = '\n'; // LF
                }
            }

            if (withPadding) { // unberuecksichtigt: wenn url-enc, dann %3d anstatt =
                for (int i = 1; i <= nPads; i++) buf[wx -i] = '=';
            } else {
                wx -= nPads;
            }
        }
        return String.valueOf(buf, 0, wx);
    }

    /**
     * Dekodiert Base64-Standard, -Mime und -PEM encoded.
     * @param s String Base64
     * @return Array of decoded Bytes.
     * @see #encode(byte[])
     */
    public static byte[] decode(String s) {
        return decode(s, false);
    }

    /**
     * Dekodiert Base64-Url-Encoded (ohne Padding).
     * @param s String Base64
     * @return Array of decoded Bytes.
     * @see #encodeUrl(byte[])
     */
    public static byte[] decodeUrl(String s) {
        return decode(s, true);
    }

    /**
     * Dekodiert Base64-Standard, -Mime und -PEM encoded.
     * @param s String Base64
     * @param url boolean true: Ausnahme fuer + und / dekodieren
     * @return Array of decoded Bytes.
     * @see #encode(byte[])
     */
    private static byte[] decode(String s, boolean url) {
        final byte[] DEC = url ? DECURL : DECDEF;
        byte[] buf = new byte[s.length()+2]; // wenn 1zu1, dann == Padding moeglich
        int rx = 0, wx = 0; // Read/WriteIndex
        char[] chars = s.toCharArray();

        WHILE: while (rx < chars.length) {
            int nPads = 0, buf32 = 0;
            for (int shift = 18; shift >= 0; shift -= 6) {
                char c = (rx == chars.length) ? '=' : chars[rx++];

                if ('\r' == c) { // CR
                    c = chars[rx++]; // LF
                    if (chars.length == rx) break WHILE;
                    c = chars[rx++]; // next char
                }

                if ('=' == c) nPads++;
                else buf32 |= DEC[c] << shift;
            }
            for (int shift = 16; shift >= 0; shift -= 8) {
                buf[wx++] = (byte) ((buf32 >> shift) & 0xFF);
            }
            wx -= nPads;
        }
        return Arrays.copyOf(buf, wx);
    }

}
