/*
 * $Id: GeoJsonBuilder.java 5712 2022-09-23 07:52:26Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;

import java.util.Map.Entry;
import java.util.Properties;

import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.reflection.JsonBuilder;

public class GeoJsonBuilder {
    // TODO Bei Zeiten auf neuen StringBuilderUtf8 umstellen

    private StringBuilder sb = new StringBuilder(0x1000);
    private Feature feature = new Feature();
    private FeatureCollection featureCollection = new FeatureCollection();
    private GeometryCollection geometryCollection = new GeometryCollection();

    private boolean fOpen, pOpen, fcOpen, fcNext, gcOpen, gcNext;
    public void reset() {
        fOpen = pOpen = fcOpen = fcNext = gcOpen = gcNext = false;
        sb.setLength(0);
    }

    @Override public String toString() {
        String s = sb.toString();
        if (fOpen || pOpen || fcOpen || gcOpen) s += "...<BUILD() MISSING>";
        return s;
    }
    public final String build() {closeFc(); closeGc(); String s = toString(); sb.setLength(0); return s;}

    public final GeoJsonBuilder pointGeometry(LatLon latLon) {openG("Point"); geometry(latLon); closeG(); return this;}
    public final GeoJsonBuilder pointsGeometry(LatLons latLons) {openG("MultiPoint"); geometry(latLons); closeG(); return this;}
    public final GeoJsonBuilder lineGeometry(LatLons latLons) {openG("LineString"); geometry(latLons); closeG(); return this;}
    public final GeoJsonBuilder linesGeometry(LatLons[] latLonsArr) {openG("MultiLineString"); geometry(latLonsArr); closeG(); return this;}
    public final GeoJsonBuilder polyGeometry(LatLons[] latLonsArr) {openG("Polygon"); geometry(latLonsArr); closeG(); return this;}

    public final Feature pointFeature(LatLon latLon) {openF(); pointGeometry(latLon); return feature;}
    public final Feature pointsFeature(LatLons latLons) {openF(); pointsGeometry(latLons); return feature;}
    public final Feature lineFeature(LatLons latLons) {openF(); lineGeometry(latLons); return feature;}
    public final Feature linesFeature(LatLons[] latLonsArr) {openF(); linesGeometry(latLonsArr); return feature;}
    public final Feature polyFeature(LatLons[] latLonsArr) {openF(); polyGeometry(latLonsArr); return feature;}

    public final FeatureCollection featureCollection() {openFc(); return featureCollection;}
    public final GeometryCollection geometryCollection() {openGc(); return geometryCollection;}

    private void openFc() {fcOpen = true; sb.append("{\"type\":\"FeatureCollection\",\"features\":[");}
    private void closeFc() {closeF(); if (fcOpen) {sb.append("]}"); fcOpen = false; fcNext = false;};}
    private void openGc() {gcOpen = true; sb.append("{\"type\":\"GeometryCollection\",\"geometries\":[");}
    private void closeGc() {if (gcOpen) {sb.append("]}"); gcOpen = false; gcNext = false;};}
    private void openF() {fOpen = true; sb.append("{\"type\":\"Feature\",\"geometry\":");}
    private void closeF() {closeP(); if (fOpen) {sb.append('}'); fOpen = false;}}
    private void openP() {pOpen = true; sb.append("\"properties\":{");}
    private void closeP() {if (pOpen) {sb.append('}'); pOpen = false;}}
    private void openG(final String type) {sb.append("{\"type\":\"").append(type).append("\",\"coordinates\":[");}
    private void closeG() {sb.append("]}");}

    public class Feature {
        private Feature(){};
        public final String build() {return GeoJsonBuilder.this.build();}
        @Override public String toString() {return GeoJsonBuilder.this.toString();}

        private void appendKey(String key) {
            sb.append(',');	if (!pOpen) openP();
            sb.append('\"').append(key).append('\"').append(':');
        }

        public final Feature with(String key, Number val) {
            appendKey(key);
            if (null == val) sb.append("null");
            else sb.append(val.toString());
            return this;
        }

        public final Feature with(String key, VarString val) {
            if (null == val) return with(key, (String)null);
            return with(key, val.toString());
        }

        public final Feature with(String key, String val) {
            appendKey(key);
            sb.append(JsonBuilder.toJsonQuoted(val));
            return this;
        }

        public final Feature with(String key, Object obj) {
            if (obj != null) {
                appendKey(key);
                sb.append(JsonBuilder.build(obj));
            }
            return this;
        }

        public final Feature with(Properties properties) {
            for (Entry<Object, Object> entry : properties.entrySet()) {
                String key = entry.getKey().toString();
                String val = entry.getValue().toString();
                this.with(key, val);
            }
            return this;
        }

        public final FeatureCollection endWith() {
            closeF(); return fcOpen ? featureCollection : null;
        }
    }

    public class FeatureCollection {
        private FeatureCollection(){}
        public final String build() {return GeoJsonBuilder.this.build();}
        @Override public String toString() {return GeoJsonBuilder.this.toString();}
        public final Feature addPoint(LatLon latLon){nextF(); return pointFeature(latLon);}
        public final Feature addPoints(LatLons latLons){nextF(); return pointsFeature(latLons);}
        public final Feature addLine(LatLons latLons){nextF(); return lineFeature(latLons);}
        public final Feature addLines(LatLons[] latLonsArr){nextF(); return linesFeature(latLonsArr);}
        public final Feature addPoly(LatLons[] latLonsArr){nextF(); return polyFeature(latLonsArr);}
        private void nextF() {closeF();	if (fcNext) sb.append(","); else fcNext = true;}
    }

    public class GeometryCollection {
        private GeometryCollection(){}
        public final String build() {return GeoJsonBuilder.this.build();}
        @Override public String toString() {return GeoJsonBuilder.this.toString();}
        public final GeometryCollection addPoint(LatLon latLon){nextG(); pointGeometry(latLon); return this;}
        public final GeometryCollection addPoints(LatLons latLons){nextG(); pointsGeometry(latLons); return this;}
        public final GeometryCollection addLine(LatLons latLons){nextG(); lineGeometry(latLons); return this;}
        public final GeometryCollection addPoly(LatLons[] latLonsArr){nextG(); polyGeometry(latLonsArr); return this;}
        private void nextG() {if (gcNext) sb.append(","); else gcNext = true;}
    }

    private void geometry(final LatLon latLon) {
        sb.append(latLon.getLon()).append(',').append(latLon.getLat());
    }

    private void geometry(final LatLons latLons) {
        int n = latLons.getSize();
        for (int i = 0; i < n; i++) {
            long coord = latLons.getCoord(i);
            if (i > 0) sb.append(',');
            sb.append('[');
            sb.append(lonOf(coord)).append(',').append(latOf(coord));
            sb.append(']');
        }
    }
    private void geometry(final LatLons[] latLonsArr) {
        sb.append('[');
        for (int i = 0; i < latLonsArr.length; i++) {
            if (i > 0) sb.append("],[");
            geometry(latLonsArr[i]);
        }
        sb.append(']');
    }

}
