/*
 * $Id: GeoUtils.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static java.lang.Math.PI;
import static java.lang.StrictMath.atan;
import static java.lang.StrictMath.atan2;
import static java.lang.StrictMath.cos;
import static java.lang.StrictMath.pow;
import static java.lang.StrictMath.sin;
import static java.lang.StrictMath.sinh;
import static java.lang.StrictMath.sqrt;
import static java.lang.StrictMath.toDegrees;
import static java.lang.StrictMath.toRadians;

import java.util.Arrays;

import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;

public class GeoUtils {

    private GeoUtils() {/*static*/}

    public final static double RADIUS_EARTH_KM = 6371d; // Mittlerer Erdradius
    public final static double GRAD_KM = 2 * RADIUS_EARTH_KM * PI / 360;
    public final static double EPSG_900913_CONST = 20037508.342789244 / 180; // Umfang / 2 [m]

    public final static double PI_HALF = PI / 2;
    public final static double RAD = PI / 180;
    public final static double DEG = 180 / PI;
    public final static double RAD_HALF = PI / 360;

    // ######################## FAST COS/SIN ############################

    private static final int SIN_BITS = 12; // Dies ist die Genauigkeit
    private static final int SIN_MASK = ~(-1 << SIN_BITS);
    private static final int SIN_COUNT = SIN_MASK + 1;
    private static final float RAD_FULL = (float) (Math.PI * 2.0);
    private static final float DEG_FULL = (float) (360.0);
    private static final float RAD2INDEX = SIN_COUNT / RAD_FULL;
    private static final float DEG2INDEX = SIN_COUNT / DEG_FULL;
    private static final float[] coss, sins;

    static {
       // see: http://www.java-gaming.org/index.php?topic=24191.0
       sins = new float[SIN_COUNT];
       coss = new float[SIN_COUNT];

       for (int i = 0; i < SIN_COUNT; i++) {
          sins[i] = (float) Math.sin((i + 0.5f) / SIN_COUNT * RAD_FULL);
          coss[i] = (float) Math.cos((i + 0.5f) / SIN_COUNT * RAD_FULL);
       }
       for (int i = 0; i < 360; i += 90) {
          sins[(int)(i * DEG2INDEX) & SIN_MASK] = (float)Math.sin(i * Math.PI / 180.0);
          coss[(int)(i * DEG2INDEX) & SIN_MASK] = (float)Math.cos(i * Math.PI / 180.0);
       }
    }
    public static final float sinRad(float rad) {
        return sins[(int) (rad * RAD2INDEX) & SIN_MASK];
    }
    public static final float sinDeg(float deg) {
        return sins[(int) (deg * DEG2INDEX) & SIN_MASK];
    }
    public static final float cosRad(float rad) {
        return coss[(int) (rad * RAD2INDEX) & SIN_MASK];
    }
    public static final float cosDeg(float deg) {
        return coss[(int) (deg * DEG2INDEX) & SIN_MASK];
    }

    // ######################## ANGLE ############################

    public static double calcAngle(double lat1, double lon1,
            double lat2, double lon2, boolean clockWise) {
        double dy = lat2 - lat1;
        double dx = lon2 - lon1;
        double deg = DEG * StrictMath.atan2(dy, dx);
        if (clockWise) deg = 90 - deg;
        if (deg < 0) deg += 360;
        return deg;
    }

    // ###################### GIS-FUNCTIONS ###########################################

    /**
     * Liefert die tatsaechliche Distanz zweier LatLon-Koordinaten in Kilometern.
     * Es wird die Erde als Kugel (Sphere) angenommen
     * und nicht als Spheroid (Ellipse).<br>
     * Die Koordinaten muessen als geodezimale Werte
     * (EPGS 4326 / WGS84) uebergeben werden.
     * <br><br><b>Gut zu wissen:</b><br>
     * Es existieren zwei Verfahen. Zum einen ist da die 'Haversine Formula',
     * zum anderen die 'Spherical Law of Cosines', die hier verwendet wird.
     * Letztere ist schneller und die Genauigkeit ist auf modernen Computern
     * mit exakter Floating-Point-Arithmetik gegeben.
     * @param lat1 y1 double
     * @param lon1 x1 double
     * @param lat2 y2 double
     * @param lon2 x2 double
     * @return double Kilometer
     * @see #calcDistanceEuklid(double, double, double, double)
     * @see #isReachableDistance(float, float, float, float, float)
     */
    public static double calcDistance(
            double lat1, double lon1, double lat2, double lon2) {
        // see: http://www.movable-type.co.uk/scripts/latlong.html
        double d = StrictMath.acos(StrictMath.sin(lat1 * RAD) * StrictMath.sin(lat2 * RAD)
                + StrictMath.cos(lat1 * RAD)* StrictMath.cos(lat2 * RAD)
                * StrictMath.cos((lon2-lon1) * RAD))
                * RADIUS_EARTH_KM;
        if (Double.isNaN(d)) return 0;
        return d;
    }

    /**
     * Liefert die Rhumb-Distanz zweier LatLon-Koordinaten in Kilometern.
     * Es wird die Erde als Kugel (Sphere) angenommen und nicht als Spheroid (Ellipse).<br>
     * Die Koordinaten muessen als geodezimale Werte (EPGS 4326 / WGS84) uebergeben werden.
     * <br><br><b>Gut zu wissen:</b><br>
     * Es existieren zwei grundsaetzliche Verfahen. Zum einen ist da die
     * 'Great Circle'-Distanz {@link #calcDistance(double, double, double, double)},
     * zum anderen die 'Rhumb Line'. Erstere liefert die tatsaechliche Distanz. Diese
     * Variante hier wird hauptsaechlich in der Schiffs-Navigation eingesetzt, die einen
     * Kurs in Grad vorgibt und nach einer gewissen Strecke einen naechsten Wendepunkt
     * setzt. Rhumb ist dadurch immer laenger, da 'Great Circle' ein permanentes
     * Nachfuehren des Ruders erfordert. Rhumb liefert auf einer Mercator projezierten
     * Karte exakt gerade Strecken, da der Tangens der Projektion herausgerechnet wird.
     * @param lat1 y1 double
     * @param lon1 x1 double
     * @param lat2 y2 double
     * @param lon2 x2 double
     * @return double Kilometer
     * @see #calcDistance(double, double, double, double)
     */
    public static double calcDistanceRhumb(
            double lat1, double lon1, double lat2, double lon2) {
        lat1 *= GeoUtils.RAD;
        lon1 *= GeoUtils.RAD;
        lat2 *= GeoUtils.RAD;
        lon2 *= GeoUtils.RAD;

        double dlon = lon2 - lon1;
        double dlat = lat2 - lat1;

        double dw = Math.log(Math.tan(Math.PI/4+lat2/2)/Math.tan(Math.PI/4+lat1/2));
        double q = Math.abs(dlat) > 10e-12 ? dlat/dw : Math.cos(lat1);
        if (Math.abs(dlon) > Math.PI) dlon = dlon>0 ? -(2*Math.PI-dlon) : (2*Math.PI+dlon);
        double dist = Math.sqrt(dlat*dlat + q*q*dlon*dlon) * GeoUtils.RADIUS_EARTH_KM;
        return dist;
    }


    /**
     * Liefert die ungefaehre aber spherische Distanz zweier LatLon-Koordinaten
     * in Kilometern. Es wird die Erde als Kugel (Sphere) angenommen
     * und nicht als Spheroid (Ellipse).<br>
     * Die Koordinaten muessen als geodezimale Werte
     * (EPGS 4326 / WGS84) uebergeben werden.<br>
     * Diese Variante umgeht diverse teure trigonometrische Aufrufe und ist
     * daher nur fuer Approximationen geeignet, so z.B. fuer AStar-Heuristiken.<br>
     * Dabei gilt: Je groesser die Distanz, desto kleiner der Fehler. Ab 50km liegt
     * die maximale Abweichungen bei ca.50%, darunter bis zu 100%.
     * @param lat1 y1 float
     * @param lon1 x1 float
     * @param lat2 y2 float
     * @param lon2 x2 float
     * @return double Kilometer
     * @see #calcDistance(double, double, double, double)
     */
    public static double calcDistanceCoarse(
            float lat1, float lon1, float lat2, float lon2) {
        // see: http://www.movable-type.co.uk/scripts/latlong.html
        double d = RADIUS_EARTH_KM * StrictMath.acos(
                sinDeg(lat1) * sinDeg(lat2)
                + cosDeg(lat1) * cosDeg(lat2)
                * cosDeg(lon2-lon1)
                );
        if (Double.isNaN(d)) return 0;
        return d;
    }

    /**
     * Liefert die tatsaechliche Distanz zweier LatLon-Koordinaten in Kilometern.
     * Es wird die Erde als Landkarte angenommen
     * und nicht als Kugel (Sphere) oder Spheroid (Ellipse).
     * Hierdurch werden Uebergaenge bei +Max nach -Max-Graden falsch gerechnet.<br>
     * Die Koordinaten muessen als geodezimale Werte
     * (EPGS 4326 / WGS84) uebergeben werden. Zudem ist diese Methode nur
     * fuer kurze Distanzen &lt; 10km geeignet.
     * @param lat1 y1 double
     * @param lon1 x1 double
     * @param lat2 y2 double
     * @param lon2 x2 double
     * @return double Kilometer
     * @see #calcDistance(double, double, double, double)
     */
    public static double calcDistanceEuklid(
            double lat1, double lon1, double lat2, double lon2) {
    	double x = (lon2 - lon1) * RAD * StrictMath.cos((lat1 + lat2) * RAD_HALF);
    	double y = (lat2 - lat1) * RAD;
    	double d = StrictMath.hypot(x, y) * RADIUS_EARTH_KM;
        return d;
    }

    /**
     * Liefert die ungefaehre Distanz zweier LatLon-Koordinaten in Kilometern.
     * Diese Variante umgeht diverse teure trigonometrische Aufrufe und ist
     * daher nur fuer Aproximationen geeignet, so z.B. fuer AStar-Heuristiken.
     * Es wird die Erde als Landkarte angenommen
     * und nicht als Kugel (Sphere) oder Spheroid (Ellipse).
     * Hierdurch werden Uebergaenge bei +Max nach -Max-Graden falsch gerechnet.<br>
     * Die Koordinaten muessen als geodezimale Werte
     * (EPGS 4326 / WGS84) uebergeben werden.
     * @param lat1 y1 float
     * @param lon1 x1 float
     * @param lat2 y2 float
     * @param lon2 x2 float
     * @return double Kilometer
     * @see #calcDistanceEuklid(double, double, double, double)
     */
    public static double calcDistanceEuklidCoarse(
            float lat1, float lon1, float lat2, float lon2) {
        float cosx = cosDeg((lat1 + lat2) / 2f);
        float x = (lon2 - lon1) * cosx;
        float y = lat2 - lat1;
        double dist = StrictMath.sqrt(x*x + y*y) * GRAD_KM;
        return dist;
    }

    /**
     * Liefert die Manhatten-Distanz zweier LatLon-Koordinaten
     * (EPGS 4326 / WGS84) in Kilometern.<br>
     * Diese Distanz ist eine reine Addition der Deltas von
     * Breiten- und Laengengrad. Diese Methode ist daher schnell und
     * kann an diversen Stellen als Filter fuer die eigentlich teuren
     * Distanzberechnungen vorgeschaltet werden.<br><br>
     * Bildlich entsteht hier eine <strong>Raute</strong>, die waagerecht
     * und senkrecht volle Ausdehnung hat. Bei 45 Grad allerdings
     * <code>sqr(((dist / 2)^2)*2)</code> berechnet. Beispiel Laenge 2:<br>
     * <code>
     * Bei&nbsp;45 Grad: _| &nbsp;(1 nach rechts plus 1 nach oben) <br>
     * Bei&nbsp;&nbsp;0 Grad: -- (2 nach rechts) <br>
     * </code>
     * Bei einer Umkreisbetrachtung mittels ManhattenBox kann dies durch
     * den Faktor<br><code>1.4 =~ 1/sin(45 Grad)</code> korrigiert werden,
     * damit sich der Kreis innerhalb der Raute befindet.<br>
     * Zudem korriegert diese Methode Plus- und Minus- Uebergaenge
     * und verkuerzt die Laengengrade mittels eines groben Kosinus
     * zu den Polen hin.
     * @param lat1 y1 float
     * @param lon1 x1 float
     * @param lat2 y2 float
     * @param lon2 x2 float
     * @return float Grobe Distanz in Kilometern
     * @see #calcDistance(double, double, double, double)
     * @see #calcDistanceEuklid(double, double, double, double)
     * @see #isReachableDistance(float, float, float, float, float)
     */
    public static float calcDistanceManhatten(
            float lat1, float lon1, float lat2, float lon2) {
        float dlat = lat2 - lat1;
        float dlon = lon2 - lon1;
        if (dlon < 0f) dlon = -dlon; // abs
        if (dlat < 0f) dlat = -dlat; // abs
        // Pseudo spherical:
        if (dlat > 90f) dlat = 180f - dlat; // Ueber Pole kuerzer
        if (dlon > 180f) dlon = 360f - dlon; // Andersrum kuerzer
        dlon *= cosDeg((lat1 + lat2) / 2); // Zu den Polen immer kuerzer
        return (float) ((dlat + dlon) * GRAD_KM);
    }

    /**
     * Berechnet die Erreichbarkeit zwischen zwei Punkten. Kombiniert hierzu
     * {@link #calcDistanceManhatten(float, float, float, float)} und
     * {@link #calcDistance(double, double, double, double)} fuer eine
     * optimale Performanz.
     *
     * @param dist float Maximale Distanz in km
     * @param lat1 y1 float
     * @param lon1 x1 float
     * @param lat2 y2 float
     * @param lon2 x2 float
     * @return boolean Erreichbar innerhalb Distanz ja/nein
     */
    public static boolean isReachableDistance(
            float dist, float lat1, float lon1, float lat2, float lon2) {
        float distManhatten = calcDistanceManhatten(lat1, lon1, lat2, lon2);
        if (distManhatten * 0.7 < dist) { // PerfomanceFilter  0.7==(1/1.4) s.o.
            return calcDistance(lat1, lon1, lat2, lon2) <= dist;
        }
        return false;
    }


    // ###################### LAT-LON-Transformationen ##########################

    public static double EPSG_4326_TO_900913_Lon(double lon) {
        double x = lon * EPSG_900913_CONST;
        return x;
    }

    public static double EPSG_900913_TO_4326_Lon(double lon) {
        double x =  lon / EPSG_900913_CONST;
        return x;
    }

    public static double EPSG_4326_TO_900913_Lat(double lat) {
        double y = StrictMath.log(StrictMath.tan((90 + lat) * RAD_HALF)) / RAD;
        y = y * EPSG_900913_CONST;
        return y;
    }

    public static double EPSG_900913_TO_4326_Lat(double lat) {
        double y = lat / EPSG_900913_CONST;
        y = DEG * (2 * StrictMath.atan(StrictMath.exp(y * RAD)) - PI_HALF);
        return y;
    }

    /**
     * Liefert den Laengengrad eines Slippy-Tiles.
     * @param x int Tile-X, x=0 ist links.
     * @param zoom int Zoom-Stufe
     * @return Laengengrad der linken Seite des Tiles
     * @see #tile2lat(int, int)
     */
    public static double tile2lon(int x, int zoom) {
        return x / pow(2.0, zoom) * 360.0 - 180;
    }

    /**
     * Liefert den Breitengrad eines Slippy-Tiles.
     * <br><b>Beispiel BBox:</b>
     * <pre>
     * minLat = tile2lat(y + 1, zoom);
     * minLon = tile2lon(x, zoom);
     * maxLat = tile2lat(y, zoom);
     * maxLon = tile2lon(x + 1, zoom);
     * </pre>
     * @param y int Tile-Y (v.o.n.u.) Hier Achtung: y=0 ist oben!
     * @param zoom int Zoom-Stufe
     * @return Breitengrad der oberen Seite des Tiles
     * @see #tile2lon(int, int)
     */
    public static double tile2lat(int y, int zoom) {
        double n = PI - (2.0 * PI * y) / pow(2.0, zoom);
        return toDegrees(atan(sinh(n)));
    }


    // ########################## QuadTile #############################

    /**
     * Generiert aus einer Koordinate einen QuadTile-Index.
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     * @return long QuadTile-Index
     */
    public static long calcQuadTile(double lat, double lon) {
        int ilat = (int) (lat * 1e7);
        int ilon = (int) (lon * 1e7);
        return interlaceBits(ilat, ilon);
    }
    /**
     * Generiert aus einer Koordinate einen QuadTile-Index.
     * @param coord LatLon-Darstellung als long
     * @return long QuadTile-Index
     */
    public static long calcQuadTile(long coord) {
        int ilat = (int) (coord >>> 32);
        int ilon = (int) (coord & 0xFFFFFFFF);
        return interlaceBits(ilat, ilon);
    }

    /**
     * Verzahnt zwei int-Werte abwechselnd bitweise zu einem long.
     * @param lo int
     * @param hi int
     * @return Bitweise verzahnte Bits als long
     */
    private static long interlaceBits(int lo, int hi) {
        long lo2 = 0, hi2 = 0, bit2 = 1;
        for (int i = 0, bit = 1; i < 32; i++, bit <<= 1) {
            if ((lo & bit) != 0) lo2 |= bit2;
            if ((hi & bit) != 0) hi2 |= bit2;
            bit2 <<= 2;
        }
        return lo2 | hi2 << 1;
    }

    // ######################### Geometry Functions ############################

    /**
     * Berechnet die Triangulation dreier Punkte
     * und somit den Abbiege-Winkel.
     * @param y1 Punkt 1 y
     * @param x1 Punkt 1 x
     * @param y2 Punkt 2 y
     * @param x2 Punkt 2 x
     * @param y3 Punkt 3 y
     * @param x3 Punkt 3 x
     * @return &lt; 0 Rechtsknick, &gt; 0 Linksknick, = 0 geradeaus.
     */
    public final static double ccw(
    		double y1, double x1, double y2, double x2, double y3, double x3) {
	    return ((x2 - x1) * (y3 - y1) - (y2 - y1) * (x3 - x1));
    }

    /**
     * Berechnet den Einheitsvektor aus einer Strecke.
     * @param y0 double Startpunkt y
     * @param x0 double Startpunkt x
     * @param y1 double Endpunkte y
     * @param x1 double Endpunkt x
     * @return {@link LatLon} der den normalisierten Vektor darstellt.
     */
    public final static LatLon calcUnitVector(
            double y0, double x0, double y1, double x1) {
        double dy = y1 - y0, dx = x1 - x0;
        double len = sqrt(dy*dy + dx*dx);
        dy /= len; dx /= len;
        return new LatLon(dy, dx);
    }

    /**
     * Liefert den Kreuzungspunkt zweier Strecken.
     * @param x1 double Line 1 Start-Coord x
     * @param y1 double Line 1 Start-Coord y
     * @param x2 double Line 1 End-Coord x
     * @param y2 double Line 1 End-Coord y
     * @param x3 double Line 2 Start-Coord x
     * @param y3 double Line 2 Start-Coord y
     * @param x4 double Line 2 End-Coord x
     * @param y4 double Line 2 End-Coord y
     * @param mustCross true: nur echte Kreuzungen liefern, false: auch Beruehrung
     * @return {@link LatLon} Kreuzungspunkt oder NULL wenn keine Ueberschneidung.
     */
    public final static LatLon intersection(double x1, double y1, double x2, double y2,
            double x3, double y3, double x4, double y4, boolean mustCross) {
        double denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1);
        if (denom == 0.0) return null;  // parallel oder identisch
        double ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denom;
        double ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / denom;
        if (mustCross) {
            if (ua > 0.0 && ua < 1.0 && ub > 0.0 && ub < 1.0) {
                return new LatLon(y1 + ua * (y2 - y1), x1 + ua * (x2 - x1));
            }
        } else {
            if (ua >= 0.0 && ua <= 1.0 && ub >= 0.0 && ub <= 1.0) {
                return new LatLon(y1 + ua * (y2 - y1), x1 + ua * (x2 - x1));
            }
        }
        return null;
    }

    /**
     * Generiert ein Teilstueck einer GreatCircle-Strecke.
     * @param lat1 double Start-Point Latitude
     * @param lon1 double Start-Point Longitude
     * @param lat2 double End-Point Latitude
     * @param lon2 double End-Point Longitude
     * @param fraction double z.B. 0:Start, 1:End, 0.5:Mittelpunt der GreatCircle-Linie.
     * @return {@link LatLon} Koordinate, abhaengig vom Anteil (fraction)
     * @see <a href="http://www.movable-type.co.uk/scripts/latlong.html">intermediatePointTo()</a>
     */
    public static LatLon computeGreatCircleLineFraction(
            double lat1, double lon1, double lat2, double lon2, double fraction) {
        lat1 = toRadians(lat1); lon1 = toRadians(lon1);
        lat2 = toRadians(lat2); lon2 = toRadians(lon2);

        double sinLat1 = sin(lat1), cosLat1 = cos(lat1), sinLon1 = sin(lon1), cosLon1 = cos(lon1);
        double sinLat2 = sin(lat2), cosLat2 = cos(lat2), sinLon2 = sin(lon2), cosLon2 = cos(lon2);

        // distance between points
        double dLat = lat2 - lat1;
        double dLon = lon2 - lon1;

        double a = sin(dLat/2) * sin(dLat/2)
            + cos(lat1) * cos(lat2) * sin(dLon/2) * sin(dLon/2);

        double o = 2 * atan2(sqrt(a), sqrt(1-a));

        double A = sin((1-fraction)*o) / sin(o);
        double B = sin(fraction*o) / sin(o);

        double x = A * cosLat1 * cosLon1 + B * cosLat2 * cosLon2;
        double y = A * cosLat1 * sinLon1 + B * cosLat2 * sinLon2;
        double z = A * sinLat1 + B * sinLat2;

        double lat3 = atan2(z, sqrt(x*x + y*y));
        double lon3 = atan2(y, x);

        lat3 = toDegrees(lat3);
        lon3 = (toDegrees(lon3) + 540) % 360 - 180;

        return new LatLon(lat3, lon3);

    }

    /**
     * Generiert ein Teilstueck einer GreatCircle-Strecke.
     * @param source {@link LatLon} Start-Point
     * @param target {@link LatLon} End-Point
     * @param fraction double z.B. 0:Start, 1:End, 0.5:Mittelpunt der GreatCircle-Linie.
     * @return {@link LatLon} Koordinate, abhaengig vom Anteil (fraction)
     * @see <a href="http://www.movable-type.co.uk/scripts/latlong.html">intermediatePointTo()</a>
     */
    public static LatLon computeGreatCircleLineFraction(
            LatLon source, LatLon target, double fraction) {
        return computeGreatCircleLineFraction(
                source.getLat(), source.getLon(), target.getLat(), target.getLon(), fraction);
    };

    /**
     * Reichert eine gerade Strecke zwischen Start und Ziel um GreatCircle-Punkte an.
     * @param source {@link LatLon} Startpunkt
     * @param target {@link LatLon} Zielpunkt
     * @param n int Anzahl der Strecken-Teilstuecke (nicht Punkte!)
     * @return {@link LatLons} Geometrie
     */
    public static LatLons computeGreatCircleLine(LatLon source, LatLon target, int n) {
        double fraction = 1d / n;
        LatLons gcLine = new LatLons();
        gcLine.append(source);
        for (double f = fraction; f < 1d; f += fraction) {
            LatLon p = computeGreatCircleLineFraction(source, target, f);
            gcLine.append(p);
        }
        gcLine.append(target);
        return gcLine;
    }

    /**
     * Ermittelt die Indizes der Kanten, die die Dateline kreuzen.
     * @param line {@link LatLons} Koordinaten
     * @return Array von Indizes von Koordinaten oder NULL.
     *      Die Zaehlung beginnt bei 1, also der Kante von Index 0 nach 1.
     *      Die Kreuzungs-Richtung liefert das Vorzeichen.
     *      Ein positiver Index bedeutet von links nach rechts, ein negativer
     *      ist die entgegengesetzte Richtung.
     */
    public static int[] dateLines(LatLons line) {
        if (null == line) return null;
        long[] coords = line.getCoords();
        if (null == coords || coords.length < 2) return null;

        double lat1 = latOf(coords[0]);
        double lon1 = lonOf(coords[0]);
        double lat2 = lat1;
        double lon2 = lon1;

        int n = 0;
        int[] dateLines = new int[0x100];
        for (int i = 1; i < coords.length; i++) {
            lat2 = latOf(coords[i]);
            lon2 = lonOf(coords[i]);
            if (Math.abs(lon2 - lon1) > 180) {
                if (dateLines.length == n) // ArrayOverflow
                    dateLines = Arrays.copyOf(dateLines, n*2);
                dateLines[n++] = lon2 < lon1 ? i : -i;
            }
            lat1 = lat2;
            lon1 = lon2;
        }

        dateLines = 0 == n ? null : Arrays.copyOf(dateLines, n);
        return dateLines;
    }


    public static LatLons[] splitAtDateLine(LatLons line) {
        if (null == line) return null;
        long[] coords = line.getCoords();
        if (null == coords || coords.length < 2) return null;

        int frax = 0;
        LatLons[] fractions = new LatLons[16];
        fractions[frax] = new LatLons();
        fractions[frax].append(coords[0]);

        double lat1 = latOf(coords[0]);
        double lon1 = lonOf(coords[0]);
        double lat2 = lat1;
        double lon2 = lon1;

        for (int i = 1; i < coords.length; i++) {
            lat2 = latOf(coords[i]);
            lon2 = lonOf(coords[i]);

            if (Math.abs(lon2 - lon1) > 180) {

                if (lon2 < lon1) {
                    double lon = lon2 + 360;
                    double dx = (180 - lon1) / (lon - lon1);
                    double lat = (lat2 - lat1) * dx + lat1;

                    fractions[frax].append(lat, 180);
                    fractions[++frax] = new LatLons();
                    fractions[frax].append(lat, -180);

//                    System.out.println("left-right " + lon1  + "->" + lon);
                } else {
                    double lon = lon2 - 360;
                    double dx = (-180 - lon1) / (lon - lon1);
                    double lat = (lat2 - lat1) * dx + lat1;

                    fractions[frax].append(lat, -180);
                    fractions[++frax] = new LatLons();
                    fractions[frax].append(lat, 180);

//                    System.out.println("right-left " + lon1  + "->" + lon);
                }
            }

            fractions[frax].append(coords[i]);

            lat1 = lat2;
            lon1 = lon2;
        }

        return Arrays.copyOf(fractions, frax +1);
    }


} // EOC
