/*
 * $Id: HttpClient.java 5857 2024-02-24 11:58:07Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.misc.Utils.UTF8;
import static java.net.HttpURLConnection.HTTP_INTERNAL_ERROR;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import de.cm.osm2po.logging.Log;
/**
 * Ein simpler Client fuer HTTP-Anfragen.
 * @author (c) 2023 - Carsten Moeller, info@osm2po.de
 */
public final class HttpClient {
    // https://www.baeldung.com/java-http-request

    private final static String MIME_JSON = "application/json";
    private final static String MIME_FORM = "application/x-www-form-urlencoded";

    public static class HttpClientResponse {
        public final int httpCode;
        public final String mimeType;
        public final byte[] responseBody;
        public final Map<String, List<String>> headers;

        private HttpClientResponse(int httpCode, String mimeType,
                Map<String, List<String>> headers, byte[] responseBody) {
            this.httpCode = httpCode;
            this.mimeType = mimeType;
            this.headers = headers;
            this.responseBody = responseBody;
        }

        public final boolean isJson() {
            return this.httpCode == 200 && MIME_JSON.equals(this.mimeType);
        }

        @Override
        public String toString() {
            String responseBody = "<NULL>";
            if (this.responseBody != null) responseBody = new String(this.responseBody, UTF8);
            return "httpCode=" + httpCode + "\nmimeType=" + mimeType + "\nheaders=" + headers
                    + "\nresponseBody=\n" + new String(responseBody);
        }
    }

    public enum Method {
        GET, POST, PUT, PATCH, DELETE
    }

    public final String baseUrl;

    private final Map<String, String> queryParams = new HashMap<>();
    private final Map<String, String> formParams = new HashMap<>();
    private final Map<String, String> headers = new HashMap<>();

    private String contentType = "text/plain";
    private int connectTimeout = 0;
    private int readTimeout = 0;
    private boolean followRedirects = true;

    private Log log = Log.stderr();

    public HttpClient(String baseUrl) {
        this.baseUrl = baseUrl;
    }

    public HttpClient setLog(Log log) {
        if (log != null) this.log = log;
        return this;
    }

    public HttpClient setReadTimeout(int readTimeout) {
        this.readTimeout = readTimeout;
        return this;
    }

    public HttpClient setConnectTimeout(int connectTimeout) {
        this.connectTimeout = connectTimeout;
        return this;
    }

    public HttpClient setFollowRedirects(boolean followRedirects) {
        this.followRedirects = followRedirects;
        return this;
    }

    public HttpClient clearParams() {
        this.queryParams.clear();
        this.formParams.clear();
        return this;
    }

    public HttpClient clearHeaders() {
        this.headers.clear();
        return this;
    }

    public HttpClient putQueryParam(String key, Object value) {
        return putParam(key, value, true);
    }

    public HttpClient putFormParam(String key, Object value) {
        return putParam(key, value, false);
    }

    // value as Object allows simple int's eg. to be passed here
    public HttpClient putParam(String key, Object value, boolean asQueryParam) {
        Map<String, String> targetMap = asQueryParam ? this.queryParams : this.formParams;
        try { // encode(key, charset) erst ab Java11
            targetMap.put(
                URLEncoder.encode(key, "UTF-8"),
                URLEncoder.encode(value.toString(), "UTF-8"));
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
        return this;
    }

    public HttpClient setContentType(String contentType) {
        this.contentType = contentType;
        return this;
    }

    public HttpClient setContentTypeJson() {
        return setContentType(MIME_JSON);
    }

    public HttpClient setContentTypeForm() {
        return setContentType(MIME_FORM);
    }

    public HttpClient putHeader(String key, String... value) {
        if ("Content-Type".equalsIgnoreCase(key)) throw new IllegalArgumentException(
                "Use setContentType() instead");

        String values = Stream.of(value).collect(Collectors.joining(", "));
        this.headers.put(key, values);
        return this;
    }

    public HttpClientResponse doGetRequest() throws Exception {
        return doGetRequest(null);
    }

    public HttpClientResponse doGetRequest(String endpoint) throws Exception {
        return doRequest(Method.GET, endpoint, null);
    }

    public HttpClientResponse doPostRequest(String endpoint, final byte[] payload) throws Exception {
        return doRequest(Method.POST, endpoint, payload);
    }

    public HttpClientResponse doPostRequest(final byte[] payload) throws Exception {
        return doPostRequest(null, payload);
    }

    public HttpClientResponse doPostRequest(String endpoint) throws Exception {
        return doPostRequest(endpoint, null);
    }

    public HttpClientResponse doPostRequest() throws Exception {
        return doPostRequest(null, null);
    }

    private String paramsToString(Map<String, String> paramMap) {
        StringBuilder sb = new StringBuilder(0x100);
        boolean isNotFirst = false;
        for (Map.Entry<String, String> entry : paramMap.entrySet()) {
            if (isNotFirst) sb.append("&");
            sb.append(entry.getKey()).append("=").append(entry.getValue());
            isNotFirst = true;
        }
        return sb.toString();
    }

    public HttpClientResponse doRequest(Method method, String endpoint, byte[] payload) {
        try {
            if (null == endpoint) endpoint = "";
            final String queryParams = this.queryParams.isEmpty() ? ""
                    : "?" + paramsToString(this.queryParams);

            URL url = new URI(this.baseUrl + endpoint + queryParams).toURL();

            HttpURLConnection con = (HttpURLConnection) url.openConnection();

            con.setRequestProperty("Content-Type", this.contentType);
            for (Map.Entry<String, String> entry : this.headers.entrySet()) {
                con.setRequestProperty(entry.getKey(), entry.getValue());
            }

            con.setRequestMethod(method.name());
            con.setDoInput(true);

            con.setConnectTimeout(this.connectTimeout);
            con.setReadTimeout(this.readTimeout);
            con.setInstanceFollowRedirects(this.followRedirects);
            // FollowRedirects itself is global static for URL

            if (null == payload && MIME_FORM.equalsIgnoreCase(this.contentType)) {
                payload = paramsToString(this.formParams).getBytes();
            }

            if (payload != null) {
                con.setDoOutput(true);
                con.getOutputStream().write(payload);
            }

            con.connect(); // muss nach getOutputStream kommen

            byte[] responseBody = new byte[0];
            try {
                ByteArrayOutputStream baos = new ByteArrayOutputStream(0x800);
                InputStream is = con.getInputStream();
                byte[] buf = new byte[4096];
                for (int len = is.read(buf); len > -1; len = is.read(buf))
                    baos.write(buf, 0, len);
                is.close();
                responseBody = baos.toByteArray();

            } catch (Exception e) {
                log.error(e);
            }

            final HttpClientResponse hcr = new HttpClientResponse(con.getResponseCode(),
                    con.getContentType(), con.getHeaderFields(), responseBody);

            con.disconnect();
            return hcr;

        } catch (Exception e) {
            return new HttpClientResponse(HTTP_INTERNAL_ERROR, "text/plain", null, e.toString().getBytes());
        }

    }
}
