/*
 * $Id: MinorStreetDetector.java 5848 2023-10-19 14:08:07Z carsten $
 *
 * Copyright 2020 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.misc.GeoUtils.calcDistance;
import static de.cm.osm2po.routing.Graph.SUPPORT_EDGEFLAGS;
import static de.cm.osm2po.routing.Graph.SUPPORT_LATLON;
import static de.cm.osm2po.routing.Graph.toEdgeId;
import static de.cm.osm2po.routing.Graph.toSegmentId;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Arrays;

import de.cm.osm2po.Config;
import de.cm.osm2po.converter.graph.GraphBuildInterceptor;
import de.cm.osm2po.converter.joiner.DefaultJoinFilter;
import de.cm.osm2po.converter.tiler.DefaultWayTagResolver;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.SegmentedWay;
import de.cm.osm2po.model.Vertex;
import de.cm.osm2po.model.Way;
import de.cm.osm2po.model.WaySegment;
import de.cm.osm2po.postp.PostProcessor;
import de.cm.osm2po.primitives.BigLongSet;
import de.cm.osm2po.primitives.IntLongBag;
import de.cm.osm2po.primitives.IntSet;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarLongArray;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.RouterConstraints;
import de.cm.osm2po.routing.RouterParameters;
import de.cm.osm2po.routing.VertexRouter;

/**
 * Multi-KLasse, die sowohl von {@link DefaultJoinFilter} erbt,
 * als auch {@link GraphBuildInterceptor} und {@link PostProcessor} implementiert
 * und eine Filterung benoetigter Wege niederer Typen ermoeglicht.<br>
 * Config-Beispiel:
 * <pre>
 * minorstreet.clazzes=51,63
 * minorstreet.detection.bean.file=[workDir]/minorstreets.bean
 * msdclass=de.cm.osm2po.misc.MinorStreetDetector
 * jobs=first,second
 * .first.cmd=cg
 * .first.graph.build.interceptor.class=[msdclass]
 * .first.postp.0.class=[msdclass]
 * .second.cmd=jsgr
 * .second.loadbean.minorstreets=[minorstreet.detection.bean.file]
 * .second.joinFilter.class=[msdclass]
 * </pre>
 * Das Ergebnis der ersten Phase ist eine {@link VarBean}-Datei vom Typ
 * {@link VarLongArray} mit einer Sammlung von OSM-IDs von Strassen der
 * konfigurierten Klassen (Beispiel oben: 51,63) ohne die einige Routen nicht
 * gefunden werden koennen.<br>
 * Der hier ueberschriebene {@link DefaultWayTagResolver} uebernimmt in
 * der zweiten Phase (2. Durchlauf) dann diese IDs und filtert diese raus.
 * @author (c) 2020 - Carsten Moeller, info@osm2po.de
 */
public class MinorStreetDetector extends DefaultJoinFilter implements GraphBuildInterceptor, PostProcessor {

    private final static String BEANNAME_SEG2OSMID = "segIdToOsmIdBag";
    private final static String BEANFILE = "minorstreets.bean";

    private boolean[] clazzSet = new boolean[128];
    private IntLongBag segIdToOsmIdBag;
    private BigLongSet idSet = new BigLongSet();
    private SegmentedWay way;

    private void initClazzSet(Config config) {
        String clazzes = config.getProperty("minorstreet.clazzes");
        if (clazzes != null) for (String clazz : clazzes.split(",")) {
            int idx = Integer.parseInt(clazz.trim());
            clazzSet[idx] = true;
        }
    }

    private static double calcVertexDistance(Graph graph, int sourceId, int targetId) {
        final double lat1 = graph.getLats()[sourceId];
        final double lon1 = graph.getLons()[sourceId];
        final double lat2 = graph.getLats()[targetId];
        final double lon2 = graph.getLons()[targetId];
        return calcDistance(lat1, lon1, lat2, lon2);
    }


    // ######################################################################################
    // ########################## IMPLEMENTS GRAPHBUILDINTERCEPTOR ##########################
    // ######################################################################################

    @Override
    public void onOpen(Config config) {
        initClazzSet(config); // Always
        this.segIdToOsmIdBag = new IntLongBag();
        config.putBean(BEANNAME_SEG2OSMID, new VarBean(this.segIdToOsmIdBag));
    }

    @Override
    public void onWay(SegmentedWay way) {
        this.way = way;
        if (this.clazzSet[way.getClazz()]) for (WaySegment ws : way.getSegments())
            segIdToOsmIdBag.add(ws.getId(), way.getId());
    }

    @Override
    public void onEdgeCreate(Edge edge, int segmentIdx, boolean reverse) {
        int wayClazz = this.way.getClazz() & 0xFF;
        int flags = wayClazz;
        edge.setFlags(flags);
    }

    @Override public void onClose(File graphFile) {}
    @Override public void onEdgePosition(int edgeIdx, Edge edge) {}
    @Override public void onVertex(Vertex vertex) {}

    // ######################################################################################
    // ############################## IMPLEMENTS POSTPROCESSOR ##############################
    // ######################################################################################

    @Override
    public void run(Config config, int index) throws Exception {
        File graphFile = config.findGraphFile(true);
        if (null == graphFile) return;
        Log log = config.getLog();

        // Pitfall: This is a different Phase in osm2po and hence a different object instance
        //          However: the config-instance is still the same
        initClazzSet(config);
        segIdToOsmIdBag = (IntLongBag) config.getBean(BEANNAME_SEG2OSMID).get();
        segIdToOsmIdBag.sort(); // enable fast binary search

        // and now ... for the fun part -------->
        // open graph and traverse from some good point to detect needed minor streets
        Graph graph = new Graph(graphFile, log,
                SUPPORT_LATLON | SUPPORT_EDGEFLAGS, false, config.getProps());

        BigLongSet osmIdSet = new BigLongSet(0x10000);

        int[] edgeFlags = graph.getEdgeFlags();
        int[] edgeIds = graph.getEdgeIds();
        int[] edgeSourceIds = graph.getEdgeSourceIds();

        MSDRouter router = new MSDRouter(graph, log, this.clazzSet);

        IntSet segIdSet = new IntSet(edgeFlags.length / 10);

        long vidsAvgSum = 0;
        long vidsAvgN = 0;

        log.info("Starting ManyToMany-Locality analysis for Clazzes "
                + config.getProperty("minorstreet.clazzes"));

        for (int i = 0; i < edgeFlags.length; i++) {

            if (i % 50000 == 0) log.progress(i + " of " + edgeFlags.length
                    + " read. " + segIdSet.getSize()
                    + " found. " + osmIdSet.getSize() + " important");

            int segmentId = toSegmentId(edgeIds[i]);

            if (this.clazzSet[(byte)edgeFlags[i]] && !segIdSet.contains(segmentId)) {
                int[] segIds = router.spanTree(edgeSourceIds[i]);
                for (int segId : segIds) segIdSet.add(segId);

                IntSet vidSet = getEntryVids(graph, this.clazzSet, segIds);
                int nVids = vidSet.getSize();

                if (nVids > 1) {
                    vidsAvgSum += nVids;
                    vidsAvgN++;

                    int[] vids = vidSet.toArray();

                    for (int x = 0; x < vids.length; x++) { // ManyToMany
                        int swap = vids[0]; // Jeder ist einmal Source
                        vids[0] = vids[x];
                        vids[x] = swap;

                        int sourceId = vids[0];
                        int[] targetIds = Arrays.copyOfRange(vids, 1, vids.length);

                        router.traverseAlternativeRoutes(sourceId, targetIds);
                        for (int targetId : targetIds) {
                            if (!router.isVisited(targetId)) {  // no alternative, so make minor street important
                                int[] path = router.findMinorStreetRoute(sourceId, targetId); // find minor street path instead
                                if (path != null) { // May happen in certain situations but ignore it for now
                                    for (int edgeIdx : path) {
                                        byte clazz = (byte) edgeFlags[edgeIdx];
                                        if (this.clazzSet[clazz]) { // Bingo! This Minor Street is important!
                                            int segId = toSegmentId(edgeIds[edgeIdx]);
                                            long osmId = segIdToOsmIdBag.get(segId);
                                            if (IntLongBag.NULLVALUE == osmId) {
                                                log.error("Upps! This must be wrong: SegmentId:" + segId);
                                            } else {
                                                osmIdSet.add(Long.valueOf(osmId));
                                            }
                                        }
                                    }
                                }

                            }
                        }
                    }
                }
            }
        }

        graph.close();
        log.debug("N=" + vidsAvgN + " AVG=" + ((double)vidsAvgSum / vidsAvgN));

        final String beanFilePath = config.getProperty("minorstreet.detection.bean.file",
                config.getWorkDir().getAbsolutePath() + File.separator + BEANFILE);
        final long[] keys = osmIdSet.getKeys();
        Arrays.sort(keys);
        {
            File beanFile = new File(beanFilePath);
            log.info("Writing " + keys.length + " important osm-IDs to internal " + beanFile);
            OutStream os = new OutStreamDisk(beanFile);
            VarBean vb = new VarBean(new VarLongArray(keys));
            vb.writeToStream(os);
            os.close();
        }{
            File beanFileTxt = new File(beanFilePath + ".txt");
            log.info("Writing " + keys.length + " important osm-IDs to readable " + beanFileTxt);
            OutputStream os = new FileOutputStream(beanFileTxt);
            for (long osmId : keys) {
                os.write((String.valueOf(osmId) + '\n').getBytes());
            }
            os.close();
        }
    }

    /**
     * Returns all Vertices of a given list of Segments which
     * are connected to major roads.
     * @param graph {@link Graph}
     * @param clazzSet Array of boolean denoting clazzes of minor streets, service = 51 e.g.
     * @param segIds Id-List of minor street segments. Clazzes must match clazzSet entries
     * @return {@link IntSet} of vertexIds which are connected to the outer world (which are not in clazzSet)
     */
    private IntSet getEntryVids(Graph graph, boolean[] clazzSet, int[] segIds) {
        IntSet vertexIdSet = new IntSet(0x100);
        int[] edgeFlags = graph.getEdgeFlags();
        int[] edgeSourceIds = graph.getEdgeSourceIds();
        int[] edgeTargetIds = graph.getEdgeTargetIds();

        for (int i = 0; i < segIds.length; i++) {
            int segId = segIds[i];
            int edgeId = toEdgeId(segId, false);
            int edgeIdx = graph.getEdgeIdxForId(edgeId);
            byte clazz = (byte) edgeFlags[edgeIdx];
            if (!clazzSet[clazz]) throw new RuntimeException();


            if (i == 0) {
                int outVertexId = edgeSourceIds[edgeIdx];
                for (int outEdgeIdx : graph.findOutgoingEdges(outVertexId)) {
                    byte outClazz = (byte) edgeFlags[outEdgeIdx];
                    if (!clazzSet[outClazz]) {
                        vertexIdSet.add(outVertexId);
                    }
                }
            }

            int outVertexId = edgeTargetIds[edgeIdx];
            for (int outEdgeIdx : graph.findOutgoingEdges(outVertexId)) {
                byte outClazz = (byte) edgeFlags[outEdgeIdx];
                if (!clazzSet[outClazz]) {
                    vertexIdSet.add(outVertexId);
                }
            }
        }

        return vertexIdSet;
    }

    private static class MSDRouter extends VertexRouter {
        private final static int PHASE_SPANTREE = 1;
        private final static int PHASE_ALTERNATIVE_SEARCH = 2;
        private final static int PHASE_IMPORTANT_RESULT = 3;

        private RouterParameters routerParams = new RouterParameters()
                .setFindShortestPath(true)
                .setIgnoreOneWays(false)
                .setIgnoreRestrictions(true);

        private boolean[] clazzSet;
        private int[] edgeFlags;
        private int[] edgeIds;
        private int phase; // see consts above

        int nSegIds = 0;
        int[] segIds = new int[0x10];

        public MSDRouter(Graph graph, Log log, boolean[] clazzSet) {
            super(graph, log);
            this.clazzSet = clazzSet;
            this.edgeFlags = graph.getEdgeFlags();
            this.edgeIds = graph.getEdgeIds();
        }

        synchronized
        final public int[] spanTree(int sourceId) {
            this.phase = PHASE_SPANTREE;
            this.nSegIds = 0;
            this.traverseFully(sourceId, this.routerParams);
            return Arrays.copyOf(this.segIds, this.nSegIds);
        }

        synchronized
        final public void traverseAlternativeRoutes(int sourceId, int[] targetIds) {
            this.phase = PHASE_ALTERNATIVE_SEARCH;
            float maxCost = getMaxDistance(sourceId, targetIds) * 10; // Locality Heuristic
            if (maxCost < 10) maxCost = 10; // Give some minimum
            this.traverse(sourceId, new RouterConstraints(maxCost, targetIds), this.routerParams);
        }

        synchronized
        final public int[] findMinorStreetRoute(int sourceId, int targetId) {
            this.phase = PHASE_IMPORTANT_RESULT;
            return this.findPath(sourceId, targetId, 0, this.routerParams);
        }

        @Override
        protected double calcEdgeCost(int edgeIdx, int fromEdgeIdx) {
            byte clazz = (byte) this.edgeFlags[edgeIdx];

            switch (this.phase) {
            case PHASE_SPANTREE:
                if (!this.clazzSet[clazz]) return -1;

                int segId = toSegmentId(this.edgeIds[edgeIdx]);
                this.segIds[this.nSegIds++] = segId;
                if (this.nSegIds >= this.segIds.length) { // ArrayOverflow
                    this.segIds = Arrays.copyOf(this.segIds, this.nSegIds * 2);
                }
                break;

            case PHASE_ALTERNATIVE_SEARCH:
                if (this.clazzSet[clazz]) return -1;
                break;

            case PHASE_IMPORTANT_RESULT:
                if (!this.clazzSet[clazz]) return -1;
                break;
            }

            return super.calcEdgeCost(edgeIdx, fromEdgeIdx);
        }

        private float getMaxDistance(int sourceId, int[] targetIds) {
            float maxDist = 0;
            for (int targetId : targetIds) {
                float dist = (float) calcVertexDistance(graph, sourceId, targetId);
                if (maxDist < dist) maxDist = dist;
            }
            return maxDist;
        }
    }


    // ######################################################################################
    // ############################## EXTENDS DEFAULTJOINFILTER #############################
    // ######################################################################################


    @Override
    public void open(Config config, String prefix) {
        super.open(config, prefix);
        initClazzSet(config); // Always

        VarBean bean = config.getBean("minorstreets");
        if (bean != null) {
            VarLongArray arr = (VarLongArray) bean.get();
            for (long l : arr.getlongs()) this.idSet.add(l);
        }
    }


    @Override
    public boolean validate(Way way) {
        boolean valid = super.validate(way);
        byte clazz = way.getClazz();
        if (valid && this.clazzSet[clazz]) {
            valid = this.idSet.contains(way.getId());
        }
        return valid;
    }

}
