/*
 * $Id: Props.java 5754 2023-02-21 19:08:03Z carsten $
 *
 * Copyright 2014 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static java.lang.System.currentTimeMillis;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * Erweitert die Standard-Java-{@link Properties} um einen simplen
 * Substitutions-Mechanismus fuer {@link Properties#getProperty(String)}.<br>
 * Zudem koennen Properties in externe Dateien ausgelagert und dort ebenso
 * substituiert werden.<br>
 * Die genaue Verarbeitung wird in {@link #getProperty(String)} beschrieben.
 * <br><br><b>Wichtig:</b><br>
 * Passwoerter sollten entweder ueber die System.Properties oder
 * Environment-Variablen transportiert werden. Diese werden in der
 * {@link #toString()} nicht mit aufgelistet und erscheinen so auch nicht im Log.
 *
 * @see #copyCase(String[])
 *
 * @author (c) 2014 - Carsten Moeller - info@osm2po.de
 */
public class Props extends Properties implements Iterable<String> {
    private static final long serialVersionUID = -6408649888081906013L;

    private boolean substituteFile = false;
    private long lastModified;

    public Props() {
        super();
    }

    public Props(Properties defaults) {
        super(defaults);
    }

    public Props(boolean substitudeFile) {
        super();
        this.substituteFile = substitudeFile;
        this.lastModified = currentTimeMillis();
    }

    public String getRawProperty(String key) {
        return super.getProperty(key);
    }

    public final long getLastModified() {
        return this.lastModified;
    }

    /**
     * Liefert den Wert zu einem Schluessel.
     * @param key String Schluessel NOT NULL.<br>
     * @return Wert als String oder <code>null</code>, wenn nicht gefunden.
     *
     * <br><br><b>Verhalten:</b><br>
     * Das Standardverhalten wird wie folgt erweitert:
     * <ol>
     *      <li>
     *          Der Roh-Wert zu <code>key</code> wird aus den internen Properties gelesen.
     *      </li>
     *      <li>
     *          Wenn <code>null</code>, dann Wert in den System.properties suchen.
     *          <br>Die Property darf hier mit "osm2po." geprefixt werden,
     *          was sogar vorrangig behandelt wird.
     *      </li>
     *      <li>
     *          Wenn <code>null</code>, dann Wert in der unterlegten HashMap suchen
     *          und {@link Object#toString()} zurueckliefern.
     *      </li>
     *      <li>
     *          Wenn <code>null</code>, dann Wert in der systemweiten
     *          Environment mittels {@link System#getenv()} suchen.
     *          <br>Die Property darf hier mit "osm2po." geprefixt werden,
     *          was sogar vorrangig behandelt wird.
     *      </li>
     *      <li>
     *          Wenn <code>null</code>, dann Schluessel als Pfad interpretieren
     *          und den Inhalt der Datei in den Wert als String einlesen.<br>
     *          Pfade muessen dafuer mindestens einen Slash beinhalten, um
     *          als solche akzeptiert zu werden. Also aktuellem Pfad ein "./"
     *          voranstellen!<br>
     *          Dateien werden default nicht substituiert. Dies kann jedoch
     *          mittels {@link #Props(boolean)} Konstruktor erzwungen werden.
     *      </li>
     *      <li>
     *          Wenn immer noch <code>null</code>, dann
     *          jetzt <code>null</code> zurueckliefern.
     *      </li>
     *      <li>
     *          Gefundenen Wert auf Substitutionen der Form <code>[key]</code>
     *          untersuchen und ggf. rekursiv aufloesen. In dieser Phase
     *          muss <code>key</code> einen Wert liefern, ansonsten wird eine
     *          {@link IllegalArgumentException} geworfen. Dies passiert
     *          uebrigens auch bei einem Syntaxfehler.
     *      </li>
     * </ol>
     *
     * <b>Wichtig:</b><br>
     * Zirkulaere Ausdruecke der Form <code>x=[x]</code> werden <b>nicht</b>
     * unterstuetzt und koennen zu Endlosschleifen fuehren.<br>
     * Aus Kompatibliaetsgruenden und fuer eine bessere Lesbarkeit
     * wird weiterhin die alte Notation <code>@[key]</code> unterstuetzt.<br>
     * Diese Klasse implementiert zudem {@link Iterable} und liefert
     * typsicher alle Schluessel als Strings anstatt Objects.<br>
     * Ist eine Substitution nicht gewuenscht, z.b. bei Patterns der
     * Form <code>[x|y]</code>, so ist der Bereich in Hochkommata <strong>"</strong>
     * einzuschliessen. <code>"[x|y]"</code><br>
     * Einzelne Sonderzeichen wie <code>"@[]'&tilde;</code> koennen auch mittels
     * <code>'</code> escaped werden. Dies gilt auch fuer <code>'</code> selbst.
     * Ein Zeilenumbruch kann mit &tilde; eingefuegt werden.
     */
    @Override
    public String getProperty(String key) {
        StringBuilder sb = new StringBuilder(0x100);

        String val = this.getRawProperty(key);
        if (null == val) {
            val = System.getProperty("osm2po." + key);
            if (null == val) {
                val = System.getProperty(key);
                if (null == val) {
                    Object o = this.get(key); // added via put()
                    if (o != null) val = o.toString();
                    if (null == val) {
                        val = System.getenv("osm2po." + key);
                        if (null == val) {
                            val = System.getenv(key);
                            if (null == val) {
                                if (this.containsPathSeparator(key)) {
                                    val = this.loadFileIntoString(key);
                                }
                                if (null == val) return null;
                                if (!this.substituteFile) return val;
                            }
                        }
                    }
                }
            }
        }
        char[] chars = val.toCharArray();
        this.recurseValue(chars, -1, sb);
        return sb.toString();
    }

    @Override
    public String getProperty(String key, String defaultValue) {
        // TODO also substitute defaultValue
        return super.getProperty(key, defaultValue);
    }

    private int recurseValue(char[] chars, int idx, StringBuilder sb) {
        boolean substitute = true;
        while (++idx < chars.length) {
            char c = chars[idx];
            switch (c) {
            case ']':
                if (substitute) return idx;
                else sb.append(c);
                break;
            case '[':
                if (substitute) {
                    StringBuilder sbSub = new StringBuilder();
                    idx = this.recurseValue(chars, idx, sbSub);
                    String subKey = sbSub.toString();
                    String subVal = this.getProperty(subKey);
                    if (null == subVal) throw new IllegalArgumentException(
                            "Property '" + subKey + "' not found or declared");
                    sb.append(subVal);
                } else {
                    sb.append(c);
                }
                break;
            case '@': // Kompatibilitaet zur alten Syntax @[var] - @ ignorieren
                if (idx +1 < chars.length && '[' == chars[idx +1]) continue;
                sb.append(c); // Wenn kein [ folgt, dann @ normales Zeichen
                break;
            case '~': // CR
                sb.append('\n');
                break;
            case '"':
                substitute = !substitute;
                break;
            case '\'': // Escape naechstes Zeichen
                if (idx +1 < chars.length) { // Wenn nicht letztes Zeichen
                    sb.append(chars[++idx]);
                    break;
                }
                // CAUTION-FallThru - no break here - see above
            default:
                sb.append(c);
            }
        }
        return idx;
    }


    private String loadFileIntoString(String filePath) {
        if (null == filePath) return null;
        try {
            File file = new File(filePath);
            if (file.exists() && file.isFile()) {
                long fileSize = file.length();
                if (fileSize > Integer.MAX_VALUE -1) { // +1 = EOF
                    return null; // Datei zu gross
                }
                // Datei ins Memory lesen
                byte[] bytes = new byte[(int) fileSize];
                FileInputStream fis = new FileInputStream(file);
                byte[] buf = new byte[4096];
                for (int pos = 0, len = fis.read(buf); len > -1; pos += len, len = fis.read(buf)) {
                    System.arraycopy(buf, 0, bytes, pos, len);
                }
                fis.close();
                return new String(bytes);
            }
        } catch (Exception ignore) {}
        return null;
    }

    @Override // Unter Java9+ ist dies ein Override, unter 8 noch nicht!.
    public synchronized void putAll(Map<?, ?> t) {
    	// Seit Java9 ? ruft putAll() intern nicht mehr put() sondern final putVal()
    	for (java.util.Map.Entry<?, ?> e : t.entrySet()) {
    		this.put(e.getKey(), e.getValue());
    	}
    }

    @Override
    public synchronized Object put(Object key, Object value) {
        // Alle loads() und setProperty() laufen hier durch!
        // null-value wird in super abgesichert
        Object previousValue = super.put(key, value);
        if (!(key instanceof String)) throw new IllegalArgumentException(
                "key Object '" + key.toString() + "' must be instanceof String");
        this.lastModified = currentTimeMillis();
        return previousValue;
    }

    @Override
    public synchronized void clear() {
        this.lastModified = currentTimeMillis();
        super.clear();
    }

    @Override
    public synchronized Object remove(Object key) {
        this.lastModified = currentTimeMillis();
        return super.remove(key);
    }

    @Override
    public Iterator<String> iterator() {
        return new Iterator<String>() {
            private Enumeration<Object> keys = keys();
            @Override public boolean hasNext() {return this.keys.hasMoreElements();}
            @Override public String next() {return this.keys.nextElement().toString();}
            @Override public void remove() {throw new UnsupportedOperationException();}
        };
    }

    private boolean containsPathSeparator(String path) {
        if (null == path) return false;
        int len = path.length();
        for (int i = 0; i < len; i++) {
            char c = path.charAt(i);
            if (File.separatorChar == c || '/' == c) return true;
        }
        return false;
    }

    /**
     * Kopiert die Properties unter Zuhilfenahme eines simplen Case-Filters.<br>
     * Das Erkennungsmerkmal ist ein fuehrender Punkt im Schluesselnamen.<br>
     * Der Trenner zum eigentlichen Schluessel ist wiederum ein Punkt.
     * Ein <code>.[keyPrefix].[key]=[value]</code>, wird so als
     * <code>[key]=[value]</code> in die neuen Properties geschrieben.<br>
     * Nicht uebereinstimmende Schluessel, die mit einem Punkt beginnen,
     * werden ignoriert. Alle anderen werden ebenfalls kopiert.
     * Zudem werden zuerst die Defaults iteriert und dann erst die moeglichen
     * Filter angewendet, was sicherstellt, dass Defaults auch ueberschrieben
     * werden koennen. Ferner werden dabei Null- oder Leer-Werte ignoriert, was
     * einem Remove entspricht und so interessante Moeglichkeiten eroeffnet.
     * Existiert bereits ein Default und ein <code>case</code>-Key ueberschreibt
     * mit Leer, dann wird die Property ebenfalls entfernt.
     * Prefix kann zudem als Array... uebergeben werden, wodurch mehrere Filter
     * wirken, was einem Case-Fall-Thru wie in z.B. Java oder C entspricht.
     * @param prefixes optional Array of String Filter
     * @return new {@link Props} NOT NULL
     */
    public Props copyCase(String... prefixes) {
        Props props = new Props(this.substituteFile);
        // Defaults ohne Prefix zuerst herausfiltern
        for (String key : this) if (!key.startsWith(".")) {
            String val = this.getRawProperty(key);
            // aber nur, wenn ein Wert gesetzt ist
            if (!(null == val || "".equals(val)))
                props.setProperty(key, val);
        }
        // Alle Prefixe als Filter anwenden
        if (prefixes != null) for (String prefix : prefixes) {
            prefix = "." + prefix + ".";
            int prefixLen = prefix.length();
            for (String key : this) {
                if (key.startsWith(prefix)) {
                    String val = this.getRawProperty(key);
                    if (val != null) {
                        String defKey = key.substring(prefixLen);
                        if ("".equals(val))
                            props.remove(defKey); // delete default
                        else props.setProperty(defKey, val);
                    }
                }
            }
        }

        return props;
    }

    /**
     * Vervielfacht Properties mit Schluesseln der Form <code>a[b|c]</code> zu
     * <code>ab</code> und <code>ac</code>.
     * <br><br><b>Wichtig:</b><br>
     * Geschachtelte Ausdruecke der Form <code>a[b|[c|d]]</code>
     * sind nicht erlaubt.
     * @return new {@link Props} mit aufgeloesten Schluesseln.
     */
    public Props expandKeys() {
        Props props = new Props(this.substituteFile);
        for (String key : this) {
            String val = this.getRawProperty(key);
            List<String> keyList = this.multiplyKey(key);
            for (String xkey : keyList) {
                props.setProperty(xkey, val);
            }
        }
        return props;
    }

    private List<String> multiplyKey(String key) {
        List<String> keyList = new ArrayList<String>();
        List<String> retList = new ArrayList<String>();
        keyList.add(key);
        while (!keyList.isEmpty()) {
            String k = keyList.remove(0);
            int pos1 = k.indexOf("[");
            if (-1 == pos1) {
                retList.add(k);
            } else {
                int pos2 = k.indexOf("]", pos1);
                String before = k.substring(0, pos1);
                String subst = k.substring(pos1 + 1, pos2);
                String after = k.substring(pos2 + 1);
                String[] tokens = subst.split("\\|");
                for (String t : tokens) {
                    k = before + t.trim() + after;
                    keyList.add(k);
                }
            }
        }
        return retList;
    }


    @Override
    public synchronized String toString() {
        final int MAX_VALUE_LEN = 50;
        List<String> lines = new ArrayList<String>(this.size());
        int len = 0;
        for (String k : this) {
            String v = null;
            try {
                v = this.getProperty(k);
            } catch (Throwable t) {
                v = "ERROR[" + t.getMessage() + "]";
            }
            if (v != null) {
                v = v.replaceAll("\n", " ");
                v = v.replaceAll("\r", " ");
                if (v.length() > MAX_VALUE_LEN) {
                    v = v.substring(0, MAX_VALUE_LEN) + "...";
                }
            }

            String line = k + "=" + v;
            len += line.length();
            lines.add(line);
        }

        StringBuilder sb = new StringBuilder(len + lines.size() * 3);
        Collections.sort(lines);
        boolean isNotFirstLine = false;
        for (String line : lines) {
            if (isNotFirstLine) sb.append('\n'); else isNotFirstLine = true;
            sb.append(line);
        }
        return sb.toString();
    }

}
