package de.cm.osm2po.misc;

import java.io.PrintStream;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;

public class ResultSetPrinter {

    private ResultSetPrinter() {/*STATIC*/}

    private static String repeatChars(char c, int n) {
        char[] chars = new char[n];
        Arrays.fill(chars, c);
        return new String(chars);
    }

    private static String formatColumnValue(Object obj, int width) {
        String s = null == obj ? "" : obj.toString();
        int l = s.length();
        if (l < width) s += repeatChars(' ', width - l);
        return s;
    }

    /**
     * Prints out an SQL-Query-Result as formatted table.<br><br>
     * @param stm {@link Statement}
     * @param sql Sql-String
     * @param out {@link PrintStream} e.g {@link System#out}
     * @throws Exception Diverse
     */
    public static void prettyPrint(final Statement stm, final String sql, final PrintStream out) throws Exception {
        PreparedStatement pstm = stm.getConnection().prepareStatement(sql);
        prettyPrint(pstm, out);
    }


    /**
     * Prints out an SQL-Query-Result as formatted table.<br><br>
     * @param stm {@link PreparedStatement}
     * @param out {@link PrintStream} e.g {@link System#out}
     * @throws Exception Diverse
     */
    public static void prettyPrint(final PreparedStatement stm, final PrintStream out) throws Exception {

        final class Column {
            private String name;
            private String type;
            private int width = 3; // Min

            public final String fitWidth(String s) {
                if (null == s) return "";
                int sl = s.length();
                if (sl > width) width = sl;
                return s;
            }
        }

        // First Pass: Get Layout ####################################
        ResultSet rst = stm.executeQuery();
        ResultSetMetaData rstm = rst.getMetaData();
        int nColumns = rstm.getColumnCount();

        Column[] columns = new Column[nColumns +1];
        for (int x = 1; x <= nColumns; x++) {
            Column c = new Column();
            columns[x] = c;
            c.name = c.fitWidth(rstm.getColumnLabel(x));
            c.type = c.fitWidth(rstm.getColumnTypeName(x));
        }

        int numRows = 0;
        while (rst.next()) {
            numRows++;
            for (int x = 1; x <= nColumns; x++) {
                Object v = rst.getObject(x);
                String s = null == v ? "<null>" : v.toString();
                columns[x].fitWidth(s);
            }
        }

        // Second Pass: Get Data ############################################
        rst.close();
        rst = stm.executeQuery();

        int nDigits = (int) Math.log10(numRows) + 1;
        String columnZero = repeatChars(' ', nDigits);
        NumberFormat nf = new DecimalFormat(repeatChars('0', nDigits));

        // Print Header
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue("C" + x, columns[x].width));
        out.println();
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue(columns[x].name, columns[x].width));
        out.println();
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue(columns[x].type, columns[x].width));
        out.println();
        out.print(repeatChars('-', nDigits));
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + repeatChars('-', columns[x].width));
        out.println();

        // Print Data
        numRows = 0;
        while (rst.next()) {
            out.print(nf.format(++numRows));
            for (int x = 1; x <= nColumns; x++) {
                Object v = rst.getObject(x);
                String s = null == v ? "<null>" : v.toString();
                out.print(" | " + formatColumnValue(s, columns[x].width));
            }
            out.println();
        }


        // Footer = Header very useful, if console buffer too small.
        out.print(repeatChars('-', nDigits));
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + repeatChars('-', columns[x].width));
        out.println();
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue("C" + x, columns[x].width));
        out.println();
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue(columns[x].name, columns[x].width));
        out.println();
        out.print(columnZero);
        for (int x = 1; x <= nColumns; x++)  out.print(" | " + formatColumnValue(columns[x].type, columns[x].width));
        out.println();

        rst.close();
    }

}
