/*
 * $Id: StatementExcecutor.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.misc.Utils.TS;
import static java.lang.System.currentTimeMillis;

import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;

import de.cm.osm2po.logging.Log;

/**
 * Ein {@link Statement#execute(String)}-Wrapper, der zusaetzlich
 * die vom DB-Server ausgehenden Meldungen, z.B. aus Stored Procedures,
 * mittels {@link Statement#getWarnings()} loggt.
 * <br><br><b>Wichtig:</b><br>
 * Einmal instanziiert, koennen hier mehrerer SQLs hintereinander
 * ausgefuehrt werden.
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 * @see #execute(String)
 */
public class StatementExcecutor {

    private final Log log;
    private final int logLevel;
    private final Statement statement;
    private Thread swlThread;

    /**
     * Konstruktor.
     * @param statement Auszufuehrendes {@link Statement}
     * @param log {@link Log}
     * @param logLevel int z.B. {@link Log#LEVEL_DEBUG}
     */
    public StatementExcecutor(Statement statement, Log log, int logLevel) {
        this.statement = statement;
        this.log = log;
        this.logLevel = logLevel;
    }

    /**
     * Konstruktor.
     * Ruft {@link #StatementExcecutor(Statement, Log, int)}
     * mit Level {@link Log#getLogLevel()}
     * @param statement Auszufuehrendes {@link Statement}
     * @param log {@link Log}
     */
    public StatementExcecutor(Statement statement, Log log) {
        this(statement, log, log.getLogLevel());
    }

    /**
     * Startet den {@link Thread} fuer {@link Statement#getWarnings()}
     * und ruft {@link Statement#execute(String)}.
     * @param sql String Auszufuehrendes SQL
     * @return Rueckgabe von {@link Statement#execute(String)} oder false, wenn Exception
     */
    public final synchronized boolean execute(final String sql) {
        boolean execOk = false;
        try {
            log.log("DB " + TS() + " " + sql, this.logLevel, null);
            this.statement.clearWarnings();
            this.swlThread = new Thread(new StatementWarningsListener());
            this.swlThread.start(); // Achtung! Thread.setName() macht Stress!
            execOk = this.statement.execute(sql);

        } catch (Exception e) {
            log.error(e);
        } finally {try {
            this.swlThread.interrupt();
            this.swlThread.join();
        } catch (Exception e) {log.error(e);}}
        return execOk;
    }

    private SQLWarning logNexts(SQLWarning lastW) {
        try {
            SQLWarning nextW = (lastW == null) ?
                    this.statement.getWarnings() : lastW.getNextWarning();
            while (nextW != null) {
                lastW = nextW;
                this.log.log("DB " + TS() + " " + lastW.getMessage(), this.logLevel, null);
                nextW = lastW.getNextWarning();
            }
        } catch (SQLException e) {
            log.error(e);
        }
        return lastW;
    }

    private class StatementWarningsListener implements Runnable {
        @Override public void run() {
            SQLWarning lastWarning = null;
            try {
                while (true) {
                    lastWarning = logNexts(lastWarning);
                    Thread.sleep(1000);
                    log.progress("Tick " + (currentTimeMillis() / 1000));
                }
            } catch (InterruptedException ie) {
                logNexts(lastWarning);
            } catch (Throwable t) {
                log.error(t);
            }
        }
    }

}
