/*
 * $Id: StringBuilderUtf8.java 5133 2020-10-20 17:06:03Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.misc.Utils.UTF8;

import java.nio.charset.Charset;
import java.util.Arrays;

/**
 * Ein optimierter Builder fuer UTF-8 kodierte Strings.
 * Vermeidet ueberfluessiges Hin- und Herkonvertieren zwischen diversen
 * {@link Charset}s und liefert das Ergebnis direkt analog
 * {@link String#getBytes(Charset)} fuer UTF-8.
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 */
public class StringBuilderUtf8 {
    private final static byte[] JSON_NULL_BYTES = "null".getBytes();

    private byte[] bytes;
    private int len;

    public StringBuilderUtf8() {
        this(0x400);
    }

    public StringBuilderUtf8(int initialCapacity) {
        this.bytes = new byte[initialCapacity];
    }

    public final byte[] getBytes() {
        if (this.bytes.length != this.len) // lazy
            this.bytes = Arrays.copyOf(this.bytes, this.len);
        return this.bytes;
    }

    public final void reset() {this.len = 0;}

    public final StringBuilderUtf8 append(byte b) {
        if (this.bytes.length == this.len) {
            this.bytes = Arrays.copyOf(this.bytes, this.len * 2);
        }
        this.bytes[len++] = b;
        return this;
    }

    public final StringBuilderUtf8 appendJson(char c, boolean uCoded) {
        switch (c) {
        case '\"': this.append((byte)'\\').append((byte)'\"'); break;
        case '\\': this.append((byte)'\\').append((byte)'\\'); break;
        case '\n': this.append((byte)'\\').append((byte)'n'); break;
        case '\r': this.append((byte)'\\').append((byte)'r'); break;
        case '\t': this.append((byte)'\\').append((byte)'t'); break;
        default:
            if (uCoded && c > 127) {
                this.append((byte)'\\').append((byte)'u');
                String hex = Integer.toHexString(c);
                int xlen = hex.length();
                for (int x = xlen; x < 4; x++) this.append('0');
                for (int x = 0; x < xlen; x++) this.append(hex.charAt(x));
            } else {
                this.append(c); // Standard-UTF8-Kodierung
            }
        }
        return this;
    }

    public final StringBuilderUtf8 append(char c) {
        int ichar = c; // 16 Bits, immer positiv, also von 0 bis 64k

        if (ichar <= 127) { // entspricht Ascii
            this.append((byte)c); // 0xxxxxxx

        } else { // UTF-8 nutzt maximal 4 Bytes fuer 21 Bits, somit 3 mehr als char
            int n = 0; //see https://en.wikipedia.org/wiki/UTF-8
            int[] ibytes = new int[8]; // maximal 4 bytes, 8 ist aber sicher
            while (ichar != 0) {
                int b = (byte) ((ichar & 0x3F) | 0x80);
                ibytes[n++] = b; // 10xxxxxx
                ichar >>>= 6;
            }

            // Hoechstes byte kodiert Anzahl der (Folge-)Bytes.
            // Bit 0x80 fuer >1 oben bereits gesetzt, daher:
            switch (--n) {
            case 1: ibytes[n] |= 0x40; break; // 2 Bytes: 110xxxxx
            case 2: ibytes[n] |= 0x60; break; // 3 Bytes: 1110xxxx
            case 3: ibytes[n] |= 0x70; break; // 4 Bytes: 11110xxx (ausserhalb char)
            }

            while (n >= 0) append((byte) ibytes[n--]);
        }
        return this;
    }

    public final StringBuilderUtf8 append(String s) {
        int len = null == s ? 0 : s.length();
        for (int i = 0; i < len; i++) append(s.charAt(i));
        return this;
    }

    public final StringBuilderUtf8 appendJson(String s, boolean uCoded) {
        int n = null == s ? 0 : s.length();
        for (int i = 0; i < n; i++) appendJson(s.charAt(i), uCoded);
        return this;
    }

    @Override
    public String toString() {
        return new String(this.getBytes(), UTF8);
    }

    public static byte[] toJson(String s, boolean uCoded) {
        if (null == s) {
            return Arrays.copyOf(JSON_NULL_BYTES, JSON_NULL_BYTES.length);
        } else {
            int n = s.length();
            StringBuilderUtf8 sb8 = new StringBuilderUtf8(n *2);
            sb8.append('\"').appendJson(s, uCoded).append('\"');
            return Arrays.copyOf(sb8.bytes, sb8.len);
        }
    }

}
