/*
 * $Id: WkbUtils.java 4351 2017-02-04 17:31:45Z carsten $
 *
 * Copyright 2010/2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;

import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;

/**
 * WKB-Converter.
 * @author (c) 2012/2015 - Carsten Moeller - info@osm2po
 * @see <a href="http://edndoc.esri.com/arcsde/9.1/general_topics/wkb_representation.htm">ESRI</a>
 * @see <a href="http://gis.stackexchange.com/questions/95585/is-it-possible-to-get-a-binary-geometry-from-postgis">::bytea Encoding-Problem</a>
 */
public class WkbUtils {
    private WkbUtils() {/*STATIC*/}
    
    /* ***********************************************
     * Konvertierung von LatLons nach WKB
     * ***********************************************/

    public static byte[] toPointWkb(LatLon latLon) {
        int size = 25; // endian[1] + type[4] + srid[4] + lon[8] + lat[8]
        byte[] buf = new byte[size];
        int[] idx = {0};
        writeByte(buf, (byte) 1, idx); // Endian NDR, nicht XDR
        int type = 1; // Point
        type |= 0x20000000; // srid != -1 , wenn nicht dann length -4
        writeInt(buf, type, idx);
        writeInt(buf, 4326, idx); // srid = 4326, wenn -1, dann weglassen und length -4. s.o.
        writeDouble(buf, latLon.getLon(), idx);
        writeDouble(buf, latLon.getLat(), idx);
        return buf;
    }

    public static byte[] toLineWkb(LatLons latLons) {
        int nPoints = latLons.getSize();
        int size = 13; // endian[1] + type[4] + srid[4] + ndCount[4]
        size += nPoints * 16; // 16 = lon[8] + lat[8]
        byte[] buf = new byte[size];
        int[] idx = {0}; // int als tricky Referenz

        writeByte(buf, (byte) 1, idx); // Endian NDR, nicht XDR
        int type = 2; // LineString
        type |= 0x20000000; // srid != -1 , wenn nicht dann length -4
        writeInt(buf, type, idx);
        writeInt(buf, 4326, idx); // srid = 4326, wenn -1, dann weglassen und length -4. s.o.
        writeInt(buf, nPoints, idx);

        for (int i = 0; i < nPoints; i++) {
            long coord = latLons.getCoord(i);
            writeDouble(buf, lonOf(coord), idx);
            writeDouble(buf, latOf(coord), idx);
        }
        return buf;
    }
    
    public static byte[] toMultiLineWkb(LatLons[] latLons) {
        int nLines = latLons.length;
        int size = 13; // endian[1] + type[4] + srid[4] + nLines[4]

        byte[][] lines = new byte[nLines][];
        for (int i = 0; i < nLines; i++) { 
            lines[i] = toLineWkb(latLons[i]);
            size += lines[i].length;
        }
        
        byte[] buf = new byte[size];
        int[] idx = {0}; // int als tricky Referenz

        // Multiline Header
        writeByte(buf, (byte) 1, idx); // Endian NDR, nicht XDR
        int type = 5; // MultilineString
        type |= 0x20000000; // srid != -1 , wenn nicht dann length -4
        writeInt(buf, type, idx);
        writeInt(buf, 4326, idx); // srid = 4326, wenn -1, dann weglassen und length -4. s.o.
        writeInt(buf, nLines, idx); // Anzahl Lines
        
        for (int i = 0; i < nLines; i++) {
            int nBytes = lines[i].length;
            System.arraycopy(lines[i], 0, buf, idx[0], nBytes);
            idx[0] += nBytes;
        }
        
        return buf;
    }
    
    public static byte[] toPolygonWkb(LatLons[] latLons) {
        int nRings = latLons.length;
        int nNodes = 0;
        for (int i = 0; i < nRings; i++) nNodes += latLons[i].getSize();

        int size = 13; // endian[1] + type[4] + srid[4] + nRings[4]
        size += nNodes * 16 + nRings * 4; // 16 = lon + lat, 4 = Integer nNodes/Ring

        byte[] buf = new byte[size];
        int[] idx = {0}; // int als tricky Referenz
        
        writeByte(buf, (byte) 1, idx); // Endian NDR, nicht XDR
        int type = 3; // Polygon
        type |= 0x20000000; // srid != -1 , wenn nicht dann length -4
        writeInt(buf, type, idx);
        writeInt(buf, 4326, idx); // srid = 4326, wenn -1, dann weglassen und length -4. s.o.
        
        writeInt(buf, nRings, idx);
        for (int i = 0; i < nRings; i++) {
            long[] coords = latLons[i].getCoords();
            int n = coords.length;
            writeInt(buf, n, idx);
            for (int j = 0; j < n; j++) {
                writeDouble(buf, lonOf(coords[j]), idx);
                writeDouble(buf, latOf(coords[j]), idx);
            }
        }

        return buf;
    }

    public static byte[] toMultiPolygonWkb(LatLons[][] latLons) {
        int nPolygons = latLons.length;
        int size = 13; // endian[1] + type[4] + srid[4] + nPolygons[4]

        byte[][] polygons = new byte[nPolygons][];
        for (int i = 0; i < nPolygons; i++) { 
            polygons[i] = toPolygonWkb(latLons[i]);
            size += polygons[i].length;
        }
        
        byte[] buf = new byte[size];
        int[] idx = {0}; // int als tricky Referenz

        // Multipolygon Header
        writeByte(buf, (byte) 1, idx); // Endian NDR, nicht XDR
        int type = 6; // Multipolygon
        type |= 0x20000000; // srid != -1 , wenn nicht dann length -4
        writeInt(buf, type, idx);
        writeInt(buf, 4326, idx); // srid = 4326, wenn -1, dann weglassen und length -4. s.o.
        writeInt(buf, nPolygons, idx); // Anzahl Polygons
        
        for (int i = 0; i < nPolygons; i++) {
            int nBytes = polygons[i].length;
            System.arraycopy(polygons[i], 0, buf, idx[0], nBytes);
            idx[0] += nBytes;
        }
        
        return buf;
    }
    
    /* ***********************************************
     * Konvertierung von WKB AutoDetect/Object
     * ***********************************************/
    
    public static Object fromWkb(byte[] geomBytes) {
        int idx[] = {0};
        int type = readWkbHeader(geomBytes, idx);
        switch (type) {
        case 1:
            return readWkbPoint(geomBytes, idx);
        case 2:
            return readWkbLine(geomBytes, idx);
        case 3:
            return readWkbPolygon(geomBytes, idx);
        case 5:
            return readWkbMultiLine(geomBytes, idx);
        case 6:
            return readWkbMultiPolygon(geomBytes, idx);
        default :
            throw new RuntimeException("Unsupported type " + type);    
        }
    }

    /* ***********************************************
     * Konvertierung von WKB nach LatLon
     * ***********************************************/

    public static LatLon fromWkbPoint(byte[] geomBytes) {
        int[] idx = {0};
        int type = readWkbHeader(geomBytes, idx);
        if (type != 1) throw new RuntimeException(
                "Wrong Type. POINT expected");
        return readWkbPoint(geomBytes, idx);
    }

    public static LatLons fromWkbLine(byte[] geomBytes) {
        int[] idx = {0};
        int type = readWkbHeader(geomBytes, idx);
        if (type != 2) throw new RuntimeException(
                "Wrong Type. LINESTRING expected");
        return readWkbLine(geomBytes, idx);
    }
    
    public static LatLons[] fromWkbMultiLine(byte[] geomBytes) {
        int[] idx = {0};
        int type = readWkbHeader(geomBytes, idx);
        if (type != 5) throw new RuntimeException(
                "Wrong Type. MULTILINESTRING expected");
        return readWkbMultiLine(geomBytes, idx);
    }
    
    public static LatLons[] fromWkbPolygon(byte[] geomBytes) {
        int[] idx = {0};
        int type = readWkbHeader(geomBytes, idx);
        if (type != 3) throw new RuntimeException(
                "Wrong Type. POLYGON expected");
        return readWkbPolygon(geomBytes, idx);
    }

    public static LatLons[][] fromWkbMultiPolygon(byte[] geomBytes) {
        int[] idx = {0};
        int type = readWkbHeader(geomBytes, idx);
        if (type != 6) throw new RuntimeException(
                "Wrong Type. MULTIPOLYGON expected");
        return readWkbMultiPolygon(geomBytes, idx);
    }

    /* ********************************************************
     * Komplexe WKB - BlockReader (Header und SubGeoms getrennt)
     * ********************************************************/

    private static int readWkbHeader(byte[] geomBytes, int[] idx) {
        byte endian = readByte(geomBytes, idx); // NDR=1, XDR=0
        if (endian != 1) throw new RuntimeException(
                "Endian XDR not supported");
        int type = readInt(geomBytes, idx);
        boolean hasSrid = (type & 0x20000000) != 0;
        type &= 0xFF; // Feature-Type im niedrigsten Byte.
        int srid = -1; // Spatial Reference System (Projektion)
        if (hasSrid) {
            srid = readInt(geomBytes, idx);
            if (srid != 4326) throw new RuntimeException(
                    "SRID " + srid + " not supported");
        }
        return type;
    }
    
    private static LatLon readWkbPoint(byte[] geomBytes, int[] idx) {
        double lon = readDouble(geomBytes, idx);
        double lat = readDouble(geomBytes, idx);
        return new LatLon(lat, lon);
    }

    private static LatLons readWkbLine(byte[] geomBytes, int[] idx) {
        int nPoints = readInt(geomBytes, idx);
        LatLons latLons = new LatLons();
        for (int i = 0; i < nPoints; i++) {
            latLons.append(readWkbPoint(geomBytes, idx));
        }
        return latLons;	
    }

    private static LatLons[] readWkbMultiLine(byte[] geomBytes, int[] idx) {
        int nLines = readInt(geomBytes, idx);
        LatLons[] ways = new LatLons[nLines];
        for (int i = 0; i < nLines; i++) {
            int type = readWkbHeader(geomBytes, idx);
            if (type != 2) throw new RuntimeException(
                    "Wrong inner Type. LINESTRING expected");
            ways[i] = readWkbLine(geomBytes, idx);
        }
        return ways;
    }

    private static LatLons[][] readWkbMultiPolygon(byte[] geomBytes, int[] idx) {
        int nPolygons = readInt(geomBytes, idx);
        LatLons[][] multiPolygon = new LatLons[nPolygons][];
        for (int i = 0; i < nPolygons; i++) {
            int type = readWkbHeader(geomBytes, idx);
            if (type != 3) throw new RuntimeException(
                    "Wrong inner Type. POLYGON expected");
            multiPolygon[i] = readWkbPolygon(geomBytes, idx);
        }
        return multiPolygon;
    }

    private static LatLons[] readWkbPolygon(byte[] geomBytes, int[] idx) {
        int nRings = readInt(geomBytes, idx);
        LatLons[] poly = new LatLons[nRings];
        for (int i = 0; i < nRings; i++) {
        	poly[i] = new LatLons(); 
        	int nCoords = readInt(geomBytes, idx);
        	for (int j = 0; j < nCoords; j++) {
        		LatLon point = readWkbPoint(geomBytes, idx);
        		poly[i].append(point);
        	}
        }
        return poly;
    }

    /*
     * LowLevel ByteArray NDR (little endian) Read/Write Manipulation.
     * Little endian schreibt bytes in umgekehrter Reihenfolge.
     * Das heisst, das niederwertigste Byte landet an der
     * niedrigsten Array-Position[0]. Standard Java ist XDR (big endian).
     * Tricky:
     * Wir benoetigen einen Index (int) als Referenz.
     * Hierzu missbrauchen wir ein int[0] (Array mit einem Element).
     */
    
    private static void writeByte(byte[] buf, byte b, int[] idx) {
        buf[idx[0]++] = b;
    }
    
    private static void writeInt(byte[] buf, int i, int[] idx) {
        writeByte(buf, (byte) i, idx); 
        writeByte(buf, (byte) (i >>> 8), idx);
        writeByte(buf, (byte) (i >>> 16), idx);
        writeByte(buf, (byte) (i >>> 24), idx);
    }
    
    private static void writeLong(byte[] buf, long l, int[] idx) {
        writeByte(buf, (byte) l, idx); 
        writeByte(buf, (byte) (l >>> 8), idx);
        writeByte(buf, (byte) (l >>> 16), idx);
        writeByte(buf, (byte) (l >>> 24), idx);
        writeByte(buf, (byte) (l >>> 32), idx);
        writeByte(buf, (byte) (l >>> 40), idx);
        writeByte(buf, (byte) (l >>> 48), idx);
        writeByte(buf, (byte) (l >>> 56), idx);
    }
    
    private static void writeDouble(byte[] buf, double d, int[] idx) {
        long longBits = Double.doubleToLongBits(d);
        writeLong(buf, longBits, idx);
    }

    private static byte readByte(byte[] buf, int[] idx) {
        return buf[idx[0]++];
    }
    
    private static int readInt(byte[] buf, int idx[]) {
        int i = 0;
        i += (readByte(buf, idx) & 0xFF) << 0;
        i += (readByte(buf, idx) & 0xFF) << 8;
        i += (readByte(buf, idx) & 0xFF) << 16;
        i += (readByte(buf, idx) & 0xFF) << 24;
        return i;
    }


    private static long readLong(byte[] buf, int[] idx) {
        long l = 0;
        l += (readByte(buf, idx) & 0xFFL) << 0;
        l += (readByte(buf, idx) & 0xFFL) << 8;
        l += (readByte(buf, idx) & 0xFFL) << 16;
        l += (readByte(buf, idx) & 0xFFL) << 24;
        l += (readByte(buf, idx) & 0xFFL) << 32;
        l += (readByte(buf, idx) & 0xFFL) << 40;
        l += (readByte(buf, idx) & 0xFFL) << 48;
        l += (readByte(buf, idx) & 0xFFL) << 56;
        return l;
    }
    
    private static double readDouble(byte[] buf, int[] idx) {
        long l = readLong(buf, idx);
        return Double.longBitsToDouble(l);
    }

}
