/*
 * $Id: WktUtils.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2010/2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import java.util.regex.PatternSyntaxException;

import de.cm.osm2po.model.LatLon;

/**
 * WKT-Converter.
 * <br><br><b>Wichtig:</b><br>
 * Aktuell wird nur POINT in 2D unterstuetzt.
 * @author (c) 2021 - Carsten Moeller - info@osm2po
 */
public class WktUtils {
    private WktUtils() {/*STATIC*/}

    private static char parseDelim(char[] chars, int[] index) {
        char delim = 0; // char 0 = nix gefunden
        while (index[0] < chars.length) {
            char c = chars[index[0]];
            if (c == ' ' || c == ',' || c == ')' || c == '(') {
                index[0]++;
                delim = c;
                if (c == ' ') continue; else break;
            } else break;
        }
        return delim;
    }

    private static double parseDouble(char[] chars, int[] index) {
        char delim = parseDelim(chars, index);
        if (!(delim == ' ' || delim == 0)) throw new PatternSyntaxException(
                "Double expected", new String(chars), index[0]);

        StringBuilder sb = new StringBuilder();
        while (index[0] < chars.length) {
            char c = chars[index[0]];
            if ((c >= '0' && c <= '9') || c == '-' || c == '+' || c == '.' ) sb.append(c); else break;
            index[0]++;
        }

        try {
            return Double.parseDouble(sb.toString());
        } catch (Exception e) {
            throw new PatternSyntaxException(
                    "Double expected", new String(chars), index[0] - sb.length());
        }
    }

    private static LatLon parseLatLon(char[] chars, int[] index) {
        double lon = parseDouble(chars, index);
        if (parseDelim(chars, index) != ' ') throw new PatternSyntaxException(
                "WhiteSpace expected", new String(chars), index[0]);
        double lat = parseDouble(chars, index);
        return new LatLon(lat, lon);
    }

    private static LatLon parsePoint(char[] chars, int[] index) {

        if (parseDelim(chars, index) != '(') throw new PatternSyntaxException(
                "( expected", new String(chars), index[0]);

        LatLon latLon = parseLatLon(chars, index);

        if (parseDelim(chars, index) != ')') throw new PatternSyntaxException(
                ") expected", new String(chars), index[0]);

        return latLon;
    }


    private static String parseGeometryType(char[] chars, int[] index) {
        StringBuilder sb = new StringBuilder();
        char delim = parseDelim(chars, index);
        if (!(delim == ' ' || delim == 0)) throw new PatternSyntaxException(
                "GeometryName expected", new String(chars), index[0]);

        while (index[0] < chars.length) {
            char c = chars[index[0]];
            if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')) sb.append(c); else break;
            index[0]++;
        }
        return sb.toString();
    }

    public static Object parseWkt(String wkt) {
        char[] chars = wkt.toCharArray();
        int[] index = new int[]{0};
        String type = parseGeometryType(chars, index).toUpperCase();
        switch (type) {
        case "POINT": return parsePoint(chars, index);
        default: throw new PatternSyntaxException(
                "GeometryName not supported", new String(chars), index[0]);
        }
    }


//    public static void main(String[] args) {
//        System.out.println(parseWkt(" point (-44.5 +66.8 )"));
//    }
}
