/*
 * $Id: ZoneUtils.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.misc;

import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static java.lang.Double.POSITIVE_INFINITY;
import static java.lang.StrictMath.abs;
import static java.lang.StrictMath.signum;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;

/**
 * Eine Zone bezeichnet eine Zelle eines quadratischen Rasters
 * auf den Erd-Koordinaten.<br>
 * Eine Zone wird durch eine Integer-Id kodiert und erlaubt
 * so Groessen von 1 - 0.01 Grad.
 * @see #createZoneId(double, double, int)
 * @author (c) 2016 - Carsten Moeller, info@osm2po.de
 */
public class ZoneUtils {

    private ZoneUtils() {/*static*/}

    // Kann auch berechnet werden. Lookups sind aber schneller.
    private final static double[] E10Scale = new double[]{1, 10, 100};
    private final static int[] E10LatMod = new int[]{180, 1800, 18000};
    private final static int[] E10LonMod = new int[]{360, 3600, 36000};

    /**
     * Liefert die erlaubten Skalierungen fuer einen Exponenten e.
     * @param e int 10er-Exponent. Erlaubte Werte sind 0, 1, 2
     * @return Skalierung, z.B. 100 fuer e=2 (10^2)
     */
    public final static double scaleForE(int e) {
        return E10Scale[e];
    }

    /**
     * Rastert eine Koordinate auf ein Grad-Raster.<br>
     * Die erste Raster-Zelle mit der Id 0 liegt links unten
     * (Lat -90 Lon -180). Hoehe und Breite betraegt 1/10^E (siehe Parameter),
     * so dass z.B. bei einer Zellengroesse von 0.1 die Koordinaten
     * 0.0 bis 0.0999.. noch zur Zelle gehoeren, 0.1 aber bereits zur naechsten,
     * egal ob in X- oder Y-Richtung. Uebergaenge von + nach - Graden werden per
     * Modulo beruecksichtigt, so dass Lon -180 = Lon 180 Grad ist.
     * Die maximale Id bei 1e-2 betraegt 1.179.618.463.
     * @param lat double
     * @param lon double
     * @param e int z.B. e=1 fuer 1/10^1 = 0.1. Erlaubte Werte fuer e sind 0,1,2.
     * @return int id der Raster-Zone.
     * @see ZoneUtils#calcZoneAnchor(int, int)
     */
    public final static int createZoneId(double lat, double lon, int e) {
        final double scale = E10Scale[e];
        final double infi = 1E-9; // Boese Floating-Point Korrektur
        int ilat = (int)(((lat + 270d + infi) % 180d) * scale);
        int ilon = (int)(((lon + 540d + infi) % 360d) * scale);
        return ilat << 16 | ilon;
    }

    /**
     * Liefert die Anker-Koordinate (links unten) einer Zone.<br>
     * Dies ist faktisch die Umkehrfunktion von
     * {@link #createZoneId(double, double, int)}
     * @param zoneId int Id der Zone
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return {@link LatLon} Ankerpunkt
     * @see #createZoneId(double, double, int)
     */
    public final static LatLon calcZoneAnchor(int zoneId, int e) {
        final double scale = E10Scale[e];
        double lat = (zoneId >>> 16) / scale;
        double lon = (zoneId & 0xFFFF) / scale;
        return new LatLon(lat - 90, lon - 180);
    }

    /**
     * Erweitert eine Zone um weitere Zonen im Umkreis quadratisch.<br>
     * Beruecksichtigt auch die Uebergaenge von + nach - Graden.
     * <p>
     * Beispiel: Z mit R(adius)=2 liefert 25 Zellen(-Ids).
     * <code>
     * <br>R R R R R
     * <br>R R R R R
     * <br>R R(Z)R R
     * <br>R R R R R
     * <br>R R R R R
     * </code>
     * </p>
     * @param zoneId int
     * @param r Anzahl der erweiterten Zonen - quadratisch.
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return int[]-Array aller zoneIds.
     */
    public final static int[] expandZone(int zoneId, int r, int e) {
        //int[] zoneArr = new int[4 * r * r + 4 * r + 1];
        int[] zoneArr = new int[(r*2 +1)*(r*2 +1)];

        int latMod = E10LatMod[e];
        int lonMod = E10LonMod[e];

        int dy = (zoneId >>> 16) + latMod;
        int dx = (zoneId & 0xFFFF) + lonMod;

        int n = 0;
        for (int y = dy - r; y <= dy + r; y++) {
            for (int x = dx - r; x <= dx + r; x++) {
                int zy = y % latMod;
                int zx = x % lonMod;
                int z = zy << 16 | zx;
                zoneArr[n++] = z;
            }
        }
        return zoneArr;
    }

    /**
     * Liefert die ungefaehre Rechteck-Form einer Zone.<br>
     * @param zoneId int Id der Zone
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return {@link LatLons} Koordinaten des Rechtecks (Polygon)
     */
    public final static LatLons calcZoneRect(int zoneId, int e) {
        final double wh = 1/E10Scale[e];
        LatLon a = calcZoneAnchor(zoneId, e);
        LatLon b = new LatLon(
                a.getLat() + wh - 1E-9,
                a.getLon() + wh - 1E-9);
        return new LatLons()
            .append(a.getLat(), a.getLon())
            .append(a.getLat(), b.getLon())
            .append(b.getLat(), b.getLon())
            .append(b.getLat(), a.getLon())
            .append(a.getLat(), a.getLon());
    }

    /**
     * Rastert eine rechteckige Box auf ein Raster.
     * Die Parameter werden vorher normalisiert, so dass z.B.
     * Lon -180 = Lon +180 ist. Dennoch wird geprueft, ob Lon1
     * links von Lon2 liegt - analog Lat - hier also Vorsicht!
     * @param lat1 Breitengrad unten
     * @param lon1 Laengengrad links
     * @param lat2 Breitengrad oben
     * @param lon2 Laengengrad rechts
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return int[] Array aller durch Box abgedeckten ZoneIds.
     * @throws IllegalArgumentException wenn keine Box
     */
    public final static int[] createZoneIdsBox(
            double lat1, double lon1, double lat2, double lon2, int e) {

        final double scale = E10Scale[e];
        final int modLat = E10LatMod[e];
        final int modLon = E10LonMod[e];

        int ilat1 = (int)((lat1 + 270d) * scale);
        int ilon1 = (int)((lon1 + 540d) * scale);
        int ilat2 = (int)((lat2 + 270d) * scale);
        int ilon2 = (int)((lon2 + 540d) * scale);

        if (lat2 < lat1 || lon2 < lon1) // = ist erlaubt, da selbe Zone moeglich
            throw new IllegalArgumentException("No BoundingBox");

        int n = (ilat2 - ilat1 +1) * (ilon2 - ilon1 +1);
        int[] zoneIds = new int[n];
        n = 0;
        for (int ilat = ilat1; ilat <= ilat2; ilat++)
            for (int ilon = ilon1; ilon <= ilon2; ilon++) {
                zoneIds[n++] = (ilat % modLat) << 16 | (ilon % modLon);

            }

        if (n != zoneIds.length) throw new RuntimeException(
                "Something went terribly wrong here");
        return zoneIds;
    }

    /**
     * Rastert eine rechteckige Box auf ein 1/10^e-Grad Raster.
     * Ermittelt hierzu die BoundingBox der uebergebenen Koordinaten.
     * Diese muessen zwingend im Bereich Lat -90 bis +89.999...
     * und Lon -180 bis -179.999... liegen.
     * @param coords long[] Koordinaten lat/lon als 1e-7-coded.
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return int[] Array von ermittelten ZoneIds oder null.
     */
    public final static int[] createZoneIdsBox(long[] coords, int e) {
        if (null == coords || coords.length < 2) return null;
        double lat1 = 90;
        double lon1 = 180;
        double lat2 = -90;
        double lon2 = -180;
        for (long coord : coords) {
            double lat = latOf(coord);
            double lon = lonOf(coord);
            if (lat < lat1) lat1 = lat;
            if (lon < lon1) lon1 = lon;
            if (lat > lat2) lat2 = lat;
            if (lon > lon2) lon2 = lon;
        }
        return createZoneIdsBox(lat1, lon1, lat2, lon2, e);
    }

    /**
     * Liefert alle Ids von Zellen, die von einer Strecke durchkreuzt werden.
     * Wird auch als Raytrace bzw. Supercover bezeichnet und ist eine Abwandlung
     * des Bresenham-Algorithmus, der allerdings Pixel als Oktagone behandelt
     * und so die Ecken der Zellen-Quadrate, wie hier verwendet, ignoriert.<br>
     * Beruecksichtigt zudem Uebergaenge von Plus- nach Minus-Graden, indem der
     * Berechnungsraum fuer Lat um 270 und Lon um 540 Grad verschoben wird.
     * @param lat1 double Startpunkt y
     * @param lon1 double Startpunkt x
     * @param lat2 double Zielpunkt y
     * @param lon2 double Zielpunkt x
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return int[] Array von ZoneIds
     */
    public final static int[] raytraceZones(
            double lat1, double lon1, double lat2, double lon2, int e) {
        // see http://playtechs.blogspot.de/2007/03/raytracing-on-grid.html

        // Scale to positive coords, so we can avoid the floor()-Method if negative
        // e.g: Wrong: (int)-5.1 = -5 Correct: (int)(floor(-5.1)) = -6
        //      but (int)5.1 = 5 is correct and equal to floor()-overhead.
        // Also take the sphere into account where Longitude -180 = +180:
        // So shift to big enough positive range where Lat -90 Mod 180 = 0
        // and Lon -180 Mod 360 = 0 respectively
        lat1 += 270d; lat2 += 270d;
        lon1 += 540d; lon2 += 540d;
        final int modLat = E10LatMod[e];
        final int modLon = E10LonMod[e];

        final double scale = E10Scale[e];
        lat1 *= scale; lon1 *= scale; // Scale the system to 1-eN
        lat2 *= scale; lon2 *= scale;

        double dxAbs = abs(lon2 - lon1);
        double dyAbs = abs(lat2 - lat1);

        int x = (int)lon1; // remove fractional part
        int y = (int)lat1;

        int n = 1;
        int incX, incY;
        double error;

        if (dxAbs == 0) { // Vertical
            incX = 0;
            error = POSITIVE_INFINITY;
        } else if (lon2 > lon1) {
            incX = 1;
            n += ((int)lon2) - x;
            error = (((int)lon1) + 1d - lon1) * dyAbs;
        } else {
            incX = -1;
            n += x - ((int)lon2);
            error = (lon1 - ((int)lon1)) * dyAbs;
        }

        if (dyAbs == 0) { // Horizontal
            incY = 0;
            error -= POSITIVE_INFINITY;
        } else if (lat2 > lat1) {
            incY = 1;
            n += ((int)lat2) - y;
            error -= (((int)lat1) + 1d - lat1) * dxAbs;
        } else {
            incY = -1;
            n += y -((int)lat2);
            error -= (lat1 - ((int)lat1)) * dxAbs;
        }

        int zoneIds[] = new int[n]; // Zone is my CellAnchorPoint[y,x]
        for (; n > 0; --n) {
            int zoneId = (y % modLat) << 16 | (x % modLon);
            zoneIds[zoneIds.length - n] = zoneId;

            // Vertical and Horizontal set via INFINITY. see above.
            if (error > 0) {
                y += incY;
                error -= dxAbs;
            } else {
                x += incX;
                error += dyAbs;
            }
        }

        return zoneIds;
    }

    /**
     * Berechnet den quadratischen Einheitsvektor und dreht diesen orthogonal
     * um 90 Grad gegen den Uhrzeigersinn.<br>
     * Der Vektor hat hier nicht konstant die Laenge 1, sondern als Einheit
     * wird eine quadratische Zelle betrachtet. Dadurch entsteht bei 45 Grad
     * eine Einheitslaenge von ca. 1.4 und bei waage- und senkrecht der echte
     * Einheitsvektor mit der Laenge 1 (im Einheitskreis).
     * @param y0 double Startpunkt y
     * @param x0 double Startpunkt x
     * @param y1 double Endpunkte y
     * @param x1 double Endpunkt x
     * @param e int z.B. e=0 fuer 1/10^0 = 1. Erlaubte Werte: 0,1,2.
     * @return {@link LatLon} der den normalisierten Vektor darstellt.
     */
    public final static LatLon calcOrthoSquareUnitVector(
            double y0, double x0, double y1, double x1, int e) {

        double scale = E10Scale[e];
        double wh = 1 / scale;

        double dy = y1 - y0, dyAbs = abs(dy);
        double dx = x1 - x0, dxAbs = abs(dx);

        double dty, dtx;
        if (dyAbs > dxAbs) {
            dty = wh * signum(dy); // Laenge 1*wh
            dtx = dx / dyAbs / scale;
        } else {
            dtx = wh * signum(dx); // Laenge 1*wh
            dty = dy / dxAbs / scale;
        }
        return new LatLon(dtx, -dty);
    }

} // EOC
