/*
 * $Id: Bounding.java 5118 2020-10-18 13:33:42Z carsten $
 *
 * Copyright 2014 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.misc.GeoUtils.tile2lat;
import static de.cm.osm2po.misc.GeoUtils.tile2lon;
import static de.cm.osm2po.misc.Utils.roundE5;
import static de.cm.osm2po.misc.Utils.roundE7;
import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;

import java.util.Arrays;

/**
 * Repraesentiert eine Begrenzung.
 * Dies kann sowohl eine rechteckige BoundingBox, als
 * auch ein valider (outer) Polygon-Ring sein.
 * <br><br><b>Wichtig:</b><br>
 * Uebergaenge von Plus- nach Minus-Graden werden
 * nicht unterstuetzt. Das heisst, die Koordinaten der
 * unteren linken Ecke muessen kleiner als die der oberen
 * rechte Ecke sein.
 *
 * @author (c) 2014 - Carsten Moeller, info@osm2po.de
 */
public class Bounding {

    private double minLat;
    private double minLon;
    private double maxLat;
    private double maxLon;

    private double[] polyLats;
    private double[] polyLons;

    public Bounding() {
        this.minLat = -90; // Achtung! Float.Min is 0!
        this.minLon = -180;
        this.maxLat = 90;
        this.maxLon = 180;

        this.polyLats = null;
        this.polyLons = null;
    }

    /**
     * Slippy-Map-Tile-Konstruktor
     * @param x int
     * @param y int
     * @param zoom int
     */
    public Bounding(int x, int y, int zoom) {
        this.minLat = tile2lat(y + 1, zoom); // y=0 ist oben!
        this.minLon = tile2lon(x, zoom); // x=0 bleibt links!
        this.maxLat = tile2lat(y, zoom);
        this.maxLon = tile2lon(x + 1, zoom);
    }

    public LatLon calcBoxCenter() {
        return new LatLon((this.maxLat + this.minLat) / 2,
                (this.maxLon + this.minLon) / 2);
    }

    public void grow(LatLon latLon) {
        this.grow(latLon.getLat(), latLon.getLon());
    }

    public void grow(double lat, double lon) {
        if (lat < this.minLat) this.minLat = lat;
        if (lon < this.minLon) this.minLon = lon;
        if (lat > this.maxLat) this.maxLat = lat;
        if (lon > this.maxLon) this.maxLon = lon;
    }

    public Bounding(LatLons polygon) {
        if (null == polygon) throw new IllegalArgumentException(
        "shape must not be null");
        if (polygon.getSize() < 2) throw new IllegalArgumentException(
        "shape must contain at least two points");

        if (2 == polygon.getSize()) {
            long c0 = polygon.getCoord(0);
            long c1 = polygon.getCoord(1);
            this.createBoundingBox(latOf(c0), lonOf(c0), latOf(c1), lonOf(c1));

        } else {
            // Achtung! Double.MIN_VALUE ist nicht negativ!
            double minLat = +90d;
            double minLon = +180d;
            double maxLat = -90d;
            double maxLon = -180d;

            long[] coords = polygon.getCoords();
            int n = coords.length;

            // Polygon schliessen, wenn noetig
            if (coords[n -1] != coords[0]) {
                coords = Arrays.copyOf(coords, ++n);
                coords[n -1] = coords[0];
            }

            this.polyLats = new double[n];
            this.polyLons = new double[n];
            for (int i = 0; i < n; i++) {
                double lat = latOf(coords[i]);
                double lon = lonOf(coords[i]);
                this.polyLats[i] = lat;
                this.polyLons[i] = lon;

                if (lat < minLat) minLat = lat;
                if (lon < minLon) minLon = lon;
                if (lat > maxLat) maxLat = lat;
                if (lon > maxLon) maxLon = lon;
            }
            this.createBoundingBox(
                    minLat, minLon, maxLat, maxLon);
        }
    }

    public Bounding(LatLon bottomLeft, LatLon topRight) {
        this.createBoundingBox(bottomLeft, topRight);
    }

    public Bounding(double minLat, double minLon, double maxLat, double maxLon) {
        this.createBoundingBox(minLat, minLon, maxLat, maxLon);
    }

    public boolean contains(LatLons latLons) {
        return this.contains(latLons.getCoords());
    }

    public boolean contains(long[] coords) {
        if (null == coords) return false;
        for (int i = 0; i < coords.length; i++) {
            if (!this.contains(latOf(coords[i]), lonOf(coords[i]))) {
                return false;
            }
        }
        return true;
    }

    public boolean contains(long coord) {
        return this.contains(latOf(coord), lonOf(coord));
    }

    public boolean contains(LatLon latLon) {
        return this.contains(latLon.getLat(), latLon.getLon());
    }

    public boolean contains(double lat, double lon) {
        if (!this.boundingBoxContains(lat, lon)) return false;
        if (null == this.polyLats) return true;

        int n = this.polyLats.length;
        boolean polygonContains = false;
        for (int i = 0, j = n - 1; i < n; j = i++) {
            double lati = this.polyLats[i];
            double loni = this.polyLons[i];
            double latj = this.polyLats[j];
            double lonj = this.polyLons[j];

            if ((lati > lat) != (latj > lat)
                    && (lon < (lonj - loni) * (lat - lati) / (latj - lati) + loni))
                polygonContains = !polygonContains;
        }
        return polygonContains;
    }

    private boolean boundingBoxContains(double lat, double lon) {
        if (lat < this.minLat) return false;
        if (lon < this.minLon) return false;
        if (lat > this.maxLat) return false;
        if (lon > this.maxLon) return false;
        return true;
    }

    private void createBoundingBox(LatLon bottomLeft, LatLon topRight) {
        this.createBoundingBox(bottomLeft.getLat(), bottomLeft.getLon(),
                topRight.getLat(), topRight.getLon());
    }

    private void createBoundingBox(double minLat, double minLon, double maxLat, double maxLon) {
        if (minLat < -90 || minLat > 90 || maxLat < -90 || maxLat > 90
                || minLon < -180 || minLon > 180 || maxLon < -180 || maxLon > 180)
            throw new IllegalArgumentException("Coordinates not on earth");
        if (minLat > maxLat || minLon > maxLon)
            throw new IllegalArgumentException("First coord must be smaller then second");

        this.minLat = minLat;
        this.minLon = minLon;
        this.maxLat = maxLat;
        this.maxLon = maxLon;
    }

    @Override
    public String toString() {
        String s = "B("
            + roundE5(this.minLat) + "," + roundE5(this.minLon) + ","
            + roundE5(this.maxLat) + "," + roundE5(this.maxLon) + ")";
        if (this.polyLats != null) {
            int n = this.polyLats.length;
            s += "P[" + n + "]("
            + roundE7(this.polyLats[0]) + ",...)";
        }
        return s;
    }


}