/*
 * $Id: Boundings.java 4548 2017-10-21 08:40:46Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import java.io.BufferedReader;
import java.io.StringReader;
import java.util.Arrays;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;


/**
 * Container fuer {@link Bounding}-Objekte, die entweder als 
 * Inner oder Outer fungieren und so eine polygone oder
 * rechteckige Begrenzung bilden.
 * 
 * @author (c) 2015 - Carsten Moeller, info@osm2po.de
 */
public class Boundings {

    private int nInners;
    private int nOuters;
    private Bounding[] inners;
    private Bounding[] outers;
    
    /** Default-Konstruktor. */
    public Boundings() {
        this.inners = new Bounding[0];
        this.outers = new Bounding[1];
    }
    
    /**
     * Konstruktor, der mehrere String-Formate interpretiert.
     * Ruft hierzu {@link #parse(String, boolean)} auf.
     * @param strPoly String
     * @param longitudeFirst true: Reihenfolge: Erst Lon, dann Lat.
     */
    public Boundings(String strPoly, boolean longitudeFirst) {
        this();
        this.parse(strPoly, longitudeFirst);
    }
    
    /**
     * Fuegt eine aeussere Begrenzung hinzu. 
     * @param bounding {@link Bounding} Outer, NULLABLE.
     */
    public final void addOuter(Bounding bounding) {
        if (null == bounding) return;
        if (this.nOuters >= this.outers.length) // ArrayOverflowCheck
            this.outers = Arrays.copyOf(this.outers, this.outers.length * 2 + 1);
        this.outers[this.nOuters++] = bounding;
    }
    
    /**
     * Fuegt eine innere Begrenzung hinzu. 
     * @param bounding {@link Bounding} Inner, NULLABLE.
     */
    public final void addInner(Bounding bounding) {
        if (null == bounding) return;
        if (this.nInners >= this.inners.length) // ArrayOverflowCheck
            this.inners = Arrays.copyOf(this.inners, this.inners.length * 2 + 1);
        this.inners[this.nInners++] = bounding;
    }
    
    /**
     * Ruft entweder {@link #addInner(Bounding)} oder
     * {@link #addOuter(Bounding)}.
     * @param bounding {@link Bounding} bounding, NULLABLE.
     * @param inner boolean true: als Inner einfuegen, sonst Outer.
     */
    public final void add(Bounding bounding, boolean inner) {
        if (inner) this.addInner(bounding);
        else this.addOuter(bounding);
    }
    
    /**
     * Prueft, ob eine Koordinate innerhalb des OsmPolygon liegt.
     * @param lat double Latitude
     * @param lon double Longitude
     * @return true: innerhalb, false: ausserhalb
     */
    public final boolean contains(double lat, double lon) {
        boolean contains = false;
        for (int i = 0; i < this.nOuters; i++)
            if (this.outers[i].contains(lat, lon)) {
                contains = true; break;
            }
        if (contains) for (int i = 0; i < this.nInners; i++)
            if (this.inners[i].contains(lat, lon)) {
                contains = false; break;
            }
        return contains;
    }
    
    /**
     * Parst einen ein- oder mehrzeiligen String zu einer Begrenzung,
     * bestehend aus aeusseren und inneren Polygonen, bzw. Rechtecken.<br>
     * Im einfachsten Falle handelt es sich um eine Kette von paarigen
     * LatLon bzw. LonLat-Werten. Als Token-Trenner werden Leerzeichen,
     * Kommata und Tabs akzeptiert.<br>
     * Ist das Token ein Plus oder ein Minuszeichen, so wird dies
     * als Schalter fuer Outer und Inner interpretiert.<br>
     * Schlaegt das Parsen fuer eine Zeile fehl, so handelt es sich
     * moeglicherweise um ein Poly-File und es werden zusaetzliche
     * Schritte unternommen: So wird z.B. auf ein fuehrendes Ausrufungszeichen
     * geprueft, was den Beginn eines Inner-Polygons markiert.
     * 
     * @param strPoly String Boundings
     * @param longitudeFirst true: Reihenfolge: Erst Lon, dann Lat.
     */
    protected void parse(String strPoly, boolean longitudeFirst) {
        if (null == strPoly) return;
        BufferedReader reader = new BufferedReader(new StringReader(strPoly));
        LatLons points = new LatLons();
        boolean isInner = false;
        String line = null;
        try {
            while ((line = reader.readLine()) != null) {
                StringTokenizer tokens = new StringTokenizer(line, " ,\t");
                while (tokens.hasMoreTokens()) {
                    try {
                        String nextFirst = tokens.nextToken();
                        if (1 == nextFirst.length()) {
                            char c = nextFirst.charAt(0);
                            if ('+' == c || '-' == c) { // osm2po-Notation
                                if (points.getSize() > 0)
                                    this.add(new Bounding(points), isInner);
                                points.clear();
                                isInner = '-' == c;
                                continue;
                            }
                        }
                        double lat = Double.parseDouble(nextFirst);
                        double lon = Double.parseDouble(tokens.nextToken());
                        if (longitudeFirst) points.append(lon, lat);
                            else points.append(lat, lon);

                    } catch (NoSuchElementException nse) {
                        // Fehler, weil lat parsebar, lon aber nicht vorhanden
                        break;
                    } catch (NumberFormatException nfe) {
                        // Wenn keine Zahl, dann moeglicherweise PolyFile - Zeile!
                        if (points.getSize() > 0) this.add(new Bounding(points), isInner);
                        points.clear();
                        isInner = line.trim().startsWith("!"); // PolyFile-Special 
                        break;
                    }
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        if (points.getSize() > 0) this.add(new Bounding(points), isInner);
    }
    
    @Override
    public String toString() {
        String s = "";
        if (this.nOuters > 0) {
            for (int i = 0; i < this.nOuters; i++)
                s += (i == 0 ? "" : "\n") + "Outer[" + i + "] " + this.outers[i]; 
        } else {
            s += "Outer[0] not found";
        }
        if (this.nInners > 0) {
            for (int i = 0; i < this.nInners; i++) 
                s += "\nInner[" + i + "] " + this.inners[i]; 
        }
        return s;
    }
    
}