/*
 * $Id: Edge.java 5243 2021-02-09 21:03:31Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.routing.Graph;

/**
 * Diese Klasse kapselt die Kanten-Attribute fuer den {@link Graph}.
 * @author (c) 2012/2017 - Carsten Moeller - info@osm2po
 */
public class Edge implements Var<Edge>, Comparable<Edge> {

    private int id;
    private int sourceId;
    private int targetId;
    private float km;
    private float h;
    private int flags;
    private byte noTurnBits;

    // Kritische ReadOnly-Attribute
    public final int getId() {return this.id;}
    public final int getSourceId() {return this.sourceId;}
    public final int getTargetId() {return this.targetId;}

    // Vergleichsweise unkritische Attribute
    public final void setKm(float km) {this.km = km;}
    public final float getKm() {return this.km;}
    public final void setH(float h) {this.h = h;}
    public final float getH() {return this.h;}
    public final void setFlags(int flags) {this.flags = flags;}
    public final void overlayFlags(int flags) {this.flags |= flags;}
    public final int getFlags() {return this.flags;}
    public final void setNoTurnBits(byte noTurnBits) {this.noTurnBits = noTurnBits;}
    public final byte getNoTurnBits() {return this.noTurnBits;}

    /** CTOR fuer Serialisierung */
    public Edge() {}

    /**
     * CTOR fuer die eigentliche Erstellung.
     * @param id int Id der Kante
     * @param sourceId int Id des SourceVertex
     * @param targetId int Id des TargetVertex
     * @param km float Laenge in km
     * @param h float Zeit in h
     * @param flags int Flags
     */
    public Edge(int id, int sourceId, int targetId, float km, float h, int flags) {
        this.id = id;
        this.sourceId = sourceId;
        this.targetId = targetId;
        this.km = km;
        this.h = h;
        this.flags = flags;
    }

    public final boolean isReverse() {return Graph.isReverse(this.id);}
    public final boolean isOneWay() {return Graph.isOneWay(this.id);}

    @Override
    public Edge readFromStream(InStream inStream) {
        this.id = inStream.readInt();
        this.sourceId = inStream.readInt();
        this.targetId = inStream.readInt();
        this.km = Float.intBitsToFloat(inStream.readInt());
        this.h = Float.intBitsToFloat(inStream.readInt());
        this.noTurnBits = inStream.readByte();
        this.flags = inStream.readInt();
        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.id);
        outStream.writeInt(this.sourceId);
        outStream.writeInt(this.targetId);
        outStream.writeInt(Float.floatToIntBits(this.km));
        outStream.writeInt(Float.floatToIntBits(this.h));
        outStream.writeByte(this.noTurnBits);
        outStream.writeInt(this.flags);
    }

    @Override
    public String toString() {
        String s = "Id:" + this.id
            + " Reverse:" + this.isReverse()
            + " OneWay:" + this.isOneWay()
            + " SourceId:" + this.sourceId
            + " TargetId:" + this.targetId
            + " km:" + this.km
            + " h:" + this.h
            + " NoTurnBits:" + this.noTurnBits
            + " (" + Integer.toBinaryString(this.noTurnBits) + ")"
		    + " Flags:" + this.flags
		    + " (" + Integer.toBinaryString(this.flags) + ")";
        return s;
    }

    @Override
    public int compareTo(Edge that) {
        if (this.sourceId < that.sourceId) return -1;
        if (this.sourceId > that.sourceId) return 1;
        return 0;
    }

}
