/*
 * $Id: LatLon.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;

/**
 * Diese Klasse repraesentiert eine Koordinate,
 * welche auf 7 Nachkommastellen begrenzt ist,
 * sich dadurch jedoch in einem einzigen <code>long</code>
 * (8 byte) unterbringen laesst.<br>
 * Zufaelligerweise ist dies der Standard von OSM.<br>
 * 7 Nachkommastellen bedeutet, dass wir am Aequator
 * in etwa auf 1.11 cm genau sind. Zu den Polen hin
 * werden wir in x-Richtung noch genauer.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class LatLon implements Var<LatLon> {

    public static final long NULL_COORDINATE = Long.MIN_VALUE;
    private static final long LONG_LOW_INT_MASK = 0x00000000FFFFFFFFl;
    private static final long LONG_HIGH_INT_MASK = 0xFFFFFFFF00000000l;

    protected long coord;

    public LatLon() {
        this.coord = NULL_COORDINATE;
    }

    public final boolean isNull() {
        return this.coord == NULL_COORDINATE;
    }

    public LatLon(double lat, double lon) {
        this.setLatLon(lat, lon);
    }


    public LatLon copy() {
        LatLon copy = new LatLon();
        copy.coord = this.coord;
        return copy;
    }

    /**
     * @return long LatLon 64Bit verschluesselt.
     */
    public long getCoord() {return this.coord;}
    /**
     * @param coord long 64Bit verschluesselte LatLon.
     * @return this 4Chaining
     */
    public LatLon setCoord(long coord) {this.coord = coord; return this;}


    /**
     * Setzt Latitude und Longitude in einem Rutsch.
     * Dies ist schneller als einzeln {@link #setLat(double)}
     * und {@link #setLon(double)} aufzurufen.
     * Wichtig: Die Werte werden dabei auf 7 Nachkommastellen
     * gerundet!
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     */
    public void setLatLon(double lat, double lon) {
        long llat = (long) StrictMath.floor(lat * 1e7 + .5d);
        long llon = (long) StrictMath.floor(lon * 1e7 + .5d);
        this.coord = (llat << 32) | (llon & LONG_LOW_INT_MASK);
    }

    /**
     * Ueberschreibt die {@link #equals(Object)} Methode
     * fuer geografische Gleichheit.
     */
    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof LatLon)) return false;
        LatLon that = (LatLon) obj;
        return this.coord == that.coord;
    }

    /**
     * Setzt den Breitengrad.
     * @param lat double
     */
    public void setLat(double lat) {
        this.coord &= LONG_LOW_INT_MASK; // HighBits nullen.
        long llat = (long) StrictMath.floor(lat * 1e7 + .5d);
        this.coord |= (llat << 32);
    }

    /**
     * Liefert den Breitengrad.
     * @return double
     */
    public double getLat() {
        int ilat = (int) (this.coord >> 32); // HighBits nach unten schieben
        return ilat / 1e7;
    }

    /**
     * Setzt den Laengengrad.
     * @param lon double
     */
    public void setLon(double lon) {
        this.coord &= LONG_HIGH_INT_MASK; // LowBits nullen.
        long llon = (long) StrictMath.floor(lon * 1e7 + .5d);
        this.coord |= (llon & LONG_LOW_INT_MASK); // Mask killt 32 1en bei neg. Werten
    }

    /**
     * Liefert den Laengengrad.
     * @return double
     */
    public double getLon() {
        int ilon = (int) (this.coord & LONG_LOW_INT_MASK);
        return ilon / 1e7;
    }

    @Override
    public LatLon readFromStream(InStream inStream) {
        this.coord = inStream.readLong();
        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeLong(this.coord);
    }

    @Override
    public String toString() {
        if (NULL_COORDINATE == this.coord) return "null";
        else return ("(" + this.getLat() + " " + this.getLon() + ")");
    }

    /**
     * Liefert den Laengengrad.
     * @param coord long
     * @return double
     * @see #latOf(long)
     * @see #toCoord(double, double)
     */
    public static double lonOf(long coord) {
        int ilon = (int) (coord & LONG_LOW_INT_MASK);
        return ilon / 1e7;
    }

    /**
     * Liefert den Breitengrad.
     * @param coord long
     * @return double
     * @see #lonOf(long)
     * @see #toCoord(double, double)
     */
    public static double latOf(long coord) {
        int ilat = (int) (coord >> 32); // HighBits nach unten schieben
        return ilat / 1e7;
    }

    /**
     * Konvertiert Latitude und Longitude in einen LongInteger.
     * Wichtig: Die Werte werden dabei auf 7 Nachkommastellen
     * gerundet!
     * @param lat double Breitengrad.
     * @param lon double Laengengrad.
     * @return Koordinate als long kodiert
     * @see #latOf(long)
     * @see #lonOf(long)
     */
    public static long toCoord(double lat, double lon) {
        long llat = (long) StrictMath.floor(lat * 1e7 + .5d);
        long llon = (long) StrictMath.floor(lon * 1e7 + .5d);
        return (llat << 32) | (llon & LONG_LOW_INT_MASK);
    }

    /**
     * Verschiebt diese Koordinate um y und x.
     * @param y double
     * @param x double
     * @return this 4 chaining
     */
    public final LatLon moveBy(double y, double x) {
        double lat = latOf(this.coord) + y;
        double lon = lonOf(this.coord) + x;
        this.coord = toCoord(lat, lon);
        return this;
    }
    /**
     * Verschiebt diese Koordinate um einen Faktor.
     * @param d double
     * @return this 4 chaining
     */
    public final LatLon moveBy(double d) {
        double lat = latOf(this.coord) * d;
        double lon = lonOf(this.coord) * d;
        this.coord = toCoord(lat, lon);
        return this;
    }

    /**
     * Betrachtet diese Koordinate als Vektor und dreht
     * ihn um 90 Grad nach links.
     * @return this 4 chaining
     */
    public final LatLon turnPiHalf() {
        double lat = latOf(this.coord);
        double lon = lonOf(this.coord);
        double z = -lat;
        lat = lon;
        lon = z;
        this.coord = toCoord(lat, lon);
        return this;
    }

}