/*
 * $Id: LatLons.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.misc.ArrayUtils.intArrayConcat;
import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;
import static de.cm.osm2po.misc.ArrayUtils.intArrayListToArray2D;
import static de.cm.osm2po.misc.GeoUtils.calcDistance;
import static de.cm.osm2po.misc.GeoUtils.intersection;
import static de.cm.osm2po.misc.ZoneUtils.raytraceZones;
import static de.cm.osm2po.model.LatLon.NULL_COORDINATE;
import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static java.lang.Math.abs;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.misc.ZoneUtils;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarNum;

/**
 * Bildet eine Array-Klammer um Koordinaten und erlaubt
 * die Serialisierung einer solchen Menge als {@link Var}.
 * Bietet zudem einige Geo-Funktionalitaeten, die je nach
 * Betrachtungsweise, z.B. als Polygon, wirken.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public final class LatLons implements Var<LatLons> {
    private long[] coords;
    private int size;

    public LatLons() {}

    public LatLons(LatLon[] latLonArr) {
        if (latLonArr != null) {
            for (int i = 0; i < latLonArr.length; i++) {
                this.append(latLonArr[i]);
            }
        }
    }

    public void clear() {
        this.coords = null;
        this.size = 0;
    }

    public int getSize() {return this.size;}

    public final LatLons copy() {
        LatLons copy = new LatLons();
        copy.coords = Arrays.copyOf(this.coords, this.size);
        copy.size = this.size;
        return copy;
    }

    public final long[] getCoords() {
        if (null == this.coords) return null;
        if (this.size != this.coords.length) {
            this.coords = Arrays.copyOf(this.coords, this.size);
        }
        return this.coords;
    }

    public final LatLon[] toLatLonArray() {
        if (null == this.coords) return null;
        LatLon[] latLonArr = new LatLon[this.size];
        for (int i = 0; i < this.size; i++) {
            latLonArr[i] = new LatLon().setCoord(this.coords[i]);
        }
        return latLonArr;
    }

    public final long getCoord(int idx) {
        return this.coords[idx];
    }

    public final LatLon getLatLon(int idx) {
        return new LatLon().setCoord(this.coords[idx]);
    }

    /**
     * Kuerzt diese Koordinatenkette auf eine neue Laenge und liefert
     * den abgeschnittenen Rest als neues Objekt zurueck.
     * @param newSize int,
     *      Neue Laenge, bzw. erster Koordinaten-Index
     *      ab dem abgeschnitten werden soll.
     * @return new {@link LatLons},
     *      Hinten abgeschnittener Teil oder <code>null</code>, wenn
     *      <code>len &gt;= this.size</code>
     */
    public final LatLons crop(int newSize) {
        if (newSize >= this.size) return null;
        int nCropped = this.size - newSize;
        long[] croppedCoords = new long[nCropped];
        System.arraycopy(this.coords, newSize, croppedCoords, 0, nCropped);

        this.size = newSize;
        return new LatLons().setCoords(croppedCoords);
    }

    /**
     * Kettet Koordinaten eines anderen {@link LatLons} an.
     * <br><br><b>Wichtig:</b><br>
     * Die erste Koordinate der anzufuegenden Kette muss mit der
     * letzten dieser Kette uebereinstimmen und wird dann entfernt.
     * Besitzt dieses Objekt noch keine Koordinaten, so wird eine
     * Kopie erstellt und <code>true</code> zurueckgeliefert.
     * @param latLons {@link LatLons} anzufuegende Koordinaten
     * @param allowReverse true: evtl. anzufuegende Kette umdrehen.
     * @return true: angefuegt ja/nein.
     */
    public final boolean concat(LatLons latLons, boolean allowReverse) {
        if (null == latLons) return false;
        long[] cc = latLons.getCoords(); // Kein Copy!
        int n = cc.length;
        if (0 == n) return false;

        if (0 == this.size) {
            this.coords = Arrays.copyOf(cc, n);
            this.size = n;
            return true;
        }

        long c = this.coords[this.size -1];

        if (allowReverse) {
            if (cc[n -1] == c) { // reverse
                cc = Arrays.copyOf(cc, n);
                for (int x = 0, y = n -1; x < y; ++x, --y) {
                    long tmp = cc[x]; cc[x] = cc[y]; cc[y] = tmp;
                }
            }
        }

        if (cc[0] == c) {
            int newSize = this.size + n - 1;
            this.coords = Arrays.copyOf(this.coords, newSize);
            System.arraycopy(cc, 1, this.coords, this.size, n -1);
            this.size = newSize;
            return true;
        }

        return false;
    }

    /**
     * Dreht die Koordinaten um.
     * @return <code>this</code>
     */
    public final LatLons reverse() {
        for (int x = 0, y = this.size -1; x < y; ++x, --y) {
            long tmp = this.coords[x];
            this.coords[x] = this.coords[y];
            this.coords[y] = tmp;
        }
        return this;
    }

    /**
     * Ueberprueft die Koordinaten auf Ring-Bildung.
     * @return true,
     *      wenn mehr als 2 Koordinaten vorhanden sind und
     *      die erste und letzte Koordinate identisch sind
     */
    public final boolean isRing() {
        if (this.size < 3) return false;
        return this.coords[0] == this.coords[this.size -1];
    }

    /**
     * Ueberprueft auf Ring-Bildung und Ueberschneidungen.
     * Zieht dabei rueckwaerts ausgehend vom letzten Punkt
     * alle Knoten in Betracht.
     * <br><br><b>Wichtig:</b><br>
     * Die Ringbildung wird hier nicht geometrisch auf Ueberschneidungen
     * geprueft. Es werden lediglich gleiche Koordinaten betrachtet.
     * @return
     *      <code>-1</code>: Kein Ring<br>
     *      <code>=0</code>: Ring, erste und letzte Koordinate gleich<br>
     *      <code>&gt;0</code>: Ring, Index der ueberscheidenden Koordinate<br>
     */
    public final int findLastRing() {
        if (this.size < 3) return -1;
        long lastCoord = this.coords[this.size -1];
        for (int i = this.size -3; i >= 0; i--) {
            if (this.coords[i] == lastCoord) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Interpretiert die Koordinaten als Streckenpunkte
     * und liefert die Laenge in Kilometern.
     * @return double km.
     */
    public final double calcLengthKm() {
        double l = 0;
        for (int i = 1; i < this.size; i++) {
            long coord1 = this.coords[i-1];
            long coord2 = this.coords[i];

            double lat1 = latOf(coord1);
            double lon1 = lonOf(coord1);
            double lat2 = latOf(coord2);
            double lon2 = lonOf(coord2);

            l += GeoUtils.calcDistance(lat1, lon1, lat2, lon2);
        }
        return l;
    }

    /**
     * Sortiert die Koordinaten nach Distanz zu einer uebergebenen Koordinate.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     */
    public final void orderByDistanceTo(double lat, double lon) {
        if (this.size < 2) return;
        for (int i = 0; i < this.size -1; i++) {
            double minDist = Double.MAX_VALUE;
            int minJ = -1;
            for (int j = i; j < this.size; j++) {
                double lat2 = latOf(this.coords[j]);
                double lon2 = lonOf(this.coords[j]);
                double dist = GeoUtils.calcDistance(lat, lon, lat2, lon2);
                if (dist < minDist) {
                    minJ = j;
                    minDist = dist;
                }
            }
            // Swap
            long tmp = this.coords[i];
            this.coords[i] = this.coords[minJ];
            this.coords[minJ] = tmp;
        }
    }

    public final LatLons setCoords(long[] coords) {
        this.coords = coords;
        if (null == coords) this.size = 0;
            else this.size = this.coords.length;
        return this;
    }

    public final LatLons setCoord(int idx, long coord) {
        int maxIdx = this.size -1;
        if (idx < 0 || idx > maxIdx) throw new ArrayIndexOutOfBoundsException(idx);
        this.coords[idx] = coord;
        return this;
    }

    public final LatLons setFirstCoord(long coord) {
        return this.setCoord(0, coord);
    }

    public final LatLons setLastCoord(long coord) {
        return this.setCoord(this.size -1, coord);
    }

    public final LatLons setCoords(LatLon[] latLons) {
    	if (null == latLons) {
    		this.coords = null;
    		this.size = 0;
    	} else {
    	    this.size = latLons.length;
    	    this.coords = new long[this.size];
    	    for (int i = 0; i < this.size; i++) {
    	        this.coords[i] = latLons[i].getCoord();
    	    }
    	}
    	return this;
    }

    public final LatLons append(double lat, double lon) {
        this.append(new LatLon(lat, lon));
        return this;
    }

    /**
     * Fuegt eine neue Koordinate hinten an.
     * @param latLon {@link LatLon}
     * @return this4Chaining
     */
    public final LatLons append(LatLon latLon) {
        this.append(latLon.coord);
        return this;
    }

    /**
     * Fuegt eine neue Koordinate hinten an.
     * @param coord long Koordinate
     * @return this4Chaining
     */
    public final LatLons append(long coord) {
        if (null == this.coords) this.coords = new long[10];
        if (this.coords.length == this.size) {
            this.coords = Arrays.copyOf(this.coords, this.coords.length * 2);
        }
        this.coords[this.size++] = coord;
        return this;
    }

    /**
     * Interpretiert die Koordinaten als Huellpunkte und liefert die Flaeche.
     * @return double Flaeche in Grad.
     * <br><br><b>Wichtig:</b><br>
     * Wenn das Polygon gegen den Uhrzeigersinn laeuft, dann wird
     * eine negative Flaeche geliefert, deren Betrag jedoch korrekt ist.
     */
    public final double area() {
        int i, j;
        double area = 0;
        for (i = 0; i < this.size; i++) {
            j = (i + 1) % this.size;
            area += lonOf(this.coords[i]) * latOf(this.coords[j]);
            area -= lonOf(this.coords[j]) * latOf(this.coords[i]);
        }
        area /= 2.0;
        return (area);
    }

    /**
     * Interpretiert die Koordinaten als Huellpunkte und liefert
     * den Mittelpunkt als Masse-Schwerpunkt bzw. Mittelpunkt des OsmPolygon.
     * <br><br><b>Wichtig:</b><br>
     * Die Koordinaten muessen zwingend eine Flaeche aufspannen.
     * Gerade Linien, z.B. Strassen mit nur zwei Koordinaten, liefern
     * Lat 0 und Lon 0.
     * @return {@link LatLon} Mittelpunkt
     * @see #center()
     */
    public final LatLon centroid() {
        double cx = 0, cy = 0;
        double area = this.area();
        int i, j;

        double factor = 0;
        for (i = 0; i < this.size; i++) {
            j = (i + 1) % this.size;
            factor = (lonOf(this.coords[i]) * latOf(this.coords[j])
                    - lonOf(this.coords[j]) * latOf(this.coords[i]));
            cx += (lonOf(this.coords[i]) + lonOf(this.coords[j])) * factor;
            cy += (latOf(this.coords[i]) + latOf(this.coords[j])) * factor;
        }
        area *= 6.0f;
        factor = 1 / area;
        cx *= factor;
        cy *= factor;
        return new LatLon(cy, cx);
    }

    /**
     * Interpretiert die Koordinaten als Wolke von
     * Punkten und liefert den Masseschwerpunkt der Haeufung.
     * Diese Mehode ist daher auch fuer Strecken mit
     * nur 2 Koordinaten geeignet.
     * @return {@link LatLon} Mittelpunkt NOT NULL
     * @see #centroid()
     */
    public final LatLon center() {
        double lonSum = 0, latSum = 0;
        for (int i = 0; i < this.size; i++) {
            latSum += latOf(this.coords[i]);
            lonSum += lonOf(this.coords[i]);
        }
        latSum /= this.size;
        lonSum /= this.size;

        return new LatLon(latSum, lonSum);
    }

    /**
     * Interpretiert die Koordinaten als Huellpunkte und prueft,
     * ob die uebergebene Koordinate in diesem Polygon liegt.
     * @param lat double Breitengrad
     * @param lon double Laengengrad
     * @return true: Punkt liegt innerhalb der Huelle
     */
    public final boolean contains(double lat, double lon) {
        int n = this.size;
        boolean polygonContains = false;
        for (int i = 0, j = n - 1; i < n; j = i++) {
            long coordi = this.coords[i], coordj = this.coords[j];
            double lati = latOf(coordi);
            double loni = lonOf(coordi);
            double latj = latOf(coordj);
            double lonj = lonOf(coordj);
            if ((lati > lat) != (latj > lat) && (lon <
                    (lonj - loni) * (lat - lati) / (latj - lati) + loni))
                polygonContains = !polygonContains;
        }
        return polygonContains;
    }

    /**
     * Interpretiert die Koordinaten als Huellpunkte und prueft,
     * ob die uebergebenen Koordinaten in diesem Polygon liegen.<br>
     * Ruft dazu {@link #contains(double, double)} fuer jeden Punkt.
     * @param latLons {@link LatLons} zu pruefende Koordinaten
     * @param fully boolean
     *      true: Alle Punkte muessen in der Huelle liegen.
     *      false: Mindestens ein Punkt muss innerhalb liegen.
     * @return true: Alle oder einige Punkte liegen in der Huelle.
     */
    public final boolean contains(LatLons latLons, boolean fully) {
        if (null == latLons || 0 == latLons.size) return false;
        for (int i = 0; i < latLons.size; i++) {
            long coord = latLons.coords[i];
            if (this.contains(latOf(coord), lonOf(coord))) {
                if (!fully) return true;
            } else {
                if (fully) return false;
            }
        }
        return true;
    }

    @Override
    public final void writeToStream(OutStream outStream) {
        VarNum.writeUnsigned(this.size, outStream);
        for (int i = 0; i < this.size; i++) {
            outStream.writeLong(this.coords[i]);
        }
    }

    @Override
    public final LatLons readFromStream(InStream inStream) {
        this.size = (int) VarNum.readUnsigned(inStream);
        this.coords = new long[this.size];
        for (int i = 0; i < this.size; i++) {
            this.coords[i] = inStream.readLong();
        }
        return this;
    }

    @Override
    public final boolean equals(Object obj) {
        LatLons that = (LatLons) obj;
        if (null == that) return false;
        if (this == that) return true;
        if (this.size != that.size) return false;
        for (int i = 0; i < this.size; i++) {
            if (this.coords[i] != that.coords[i]) return false;
        }
        return true;
    }

    @Override
    public final String toString() {
    	StringBuilder sb = new StringBuilder(0x1000 /*4k*/);
    	for (int i = 0; i < this.size; i++) {
    		if (i > 0) sb.append(',');
    		long coord = this.coords[i];
    		if (NULL_COORDINATE == coord) {
    		    sb.append("NULL");
    		    continue;
    		}
    		sb.append('(');
    		sb.append(latOf(coord));
    		sb.append(' ');
    		sb.append(lonOf(coord));
            sb.append(')');
    	}
    	return sb.toString();
    }

    /**
     * Interpretiert diese Koordinaten als Kette von Teilstrecken
     * und rastert sie mittels
     * {@link ZoneUtils#raytraceZones(double, double, double, double, int)}
     * @param zoneSize int Groesse einer Zonen-Zelle.
     * @return Array von ints von ZoneIds
     */
    public final int[] getZoneIds(int zoneSize) {
        List<int[]> zonesList = new ArrayList<int[]>();
        for (int i = 1; i < this.size; i++) {
            long coord1 = this.coords[i-1];
            long coord2 = this.coords[i];

            double lat1 = latOf(coord1);
            double lon1 = lonOf(coord1);
            double lat2 = latOf(coord2);
            double lon2 = lonOf(coord2);

            zonesList.add(raytraceZones(lat1, lon1, lat2, lon2, zoneSize));
        }
        int[] zones = intArrayDistinct(
                intArrayConcat(intArrayListToArray2D(zonesList)));
        return zones;
    }

    /**
     * Prueft diese Koordinatenkette gegen einen Polygon-Ring und ermittelt
     * die Laenge der Ueberschneidung.
     * @param poly {@link LatLons} NOT NULL
     * @return double Laenge der Ueberschneidung
     */
    public final double calcIntersectionLengthKm(LatLons poly) {
        LatLons isecs = new LatLons();
        double isecLength = 0;

        long coord1 = this.coords[0];
        double lat1 = latOf(coord1);
        double lon1 = lonOf(coord1);
        boolean inside = poly.contains(lat1, lon1);
        // Itereriere ueber die Teilstrecken dieser Koordinaten
        for (int i = 1; i < this.size; i++) {
            long coord2 = this.coords[i];
            if (coord1 == coord2) {
                continue;
            }

            double lat2 = latOf(coord2);
            double lon2 = lonOf(coord2);

            isecs.clear();

            if (abs(lon2 - lon1) > 180 || abs(lat2 - lat1) > 90) {
                continue; // FIXME handle dateline or throw an error instead?
            }

            // Emittle alle Ueberschneidungen der aktuellen Teilstrecke mit allen
            // Teilstrecken des Polygon-Rings.
            long coord3 = poly.coords[0];
            double lat3 = latOf(coord3);
            double lon3 = lonOf(coord3);
            for (int j = 1; j < poly.size; j++) {
                long coord4 = poly.coords[j];
                double lat4 = latOf(coord4);
                double lon4 = lonOf(coord4);

                LatLon isec = intersection(
                        lon1, lat1, lon2, lat2, lon3, lat3, lon4, lat4, false);
                if (isec != null) { // Es koennen mehrere pro Polygon sein
                    isecs.append(isec);
                }
                coord3 = coord4;
                lat3 = lat4;
                lon3 = lon4;
            }

            // Wenn mehr als eine Ueberschneidung, dann Reihenfolge sicherstellen
            if (isecs.size > 1) {
                isecs.orderByDistanceTo(lat1, lon1);
            }

            long lastIsec = coord1;
            for (int j = 0; j < isecs.size; j++) {
                long isec = isecs.coords[j];
                if (lastIsec == isec) {
                    continue; // Doppelerkennung von Ueberschneidung. Kann passieren.
                } else {
                    if (inside) {
                        isecLength += calcDistance(
                                latOf(lastIsec), lonOf(lastIsec), latOf(isec), lonOf(isec));
                    }
                    inside = !inside;
                    lastIsec = isec;
                }
            }

            if (coord2 != lastIsec && inside) {
                isecLength += calcDistance(latOf(lastIsec), lonOf(lastIsec), lat2, lon2);
            }

            coord1 = coord2;
            lat1 = lat2;
            lon1 = lon2;
        }

        return isecLength;
    }
}
