/*
 * $Id: Membership.java 5795 2023-04-20 10:02:40Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import java.util.Arrays;

import de.cm.osm2po.converter.Joiner;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarNum;
import de.cm.osm2po.primitives.VarString;

/**
 * Diese Klasse spiegelt eine umgekehrte Sicht einer
 * {@link OsmRelation} wider. Sie wird direkt vom
 * {@link Joiner} an den entsprechenden Members, also
 * {@link Node}, {@link Way} und {@link Membership}
 * angebracht.
 *
 * @author (c) 2014 - Carsten Moeller - info@osm2po.de
 */
public class Membership implements Var<Membership> {

    // Achtung! Internal Flags werden als Byte serialisiert. Bit 8 meiden.
    private static final int INTERN_FLAG_HASCLAZZ = 1;
    private static final int INTERN_FLAG_HASROLE = 2;
    private static final int INTERN_FLAG_HASFLAGS = 4;
    private static final int INTERN_FLAG_HASNAME = 8;
    private static final int INTERN_FLAG_HASPARENTS = 16;

    private long id; // OSM-Id der Relation
    private byte clazz;
    private byte role;
    private int order;
    private int flags;
    private VarString name;
    private Membership[] parents;

    public long getId() {return this.id;}
    public byte getClazz() {return this.clazz;}
    public int getFlags() {return this.flags;}
    public Membership[] getParents() {return this.parents;}
    public VarString getName() {return this.name;}

    public int getOrder() {return this.order;}
    public byte getRole() {return this.role;}

    public Membership() {}

    public Membership(byte role, int order, OsmRelation osmRelation) {
        this.id = osmRelation.getId();
        this.role = role;
        this.order = order;
        this.clazz = osmRelation.getClazz();
        this.flags = osmRelation.getFlags();

        VarString name = osmRelation.getName();
        if (name != null) this.name = name.copy();
    }

    public void setParents(Membership[] parents) {
        this.parents = parents;
    }

    public Membership findDistinctParentOrSelf() {
        Membership membership = this;
        while (membership.parents != null && membership.parents.length == 1) {
            membership = membership.parents[0];
        }
        return membership;
    }

    public void clear() {
        this.id = 0;
        this.role = 0;
        this.order = 0;
        this.clazz = 0;
        this.flags = 0;
        this.name = null;
        this.parents = null;
    }

    @Override
    public String toString() {
        return "Id:" + this.id
            + " Name:" + (null == this.name ? "null" : "\"" + this.name + "\"")
            + " Clazz:" + this.clazz
            + " Order:" + this.order
            + " Role:" + this.role
            + " Flags:" + this.flags
            + " Parents:" + Arrays.toString(this.parents); // NullSafe
    }

    @Override
    public Membership readFromStream(InStream inStream) {
        this.clear();
        this.id = VarNum.readUnsigned(inStream);
        this.order = (int) VarNum.readUnsigned(inStream);

        // Kompression
        int internFlags = inStream.readByte() & 0xFF;

        if ((internFlags & INTERN_FLAG_HASCLAZZ) != 0)
            this.clazz = inStream.readByte();
        if ((internFlags & INTERN_FLAG_HASROLE) != 0)
            this.role = inStream.readByte();
        if ((internFlags & INTERN_FLAG_HASFLAGS) != 0)
            this.flags = (int) VarNum.readUnsigned(inStream);
        if ((internFlags & INTERN_FLAG_HASNAME) != 0) {
            this.name = new VarString();
            this.name.readFromStream(inStream);
        }
        if ((internFlags & INTERN_FLAG_HASPARENTS) != 0) {
            int nParents = (int) VarNum.readUnsigned(inStream);
            this.parents = new Membership[nParents];
            for (int i = 0; i < nParents; i++) {
                this.parents[i] = new Membership().readFromStream(inStream);
            }
        }

        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        // Kompression
        int internFlags = 0;
        boolean hasClazz = this.clazz != 0;
        if (hasClazz) internFlags |= INTERN_FLAG_HASCLAZZ;
        boolean hasRole = this.role != 0;
        if (hasRole) internFlags |= INTERN_FLAG_HASROLE;
        boolean hasFlags = this.flags != 0;
        if (hasFlags) internFlags |= INTERN_FLAG_HASFLAGS;
        boolean hasName = this.name != null && !this.name.isEmpty();
        if (hasName) internFlags |= INTERN_FLAG_HASNAME;
        boolean hasParents = this.parents != null && this.parents.length != 0;
        if (hasParents) internFlags |= INTERN_FLAG_HASPARENTS;

        VarNum.writeUnsigned(this.id, outStream);
        VarNum.writeUnsigned(this.order, outStream);

        outStream.writeByte((byte) internFlags);

        if (hasClazz) outStream.writeByte(this.clazz);
        if (hasRole) outStream.writeByte(this.role);
        if (hasFlags) VarNum.writeUnsigned(this.flags, outStream);
        if (hasName) this.name.writeToStream(outStream);
        if (hasParents) {
            int nParents = this.parents.length;
            VarNum.writeUnsigned(nParents, outStream);
            for (int i = 0; i < nParents; i++) {
                this.parents[i].writeToStream(outStream);
            }
        }
    }

}
