/*
 * $Id: OsmNode.java 5490 2021-12-21 21:00:35Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.primitives.VarTypeDesk.newInstanceOf;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;

import de.cm.osm2po.converter.tiler.NodeTagResolver;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarNum;
import de.cm.osm2po.primitives.VarString;

/**
 * Repraesentiert einen OSM-Node.
 * @see OsmWay
 * @see OsmRelation
 * @see Node
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class OsmNode extends LatLon implements OsmObject {

    // Achtung! Internal Flags werden als Byte serialisiert. Bit 8 meiden.
    protected final static int INTERN_FLAG_NULL = 0x0;
    protected final static int INTERN_FLAG_HAS_ID = 0x1;
    protected final static int INTERN_FLAG_HAS_CLAZZ = 0x2;
    protected final static int INTERN_FLAG_HAS_FLAGS = 0x4;
    protected final static int INTERN_FLAG_HAS_NAME = 0x8;
    protected final static int INTERN_FLAG_HAS_CUSTOM = 0x10;
    protected final static int INTERN_FLAG_LEVELCROSSING = 0x20;
    protected final static int INTERN_FLAG_BARRIER = 0x40;

    private long id;
    private byte clazz;
    private int flags;
    private int internFlags;
    private VarString name;
    private Var<?> custom;

    /** STD-CTOR. */
    public OsmNode() {}

    /**
     * CTOR.
     * Maximal 7 Nachkommastellen!
     * Siehe auch {@link #setCoord(long)}
     * @param lon double Laengengrad
     * @param lat double Breitengrad
     */
    public OsmNode(double lat, double lon) {
        this.setLatLon(lat, lon);
        this.clazz = NodeTagResolver.CLAZZ_NULL;
        this.flags = NodeTagResolver.FLAGS_NULL;
    }


    private transient int index; // set by find for refetch

    // Transient ReFetch-Index
    @Override public int getIndex() {return index;}
    @Override public void setIndex(int index) {this.index = index;}
    @Override public long getId() {return id;}
    @Override public byte getClazz() {return this.clazz;}
    @Override public VarString getName() {return this.name;}
    @Override public int getFlags() {return this.flags;}
    @Override public Var<?> getCustom() {return this.custom;}

    public boolean isLevelCrossing() {
        return (this.internFlags & INTERN_FLAG_LEVELCROSSING) != 0;}
    public boolean isBarrier() {
        return (this.internFlags & INTERN_FLAG_BARRIER) != 0;}

    public void setId(long id) {
        if (id != 0) this.internFlags |= INTERN_FLAG_HAS_ID;
        else this.internFlags &=~ INTERN_FLAG_HAS_ID;
        this.id = id;
    }
    public void setClazz(byte clazz) {
        if (clazz != 0) this.internFlags |= INTERN_FLAG_HAS_CLAZZ;
        else this.internFlags &=~ INTERN_FLAG_HAS_CLAZZ;
        this.clazz = clazz;
    }
    public void setFlags(int flags) {
        if (flags != 0) this.internFlags |= INTERN_FLAG_HAS_FLAGS;
        else this.internFlags &=~ INTERN_FLAG_HAS_FLAGS;
        this.flags = flags;
    }
    public void setName(VarString name) {
        if (name != null) this.internFlags |= INTERN_FLAG_HAS_NAME;
        else this.internFlags &=~ INTERN_FLAG_HAS_NAME;
        this.name = name;
    }

    public void setCustom(Var<?> custom) {
        if (custom != null) this.internFlags |= INTERN_FLAG_HAS_CUSTOM;
        else this.internFlags &=~ INTERN_FLAG_HAS_CUSTOM;
        this.custom = custom;
    }

    public void setLevelCrossing(boolean isLevelCrossing) {
        this.internFlags &=~ INTERN_FLAG_LEVELCROSSING;
        if (isLevelCrossing) this.internFlags |= INTERN_FLAG_LEVELCROSSING;
    }

    public void setBarrier(boolean isBarrier) {
        this.internFlags &=~ INTERN_FLAG_BARRIER;
        if (isBarrier) this.internFlags |= INTERN_FLAG_BARRIER;
    }

    /**
     * Reinitialisiert.
     */
    public void clear() {
        this.id = 0;
        this.coord = NULL_COORDINATE;
        this.clazz = NodeTagResolver.CLAZZ_NULL;
        this.flags = NodeTagResolver.FLAGS_NULL;
        this.name = null;
        this.internFlags = INTERN_FLAG_NULL;
        this.index = 0;
        this.custom = null;
    }

    /**
     * Pruefung, ob dieser Knoten lediglich eine Koordinate ist
     * und somit bei weiteren Schritten platzsparend gespeichert
     * werden kann.
     * @return Nur Koordinate und Id, sonst keine weiteren Infos.
     */
    public boolean isLatLonOnly() {
        return (this.internFlags | INTERN_FLAG_HAS_ID) == INTERN_FLAG_HAS_ID;
    }

    @Override
    public String toString() {
        String s = "Id:" + this.id
            + " Clazz:" + this.clazz + " Flags:" + this.flags
            + " Name:" + (this.name != null ? "'" + this.name + "'" : "null")
            + " InternFlags:" + this.internFlags + " Custom:" + this.custom
            + " LatLon:" + super.toString();
        return s;
    }

    @Override
    public void writeToStream(OutStream outStream) {

        outStream.writeLong(this.coord);
        outStream.writeByte((byte) this.internFlags);

        // Kompression
        if ((INTERN_FLAG_HAS_ID & this.internFlags) != 0)
            VarNum.writeUnsigned(this.id, outStream);
        if ((INTERN_FLAG_HAS_CLAZZ & this.internFlags) != 0)
            outStream.writeByte(this.clazz);
        if ((INTERN_FLAG_HAS_FLAGS & this.internFlags) != 0)
            VarNum.writeUnsigned(this.flags, outStream);
        if ((INTERN_FLAG_HAS_NAME & this.internFlags) != 0)
            this.name.writeToStream(outStream);
        if ((INTERN_FLAG_HAS_CUSTOM & this.internFlags) != 0) {
            @SuppressWarnings("unchecked")
            Class<? extends Var<?>> varClass =
                (Class<? extends Var<?>>) this.custom.getClass();
            outStream.writeByte(typeIdOf(varClass));
            this.custom.writeToStream(outStream);
        }
    }

    @Override
    public OsmNode readFromStream(InStream inStream) {
        this.clear();

        this.coord = inStream.readLong();
        this.internFlags = inStream.readByte() & 0xFF;

        // De-Kompression
        if ((this.internFlags & INTERN_FLAG_HAS_ID) != 0)
            this.id = VarNum.readUnsigned(inStream);
        if ((this.internFlags & INTERN_FLAG_HAS_CLAZZ) != 0)
            this.clazz = inStream.readByte();
        if ((this.internFlags & INTERN_FLAG_HAS_FLAGS) != 0)
            this.flags = (int) VarNum.readUnsigned(inStream);
        if ((this.internFlags & INTERN_FLAG_HAS_NAME) != 0) {
            this.name = new VarString();
            this.name.readFromStream(inStream);
        }
        if ((this.internFlags & INTERN_FLAG_HAS_CUSTOM) != 0) {
            byte varType = inStream.readByte();
            this.custom = newInstanceOf(varType);
            this.custom.readFromStream(inStream);
        }
        return this;
    }

}
