/*
 * $Id: OsmPolygon.java 5243 2021-02-09 21:03:31Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.primitives.VarNum.readUnsigned;
import static de.cm.osm2po.primitives.VarNum.writeUnsigned;
import static de.cm.osm2po.primitives.VarTypeDesk.TYPE_NULL;
import static de.cm.osm2po.primitives.VarTypeDesk.newInstanceOf;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;
import static java.lang.StrictMath.abs;

import java.util.Arrays;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;

/**
 * Ein Multipolygon. Hier normalerweise eine Sammlung von Polygonen, die genau
 * einer Relation (Id) zugeordnet ist.<br>
 * {@link Comparable} ordnet lediglich nach area-size des outer
 * Polygons, wodurch grosse Regionen z.B. Laender vorne stehen.
 *
 * @see OsmPolygonBuilder
 * @author (c) 2015 - Carsten Moeller, info@osm2po.de
 */
public final class OsmPolygon
implements Var<OsmPolygon>, OsmObject, Comparable<OsmPolygon> {
    private long id;
    private byte clazz;
    private int flags;
    private VarString name;
    private Var<?> custom;
    private LatLons[][] polygons;
    private long centroid; // vom Hauptpolygon
    private float areaSize; // vom Hauptpolygon

    private transient int index; // set by find for refetch

    /** CTOR. */
    public OsmPolygon() {}

    /**
     * CTOR.
     * @param id long OsmId der Polygon-Relation
     * @param clazz byte Klasse
     * @param flags byte Flags
     * @param name {@link VarString} Name NULLABLE =&gt; EMPTY
     * @param polygons Array von Polygonen
     *      als {@link LatLons}[] NOT NULL<br>
     *      Index 0 ist jeweils Outer, nachfolgende sind Inners.
     */
    public OsmPolygon(long id, byte clazz, int flags,
            VarString name, LatLons[][] polygons) {
        if (null == name) name = new VarString();
        this.id = id;
        this.clazz = clazz;
        this.flags = flags;
        this.name = name;
        this.polygons = polygons;
        this.calcHints();
    }

    public void clear() {
        this.id = 0;
        this.clazz = 0;
        this.flags = 0;
        this.name = null;
        this.areaSize = 0;
        this.centroid = 0;
        this.index = 0;
        this.polygons = null;
        this.custom = null;
    }

    /**
     * Liefert die Polygone als Array. Jedes Polygon ist
     * wiederum ein Array von {@link LatLons}. Jede Teilkette bildet
     * einen Ring. Index [0] ist der aeussere, alle weiteren
     * die inneren Ringe.
     * @return Array von Array von {@link LatLons}
     */
    public final LatLons[][] getPolygons() {return this.polygons;}

    @Override public long getId() {return this.id;}
    @Override public byte getClazz() {return this.clazz;}
    @Override public VarString getName() {return this.name;}
    @Override public int getFlags() {return this.flags;}
    @Override public Var<?> getCustom() {return this.custom;}

    // Transient ReFetch-Index
    @Override
    public int getIndex() {return this.index;}
    @Override
    public void setIndex(int index) {this.index = index;}
    // calculated hints
    public long getCentroid() {return this.centroid;}
    public float getAreaSize() {return this.areaSize;}
    public void setCustom(Var<?> custom) {this.custom = custom;}

    /**
     * Prueft, ob eine Koordinate innerhalb eines der Polygone liegt.
     * @param lat double Latitude
     * @param lon double Longitude
     * @return true: innerhalb, false: ausserhalb
     */
    public final boolean contains(double lat, double lon) {
        if (this.polygons != null) for (LatLons[] plg : this.polygons) {
            int nRings = null == plg ? 0 : plg.length;
            int i = 0; for (; i < nRings; i++) {
                LatLons ring = plg[i];
                if (0 == i) if (ring.contains(lat, lon)) continue; else break;
                if (ring.contains(lat, lon)) break; // Inner
            }
            if (nRings == i) return true;
        }
        return false;
    }

    @Override
    public String toString() {
        if (null == this.polygons) return "NULL";
        String s = "Id:" + this.id + " Name:'" + this.name + "'"
            + " Clazz:" + this.clazz + " Flags:" + this.flags
            + " Cenroid:" + (new LatLon().setCoord(this.centroid))
            + " AreaSize:" + this.areaSize
            + " Polygons[" + this.polygons.length + "]:";

        for (int i = 0; i < this.polygons.length; i++) {
            LatLons[] poly = this.polygons[i];
            for (int j = 0; j < poly.length; j++) {
                s += "\n[" + i + "]";
                s += (0 == j ? "Outer:" : "Inner" + j + ":") + poly[j];
            }
        }
        return s;
    }

    @Override
    public boolean equals(Object obj) {
        // Name wird hier nicht vergliichen
        if (null == obj) return false;
        if (!(obj instanceof OsmPolygon)) return false;
        OsmPolygon that = (OsmPolygon) obj;
        if (that.id != this.id) return false;
        if (that.clazz != this.clazz) return false;
        if (that.flags != this.flags) return false;
        int nPolygons = this.polygons.length;
        if (that.polygons.length != nPolygons) return false;
        for (int i = 0; i < nPolygons; i++) {
            LatLons[] thatRings = that.polygons[i];
            LatLons[] thisRings = this.polygons[i];
            Arrays.equals(thatRings, thisRings);
        }
        return true;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeByte(this.clazz);
        writeUnsigned(this.flags & 0xffffffffL, outStream);
        writeUnsigned(this.id, outStream);
        if (null == this.name) VarString.EMPTY.writeToStream(outStream);
            else this.name.writeToStream(outStream);

        if (this.custom != null) {
            @SuppressWarnings("unchecked")
            Class<? extends Var<?>> varClass =
                (Class<? extends Var<?>>) this.custom.getClass();
            outStream.writeByte(typeIdOf(varClass));
            this.custom.writeToStream(outStream);
        } else {
            outStream.writeByte(TYPE_NULL);
        }

        int nPolygons = null == this.polygons ? 0 : this.polygons.length;
        writeUnsigned(nPolygons, outStream);
        if (nPolygons > 0) for (LatLons[] poly : this.polygons) {
            writeUnsigned(poly.length, outStream);
            for (LatLons rings : poly) rings.writeToStream(outStream);
        }
        if (nPolygons > 0) {
            outStream.writeLong(this.centroid);
            outStream.writeInt(Float.floatToIntBits(this.areaSize));
        }
    }

    @Override
    public OsmPolygon readFromStream(InStream inStream) {
        this.clazz = inStream.readByte();
        this.flags = (int) readUnsigned(inStream);
        this.id = readUnsigned(inStream);
        this.name = new VarString().readFromStream(inStream); // not null!

        byte varType = inStream.readByte();
        if (varType != TYPE_NULL) {
            this.custom = newInstanceOf(varType);
            this.custom.readFromStream(inStream);
        }

        int nPolygons = (int) readUnsigned(inStream);
        this.polygons = new LatLons[nPolygons][];
        for (int i = 0; i < nPolygons; i++) {
            int nRings = (int) readUnsigned(inStream);
            LatLons[] poly = new LatLons[nRings];
            for (int j = 0; j < nRings; j++)
                poly[j] = new LatLons().readFromStream(inStream);
            this.polygons[i] = poly;
        }
        if (nPolygons > 0) {
            this.centroid = inStream.readLong();
            this.areaSize = Float.intBitsToFloat(inStream.readInt());
        }
        return this;
    }

    /**
     * Berechnet intern die Groesse und den Mittelpunkt
     * des dominierenden Outer-Polygons, falls mehrere vorhanden.
     */
    private void calcHints() {
        this.areaSize = 0f;
        this.centroid = 0l;
        if (this.polygons != null) for (LatLons[] plg : this.polygons) {
            float ringSize = 0f;
            long ringCentroid = 0l;
            int nRings = null == plg ? 0 : plg.length;
            int i = 0; for (; i < nRings; i++) {
                LatLons ring = plg[i];
                if (0 == i) { // Outer
                    ringCentroid = ring.centroid().getCoord();
                    ringSize += abs(ring.area());
                } else { // Inners
                    ringSize -= abs(ring.area());
                }
            }
            if (ringSize > this.areaSize) {
                this.areaSize = ringSize;
                this.centroid = ringCentroid;
            }
        }
    }

    @Override
    public int compareTo(OsmPolygon that) {
        // Grosse Gebiete vorne, kleine so in UI vorne und anklickbar
        if (null == that || this.areaSize < that.areaSize) return 1;
        else if (this.areaSize > that.areaSize) return -1;
        return 0;
    }

}
