/*
 * $Id: OsmPolygonBuilder.java 4549 2017-10-31 09:44:07Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_MASK;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_POLYGON;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_INNER;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_OUTER;
import static de.cm.osm2po.model.OsmRelation.MEMBER_ROLE_UNKNOWN;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.VarString;

/**
 * Sammelt {@link Way}-Fragmente, die Polygon-{@link Membership}s
 * zugeordnet sind, und erstellt daraus {@link OsmPolygon}s.
 *
 * @author (c) 2015 - Carsten Moeller, info@osm2po.de
 */
public class OsmPolygonBuilder implements Iterable<OsmPolygon>{

    private Log log;
    private Map<Long, Polygonizer> polygonizerMap;
    private Iterator<Polygonizer> iterator;

    public OsmPolygonBuilder(Log log) {
        this.log = null == log ? Log.empty() : log;
        this.polygonizerMap = new HashMap<Long, Polygonizer>(0x400);
    }

    public void collectMemberships(Way way) {
        if (null == way.getMemberships()) return;
        for (Membership membership : way.getMemberships()) {

            byte clazz = membership.getClazz();
            if (CLAZZ_BOUNDARY == (CLAZZ_MASK & clazz)
                    || CLAZZ_POLYGON == (CLAZZ_MASK & clazz)) {
                boolean isOuter;
                switch (membership.getRole()) {
                case MEMBER_ROLE_UNKNOWN:
                case MEMBER_ROLE_OUTER: isOuter = true; break;
                case MEMBER_ROLE_INNER: isOuter = false; break;
                default: continue;}

                long id = membership.getId();
                Polygonizer polygonizer = this.polygonizerMap.get(id);
                if (null == polygonizer) {
                    int flags = membership.getFlags();
                    VarString name = membership.getName();
                    polygonizer = new Polygonizer(id, clazz, flags, name);
                    this.polygonizerMap.put(membership.getId(), polygonizer);
                }
                polygonizer.addPart(
                        new LatLons().setCoords(way.getNodes()), isOuter);
            }
        }
    }

    private final class Polygonizer {
        private long id;
        private byte clazz;
        private int flags;
        private VarString name;
        private LatLons[] outerWays;
        private int nOuterWays;
        private LatLons[] innerWays;
        private int nInnerWays;

        private Polygonizer(long id, byte clazz, int flags, VarString name) {
            this.id = id;
            this.clazz = clazz;
            this.flags = flags;
            this.name = name;
            this.outerWays = null;
            this.nOuterWays = 0;
            this.innerWays = null;
            this.nInnerWays = 0;
        }

        private final void addPart(LatLons part, boolean isOuter) {
            if (null == part) return;
            if (isOuter) {
                if (null == this.outerWays) this.outerWays = new LatLons[8];
                if (this.nOuterWays == this.outerWays.length)
                    this.outerWays = Arrays.copyOf(this.outerWays, this.nOuterWays << 1);
                this.outerWays[this.nOuterWays++] = part;
            } else {
                if (null == this.innerWays) this.innerWays = new LatLons[8];
                if (this.nInnerWays == this.innerWays.length)
                    this.innerWays = Arrays.copyOf(this.innerWays, this.nInnerWays << 1);
                this.innerWays[this.nInnerWays++] = part;
            }
        }

        private OsmPolygon polygonize() {
            List<LatLons[]> polygonList = new ArrayList<LatLons[]>();

            List<LatLons> outerRingList = //NULL-safe
                this.ringify(this.outerWays, this.nOuterWays);
            List<LatLons> innerRingList =
                this.ringify(this.innerWays, this.nInnerWays);

            for (LatLons outerRing : outerRingList) {
                if (outerRing.area() < 0) {
                    outerRing.reverse();
                }

                LatLons[] rings = new LatLons[]{outerRing};

                if (!innerRingList.isEmpty()) {
                    // TODO Teure Operation, da Bounding alle Koordinaten nochmals kopiert
                    Bounding bounding = new Bounding(outerRing);

                    for (LatLons innerRing : innerRingList) {
                        if (innerRing.area() < 0) {
                            innerRing.reverse();
                        }
                        if (bounding.contains(innerRing)) {
                            int size = rings.length;
                            rings = Arrays.copyOf(rings, size +1);
                            rings[size] = innerRing;
                        }
                    }
                }
                polygonList.add(rings);
            }
            // ClearMem
            this.outerWays = null;
            this.nOuterWays = 0;
            this.innerWays = null;
            this.nInnerWays = 0;

            // Transformiere nach Array statt Liste
            int n = 0;
            LatLons[][] osmMultiPoly = new LatLons[polygonList.size()][];
            for (LatLons[] poly : polygonList) osmMultiPoly[n++] = poly;

            return new OsmPolygon(
                    this.id, this.clazz, this.flags, this.name, osmMultiPoly);
        }

        private List<LatLons> ringify(LatLons[] segs, int n) {
            List<LatLons> returnList = new ArrayList<LatLons>();
            LatLons line = new LatLons();

            int nChained = 0;
            for (int j = 0; j < n; j++) {
                for (int i = j; i < n; i++) {
                    if (line.concat(segs[i], true)) {
                        nChained++;
                        LatLons tmp = segs[i]; segs[i] = segs[j]; segs[j] = tmp;
                        int ringBeginIdx = line.findLastRing();
                        // TODO Increase performance by checking concat-points only
                        if (ringBeginIdx > 0) { // Invalid Polygon! Self-Intersection!
                            LatLons ring = line.crop(ringBeginIdx);
                            returnList.add(ring);
                            line.append(ring.getCoord(0));

                        } else if (ringBeginIdx == 0) { // Full valid Ring
                            returnList.add(line);
                            line = new LatLons();
                        }

                        break;
                    }
                }
            }

            int nMissing = n - nChained;
            if (nMissing > 0) {
                if (log.isEnabled(Log.LEVEL_DATA)) log.data(
                        "Relation " + this.id + " (" + this.clazz + " " + this.name
                        + ") " + nMissing + " ways missing");
            }

            return returnList;
        }

    }

    @Override
    public Iterator<OsmPolygon> iterator() {
        return new Iterator<OsmPolygon>() {
            @Override
            public boolean hasNext() {
                if (null == iterator) iterator = // Open Iterator
                    polygonizerMap.values().iterator();
                if (!iterator.hasNext()) {
                    iterator = null; // Close Iterator
                    return false;
                } else {
                    return true;
                }
            }
            @Override
            public OsmPolygon next() {
                Polygonizer polygonizer = iterator.next();
                OsmPolygon polygons = polygonizer.polygonize();
                return polygons;
            }
            @Override
            public void remove() {
                throw new UnsupportedOperationException(
                        "Iterator.remove() not supported");
            }
        };
    }

}
