/*
 * $Id: OsmRelation.java 5258 2021-04-03 16:20:36Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.converter.tiler.TagResolver.CLAZZ_NULL;
import static de.cm.osm2po.converter.tiler.TagResolver.FLAGS_NULL;
import static de.cm.osm2po.primitives.VarTypeDesk.newInstanceOf;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;

import java.util.ArrayList;
import java.util.List;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarNum;
import de.cm.osm2po.primitives.VarString;

/**
 * Repraesentiert eine OSM-Relation.
 * @see OsmWay
 * @see OsmNode
 * @see Membership
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class OsmRelation implements Var<OsmRelation>, OsmObject {

    private static final int INTERN_FLAG_HASCLAZZ = 0x1;
    private static final int INTERN_FLAG_HASFLAGS = 0x2;
    private static final int INTERN_FLAG_HASNAME = 0x4;
    private static final int INTERN_FLAG_INVALID = 0x8;
    private static final int INTERN_FLAG_HASCUSTOM = 0x10;

    public final static byte MEMBER_TYPE_UNKNOWN = 0;
    public final static byte MEMBER_TYPE_NODE = 1;
    public final static byte MEMBER_TYPE_WAY = 2;
    public final static byte MEMBER_TYPE_RELATION = 3;

    public final static byte MEMBER_ROLE_UNKNOWN = 0;
    public final static byte MEMBER_ROLE_FROM = 1;
    public final static byte MEMBER_ROLE_TO = 2;
    public final static byte MEMBER_ROLE_VIA = 3;
    public final static byte MEMBER_ROLE_FORWARD = 4;
    public final static byte MEMBER_ROLE_BACKWARD = 5;
    public final static byte MEMBER_ROLE_OUTER = 6;
    public final static byte MEMBER_ROLE_INNER = 7;

    /**
     * Inner Class fuer ein Member einer Relation.
     * @author (c) 2012 - Carsten Moeller - info@osm2po
     */
    public static class Member implements Var<Member> {
        private byte type;
        private long ref;
        private byte role;

        private Member(){};

        public byte getType() {return this.type;}
        public long getRef() {return this.ref;}
        public byte getRole() {return this.role;}

        private Member(byte type, long ref, byte role) {
            this.type = type;
            this.ref = ref;
            this.role = role;
        }

        @Override
        public String toString() {
            return "Type:" + this.type
                + " Ref:" + this.ref + " Role:" + this.role;
        }

        @Override
        public Member readFromStream(InStream inStream) {
            this.type = inStream.readByte();
            this.ref = inStream.readLong();
            this.role = inStream.readByte();
            return this;
        }
        @Override
        public void writeToStream(OutStream outStream) {
            outStream.writeByte(this.type);
            outStream.writeLong(this.ref);
            outStream.writeByte(this.role);
        }
    }

    private long id;
    private byte clazz;
    private int flags;
    private boolean invalid;
    private VarString name;
    private Var<?> custom;
    private int index;

    private List<Member> members = new ArrayList<Member>();

    @Override public long getId() {return this.id;}
    @Override public byte getClazz() {return this.clazz;}
    @Override public VarString getName() {return this.name;};
    @Override public int getFlags() {return this.flags;}
    @Override public Var<?> getCustom() {return this.custom;}
    @Override public int getIndex() {return index;}
    @Override public void setIndex(int index) {this.index = index;}

    public void setId(long id) {this.id = id;}
    public void setClazz(byte clazz) {this.clazz = clazz;}
    public void setFlags(int flags) {this.flags = flags;}
    public void setName(VarString name) {this.name = name;}
    public void setCustom(Var<?> custom) {this.custom = custom;}

    /** @return List of {@link Member} NOT NULL */
    public List<Member> getMembers() {return this.members;}

    /**
     * Fuegt der Relation einen Member hinzu.
     * @param type byte, gueltige Werte sind: {@link #MEMBER_TYPE_NODE},
     *        {@link #MEMBER_TYPE_WAY}, {@link #MEMBER_TYPE_RELATION}
     *        oder {@link #MEMBER_TYPE_UNKNOWN}, letzteres eigentlich nicht.
     * @param refId long Ref-Id auf Node, Way oder Relation (Foreign Key)
     * @param role byte z.B. {@link #MEMBER_ROLE_FROM}, {@link #MEMBER_ROLE_TO},
     *        {@link #MEMBER_ROLE_VIA}, {@link #MEMBER_ROLE_UNKNOWN}
     *        - aber auch frei definierbar.
     */
    protected void addMember(byte type, long refId, byte role) {
        this.members.add(new Member(type, refId, role));
    }

    /**
     * Ruft {@link #addMember(byte, long, byte)}, uebersetzt allerdings
     * Member und Role in die intern verwendeten Konstanten.
     * Wird keine Konstante gefunden, so wird entsprechen auf
     * {@link #MEMBER_TYPE_UNKNOWN} bzw {@link #MEMBER_ROLE_UNKNOWN}
     * gesetzt.
     * @param strType String wie vom Parser geliefert.
     * @param lngRef long Ref-Id auf Node, Way oder Relation.
     * @param strRole String wie vom Parser geliefert.
     */
    public void addMember(String strType, long lngRef, String strRole) {
        byte memberType = OsmRelation.MEMBER_TYPE_UNKNOWN;
        if ("node".equals(strType))
            memberType = OsmRelation.MEMBER_TYPE_NODE;
        else if ("way".equals(strType))
            memberType = OsmRelation.MEMBER_TYPE_WAY;
        else if ("relation".equals(strType))
            memberType = OsmRelation.MEMBER_TYPE_RELATION;

        byte memberRole = OsmRelation.MEMBER_ROLE_UNKNOWN;
        if ("from".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_FROM;
        else if ("to".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_TO;
        else if ("via".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_VIA;
        else if ("forward".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_FORWARD;
        else if ("backward".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_BACKWARD;
        else if ("outer".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_OUTER;
        else if ("inner".equals(strRole))
            memberRole = OsmRelation.MEMBER_ROLE_INNER;

        this.addMember(memberType, lngRef, memberRole);
    }

    /** CTOR. */
    public OsmRelation() {
        this.clear();
    }

    public boolean isInvalid() {
        return this.invalid;
    }

    public void setInvalid(boolean invalid) {
        this.invalid = invalid;
    }


    /** Reinitialisiert. */
    public void clear() {
        this.id = 0;
        this.clazz = CLAZZ_NULL;
        this.flags = FLAGS_NULL;
        this.name = null;
        this.invalid = false;
        this.members.clear();
        this.custom = null;
    }

    @Override
    public String toString() {
        String r = "Id:" + this.id
            + " Name:'" + this.name + "'"
            + " Clazz:" + this.clazz + " Flags:" + this.flags
            + " Valid:" + !this.invalid
            + " Members[" + this.members.size() + "]:";
        for (int i = 0; i < this.members.size(); i++) {
            r += "\n" + this.members.get(i).toString();
        }
        return r;
    }

    public void writeToStreamWithoutMembers(OutStream outStream) {
        // Kompression
        int internFlags = 0;
        if (this.invalid) internFlags |= INTERN_FLAG_INVALID;
        boolean hasClazz = this.clazz != 0;
        if (hasClazz) internFlags |= INTERN_FLAG_HASCLAZZ;
        boolean hasFlags = this.flags != 0;
        if (hasFlags) internFlags |= INTERN_FLAG_HASFLAGS;
        boolean hasName = this.name != null && !this.name.isEmpty();
        if (hasName) internFlags |= INTERN_FLAG_HASNAME;
        boolean hasCustom = this.custom != null;
        if (hasCustom) internFlags |= INTERN_FLAG_HASCUSTOM;

        VarNum.writeUnsigned(this.id, outStream);
        outStream.writeByte((byte) internFlags);

        if (hasClazz) outStream.writeByte(this.clazz);
        if (hasFlags) VarNum.writeUnsigned(this.flags, outStream);
        if (hasName) this.name.writeToStream(outStream);
        if (hasCustom) {
            @SuppressWarnings("unchecked")
            Class<? extends Var<?>> varClass =
                (Class<? extends Var<?>>) this.custom.getClass();
            outStream.writeByte(typeIdOf(varClass));
            this.custom.writeToStream(outStream);
        }
    }

    public OsmRelation readFromStreamWithoutMembers(InStream inStream) {
        this.clear();
        this.id = VarNum.readUnsigned(inStream);

        // Kompression
        int internFlags = inStream.readByte() & 0xFF;
        this.invalid = (internFlags & INTERN_FLAG_INVALID) != 0;
        if ((internFlags & INTERN_FLAG_HASCLAZZ) != 0)
            this.clazz = inStream.readByte();
        if ((internFlags & INTERN_FLAG_HASFLAGS) != 0)
            this.flags = (int) VarNum.readUnsigned(inStream);
        if ((internFlags & INTERN_FLAG_HASNAME) != 0) {
            this.name = new VarString();
            this.name.readFromStream(inStream);
        }
        if ((internFlags & INTERN_FLAG_HASCUSTOM) != 0) {
            byte varType = inStream.readByte();
            this.custom = newInstanceOf(varType);
            this.custom.readFromStream(inStream);
        }

        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        this.writeToStreamWithoutMembers(outStream);

        int membersSize = this.members.size();
        outStream.writeInt(membersSize);
        for (int i = 0; i < membersSize; i++) {
            Member member = this.members.get(i);
            member.writeToStream(outStream);
        }
    }

    @Override
    public OsmRelation readFromStream(InStream inStream) {
        this.readFromStreamWithoutMembers(inStream);

        int membersSize = inStream.readInt();
        for (int i = 0; i < membersSize; i++) {
            this.members.add(new Member().readFromStream(inStream));
        }
        return this;
    }

}
