/*
 * $Id: OsmWay.java 5243 2021-02-09 21:03:31Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.converter.tiler.TagResolver.CLAZZ_NULL;
import static de.cm.osm2po.primitives.VarTypeDesk.TYPE_NULL;
import static de.cm.osm2po.primitives.VarTypeDesk.newInstanceOf;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;

import java.util.Arrays;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarNum;
import de.cm.osm2po.primitives.VarString;

/**
 * Repraesentiert einen OSM-Way.
 * @see OsmNode
 * @see OsmRelation
 * @see Way
 * @see SegmentedWay
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class OsmWay implements Var<OsmWay>, OsmObject {

    // InternalFlags sind VarNum-serialisiert. Werte daher aufsteigend vergeben
    public static final int INTERN_FLAG_NULL = 0x0;
    public static final int INTERN_FLAG_INVALID = 0x1;
    public static final int INTERN_FLAG_ONEWAY = 0x2;
    public static final int INTERN_FLAG_REVERSED = 0x4;
    public static final int INTERN_FLAG_RAIL = 0x8;
    public static final int INTERN_FLAG_ROUNDABOUT = 0x10;
    public static final int INTERN_FLAG_ROUTEMEMBER = 0x20;
    public static final int INTERN_FLAG_POLYGON = 0x40;

    private int internFlags; // siehe oben
    private long id;
    private byte clazz;
    private int flags;
    private VarString name;
    private int kmh;
    private Var<?> custom;
    private int index;

    private static final int INITIAL_REF_ARRAY_SIZE = 10;
    private long[] refs;
    private int nRefs;

    /** CTOR. */
    public OsmWay() {clear();}

    public void clear() {
        this.id = 0;
        this.name = new VarString();
        this.custom = null;
        this.clazz = CLAZZ_NULL;
        this.flags = INTERN_FLAG_NULL;
        this.internFlags = INTERN_FLAG_NULL;
        this.kmh = 0;
        this.clearRefs();
    }

    @Override public long getId() {return this.id;}
    @Override public byte getClazz() {return this.clazz;}
    @Override public VarString getName() {return this.name;};
    @Override public int getFlags() {return this.flags;}
    @Override public Var<?> getCustom() {return this.custom;}
    @Override public int getIndex() {return index;}
    @Override public void setIndex(int index) {this.index = index;}

    public int getRefsSize() {return this.nRefs;}
    public long getRef(int idx) {return this.refs[idx];}
    public void setClazz(byte clazz) {this.clazz = clazz;}
    public void setId(long id) {this.id = id;}
    public int getKmh() {return this.kmh;}
    public void setKmh(int kmh) {this.kmh = kmh;}
    public void setName(VarString name) {this.name = name;}
    public void setCustom(Var<?> custom) {this.custom = custom;}
    public void clearRefs() {this.refs = null; this.nRefs = 0;}

    public void setFlags(int flags) {this.flags = flags;}
    public int getInternFlags() {return this.internFlags;}

    public OsmWay addRef(long refId) throws IllegalArgumentException {
        if (null == this.refs) {
            this.refs = new long[INITIAL_REF_ARRAY_SIZE];
        }
        if (this.nRefs > 0 && refId == this.refs[this.nRefs -1]) {
            throw new IllegalArgumentException("Redundant Ref " + refId
                    + " in way " + this.id);
        }
        if (this.nRefs == this.refs.length) {
            // Index-Range-Check
            this.refs = Arrays.copyOf(this.refs, this.refs.length * 2);
        }
        this.refs[this.nRefs++] = refId;
        return this;
    }

    public boolean isInvalid() {
        return (this.internFlags & INTERN_FLAG_INVALID) != 0;
    }

    public void setInvalid(boolean isInvalid) {
        if (isInvalid) this.internFlags |= INTERN_FLAG_INVALID;
        else this.internFlags &=~ INTERN_FLAG_INVALID;
    }

    public boolean isOneWay() {
        return (this.internFlags & INTERN_FLAG_ONEWAY) != 0;
    }

    public void setOneWay(boolean isOneWay) {
        if (isOneWay) this.internFlags |= INTERN_FLAG_ONEWAY;
        else this.internFlags &=~ INTERN_FLAG_ONEWAY;
    }

    public boolean isRail() {
        return (this.internFlags & INTERN_FLAG_RAIL) != 0;
    }

    public void setRail(boolean isRail) {
        if (isRail) this.internFlags |= INTERN_FLAG_RAIL;
        else this.internFlags &=~ INTERN_FLAG_RAIL;
    }

    public boolean isRoundabout() {
        return (this.internFlags & INTERN_FLAG_ROUNDABOUT) != 0;
    }

    public void setRoundabout(boolean isRoundabout) {
        if (isRoundabout) this.internFlags |= INTERN_FLAG_ROUNDABOUT;
        else this.internFlags &=~ INTERN_FLAG_ROUNDABOUT;
    }

    public boolean isRouteMember() {
        return (this.internFlags & INTERN_FLAG_ROUTEMEMBER) != 0;
    }

    public void setRouteMember(boolean isRouteMember) {
        if (isRouteMember) this.internFlags |= INTERN_FLAG_ROUTEMEMBER;
        else this.internFlags &=~ INTERN_FLAG_ROUTEMEMBER;
    }

    public boolean isPolygon() {
        return (this.internFlags & INTERN_FLAG_POLYGON) != 0;
    }

    public void setPolygon(boolean isPolygon) {
        if (isPolygon) this.internFlags |= INTERN_FLAG_POLYGON;
        else this.internFlags &=~ INTERN_FLAG_POLYGON;
    }

    public boolean isRing() {
        return this.nRefs > 2 && this.refs[0] == this.refs[this.nRefs -1];
    }

    public final void reverse() {
        this.internFlags ^= INTERN_FLAG_REVERSED;
        for (int x = 0, y = this.nRefs -1; x < y; ++x, --y) {
            long tmp = this.refs[x];
            this.refs[x] = this.refs[y];
            this.refs[y] = tmp;
        }
    }

    @Override
    public String toString() {
        String w = "Id:" + this.id
            + " Name:'" + this.name + "'"
            + " Clazz:" + this.clazz
            + " Flags:" + this.flags
            + " Kmh:" + this.kmh;
        if (this.nRefs > 0) {
            w += "\nRefs[" + this.nRefs + "]:";
            for (int i = 0; i < this.nRefs; i++) { // Null-sicher!!
                if (i > 0) w += ",";
                w += this.refs[i];
            }
        }
        return w;
    }


    @Override
    public OsmWay readFromStream(InStream inStream) {
        this.clear(); // ruft ggf. clear() abgeleiteter Klassen

        this.id = inStream.readLong();
        this.clazz = inStream.readByte();

        this.flags = (int) VarNum.readUnsigned(inStream);
        this.internFlags = (int) VarNum.readUnsigned(inStream);
        this.kmh = (int) VarNum.readUnsigned(inStream);

        this.name.readFromStream(inStream);

        byte varType = inStream.readByte();
        if (varType != TYPE_NULL) {
            this.custom = newInstanceOf(varType);
            this.custom.readFromStream(inStream);
        }

        this.nRefs = (int) VarNum.readUnsigned(inStream);
        this.refs = new long[this.nRefs];
        for (int i = 0; i < this.nRefs; i++) {
            this.refs[i] = inStream.readLong();
        }
        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeLong(this.id);
        outStream.writeByte(this.clazz);

        VarNum.writeUnsigned(this.flags, outStream);
        VarNum.writeUnsigned(this.internFlags, outStream);
        VarNum.writeUnsigned(this.kmh, outStream);

        this.name.writeToStream(outStream);

        if (this.custom != null) {
            @SuppressWarnings("unchecked")
            Class<? extends Var<?>> varClass =
                (Class<? extends Var<?>>) this.custom.getClass();
            outStream.writeByte(typeIdOf(varClass));
            this.custom.writeToStream(outStream);
        } else {
            outStream.writeByte(TYPE_NULL);
        }

        VarNum.writeUnsigned(this.nRefs, outStream);
        for (int i = 0; i < this.nRefs; i++) {
            outStream.writeLong(this.refs[i]);
        }
    }

}

