/*
 * $Id: Restriction.java 4423 2017-04-16 08:26:55Z carsten $
 *
 * Copyright 2014 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_ENTRY;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_EXIT;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_LEFT_TURN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_RIGHT_TURN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_STRAIGHT_ON;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_NO_U_TURN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_ONLY_LEFT_TURN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_ONLY_RIGHT_TURN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_RESTRICTION_ONLY_STRAIGHT_ON;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;

/**
 * Diese Klasse spiegelt eine Restriction-{@link Membership}
 * in der Rolle "Via" eines {@link Node} Objekts,
 * wenn vorhanden, wider. Sie ist initial mit 0-en besetzt und
 * wird erst nach der Segmentierung mit echten Ids versorgt.
 * Da Objekte dieser Klasse Kinder des {@link Vertex}
 * sind, wird die Via-Id hier nicht gespeichert, entspricht
 * jedoch der Id des Vertex.
 */

public class Restriction implements Var<Restriction> {
    
    private long id;
    private int from;
    private int to;
    private byte clazz;

    public Restriction(){} // 4Var
    public Restriction(long id, byte clazz, int from, int to) {
        this.id = id; // OsmId aus Relation
        this.clazz = clazz;
        this.from = from;
        this.to = to;
    }
    
    public void setId(long id) {this.id = id;}
    public long getId() {return this.id;}
    public void setClazz(byte clazz) {this.clazz = clazz;}
    public byte getClazz() {return this.clazz;}
    public void setFrom(int from) {this.from = from;}
    public int getFrom() {return this.from;}
    public void setTo(int to) {this.to = to;}
    public int getTo() {return this.to;}
    
    @Override
    public String toString() {
        return "Id:" + this.id 
            + " From:" + this.from + " To:" + this.to
            + " Clazz:" + this.clazz 
            + "(" + clazzToString(clazz) + ")";
    }
    
    /**
     * Liefert eine lesbare und parsbare Darstellung der Clazz.
     * @param clazz Restriction-Type
     * @return String-Darstellung.
     */
    public static String clazzToString(byte clazz) {
        if (CLAZZ_RESTRICTION_NO_LEFT_TURN == clazz) return "NLT";
        if (CLAZZ_RESTRICTION_NO_RIGHT_TURN == clazz) return "NRT";
        if (CLAZZ_RESTRICTION_NO_STRAIGHT_ON == clazz) return "NSO";
        if (CLAZZ_RESTRICTION_NO_U_TURN == clazz) return "NUT";
        if (CLAZZ_RESTRICTION_NO_ENTRY == clazz) return "NNY";
        if (CLAZZ_RESTRICTION_NO_EXIT == clazz) return "NXT";
        if (CLAZZ_RESTRICTION_ONLY_LEFT_TURN == clazz) return "OLT";
        if (CLAZZ_RESTRICTION_ONLY_RIGHT_TURN == clazz) return "ORT";
        if (CLAZZ_RESTRICTION_ONLY_STRAIGHT_ON == clazz) return "OSO";
        return "";
    }
    
    @Override
    public Restriction readFromStream(InStream inStream) {
        // Wichtig! Keine variablen Laengen!
        this.id = inStream.readLong();
        this.from = inStream.readInt();
        this.to = inStream.readInt();
        this.clazz = inStream.readByte();
        return this;
    }
    
    @Override
    public void writeToStream(OutStream outStream) {
        // Wichtig! Keine variablen Laengen!
        outStream.writeLong(this.id);
        outStream.writeInt(this.from);
        outStream.writeInt(this.to);
        outStream.writeByte(this.clazz);
    }

}
