/*
 * $Id: SegmentedWay.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.VarNum;

/**
 * Abgeleitet von {@link Way}, werden hier die nach der Segmentierung
 * ermittelten Teilstuecke {@link WaySegment} vermerkt.
 * @see OsmWay
 * @see Way
 * @author (c) 2012, 2016 - Carsten Moeller - info@osm2po
 */
public class SegmentedWay extends Way {

    protected WaySegment[] waySegments;

    public void setSegments(WaySegment[] segments) {
        this.waySegments = segments;
    }
    
    public WaySegment[] getSegments() {
        return this.waySegments;
    }
    
    @Override
    public void clear() {
        super.clear();
        this.waySegments = null;
    }
    
    /**
     * Upcast-Stream-Reader von {@link Way} nach {@link SegmentedWay}.
     * @param inStream {@link InStream}
     * @return {@link SegmentedWay} <code>this</code>
     */
    public SegmentedWay readFromWayStream(InStream inStream) {
        super.readFromStream(inStream); // ruft clear() dieser Klasse
        return this;
    }
    
    @Override
    public void writeToStream(OutStream outStream) {
        super.writeToStream(outStream);
        int nSegments = null == this.waySegments ? 0 : this.waySegments.length;
        VarNum.writeUnsigned(nSegments, outStream);
        for (int i = 0; i < nSegments; i++) {
            this.waySegments[i].writeToStream(outStream);
        }
    }
    
    @Override
    public SegmentedWay readFromStream(InStream inStream) {
        super.readFromStream(inStream); // Ruft clear() dieser Klasse
        int nSegments = (int) VarNum.readUnsigned(inStream);
        this.waySegments = new WaySegment[nSegments];
        for (int i = 0; i < nSegments; i++) {
            this.waySegments[i] = new WaySegment();
            this.waySegments[i].readFromStream(inStream);
        }
        return this;
    }
    
    @Override
    public String toString() {
        String s = super.toString();
        int nSegments = 0;
        if (this.waySegments != null) {
            nSegments = this.waySegments.length;
            s += "\nSegments[" + nSegments + "]";
            for (int i = 0; i < nSegments; i++) {
                s += " (" + this.waySegments[i].toString() + ")";
            }
        }
        return s;
    }
    
    public Node getSourceNode(int segmentIdx) {
        return this.getNodes()[this.waySegments[segmentIdx].getSourceIdx()];
    }

    public Node getTargetNode(int segmentIdx) {
        return this.getNodes()[this.waySegments[segmentIdx].getTargetIdx()];
    }
    
    public LatLons getSegmentGeometry(int segmentIdx) {
        int sx = this.waySegments[segmentIdx].getSourceIdx();
        int tx = this.waySegments[segmentIdx].getTargetIdx();
        int n = tx - sx +1;
        
        long[] coords = new long[tx - sx +1];
        for (int i = 0; i < n; i++) {
            coords[i] = this.getNodes()[i + sx].coord; 
        }
        return new LatLons().setCoords(coords);
    }
    

}
