/*
 * $Id: Vertex.java 5186 2020-10-28 20:09:31Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.primitives.VarTypeDesk.newInstanceOf;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;

import de.cm.osm2po.converter.Segmenter;
import de.cm.osm2po.converter.segmenter.VertexHashMap;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;

/**
 * Vertex zur Ermittlung von Kreuzungspunkten innerhalb der Topologie.<br>
 * Im Kern werden hier aehnliche Attribute wie in {@link Node} abgebildet.
 * Jedoch kann aus Speichergruenden meistens auf long-IDs verzichtet werden
 * und Koordinaten und OsmIds spielen hier ueberhaupt keine Rolle, werden
 * aus praktischen Gruenden jedoch aufgenommen. Ferner wird hier auch das
 * ReferenceCounting realisiert, was im {@link Node} nichts zu suchen hat.
 * Die Restriktionen werden hier ebenfalls hineingespiegelt, jedoch ebenfalls
 * unter der Verwendung kleinerer IDs.<br> Initial hat ein Vertex immer die
 * Id 0. Erst die Erkennung von Start und Ziel bzw. Kreuzungen setzt selbige.
 * Vertices mit einer 0-Id werden abschliessend im {@link Segmenter} wieder
 * verworfen. <br><b>Hinweis:</b><br>
 * Diese Klasse ruft {@link VertexHashMap#override(Vertex)} obwohl sie zwei
 * Members mit dynamischer Laenge besitzt. Hierfuer wurde eine zusaetzliche
 * Absicherung in {@link #setRestrictions(Restriction[])} und
 * {@link #setOsmName(VarString)} eingebaut, so dass dies nur bei der initialen
 * Befuellung erlaubt ist.
 * @see Node
 * @see Membership
 * @see Restriction
 * @author (c) 2014 - Carsten Moeller - info@osm2po
 */
public class Vertex extends LatLon {
    private static final int INTERN_FLAG_NULL = 0;
    private static final int INTERN_FLAG_HASOSMID = 1;
    private static final int INTERN_FLAG_HASCLAZZ = 2;
    private static final int INTERN_FLAG_HASNAME = 4;
    private static final int INTERN_FLAG_HASCOORD = 8;
    private static final int INTERN_FLAG_HASRESTRICTIONS = 16;
    private static final int INTERN_FLAG_HASCUSTOM = 32;

    /** Interne Verwaltung fuer {@link VertexHashMap} */
    private transient long cacheIndex;
    /** Fuer {@link VertexHashMap}-Entscheidung im {@link Segmenter} */
    private transient boolean isShared;
    /** Absicherung, da Verwendung in {@link VertexHashMap#override(Vertex)} */
    private transient boolean isConst;

    private int id;
    private int refCounter;
    private Restriction[] restrictions;
    private long osmId;
    private byte clazz;
    private VarString osmName;
    private Var<?> custom;

    public void clear() {
        this.id = 0;
        this.refCounter = 0;
        this.restrictions = null;
        this.osmId = 0;
        this.clazz = 0;
        this.osmName = null;
        this.custom = null;
        this.coord = NULL_COORDINATE;
        this.isConst = false;
        this.isShared = false;
        this.cacheIndex = 0;
    }

    public final void setId(int id) {this.id = id;}
    public final int getId() {return this.id;}
    public final int getRefCounter() {return this.refCounter;}
    public final int incRefCounter() {return ++this.refCounter;}
    public final void setOsmId(long osmId) {this.osmId = osmId;}
    public final long getOsmId() {return this.osmId;}
    public final byte getClazz() {return this.clazz;}
    public final void setClazz(byte clazz) {this.clazz = clazz;}
    public final VarString getOsmName() {return this.osmName;}
    public final Var<?> getCustom() {return this.custom;}

    /**
     * @throws IllegalStateException <br>
     *      wenn bereits eine Serialisierung stattgefunden hat.<br>
     *      Der Grund hierfuer ist die Verwendung mit
     *      {@link VertexHashMap#override(Vertex)}
     */
    private void assertNonConstState() {
        if (this.isConst) throw new IllegalStateException(
                "Vertex-Restrictions cannot be reset after being used in Serialization.");
    }

    /**
     * @param osmName {@link VarString}
     * @throws IllegalStateException <br>
     *      wenn bereits eine Serialisierung stattgefunden hat.<br>
     *      Der Grund hierfuer ist die Verwendung mit
     *      {@link VertexHashMap#override(Vertex)}
     */
    public void setOsmName(VarString osmName) {
        this.assertNonConstState();
        this.osmName = osmName;
    }

    /**
     * @param custom {@link Var}
     * @throws IllegalStateException <br>
     *      wenn bereits eine Serialisierung stattgefunden hat.<br>
     *      Der Grund hierfuer ist die Verwendung mit
     *      {@link VertexHashMap#override(Vertex)}
     */
    public void setCustom(Var<?> custom) {
        this.assertNonConstState();
        this.custom = custom;
    }

    /**
     * @param restrictions Array von {@link Restriction} NULLABLE
     * @throws IllegalStateException <br>
     *      wenn bereits eine Serialisierung stattgefunden hat.<br>
     *      Der Grund hierfuer ist die Verwendung mit
     *      {@link VertexHashMap#override(Vertex)}
     */
    public void setRestrictions(Restriction[] restrictions) {
       this.assertNonConstState();
       this.restrictions = restrictions;
    }
    public Restriction[] getRestrictions() {
        return this.restrictions;}

    public boolean isShared() {return this.isShared;}
    public void setShared(boolean isShared) {this.isShared = isShared;}

    public long getReadIndex() {
        return this.cacheIndex;
    }

    public void setReadIndex(long index) {
        this.cacheIndex = index;
    }

    @Override
    public Vertex readFromStream(InStream inStream) {
        this.clear();
        this.isConst = true;

        this.id = inStream.readInt();
        this.refCounter = inStream.readByte();
        int flags = inStream.readByte();

        // Kompression
        boolean hasOsmId = false;
        boolean hasClazz = false;
        boolean hasName = false;
        boolean hasCustom = false;
        boolean hasCoord = false;
        boolean hasRestrictions = false;

        if (flags != INTERN_FLAG_NULL) {
            hasOsmId = (flags & INTERN_FLAG_HASOSMID) != 0;
            hasClazz = (flags & INTERN_FLAG_HASCLAZZ) != 0;
            hasName = (flags & INTERN_FLAG_HASNAME) != 0;
            hasCoord = (flags & INTERN_FLAG_HASCOORD) != 0;
            hasRestrictions = (flags & INTERN_FLAG_HASRESTRICTIONS) != 0;
            hasCustom = (flags & INTERN_FLAG_HASCUSTOM) != 0;
        }

        if (hasOsmId) this.osmId = inStream.readLong();
        if (hasClazz) this.clazz = inStream.readByte();
        if (hasName)  this.osmName = new VarString().readFromStream(inStream);
        if (hasCoord) this.coord = inStream.readLong();

        if (hasRestrictions) {
            int nVnr = inStream.readByte();
            this.restrictions = new Restriction[nVnr];
            for (int i = 0; i < nVnr; i++) {
                this.restrictions[i] = new Restriction();
                this.restrictions[i].readFromStream(inStream);
            }
        }

        if (hasCustom) {
            byte varType = inStream.readByte();
            this.custom = newInstanceOf(varType);
            this.custom.readFromStream(inStream);
        }

        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        this.isConst = true;

        int flags = INTERN_FLAG_NULL;
        boolean hasOsmId = false;
        boolean hasClazz = false;
        boolean hasName = false;
        boolean hasCoord = false;
        boolean hasRestrictions = false;
        boolean hasCustom = false;

        // Kompression
        if (this.osmId != 0) {
            hasOsmId = true;
            flags |= INTERN_FLAG_HASOSMID;
        }
        if (this.clazz != 0) {
            hasClazz = true;
            flags |= INTERN_FLAG_HASCLAZZ;
        }
        if (this.osmName != null) {
            hasName = true;
            flags |= INTERN_FLAG_HASNAME;
        }
        if (this.coord != NULL_COORDINATE) {
            hasCoord = true;
            flags |= INTERN_FLAG_HASCOORD;
        }
        if (this.restrictions != null) {
            hasRestrictions = true;
            flags |= INTERN_FLAG_HASRESTRICTIONS;
        }
        if (this.custom != null) {
            hasCustom = true;
            flags |= INTERN_FLAG_HASCUSTOM;
        }

        outStream.writeInt(this.id);
        outStream.writeByte((byte) this.refCounter);
        outStream.writeByte((byte) flags);

        if (hasOsmId) outStream.writeLong(this.osmId);
        if (hasClazz) outStream.writeByte(this.clazz);
        if (hasName) this.osmName.writeToStream(outStream);
        if (hasCoord) outStream.writeLong(this.coord);

        if (hasRestrictions) {
            int nVnr = this.restrictions.length;
            outStream.writeByte((byte) nVnr);
            for (int i = 0; i < nVnr; i++) {
                this.restrictions[i].writeToStream(outStream);
            }
        }

        if (hasCustom) {
            @SuppressWarnings("unchecked")
            Class<? extends Var<?>> varClass =
                (Class<? extends Var<?>>) this.custom.getClass();
            outStream.writeByte(typeIdOf(varClass));
            this.custom.writeToStream(outStream);
        }
    }

    @Override
    public String toString() {
        String s = "Id:" + this.id + " Clazz:" + this.clazz
            + " Name:" + (this.osmName != null ? "'" + this.osmName + "'" : "null")
            + " LatLon:" + super.toString()
            + " RefCounter:" + this.refCounter + " OsmId:" + this.osmId;
        if (this.restrictions != null) {
            int nVnr = this.restrictions.length;
            s += " ViaNodeRestrictions[" + nVnr + "]:";
            for (int i = 0; i < nVnr; i++) {
                Restriction vnr = this.restrictions[i];
                s += " [" + i + "] " + vnr.toString();
            }
        }
        return s;
    }
}
