/*
 * $Id: Way.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.model;

import static de.cm.osm2po.misc.GeoUtils.calcDistance;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.VarNum;

/**
 * Abgeleitet von {@link OsmWay}, werden in dieser spezialisierten
 * Way-Class anstatt Refs (long Referenzen/Zeiger auf Knoten)
 * echte {@link Node} Objekte verarbeitet und serialisiert.
 * Zudem befinden sich hier noch weitere Attribute,
 * z.B. Informationen aus Relationen.
 * @see OsmWay
 * @see SegmentedWay
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class Way extends OsmWay {

    private Node[] nodes;
    private Membership[] memberships;

    public Way() {
        this.clear();
    }
    
    public void setNodes(Node[] nodes) {
        this.nodes = nodes;
    }
    
    public Node[] getNodes() {
        return this.nodes;
    }
    
    public Membership[] getMemberships() {
        return this.memberships;
    }
    
    public void setMemberships(Membership[] memberships) {
        this.memberships = memberships;
    }
    
    @Override
    public boolean isRing() {
        if (null == this.nodes) return false;
        int n = this.nodes.length;
        return (n > 2 && this.nodes[0].getId() == this.nodes[n-1].getId());
    }
    
    @Override
    public void clear() {
        super.clear();
        this.nodes = null;
        this.memberships = null;
    }
    
    public void clearMemberships() {
        this.memberships = null;
    }
    
    @Override
    public void setInvalid(boolean isInvalid) {
        throw new IllegalStateException(
                "osmWay.setInvalid() locked for derived classes");
    }
    
    /**
     * Upcast-Stream-Reader von {@link OsmWay} nach {@link Way}.
     * @param inStream {@link InStream}
     * @return {@link Way} <code>this</code>
     */
    public Way readFromOsmWayStream(InStream inStream) {
        super.readFromStream(inStream); // ruft clear() dieser Klasse
        return this;
    }
    
    public void writeToOsmWayStream(OutStream outStream) {
        // Wird beim Zurueckschreiben von orphaned ways im Joiner gerufen
        super.writeToStream(outStream);
    }
    
    @Override
    public void writeToStream(OutStream outStream) {
        super.clearRefs();
        super.writeToStream(outStream);
        int nNodes = null == this.nodes ? 0 : this.nodes.length;
        VarNum.writeUnsigned(nNodes, outStream);
        for (int i = 0; i < nNodes; i++) {
            this.nodes[i].writeToStream(outStream);
        }
        int nMemberships = 0;
        if (this.memberships != null) nMemberships = this.memberships.length;
        VarNum.writeUnsigned(nMemberships, outStream);
        for (int i = 0; i < nMemberships; i++) {
            this.memberships[i].writeToStream(outStream);
        }
    }
    
    @Override
    public Way readFromStream(InStream inStream) {
        super.readFromStream(inStream); // ruft clear() dieser Klasse
        int nNodes = (int) VarNum.readUnsigned(inStream);
        this.nodes = new Node[nNodes];
        for (int i = 0; i < nNodes; i++) {
            this.nodes[i] = new Node();
            this.nodes[i].readFromStream(inStream);
        }
        int nMemberships = (int) VarNum.readUnsigned(inStream);
        this.memberships = 0 == nMemberships ? null : new Membership[nMemberships];
        for (int i = 0; i < nMemberships; i++) {
            this.memberships[i] = new Membership();
            this.memberships[i].readFromStream(inStream);
        }
        return this;
    }
    
    @Override
    public String toString() {
        String s = super.toString();
        int nNodes = 0;
        if (this.nodes != null) nNodes = this.nodes.length;
        if (nNodes > 0) {
            s += "\nNodes[" + nNodes + "]:";
            for (int i = 0; i < nNodes; i++) {
                s += " (" + this.nodes[i].toString() + ")";
            }
        }
        int nMemberships = 0;
        if (this.memberships != null) nMemberships = this.memberships.length;
        if (nMemberships > 0) {
            s += "\nMemberships[" + nMemberships + "]:";
            for (int i = 0; i < nMemberships; i++) {
                s += " (" + this.memberships[i].toString() + ")";
            }
        }
        return s;
    }
   
    /**
     * Kalkuliert die echte Weglaenge in Kilometer.
     * @return double km
     */
    public double calcLengthKm() {
        double km = 0;
        if (this.nodes.length > 1) {
            Node ndFrom = this.nodes[0];
            for (int i = 1; i < this.nodes.length; i++) {
                Node ndTo = this.nodes[i];
                km += calcDistance(ndFrom.getLat(), ndFrom.getLon(),
                        ndTo.getLat(), ndTo.getLon());
                ndFrom = ndTo;
            }
        }
        return km;
    }

}
