/*
 * $Id: BndExtensionBuilderDb.java 5295 2021-05-18 18:49:13Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.misc.ArrayUtils.intArrayConcat;
import static de.cm.osm2po.misc.WkbUtils.fromWkbMultiPolygon;
import static de.cm.osm2po.misc.ZoneUtils.createZoneIdsBox;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Iterator;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.BndExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.model.OsmObject;
import de.cm.osm2po.model.OsmPolygon;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.Graph;
/**
 * Erweitert die {@link Graph}-Datei um Boundaries (Multipolygone)
 * aus einer Datenbank-Tabelle.<br> Die Datenbankverbindung wird
 * mittels {@link Config#openDbConnection(String)} aufgebaut.<br>
 * Das Abfrage-SQL muss folgende Spalten liefern, wobei die Namen
 * der Spalten auch abweichen duerfen. Die ersten vier Spalten
 * entsprechen uebrigens dem Kontrakt aus {@link OsmObject}.
 * <ol>
 * <li><strong>id</strong> : <code>bigint</code></li>
 * <li><strong>clazz</strong> : <code>byte</code></li>
 * <li><strong>flags</strong> : <code>integer</code></li>
 * <li><strong>name</strong> : <code>text</code></li>
 * <li>
 * <strong>geom</strong> : <code>geometry(MULTIPOLYGON, 4326)</code>
 * </li>
 * </ol>
 * Das SQL sollte mittels der Property <code><strong>sql</strong></code>
 * gesetzt werden. Default ist:<br>
 * <code>SELECT id, clazz, flags, name, geom FROM multipolygons</code><br><br>
 * Muessen weitere Spalten verarbeitet werden, so kann eine abgeleitete
 * Klasse {@link #handleCustom(OsmPolygon, ResultSet)} ueberschreiben.<br>
 * Das Prefix fuer die Datenbankverbindung in der Konfiguration kann abweichend
 * mittels <code><strong>db</strong></code> gesetzt werden. Ist dies nicht
 * gesetzt, so wird default der Wert <code>db</code> aus
 * {@link Config#openDbConnection(String)} verwendet.
 * <br><br>Da dies ein {@link PostProcessor} ist, muessen die Properties
 * entsprechend mit <code>postp.X.</code> geprefixt sein, wobei das X der Id
 * des konfigurierten Postprozessors entsprechen muss.
 *
 * @author (c) 2017/2021 - Carsten Moeller, info@osm2po.de
 * @see GeoExtensionBuilderBase
 */
public class BndExtensionBuilderDb
extends GeoExtensionBuilderBase<OsmPolygon>
implements Iterator<OsmPolygon> {

    private Config config;
    private int index;
    private ResultSet rst;
    private boolean hasNext;
    private Connection dbConn;

    @Override
    public Class<? extends Var<?>> forExtensionClass() {
        return BndExtension.class;
    }

    @Override
    public void run(Config config, int index) throws Exception {
        this.config = config;
        this.index = index;
        super.run(config, index);
        if (this.rst != null) this.rst.close();
        if (this.dbConn != null) dbConn.close();
    }

    @Override
    public Iterable<OsmPolygon> getSource() {
        final String prep = Config.POSTP(this.index) + ".";
        final String sql = this.config.getProperty(prep + "sql",
                "SELECT id, clazz, flags, name, geom FROM multipolygons");
        log.debug(sql);
        try {
            String dbPrefix = this.config.getProperty(prep + "db");
            this.dbConn = this.config.openDbConnection(dbPrefix);
            this.dbConn.setAutoCommit(false);
            Statement stm = this.dbConn.createStatement();
            stm.setFetchSize(0x400/*1k*/);
            this.rst = stm.executeQuery(sql);
            this.hasNext = rst.next();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        return new Iterable<OsmPolygon>() {
            @Override
            public Iterator<OsmPolygon> iterator() {
                return BndExtensionBuilderDb.this;
            }
        };
    }

    @Override
    public boolean hasNext() {
        return hasNext;
    }

    @Override
    public OsmPolygon next() {
        try {
            if (!hasNext) return null;

            long id = rst.getLong(1);
            byte clazz = rst.getByte(2);
            int flags = rst.getInt(3);
            String name = rst.getString(4);
            String geomHex = rst.getString(5);

            // Filter auf MultiPolygone aus Tabelle von beliebigen Geometrien moeglich!
            // Dazu stattdessen fromWkb(geomBytes) rufen und Typ pruefen.
            LatLons[][] multiPolygon = fromWkbMultiPolygon(Utils.hexToBytes(geomHex));
            VarString vsName = new VarString(name);
            OsmPolygon mp = new OsmPolygon(id, clazz, flags, vsName, multiPolygon);

            this.handleCustom(mp, rst);

            hasNext = rst.next();
            return mp;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Ermoeglicht einer abgeleiteteten Klasse zusaetzliche Informationen
     * anzubringen. Diese Implementation ist leer.
     * @param op {@link OsmPolygon}
     * @param rst {@link ResultSet}
     * @throws Exception beliebiger Natur
     */
    protected void handleCustom(OsmPolygon op, ResultSet rst) throws Exception {}

    @Override
    public void remove() {
        throw new UnsupportedOperationException();
    }

    @Override
    public int[] zoneIdsOf(OsmPolygon polygon, byte zoneSize) {
        int[] zoneIds = null;
        // ZoneIds der OuterPolygons ermitteln (Plural!)
        for (LatLons[] polyCoords : polygon.getPolygons()) {
            try {
                zoneIds = intArrayConcat( // polygon[0] ist Outer
                    zoneIds, createZoneIdsBox(polyCoords[0].getCoords(), zoneSize));
            } catch (Throwable t) {
                log.error(t.toString());
                log.debug(polygon.toString() + "\n" + Log.translateError(t, 3));
            }
        }
        return zoneIds;
    }

}
