/*
 * $Id: DbExtensionBuilder.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.misc.Utils.createTmpFile;
import static java.lang.reflect.Modifier.PUBLIC;
import static java.lang.reflect.Modifier.STATIC;

import java.io.File;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.cm.osm2po.Config;
import de.cm.osm2po.annotations.DbColumn;
import de.cm.osm2po.annotations.DbSource;
import de.cm.osm2po.extensions.DbExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.LongLongBag;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;

/**
 * Schreibt eine {@link VarBean}-Datei mit einer Sammlung
 * von DB-Abzuegen (Views und Tables).<br>
 * Die Datenbankverbindung wird mittels
 * {@link Config#openDbConnection(String)} aufgebaut.<br>
 * Die korrespondierende Entity-Klasse muss mit {@link DbSource}
 * und deren Felder mittels {@link DbColumn} annotiert sein.
 * Fuer <code>public</code> Felder ist keine Annotation noetig.
 * <br><br><b>Wichtig:</b><br>
 * Die View muss zwingend eine Index-Spalte vom Typ int oder long
 * besitzen und entweder <code>id</code> heissen oder mit dem
 * in {@link DbSource#index()} gelieferten Namen uebereinstimmen.
 * Der Index ist normalerweise die Id (also unique), allerdings koennen
 * hier auch Fremdschluessel gesetzt werden, so dass auch N-Saetze,
 * also z.B. alle Kontakte einer Person, geliefert werden koennen.
 * <ol>
 * <li><strong>classes</strong> : CSV von voll qualifizierten Entity-Klassen</li>
 * <li><strong>db</strong> : Prefix fuer DB-Verbindung, default ist 'db'</li>
 * <li><strong>file</strong> : Abweichender Dateipfad der Bean-Datei</li>
 * </ol>
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 * @see DbExtension
 * @see DbSource
 * @see DbColumn
 */
public class DbExtensionBuilder implements PostProcessor {
    public final static String BEAN_FILE = "dbviews.bean";

    private class DbExtract {
        final String className;
        final String key;
        final File dataFile;
        final File indexFile;
        DbExtract(String className, String alias,
                File dataFile, File indexFile) {
            this.className = className;
            this.key = alias;
            this.dataFile = dataFile;
            this.indexFile = indexFile;
        }
    }

    @Override
    public void run(Config config, int index) throws Exception {
        Log log = config.getLog();

        final String prep = Config.POSTP(index) + ".";
        final String classesCsv = config.getProperty(prep + "classes");
        final String dbPrefix = config.getProperty(prep + "db", "db");

        File beanFile = new File(config.getProperty(prep + "file", BEAN_FILE));
        if (null == beanFile.getParentFile()) // wenn ohne Verzeichnis, dann WorkDir
            beanFile = new File(config.getWorkDir(), beanFile.getName());

        List<DbExtract> dbExtractList = new ArrayList<DbExtract>();

        String[] classNames = classesCsv.split(",");

        for (String className : classNames) {
            Class<?> varClass = Config.loadClass(className.trim());
            DbSource dbSource = varClass.getAnnotation(DbSource.class);
            if (dbSource != null) { // TODO Meckern, wenn nicht!
                String viewKey = dbSource.alias();
                if ("".equals(viewKey)) viewKey = className;
                File dataFile = createTmpFile(config.getWorkDir());
                LongLongBag indexBag = extractData(varClass, dataFile, config, dbPrefix);
                if (indexBag != null && indexBag.getSize() > 0) {
                    File indexFile = createTmpFile(config.getWorkDir());
                    OutStream outStream = new OutStreamDisk(indexFile);
                    indexBag.sort();
                    indexBag.writeToStream(outStream);
                    outStream.close();
                    dbExtractList.add(new DbExtract(varClass.getName(), viewKey,
                            dataFile, indexFile));
                }
            }
        }

        OutStreamDisk osd = new OutStreamDisk(beanFile);
        osd.writeByte(VarTypeDesk.typeIdOf(DbExtension.class));
        osd.writeInt(dbExtractList.size());
        Set<String> viewKeySet = new HashSet<String>();

        for (DbExtract dbExtract : dbExtractList) {
            if (viewKeySet.contains(dbExtract.key)) {
                log.error("Duplicate key '" + dbExtract.key
                        + "' in " + dbExtract.className);
                continue;
            }
            viewKeySet.add(dbExtract.key);

            new VarString(dbExtract.className).writeToStream(osd);
            new VarString(dbExtract.key).writeToStream(osd);
            osd.writeLong(dbExtract.dataFile.length());
            osd.concatFile(dbExtract.indexFile, 0);
        }
        for (DbExtract dbExtract : dbExtractList) {
            osd.concatFile(dbExtract.dataFile, 0);
            dbExtract.indexFile.delete();
            dbExtract.dataFile.delete();
        }
        osd.close();
    }

    private static LongLongBag extractData(Class<?> varClass, File file,
            Config config, String dbPrefix) throws Exception {

        DbSource dbSource = varClass.getAnnotation(DbSource.class);

        Connection con = config.openDbConnection(dbPrefix);
        con.setAutoCommit(false);
        Statement stm = con.createStatement();
        stm.setFetchSize(0x400/*1k*/);
        ResultSet rst = stm.executeQuery(dbSource.sql());

        int nColumns = rst.getMetaData().getColumnCount();

        Field[] matchingFields = new Field[nColumns];
        boolean[] matchingNotNulls = new boolean[nColumns];
        int[] matchingColumns = new int[nColumns];
        int nMatchingFields = 0;

        for (int i = 1; i <= nColumns; i++) {
            Class<?> colClass = Class.forName(rst.getMetaData().getColumnClassName(i));
            String colName = rst.getMetaData().getColumnLabel(i); // col AS ...

            for (Field field : varClass.getDeclaredFields()) {
                DbColumn dbColumn = field.getAnnotation(DbColumn.class);
                String fieldName = "";
                boolean nullable = true;
                if (dbColumn != null) {
                    fieldName = dbColumn.value();
                    if ("".equals(fieldName)) fieldName = field.getName();
                    nullable = dbColumn.nullable();
                } else if (PUBLIC == (field.getModifiers() & (PUBLIC|STATIC))) {
                    fieldName = field.getName();
                } else {
                    continue;
                }

                if (colName.equals(fieldName)) {
                    Class<?> fieldClass = field.getType();
                    if (fieldClass.isAssignableFrom(colClass)
                            || Number.class.isAssignableFrom(colClass)
                            || Character.class.isAssignableFrom(colClass)
                            || Boolean.class.isAssignableFrom(colClass)) {
                        field.setAccessible(true);
                        matchingColumns[nMatchingFields] = i;
                        matchingFields[nMatchingFields] = field;
                        matchingNotNulls[nMatchingFields] = !nullable;
                        nMatchingFields++;
                        break;
                    }
                }
            }
        }

        LongLongBag indexBag = new LongLongBag();
        Var<?> var = (Var<?>) varClass.getDeclaredConstructor().newInstance();
        OutStream outStream = new OutStreamDisk(file);

        while (rst.next()) {
            boolean doWrite = true;
            for (int i = 0; i < nMatchingFields; i++) {
                Object colVal = rst.getObject(matchingColumns[i]);
                if (null == colVal && matchingNotNulls[i]) {
                    doWrite = false;
                    break;
                }

                if (!matchingFields[i].getType().isPrimitive()) {
                    matchingFields[i].set(var, colVal);

                } else {
                    Class<?> primitiveClazz = matchingFields[i].getType();
                    if (Long.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setLong(var, 0);
                        else matchingFields[i].setLong(var, ((Number)colVal).longValue());
                    } else if (Integer.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setInt(var, 0);
                        else matchingFields[i].setInt(var, ((Number)colVal).intValue());
                    } else if (Short.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setShort(var, (short)0);
                        else matchingFields[i].setShort(var, ((Number)colVal).shortValue());
                    } else if (Byte.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setByte(var, (byte)0);
                        else matchingFields[i].setByte(var, ((Number)colVal).byteValue());
                    } else if (Double.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setDouble(var, 0);
                        else matchingFields[i].setDouble(var, ((Number)colVal).doubleValue());
                    } else if (Float.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setFloat(var, 0);
                        else matchingFields[i].setFloat(var, ((Number)colVal).floatValue());
                    } else if (Boolean.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setBoolean(var, false);
                        else matchingFields[i].setBoolean(var, ((Boolean)colVal).booleanValue());
                    } else if (Character.TYPE == primitiveClazz) {
                        if (null == colVal) matchingFields[i].setChar(var, '\0');
                        else matchingFields[i].setChar(var, ((Character)colVal).charValue());
                    }
                }
            }

            if (doWrite) {
                long index = rst.getLong(dbSource.index());
                indexBag.add(index, outStream.getWriteIndex());
                var.writeToStream(outStream);
            }
        }
        outStream.close();

        rst.close();
        stm.close();
        con.close();

        return indexBag;
    }
}
