/*
 * $Id: GeoExtensionBuilderBase.java 5848 2023-10-19 14:08:07Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.converter.GraphBuilder.addExtension;
import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;
import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.misc.Utils.createTmpFile;
import static de.cm.osm2po.primitives.VarTypeDesk.typeIdOf;

import java.io.File;
import java.util.Arrays;
import java.util.StringTokenizer;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.GraphExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Props;
import de.cm.osm2po.misc.ZoneUtils;
import de.cm.osm2po.model.OsmObject;
import de.cm.osm2po.model.OsmObjectTagged;
import de.cm.osm2po.primitives.Dictionary;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.IndexRaw;
import de.cm.osm2po.primitives.IntIntBag;
import de.cm.osm2po.primitives.LongIntBag;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.OutStreamDisk;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarBean;
import de.cm.osm2po.primitives.VarByteArray;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.GraphHeader;

/**
 * Abstrakte Basis-Implementation, um eine Sammlung von {@link OsmObject}s als
 * {@link GraphExtension} zu schreiben. Und zwar direkt in den Graphen und/oder
 * isoliert als eigene {@link VarBean}-Datei.<br>
 * Diese Klasse abstrahiert einen sehr komplexen Mechanismus und erlaubt so
 * mit geringem Aufwand unterschiedliche Quellen und Typen zu implementieren.<br>
 * Das Ergebnis ist eine Komposition aus einer Liste von Geo-Objekten, einem
 * Dictionary, und einem einfachen Geo-Index auf Rasterbasis.
 * Objekte koennen so schnell per Koordinate oder Namen gefunden werden.
 * Die folgenden {@link Props} werden unterstuetzt. Defaults in Klammern:
 * <ul>
 * <li>
 *      <strong>zoneSize</strong> : {@link ZoneUtils}-Raster fuer Suche (1).
 *      Die eigentliche Abdeckung obliegt der jeweiligen Implementation
 *      von {@link #zoneIdsOf(Var, byte)} und dem darin enthaltenen
 *      Aufruf zu z.B. {@link ZoneUtils#expandZone(int, int, int)}.
 * </li>
 * <li><strong>noDic</strong> : Kein {@link Dictionary} erzeugen (false)</li>
 * <li><strong>beanFile</strong> : Nur externe Extension-Datei erzeugen (null)</li>
 * </ul>
 * Alle oben gelisteten Properties muessen mit
 * <code><strong>postp.X.</strong></code> geprefixt werden, wobei X durch
 * die Id des Postprozessors ersetzt werden muss.
 * @author (c) 2016/2017 - Carsten Moeller, info@osm2po.de
 * @param <T> extends {@link OsmObject}
 * @see BndExtensionBuilder
 * @see BndExtensionBuilderDb
 * @see PoiExtensionBuilder2po
 * @see PoiExtensionBuilderDb
 */
public abstract class GeoExtensionBuilderBase<T extends Var<?> & OsmObject>
implements PostProcessor {
    public final static int COMPAT_VERSION = 1;

    protected Log log;
    protected byte EXTENSION_ID;

    protected abstract Iterable<T> getSource();
    protected abstract int[] zoneIdsOf(T osmObject, byte zoneSize);
    protected abstract Class<? extends Var<?>> forExtensionClass();

    private String tokenizers;

    /**
     * Liefert die assozierten Namen/Tags fuer das interne {@link Dictionary}.<br>
     * Der eigentliche Name wird in Gaenze aufgenommen und danach auf
     * Tokens untersucht. Optional koennen weitere Tokens ueber
     * {@link OsmObjectTagged#getTags()} hineinkommen.<br>
     * Mittels {@link StringTokenizer} werden die Einzelteile dann ebenfalls
     * hinzugefuegt, vorausgesetzt, sie beginnen nicht selbst mit dem
     * eigentlichen Namen.<br>
     * Als Token-Trenner werden <code>| .:-[]()</code> erkannt.
     * Dies kann jedoch mittels der Config-Property <code>tokenizers</code>
     * ueberschrieben werden.<br>Wichtig:
     * Tokens werden hier immer als {@link String#toLowerCase()} aufgenommen.
     * @param osmObject {@link OsmObject}, NULLABLE
     * @return Array von {@link VarString} von assoziierenden Namen, NULLABLE
     * @see OsmObjectTagged
     */
    protected VarString[] associateNames(T osmObject) {
        VarString fullName = osmObject.getName();
        if (null == fullName || fullName.isEmpty()) return new VarString[0];

        String tags = fullName.toString();
        if (osmObject instanceof OsmObjectTagged) {
            VarString moreTags = ((OsmObjectTagged) osmObject).getTags();
            if (moreTags != null && !moreTags.isEmpty()) {
                tags += this.tokenizers.charAt(0) + moreTags.toString();
            }
        }
        tags = tags.toLowerCase();

        VarString[] vsTags = new VarString[1];
        int nTags = 0;
        vsTags[nTags++] = fullName.toLowerCase();

        StringTokenizer tokenizer = new StringTokenizer(tags, this.tokenizers);
        while (tokenizer.hasMoreElements()) {
            String token = tokenizer.nextToken().trim();
            if (token.length() < 3) continue;
            if (tags.startsWith(token)) continue; // dont find it twice via startswith

            if (nTags == vsTags.length) // ArrayOverflow
                vsTags = Arrays.copyOf(vsTags, nTags * 2);
            vsTags[nTags++] = new VarString(token);
        }

        return Arrays.copyOf(vsTags, nTags);
    }

    @Override
    public void run(Config config, int index)
    throws Exception {
        this.log = config.getLog();
        File workDir = config.getWorkDir();
        File graphFile = null;

        final String prep = Config.POSTP(index) + ".";

        this.EXTENSION_ID = Byte.valueOf(
                config.getProperty(prep + "extensionId", "0"));

        this.tokenizers = config.getProperty(prep + "tokenizers", "| .:-[]()");

        final byte zoneSize = Byte.parseByte(
                config.getProperty(prep + "zoneSize", "1"));
        final boolean buildDic = !Boolean.valueOf( // Dictionary builden ja/nein
                config.getProperty(prep + "noDic"));

        final String beanFilePath = config.getProperty(prep + "beanFile");
        final boolean isolateBean = beanFilePath != null;
        final File beanFile = isolateBean ? new File(beanFilePath) : createTmpFile(workDir);

        if (isolateBean) {
            log.info("Building isolated " + this.forExtensionClass().getSimpleName() + " " + beanFile);

        } else {
            graphFile = config.findGraphFile(true);
            if (null == graphFile) throw new IllegalStateException("Graphfile not found");

            if (new GraphHeader(graphFile).isExtensionExist(this.forExtensionClass(), EXTENSION_ID)) {
                log.error("Extension " + this.forExtensionClass().getSimpleName()
                        + "#" + EXTENSION_ID + " already exists");
                return;
            }
        }

        // Rechteckige Raster-Abdeckungen der Geometry
        IntIntBag objZoneBag = new IntIntBag();
        // Index der Geometrien
        IndexRaw<T> objIndex = new IndexRaw<T>();
        objIndex.create(workDir, log);
        // Dictionary von Namen der Geometrien
        Dictionary objDic = new Dictionary();
        objDic.create(workDir, log);
        // Fuer Schnellfilter Clazzes pro Index extra halten
        byte[] clazzes = new byte[0x100];
        // Schnellsucher von ID(s) zu Idx
        LongIntBag id2IdxBag = new LongIntBag();

        int nObjects = 0;
        for (T osmObject : this.getSource()) {
            int[] zoneIds = this.zoneIdsOf(osmObject, zoneSize);

            // Geografische Abdeckung (Zonen), Indexe und Namen vermerken
            if (zoneIds != null) {
                VarString[] dicNames = this.associateNames(osmObject);
                int idx = objIndex.push(osmObject);

                if (idx == clazzes.length)
                    clazzes = Arrays.copyOf(clazzes, idx *2);
                clazzes[idx] = osmObject.getClazz();

                id2IdxBag.add(osmObject.getId(), idx);

                if (dicNames != null) for (VarString dicName : dicNames) {
                    if (buildDic && dicName != null && !VarString.EMPTY.equals(dicName))
                        objDic.add(idx, dicName.toString().toLowerCase());
                }
                zoneIds = intArrayDistinct(zoneIds);
                for (int zoneId : zoneIds) objZoneBag.add(zoneId, idx);
            }

            if (++nObjects % 2500 == 0) log.progress(DF(nObjects) + " "
                    + osmObject.getClass().getSimpleName() + "s processed.");
        }

        clazzes = Arrays.copyOf(clazzes, objIndex.getIndexSize());

        log.progress("Sorting IdToIdxs");
        id2IdxBag.sort(); // Fuer spaetere fast Lookups aufgrund LatLon

        log.progress("Sorting touched ZoneIds");
        objZoneBag.sort(); // Fuer spaetere fast Lookups aufgrund LatLon

        log.info(DF(objIndex.getIndexSize())
                + " of " + DF(nObjects) + " Elements written, "
                + DF(objZoneBag.getSize()) + " ZoneIds touched.");

        // Da wir spaeter mehrere Dateipositionen im Vorraus kennen muessen,
        // wird hier nochmals eine finale Temp-Datei gebaut. Dies ist vor allem
        // deswegen noetig, weil die Groessen der Indexe, z.B. die des IndexSet
        // erst nach einmaligem Rausschreiben bekannt sind.
        if (beanFile.getParentFile() != null) beanFile.getParentFile().mkdirs();
        OutStream osExt = new OutStreamDisk(beanFile);

        // 1.Byte im VarBean-File ist der eigentliche Typ
        osExt.writeByte(typeIdOf(forExtensionClass()));

        // Headerdaten hier noch nicht aller ermittelt, daher mit 0 fuellen
        osExt.writeInt(0); // version
        osExt.writeByte((byte)0); // zoneSize
        osExt.writeLong(0); // objIndexPos
        osExt.writeLong(0); // objDictionaryPos
        osExt.writeLong(0); // eofPos (size in bytes)

        log.progress("Writing Clazzes for Entries");
        new VarByteArray(clazzes).writeToStream(osExt);

        log.progress("Writing IdToIdxBag");
        id2IdxBag.writeToStream(osExt);
        id2IdxBag.clear();

        log.progress("Writing ZoneBag");
        objZoneBag.writeToStream(osExt);
        log.debug(DF(objZoneBag.getSize()) + " Zones written.");
        objZoneBag.clear();

        log.progress("Building Index");
        long objIndexPos = osExt.getWriteIndex();
        objIndex.build(osExt);

        log.progress("Building Dictionary");
        long objDicPos = osExt.getWriteIndex();
        objDic.build(osExt);

        long eofPos = osExt.getWriteIndex();
        osExt.close();

        // Headerdaten sind jetzt bekannt. Ueberschreiben.
        RandomAccessDisk raExt = new RandomAccessDisk(beanFile, false);
        raExt.setWriteIndex(1); // 1.Byte ist der eigentliche VarType
        raExt.writeInt(COMPAT_VERSION);
        raExt.writeByte(zoneSize);
        raExt.writeLong(objIndexPos - 1);
        raExt.writeLong(objDicPos - 1);
        raExt.writeLong(eofPos - 1);
        raExt.close();

        // Jetzt kann die Extension geschrieben werden
        if (isolateBean) log.info(beanFile.getAbsolutePath() + " created");
        else this.addToGraphFile(beanFile, graphFile);
        if (!isolateBean) beanFile.delete();
    }


    private void addToGraphFile(File extensionFile, File graphFile) throws Exception {
        long size = extensionFile.length() -1; // size(Var) is size(VarBean - TypeByte)

        Class<? extends Var<?>> varClass = this.forExtensionClass();
        byte varType = typeIdOf(varClass);

        log.info("Writing " + varClass.getSimpleName() + "#" + EXTENSION_ID
                + " from " + extensionFile.getName()
                + " to " + graphFile.getName()
                + " [" + (DF2((double)size / 0x100000) + "M]"));

        InStream inStream = new InStreamDisk(extensionFile);
        byte varBeanType = inStream.readByte();
        if (varType != varBeanType) { // Check!
            throw new RuntimeException("VarBean- and Var-Type do not match");
        }
        addExtension(graphFile, varType, EXTENSION_ID, size, inStream);
        inStream.close();
    }

}
