/*
 * $Id: MlgExtensionBuilderRouter.java 3897 2015-08-28 11:51:19Z cmindividual $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.routing.Graph;

/**
 * Router fuer den {@link MlgExtensionBuilder}.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class MlgExtensionBuilderRouter {

    private Log log = new Log();

    private final static int PQ_INITIAL_CAPACITY = 10000;
    private int pqSize; // Current PQ size.
    private int pqCapacity; // Current PQ Capacity
    private int[] pqEdgeIdxs;
    private double[] pqCosts;

    private int[] wgbs; // OpenVisitedClosed-List

    // Current Edge (after pqRemove)
    private int pqrEdgeIdx;
    private double pqrCost;

    private Graph graph;

    private float[] edgeCosts;

    // Artificial Semaphore/Synchronize
    private boolean dirty;

    // Mode
    private boolean ignoreOneWays;
    private boolean ignoreRestrictions;
    
    public void reset() {
        this.dirty = false;
        this.pqEdgeIdxs = null;
        this.pqCosts = null;
        this.wgbs = null;
    }

    public void setLog(Log log) {
        this.log = log;
    }

    private void initPqAndWgb() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Max. ArraySize PQ
        this.pqEdgeIdxs = new int[pqCapacity];
        this.pqCosts = new double[pqCapacity];

        int nVertices = this.graph.getEdgeEntries().length;
        this.wgbs = new int[nVertices];

        // DeadEdge references this Vertex.
        this.wgbs[0] = Integer.MAX_VALUE | 0x80000000;
    }

    public void traverse(Graph graph, int cellId,
            int[] cellIds, byte level, byte[] levels, int sourceId,
            boolean findShortest, boolean ignoreOneWays, boolean ignoreRestrictions) {

        this.ignoreOneWays = ignoreOneWays;
        this.ignoreRestrictions = ignoreRestrictions;
        
        this.init(graph, findShortest);
        this.initPqAndWgb();
        this.startWithVertex(sourceId, cellId, cellIds, level, levels);
        this.traverseInternal(cellId, cellIds, level, levels);
    }

    private void init(Graph graph, boolean shortest) {
        this.graph = graph;

        if (this.dirty) {
            throw new RuntimeException(
            "Router already in use. Call reset() first.");
        } 
        this.dirty = true;

        if (shortest) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
        }
    }
    
    private void startWithVertex(int fromSourceId,
            int cellId, int[] cellIds, byte level, byte[] levels) {

        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();

        // StartVertex has no incoming edge. (all Bits Set = -1)
        this.wgbs[fromSourceId] = Integer.MAX_VALUE | 0x80000000;

        int edgeIdx = edgeEntries[fromSourceId];
        while (edgeSourceIds[edgeIdx] == fromSourceId) {
            int targetId = edgeTargetIds[edgeIdx];
            if (cellId == cellIds[targetId]) {
                int toLevel = levels[edgeIdx];
                if (toLevel >= level) {
                    if (this.wgbs[targetId] >= 0) {
                        boolean isWrongWay = (edgeIds[edgeIdx] & 0x80000001) == 0x80000000;
                        if (!isWrongWay || this.ignoreOneWays) {
                            double cost = this.edgeCosts[edgeIdx];
                            this.pqInsertOrUpdate(edgeIdx, cost, targetId);
                        }
                    }
                }
            }
            edgeIdx++;
        }
    }

    private void traverseInternal(int cellId,
            int[] cellIds, byte level, byte[] levels) {
        
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        byte[] edgeNoTurnBits = this.graph.getEdgeNoTurnBits();

        while (this.pqSize > 0) {
            this.pqRemove();
            int targetId = edgeTargetIds[this.pqrEdgeIdx];
            this.wgbs[targetId] = this.pqrEdgeIdx | 0x80000000;
            
            if (cellId != cellIds[targetId]) {
                continue;
            }

            int sourceId = targetId;
            int edgeIdx = edgeEntries[sourceId];

            int noTurnBits = edgeNoTurnBits[this.pqrEdgeIdx];
            int noTurnBit = 1;

            do { // Outgoing Edges.
                int toLevel = levels[edgeIdx];
                
                if (toLevel >= level) {
                    
                    targetId = edgeTargetIds[edgeIdx];
                    if (this.wgbs[targetId] >= 0) {
                        boolean mayTurn = true;
                        if (noTurnBits != 0 && !this.ignoreRestrictions)
                            mayTurn &= ((noTurnBits & noTurnBit) == 0);
                        if (!this.ignoreOneWays)
                            mayTurn &= (!((edgeIds[edgeIdx] & 0x80000001) == 0x80000000));
                        
                        if (mayTurn) {
                            double cost = this.pqrCost + this.edgeCosts[edgeIdx];
                            this.pqInsertOrUpdate(edgeIdx, cost, targetId);
                        }
                    }
                    
                }
                noTurnBit <<= 1;
                edgeIdx++;
            } while (sourceId == edgeSourceIds[edgeIdx]);
        }
    }

    public final int[] makePath(int toTargetId) {
        if (!this.isVisited(toTargetId)) return null;

        int[] edgeSourceIds = this.graph.getEdgeSourceIds();

        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        int targetId = toTargetId;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY); 
            int edgeIdx = this.wgbs[targetId] & 0x7FFFFFFF; // Bit 32 (Visited-Black-Flag).
            if (edgeIdx == Integer.MAX_VALUE) break;
            int sourceId = edgeSourceIds[edgeIdx];
            pathIds[n++] = edgeIdx;
            targetId = sourceId;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        // Reverse
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    public final int[] getVisited() {
        int len = this.wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) {
            int edgeIdx = this.wgbs[vertexId];
            if (edgeIdx < 0) {
                if (buffer.length < vertexId)
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    public final boolean isVisited(int vertexId) {
        if (vertexId < 1 || vertexId > this.wgbs.length -1) return false;
        return this.wgbs[vertexId] < 0;
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqEdgeIdxs = Arrays.copyOf(this.pqEdgeIdxs, this.pqCapacity);
        this.pqCosts = Arrays.copyOf(this.pqCosts, this.pqCapacity);
        log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsertOrUpdate(int edgeIdx, double cost, int targetId) {
        int pqPos = this.wgbs[targetId];

        if (0 == pqPos) { // new Target-Node
            pqPos = ++this.pqSize;

            if (this.pqSize >= this.pqCapacity) this.pqGrow();

            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCosts[pqPos] = cost;

            this.wgbs[targetId] = pqPos;

        } else if (this.pqCosts[pqPos] > cost) { // shorter way found
            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCosts[pqPos] = cost;

        } else {
            return;
        }

        this.pqUpheap(pqPos);
    }

    private void pqRemove() {
        this.pqrEdgeIdx = this.pqEdgeIdxs[1];
        this.pqrCost = this.pqCosts[1];
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap(1);
    }

    private void pqSwap(int pos1, int pos2) {
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();

        // PQ-Swap
        int edgeIdx1 = this.pqEdgeIdxs[pos1];
        double cost1 = this.pqCosts[pos1];
        int edgeIdx2 = this.pqEdgeIdxs[pos2];
        double cost2 = this.pqCosts[pos2];

        this.pqEdgeIdxs[pos1] = edgeIdx2;
        this.pqCosts[pos1] = cost2;
        this.pqEdgeIdxs[pos2] = edgeIdx1;
        this.pqCosts[pos2] = cost1;

        // WGBS-Swap
        int targetId1 = edgeTargetIds[edgeIdx1];
        int targetId2 = edgeTargetIds[edgeIdx2];

        this.wgbs[targetId1] = pos2;
        this.wgbs[targetId2] = pos1;
    }

    private void pqUpheap(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            if (this.pqCosts[parentPos] < this.pqCosts[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos) {
        int childPos = pos << 1; // * 2 = left child 
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // right child exists?
                if (this.pqCosts[childRightPos] < this.pqCosts[childPos]) {
                    childPos = childRightPos;
                }
            }
            if (this.pqCosts[pos] < this.pqCosts[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
