/*
 * $Id: PoiExtensionBuilderDb.java 5295 2021-05-18 18:49:13Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.postp;

import static de.cm.osm2po.misc.WkbUtils.fromWkbPoint;
import static de.cm.osm2po.misc.WktUtils.parseWkt;
import static de.cm.osm2po.misc.ZoneUtils.createZoneId;
import static de.cm.osm2po.misc.ZoneUtils.expandZone;
import static java.sql.Types.OTHER;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.Statement;
import java.util.Iterator;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.PoiExtension;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.OsmObject;
import de.cm.osm2po.model.OsmObjectTagged;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.Graph;

/**
 * Erweitert die {@link Graph}-Datei um POIs ({@link OsmNode}s)
 * aus einer Datenbank-Tabelle.<br> Die Datenbankverbindung wird
 * mittels {@link Config#openDbConnection(String)} aufgebaut.<br>
 * Das Abfrage-SQL muss die folgende Spalten liefern, wobei die Namen
 * der Spalten auch abweichen duerfen. Die ersten vier Spalten
 * entsprechen uebrigens dem Kontrakt aus {@link OsmObject}.
 * <ol>
 * <li><strong>id</strong> : <code>bigint</code></li>
 * <li><strong>clazz</strong> : <code>byte</code></li>
 * <li><strong>flags</strong> : <code>integer</code></li>
 * <li><strong>name</strong> : <code>text</code></li>
 * <li><strong>tags</strong> : <code>text</code></li>
 * <li><strong>geom</strong> : <code>geometry(POINT, 4326) oder WKT</code></li>
 * </ol>
 *
 * Das SQL sollte mittels der Property <code><strong>sql</strong></code>
 * gesetzt werden. Default ist:<br>
 * <code>SELECT id, clazz, flags, name, tags, geom FROM nodes</code><br><br>
 * Muessen weitere Spalten verarbeitet werden, so kann eine abgeleitete
 * Klasse {@link #handleCustom(OsmNode, ResultSet)} ueberschreiben.<br>
 * Das Prefix fuer die Datenbankverbindung in der Konfiguration kann abweichend
 * mittels <code><strong>db</strong></code> gesetzt werden. Ist dies nicht
 * gesetzt, so wird default der Wert <code>db</code> aus
 * {@link Config#openDbConnection(String)} verwendet.
 * <br><br>Da dies ein {@link PostProcessor} ist, muessen die Properties
 * entsprechend mit <code>postp.X.</code> geprefixt sein, wobei das X der Id
 * des konfigurierten Postprozessors entsprechen muss.
 *
 * @author (c) 2017/2021 - Carsten Moeller, info@osm2po.de
 * @see GeoExtensionBuilderBase
 */
public class PoiExtensionBuilderDb
extends GeoExtensionBuilderBase<OsmNode>
implements Iterator<OsmNode> {

    private Config config;
    private int index;
    private ResultSet rst;
    private boolean hasNext;
    private Connection dbConn;
    private boolean geomIsWkb;

    public final static class TaggedOsmNode extends OsmNode implements OsmObjectTagged {
        private VarString tags;
        @Override public VarString getTags() {return this.tags;}
    }

    @Override
    public Class<? extends Var<?>> forExtensionClass() {
        return PoiExtension.class;
    }

    @Override
    public void run(Config config, int index) throws Exception {
        this.config = config;
        this.index = index;

        super.run(config, index);

        if (this.rst != null) this.rst.close();
        if (this.dbConn != null) dbConn.close();
    }

    @Override
    public Iterable<OsmNode> getSource() {
        try {
            final String prep = Config.POSTP(this.index) + ".";

            String sql = this.config.getProperty(prep + "sql",
                "SELECT id, clazz, flags, name, tags, geom FROM nodes");
            log.debug(sql);

            String dbPrefix = this.config.getProperty(prep + "db");
            this.dbConn = this.config.openDbConnection(dbPrefix);
            this.dbConn.setAutoCommit(false);
            Statement stm = this.dbConn.createStatement();
            stm.setFetchSize(0x400/*1k*/);
            this.rst = stm.executeQuery(sql);
            this.hasNext = rst.next();

            ResultSetMetaData metaData = this.rst.getMetaData();
            this.geomIsWkb = metaData.getColumnType(6) == OTHER; // PostGis / PgObject if onlny Postgres

        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        return new Iterable<OsmNode>() {
            @Override
            public Iterator<OsmNode> iterator() {
                return PoiExtensionBuilderDb.this;
            }
        };
    }

    @Override
    public boolean hasNext() {
        return hasNext;
    }

    @Override
    public OsmNode next() {
        try {
            if (!hasNext) return null;
            long id = rst.getLong(1);
            byte clazz = rst.getByte(2);
            int flags = rst.getInt(3);
            String name = rst.getString(4);
            String tags = rst.getString(5);
            String geom = rst.getString(6);
            LatLon coord = this.geomIsWkb ? fromWkbPoint(Utils.hexToBytes(geom)) : (LatLon) parseWkt(geom);

            TaggedOsmNode nd = new TaggedOsmNode();
            nd.setLat(coord.getLat());
            nd.setLon(coord.getLon());
            nd.setId(id);
            nd.setClazz(clazz);
            nd.setFlags(flags);
            nd.setName(new VarString(name));

            if (tags != null) nd.tags = new VarString(tags);

            this.handleCustom(nd, rst);
            hasNext = rst.next();
            return nd;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Ermoeglicht einer abgeleiteteten Klasse zusaetzliche Informationen
     * anzubringen. Diese Implementation ist leer.
     * @param nd {@link OsmNode}
     * @param rst {@link ResultSet}
     * @throws Exception beliebiger Natur
     */
    protected void handleCustom(OsmNode nd, ResultSet rst) throws Exception {}

    @Override
    public void remove() {
        throw new UnsupportedOperationException();
    }

    @Override
    public int[] zoneIdsOf(OsmNode osmNode, byte zoneSize) {
        int zoneId = createZoneId(osmNode.getLat(), osmNode.getLon(), zoneSize);
        int[] zoneIds = expandZone(zoneId, 1, zoneSize);
        return zoneIds;
    }

}
