/*
 * $Id: BigLongLongHashMap.java 5451 2021-11-26 20:45:59Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import de.cm.osm2po.misc.Utils;

/**
 * Eine Hashmap auf Grundlage von primitiven Datentypen fuer
 * Key und Value (long / long).
 * <br><br><b>Wichtig:</b><br>
 * Es koennen maximal 2Gig Elemente adressiert werden.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class BigLongLongHashMap {
    // TODO LongLongHashMap ist kein Var

    public static final long NULLVALUE = Long.MIN_VALUE;

    // Ein MapEntry benoetigt immer Key, Value und einen Zeiger
    // auf einen kollidierenden Naechsten mit gleichem Key.
    // Anstatt eine LinkedList eines solchen Objekts zu erstellen,
    // verwenden wir drei synchrone Arrays.
    private LongArray keys = new LongArray();
    private LongArray values = new LongArray();
    private IntArray nexts = new IntArray();

    private int size; // Anzahl neuer Eintraege ohne Ueberschreiben
    private int hashSize; // Hash-Array-Groesse
    private int collisionsMax; // Anzahl der Kollisionen (gleicher Hash)
    private int collisionsSum  = 0; // Fuer durchschnittliche Kollisionen.

    // TODO map und reuseIndex aus Speichergruenden nicht zwangslaeufig als IntArray
    // TODO Aufgrund der internen IntArray und LongArray's koennte der Adress-Bereich groesser sein

    // Dies ist die Eigentliche HashMap
    private IntArray map = new IntArray();

    // Hier merken wir uns geloeschte Speicherplaetze fuer Wiederverwendung
    private IntArray reuseArray = new IntArray();
    private int reuseArraySize = 0;


    public BigLongLongHashMap() {this(0x100000);}

    public BigLongLongHashMap(int hashSize) {
        this.hashSize = (int) Utils.findNextPrime(hashSize);
    }

    public LongArray getRawKeys() {
        return this.keys;
    }

    public void clear() {
        this.keys.clear();
        this.values.clear();
        this.nexts.clear();
        this.map.clear();
        this.reuseArray.clear();

        this.collisionsMax = 0;
        this.size = 0;
        this.reuseArraySize = 0;
    }

    public int getSize() {
        return this.size;
    }

    public int getHashSize() {
        return this.hashSize;
    }

    public int getCollisionsMax() {
        return this.collisionsMax;
    }

    public double getCollisionsAvg() {
        if (this.collisionsSum == 0) return 0;
        return (double) this.collisionsSum / this.size;
    }

    private int calcHashMapIndex(long key) {
        int index = (int) (StrictMath.abs(key) % this.hashSize);
        return index;
    }

    // Kritisch? Der Index ist auf 2Gig begrenzt
    private int addNewKeyValue(int prevIndex, long key, long value) {
        // Koennen wir Speicher wiederverwenden?
        int index; // An dieser Stelle werden Key/Value gespeichert.
        if (this.reuseArraySize > 0) {
            // Speicherloecher wiederverwenden.
            index = this.reuseArray.get(--this.reuseArraySize);
        } else { // Sonst hinten anfuegen.
            index = (int) keys.getSize();
        }
        this.size++;
        this.keys.set(index, key);
        this.values.set(index, value);
        if (prevIndex != IntArray.NULLVALUE) {
            this.nexts.set(prevIndex, index);
        }
        return index;
    }

    public void put(long key, long value) {
        int hashIndex = this.calcHashMapIndex(key);
        int hashEntry = this.map.get(hashIndex);
        int listCount = 0;
        if (IntArray.NULLVALUE == hashEntry) {
            int kvIndex = this.addNewKeyValue(IntArray.NULLVALUE, key, value);
            this.map.set(hashIndex, kvIndex);
        } else {
            int kvIndex = hashEntry;
            int lastKvIndex = IntArray.NULLVALUE;
            while (kvIndex != IntArray.NULLVALUE) {
                listCount++;
                long foundKey = this.keys.get(kvIndex);
                if (foundKey == key) {
                    this.values.set(kvIndex, value);
                    return;
                }
                lastKvIndex = kvIndex;
                kvIndex = this.nexts.get(kvIndex);
            }
            this.addNewKeyValue(lastKvIndex, key, value);
            if (listCount > this.collisionsMax) this.collisionsMax = listCount;
            this.collisionsSum += listCount;
        }
    }

    public long get(long key) {
        long foundValue = NULLVALUE;
        int hashIndex = this.calcHashMapIndex(key);
        int kvIndex = this.map.get(hashIndex);

        while (kvIndex != IntArray.NULLVALUE) {
            long foundKey = this.keys.get(kvIndex);
            if (foundKey == key) {
                foundValue = this.values.get(kvIndex);
                break;
            }
            kvIndex = this.nexts.get(kvIndex);
        }

        return foundValue;
    }

    public long remove(long key) {
        int hashIndex = this.calcHashMapIndex(key);
        if (hashIndex != IntArray.NULLVALUE) {
            int kvIndex = this.map.get(hashIndex);
            int lastKvIndex = IntArray.NULLVALUE;
            while (kvIndex != IntArray.NULLVALUE) {
                long foundKey = this.keys.get(kvIndex);
                if (foundKey == key) {
                    int next = this.nexts.get(kvIndex);
                    if (lastKvIndex == IntArray.NULLVALUE) {
                        this.map.set(hashIndex, next);
                    } else {
                        this.nexts.set(lastKvIndex, next);
                    }
                    this.keys.set(kvIndex, LongArray.NULLVALUE);
                    this.reuseArray.set(this.reuseArraySize++, kvIndex);
                    this.size--;
                    return this.values.get(kvIndex);
                }
                lastKvIndex = kvIndex;
                kvIndex = this.nexts.get(kvIndex);
            }
        }
        return NULLVALUE;
    }

}
