/*
 * $Id: BigLongSet.java 5448 2021-11-26 17:17:57Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import de.cm.osm2po.misc.Utils;

/**
 * Ein Hashset fuer kleine primitive longs.
 * <br><b>Wichtig:</b><br>
 * {@link Long#MIN_VALUE} ist der NULL-Wert und darf
 * daher nicht als Schluessel uebergeben werden.
 * <br><br><b>Wichtig:</b><br>
 * Der maximal Anzahl der Elemente ist auf 2Gig begrenzt
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class BigLongSet {

    private LongArray keys = new LongArray();
    private IntArray nexts = new IntArray();

    private int size; // Anzahl neuer Eintraege ohne Ueberschreiben
    private int hashSize; // Array-Groesse der HashMap
    private int collisions; // Anzahl der Kollisionen (gleicher Hash)

    // Dies ist die Eigentliche HashMap
    private IntArray map = new IntArray();
    
    public BigLongSet() {this(0x100000);}
    
    public BigLongSet(int hashSize) {
        this.hashSize = (int) Utils.findNextPrime(hashSize);
    }

    public void clear() {
        this.keys.clear();
        this.nexts.clear();
        this.map.clear();
        
        this.collisions = 0;
        this.size = 0;
    }
    
    public long getSize() {
        return this.size;
    }
    
    public int getHashSize() {
        return this.hashSize;
    }
    
    public int getCollisions() {
        return this.collisions;
    }
    
    private int calcHashMapIndex(long key) {
        int hash = (int) (StrictMath.abs(key) % this.hashSize);
        return hash;
    }
    
    // Kritisch? Der Index ist auf 2Gig begrenzt
    private int addNewKey(int prevIndex, long key) {
        this.size++;
        int index = (int) keys.getSize();

        this.keys.set(index, key);
        if (prevIndex != IntArray.NULLVALUE) {
            this.nexts.set(prevIndex, index);
        }
        return index;
    }
    
    public void add(long key) {
        int hashIndex = this.calcHashMapIndex(key);
        int hashEntry = this.map.get(hashIndex);
        int listCount = 0;
        if (IntArray.NULLVALUE == hashEntry) {
            int index = this.addNewKey(IntArray.NULLVALUE, key);
            this.map.set(hashIndex, index);
        } else {
            int index = hashEntry;
            int lastIndex = IntArray.NULLVALUE;
            while (index != IntArray.NULLVALUE) {
                listCount++;
                long foundKey = this.keys.get(index);
                if (foundKey == key) {
                    return; // Bereits vorhanden
                }
                lastIndex = index;
                index = this.nexts.get(index);
            }
            this.addNewKey(lastIndex, key);
            if (listCount > this.collisions) this.collisions = listCount;
        }
    }
    
    public boolean remove(long key) {
        int hashIndex = this.calcHashMapIndex(key);
        int hashEntry = this.map.get(hashIndex);
        if (hashEntry != IntArray.NULLVALUE) {
            int kvIndex = hashEntry;
            int lastKvIndex = IntArray.NULLVALUE;
            while (kvIndex != IntArray.NULLVALUE) {
                long foundKey = this.keys.get(kvIndex);
                if (foundKey == key) {
                    int next = this.nexts.get(kvIndex);
                    if (lastKvIndex == IntArray.NULLVALUE) {
                        this.map.set(hashIndex, next);
                    } else {
                        this.nexts.set(lastKvIndex, next);
                    }
                    this.keys.set(kvIndex, LongArray.NULLVALUE);
                    this.size--;
                    return true;
                }
                lastKvIndex = kvIndex;
                kvIndex = this.nexts.get(kvIndex);
            }
        }
        return false;
    }
    
    public boolean contains(long key) {
        int hashIndex = this.calcHashMapIndex(key);
        int index = this.map.get(hashIndex);

        while (index != IntArray.NULLVALUE) {
            long foundKey = this.keys.get(index);
            if (foundKey == key) {
                return true;
            }
            index = this.nexts.get(index);
        }
        return false;
    }
    
    public long[] getKeys() {
        long[] ks = new long[this.size];
        int nKeys = (int) this.keys.getSize();
        for (int i = 0, j = 0; i < nKeys; i++) {
            long key = this.keys.get(i);
            if (key != LongArray.NULLVALUE) {
                ks[j++] = key; 
            }
        }
        return ks;
    }
 
}
