/*
 * $Id: ByteArray.java 4712 2018-03-17 15:48:59Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;


/**
 * Dieses Array waechst nach Bedarf und arbeitet auf Blockbasis.
 * Das heisst, es werden nur Speicherbereiche verbraucht, die
 * tatsaechlich indiziert werden.<br>
 * <strong>Wichtig:</strong><br>
 * Dieses Array indiziert mehr als {@link Integer#MAX_VALUE}
 * Bytes. Dennoch wird aus Speichergruenden darauf verzichtet,
 * den kompletten {@link Long} Bereich zu adressieren.
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public final class ByteArray implements Swappable {

    public  static final long MAXMEM = 32 * 0x40000000L; // Maximaler Speicher in GigaBytes.
    private static final int TYPELEN = 1; // Byte hat 1 Byte - Int haette 4 Bytes
    private static final int BLOCKBITS = 20; // 1M
    private static final int BLOCKSIZE = 1 << BLOCKBITS; // BlockSize = 1M
    private static final long BLOCKMASK = BLOCKSIZE - 1; // 20 Bits Mask
    private static final int SLOTSIZE = (int) (MAXMEM / (TYPELEN * BLOCKSIZE));

    private long maxIdx = -1;
    private byte[][] slots = new byte[SLOTSIZE][];

    private Swappable swappable;

    public ByteArray(Swappable swappable) {
        this.swappable = swappable;
    }

    public ByteArray() {
    }

    public long getSize() {
        return this.maxIdx +1;
    }

    @Override
    public void swap(long i, long j) {
        byte tmp = this.get(i);
        this.set(i, this.get(j));
        this.set(j, tmp);
        if (this.swappable != null) swappable.swap(i, j);
    }

    public void push(byte value) {
        this.set(this.maxIdx +1, value);
    }

    public void clear() {
        for (int i = 0; i < SLOTSIZE; i++) {
            this.slots[i] = null;
        }
        this.maxIdx = -1;
    }

    public byte get(long idx) {
        if (idx < 0)
            throw new ArrayIndexOutOfBoundsException(String.valueOf(idx));
        int slotIdx = (int) (idx >>> BLOCKBITS);
        byte[] block = this.slots[slotIdx];
        if (null == block) return 0;
        int blockIdx = (int) (idx & BLOCKMASK);
        return block[blockIdx];
    }

    public void set(long idx, byte value) {
        if (idx < 0)
            throw new ArrayIndexOutOfBoundsException(String.valueOf(idx));
        int slotIdx = (int) (idx >>> BLOCKBITS);
        byte[] block = this.slots[slotIdx];
        if (null == block) {
            block = new byte[BLOCKSIZE];
            this.slots[slotIdx] = block;
        }
        if (idx > this.maxIdx) this.maxIdx = idx;
        int blockIdx = (int) (idx & BLOCKMASK);
        block[blockIdx] = value;
    }

    public int get(long idx, byte[] bytes) {
        if (idx < 0)
            throw new ArrayIndexOutOfBoundsException(String.valueOf(idx));

        int len = bytes.length;
        long lastIdx = idx + len -1;
        if (lastIdx > this.maxIdx)
            len -= (lastIdx - this.maxIdx);

        int nCopied = 0;
        int slotIdx = (int) (idx >>> BLOCKBITS);
        int blockIdx = (int) (idx & BLOCKMASK);

        while (nCopied < len) {
            byte[] block = this.slots[slotIdx];
            int n = BLOCKSIZE - blockIdx;
            if (len - nCopied < n) n = len - nCopied;

            System.arraycopy(block, blockIdx, bytes, nCopied, n);

            nCopied += n;
            slotIdx++;
            blockIdx = 0;
        }

        return nCopied;
    }

    public void set(long idx, byte[] bytes, int len) {
        if (idx < 0)
            throw new ArrayIndexOutOfBoundsException(String.valueOf(idx));
        if (bytes.length < len)
            throw new IllegalArgumentException("len > bytes.length");

        int nCopied = 0;
        int slotIdx = (int) (idx >>> BLOCKBITS);
        int blockIdx = (int) (idx & BLOCKMASK);

        while (nCopied < len) {
            byte[] block = this.slots[slotIdx];
            if (null == block) {
                block = new byte[BLOCKSIZE];
                this.slots[slotIdx] = block;
            }
            int n = BLOCKSIZE - blockIdx;
            if (len - nCopied < n) n = len - nCopied;

            System.arraycopy(bytes, nCopied, block, blockIdx, n);

            nCopied += n;
            slotIdx++;
            blockIdx = 0;
        }

        if (idx + nCopied > this.maxIdx)
            this.maxIdx = idx + nCopied;
    }

}
