/*
 * $Id: Dictionary.java 5265 2021-04-10 08:57:14Z carsten $
 *
 * Copyright 2016 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static de.cm.osm2po.misc.ArrayUtils.intArrayConcat;
import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;
import static de.cm.osm2po.misc.ArrayUtils.intArrayListToArray2D;
import static java.util.Arrays.copyOf;

import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.postp.GeoExtensionBuilderBase;
import de.cm.osm2po.postp.SndExtensionBuilder;

/**
 * Ein Bidi-Bag (Dictionary) fuer grosse Datenmengen.<br>
 * Die add()-Methode erwartet einen einfachen Schluessel (int)
 * und den Wert (String), so dass intern eine M-zu-N-Beziehung ueber alle
 * Assoziationen entsteht. Diese kann dann von beiden Seiten her
 * abgefragt werden.
 * @author (c) 2020 - Carsten Moeller, info@osm2po.de
 * @see SndExtensionBuilder
 * @see GeoExtensionBuilderBase
 */
public class Dictionary {

    private Log log;
    boolean buildMode;
    private int nIds;
    private int[] ids;
    private File tmpFile;
    private String lastAddedName;

    private VarMergeSorter<VarStringInts> nameSorter;
    private VarMergeSorter<VarIntInt> idSorter;

    private Index<VarStringInts> nameIndex = new IndexOpt<VarStringInts>();
    private Index<VarIntInts> idIndex = new IndexOpt<VarIntInts>();

    public Dictionary create(File workDir, Log log) {
        if (null == log) log = Log.stdout();
        this.log = log;
        this.buildMode = true;

        this.tmpFile = Utils.createTmpFile(workDir);

        this.nameSorter = new VarMergeSorter<>(VarStringInts.class, 0x400000/*4M*/, workDir);
        this.idSorter = new VarMergeSorter<>(VarIntInt.class, 0x1000000/*16M*/, workDir);
        this.nameIndex = new IndexOpt<VarStringInts>().create(workDir, log);
        this.idIndex = new IndexOpt<VarIntInts>().create(workDir, log);

        this.nIds = 0; // Number of Deferred/Pending after add()
        this.ids = new int[1];

        return this;
    }

    public void connect(RandomRead rr) {
        long idxIndexOffset = rr.readLong();
        long offset = rr.getReadIndex();
        this.nameIndex.connect(rr);
        synchronized (rr) {rr.setReadIndex(offset + idxIndexOffset);}
        this.idIndex.connect(rr);
    }

    /**
     * Fuegt ein neues Key-Value-Pair hinzu.
     * @param id int Id bzw. Key
     * @param name String Wert
     * @see #get(int)
     * @see #get(String)
     */
    public void add(int id, String name) {
        if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");

        // log.progress("add(" + id + ", " + var + ")");
        if (name != null && name.equals(this.lastAddedName)) {
            if (this.ids.length == this.nIds) // ArrayOverflow
                this.ids = copyOf(this.ids, nIds * 2);
        } else if (this.nIds > 0) {
            int[] ids = copyOf(this.ids, this.nIds);
            
            VarStringInts kv = new VarStringInts(this.lastAddedName, ids);

            this.nameSorter.add(kv);
            this.nIds = 0;
            this.lastAddedName = null;
        }
        this.ids[this.nIds++] = id;
        this.lastAddedName = name;
    }

    public void build(OutStream os) {
        if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");

        log.debug("Building " + this.getClass().getSimpleName()
                + " at " + new Date().toString() + " - " + Utils.MB());

        OutStream osTmp = new OutStreamDisk(this.tmpFile);

        if (this.nIds > 0) {
            this.add(0, null); // Flush

            log.debug("Building Value-Lookup-Index");
            this.lastAddedName = null;
            List<int[]> idList = new ArrayList<int[]>();
            for (VarStringInts kv : this.nameSorter) {
                if (this.lastAddedName != null && !kv.getKey().equals(this.lastAddedName)) {
                    int[][] arr2D = intArrayListToArray2D(idList);
                    idList.clear();
                    int[] arr1D = intArrayConcat(arr2D);
                    arr1D = intArrayDistinct(arr1D);
                    this.ids = arr1D;
                    this.writeToNameIndexAndIndexSorter();
                    this.ids = null;
                }
                this.lastAddedName = kv.getKey();
                idList.add(kv.getVal());
            }
            int[][] arr2D = intArrayListToArray2D(idList);
            idList.clear();
            int[] arr1D = intArrayConcat(arr2D);
            arr1D = intArrayDistinct(arr1D);
            this.ids = arr1D;
            this.writeToNameIndexAndIndexSorter();
            this.ids = null;
            this.nameSorter.clear();
        }

        // Umweg ueber TempFile weil Groesse des ersten Index
        // bei connect fuer Offset des zweiten Index bekannt sein muss.
        this.nameIndex.build(osTmp);
        // Header schreiben
        os.writeLong(osTmp.getWriteIndex()); // Groesse des Index in Bytes
        osTmp.close();

        // Index aus TmpFile uebertragen
        byte[] buf = new byte[0x1000];
        InStream is = new InStreamDisk(this.tmpFile);
        while (!is.isEof()) {int n = is.readBytes(buf); os.writeBytes(buf, n);}
        is.close();
        this.tmpFile.delete();

        // Der zweite Index kann direkt geschrieben werden
        log.debug("Building Id-Lookup-Index");
        int lastKey = Integer.MIN_VALUE;
        this.ids = new int[1];
        this.nIds = 0;
        for (VarIntInt kv : this.idSorter) {
            if (lastKey != Integer.MIN_VALUE && kv.getKey() != lastKey) {
                this.ids = copyOf(this.ids, this.nIds);
                this.ids = intArrayDistinct(this.ids);
                this.idIndex.push(new VarIntInts(lastKey, ids));
                this.nIds = 0;
            }
            lastKey = kv.getKey();
            if (this.ids.length == this.nIds) // ArrayOverflow
                this.ids = copyOf(this.ids, nIds * 2);
            this.ids[this.nIds++] = kv.getVal();
        }
        if (this.nIds > 0) { // Pending schreiben und flushen
            this.ids = copyOf(this.ids, this.nIds);
            this.ids = intArrayDistinct(this.ids);
            this.idIndex.push(new VarIntInts(lastKey, ids));
            this.nIds = 0;
        }
        this.idIndex.build(os);
        this.idSorter.clear();

        log.debug(this.getClass().getSimpleName() + " created at "
                + new Date().toString() + " - " + Utils.MB());

        this.buildMode = false;
    }


    /**
     * Build-Methode, die einen Var-Index-Eintrag schreibt und
     * dabei gleichzeitig die IDs fuer den naechsten Schritt
     * vorbereitet.
     */
    private void writeToNameIndexAndIndexSorter() {
        VarStringInts kvOut = new VarStringInts(this.lastAddedName, this.ids);
        // log.progress(kvOut.toString());
        int pos = this.nameIndex.push(kvOut);

        // Transfer association to idxSorter
        for (int i = 0; i < this.ids.length; i++) {
            int idx = this.ids[i];
            this.idSorter.add(new VarIntInt (idx, pos));
        }
    }


    /**
     * Holt einen Eintrag aufgrund seines internen Adress-Index und liefert
     * so den Wert und die mittels {@link #add(int, String)} assoziierten IDs.
     * @param idx int Index/Adresse
     * @return {@link VarStringInts}, also Name und assozizierte IDs oder NULL.
     * @see #get(int)
     */
    public final VarStringInts fetch(int idx) {
        if (idx < 0 ||idx >= this.getIndexSize()) return null;
        else return this.nameIndex.fetch(idx, new VarStringInts());
    }

    /**
     * Liefert die Anzahl der sortiert und eindeutigen Namen im Index.
     * @return int Anzahl gespeicherter Namen.
     */
    public final int getIndexSize() {
        return this.nameIndex.getIndexSize();
    }

    /**
     * Findet den ersten internen Adress-Index zu einem Vergleichswert.
     * <br><br><b>Wichtig:</b><br>
     * Findet diese Methode keine Uebereinstimmung, so wird dennoch
     * eine Adresse &gt;= 0 geliefert. In diesem Fall bezeichnet sie
     * die Stelle, an der sich der gesuchte Wert haette befinden muessen.
     *
     * @param searchName final String Suchwert
     * @return int adresse &gt;= 0 und &lt;= {@link #getIndexSize()} <b>!!!</b>
     * @see #fetch(int)
     */
    public final int find(final String searchName) {
        if (this.buildMode) throw new IllegalStateException("Not in Runtime-Mode");
        return this.searchFirstVal(searchName);
    }

    /**
     * Liefert alle internen Adressen zu Werten, die mit einem
     * bestimmten Schluessel hinzugefuegt wurden.
     * @param id int Schluessel aus {@link #add(int, String)}
     * @return Array von int-Adressen oder NULL.
     */
    public final int[] get(int id) {
        if (this.buildMode) throw new IllegalStateException("Not in Runtime-Mode");

        int pos = this.searchIdx(id);
        if (pos < 0) return null;

        VarIntInts kv = this.idIndex.fetch(pos, new VarIntInts());

        if (null == kv || id != kv.getKey()) return null;

        int[] idxs = kv.getVal();
        if (null == idxs || 0 == idxs.length) return null;

        return idxs;
    }

    /**
     * Liefert alle IDs bzw. Schluessel, unter denen ein Wert
     * mittels {@link #add(int, String)} hinzugefuegt wurde.
     * @param name {@link String} gesuchter Wert NULLABLE
     * @return Array von int aller assozierter IDs oder NULL.
     */
    public final int[] get(final String name) {
        if (this.buildMode) throw new IllegalStateException("Not in Runtime-Mode");
        if (null == name) return null;

        int pos = this.searchFirstVal(name);
        
        VarStringInts kv = this.nameIndex.fetch(pos, new VarStringInts());

        if (null == kv || !name.equals(kv.getKey())) return null;

        int[] ids = kv.getVal();
        if (null == ids || 0 == ids.length) return null;

        return ids;
    }

    /**
     * Findet die erste Position zu einem Vergleichswert.
     * <br><br><b>Wichtig:</b><br>
     * Findet diese Methode keine Uebereinstimmung, so wird dennoch
     * eine Position >= 0 geliefert. In diesem Fall bezeichnet sie
     * die Stelle, an der sich der gesuchte Wert haette befinden muessen.
     * Der gefundene Wert befindet sich danach in var selbst.
     * @param id int Suchwert
     * @return int position &gt;= 0 und &lt;= {@link Index#getIndexSize()} <b>!!!</b>
     */
    private int searchIdx(int id) {
        int left = 0;
        int right = this.idIndex.getIndexSize() -1;
        VarIntInts kv = new VarIntInts();
        return this.binarySearchIdx(id, left, right, kv);
    }

    private int binarySearchIdx(int id, int left, int right, VarIntInts kv) {
        if (right < left) return left;
        int mid = (left + right) >>> 1;
        int x = this.idIndex.fetch(mid, kv).getKey();
        if (id > x) return binarySearchIdx(id, mid +1, right, kv);
        else return binarySearchIdx(id, left, mid -1, kv);
    }

    /**
     * Findet die erste Position zu einem Vergleichswert.
     * <br><br><b>Wichtig:</b><br>
     * Findet diese Methode keine Uebereinstimmung, so wird dennoch
     * eine Position &gt;= 0 geliefert. In diesem Fall bezeichnet sie
     * die Stelle, an der sich der gesuchte Wert haette befinden muessen.
     * Der gefundene Wert befindet sich danach in var selbst.
     *
     * @param searchValue final String Suchwert
     * @return int position &gt;= 0 und &lt;= {@link Index#getIndexSize()} <b>!!!</b>
     */
    private int searchFirstVal(final String searchValue) {
        int left = 0;
        int right = this.nameIndex.getIndexSize() -1;
        VarStringInts compareValueHolder = new VarStringInts(searchValue);
        return this.binarySearchVal(searchValue, compareValueHolder, left, right);
    }

    private int binarySearchVal(String searchValue,
            VarStringInts compareValueHolder, int left, int right) {
        
        if (right < left) return left;
        int mid = (left + right) >>> 1;
        
        String v = this.nameIndex.fetch(mid, compareValueHolder).getKey();
        if (searchValue.compareTo(v) > 0) return binarySearchVal(
                searchValue, compareValueHolder, mid +1, right);
        else return binarySearchVal(searchValue, compareValueHolder, left, mid -1);
    }

    
    /**
     * Fuehrt eine StartsWith-Suche durch und liefert die Ids.
     * @param name String Name NULLABLE
     * @return int Array von assoziierten Ids oder null
     */
    public final int[] findByStartsWith(final String name) {
        int[] foundIds = null;
        if (name != null) for (int idx = this.find(name);;) {
            VarStringInts found = this.fetch(idx++);
            if (null == found || !found.getKey().startsWith(name)) break;
            foundIds = intArrayConcat(foundIds, found.getVal());
        }
        return foundIds;
    }
}
