/*
 * $Id: FastIntSet.java 4310 2016-10-19 17:28:35Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.math.BigInteger;
import java.util.Arrays;


/**
 * Ein Hashset fuer kleine primitive ints.
 * Allerdings mit der Besonderheit, dass hier keine
 * Kollisionen behandelt werden, sondern stattdessen
 * das Hash-Array solange vergroessert wird, bis
 * es kollisionsfrei ist. Das heisst im Klartext,
 * dass hier nur kleine Mengen verarbeitet werden
 * koennen. Der Vorteil liegt dabei im Zugriff,
 * der konstant O(1) ist und ueber eine
 * einfache Modulo-Operation laeuft.
 * 
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class FastIntSet {

    public final static int NULLVALUE = Integer.MIN_VALUE; 

    private int[] hash;
    private int prime;
    private int nValues;
    
    public FastIntSet(int[] values) {
        this.nValues = values.length;
        this.prime = this.nValues * 10;
        for (;;) {
            this.prime = (int) this.findNextPrime(this.prime);
            this.hash = new int[this.prime];
            Arrays.fill(this.hash, NULLVALUE);
            int i;
            for (i = 0; i < this.nValues; i++) {
                int hashIndex = values[i] % this.prime;
                if (this.hash[hashIndex] != NULLVALUE) break;
                this.hash[hashIndex] = values[i];
            }
            if (i == this.nValues) break;
        }
    }
    
    public boolean contains(int value) {
        int hashIndex = value % this.prime;
        return this.hash[hashIndex] == value;
    }
    
    public boolean remove(int value) {
        int hashIndex = value % this.prime;
        if (this.hash[hashIndex] == value) {
            this.hash[hashIndex] = NULLVALUE;
            this.nValues--;
            return true;
        }
        return false;
    }
    
    public boolean isEmpty() {
        return this.nValues == 0;
    }
    
    public int[] getHash() {return this.hash;}
    public int getPrime() {return this.prime;}
 
    /**
     * @param longNumber long zahl
     * @return Primzahl > longNumber
     */
    private long findNextPrime(long longNumber){
        final BigInteger bigTwo = BigInteger.valueOf(2);
        BigInteger bigi = BigInteger.valueOf((longNumber+1)|1);
        while(!bigi.isProbablePrime(128)) bigi = bigi.add(bigTwo);
        return bigi.longValue();
    }

}
