/*
 * $Id: InStreamAdapter.java 2355 2013-06-05 10:24:28Z cmindividual $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.io.BufferedInputStream;
import java.io.InputStream;

/**
 * Wrapped einen beliebigen {@link InputStream} fuer
 * die Verwendung mit {@link InStream}.
 * Verwendet intern zusaetzlich einen Buffer,
 * was den Einsatz eines {@link BufferedInputStream}
 * ueberfluessig macht.
 * 
 * @author (c) 2012 - Carsten Moeller - info@osm2po.de
 */
public class InStreamAdapter implements InStream {

    private InputStream inputStream;
    private byte[] buf;
    private int bufFillSize; // Aktuelle Fuellung des Puffers
    private int bufIndex;
    private long offset;
    
    public InStreamAdapter(InputStream inputStream) {
        this.inputStream = inputStream;
        this.buf = new byte[0x8000/*32k*/];
    }
    
    @Override
    public void close() {
        try {
            this.inputStream.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        this.buf = null;
    }

    
    @Override
    public long getReadIndex() {
        return this.offset + this.bufIndex;
    }

    @Override
    public boolean isEof() {
        if (this.bufIndex >= this.bufFillSize) {
            this.fillBuffer();
            return this.bufFillSize == 0;
        } else {
            return false;
        }
    }

    @Override
    public byte readByte() {
        if (this.bufFillSize - this.bufIndex < 1) {
            this.fillBuffer();
            if (this.bufFillSize < 1) {
                throw new IllegalStateException(
                        "Buffer Overflow " + this.bufFillSize);
            }
        }
        return this.buf[bufIndex++];
    }

    @Override
    public short readShort() {
        if (this.bufFillSize - this.bufIndex < 2) {
            this.fillBuffer();
            if (this.bufFillSize < 2) {
                throw new IllegalStateException(
                        "Buffer Overflow " + this.bufFillSize);
            }
        }
        short s = Bits.getShort(this.buf, this.bufIndex);
        this.bufIndex += 2;
        return s;
    }

    @Override
    public int readInt() {
        if (this.bufFillSize - this.bufIndex < 4) {
            this.fillBuffer();
            if (this.bufFillSize < 4) {
                throw new IllegalStateException(
                        "Buffer Overflow " + this.bufFillSize);
            }
        }
        int i = Bits.getInt(this.buf, this.bufIndex);
        this.bufIndex += 4;
        return i;
    }

    @Override
    public long readLong() {
        if (this.bufFillSize - this.bufIndex < 8) {
            this.fillBuffer();
            if (this.bufFillSize < 8) {
                throw new IllegalStateException(
                        "Buffer Overflow " + this.bufFillSize);
            }
        }
        long l = Bits.getLong(this.buf, this.bufIndex);
        this.bufIndex += 8;
        return l;
    }

    @Override
    public int readBytes(byte[] buffer) {
        try {
            int nRest = this.bufFillSize - this.bufIndex;
            int nFlush = buffer.length <= nRest ? buffer.length : nRest;
            System.arraycopy(this.buf, this.bufIndex, buffer, 0, nFlush);
            this.bufIndex += nFlush;
            
            nRest = buffer.length - nFlush;
            if (nRest == 0) return nFlush; // Bytes nur aus Puffer gelesen.
            
            // Bytes nachladen
            int nRead = this.inputStream.read(buffer, nFlush, nRest);
            if (nRead < 0) nRead = 0; // -1 EOF
            
            this.offset += nRead;
            
            return nFlush + nRead;
            
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        
    }

    private void fillBuffer() {
        try {
            this.bufFillSize -= this.bufIndex;
            // ungelesene Bytes hinten im Puffer nach vorne kopieren
            System.arraycopy(this.buf, this.bufIndex, this.buf, 0, this.bufFillSize);
            
            // Dahinter wieder den Puffer fuellen.
            int bufSize = this.buf.length;
            int nRest = bufSize - this.bufFillSize;
            while (nRest > 0) {
                int nRead = inputStream.read(
                        this.buf, bufSize - nRest, nRest);
                if (nRead < 0) { // -1 EOF
                    nRest = 0;
                } else {
                    this.bufFillSize += nRead;
                    nRest -= nRead;
                }
            }
            this.offset += this.bufIndex;
            this.bufIndex = 0;
            
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void skip(long n) {
        try {
            int nRest = this.bufFillSize - this.bufIndex;
            if (n <= nRest) {
                this.bufIndex += n;
            } else {
                this.bufIndex = this.bufFillSize;
                n -= nRest;
                while (n > 0) {
                    long nSkipped = this.inputStream.skip(n); 
                    n -= nSkipped;
                    this.offset += nSkipped;
                }
            }

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
}
