/*
 * $Id: InStreamAdapterBuffered.java 4713 2018-03-19 07:29:18Z carsten $
 *
 * Copyright 2014 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;

public class InStreamAdapterBuffered extends Thread implements InStream {

    private final static int BUF_SIZE_DEFAULT = 0x100000; // 1M

    private byte[] fgBuf, bgBuf; // Foreground-, Background-Buffer
    private int fgBufSize, bgBufSize; // Fillsizes of Buffers
    private int fgIdx; // Read Pointer of forgeground buffer
    private long fgIdxHistory; // Number of bytes in already read buffers

    private InputStream is;

    private volatile boolean loadSignal; // Backgound is reading data
    private volatile boolean closeSignal;

    private void construct(InputStream is, int bufferSize) {
        fgBuf = new byte[bufferSize]; // Foreground read buffer
        bgBuf = new byte[bufferSize]; // Background load buffer

        try {
            this.is = is;
            bgBufSize = is.read(bgBuf);
            this.setName(this.getClass().getSimpleName() + "Thread");

            start();
            swapBuffersAndLoad();

        } catch (Throwable t) {
            throw new RuntimeException(t);
        }
    }

    public InStreamAdapterBuffered(InputStream is) {
        this.construct(is, BUF_SIZE_DEFAULT);
    }

    public InStreamAdapterBuffered(InputStream is, int bufferSize) {
        this.construct(is, bufferSize);
    }

    public InStreamAdapterBuffered(File file) {
        try {
            FileInputStream fis = new FileInputStream(file);
            this.construct(fis, BUF_SIZE_DEFAULT);
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final boolean isEof() {
        return fgBufSize < 0;
    }

    @Override
    public final long getReadIndex() {
        return fgIdxHistory + fgIdx;
    }

    @Override
    public final synchronized void close() {
        try {
            ensureBgLoaded();
            closeSignal = true;
            notify();
            join();
            is.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final void run() {
        try {
            synchronized (this) {
                do {
                    while (!(loadSignal || closeSignal)) wait();
                    if (loadSignal) {
                        bgBufSize = is.read(bgBuf);
                        loadSignal = false;
                        notify();
                    }
                } while (!closeSignal);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Stellt sicher dass, der Hintergrundpuffer vollstaendig
     * geladen wird und somit die beteiligten Variablen
     * nicht auf dem vorigen Wert stehen.
     * @throws InterruptedException
     */
    private final synchronized void ensureBgLoaded()
    throws InterruptedException {
        while (loadSignal) wait();
    }

    private final synchronized void swapBuffersAndLoad() {
        try {
            ensureBgLoaded();
            loadSignal = true;

            // Puffer tauschen
            byte[] buf = fgBuf;
            fgBuf = bgBuf;
            bgBuf = buf;
            fgBufSize = bgBufSize;

            fgIdxHistory += fgIdx;
            fgIdx = 0;

            notify();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final byte readByte() {
        byte b = fgBuf[fgIdx++];
        if (fgIdx == fgBufSize) swapBuffersAndLoad();
        return b;
    }

    @Override
    public final short readShort() {
        short s;
        if (fgBufSize - fgIdx > 2) {
            s = Bits.getShort(fgBuf, fgIdx);
            fgIdx += 2;
        } else {
            s = (short) (((readByte() & 0xFF) << 8)
                    | (readByte() & 0xFF));
        }
        return s;
    }

    @Override
    public final int readInt() {
        int i;
        if (fgBufSize - fgIdx > 4) {
            i = Bits.getInt(fgBuf, fgIdx);
            fgIdx += 4;
        } else {
            i = (((readByte() & 0xFF) << 24)
                | ((readByte() & 0xFF) << 16)
                | ((readByte() & 0xFF) << 8)
                | (readByte() & 0xFF));
        }
        return i;
    }

    @Override
    public final long readLong() {
        long l;
        if (fgBufSize - fgIdx > 8) {
            l = Bits.getLong(fgBuf, fgIdx);
            fgIdx += 8;
        } else {
            return (((readByte() & 0xFFL) << 56)
                    | ((readByte() & 0xFFL) << 48)
                    | ((readByte() & 0xFFL) << 40)
                    | ((readByte() & 0xFFL) << 32)
                    | ((readByte() & 0xFFL) << 24)
                    | ((readByte() & 0xFFL) << 16)
                    | ((readByte() & 0xFFL) << 8)
                    | (readByte() & 0xFFL));
        }
        return l;
    }

    @Override
    public final void skip(long n) {
        try {
            ensureBgLoaded();

            long nRest = n;
            int nBuffered = fgBufSize - fgIdx;
            long nAvailable = (nRest <= nBuffered ? nRest : nBuffered);
            nRest -= nAvailable;
            fgIdx += nAvailable;

            if (nRest == 0) { // Nur Foreground betroffen
                if (fgIdx == fgBufSize) swapBuffersAndLoad();
            } else if (nRest <= bgBufSize) { // Auch Background betroffen
                swapBuffersAndLoad();
                fgIdx = (int) nRest;
                if (fgIdx == fgBufSize) swapBuffersAndLoad();
            } else { // Beides betroffen
                fgIdxHistory += nRest;
                nRest -= bgBufSize;
                while (nRest > 0) {
                    long nSkipped = is.skip(nRest);
                    if (0 == nSkipped) throw new EOFException(
                            "Failed to skip " + nRest + " bytes");
                    nRest -= nSkipped;
                }
                bgBufSize = is.read(bgBuf); // Neu laden
                swapBuffersAndLoad();
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final int readBytes(byte[] buffer) {

        int nRead = 0;
        int nRest = buffer.length;

        while (nRest > 0 && fgBufSize != -1) {
            int nBuffered = fgBufSize - fgIdx;
            int nAvailable = nRest <= nBuffered ? nRest : nBuffered;
            System.arraycopy(fgBuf, fgIdx, buffer, nRead, nAvailable);
            fgIdx += nAvailable;
            nRest -= nAvailable;
            nRead += nAvailable;
            if (fgIdx == fgBufSize) swapBuffersAndLoad();
        }

        return nRead;
    }

}
