/*
 * $Id: Index.java 4577 2017-12-14 16:30:52Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.io.File;

import de.cm.osm2po.logging.Log;

/**
 * Interface fuer indizierte Listen von z.B. {@link Var}-Objekten
 * innerhalb eines {@link RandomRead} oder auch anderen
 * Anwendungsfaellen, wo serialisierte Objekte keine konstante
 * Groesse besitzen.<br>
 * Klassen, die hiervon erben, bieten sowohl Build- als
 * auch Laufeit-Methoden.
 * <br><br><b>Wichtig:</b><br>
 * Indizes werden hier nicht gesetzt, sondern fortlaufend
 * vergeben.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 * @see #push(Var)
 * @see IndexRaw
 * @param <T> Class of Var
 */
public interface Index<T extends Var<?>> {
    /**
     * Laufzeit-Methode, um eine Verbindung zu einem
     * geoeffneteten Daten-Container herzustellen.
     * <br><br><b>Wichtig:</b><br>
     * {@link RandomAccess#setReadIndex(long)} muss
     * richtig gesetzt sein.
     * @param rr {@link RandomAccess}
     */
    void connect(RandomRead rr);

    /**
     * Liefert die Groesse des Index sowohl zur
     * Build- als auch zur Laufzeit.
     * @return int Anzahl der Indizes, bzw. der Elemente
     */
    int getIndexSize();

    /**
     * Laufzeit-Methode, die ein {@link Var} aufgrund eines Index liefert,
     * der implizit mittels {@link #push(Var)} aufsteigend vergeben wurde.
     * @param idx int Index 0-wertig.
     * @param var {@link Var} in den der Wert gelesen werden soll. NOT NULL.
     * @return der uebergebene Wert selbst oder null, wenn idx out of range.
     */
    T fetch(int idx, T var);

    /**
     * Initialer Quasi-Konstruktor fuer den Build-Modus.
     * @param workDir File optionales Arbeitsverzeichnis NULLABLE
     * @param log {@link Log} optionaler Logger NULLABLE
     * @return this 4Chaining
     * @see #build(OutStream)
     */
    Index<T> create(File workDir, Log log);

    /**
     * Build-Methode, die ein {@link Var}-Objekt vermerkt und es
     * dabei mit den naechsten Index (Position +1) assoziiert.
     * @param var {@link Var} der Wert.
     * @return int Ablage-Index, normalerweise aktuelle Groesse -1.
     * @see #create(File, Log)
     * @see #build(OutStream)
     * @see #fetch(int, Var)
     */
    int push(T var);

    /**
     * Build-Methode, die abschliessend den gesamten Index
     * inklusive aller Werte in den Ausgabe-Strom schreibt.
     * @param os {@link OutStream}
     * @see #create(File, Log)
     */
    void build(OutStream os);
}