/*
 * $Id: IndexOpt.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static de.cm.osm2po.misc.Utils.createTmpFile;
import static java.util.Arrays.copyOf;

import java.io.File;

import de.cm.osm2po.logging.Log;

/**
 * Diese Implementation von {@link Index} arbeitet analog
 * {@link IndexRaw}, jedoch werden Indizes mittels eines
 * vorgeschalteten Offset-Arrays als Ints (4 bytes) und nicht als
 * Longs (8 bytes) gespeichert, was eine erhebliche Optimierung
 * des Speicherverbrauchs darstellt.
 * <br><br><b>Wichtig:</b><br>
 * Die Kompression der Offsets auf Ints beruht auf der Annahme,
 * dass zu serialisierende Objekte eine bestimmte Groesse nicht
 * ueberschreiten. Konkret heisst dies hier, dass 1 Mio Elemente
 * serialisiert nicht groesser als MaxInt, also 2 Gig sein duerfen
 * und somit den indizierbaren Raum nicht verlassen.<br>
 * Nach jeweils 1 Mio Elemente wird der interne Offset erhoeht
 * und das gleiche Spiel beginnt von vorne. 
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 * @see IndexSeqOpt
 * @see IndexRaw
 * @param <T> Class of Var
 */
public class IndexOpt<T extends Var<?>> implements Index<T> {
    private static final int BLOCK_SIZE = 0x100000; // 1MB;
    
    private File varFile;
    private File idxFile;
    private long offsIdx;
    private long offsVar;
    private RandomRead rr;
    
    protected boolean buildMode;
    protected OutStream osVar;
    protected OutStream osIdx;
    protected int size;
    protected long[] deltas;
    protected int deltasIdx;
    protected int blockSize;
    
    @Override
    public Index<T> create(File workDir, Log log) {
        this.buildMode = true;
        this.varFile = createTmpFile(workDir);
        this.idxFile = createTmpFile(workDir);
        this.osVar = new OutStreamDisk(this.varFile);
        this.osIdx = new OutStreamDisk(this.idxFile);
        
        this.size = 0;
        this.blockSize = 0;
        this.deltas = new long[1];
        this.deltasIdx = 0;
        return this;
    }

    @Override
    public void connect(RandomRead rr) {
        int deltasLength = rr.readInt();
        this.deltas = new long[deltasLength];
        for (int i = 0; i < deltasLength; i++) {
            this.deltas[i] = rr.readLong();
        }
        this.size = rr.readInt();
        this.offsIdx = rr.getReadIndex();
        this.offsVar = this.offsIdx + (this.size << 2); // 1int=4bytes
        this.rr = rr;
    }

    @Override
    public int getIndexSize() {return this.size;}
	
    @Override
    public T fetch(int idx, T var) {
		if (this.buildMode) throw new IllegalStateException("Not in Runtime-Mode");
        if (idx < 0 || idx >= this.size) return null;
		synchronized (this.rr) {
			this.rr.setReadIndex(this.offsIdx + (idx << 2));
			int intPtr = this.rr.readInt();
			int deltaIdx = idx / BLOCK_SIZE;
			long ptr = intPtr + this.deltas[deltaIdx];
			this.rr.setReadIndex(this.offsVar + ptr);
			var.readFromStream(this.rr);
		}
		return var;
    }
    
    @Override
	public int push(T var) {
		if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");
	    long ptr = this.osVar.getWriteIndex();
	    var.writeToStream(this.osVar);

        if (this.blockSize >= BLOCK_SIZE) {
            this.deltasIdx++;
            if (this.deltas.length == this.deltasIdx)
                this.deltas = copyOf(this.deltas, this.deltas.length * 2);
            this.deltas[this.deltasIdx] = ptr;
            this.blockSize = 0;
        }
        
        int intPtr = (int) (ptr - deltas[this.deltasIdx]);
        this.osIdx.writeInt(intPtr);

        this.blockSize++;
        int idx = this.size++;
        return idx;
	}

    @Override
	public void build(OutStream os) {
		if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");
		InStream is;
		this.osIdx.close();
		this.osVar.close();
        int deltasSize = this.deltasIdx +1;
        os.writeInt(deltasSize);
        for (int i = 0; i < deltasSize; i++) os.writeLong(this.deltas[i]);
        os.writeInt(this.size);
		byte[] buf = new byte[0x1000];
    	is = new InStreamDisk(this.idxFile);
    	while (!is.isEof()) {int n = is.readBytes(buf);	os.writeBytes(buf, n);}
		is.close();
		this.idxFile.delete();
		is = new InStreamDisk(this.varFile);
    	while (!is.isEof()) {int n = is.readBytes(buf);	os.writeBytes(buf, n);}
		is.close();
		this.varFile.delete();
		this.buildMode = false;
	}

}
