/*
 * $Id: IndexSeqOpt.java 5244 2021-02-10 19:53:01Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static java.util.Arrays.copyOf;

/**
 * Diese Implementation von {@link Index} arbeitet analog
 * {@link IndexSeq}, eliminiert also aufeinanderfolgende
 * gleiche {@link Var}-Werte, zusaetzlich werden jedoch Indizes
 * mittels eines vorgeschalteten Offset-Arrays als Ints (4 bytes)
 * und nicht als Longs (8 bytes) gespeichert, was eine erhebliche
 * Optimierung des Speicherverbrauchs darstellt.
 * <br><br><b>Wichtig:</b><br>
 * Damit dieser Mechanismus sinnvoll funktioniert, muessen verwendete
 * {@link Var}-Klassen {@link #equals(Object)} ueberschreiben.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 * @see IndexSeq
 * @param <T> Class of Var
 */
public class IndexSeqOpt<T extends Var<?>> extends IndexOpt<T> {
    private static final int BLOCK_SIZE = 0x100000; // 1MB;

	private T var;
    private int nPending;

    @Override
	public int push(T var) {
		if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");
		if (null == this.var) {
            try {
                // called once and checks for an implented equals method
                var.getClass().getDeclaredMethod("equals", Object.class);
            } catch (Exception e) {
                throw new IllegalArgumentException(
                        "Class '" + var.getClass().getSimpleName()
                        + "' does not provide equals()", e);
            }
            this.var = var;
		}
		if (this.var.equals(var)) this.nPending++;
		else {
		    long ptr = this.osVar.getWriteIndex();
		    this.var.writeToStream(this.osVar);

		    for (int i = 0; i < this.nPending; i++) {
                if (this.blockSize >= BLOCK_SIZE) {
                    this.deltasIdx++;
                    if (this.deltas.length == this.deltasIdx)
                        this.deltas = copyOf(this.deltas, this.deltas.length * 2);
                    this.deltas[this.deltasIdx] = ptr;
                    this.blockSize = 0;
                }

                this.blockSize++;
                this.size++;
                int intPtr = (int) (ptr - deltas[this.deltasIdx]);
                this.osIdx.writeInt(intPtr);
		    }

			this.var = var;
			this.nPending = null == var ? 0 : 1;
		}
		return this.size + this.nPending -1;
	}

    @Override
	public void build(OutStream os) {
        if (!this.buildMode) throw new IllegalStateException("Not in Build-Mode");
        if (this.nPending > 0) this.push(null); // Flush buffer and lock
        super.build(os);
	}

}
