/*
 * $Id: IntFloatHashMap.java 5468 2021-12-03 20:06:15Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static java.lang.Float.floatToIntBits;
import static java.lang.Float.intBitsToFloat;

import java.util.Arrays;
import java.util.Iterator;

import de.cm.osm2po.primitives.KeyVal.IntFloat;

/**
 * Dies ist eine Hashmap fuer primitive floats als Werte
 * und einem primitiven <code>int</code> als Schluessel.<br>
 * Die Kapazitaet ist auf 2 Gig begrenzt.<br>
 * <br><b>Wichtig:</b><br>
 * {@link #rehash()} wird nicht automatisch ausgeloest. Die Statistiken koennen
 * jedoch mittels {@link #getHashPopulation()}, {@link #getHashSize()},
 * {@link #getSize()} und {@link #getMaxCollisions()} abgefragt werden.
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class IntFloatHashMap implements Var<IntFloatHashMap>, Iterable<IntFloat> {
    public static final float NULLVALUE = Float.NaN;
    private static final int NULLINDEX = -1;
    private static final int REMOVEDINDEX = -2;

    private final static int INTMAXHALF = Integer.MAX_VALUE >>> 1;

    // Ein MapEntry benoetigt immer Key, Value und einen Zeiger
    // auf einen evtl. kollidierenden Naechsten mit gleichem Hash(Key).
    // Anstatt eine Linked-List eines solchen Objekts zu erstellen,
    // verwenden wir drei synchrone Arrays.
    private int[] keys;
    private float[] values;
    private int[] nexts;
    private int kvSize; // Arraygroesse key/val - nicht verwechseln mit Anzahl Elemente

    // Dies ist die eigentliche HashMap, also die Zeiger auf die keys und values
    private int[] hash;
    private int hashMask; // Arraygroesse des Hashes -1 (Mask).

    // Hier merken wir uns geloeschte Speicherplaetze fuer Wiederverwendung
    private int[] reuseArray;
    private int reuseArraySize = 0;

    private int size; // Anzahl der Elemente, auch nach Add/Remove

    private int collisions; // Laengste Kollisions-Liste (Elemente mit gleichem Hash)
    private int population; // Belegung des eigentlichen Hash-Arrays

    public IntFloatHashMap() {this.construct(0x400);}

    /**
     * @param capacity Initial erwartete maximale Anzahl der Elemente.
     * <br><br><b>Wichtig:</b><br>
     * Die Groesse des internen HashArrays liegt
     * zwischen 2 (bei n^2 -1) und 4 (bei n^2).
     */
    public IntFloatHashMap(int capacity) {this.construct(capacity);}

    private void construct(int capacity) {
        if (capacity < 0) capacity = 1;
        int hashSize = 1;
        for (int bit = 1; bit < 30 && hashSize < capacity; bit++) hashSize <<= 1; // Max 30 Bits
        hashSize <<= 1; // mindestens 2x so gross // Max 31 Bits

        this.hashMask = hashSize -1;
        this.hash = new int[hashSize];
        this.nexts = new int[capacity];
        this.keys = new int[capacity];
        this.values = new float[capacity];
        this.reuseArray = new int[capacity / 5 + 1];
        this.clear();
    }

    /** Leert die Map */
    public void clear() {
        this.collisions = 0;
        this.population = 0;
        this.size = 0;
        this.kvSize = 0;
        this.reuseArraySize = 0;
        Arrays.fill(this.hash, NULLINDEX);
    }

    /** @return Anzahl der Elemente in der Map. */
    public int getSize() {
        return this.size;
    }

    /** @return Groesse des internen HashArrays. */
    public int getHashSize() {
        return this.hashMask +1;
    }

    /**
     * @return Groesse der laengsten Kollisionskette fuer gleiche Hashwerte
     * nach {@link #put(int, float)} oder {@link #rehash()}.
     * Aus Performanz-Gruenden jedoch nicht nach {@link #remove(int)}.
     */
    public int getMaxCollisions() {
        return this.collisions;
    }

    /** @return Anzahl der verwendeten Slots im internen HashArray. */
    public int getHashPopulation() {
        return this.population;
    }

    private int addNewKeyValue(int prevIndex, int key, float value) {
        int index; // An dieser Stelle werden Key/Value gespeichert
        if (this.reuseArraySize > 0) {
            // Speicherloecher wiederverwenden.
            index = this.reuseArray[--this.reuseArraySize];
        } else { // Sonst hinten anfuegen.
            int capacity = this.values.length;
            if (this.kvSize == capacity) {
                // ArrayOverflow absichern
                if (capacity < 0x1000000 /*16MB*/) capacity *= 2;
                else if (capacity < INTMAXHALF) capacity *= 1.5;
                else capacity += ((Integer.MAX_VALUE - capacity) / 3); // grow third of rest till 0

                this.nexts = Arrays.copyOf(this.nexts, capacity);
                this.keys = Arrays.copyOf(this.keys, capacity);
                this.values = Arrays.copyOf(this.values, capacity);
            }
            index = this.kvSize++;
        }
        this.size++;
        this.keys[index] = key;
        this.values[index] = value;
        this.nexts[index] = NULLINDEX;

        if (prevIndex != NULLINDEX) {
            this.nexts[prevIndex] = index;
        }
        return index;
    }

    /**
     * Fuegt ein Schluessel-Wert-Paar in die Map ein.
     * @param key int Schluessel
     * @param value float Wert
     */
    public void put(int key, float value) {
        int hashIndex = key & this.hashMask;
        int kvIndex = this.hash[hashIndex];
        if (NULLINDEX == kvIndex) {
            kvIndex = this.addNewKeyValue(NULLINDEX, key, value);
            this.hash[hashIndex] = kvIndex;
            this.population++;
        } else {
            int listCount = 0;
            int lastKvIndex = NULLINDEX;
            while (kvIndex != NULLINDEX) {
                listCount++;
                int foundKey = this.keys[kvIndex];
                if (foundKey == key) {
                    this.values[kvIndex] = value;
                    return;
                }
                lastKvIndex = kvIndex;
                kvIndex = this.nexts[kvIndex];
            }
            this.addNewKeyValue(lastKvIndex, key, value);
            if (listCount > this.collisions) this.collisions = listCount;
        }
    }

    /**
     * Entfernt ein Element aus der Map.
     * @param key int Schluessel
     * @return float entfernter Wert.
     */
    public float remove(int key) {
        // Kollisions-Statistiken werden hier nicht aktualisiert.
        // Sonst muessten wir immer die komplette Kollisions-Liste durchlaufen.
        int hashIndex = key & this.hashMask;
        if (hashIndex != NULLINDEX) {
            int prevKvIndex = NULLINDEX;
            int thisKvIndex = this.hash[hashIndex];
            while (thisKvIndex != NULLINDEX) {
                int nextKvIndex = this.nexts[thisKvIndex];
                int foundKey = this.keys[thisKvIndex];
                if (foundKey == key) {
                    if (prevKvIndex == NULLINDEX) {
                        this.hash[hashIndex] = nextKvIndex;
                        if (nextKvIndex == NULLINDEX) this.population--;
                    } else {
                        this.nexts[prevKvIndex] = nextKvIndex;
                    }

                    float removed = this.values[thisKvIndex];

                    this.values[thisKvIndex] = NULLVALUE;
                    this.keys[thisKvIndex] = 0;
                    this.nexts[thisKvIndex] = REMOVEDINDEX;
                    this.size--;

                    if (this.reuseArraySize == this.reuseArray.length) {
                        // ArrayOverflow absichern
                        this.reuseArray = Arrays.copyOf(
                                this.reuseArray, this.reuseArray.length * 2);
                    }
                    this.reuseArray[this.reuseArraySize++] = thisKvIndex;
                    return removed;
                }
                prevKvIndex = thisKvIndex;
                thisKvIndex = this.nexts[thisKvIndex];
            }
        }
        return NULLVALUE;
    }

    /**
     * Prueft, ob ein Schluessel existiert.
     * @param key int Schluessel.
     * @return Existiert in Map ja/nein.
     */
    public boolean contains(int key) {
        int hashIndex = key & this.hashMask;
        int kvIndex = this.hash[hashIndex];

        while (kvIndex != NULLINDEX) {
            int foundKey = this.keys[kvIndex];
            if (foundKey == key) {
                return true;
            }
            kvIndex = this.nexts[kvIndex];
        }
        return false;
    }

    /**
     * Liefert einen Wert aus der Map.
     * @param key int Schluessel
     * @return float value oder {@link #NULLVALUE}
     */
    public float get(int key) {
        float foundValue = NULLVALUE;
        int hashIndex = key & this.hashMask;
        int kvIndex = this.hash[hashIndex];

        while (kvIndex != NULLINDEX) {
            int foundKey = this.keys[kvIndex];
            if (foundKey == key) {
                foundValue = this.values[kvIndex];
                break;
            }
            kvIndex = this.nexts[kvIndex];
        }

        return foundValue;
    }

    public final int[] getKeys() {
        int[] validKeys = new int[0x100];
        int nKeys = 0;
        for (int idx = 0; idx < this.kvSize; idx++) {
            if (REMOVEDINDEX == nexts[idx]) continue; // invalid key after remove
            if (nKeys == validKeys.length) { // ArrayOverflow
                validKeys = Arrays.copyOf(validKeys, nKeys * 2);
            }
            validKeys[nKeys++] = this.keys[idx];
        }
        return Arrays.copyOf(validKeys, nKeys);
    }

    /**
     * Vergroessert das interne HashArray um Faktor 2, reorganisiert
     * die Schluessel-Wert-Paare und steigert so die Performanz.
     */
    public void rehash() {
        final int oldHashSize = this.getHashSize();
        final int newHashSize = oldHashSize << 1;
        if (newHashSize < 0) return; // 31 Bit overflow check

        this.hash = Arrays.copyOf(this.hash, newHashSize);
        Arrays.fill(this.hash, oldHashSize, this.hash.length, NULLINDEX);

        this.hashMask = this.hash.length -1;
        this.collisions = 0;

        for (int oldHashIndex = 0; oldHashIndex < oldHashSize; oldHashIndex++) {
            int newHashIndex = oldHashIndex | oldHashSize;

            int nOld = 0, nNew = 0;

            int thisKvIndex = this.hash[oldHashIndex];
            int prevKvIndex = NULLINDEX;
            int prevKvIndexNew = NULLINDEX;
            while (thisKvIndex != NULLINDEX) {
                int nextKvIndex = this.nexts[thisKvIndex];
                int foundKey = this.keys[thisKvIndex];
                if ((foundKey & oldHashSize) == 0) {
                    nOld++;
                    prevKvIndex = thisKvIndex;

                } else {
                    nNew++;

                    // Detach at old position
                    if (prevKvIndex == NULLINDEX) {
                        this.hash[oldHashIndex] = nextKvIndex;
                    } else {
                        this.nexts[prevKvIndex] = nextKvIndex;
                    }

                    // Move to new position
                    this.nexts[thisKvIndex] = NULLINDEX;
                    if (prevKvIndexNew == NULLINDEX) {
                        this.hash[newHashIndex] = thisKvIndex;
                    } else {
                        this.nexts[prevKvIndexNew] = thisKvIndex;
                    }

                    prevKvIndexNew = thisKvIndex;
                }

                thisKvIndex = nextKvIndex;
            }

            int collis = (nOld > nNew ? nOld : nNew) -1;
            if (collis > this.collisions) this.collisions = collis;
            if (nOld != 0 && nNew != 0) this.population++;
        }
    }

    @Override
    public String toString() {
        int n = 0; String s = null;
        for (int i = 0; i < this.kvSize; i++) {
            if (REMOVEDINDEX == this.nexts[i]) continue;
            if (null == s) s = ""; else s += "\n";
            if (++n > 10) {s += "more..."; break;}
            s += "[" + this.keys[i] + "]:" + this.values[i];
        }
        return s;
    }

    public boolean isRehashNeeded() {
        double dist = (double) this.size / this.hash.length;
        return dist > 1.0;
    }

    @Override
    public Iterator<IntFloat> iterator() {
        return new Iterator<IntFloat>() {
            private int idx = 0;
            @Override public boolean hasNext() {
                for (;;) {
                    if (idx >= kvSize) return false;
                    if (REMOVEDINDEX == nexts[idx]) idx++;
                    else return true;
                }
            }
            @Override public IntFloat next() {
                int key = keys[this.idx];
                float val = values[this.idx++];
                return new IntFloat(key, val);
            }
            @Override public void remove() {
                throw new UnsupportedOperationException("remove() not supported");
            }
        };
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.size);
        int sizeCheck = 0;
        for (int i = 0; i < this.kvSize; i++) {
            if (REMOVEDINDEX == this.nexts[i]) continue;
            sizeCheck++;
            outStream.writeInt(this.keys[i]);
            outStream.writeInt(floatToIntBits(this.values[i]));
        }
        if (sizeCheck != this.size) {
            throw new RuntimeException("Sth. went terribly wrong");
        }
    }

    @Override
    public IntFloatHashMap readFromStream(InStream inStream) {
        int n = inStream.readInt();
        this.construct(n);
        for (int i = 0; i < n; i++) {
            int key = inStream.readInt();
            float val = intBitsToFloat(inStream.readInt());
            this.put(key, val);
        }
        return this;
    }


}
