/*
 * $Id: IntLongBag.java 4470 2017-07-14 19:34:41Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Ein Bag auf Grundlage von primitiven Datentypen fuer
 * Key und Value (int / long).
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class IntLongBag implements Var<IntLongBag> {

    private LongArray values = new LongArray();
    private IntArray keys = new IntArray(values);

    public final static long NULLVALUE = Long.MIN_VALUE;

    private long size;

    public long getSize() {
        return this.size;
    }

    public void clear() {
        this.keys.clear();
        this.values.clear();
        this.size = 0;
    }

    public IntArray getKeys() {
        return this.keys;
    }

    public LongArray getValues() {
        return this.values;
    }

    /**
     * Fuegt ein neues Schluessel-Wert-Paar hinzu.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig. Es koennen also mehrere Werte den gleichen
     * Schluessel haben.
     * @param key Schluessel
     * @param value Wert
     */
    public void add(int key, long value) {
        keys.set(this.size, key);
        values.set(this.size, value);
        this.size++;
    }

    /**
     * Findet einen Wert zu einem Schluessel.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig. Es wird also nur der zuerst gefundene geliefert.
     * @param key Schluessel
     * @return Wert oder NULLVALUE
     */
    public long get(int key) {
        long idx = this.keys.searchAny(key);
        if (IntArray.NOTFOUND == idx) return NULLVALUE;
        return this.values.get(idx);
    }

    /**
     * Findet alle Werte zu einem Schluessel.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig, daher koennen mehrere Werte folgen.
     * @param key Schluessel
     * @return long[] Werte oder null, wenn keine.
     */
    public long[] getAll(int key) {
        long idxs[] = this.keys.searchAll(key);
        if (null == idxs) return null;
        long vals[] = new long[idxs.length];
        for (int i = 0; i < idxs.length; i++) {
            vals[i] = this.values.get(idxs[i]);
        }
        return vals;
    }

    /**
     * Liefert einen Wert aus dem Cache,
     * allerdings nicht ueber den Schluessel, sondern
     * mittels Positionsangabe.
     * @param idx Index Position 0-wertig.
     * @return Wert oder NULLVALUE
     */
    public long getAt(long idx) {
        return this.values.get(idx);
    }

    public void sort() {
        this.keys.sort();
    }

    @Override
    public void writeToStream(OutStream outStream) {
        if (this.size != this.keys.getSize())
            throw new RuntimeException("Size of Bag and underlying data differ");
        outStream.writeLong(this.size);
        outStream.writeByte(this.keys.isSorted() ? (byte)1 : (byte)0);
        for (long i = 0; i < this.size; i++) {
            outStream.writeInt(this.keys.get(i));
            outStream.writeLong(this.values.get(i));
        }
    }

    @Override
    public IntLongBag readFromStream(InStream inStream) {
        this.clear();
        long n = inStream.readLong();
        boolean isSorted = inStream.readByte() != 0;
        for (long i = 0; i < n; i++) {
            int key = inStream.readInt();
            long val = inStream.readLong();
            this.add(key, val);
        }
        if (isSorted) this.sort();
        return this;
    }

}
