/*
 * $Id: IntSet.java 5468 2021-12-03 20:06:15Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.util.Arrays;

/**
 * Dies ist ein Set von primitiven ints.<br>
 * Die Kapazitaet ist auf 2 Gig begrenzt.<br>
 * <br><b>Wichtig:</b><br>
 * Ein ReHash wird derzeit nicht unterstuetzt!
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class IntSet implements Var<IntSet> {
    public static final int NULLVALUE = Integer.MIN_VALUE;
    private static final int NULLINDEX = -1;
    private static final int REMOVEDINDEX = -2;

    private final static int INTMAXHALF = Integer.MAX_VALUE >>> 1;

    // Ein Entry benoetigt immer den Wert und einen Zeiger
    // auf einen evtl. kollidierenden Naechsten mit gleichem Hash(Wert).
    // Anstatt eine Linked-List eines solchen Objekts zu erstellen,
    // verwenden wir drei synchrone Arrays.
    private int[] values;
    private int[] nexts;
    private int vSize; // Arraygroesse der Werte - nicht verwechseln mit Anzahl Elemente

    // Dies ist die eigentliche HashMap, also die Zeiger auf die values
    private int[] hash;
    private int hashMask; // Arraygroesse des Hashes -1 (Mask).

    // Hier merken wir uns geloeschte Speicherplaetze fuer Wiederverwendung
    private int[] reuseArray;
    private int reuseArraySize = 0;

    private int size; // Anzahl der Elemente, auch nach Add/Remove

    public IntSet() {this.construct(0x400);}

    /**
     * @param capacity Initial erwartete maximale Anzahl der Elemente.
     * <br><br><b>Wichtig:</b><br>
     * Die Groesse des internen HashArrays liegt
     * zwischen 2 (bei n^2 -1) und 4 (bei n^2).
     */
    public IntSet(int capacity) {this.construct(capacity);}

    private void construct(int capacity) {
        if (capacity < 0) capacity = 1;
        int hashSize = 1;
        for (int bit = 1; bit < 30 && hashSize < capacity; bit++) hashSize <<= 1; // Max 30 Bits
        hashSize <<= 1; // mindestens 2x so gross // Max 31 Bits

        this.hashMask = hashSize -1;
        this.hash = new int[hashSize];
        this.nexts = new int[capacity];
        this.values = new int[capacity];
        this.reuseArray = new int[capacity / 5 + 1];
        this.clear();
    }

    /** Leert den Set */
    public void clear() {
        this.size = 0;
        this.vSize = 0;
        this.reuseArraySize = 0;
        Arrays.fill(this.hash, NULLINDEX);
    }

    /** @return Anzahl der Elemente im Set. */
    public int getSize() {
        return this.size;
    }

    /** @return Groesse des internen HashArrays. */
    public int getHashSize() {
        return this.hashMask +1;
    }

    private int addNewValue(int prevIndex, int value) {
        int index; // An dieser Stelle werden Key/Value gespeichert
        if (this.reuseArraySize > 0) {
            // Speicherloecher wiederverwenden.
            index = this.reuseArray[--this.reuseArraySize];
        } else { // Sonst hinten anfuegen.
            int capacity = this.values.length;
            if (this.vSize == capacity) {
                // ArrayOverflow absichern
                if (capacity < 0x1000000 /*16MB*/) capacity *= 2;
                else if (capacity < INTMAXHALF) capacity *= 1.5;
                else capacity += ((Integer.MAX_VALUE - capacity) / 3); // grow third of rest till 0

                this.nexts = Arrays.copyOf(this.nexts, capacity);
                this.values = Arrays.copyOf(this.values, capacity);
            }
            index = this.vSize++;
        }
        this.size++;
        this.values[index] = value;
        this.nexts[index] = NULLINDEX;

        if (prevIndex != NULLINDEX) {
            this.nexts[prevIndex] = index;
        }
        return index;
    }

    /**
     * Fuegt einen Wert hinzu.
     * @param value int Wert
     */
    public void add(int value) {
        int hashIndex = value & this.hashMask;
        int vIndex = this.hash[hashIndex];
        if (NULLINDEX == vIndex) {
            vIndex = this.addNewValue(NULLINDEX, value);
            this.hash[hashIndex] = vIndex;
        } else {
            int lastIndex = NULLINDEX;
            while (vIndex != NULLINDEX) {
                int foundValue = this.values[vIndex];
                if (foundValue == value) {
                    return;
                }
                lastIndex = vIndex;
                vIndex = this.nexts[vIndex];
            }
            this.addNewValue(lastIndex, value);
        }
    }

    /**
     * Prueft, ob ein Wert existiert,
     * wobei der Wert auch <code>null</code> sein kann.
     * @param value int Schluessel.
     * @return Existiert in Map ja/nein.
     */
    public boolean contains(int value) {
        int hashIndex = value & this.hashMask;
        int vIndex = this.hash[hashIndex];

        while (vIndex != NULLINDEX) {
            int foundValue = this.values[vIndex];
            if (foundValue == value) {
                return true;
            }
            vIndex = this.nexts[vIndex];
        }
        return false;
    }

    /**
     * Entfernt ein Element aus dem Set.
     * @param value int
     * @return entfernte ja/nein
     */
    public boolean remove(int value) {
        int hashIndex = value & this.hashMask;
        if (hashIndex != NULLINDEX) {
            int prevIndex = NULLINDEX;
            int thisIndex = this.hash[hashIndex];
            while (thisIndex != NULLINDEX) {
                int nextIndex = this.nexts[thisIndex];
                int foundValue = this.values[thisIndex];
                if (foundValue == value) {
                    if (prevIndex == NULLINDEX) {
                        this.hash[hashIndex] = nextIndex;
                    } else {
                        this.nexts[prevIndex] = nextIndex;
                    }

                    this.values[thisIndex] = NULLVALUE;
                    this.nexts[thisIndex] = REMOVEDINDEX;
                    this.size--;

                    if (this.reuseArraySize == this.reuseArray.length) {
                        // ArrayOverflow absichern
                        this.reuseArray = Arrays.copyOf(
                                this.reuseArray, this.reuseArray.length * 2);
                    }
                    this.reuseArray[this.reuseArraySize++] = thisIndex;
                    return true;
                }
                prevIndex = thisIndex;
                thisIndex = this.nexts[thisIndex];
            }
        }
        return false;
    }


    @Override
    public String toString() {
        int n = 0; String s = null;
        for (int i = 0; i < this.vSize; i++) {
            if (REMOVEDINDEX == this.nexts[i]) continue;
            if (null == s) s = ""; else s += ", ";
            if (++n > 100) {s += "more..."; break;}
            s += this.values[i];
        }
        return s;
    }

    public int[] toArray() {
        int[] arr = new int[this.size];
        int n = 0;
        for (int i = 0; i < this.vSize; i++) {
            if (REMOVEDINDEX == this.nexts[i]) continue;
            arr[n++] = this.values[i];
        }
        return arr;
    }

    public void add(IntSet otherSet) {
        if (null == otherSet) return;
        for (int i = 0; i < otherSet.vSize; i++) {
            if (REMOVEDINDEX == otherSet.nexts[i]) continue;
            this.add(otherSet.values[i]);
        }
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.size);
        int sizeCheck = 0;
        for (int i = 0; i < this.vSize; i++) {
            if (REMOVEDINDEX == this.nexts[i]) continue;
            sizeCheck++;
            outStream.writeInt(this.values[i]);
        }
        if (sizeCheck != this.size) {
            throw new RuntimeException("Sth. went terribly wrong");
        }
    }

    @Override
    public IntSet readFromStream(InStream inStream) {
        int n = inStream.readInt();
        this.construct(n);
        for (int i = 0; i < n; i++) {
            int val = inStream.readInt();
            this.add(val);
        }
        return this;
    }


}
