/*
 * $Id: LongArray.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static java.lang.System.arraycopy;
import static java.util.Arrays.fill;

import java.util.Arrays;

/**
 * Dieses Array waechst nach Bedarf und arbeitet auf Blockbasis.
 * Das heisst, es werden nur Speicherbereiche verbraucht, die
 * tatsaechlich indiziert werden.
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public final class LongArray implements Swappable {
    private static final long MAXMEM = 32 * 0x40000000L; // Maximaler Speicher in GigaBytes.
    private static final int TYPELEN = 8; // Byte hat 1 Byte - Long hat 8 Bytes
    private static final int BLOCKBITS = 17;
    private static final int BLOCKSIZE = 1 << BLOCKBITS; // BlockSize = 128k
    private static final int BLOCKMASK = BLOCKSIZE - 1; // 17 Bits Mask
    private static final int SLOTSIZE = (int) (MAXMEM / (TYPELEN * BLOCKSIZE));

    public static final long NOTFOUND = Long.MIN_VALUE;
    public static final long NULLVALUE = Long.MIN_VALUE;

    private long maxIdx = -1L;
    private long[][] slots = new long[SLOTSIZE][];
    private boolean isSorted = false;

    private Swappable swappable;

    public LongArray(Swappable swappable) {
        this.swappable = swappable;
    }

    public LongArray() {
    }

    public void clear() {
        for (int i = 0; i < SLOTSIZE; i++) {
            this.slots[i] = null;
        }
        this.maxIdx = -1;
        this.isSorted = false;
    }

    public void push(long value) {
    	this.set(this.maxIdx +1, value);
    }

    public void push(LongArray longArray) {
        long n = longArray.getSize();
        for (long i = 0; i < n; i++) {
            this.push(longArray.get(i));
        }
    }

    public boolean contains(long value) {
    	return this.searchAny(value) != NOTFOUND;
    }

    public boolean isSorted() {
        return this.isSorted;
    }

    public void set(long idx, long value) {
        if (idx < 0) throw new ArrayIndexOutOfBoundsException("" + idx);
        int slotIdx = (int) (idx >>> BLOCKBITS);
        long[] block = this.slots[slotIdx];
        if (null == block) {
            if (NULLVALUE == value) return;
            block = new long[BLOCKSIZE];
            Arrays.fill(block, NULLVALUE);
            this.slots[slotIdx] = block;
        }
        // Wichtig! maxIdx erst hier setzen, falls oben OutOfMemory.
        if (idx > this.maxIdx) this.maxIdx = idx;
        int blockIdx = (int) (idx & BLOCKMASK);
        block[blockIdx] = value;
        this.isSorted = false;
    }

    public long get(long idx) {
        if (idx < 0) throw new ArrayIndexOutOfBoundsException("" + idx);
        int slotIdx = (int) (idx >>> BLOCKBITS);
        long[] block = this.slots[slotIdx];
        if (null == block) return NULLVALUE;
        int blockIdx = (int) (idx & BLOCKMASK);
        return block[blockIdx];
    }

    public long getSize() {
        return this.maxIdx +1;
    }

    public synchronized void sort() {
        quicksort(0, this.maxIdx);
        this.isSorted = true;
    }

    private void quicksort(long low, long high) {
        long i = low, j = high;
        long pivot = this.get(low + (high-low)/2);

        while (i <= j) {
            while (this.get(i) < pivot) i++;
            while (this.get(j) > pivot) j--;

            if (i <= j) {
                this.swap(i, j);
                i++;
                j--;
            }
        }
        if (low < j) quicksort(low, j);
        if (i < high) quicksort(i, high);
    }

    /**
     * Sortiert und eliminiert mehrfach vorkommende Werte im Array.
     * Freigewordene Slots werden danach wieder auf null gesetzt.
     */
    public void shrinkToSet() {
        if (this.maxIdx <= 0) return;
        if (!this.isSorted) this.sort();
        long lastWriteIdx = -1;
        long lastWrittenVal = NULLVALUE; // MaxNegative
        for (long idx = 0; idx <= this.maxIdx; idx++) {
            long val = this.get(idx);
            if (val > lastWrittenVal) {
                lastWriteIdx++;
                lastWrittenVal = val;
                this.set(lastWriteIdx, lastWrittenVal); // setzt isSorted auf false!
            }
        }
        this.maxIdx = lastWriteIdx;
        int slotIdx = (int) (lastWriteIdx >>> BLOCKBITS);
        for (int i = slotIdx + 1; i < slots.length && slots[i] != null; i++) {
            slots[i] = null;
        }
        this.isSorted = true; // Wichtig!!!
    }

    @Override
    public void swap(long i, long j) {
        long tmp = this.get(i);
        this.set(i, this.get(j));
        this.set(j, tmp);
        if (this.swappable != null) this.swappable.swap(i, j);
    }


    /**
     * Findet den Wert im Array und liefert den ersten
     * beliebigen Treffer-Index im Array.
     * @param searchValue zu suchender Wert
     * @return Index oder {@link #NOTFOUND}
     */
    public final long searchAny(long searchValue) {
        if (this.isSorted) {
            long left = 0;
            long right = this.maxIdx;
            return binarySearch(searchValue, left, right);
        } else {
            for (long i = 0; i <= this.maxIdx; i++) {
                if (this.get(i) == searchValue) return i;
            }
            return NOTFOUND;
        }
    }

    public long[] toPrimitive() {
    	if (this.maxIdx >= Integer.MAX_VALUE)
    		throw new IllegalStateException("Index exceeds int range " + this.maxIdx);
        long[] arr = new long[(int) (this.maxIdx +1)];
        int maxSlotIdx = (int) (this.maxIdx >>> BLOCKBITS);
        int writeIdx = 0;
        for (int i = 0; i < maxSlotIdx; i++) {
            if (null == this.slots[i])
                fill(arr, writeIdx, writeIdx + BLOCKSIZE, NULLVALUE);
            else
                arraycopy(this.slots[i], 0, arr, writeIdx, BLOCKSIZE);
            writeIdx += BLOCKSIZE;
        }
        if (writeIdx < this.maxIdx) {
            System.arraycopy(this.slots[maxSlotIdx],
                    0, arr, writeIdx, (int) (this.maxIdx - writeIdx +1));
        }
        return arr;
    }

    /**
     * Sucht den Wert im Array und liefert alle
     * gefundenen Indizes als long[]-Array.
     * Die Ergebnis-Reihenfolge entspricht dem Original.
     * @param searchValue zu suchender Wert.
     * @return <code>long[&gt;0]</code> oder <code>null</code>
     */
    public final long[] searchAll(long searchValue) {

        if (this.isSorted) {
            long anyIndex = this.searchAny(searchValue);
            if (NOTFOUND == anyIndex) return null;
            long left = anyIndex;
            long right = anyIndex;
            while (left > 0 && this.get(left -1) == searchValue) left--;
            while (right < this.maxIdx && this.get(right +1) == searchValue) right++;
            long[] retArr = new long[(int) (right - left +1)];
            for (long i = left; i <= right; i++) retArr[(int) (i - left)] = i;
            return retArr;

        } else { // Nicht sortiert, dann sequentiell suchen = Langsam.
            long retArr[] = new long[10];
            int retArrSize = 0;
            for (long i = 0; i <= this.maxIdx; i++) {
                if (searchValue == this.get(i)) {
                    if (retArrSize == retArr.length) { // Ueberlaufschutz.
                        retArr = Arrays.copyOf(retArr, retArr.length * 2);
                    }
                    retArr[retArrSize++] = i;
                }
            }
            if (retArrSize == 0) return null;
            return Arrays.copyOf(retArr, retArrSize);
        }
    }

    private final long binarySearch(long searchValue, long left, long right) {
        if (right < left) return NOTFOUND;
        long mid = (left + right) >>> 1;
        long midVal = this.get(mid);

        if (searchValue > midVal) {
            return binarySearch(searchValue, mid + 1, right);
        } else if (searchValue < midVal) {
            return binarySearch(searchValue, left, mid - 1);
        } else {
            return mid;
        }
    }

}
