/*
 * $Id: LongArrayAsc.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Sehr spezielle Array-Klasse, deren Werte lediglich mittels
 * {@link #push(long)} hinzugefuegt werden koennen.
 * Neue Werte duerfen dabei nicht kleiner als die Vorgaenger sein.
 * Dies ermoeglicht den Einsatz einer internen Offset-Verwaltung,
 * die es ermoeglicht, intern Werte als ints (4-Bytes) zu speichern. 
 * <br><br><b>Wichtig:</b><br>
 * Die interne Offset-Verwaltung reinitialisiert sich nach
 * einer Million Werte. Das heisst, obwohl Long, darf eine Spanne
 * den Int-Bereich (2 Gig) nicht verlassen. Nach einer Reinitialisierung
 * verschiebt sich die Spanne dann in den naechst hoeheren Wertebereich.
 * 
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class LongArrayAsc {
	
	private static final int BLOCK_SIZE = 0x100000; // 1M;
	
	private IntArray ptrs;
	private long[] deltas;
	private int deltasIdx;
	private int blockSize;
	private long valGuard;
	
	public LongArrayAsc() {
	    this.clear();
	}
	
	/** @return Anzahl der Elemente. */
	public long getSize() {return this.ptrs.getSize();}
	
	public void clear() {
        this.ptrs = new IntArray();
        this.valGuard = -1;
        this.blockSize = 0;
        this.deltas = new long[1024];
        this.deltasIdx = 0;
	}
	
	/**
	 * Liefert den Pointer fuer einen Index.
	 * <br><b>Wichtig:</b><br>
	 * Dieses Methode ist nicht ThreadSafe!
	 * @param index int Index-Position (0-wertig).
	 * @return long Wert.
	 */
	public long get(int index) {
		int deltaIdx = index / BLOCK_SIZE;
		return this.ptrs.get(index) + this.deltas[deltaIdx];
	}

	/**
	 * Fuegt einen weiteren Index (normalerweise FilePointer)
	 * sequentiell hinzu.
	 * @param index long-Wert
	 *     PRE: nur aufsteigende oder gleiche Werte erlaubt.
	 * @return long Index-Position (n-tes Element - 0-Wertig)
	 */
	public long push(long index) {
		if (index < this.valGuard) throw new ArrayIndexOutOfBoundsException(
				"Added value not in ascending sequence");
		this.valGuard = index;

		if (this.blockSize >= BLOCK_SIZE) {
			this.deltas[++this.deltasIdx] = index;
			this.blockSize = 0;
		}
		this.blockSize++;

		long idx = index - deltas[this.deltasIdx];
		if (idx > Integer.MAX_VALUE)
		    throw new IllegalArgumentException("Index not in delta range");
		
		int intIdx = (int) idx;
		this.ptrs.push(intIdx);
		return this.ptrs.getSize() -1;
	}
	
}
