/*
 * $Id: LongIntBag.java 4470 2017-07-14 19:34:41Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Ein Bag auf Grundlage von primitiven Datentypen fuer
 * Key und Value (long / int).
 *
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class LongIntBag implements Var<LongIntBag> {

    private IntArray values = new IntArray();
    private LongArray keys = new LongArray(values);

    public final static int NULLVALUE = IntArray.NULLVALUE;

    private long size;

    public long getSize() {
        return this.size;
    }

    public void clear() {
        this.keys.clear();
        this.values.clear();
        this.size = 0;
    }

    public LongArray getKeys() {
        return this.keys;
    }

    public IntArray getValues() {
        return this.values;
    }

    /**
     * Fuegt ein neues Schluessel-Wert-Paar hinzu.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig. Es koennen also mehrere Werte den gleichen
     * Schluessel haben.
     * @param key Schluessel
     * @param value Wert
     */
    public void add(long key, int value) {
        keys.set(this.size, key);
        values.set(this.size, value);
        this.size++;
    }

    /**
     * Findet einen Wert zu einem Schluessel.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig. Es wird also nur der zuerst gefundene geliefert.
     * @param key Schluessel
     * @return Wert oder NULLVALUE
     */
    public int get(long key) {
        long idx = this.keys.searchAny(key);
        if (LongArray.NOTFOUND == idx) return NULLVALUE;
        return this.values.get(idx);
    }

    /**
     * Findet alle Werte zu einem Schluessel.<br>
     * <b>Wichtig</b> Da dies ein Bag ist, ist der Schluessel
     * nicht eindeutig, daher koennen mehrere Werte folgen.
     * @param key Schluessel
     * @return int[] Werte oder null, wenn keine.
     */
    public int[] getAll(long key) {
        long idxs[] = this.keys.searchAll(key);
        if (null == idxs) return null;
        int vals[] = new int[idxs.length];
        for (int i = 0; i < idxs.length; i++) {
            vals[i] = this.values.get(idxs[i]);
        }
        return vals;
    }

    public void sort() {
        this.keys.sort();
    }

    @Override
    public void writeToStream(OutStream outStream) {
        if (this.size != this.keys.getSize())
            throw new RuntimeException("Size of Bag and underlying data differ");
        outStream.writeLong(this.size);
        outStream.writeByte(this.keys.isSorted() ? (byte)1 : (byte)0);
        for (long i = 0; i < this.size; i++) {
            outStream.writeLong(this.keys.get(i));
            outStream.writeInt(this.values.get(i));
        }
    }

    @Override
    public LongIntBag readFromStream(InStream inStream) {
        this.clear();
        long n = inStream.readLong();
        boolean isSorted = inStream.readByte() != 0;
        for (long i = 0; i < n; i++) {
            long key = inStream.readLong();
            int val = inStream.readInt();
            this.add(key, val);
        }
        if (isSorted) this.sort();
        return this;
    }

}
