/*
 * $Id: OutStreamAdapterBuffered.java 4347 2017-02-02 19:50:36Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * @author (c) 2014 - Carsten Moeller, info@osm2po.de
 */
public class OutStreamAdapterBuffered extends Thread implements OutStream, Concatenable {

    private final static int BUF_SIZE_DEFAULT = 0x100000; // 1M

    // Aktueller Schreib- und Hintergrund-Flushpuffer.
    private int bufferSize;
    private byte[] fgBuf, bgBuf;  // Foreground-, Background-Buffer
    private int fgFillSize, bgFillSize;  // Fillsizes of Buffers
    private long fgHistorySize; // Number of fully written buffers

    private OutputStream os;

    private volatile boolean flushSignal; // background is flushing data
    private volatile boolean closeSignal;

    private void construct(OutputStream os, int bufferSize) {
        try {
            this.os = os;
            this.bufferSize = bufferSize;
            fgBuf = new byte[bufferSize];
            bgBuf = new byte[bufferSize];
            
            this.setName(this.getClass().getSimpleName() + "Thread");
            start();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    public OutStreamAdapterBuffered(OutputStream os) {
        this.construct(os, BUF_SIZE_DEFAULT);
    }
    
    public OutStreamAdapterBuffered(OutputStream os, int bufferSize) {
        this.construct(os, bufferSize);
    }

    public OutStreamAdapterBuffered(File file) {
        try {
            FileOutputStream fos = new FileOutputStream(file);
            this.construct(fos, BUF_SIZE_DEFAULT);
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final synchronized void close() {
        try {
            while (flushSignal) wait();
            closeSignal = true;
            swapBuffersAndFlush();
            join();
            os.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final void run() {
        try {
            synchronized (this) {
                do {
                    while (!flushSignal) wait();
                    os.write(bgBuf, 0, bgFillSize);
                    flushSignal = false;
                    notify();
                } while (!closeSignal);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private final synchronized void swapBuffersAndFlush() {
        try {
            while (flushSignal) wait();
            flushSignal = true;
            fgHistorySize += fgFillSize;
            
            // Puffer tauschen
            byte[] buf = fgBuf;
            fgBuf = bgBuf;
            bgBuf = buf;
            bgFillSize = fgFillSize;
            fgFillSize = 0;

            notify();

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public final void writeByte(byte b) {
        fgBuf[this.fgFillSize++] = b;
        if (bufferSize == fgFillSize) swapBuffersAndFlush();
    }

    @Override
    public final void writeShort(short s) {
        if (bufferSize - fgFillSize > 2) {
            Bits.putShort(fgBuf, fgFillSize, s);
            fgFillSize += 2;
        } else {
            writeByte((byte) (s >>> 8));
            writeByte((byte) (s >>> 0));
        }
    }

    @Override
    public final void writeInt(int i) {
        if (bufferSize - fgFillSize > 4) {
            Bits.putInt(fgBuf, fgFillSize, i);
            fgFillSize += 4;
        } else {
            writeByte((byte) (i >>> 24));
            writeByte((byte) (i >>> 16));
            writeByte((byte) (i >>> 8));
            writeByte((byte) (i >>> 0));
        }
    }
    
    @Override
    public final void writeLong(long l) {
        if (bufferSize - fgFillSize > 8) {
            Bits.putLong(fgBuf, fgFillSize, l);
            fgFillSize += 8;
        } else {
            writeByte((byte) (l >>> 56));
            writeByte((byte) (l >>> 48));
            writeByte((byte) (l >>> 40));
            writeByte((byte) (l >>> 32));
            writeByte((byte) (l >>> 24));
            writeByte((byte) (l >>> 16));
            writeByte((byte) (l >>> 8));
            writeByte((byte) (l >>> 0));
        }
    }


    @Override
    public final long getWriteIndex() {
        return fgFillSize + fgHistorySize;
    }

    @Override
    public final void writeBytes(byte[] bytes, int len) {
        // TODO Schleife durch direktes BufWrite ersetzen
        for (int i = 0; i < len; i++) writeByte(bytes[i]);
    }
    
    @Override
    public void concatFile(File file, long skip) {
        try {
            byte[] buffer = new byte[4096];
            InputStream is = new BufferedInputStream(
                    new FileInputStream(file), 0x8000 /*32k*/);
            is.skip(skip);
            for (int nRead = is.read(buffer); nRead >= 0; nRead = is.read(buffer)) {
                this.writeBytes(buffer, nRead);
            }
            is.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

}
