/*
 * $Id: RandomAccessArray.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;
import static java.lang.System.arraycopy;
import static java.util.Arrays.copyOf;

/**
 * Diese Klasse dient als Container {@link RandomAccess} fuer beliebige
 * Objekte, die einmal gecacht und dann haeufig gelesen werden muessen.<br>
 * Die Speicherverwaltung basiert auf einem simplen ByteArray und adressiert
 * so maximal <b>2Gig</b>. Fuer groessere Anfoderungen sollte {@link RandomAccessMemory}
 * verwendet werden. Der Vorteil dieser Variante liegt in der initial minimalen Groesse.
 * <br><br><b>Wichtig:</b><br>
 * Der Gebrauch der setIndex() oder getIndex()-Methoden
 * ist nicht thread-safe! 
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public final class RandomAccessArray implements RandomAccess, Var<RandomAccessArray> {

    private int writeIndex;
    private int readIndex;
    private byte[] bytes;
    private int size;

    /** STDCTOR. */
    public RandomAccessArray() {this.clear();}
    
    public byte[] toBytes() {
        return copyOf(this.bytes, this.size);
    }
    
    @Override
    public void clear() {
        this.bytes = new byte[0x20];
        this.size = 0;
        this.writeIndex = 0;
        this.readIndex = 0;
    }
    
    @Override
    public void setReadIndex(long index) {
        this.readIndex = (int) index;
    }

    @Override
    public long getReadIndex() {
        return this.readIndex;
    }

    @Override
    public long getWriteIndex() {
        return this.writeIndex;
    }

    @Override
    public void setWriteIndex(long index) {
        this.writeIndex = (int) index;
    }

    @Override
    public void setWriteIndexEof() {
        this.writeIndex = this.size;
    }

    private void ensureCapacity(int nBytes) {
        int capacity = this.bytes.length;
        int ensure = this.writeIndex + nBytes;
        if (ensure < capacity) return;
        while (ensure >= capacity) {
            capacity <<= 1;
        }
        this.bytes = copyOf(this.bytes, capacity);
    }
    
    @Override
    public void writeByte(byte b) {
        ensureCapacity(1);
        this.bytes[this.writeIndex] = b;
        this.writeIndex += 1;
        if (this.writeIndex > this.size) this.size = this.writeIndex;
    }

    @Override
    public byte readByte() {
        byte b = this.bytes[this.readIndex];
        this.readIndex += 1;
        return b;
    }

    @Override
    public void writeLong(long l) {
        ensureCapacity(8);
        Bits.putLong(this.bytes, this.writeIndex, l);
        this.writeIndex += 8;
        if (this.writeIndex > this.size) this.size = this.writeIndex;
    }
    
    @Override
    public long readLong() {
        long l = Bits.getLong(this.bytes, this.readIndex);
        this.readIndex += 8;
        return l;
    }

    @Override
    public void writeInt(int i) {
        ensureCapacity(4);
        Bits.putInt(this.bytes, this.writeIndex, i);
        this.writeIndex += 4;
        if (this.writeIndex > this.size) this.size = this.writeIndex;
    }

    @Override
    public int readInt() {
        int i = Bits.getInt(this.bytes, this.readIndex);
        this.readIndex += 4;
        return i;
    }

    @Override
    public void writeShort(short s) {
        ensureCapacity(2);
        Bits.putShort(this.bytes, this.writeIndex, s);
        this.writeIndex += 2;
        if (this.writeIndex > this.size) this.size = this.writeIndex;
    }

    @Override
    public short readShort() {
        short s = Bits.getShort(this.bytes, this.readIndex);
        this.readIndex += 2;
        return s;
    }

    @Override
    public boolean isEof() {
        return this.readIndex >= this.size;
    }

    @Override
    public void close() {
        this.clear();
    }

    @Override
    public int readBytes(byte[] buffer) {
        int n = this.size - this.readIndex;
        if (buffer.length < n) n = buffer.length;
        arraycopy(this.bytes, this.readIndex, buffer, 0, n);
        this.readIndex += n;
        return n;
    }

    @Override
    public void writeBytes(byte[] buffer, int len) {
        ensureCapacity(len);
        arraycopy(buffer, 0, this.bytes, this.writeIndex, len);
        this.writeIndex += len;
        if (this.writeIndex > this.size) this.size = this.writeIndex;
    }

    @Override
    public void skip(long n) {
        this.readIndex += n;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.size);
        outStream.writeBytes(this.bytes, this.size);
    }

    @Override
    public RandomAccessArray readFromStream(InStream inStream) {
        this.size = inStream.readInt();
        this.bytes = new byte[this.size];
        inStream.readBytes(this.bytes);
        return this;
    }

}
