/*
 * $Id: RandomAccessDisk.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.RandomAccessFile;

/**
 * Implementiert das {@link RandomAccess} Interface als
 * {@link RandomAccessFile}.<br>
 * <br><b>Wichtig</b><br>
 * Der Gebrauch der setIndex() oder getIndex()-Methoden
 * ist nicht thread-safe!
 *  
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 * @see RandomAccessMemory
 */
public class RandomAccessDisk implements RandomAccess, Concatenable {
    
    private RandomAccessFile raf;
    private long readPointer;
    private long writePointer;
    private boolean isReadPointerDirty;
    private boolean isWritePointerDirty;

    public RandomAccessDisk(File file, boolean readonly) {
        try {
            this.raf = new RandomAccessFile(file, readonly ? "r" : "rw");
            this.readPointer = 0;
            this.writePointer = 0;
            this.isReadPointerDirty = true;
            this.isWritePointerDirty = true;
            
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    @Override
    public void concatFile(File file, long skip) {
        try {
        	byte[] buffer = new byte[4096];
            InputStream is = new BufferedInputStream(
                    new FileInputStream(file), 0x8000 /*32k*/);
            is.skip(skip);
            for (int nRead = is.read(buffer); nRead >= 0; nRead = is.read(buffer)) {
	            this.raf.write(buffer, 0, nRead);
	            this.writePointer += nRead;
            }
            is.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    @Override
    public void clear() {
        try {
            this.readPointer = 0;
            this.writePointer = 0;
            this.isReadPointerDirty = true;
            this.isWritePointerDirty = true;
            this.raf.setLength(0);
            
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        
    }

    @Override
    public void close() {
        try {
            if (this.raf != null) {
                this.raf.close();
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    @Override
    public long getReadIndex() {
        return this.readPointer;
    }

    @Override
    public long getWriteIndex() {
        return this.writePointer;
    }

    private long getFileLength() {
        try {
            return this.raf.length();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setReadIndex(long index) {
    	this.readPointer = index;
    	this.isWritePointerDirty = true;
    	this.isReadPointerDirty = true;
    }
    
    @Override
    public void setWriteIndexEof() {
        this.writePointer = this.getFileLength();
        this.isWritePointerDirty = true;
        this.isReadPointerDirty = true;
    }


    @Override
    public void setWriteIndex(long index) {
        this.writePointer = index;
        this.isWritePointerDirty = true;
        this.isReadPointerDirty = true;
    }

    @Override
    public byte readByte() {
        try {
        	if (this.isReadPointerDirty) {
        		this.isReadPointerDirty = false;
        		this.raf.seek(this.readPointer);
        	}
            byte b = this.raf.readByte();
            this.readPointer += 1;
            return b;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public short readShort() {
        try {
        	if (this.isReadPointerDirty) {
        		this.isReadPointerDirty = false;
        		this.raf.seek(this.readPointer);
        	}
            short s = this.raf.readShort();
            this.readPointer += 2;
            return s;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public int readInt() {
        try {
        	if (this.isReadPointerDirty) {
        		this.isReadPointerDirty = false;
        		this.raf.seek(this.readPointer);
        	}
            int n = this.raf.readInt();
            this.readPointer += 4;
            return n;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public long readLong() {
        try {
        	if (this.isReadPointerDirty) {
        		this.isReadPointerDirty = false;
        		this.raf.seek(this.readPointer);
        	}
            long l = this.raf.readLong();
            this.readPointer += 8;
            return l;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void writeByte(byte b) {
        try {
        	if (this.isWritePointerDirty) {
        		this.isWritePointerDirty = false;
        		this.raf.seek(this.writePointer);
        	}
            this.raf.writeByte(b);
            this.writePointer += 1;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void writeShort(short s) {
        try {
        	if (this.isWritePointerDirty) {
        		this.isWritePointerDirty = false;
        		this.raf.seek(this.writePointer);
        	}
            this.raf.writeShort(s);
            this.writePointer += 2;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void writeInt(int i) {
        try {
        	if (this.isWritePointerDirty) {
        		this.isWritePointerDirty = false;
        		this.raf.seek(this.writePointer);
        	}
            this.raf.writeInt(i);
            this.writePointer += 4;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void writeLong(long l) {
        try {
        	if (this.isWritePointerDirty) {
        		this.isWritePointerDirty = false;
        		this.raf.seek(this.writePointer);
        	}
            this.raf.writeLong(l);
            this.writePointer += 8;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public boolean isEof() {
        try {
            return this.readPointer >= this.raf.length();
            
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public int readBytes(byte[] buffer) {
        int nRead = 0;
        try {
            if (this.isReadPointerDirty) {
                this.isReadPointerDirty = false;
                this.raf.seek(this.readPointer);
            }
            nRead = this.raf.read(buffer);
            if (nRead > 0) {
                this.readPointer += nRead;
            } else {
                nRead = 0; // -1 moeglich
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        return nRead;
    }

    @Override
    public void writeBytes(byte[] bytes, int len) {
        try {
            if (this.isWritePointerDirty) {
                this.isWritePointerDirty = false;
                this.raf.seek(this.writePointer);
            }
            this.raf.write(bytes, 0, len);
            this.writePointer += len;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    @Override
    public void skip(long n) {
        throw new UnsupportedOperationException();
    }

}
