/*
 * $Id: RandomAccessMemory.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Implementiert ein InMemory-{@link RandomAccess}.<br>
 * Die Speicherverwaltung ist blockorientiert und somit
 * speicherfreundlich und kompakt.<br>
 * Um Objekte nach dem Schreiben wiederzufinden, muss vor dem Schreiben
 * {@link #getWriteIndex()} abgefragt und dieser Wert in einer
 * beliebigen zusaetzlichen Map hinterlegt werden. Der Rueckweg laeuft analog,
 * also auslesen der Leseposition aus der Map und setzen mittels
 * {@link #setReadIndex(long)} vor dem ersten Lesen.<br>
 * <br><b>Wichtig</b><br>
 * Der Gebrauch der setIndex() oder getIndex()-Methoden
 * ist nicht thread-safe! 
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public final class RandomAccessMemory implements RandomAccess {

    public static final long MAXMEM = ByteArray.MAXMEM;
    
    private ByteArray byteArray = new ByteArray();
    private long writeIndex;
    private long readIndex;
    
    private byte[] buffer = new byte[1024];

    /**
     * STDCTOR.
     */
    public RandomAccessMemory() {}
    
    @Override
    public void clear() {
        this.byteArray.clear();
        this.writeIndex = 0;
        this.readIndex = 0;
    }
    
    @Override
    public void setReadIndex(long index) {
        this.readIndex = index;
    }

    @Override
    public long getReadIndex() {
        return this.readIndex;
    }

    @Override
    public long getWriteIndex() {
        return this.writeIndex;
    }

    @Override
    public void setWriteIndex(long index) {
        this.writeIndex = index;
    }

    @Override
    public void setWriteIndexEof() {
        this.writeIndex = this.byteArray.getSize();
    }
    
    @Override
    public void writeByte(byte b) {
        this.byteArray.set(this.writeIndex++, b);
    }

    @Override
    public byte readByte() {
        return this.byteArray.get(this.readIndex++);
    }

    @Override
    public void writeLong(long l) {
        Bits.putLong(this.buffer, 0, l);
        for (int i = 0; i < 8; i++) {
            this.writeByte(this.buffer[i]);
        }
    }
    
    @Override
    public long readLong() {
        for (int i = 0; i < 8; i++) {
            this.buffer[i] = this.readByte();
        }
        return Bits.getLong(this.buffer, 0);
    }

    @Override
    public void writeInt(int i) {
        Bits.putInt(this.buffer, 0, i);
        for (int j = 0; j < 4; j++) {
            this.writeByte(this.buffer[j]);
        }
    }

    @Override
    public int readInt() {
        for (int i = 0; i < 4; i++) {
            this.buffer[i] = this.readByte();
        }
        return Bits.getInt(this.buffer, 0);
    }

    @Override
    public void writeShort(short s) {
        Bits.putShort(this.buffer, 0, s);
        for (int j = 0; j < 2; j++) {
            this.writeByte(this.buffer[j]);
        }
    }

    @Override
    public short readShort() {
        for (int i = 0; i < 2; i++) {
            this.buffer[i] = this.readByte();
        }
        return Bits.getShort(this.buffer, 0);
    }

    @Override
    public boolean isEof() {
        return this.readIndex >= this.byteArray.getSize();
    }

    @Override
    public void close() {
        this.clear();
        this.byteArray = null;
    }

    @Override
    public int readBytes(byte[] buffer) {
        int nRead = this.byteArray.get(this.readIndex, buffer);
        this.readIndex += nRead;
        return nRead;
    }

    @Override
    public void writeBytes(byte[] bytes, int len) {
        this.byteArray.set(this.writeIndex, bytes, len);
        this.writeIndex += len;
    }

    @Override
    public void skip(long n) {
        throw new UnsupportedOperationException();
    }

}
