/*
 * $Id: Ranking.java 4353 2017-02-05 11:48:42Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;


/**
 * Sortierter Container fuer key(int)-val(double)-Paare
 * mit begrenzter Aufnahmekapazitaet.
 * D.h. es werden unbegrenzt Werte entgegengenommen,
 * jedoch nur die kleinsten (besten) behalten.
 */
public class Ranking {
    private int[] pqIds;
    private double[] pqValues;
    private int pqSize;
    private int pqCapacity;
    
    /**
     * Konstruktor.
     * @param capacity int Aufnahmekapazitaet.
     */
    public Ranking(int capacity) {
        this.pqSize = 0;
        this.pqCapacity = capacity;
        this.pqIds = new int[capacity +1];
        this.pqValues = new double[capacity +1];
    }
    
    /**
     * @return boolean ist leer ja/nein.
     */
    public boolean isEmpty() {return this.pqSize == 0;}
    
    /**
     * Fuegt eine neue Bewertung hinzu.
     * @param id int
     * @param value double Bewertung.
     */
    public void add(int id, double value) {
        if (this.pqSize < this.pqCapacity) {
            this.pqInsert(id, value);
        // PQ ist absteigend sortiert, so dass der schlechteste
        // (groesste) Wert oben ist.
        } else if (value < this.pqValues[1]) {
            this.pqRemove();
            this.pqInsert(id, value);
        }
    }
    
    /**
     * Liefert die finalen IDs des aktuellen Rankings.
     * @return int[] Array aufsteigend nach Rating sortiert,
     * NOT NULL.
     * <br><b>Wichtig:</b><br>
     * Diese Methode leert die PQ, und reinitialisiert daher.
     */
    public int[] flushIds() {
        int ids[] = new int[this.pqSize];
        while (this.pqSize > 0) {
            ids[this.pqSize -1] = this.pqIds[1];
            this.pqRemove();
        }
        return ids;
    }
    
    private void pqInsert(int id, double rating) {
        int pqPos = ++this.pqSize;
        this.pqIds[pqPos] = id;
        this.pqValues[pqPos] = rating;
        this.pqUpheap();
    }

    private void pqRemove() {
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap();
    }

    private void pqSwap(int pos1, int pos2) {
        int id1 = this.pqIds[pos1];
        this.pqIds[pos1] = this.pqIds[pos2];
        this.pqIds[pos2] = id1;
        double rating1 = this.pqValues[pos1];
        this.pqValues[pos1] = this.pqValues[pos2];
        this.pqValues[pos2] = rating1;
    }

    private void pqUpheap() {
        int pos = this.pqSize;
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqValues[parentPos] > this.pqValues[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap() {
        int pos = 1;
        int childPos = pos << 1; // * 2 = linkes Kind 
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqValues[childRightPos] > this.pqValues[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqValues[pos] > this.pqValues[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

}

