/*
 * $Id: SortableTArray.java 3879 2015-08-01 18:26:58Z cmindividual $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Dieses Array waechst nach Bedarf und arbeitet auf Blockbasis.
 * Das heisst, es werden nur Speicherbereiche verbraucht, die
 * tatsaechlich indiziert werden.
 * 
 * @param <T> Generic Type.
 *  
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class SortableTArray<T extends Comparable<T>>
extends TArray<T> implements Swappable {

    private boolean isSorted = false;
    private Swappable swappable;
    
    public SortableTArray() {
    }
    public SortableTArray(Swappable swappable) {
        this.swappable = swappable;
    }
    
    @Override
    public void set(long idx, T obj) {
        if (idx > this.maxIdx +1 || null == obj)
            throw new RuntimeException("Illegal Index or NULL-Object");
        super.set(idx, obj);
        this.isSorted = false;
    };
    
    public synchronized void sort() {
        quicksort(0, this.maxIdx);
        this.isSorted = true;
    }

    private void quicksort(long low, long high) {
        long i = low, j = high;
        T pivot = this.get(low + (high-low)/2);

        while (i <= j) {
            while (this.get(i).compareTo(pivot) < 0) i++;
            while (this.get(j).compareTo(pivot) > 0) j--;

            if (i <= j) {
                this.swap(i, j);
                i++;
                j--;
            }
        }
        if (low < j) quicksort(low, j);
        if (i < high) quicksort(i, high);
    }

    @Override
    public void swap(long i, long j) {
        T tmp = this.get(i);
        this.set(i, this.get(j));
        this.set(j, tmp);
        if (this.swappable != null) swappable.swap(i, j);
    }
    
    @Override
    public final long searchAny(T searchValue) {
        if (this.isSorted) {
            long left = 0;
            long right = this.maxIdx;
            return binarySearch(searchValue, left, right);
        } else {
            for (long i = 0; i <= this.maxIdx; i++) {
                if (this.get(i).equals(searchValue)) return i;
            }
            return NOTFOUND;
        }
    }
 
    private final long binarySearch(T searchValue, long left, long right) {
        if (right < left) return NOTFOUND;
        long mid = (left + right) >>> 1;
        int cmp = searchValue.compareTo(this.get(mid)); 
        if (cmp > 0) {
            return binarySearch(searchValue, mid + 1, right);
        } else if (cmp < 0) {
            return binarySearch(searchValue, left, mid - 1);
        } else {
            return mid;
        }       
    }
    
}
