/*
 * $Id: TArray.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

/**
 * Dieses Array waechst nach Bedarf und arbeitet auf Blockbasis.
 * Das heisst, es werden nur Speicherbereiche verbraucht, die
 * tatsaechlich indiziert werden.<br><br>
 *
 * @param <T> Generic Type.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class TArray<T> implements Iterable<T> {

    private static final int BLOCKBITS = 17;
    private static final int BLOCKSIZE = 1 << BLOCKBITS; // BlockSize = 128k
    private static final int BLOCKMASK = BLOCKSIZE - 1; // 17 Bits Mask
    private static final int SLOTSIZE = 0x2000; // 8K

    public static final int NOTFOUND = Integer.MIN_VALUE;

    protected long maxIdx = -1L;
    private Object[] slots = new Object[SLOTSIZE];

    public void clear() {
        for (int i = 0; i < SLOTSIZE; i++) {
            this.slots[i] = null;
        }
        this.maxIdx = -1;
    }

    /**
     * Liefert den Inhalt als Java-Liste.
     * @return List&lt;T&gt;
     */
    public List<T> asList() {
        List<T> list = new ArrayList<T>();
        for (int i = 0; i < this.getSize(); i++) {
            list.add(this.get(i));
        }
        return list;
    }

    public void push(T t) {
        this.set(this.maxIdx +1, t);
    }

    @SuppressWarnings("unchecked")
    public void set(long idx, T obj) {
        if (idx < 0) throw new ArrayIndexOutOfBoundsException("" + idx);
        if (idx > this.maxIdx) this.maxIdx = idx;
        int slotIdx = (int) (idx >>> BLOCKBITS);
        T[] block = (T[]) this.slots[slotIdx];
        if (null == block) {
            block = (T[]) new Object[BLOCKSIZE];
            this.slots[slotIdx] = block;
        }
        int blockIdx = (int) (idx & BLOCKMASK);
        block[blockIdx] = obj;
    }

    @SuppressWarnings("unchecked")
    public T get(long idx) {
        if (idx < 0) throw new ArrayIndexOutOfBoundsException("" + idx);
        int slotIdx = (int) (idx >>> BLOCKBITS);
        T[] block = (T[]) this.slots[slotIdx];
        if (null == block) return null;
        int blockIdx = (int) (idx & BLOCKMASK);
        return block[blockIdx];
    }

    public long getSize() {
        return this.maxIdx +1;
    }

    /**
     * Sucht den Wert im Array und liefert den ersten
     * (beliebigen) Treffer-Index.
     * @param searchValue zu suchender Wert
     * @return Index oder {@link #NOTFOUND}
     */
    public long searchAny(T searchValue) {
        for (long i = 0; i <= this.maxIdx; i++) {
            if (this.get(i).equals(searchValue)) return i;
        }
        return NOTFOUND;
    }

    /**
     * Sucht den Wert im Array und liefert alle
     * gefundenen Indizes als long[]-Array.
     * @param searchValue NOT NULL zu suchender Wert.
     * @return NOT NULL <code>long[]</code> (Indizes)
     */
    public final long[] searchAll(T searchValue) {
        long retArr[] = new long[10];
        int retArrSize = 0;
        for (long i = 0; i <= this.maxIdx; i++) {
            if (searchValue.equals(this.get(i))) {
                if (retArrSize == retArr.length) { // Ueberlaufschutz.
                    retArr = Arrays.copyOf(retArr, retArr.length * 2);
                }
                retArr[retArrSize++] = i;
            }
        }
        return Arrays.copyOf(retArr, retArrSize);
    }


    @Override
    public Iterator<T> iterator() {
        return new Iterator<T>() {
            private int idx = 0;
            @Override
            public boolean hasNext() {
                return idx < getSize();
            }
            @Override
            public T next() {
                return get(this.idx++);
            }
            @Override
            public void remove() {
                throw new RuntimeException("remove() not supported");
            }
        };
    }

    @Override
    public String toString() {
        // TODO Kritisch ab 2Gig!
        int n = (int) this.getSize();
        String s = "[" + n + "] ";
        for (int i = 0; i < n; i++) {
            if (i > 0) s += ", ";
            String v = this.get(i).toString();
            if (s.length() + v.length() > 70) {
                s += "...";
                break;
            }
            s += v;
        }
        return s;
    }

}
