package de.cm.osm2po.primitives;

/**
 * Kombiniert zwei {@link VarBooleanArray}s und adressiert sie
 * als zweidimensionales Array. Der Wertebereich besteht also nur
 * aus 2 Bits und liegt so bei 0-3. Auf Bit 0 und 1 kann zudem auch
 * direkt zugegriffen werden.
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 */
public class VarDoubleBooleanArray2D implements Var<VarDoubleBooleanArray2D> {

    private int sizeY;
    private int sizeX;
    private VarBooleanArray ba0;
    private VarBooleanArray ba1;

    public VarDoubleBooleanArray2D() {}
    public VarDoubleBooleanArray2D(int sizeY, int sizeX) {
        this.sizeX = sizeX;
        this.sizeY = sizeY;
        int capacity = sizeY * sizeX ;
        ba0 = new VarBooleanArray(capacity);
        ba1 = new VarBooleanArray(capacity);
    }

    public int getSizeY() {return sizeY;}
    public int getSizeX() {return sizeX;}

    private void checkArrayBounds(int y, int x) {
        if (y < 0 || y >= sizeY) throw new ArrayIndexOutOfBoundsException(y);
        if (x < 0 || x >= sizeX) throw new ArrayIndexOutOfBoundsException(x);
    }

    public void set(int y, int x, int value) {
        this.checkArrayBounds(y, x);
        if (value < 0 || value > 3) throw new IllegalArgumentException(""+value);
        int idx = y * sizeX + x;
        boolean bit0 = (value & 0x1) != 0;
        boolean bit1 = (value & 0x2) != 0;
        ba0.set(idx, bit0);
        ba1.set(idx, bit1);
    }

    public int get(int y, int x) {
        this.checkArrayBounds(y, x);
        int idx = y * sizeX + x;
        int val = 0;
        if (ba0.get(idx)) val |= 0x1;
        if (ba1.get(idx)) val |= 0x2;
        return val;
    }

    public void setBit0(int y, int x, boolean bit) {
        this.checkArrayBounds(y, x);
        int idx = y * sizeX + x;
        ba0.set(idx, bit);
    }

    public boolean getBit0(int y, int x) {
        this.checkArrayBounds(y, x);
        int idx = y * sizeX + x;
        return ba0.get(idx);
    }

    public void setBit1(int y, int x, boolean bit) {
        this.checkArrayBounds(y, x);
        int idx = y * sizeX + x;
        ba1.set(idx, bit);
    }

    public boolean getBit1(int y, int x) {
        this.checkArrayBounds(y, x);
        int idx = y * sizeX + x;
        return ba1.get(idx);
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(sizeY);
        outStream.writeInt(sizeX);
        if (sizeY > 0 && sizeX > 0) {
            ba0.writeToStream(outStream);
            ba1.writeToStream(outStream);
        }
    }

    @Override
    public VarDoubleBooleanArray2D readFromStream(InStream inStream) {
        sizeY = inStream.readInt();
        sizeX = inStream.readInt();
        if (sizeY > 0 && sizeX > 0) {
            ba0 = new VarBooleanArray().readFromStream(inStream);
            ba1 = new VarBooleanArray().readFromStream(inStream);
        }
        return this;
    }
}
