/*
 * $Id: VarMergeSorter.java 5244 2021-02-10 19:53:01Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;
import static de.cm.osm2po.misc.Utils.createTmpFile;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;

/**
 * Big MergeSort fuer beliebige {@link Var}-Objekte unter
 * Zuhilfenahme temporaerer Auslagerungsdateien.
 * <br><br><b>Wichtig:</b><br>
 * Die Entnahme der sortieren Werte erfolgt mittels {@link Iterable}.<br>
 * Wird nach der Iteration erneut {@link #add(Var)} gerufen,
 * so wird der Mechanismus schmerzfrei fortgesetzt.<br>
 * Nach Gebrauch sollte {@link #clear()} aufgerufen werden.
 * @author (c) 2015 - Carsten Moeller - info@osm2po.de
 * @param <T> Class of {@link Var}
 */
public class VarMergeSorter<T extends Var<?> & Comparable<T>>
implements Iterable<T> {

    private Class<T> implClass;
    private File workDir; // Directory for TempFiles
    private List<File> mergeFileList = new ArrayList<File>();
    private int threshold; // Number of buffered Elements before flushing to file
    private SortableTArray<T> buffer = new SortableTArray<T>();
    private PriorityQueue<Cursor> pq = new PriorityQueue<Cursor>(0x100);

    /**
     * CTOR.
     * @param implClass ImplClass of T
     * @param threshold int
     *      Maximale Anzahl Datensaetze im Memory bevor in
     *      temporaere Datei ausgelagert wird.
     * @param workDir Directory fuer temporaere Dateien.
     */
    public VarMergeSorter(Class<T> implClass, int threshold, File workDir) {
        this.implClass = implClass;
        this.threshold = threshold;
        this.workDir = workDir;
    }

    /**
     * Fuegt einen neuen Wert hinzu.
     * <br><br><b>Wichtig:</b><br>
     * Der Wert muss zwingend als eigene Instanz, die mittels
     * new-Operator erzeugt wurde, uebergeben werden.
     * Reference-Holder koennen zu unerwarteten Ergebnissen
     * fuehren.
     * @param val Class&lt;T&gt;
     */
    public void add(T val) {
        if (this.buffer.getSize() >= this.threshold) this.flush();
        this.buffer.push(val);
    }

    /**
     * Setzt auf Anfangszustand zurueck und loescht sowohl den Speicher
     * als auch temporaere Dateien
     */
    public void clear() {
        this.buffer.clear();
        for (File file : this.mergeFileList) file.delete();
        this.mergeFileList.clear();
    }

    /**
     * Schreibt sortiert in eine temporaere Datei und leert den Puffer.<br>
     * Dabei wird die Datei intern in die TempFile-Liste aufgenommen.
     */
    private void flush() {
        if (this.buffer.getSize() > 0) {
            File file = createTmpFile(this.workDir);
            this.buffer.sort();
            OutStream outStream = new OutStreamDisk(file);
            for (T val : this.buffer) val.writeToStream(outStream);
            outStream.close();
            this.buffer.clear();
            this.mergeFileList.add(file);
        }
    }

    private class Cursor implements Comparable<Cursor> {
        private InStream inStream;
        private T val;
        private Cursor(File f) {this.inStream = new InStreamDisk(f);}
        private boolean hasNext() {return !this.inStream.isEof();}
        @Override public int compareTo(Cursor that) {return this.val.compareTo(that.val);}
        private Cursor next() {
            try {
                val = implClass.getDeclaredConstructor().newInstance();
                val.readFromStream(this.inStream);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
            return this;
        }
    }

    @Override public Iterator<T> iterator() {
        this.flush();
        this.pq.clear();
        for (File file : mergeFileList) pq.add(new Cursor(file).next());
        return new Iterator<T>() {
            @Override public boolean hasNext() {return !pq.isEmpty();}
            @Override public T next() {
                Cursor cursor = pq.remove();
                T val = cursor.val;
                if (cursor.hasNext()) pq.add(cursor.next());
                else {cursor.inStream.close();}
                return val;
            }
            @Override
            public void remove() {
                throw new UnsupportedOperationException(
                        "Iterator.remove() not supported");
            }
        };
    }
} // EOC
