/*
 * $Id: VarNum.java 5243 2021-02-09 21:03:31Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

/**
 * Diese Klasse kapselt einen <code>long</code> und ist aufgrund seiner
 * {@link Var} Ableitung serialisierbar. Dies allerdings mit der
 * Besonderheit, dass der Longwert in einen je nach Groesse unterschiedlich
 * langen Byte-Strom verwandelt wird. Eine "66" wuerde z.B. nur als
 * ein einziges Byte geschrieben werden.
 * Hingegen wuerde ein MaxLong tatsaechlich 9 Bytes benoetigen
 * und negative Werte immer 10 Bytes.
 * Diese Klasse bietet aber auch statische Methoden, mit denen
 * direkt gearbeitet werden kann.
 * <br><br><b>Wichtig:</b><br>
 * Es existieren hier zwei statische Varianten:
 * Signed und Unsigned. Sind ueberwiegend positive Werte zu
 * erwarten, so ist die Standard-Unsigned-Variante zu bevorzugen,
 * da sie kein zusaetzliches Vorzeichen-Bit benoetigt.
 * Negative Werte werden auch hier korrekt serialisiert, verbrauchen
 * jedoch unnoetig viele Bytes, da das Vorzeichen in Bit-63 sitzt.
 * Achtung also bei casts von negativen bytes, shorts oder ints nach long.
 * Werden auch viele negative Werte verarbeitet, so sollte die
 * signed-Variante verwendet werden.<br>
 * Die {@link Var}-Serialisierung laeuft hingegen immer auf
 * unsigned!
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class VarNum implements Var<VarNum>, Comparable<VarNum> {

    private long longValue;

    public VarNum() {}


    public VarNum(long longValue) {
        this.longValue = longValue;
    }

    public long getLong() {
        return this.longValue;
    }

    public void setLong(long longValue) {
        this.longValue = longValue;
    }

    @Override
    public String toString() {
        return String.valueOf(this.longValue);
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof VarNum)) return false;
        VarNum that = (VarNum) obj;
        return this.longValue == that.longValue;
    }

    @Override
    public int hashCode() {
        // HashCode wie gross Long
        return (int)(this.longValue ^ (this.longValue >>> 32));
    }

    @Override
    public int compareTo(VarNum that) {
        if (this.longValue < that.longValue) return -1;
        else if (this.longValue > that.longValue) return 1;
        else return 0;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        writeUnsigned(this.longValue, outStream);
    }

    @Override
    public VarNum readFromStream(InStream inStream) {
        this.longValue = readUnsigned(inStream);
        return this;
    }

    /**
     * Schreibt einen Long-Wert in einen Stream.
     * Der Wert selbst ist dabei derart Byte-Codiert,
     * dass kleine Zahlen wenig Speicher verbrauchen.
     * Negative Longs werden auch korrekt verarbeitet,
     * verbrauchen jedoch unnoetig viele Bytes.
     *
     * @param lval long
     * @param outStream {@link OutStream}
     *
     * @see #writeSigned(long, OutStream)
     */
    public static void writeUnsigned(
            long lval, OutStream outStream) {
        long highBits = lval;
        do {
            long lowBits = highBits & 0x7FL;
            byte b = (byte) lowBits;
            highBits >>>= 7;
            if (highBits > 0) b |= 0x80;
            outStream.writeByte(b);

        } while (highBits > 0);
    }

    /**
     * Schreibt einen Int-Wert in den Stream.
     * Deligiert nach Ausmaskierung der 32 Bits an
     * {@link #writeUnsigned(long, OutStream)}
     * @param ival int
     * @param outStream {@link OutStream}
     */
    public static void writeUnsigned(
            int ival, OutStream outStream) {
        long lval = ival & 0xFFFFFFFFL;
        writeUnsigned(lval, outStream);
    }

    /**
     * Liest einen Long-Wert aus einem Stream.
     * Der Wert selbst ist dabei derart Byte-Codiert,
     * dass kleine Zahlen wenig Speicher verbrauchen.
     * @param inStream {@link InStream}
     * @return long
     */
    public static long readUnsigned(InStream inStream) {
        int idx = 0;
        long lval = 0;
        byte b = 0;
        do {
            b = inStream.readByte();
            long lowByte = b & (byte)0x7F;
            lowByte <<= idx * 7;
            lval |= lowByte;
            idx++;
        } while (b < 0);
        return lval;
    }

    /**
     * Schreibt einen Long-Wert in einen Stream.
     * Kodiert jedoch das Vorzeichen-Bit an die vorderste
     * Position, so dass entgegen der Unsigned-Variante
     * auch kleine negative Zahlen nur wenige Bytes
     * verbrauchen.
     *
     * @param longValue long
     * @param outStream {@link OutStream}
     *
     * @see #writeUnsigned(long, OutStream)
     */
    public static void writeSigned(
            long longValue, OutStream outStream) {
        if (longValue < 0) {
            writeUnsigned(((((longValue + 1) * -1) << 1) | 1), outStream);
        } else {
            writeUnsigned(longValue << 1, outStream);
        }
    }

    /**
     * Liest einen Long-Wert aus einem Stream.
     * Interpretiert jedoch das niedrigste Bit als
     * Vorzeichen.
     *
     * @param inStream {@link InStream}
     * @return long
     *
     * @see #readUnsigned(InStream)
     */
    public static long readSigned(InStream inStream) {
        long longValue = readUnsigned(inStream);
        if ((longValue & 1) == 0) {
            return longValue >>> 1;
        } else {
            return ((longValue >>> 1) + 1) * -1;
        }
    }

}
