/*
 * $Id: VarString.java 5261 2021-04-04 17:37:39Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import static de.cm.osm2po.misc.Utils.UTF8;

import java.util.Arrays;

/**
 * Diese Klasse kapselt einen <code>String</code> und ist aufgrund seiner
 * {@link Var} Ableitung serialisierbar.
 * @author (c) 2012/2020 - Carsten Moeller - info@osm2po
 */
public class VarString implements Var<VarString>, Comparable<VarString> {

    private String string;
    private boolean immutable;

    public final static VarString NULL_JSON = new VarString("null").makeImmutable();
    public final static VarString NULL_SQL = new VarString("NULL").makeImmutable();
    public final static VarString EMPTY = new VarString("").makeImmutable();

    public final static byte[] NULL_JSON_BYTES = new VarString("null").getBytes();
    public final static byte[] NULL_SQL_BYTES = new VarString("NULL").getBytes();

    /**
     * STD-CTOR.
     * Setzt intern eine leere Zeichenfolge ungleich null.
     */
    public VarString() {
        this.string = "";
    }

    public VarString copy() {
        return new VarString(this.string);
    }

    public VarString(char[] c) {
        this.string = null == c ? "" : new String(c);
    }

    public VarString(byte[] b) {
        this.string = null == b ? "" : new String(b, UTF8);
    }

    public VarString(String s) {
        this.string = null == s ? "" : s;
    }

    /** @return true, wenn leere Zeichenfolge. */
    public final boolean isEmpty() {
        return this.string.length() == 0;
    }

    /** @return byte-Array UTF-8 des String */
    public final byte[] getBytes() {
        return this.string.getBytes(UTF8);
    }

    /**
     * Setzt den internen Wert auf die uebergebenen Bytes.
     * @param bytes Array, UTF-8 erwartet.
     * @return this
     */
    public final VarString setBytes(byte[] bytes) {
        checkImmutable(this);
        if (null == bytes) this.string = "";
        else this.string = new String(bytes, UTF8);
        return this;
    }

    /**
     * Setzt den Wert auf nicht veraenderbar.
     * Wird u.a. in {@link #setBytes(byte[])} geprueft.
     * @return this
     */
    public final VarString makeImmutable() {
        this.immutable = true;
        return this;
    }

    private static void checkImmutable(VarString vs) {
        if (null == vs) return;
        if (vs.immutable) throw new IllegalAccessError(
            "Manipulation of immutable VarStrings not allowed");
    }

    public final VarString toLowerCase() {
        return new VarString(this.string.toLowerCase());
    }

    public final VarString toUpperCase() {
        return new VarString(this.string.toUpperCase());
    }

    /**
     * Kettet einen anderen String an diesen an.
     * @param that {@link VarString}
     * @return this
     */
    public final VarString concat(VarString that) {
        checkImmutable(this);
        if (that != null) this.string += that.string;
        return this;
    }

    /**
     * Delegiert an {@link String#split(String)}
     * @param regexp String
     * @return Array von {@link VarString} - NOT NULL.
     * @see #split(char)
     */
    public final VarString[] split(String regexp) {
        String[] strings = this.string.split(regexp);
        int n = strings.length;
        VarString[] varStrings = new VarString[n];
        for (int i = 0; i < n; i++) varStrings[i] = new VarString(strings[i]);
        return varStrings;
    }

    /**
     * Splittet diesen String an einem Trennzeichen.
     * Diese Variante ist sicherer und schneller als
     * die Benutzung von {@link #split(String)}, da keine
     * komplizierte RegularExpression verwendet wird.
     * @param divider char Trenner.
     * @return Array von {@link VarString} - NOT NULL.
     * @see #split(String)
     */
    public final VarString[] split(char divider) {
        VarString[] vsArr = new VarString[1];
        int n = 0, x = 0, y = 0, len = this.string.length();
        while (y < len) {
            y = this.string.indexOf(divider, x);
            if (-1 == y) y = len;
            if (vsArr.length == n) vsArr = Arrays.copyOf(vsArr, n * 2);
            vsArr[n++] = new VarString(this.string.substring(x, y));
            x = y +1;
        }
        return Arrays.copyOf(vsArr, n);
    }

    /**
     * Entfernt Leerzeichen am Anfang und Ende der Zeichenkette.
     * @return this !!!
     */
    public final VarString trim() {
        checkImmutable(this);
        this.string = this.string.trim();
        return this;
    }

    public static VarString toQuotedSql(VarString vs, boolean nullIfEmpty) {

        if (null == vs || (vs.isEmpty() && nullIfEmpty)) return NULL_SQL;

        char[] oldchars = vs.string.toCharArray();
        char[] newchars = new char[oldchars.length * 2 + 2];

        int n = 0;
        newchars[n++] = '\'';
        for (int i = 0; i < oldchars.length; i++) {
            char c = oldchars[i];
            if ('\\' == c) newchars[n++] = '\\';
            if ('\'' == c) newchars[n++] = '\'';
            newchars[n++] = oldchars[i];
        }
        newchars[n++] = '\'';

        return new VarString(Arrays.copyOf(newchars, n));
    }

    public static VarString toQuotedJson(VarString vs, boolean nullIfEmpty) {

        if (null == vs || (vs.isEmpty() && nullIfEmpty)) return NULL_JSON;

        char[] oldchars = vs.string.toCharArray();
        char[] newchars = new char[oldchars.length * 2 + 2];

        int n = 0;
        newchars[n++] = '"';
        for (int i = 0; i < oldchars.length; i++) {
            char c = oldchars[i];
            if ('\"' == c || '\\' == c) newchars[n++] = '\\'; // Escape
            newchars[n++] = oldchars[i];
        }
        newchars[n++] = '"';

        return new VarString(Arrays.copyOf(newchars, n));
    }

    public static byte[] toQuotedJsonBytes(VarString vs, boolean nullIfEmpty) {
        if (null == vs || (vs.isEmpty() && nullIfEmpty)) return NULL_JSON_BYTES;
        return toQuotedJson(vs, nullIfEmpty).getBytes();
    }

    public static byte[] toQuotedSqlBytes(VarString vs, boolean nullIfEmpty) {
        if (null == vs || (vs.isEmpty() && nullIfEmpty)) return NULL_SQL_BYTES;
        return toQuotedSql(vs, nullIfEmpty).getBytes();
    }

    @Override
    public String toString() {
        return this.string;
    }

    @Override
    public int hashCode() {
        return this.string.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof VarString)) return false;
        if (obj == this) return true;
        VarString that = (VarString) obj;
        return this.string.equals(that.string);
    }

    @Override
    public VarString readFromStream(InStream inStream) {
        checkImmutable(this);
        int len = (int) VarNum.readUnsigned(inStream);
        byte[] bytes = new byte[len];
        for (int i = 0; i < len; i++) {
            bytes[i] = inStream.readByte();
        }
        this.string = new String(bytes, UTF8);
        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        byte[] bytes = this.string.getBytes(UTF8);
        int len = bytes.length;
        VarNum.writeUnsigned(len, outStream);
        for (int i = 0; i < len; i++) {
            outStream.writeByte(bytes[i]);
        }
    }

    @Override
    public int compareTo(VarString that) {
        return this.string.compareTo(that.string);
    }

    public boolean startsWith(VarString that) {
        return this.string.startsWith(that.string);
    }

}
