/*
 * $Id: VarTypeDesk.java 5454 2021-11-27 13:04:44Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.primitives;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Zentrales Factory-Register zum Erzeugen von {@link Var} Typen.
 * Diese Typen sollten nicht per <code>new</code> erzeugt werden,
 * sondern stattdessen aus dieser Factory entnommen werden.
 * Ein weiterer Punkt ist die statische Vergabe von Klassen-IDs,
 * die spaeter zur Identifikation in {@link OutStream}
 * bzw. {@link InStream} Klassen Verwendung findet.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class VarTypeDesk {
    // TODO Bidi-Map geschickter loesen.
    private static Map<Byte, Class<? extends Var<?>>> registryForId;
    private static Map<Class<? extends Var<?>>, Byte> registryForClass;

    public static final byte TYPE_NULL = 0;

    static {
        // Instanziierung muss hier erfolgen.
        registryForId = new HashMap<Byte, Class<? extends Var<?>>>(100);
        registryForClass = new HashMap<Class<? extends Var<?>>, Byte>(100);
        registerPrimitiveTypes();
    }

    private VarTypeDesk() {/*STATIC*/}

    @SuppressWarnings("unchecked")
    public static void registerType(byte typeId, String clazz, ClassLoader classLoader) {
        try {
            Class<?> varClass = Class.forName(clazz, true, classLoader);
            registerType(typeId, (Class<? extends Var<?>>) varClass);
        } catch (ClassNotFoundException e) {
            throw new RuntimeException("Class not found: " + clazz);
        } catch (Throwable t) {
            throw new RuntimeException(t.getMessage());
        }
    }

    /**
     * Registriert einen neuen {@link Var}-Typen.
     * @param typeId byte. Custom-Werte sollten &gt;0 sein.
     * @param clazz {@link Class} of {@link Var}
     */
    public static void registerType(byte typeId, Class<? extends Var<?>> clazz) {
        if (TYPE_NULL == typeId)
            throw new RuntimeException("Type 0 is reserved for NULL");

        checkGetterType(clazz);

        Class<?> c = registryForId.get(Byte.valueOf(typeId));
        Byte t = registryForClass.get(clazz);
        if (null == t && null == c) {
            registryForId.put(Byte.valueOf(typeId), clazz);
            registryForClass.put(clazz, Byte.valueOf(typeId));
        } else if (!(Byte.valueOf(typeId).equals(t) && clazz.equals(c))) {
            throw new RuntimeException("VarType Id " + typeId
                    + " or " + clazz.getName() + " already reserved");
        }
    }

    private static void checkGetterType(Class<? extends Var<?>> clazz) {
        try {
            Method method = clazz.getMethod(
                    "readFromStream", InStream.class);
            Class<?> returnType = method.getReturnType();
            if (!returnType.isAssignableFrom(clazz)) throw new RuntimeException(
                    "Wrong return type in " + clazz.getName()
                    + "." + "readFraomStream()");

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static void registerPrimitiveTypes() {
        registerType((byte) -128, VarBean.class);
        registerType((byte) -127, VarByte.class);
        registerType((byte) -126, VarShort.class);
        registerType((byte) -125, VarInt.class);
        registerType((byte) -124, VarLong.class);
        registerType((byte) -123, VarNum.class);
        registerType((byte) -122, VarString.class);
        registerType((byte) -121, VarByteArray.class);
        registerType((byte) -120, VarShortArray.class);
        registerType((byte) -119, VarIntArray.class);
        registerType((byte) -118, VarLongArray.class);
        registerType((byte) -117, IntIntBag.class);
        registerType((byte) -116, IntLongBag.class);
        registerType((byte) -115, LongIntBag.class);
        registerType((byte) -114, LongLongBag.class);
        registerType((byte) -113, IntSet.class);
        registerType((byte) -112, LongSet.class);
        registerType((byte) -111, IntIntHashMap.class);
        registerType((byte) -110, IntFloatHashMap.class);
        registerType((byte) -109, LongLongHashMap.class);
        registerType((byte) -100, RandomAccessArray.class);
        registerType((byte)  -99, VarBooleanArray.class);
        registerType((byte)  -98, VarBooleanArray2D.class);
        registerType((byte)  -97, VarDoubleBooleanArray2D.class);
        registerType((byte)  -96, VarByteArray2D.class);
    }

    public static Class<? extends Var<?>> classOfType(byte typeId) {
        try {
            return registryForId.get(Byte.valueOf(typeId));
        } catch (Throwable t) {
            throw new RuntimeException("No Var-Type registered for Type ID "
                    + typeId + " or Default Constructor missing.");
        }
    }

    public static Var<?> newInstanceOf(byte typeId) {
        try {
            Class<?> clazz = registryForId.get(Byte.valueOf(typeId));
            return (Var<?>) clazz.getDeclaredConstructor().newInstance();
        } catch (Throwable t) {
            throw new RuntimeException("No Var-Type registered for Type ID "
                    + typeId + " or Default Constructor missing.");
        }
    }

    public static byte typeIdOf(Var<?> var) {
        @SuppressWarnings("unchecked")
        Class<? extends Var<?>> clazz = (Class<? extends Var<?>>) var.getClass();
        return typeIdOf(clazz);
    }
    public static byte typeIdOf(Class<? extends Var<?>> clazz) {
        try {
            return registryForClass.get(clazz).byteValue();
        } catch (Throwable t) {
            throw new RuntimeException(
                    "No VarType registered for Class " + clazz.getName());
        }
    }

    public static byte[] getRegisteredTypes() {
        Set<Byte> keys = registryForId.keySet();
        byte[] types = new byte[keys.size()];
        int n = 0;
        for (Byte type : keys) {
            types[n++] = type.byteValue();
        }
        return types;
    }

}
