/*
 * $Id: JsonBuilder.java 5851 2024-02-15 18:32:00Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.reflection;

import static java.time.format.DateTimeFormatter.ISO_OFFSET_DATE_TIME;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.Temporal;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import de.cm.osm2po.reflection.JsonDeserializer.JsonString;
import de.cm.osm2po.reflection.Reflector.Getter;
import de.cm.osm2po.reflection.Reflector.Reflectable;

/**
 * Ein simpler Json-Builder.<br>
 * Komplexe Objekte werden per Reflection auf Feld- oder Mehoden-Ebene
 * untersucht. Das {@link Field} muss dafuer <code>public</code>
 * oder mittels {@link Reflectable} annotiert sein.
 * {@link Method}s oder <code>static</code> Fields muessen explizit
 * annotiert werden. Methods muessen zudem parameterlos sein.
 * Statische Methoden werden ignoriert.
 * @author (c) 2018/2021 - Carsten Moeller, info@osm2po.de
 * @see Reflector
 * @see Reflectable
 */
public final class JsonBuilder {
    // TODO Bei Zeiten auf neuen StringBuilderUtf8 umstellen

    private JsonBuilder() {/*STATIC*/}

    /**
     * Generiert einen Json-String.
     * @param obj beliegibes {@link Object}
     * @return Json-String
     */
    public static String build(Object obj) {
        return build(obj, null, 0L);
    }

    /**
     * Generiert einen Json-String.
     * @param obj beliegibes {@link Object}
     * @param filter long, BitMask.
     * @return Json-String
     */
    public static String build(Object obj, long filter) {
        return build(obj, null, filter);
    }

    /**
     * Generiert einen Json-String.
     * @param obj beliegibes {@link Object}
     * @param name String NULLABLE, Wert als Objekt mit diesem Bezeichner bauen.
     * @return Json-String
     */
    public static String build(Object obj, String name) {
        return build(obj, name, 0L);
    }


    /**
     * Generiert einen Json-String.
     * @param obj beliegibes {@link Object}
     * @param name String NULLABLE, Wert als Objekt mit diesem Bezeichner bauen.
     * @param filter long, BitMask
     * @return Json-String
     */
    public static String build(Object obj, String name, long filter) {
        StringBuilder sb = new StringBuilder();
        if (null == name) {
            appendValue(obj, sb, filter);
        } else {
            sb.append('{').append(toJsonQuoted(name)).append(':');
            appendValue(obj, sb, filter);
            sb.append('}');
        }
        return sb.toString();
    }

    private static void appendValue(Object obj, StringBuilder sb, long filter) {
        if (null == obj) {
            sb.append("null");
            return;
        }

        Class<? extends Object> clazz = obj.getClass();

        if (clazz.isArray()) {
            Class<?> elmClazz = clazz.getComponentType();
            if (elmClazz.isPrimitive()) {
                if (Long.TYPE == elmClazz)
                    sb.append(Arrays.toString((long[])obj));
                else if (Integer.TYPE == elmClazz)
                    sb.append(Arrays.toString((int[])obj));
                else if (Short.TYPE == elmClazz)
                    sb.append(Arrays.toString((short[])obj));
                else if (Byte.TYPE == elmClazz)
                    sb.append(Arrays.toString((byte[])obj));
                else if (Double.TYPE == elmClazz)
                    sb.append(Arrays.toString((double[])obj));
                else if (Float.TYPE == elmClazz)
                    sb.append(Arrays.toString((float[])obj));

            } else {
                sb.append('[');
                boolean notfirst = false;
                for (Object o : (Object[]) obj) {
                    if (notfirst) sb.append(','); else notfirst = true;
                    appendValue(o, sb, filter);
                }
                sb.append(']');
            }

        } else if (obj instanceof JsonString) {
            sb.append(obj.toString());

        } else if (obj instanceof String) {
            sb.append(toJsonQuoted(obj.toString()));

        } else if (obj instanceof Number) { // auch fuer Primitive
            sb.append(obj.toString());

        } else if (obj instanceof Boolean) { // auch fuer Primitive
            sb.append(obj.toString());

        } else if (obj instanceof Temporal) {

            if (obj instanceof LocalDate) {
                sb.append(toJsonQuoted((LocalDate) obj));

            } else if (obj instanceof LocalDateTime) {
                sb.append(toJsonQuoted((LocalDateTime) obj));

            } else if (obj instanceof ZonedDateTime) {
                sb.append(toJsonQuoted((ZonedDateTime) obj));

            } else if (obj instanceof OffsetDateTime) {
                sb.append(toJsonQuoted(((OffsetDateTime) obj).toZonedDateTime()));
            }

        } else if (obj instanceof Date) {
            sb.append(toJsonQuoted((Date) obj));

        } else if (obj instanceof Calendar) {
            sb.append(toJsonQuoted((Calendar) obj));

        } else if (obj instanceof Enum) {
            sb.append(toJsonQuoted(obj.toString()));

        } else if (obj instanceof Collection) {
            sb.append('[');
            boolean notfirst = false;
            for (Object o : (Collection<?>) obj) {
                if (notfirst) sb.append(','); else notfirst = true;
                appendValue(o, sb, filter);
            }
            sb.append(']');

        } else if (obj instanceof Map) {
            boolean isPropMap = obj instanceof Properties;
            @SuppressWarnings("unchecked")
            Map<Object, Object> map = (Map<Object, Object>) obj;

            sb.append('{');
            boolean notfirst = false;
            for (Entry<Object, Object> entry : map.entrySet()) {
                Object value = entry.getValue();
                if (isPropMap && !(value instanceof String)) {
                    continue; // Bei echten Properties nur Strings durchlassen
                }

                if (notfirst) sb.append(','); else notfirst = true;
                appendValue(entry.getKey().toString(), sb, filter);
                sb.append(':');
                appendValue(entry.getValue(), sb, filter);
            }
            sb.append('}');

        } else if (clazz.getName().startsWith("java.")) {
            sb.append(toJsonQuoted(obj.toString())); // Fallback

        } else { // Custom Object

            boolean notfirst = false;
            sb.append('{');
            Iterator<Getter> getters = Reflector.getGetters(clazz, filter);
            while (getters.hasNext()) {
                try {
                    Getter getter = getters.next();
                    Object objVal = getter.getValue(obj);

                    if (null == objVal && !getter.nullable) {
                        continue;
                    }

                    if (notfirst) sb.append(','); else notfirst = true;
                    appendValue(getter.name, sb, filter);
                    sb.append(':');
                    appendValue(objVal, sb, filter);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
            sb.append('}');
        }
    }


    public static String toJsonQuoted(Calendar cal) {
        if (null == cal) return "null";
        if (cal instanceof GregorianCalendar) {
            ZonedDateTime zdt = ((GregorianCalendar) cal).toZonedDateTime();
            return toJsonQuoted(zdt);
        } else {
            return toJsonQuoted(cal.getTime());
        }
    }

    public static String toJsonQuoted(Date date) {
        if (null == date) return "null";
        ZonedDateTime zdt = ZonedDateTime.ofInstant(
                date.toInstant(), ZoneId.systemDefault());
        return toJsonQuoted(zdt);
    }

    public static String toJsonQuoted(LocalDate ld) {
        if (null == ld) return "null";
        ZonedDateTime zdt = ZonedDateTime.of(ld, LocalTime.ofSecondOfDay(0), ZoneId.systemDefault());
        return toJsonQuoted(zdt);
    }

    public static String toJsonQuoted(LocalDateTime ldt) {
        if (null == ldt) return "null";
        ZonedDateTime zdt = ZonedDateTime.of(ldt, ZoneId.systemDefault());
        return toJsonQuoted(zdt);
    }

    public static String toJsonQuoted(ZonedDateTime zdt) {
        if (null == zdt) return "null";
        final String json = '\"' + zdt.format(ISO_OFFSET_DATE_TIME) + '\"';
        // ISO_OFFSET_DATE_TIME bedeutet ohne Zonen-Suffix wie [Europe/Berlin]
        return json;
    }

    public static String toJsonQuoted(String s) {
        if (null == s) return "null";
        int len = s.length();
        char[] newc = new char[len * 6 + 2];
        int n = 0;
        newc[n++] = '\"';
        for (int i = 0; i < len; i++) {
            char c = s.charAt(i);

            switch (c) {
            case '\"':
                newc[n++] = '\\';
                newc[n++] = '\"';
                break;

            case '\b':
                newc[n++] = '\\';
                newc[n++] = 'b';
                break;

            case '\f':
                newc[n++] = '\\';
                newc[n++] = 'f';
                break;

            case '\n':
                newc[n++] = '\\';
                newc[n++] = 'n';
                break;

            case '\r':
                newc[n++] = '\\';
                newc[n++] = 'r';
                break;

            case '\t':
                newc[n++] = '\\';
                newc[n++] = 't';
                break;

            case '\\':
                newc[n++] = '\\';
                newc[n++] = '\\';
                break;

            default:
                if (c > 127) { // char ist immer positiv (int)0->64k
                    newc[n++] = '\\';
                    newc[n++] = 'u';
                    String hex = Integer.toHexString(c);
                    int xlen = hex.length();
                    for (int x = xlen; x < 4; x++) newc[n++] = '0';
                    for (int x = 0; x < xlen; x++) newc[n++] = hex.charAt(x);

                } else {
                    newc[n++] = c;
                }
            }
        }
        newc[n++] = '\"';
        return new String(newc, 0, n);
    }

    // -------------------- Pretty Print -------------------------

    private static void prettyNewLine(String indent, int level, StringBuilder sb) {
        sb.append('\n');
        for (int i = 0; i < level; i++) sb.append(indent);
    }

    public static String prettyPrint(String json) {
        return prettyPrint(json, "  ");
    }

    // TODO prettyPrint in eigene Klasse auslagern
    public static String prettyPrint(String json, String indent) {

        StringBuilder sb = new StringBuilder(json.length() * 2);
        int level = 0;
        boolean stringMode = false;
        char prevChar = ' ';

        for (char c : json.toCharArray()) {
            if (stringMode) {
                if ('\"'== c && '\\'!=prevChar) stringMode = false;
                prevChar = c;
                sb.append(c);

            } else {

                switch(c) {
                case ' ':
                case '\n':
                case '\r':
                case '\t': break; // whitespace

                case ']': case'}':
                    prettyNewLine(indent, --level, sb);
                    sb.append(c);
                    break;

                case '[': case '{': case ',':
                    sb.append(c);
                    if (c != ',') level++;
                    prettyNewLine(indent, level, sb);
                    break;

                default:
                    sb.append(c);
                    stringMode = ('\"'==c);
                    break;
                }
            }
        }
        return sb.toString();
    }


}
