/*
 * $Id: BidiVertexRouter.java 5466 2021-12-02 20:45:54Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.routing.Graph.SUPPORT_REVERSE;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;

/**
 * Bidirektionale Dijkstra-Implemetation eines Routings auf einem
 * statischen Graphen.<br>
 * Dieser Router beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * </ul>
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see Graph
 * @see VertexRouter
 * @see VisitedVerticesAware
 */
public class BidiVertexRouter extends AbstractRouter<Graph>
implements VisitedVerticesAware {

    public BidiVertexRouter(Graph graph, Log log) {
        super(graph, log);

        if (!graph.isSupported(SUPPORT_REVERSE)) throw new IllegalStateException(
                "Graph does not support Reverse-Routing");

        this.edgeIds = graph.getEdgeIds();
        this.edgeEntries = graph.getEdgeEntries();
        this.edgeSourceIds = graph.getEdgeSourceIds();
        this.edgeTargetIds = graph.getEdgeTargetIds();
        this.edgeNoTurnBits = graph.getEdgeNoTurnBits();

        this.edgeEntriesR = graph.getEdgeEntriesR();
        this.edgeIdxsR = graph.getEdgeIdxsR();
        this.edgeNoTurnBitsR = graph.getEdgeNoTurnBitsR();
    }

    // BIDI-Router is not optimal. See:
    // http://cs.stackexchange.com/questions/53943/is-the-bidirectional-dijkstra-algorithm-optimal
    private final static int FLAG_SETTLED = 0x80000000; // highest bit
    private final static int PQ_INITIAL_CAPACITY = 10000;

    // PQ and OpenClosedList (WBG WhiteGreyBlack)
    private int fwdPqSize, revPqSize; // Number of elements in PQ
    private int fwdPqCapacity, revPqCapacity; // Size of PQ-Array
    private int[] fwdPqEdgeIdxs, revPqEdgeIdxs; // Edge inside PQ
    private double[] fwdPqCost, revPqCost; // Corresponding path cost inside PQ
    private int fwdPqrEdgeIdx, revPqrEdgeIdx; // Current Edge after pqRemove()
    private double fwdPqrCost, revPqrCost; // Cumulated cost after pqRemove()
    private int[] fwdWgbs, revWgbs; // OpenClosed (WhiteGreyBlack) -List

    // Graph accessors
    private int[] edgeSourceIds, edgeTargetIds, edgeIds, edgeEntries;
    private int[] edgeIdxsR, edgeEntriesR;
    private byte[] edgeNoTurnBits;
    private byte[] edgeNoTurnBitsR;

    private float[] edgeCosts; // Runtime-Costs
    private boolean ignoreOneWays; // Runtime-Param
    private boolean findShortestPath; // Runtime-Param
    private boolean ignoreRestrictions; // Runtime-Param

    private int junctionVertexId; // >0 after successful routing
    private int targetIds0;

    private void initParams(RouterParameters params) {
        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();

        if (findShortestPath) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
        }
    }

    private void initPqAndWgb() {
        int nVertices = this.edgeEntries.length;

        this.fwdPqSize = 0;
        this.fwdPqCapacity = PQ_INITIAL_CAPACITY;
        this.fwdPqEdgeIdxs = new int[fwdPqCapacity];
        this.fwdPqCost = new double[fwdPqCapacity];
        this.fwdPqrEdgeIdx = 0;
        this.fwdPqrCost = 0d; // Important
        this.fwdWgbs = new int[nVertices];
        this.fwdWgbs[0] = FLAG_SETTLED;

        this.revPqSize = 0;
        this.revPqCapacity = PQ_INITIAL_CAPACITY;
        this.revPqEdgeIdxs = new int[revPqCapacity];
        this.revPqCost = new double[revPqCapacity];
        this.revPqrEdgeIdx = 0;
        this.revPqrCost = 0d; // Important
        this.revWgbs = new int[nVertices];
        this.revWgbs[0] = FLAG_SETTLED;
    }


    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (constraints.targetIds.length > 1) throw new IllegalArgumentException(
                "OneToMany-Routing with more than One Target not supported");

        this.targetIds0 = constraints.targetIds[0]; // merken fuer isVisited()-Check

        initParams(params);
        initPqAndWgb();

        this.junctionVertexId = this.traverseInternal(
                sourceId, this.targetIds0, constraints.maxCost);
        return this.junctionVertexId > 0;
    }

    @Override
    public boolean isVisited(int targetId) {
        return this.junctionVertexId > 0 && this.targetIds0 == targetId;
    }

    @Override
    public int[] makePath(int targetId) {
        if (!isVisited(targetId)) return null;

        int fwdPath[] = this.makePath(junctionVertexId, false);
        int revPath[] = this.makePath(junctionVertexId, true);
        int path[] = new int[fwdPath.length + revPath.length];
        System.arraycopy(fwdPath, 0, path, 0, fwdPath.length);
        System.arraycopy(revPath, 0, path, fwdPath.length, revPath.length);

        return path;
    }

    private int traverseInternal(int sourceId, int targetId, double maxCost) {

        this.fwdWgbs[sourceId] = FLAG_SETTLED;
        this.revWgbs[targetId] = FLAG_SETTLED;
        this.expandFwd(sourceId);
        this.expandRev(targetId);

        for (;;) {
            if (maxCost < this.fwdPqrCost + this.revPqrCost) return 0;
            if (this.fwdPqrCost <= this.revPqrCost) {
                if (0 == this.fwdPqSize) return 0;
                this.pqRemoveFwd();
                int settledTargetId = this.edgeTargetIds[this.fwdPqrEdgeIdx];
                this.fwdWgbs[settledTargetId] = this.fwdPqrEdgeIdx | FLAG_SETTLED;
                if ((this.revWgbs[settledTargetId] & FLAG_SETTLED) != 0) return settledTargetId;
                this.expandFwd(settledTargetId);
            } else {
                if (0 == this.revPqSize) return 0;
                this.pqRemoveRev();
                int settledSourceId = this.edgeSourceIds[this.revPqrEdgeIdx];
                this.revWgbs[settledSourceId] = this.revPqrEdgeIdx | FLAG_SETTLED;
                if ((this.fwdWgbs[settledSourceId] & FLAG_SETTLED) != 0) return settledSourceId;
                this.expandRev(settledSourceId);
            }
        }
    }

    private int expandFwd(int sourceId) { // Default Forward Routing
        int fromEdgeIdx = this.fwdPqrEdgeIdx;
        int noTurnBits = this.edgeNoTurnBits[fromEdgeIdx];
        int noTurnBit = 1;
        int edgeIdx = this.edgeEntries[sourceId];
        do {
            int targetId = this.edgeTargetIds[edgeIdx];
            if ((0 == (FLAG_SETTLED & this.fwdWgbs[targetId]))) { // not settled
                boolean mayTurn = true;
                if (noTurnBits != 0 && !this.ignoreRestrictions)
                    mayTurn &= ((noTurnBits & noTurnBit) == 0);
                if (!this.ignoreOneWays)
                    mayTurn &= (!((this.edgeIds[edgeIdx] & 0x80000001) == 0x80000000));
                if (mayTurn) {
                    double cost = this.calcEdgeCost(
                            edgeIdx, fromEdgeIdx, this.fwdPqrCost, false);
                    if (cost >= 0d) // Edge ignorieren bei negativen Kosten
                        this.pqInsertOrUpdateFwd(edgeIdx, this.fwdPqrCost + cost, targetId);
                }
            }
            edgeIdx++;
            noTurnBit <<= 1;
        } while (sourceId == this.edgeSourceIds[edgeIdx]);
        return 0;
    }

    private int expandRev(int targetId) { // Redundant but slightly different code
        int fromEdgeIdx = this.revPqrEdgeIdx;
        int noTurnBits = this.edgeNoTurnBitsR[fromEdgeIdx];
        int noTurnBit = 1;
        int edgeIdxIdx = this.edgeEntriesR[targetId];
        int edgeIdx = this.edgeIdxsR[edgeIdxIdx]; // real edgeIdx
        do {
            int sourceId = this.edgeSourceIds[edgeIdx];
            if (0 == (FLAG_SETTLED & this.revWgbs[sourceId])) { // Not settled
                boolean mayTurn = true;
                if (noTurnBits != 0 && !this.ignoreRestrictions)
                    mayTurn &= ((noTurnBits & noTurnBit) == 0);
                if (!this.ignoreOneWays)
                    mayTurn &= (!((this.edgeIds[edgeIdx] & 0x80000001) == 0x80000000));
                if (mayTurn) {
                    double cost = this.calcEdgeCost(
                            edgeIdx, fromEdgeIdx, this.revPqrCost, true);
                    if (cost >= 0d)  // Edge ignorieren bei negativen Kosten
                        this.pqInsertOrUpdateRev(edgeIdx, this.revPqrCost + cost, sourceId);
                }
            }
            edgeIdx = this.edgeIdxsR[++edgeIdxIdx];
            noTurnBit <<= 1;
        } while (targetId == this.edgeTargetIds[edgeIdx]);
        return 0;
    }

    private final int[] makePath(int junctionId, boolean reverse) {
        int[] edgeVertexIds = reverse ? this.edgeTargetIds : this.edgeSourceIds;
        int[] wgbs = reverse ? this.revWgbs : this.fwdWgbs;
        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY);
            int edgeIdx = wgbs[junctionId] &~ FLAG_SETTLED;
            if (edgeIdx == 0) break;
            junctionId = edgeVertexIds[edgeIdx];
            pathIds[n++] = edgeIdx;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        if (!reverse) for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    @Override
    public final int[] getVisited() {
        int fwdVisited[] = this.getVisited(false);
        int revVisited[] = this.getVisited(true);
        int visited[] = new int[fwdVisited.length + revVisited.length];
        System.arraycopy(fwdVisited, 0, visited, 0, fwdVisited.length);
        System.arraycopy(revVisited, 0, visited, fwdVisited.length, revVisited.length);
        return visited;
    }

    private final int[] getVisited(boolean reverse) {
        int[] wgbs = reverse ? this.revWgbs : this.fwdWgbs;
        int len = wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) { // wgbs starts with Idx 1
            int edgeIdx = wgbs[vertexId];
            if (edgeIdx < 0) { // Edge is visited
                if (buffer.length < vertexId) // Overflow check
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    /**
     * Liefert die Kosten einer Kante.
     * Diese Implementation liefert entweder die Distanz oder die Zeit.
     * @param edgeIdx int Index in der Adjazenzliste.
     * @param fromEdgeIdx int Index der Vorgaenger-Edge
     * @param costToHere double bis hier aufgelaufende Weg-Kosten.
     * @param reverse true: Wird vom Rueckwaerts-Router gerufen
     * @return double Kosten.
     */
    protected double calcEdgeCost(
            int edgeIdx, int fromEdgeIdx, double costToHere, boolean reverse) {
        return this.edgeCosts[edgeIdx];
    }


    /* ------------------------------------------------ *
     *            PriorityQueue Forward                 *
     * ------------------------------------------------ */

    private void pqGrowFwd() {
        this.fwdPqCapacity += PQ_INITIAL_CAPACITY;
        this.fwdPqEdgeIdxs = Arrays.copyOf(this.fwdPqEdgeIdxs, this.fwdPqCapacity);
        this.fwdPqCost = Arrays.copyOf(this.fwdPqCost, this.fwdPqCapacity);
        log.debug("Forward priority queue capacity increased to " + this.fwdPqCapacity);
    }

    private void pqInsertOrUpdateFwd(int edgeIdx, double cost, int targetId) {
        int pqPos = this.fwdWgbs[targetId];

        if (0 == pqPos) { // new unvisited target vertex
            pqPos = ++this.fwdPqSize;
            if (this.fwdPqSize >= this.fwdPqCapacity) this.pqGrowFwd();

            this.fwdPqEdgeIdxs[pqPos] = edgeIdx;
            this.fwdPqCost[pqPos] = cost;
            this.fwdWgbs[targetId] = pqPos; // Caution, can be Source or Target

        } else if (this.fwdPqCost[pqPos] > cost) { // shorter path to vertex found
            this.fwdPqEdgeIdxs[pqPos] = edgeIdx;
            this.fwdPqCost[pqPos] = cost;

        } else {
            return;
        }

        this.pqUpheapFwd(pqPos);
    }

    private void pqRemoveFwd() {
        this.fwdPqrEdgeIdx = this.fwdPqEdgeIdxs[1];
        this.fwdPqrCost = this.fwdPqCost[1];
        this.pqSwapFwd(1, this.fwdPqSize);
        this.fwdPqSize--;
        this.pqDownheapFwd(1);
    }

    private void pqSwapFwd(int pos1, int pos2) {
        // PQ-Swap
        int edgeIdx1 = this.fwdPqEdgeIdxs[pos1];
        int edgeIdx2 = this.fwdPqEdgeIdxs[pos2];
        double cost1 = this.fwdPqCost[pos1];
        double cost2 = this.fwdPqCost[pos2];

        this.fwdPqEdgeIdxs[pos1] = edgeIdx2;
        this.fwdPqCost[pos1] = cost2;
        this.fwdPqEdgeIdxs[pos2] = edgeIdx1;
        this.fwdPqCost[pos2] = cost1;

        // Reconstruct (vertex)position in WGB-Array
        int targetId1 = this.edgeTargetIds[edgeIdx1];
        int targetId2 = this.edgeTargetIds[edgeIdx2];

        // WGBS-Swap
        this.fwdWgbs[targetId1] = pos2;
        this.fwdWgbs[targetId2] = pos1;
    }

    private void pqUpheapFwd(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            if (this.fwdPqCost[parentPos] < this.fwdPqCost[pos]) break;
            this.pqSwapFwd(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheapFwd(int pos) {
        int childPos = pos << 1; // * 2 = left child
        while (childPos <= this.fwdPqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.fwdPqSize) { // does right child exist?
                if (this.fwdPqCost[childRightPos] < this.fwdPqCost[childPos]) {
                    childPos = childRightPos;
                }
            }
            if (this.fwdPqCost[pos] < this.fwdPqCost[childPos]) break;
            this.pqSwapFwd(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

    /* ------------------------------------------------ *
     *            PriorityQueue Reverse                 *
     * ------------------------------------------------ */

    private void pqGrowRev() {
        this.revPqCapacity += PQ_INITIAL_CAPACITY;
        this.revPqEdgeIdxs = Arrays.copyOf(this.revPqEdgeIdxs, this.revPqCapacity);
        this.revPqCost = Arrays.copyOf(this.revPqCost, this.revPqCapacity);
        log.debug("Reverse priority queue capacity increased to " + this.revPqCapacity);
    }

    private void pqInsertOrUpdateRev(int edgeIdx, double cost, int sourceId) {
        int pqPos = this.revWgbs[sourceId];

        if (0 == pqPos) { // new unvisited target vertex
            pqPos = ++this.revPqSize;
            if (this.revPqSize >= this.revPqCapacity) this.pqGrowRev();

            this.revPqEdgeIdxs[pqPos] = edgeIdx;
            this.revPqCost[pqPos] = cost;
            this.revWgbs[sourceId] = pqPos; // Caution, can be Source or Target

        } else if (this.revPqCost[pqPos] > cost) { // shorter path to vertex found
            this.revPqEdgeIdxs[pqPos] = edgeIdx;
            this.revPqCost[pqPos] = cost;

        } else {
            return;
        }
        this.pqUpheapRev(pqPos);
    }

    private void pqRemoveRev() {
        this.revPqrEdgeIdx = this.revPqEdgeIdxs[1];
        this.revPqrCost = this.revPqCost[1];
        this.pqSwapRev(1, this.revPqSize);
        this.revPqSize--;
        this.pqDownheapRev(1);
    }

    private void pqSwapRev(int pos1, int pos2) {
        // PQ-Swap
        int edgeIdx1 = this.revPqEdgeIdxs[pos1];
        int edgeIdx2 = this.revPqEdgeIdxs[pos2];
        double cost1 = this.revPqCost[pos1];
        double cost2 = this.revPqCost[pos2];

        this.revPqEdgeIdxs[pos1] = edgeIdx2;
        this.revPqCost[pos1] = cost2;
        this.revPqEdgeIdxs[pos2] = edgeIdx1;
        this.revPqCost[pos2] = cost1;

        // Reconstruct (vertex)position in WGB-Array
        int sourceId1 = this.edgeSourceIds[edgeIdx1];
        int sourceId2 = this.edgeSourceIds[edgeIdx2];

        // WGBS-Swap
        this.revWgbs[sourceId1] = pos2;
        this.revWgbs[sourceId2] = pos1;
    }

    private void pqUpheapRev(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            if (this.revPqCost[parentPos] < this.revPqCost[pos]) break;
            this.pqSwapRev(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheapRev(int pos) {
        int childPos = pos << 1; // * 2 = left child
        while (childPos <= this.revPqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.revPqSize) { // does right child exist?
                if (this.revPqCost[childRightPos] < this.revPqCost[childPos]) {
                    childPos = childRightPos;
                }
            }
            if (this.revPqCost[pos] < this.revPqCost[childPos]) break;
            this.pqSwapRev(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
