/*
 * $Id: Graph.java 5555 2022-02-25 20:11:41Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;
import static de.cm.osm2po.misc.ArrayUtils.intArrayConcat;
import static de.cm.osm2po.misc.ArrayUtils.intArrayDistinct;
import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.ZoneUtils.createZoneId;
import static de.cm.osm2po.misc.ZoneUtils.expandZone;
import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static java.util.Arrays.copyOf;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import de.cm.osm2po.converter.GraphBuilder;
import de.cm.osm2po.extensions.GraphExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.ArrayUtils;
import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.model.Edge;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamAdapter;
import de.cm.osm2po.primitives.InStreamAdapterBuffered;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.IndexOpt;
import de.cm.osm2po.primitives.IndexSeqOpt;
import de.cm.osm2po.primitives.IntIntBag;
import de.cm.osm2po.primitives.RandomAccess;
import de.cm.osm2po.primitives.RandomAccessDisk;
import de.cm.osm2po.primitives.RandomAccessMemory;
import de.cm.osm2po.primitives.RandomRead;
import de.cm.osm2po.primitives.Ranking;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;
import de.cm.osm2po.routing.GraphHeader.ExtensionEntry;

/**
 * Graph / Topologie fuer das osm2po-Routing.
 * Liest eine *.gph und stellt diese fuer
 * ein beliebiges Routing zur Verfuegung.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 * @see GraphBuilder
 */
public class Graph implements AbstractGraph {

    /** Keine Unterstuetzung zusaetzlicher Features. */
    public static final int SUPPORT_NOTHING = 0x0;
    /** Untestuetzung von Koordinaten. z.B. fuer AStar oder Vertex-Lookups. */
    public static final int SUPPORT_LATLON = 0x1;
    /** Unterstuetzung von EdgeFlags. */
    public static final int SUPPORT_EDGEFLAGS = 0x2;
    /** Es wird versucht alle Segmente in den Speicher zu laden. Sonst HardDisk. */
    public static final int SUPPORT_INMEMORY = 0x4;
    /** Dynamische Anreicherung des Graphen um Reverse-Routing-Infos. */
    public static final int SUPPORT_REVERSE = 0x8;
    /** Dynamische Einsortierung von Segmenten in ein Raster. */
    public static final int SUPPORT_RASTER_EDGE = 0x10;
    /** Dynamische Einsortierung von Vertices in ein Raster. */
    public static final int SUPPORT_RASTER_VERTEX = 0x20;
    /** Barriers werden in die TurnRestrictions gemergt */
    public static final int SUPPORT_BARRIERS = 0x40;
    /** Extensions werden gecacht */
    public static final int SUPPORT_EXTENSIONS = 0x80;

    /** Zusaetzlich unterstuetzte Features. */
    private int supports = SUPPORT_NOTHING;

    private File graphFile = null;

    private boolean containsWrongWays;
    private int id;

    // VertexZuEdgeArray, dessen Wert auf die IndexPostion der ersten Edge
    // mit diesem Vertex als Source zeigt  (Vertex ID 1 ist an Index 1)
    private int[] edgeEntries;
    // Einzel-EdgeArrays (Size = NumberOfEdges) - Kuenstlicher composite-type.
    private int[] edgeIds; // Die echten EdgeIds.
    private int[] edgeSourceIds; // Sources aller Edges als Liste (0-wertig)
    private int[] edgeTargetIds; // Targets aller Edges als Liste (0-wertig)
    private float[] edgeCostsKm; // Costs als Laenge in Km.
    private float[] edgeCostsH; // Costs als Zeit in Stunden.
    private byte[] edgeNoTurnBits; // Abbiegevorschriften.
    private int[] edgeFlags; // Weitere Flags

    private int[] edgeIdToIdxCache; // Transformed cache von EdgeId zu EdgeIdx - Lazy.

    // Reverse
    private int[] edgeIdxsR; // Indirektion auf EdgeIdxs, sortiert nach targetId
    private int[] edgeEntriesR; // EntrySet aller Vertices nach targetId
    private byte[] edgeNoTurnBitsR; // Abbiegevorschriften rueckwaerts

    // Raster
    private IntIntBag edgeRasterBag;
    private IntIntBag vertexRasterBag;

    private float[] lats; // Y-Koordinaten
    private float[] lons; // X-Koordinaten
    private byte[] vertexClazzes;

    private float minLat;
    private float minLon;
    private float maxLat;
    private float maxLon;

    private RandomRead rrSegments;
    private IndexSeqOpt<VarString> edgeNameIndex;
    private IndexOpt<LatLons> geomsIndex;

    private long fptrEdgeNameIndex;
    private long fptrGeomsIndex;
    private long fptrBaseEof;

    private int nSegments;

    private Log log;

    private Thread constructorThread;
    private volatile boolean halted;

    // Extensions
    private ExtensionEntry[] extensionEntries;
    private Map<Short, Var<?>> extensionDataMap = new HashMap<Short, Var<?>>();

    @Override
    protected Object clone() throws CloneNotSupportedException {
        throw new CloneNotSupportedException();
    }

    /**
     * CTOR.
     * Ruft {@link #Graph(File, Log, int, boolean, Properties)} mit
     * <code>new Log(), {@link #SUPPORT_LATLON}, false</code>.
     * @param graphFile {@link File} Graph-Datei.
     */
    public Graph(File graphFile) {
        this(graphFile, Log.empty(), SUPPORT_LATLON, false, null);
    }

    /**
     * CTOR.
     * Ruft {@link #Graph(InStream, Log, int, boolean, Properties)} mit
     * <code>new Log(), {@link #SUPPORT_LATLON}, false</code>.
     * @param graphInStream {@link InStream}
     */
    public Graph(InStream graphInStream) {
        this(graphInStream, Log.empty(), SUPPORT_LATLON, false, null);
    }

    /**
     * CTOR.
     * Verpackt den {@link InputStream} in einen
     * {@link InStreamAdapter} und ruft {@link #Graph(InStream)}.
     * @param graphInputStream {@link InputStream}
     */
    public Graph(InputStream graphInputStream) {
        this(new InStreamAdapter(graphInputStream));
    }

    /**
     * CTOR.
     * @param graphInStream {@link InStream} Graph-als Stream NOT NULL.
     * @param log {@link Log} NULLABLE.
     * @param supports int Flags. z.B {@link #SUPPORT_LATLON}.
     * @param async true: Der Graph wird asynchron geladen.
     * @param props zusaetzliche {@link Properties} NULLABLE.
     */
    public Graph(final InStream graphInStream, final Log log,
            final int supports, boolean async, final Properties props) {

        this.log = log;
        if (null == this.log) this.log = Log.empty();
        this.supports = supports | SUPPORT_INMEMORY; // Obligatorisch
        this.halted = false;

        if (async) {
            this.constructorThread = new Thread(new Runnable(){
                @Override
                public void run() {
                    try {
                        construct(graphInStream, null, props);
                    } catch (Throwable e) {
                        halted = true;
                        log.fatal(e);
                    }
                }
            });

            this.constructorThread.setName("osm2po-GraphConstructorThread");
            this.constructorThread.start();

        } else {
            construct(graphInStream, null, props);
        }
    }

    /**
     * CTOR.
     * @param graphFile {@link File} Graph-Datei NOT NULL.
     * @param log {@link Log} NULLABLE.
     * @param supports int Flags. z.B {@link #SUPPORT_LATLON}.
     * @param async true: Der Graph wird asynchron geladen.
     * @param props zusaetzliche {@link Properties} NULLABLE.
     */
    public Graph(final File graphFile, final Log log, final int supports,
            final boolean async, final Properties props) {

        this.log = log;
        if (null == this.log) this.log = Log.empty();
        this.supports = supports;
        this.halted = false;

        if (async) {
            this.constructorThread = new Thread(new Runnable(){
                @Override
                public void run() {
                    InStream graphInStream = null;
                    try {
                        graphInStream = new InStreamAdapterBuffered(
                                new FileInputStream(graphFile));
                        construct(graphInStream, graphFile, props);
                        Graph.this.graphFile = graphFile;
                    } catch (Throwable e) {
                        halted = true;
                        log.fatal(e);
                    } finally {
                        if (graphInStream != null) {
                            graphInStream.close();
                        }
                    }
                }
            });

            this.constructorThread.setName("osm2po-GraphConstructorThread");
            this.constructorThread.start();

        } else {
            InStream graphInStream = null;
            try {
                graphInStream = new InStreamAdapterBuffered(
                        new FileInputStream(graphFile));
                construct(graphInStream, graphFile, props);
                this.graphFile = graphFile;
            } catch (Throwable t) {
                throw new RuntimeException(t);
            } finally {
                // Wichtig wg. zus. Thread
                if (graphInStream != null) graphInStream.close();
            }
        }


    }

    /**
     * Wartet bis der Konstruktor den Graphen
     * im asynchronen Modus geladen hat.
     * @return true: ok, false: Das Laden wurde abgebrochen.
     */
    public boolean sync() {
        if (null == this.constructorThread) return true;
        try {
            this.constructorThread.join();
            return !this.halted;
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Startet dynamisch berechnete Erweiterungen.
     */
    private void supportDynamics() {
        if (this.halted) return;

        // Nachgelagerte Anreicherungen
        if (this.isSupported(SUPPORT_REVERSE)) this.createReverseGraph(); // MemorySensitive!
        if (this.isSupported(SUPPORT_RASTER_VERTEX)) this.createVertexRaster(); // MemorySensitive!
        if (this.isSupported(SUPPORT_RASTER_EDGE)) this.createEdgeRaster(); // MemorySensitive!
        if (this.isSupported(SUPPORT_BARRIERS)) this.createCarBarriersAsNoTurnBits();
    }

    /**
     * Prueft ob der Graph ein oder mehrere Zusatzfeatures unterstuetzt.
     * Jedes dieser Features benoetigt entsprechend mehr RAM.
     * @param SUPPORT int-Flags, z.B {@value #SUPPORT_LATLON}
     * @return Unterstuetzung ja/nein.
     */
    public boolean isSupported(int SUPPORT) {
        if (this.halted) return false;
        return (this.supports & SUPPORT) == SUPPORT;
    }

    public boolean containsWrongWays() {return this.containsWrongWays;}
    public int getGraphId() {return this.id;}

    // Edges
    public final int[] getEdgeEntries() {return this.edgeEntries;}
    public final int[] getEdgeIds() {return this.edgeIds;}
    public final int[] getEdgeSourceIds() {return this.edgeSourceIds;}
    public final int[] getEdgeTargetIds() {return this.edgeTargetIds;}
    public final float[] getEdgeCostsKm() {return this.edgeCostsKm;}
    public final float[] getEdgeCostsH() {return this.edgeCostsH;}
    public final byte[] getEdgeNoTurnBits() {return this.edgeNoTurnBits;}
    public final int[] getEdgeFlags() {return this.edgeFlags;}

    // Reverse
    public final int[] getEdgeIdxsR() {return this.edgeIdxsR;}
    public final int[] getEdgeEntriesR() {return this.edgeEntriesR;}
    public final byte[] getEdgeNoTurnBitsR() {return this.edgeNoTurnBitsR;}

    // Vertices
    public final float[] getLats() {return this.lats;}
    public final float[] getLons() {return this.lons;}
    public final float getMinLat() {return this.minLat;}
    public final float getMinLon() {return this.minLon;}
    public final float getMaxLat() {return this.maxLat;}
    public final float getMaxLon() {return this.maxLon;}

    public final byte[] getVertexClazzes() {return this.vertexClazzes;}

    /**
     * Liefert eine Extension-Info.
     * @param varClass Klasse von Var
     * @param id byte ExtensionId
     * @return {@link Var} {@link ExtensionEntry} oder null.
     */
    public final ExtensionEntry findExtensionEntry(Class<? extends Var<?>> varClass, byte id) {
        if (null == this.extensionEntries) return null;
        byte varType = VarTypeDesk.typeIdOf(varClass);
        for (ExtensionEntry ee : this.extensionEntries) {
            if (ee.getVarType() == varType && ee.getId() == id) {
                return ee;
            }
        }
        return null;
    }

    /**
     * Bequemlichkeits-Methode, welche die erste Extension im Graphen
     * findet, die von einem bestimmten Typ ist.
     * @param varClass Class of {@link Var}
     * @param <T> Class of T
     * @return {@link Var}
     * @see #loadFirstExtension(Class)
     * @see #getExtension(Class, byte)
     * @see ArrayUtils#csvBytesToBooleanArray(String)
     */
    public final <T extends Var<?>> T getFirstExtension(Class<T> varClass) {
        if (this.extensionEntries != null)
            for (ExtensionEntry ee : this.extensionEntries) {
                byte varType = VarTypeDesk.typeIdOf(varClass);
                if (ee.getVarType() == varType)
                    return this.getExtension(varClass, ee.getId());
            }
        return null;
    }
    /**
     * Bequemlichkeits-Methode, welche die erste Extension im Graphen
     * laedt, die von einem bestimmten Typ ist.
     * @param varClass Class of {@link Var}
     * @param <T> Class of T
     * @return {@link Var}
     * @see #getFirstExtension(Class)
     * @see #loadExtension(Class, byte)
     */
    public final <T extends Var<?>> T loadFirstExtension(Class<T> varClass) {
        if (this.extensionEntries != null)
            for (ExtensionEntry ee : this.extensionEntries) {
                byte varType = VarTypeDesk.typeIdOf(varClass);
                if (ee.getVarType() == varType)
                    return this.loadExtension(varClass, ee.getId());
            }
        return null;
    }

    /**
     * Liefert eine bereits gecachte Extension.
     * @param varClass Klasse von Var
     * @param id byte ExtensionId
     * @param <T> Class of T
     * @return {@link Var} Extension oder null
     */
    @SuppressWarnings("unchecked")
    public final <T extends Var<?>> T getExtension(Class<T> varClass, byte id) {
        byte varType = VarTypeDesk.typeIdOf(varClass);
        short key = (short) (varType << 8 | id);
        return (T) this.extensionDataMap.get(Short.valueOf(key));
    }


    /**
     * Liefert oder liest eine Extension in den Cache.<br>
     * Funktioniert analog {@link #getExtension(Class, byte)}, allerdings
     * wird die Extension, wenn nicht vorhanden, mittels
     * {@link #loadExtension(Class, byte)} geladen und nachgelagert
     * gecacht, so dass sie beim naechsten Call im Speicher liegt.
     * @param varClass Klasse von Var
     * @param id byte ExtensionId
     * @param <T> Class of T
     * @return {@link Var} Extension oder null
     */
    @SuppressWarnings("unchecked")
    public final <T extends Var<?>> T getOrLoadAndCacheExtension(Class<T> varClass, byte id) {
        byte varType = VarTypeDesk.typeIdOf(varClass);
        short key = (short) (varType << 8 | id);
        Var<?> var = this.extensionDataMap.get(Short.valueOf(key));
        if (null == var) {
            log.info("Caching extension " + varClass.getSimpleName() + "#" + id);
            var = this.loadExtension(varClass, id);
            if (var != null) {
                this.extensionDataMap.put(Short.valueOf(key), var);
            }
        }
        return (T) var;
    }

    /**
     * Sucht eine Extension in der Graphdatei und ruft
     * fuer diese {@link Var#readFromStream(InStream)}.
     * Delegiert an {@link #loadExtension(Class, byte, Properties)}
     * mit Properties = <code>null</code>
     * @param varClass Klasse von Var
     * @param id byte ExtensionId
     * @param <T> Class of T
     * @return {@link Var} Extension oder null
     */
    public final <T extends Var<?>> T loadExtension(Class<T> varClass, byte id) {
        return this.loadExtension(varClass, id, null);
    }

    /**
     * Sucht eine Extension in der Graphdatei und ruft
     * fuer diese {@link Var#readFromStream(InStream)}.
     * @param varClass Klasse of Var
     * @param id byte ExtensionId
     * @param props {@link Properties}, wenn {@link GraphExtension} NULLABLE
     * @param <T> Class of T
     * @return {@link Var} Extension oder null
     * <br><br><b>Wichtig:</b><br>
     * Handelt es sich bei der zu ladenden Erweiterung um eine
     * {@link GraphExtension}, so wird vorher
     * {@link GraphExtension#connect(File, long, Log, Properties)}
     * aufgerufen. Jedoch Achtung! {@link InStream#getReadIndex()}
     * entspricht <strong>nicht</strong> der absoluten Dateiposition!
     */
    @SuppressWarnings("unchecked")
    public final <T extends Var<?>> T loadExtension(Class<T> varClass, byte id, Properties props) {
        try {
            ExtensionEntry ee = this.findExtensionEntry(varClass, id);
            FileInputStream fis = new FileInputStream(this.graphFile);
            fis.getChannel().position(ee.getOffset());
            InStream is = new InStreamAdapterBuffered(fis);
            Var<?> var = VarTypeDesk.newInstanceOf(ee.getVarType());
            if (var instanceof GraphExtension) {
                GraphExtension<?> gex = (GraphExtension<?>) var;
                gex.connect(this.graphFile, ee.getOffset(), this.log, props);
            }
            var.readFromStream(is);
            is.close(); // Closes underlying inputstream
            return (T) var;

        } catch (Exception e) {
            return null;
        }
    }

    /**
     * Ermittelt aus der EdgeId die SegmentId.
     * @param edgeId int
     * @return int SegmentId.
     */
    public static int toSegmentId(int edgeId) {
        return (edgeId & 0x7FFFFFFF) + 1 >>> 1;
    }

    /**
     * Ermittelt aus einer SegmentId die EdgeId ohne
     * Beruecksichtigung der Information OneWay (Bit 32).
     * @param segmentId int
     * @param reverse boolean in Rueckwaertsrichung
     * @return int edgeId
     */
    public static int toEdgeId(int segmentId, boolean reverse) {
        return ((segmentId - 1) << 1) + (reverse ? 2 : 1);
    }

    /**
     * Ermittelt aus der EdgeId, ob es sich um eine
     * Einbahnstrasse in Gegenrichung handelt.
     * @param edgeId int
     * @return boolean, Einbahnstrasse in Gegenrichtung ja/nein.
     */
    public static boolean isWrongWay(int edgeId) {
        return (edgeId & 0x80000001) == 0x80000000;
    }
    /**
     * Ermittelt aus der EdgeId, ob es sich um eine
     * Einbahnstrasse handelt.
     * @param edgeId int
     * @return boolean, Einbahnstrasse ja/nein.
     */
    public static boolean isOneWay(int edgeId) {
        return (edgeId & 0x80000000) == 0x80000000;
    }
    /**
     * Ermittelt aus der EdgeId, ob dies eine Rueckwaerts-Kante
     * ist, es sich hier also um die Spiegelung des
     * Original-Segments handelt.
     * @param edgeId int
     * @return Rueckrichtung, bzw. gespiegeltes Segment.
     */
    public static boolean isReverse(int edgeId) {
        return (edgeId & 0x1) == 0;
    }

    public boolean isInitializing() {
        return this.constructorThread != null && this.constructorThread.isAlive();
    }

    /**
     * Schliesst den Graphen und raeumt den Speicher frei.
     * Ruft ggf. auch alle {@link GraphExtension#disconnect()}.
     * <br><b>Hinweis:</b><br>
     * Diese Methode kann ebenfalls asynchron waehrend
     * des Ladevorgangs gerufen werden, wodurch dieser
     * umgehend unterbrochen und sauber beendet wird.
     */
    public void close() {

        if (this.isInitializing()) {
            log.warn("Graph loading interrupted. Sending HALT signal!");
            this.halted = true;
            this.sync();
            log.warn("HALT signal received. Processing close()");
        }

        if (this.rrSegments != null) {
            this.rrSegments.close();
            this.rrSegments = null;
        }

        this.edgeEntries = null;
        this.edgeCostsH = null;
        this.edgeCostsKm = null;
        this.edgeFlags = null;
        this.edgeIds = null;
        this.edgeNoTurnBits = null;
        this.edgeSourceIds = null;
        this.edgeTargetIds = null;

        this.lats = null;
        this.lons = null;
        this.vertexClazzes = null;

        if (this.edgeRasterBag != null) {
            this.edgeRasterBag.clear();
            this.edgeRasterBag = null;
        }
        if (this.vertexRasterBag != null) {
            this.vertexRasterBag.clear();
            this.vertexRasterBag = null;
        }
        for (Var<?> var : this.extensionDataMap.values())
            if (var instanceof GraphExtension)
                ((GraphExtension<?>) var).disconnect();
    }

    /**
     * Prueft, ob die Datei eine {@link Graph}-Datei ist.
     * @param file {@link File} NULLABLE plus isFile()-Check
     * @return true if GraphFile
     */
    public static boolean isGraphFile(File file) {
        if (null == file || !file.isFile()) return false;
        InStream inStream = null;
        boolean isGraph = false;
        try {
            inStream = new InStreamDisk(file);
            int fileType = inStream.readInt();
            isGraph = GraphBuilder.COMPAT_FILE_TYPE == fileType;
        } catch (Throwable ignore) {}
        if (inStream != null) inStream.close();
        return isGraph;
    }

    /**
     * Oeffnet den Graphen.
     * <br><b>Wichtig:</b>
     * <ul>
     * <li>Der hier uebergebene Stream wird am Ende NICHT geschlossen!</li>
     * <li>Das Log muss vom Konstruktor gesetzt sein NOT NULL.</li>
     * <li>Graph.SUPPORTS muss vom Konstruktor gesetzt sein.</li>
     * </ul>
     * @param graphInStream {@link InStream}
     *      Graph als Stream. NOT NULL.<br>
     * @param graphFile {@link File}
     *      Graph-Datei fuer RandomAccessDisk-Modus. NULLABLE.
     *      Sollte sich, wenn uebergeben um die gleiche Quelle wie
     *      <code>graphInStream</code> handeln.
     * @param props zusaetzliche {@link Properties}. NULLABLE.
     */
    private void construct(final InStream graphInStream, final File graphFile, final Properties props) {
        long offset = graphInStream.getReadIndex();

        if (this.halted) return;

        boolean supportsLatLons = this.isSupported(SUPPORT_LATLON);
        boolean supportsEdgeFlags = this.isSupported(SUPPORT_EDGEFLAGS);
        boolean supportsInMemory = this.isSupported(SUPPORT_INMEMORY);

        if (null == graphFile) {
            if (!supportsInMemory)
                throw new IllegalStateException("InMemorySegments must be supported");
            log.info("Loading Graph from InputStream");
        } else {
            log.info("Loading Graph-File " + graphFile.getAbsolutePath());
        }
        System.gc();

        GraphHeader gh = new GraphHeader();
        gh.readFromStream(graphInStream);

        this.extensionEntries = gh.getExensions();

        // Header Compatibility
        if (gh.getCompatFileType() != GraphBuilder.COMPAT_FILE_TYPE)
            throw new RuntimeException("No osm2po-GraphFile");
        log.info("Created with Version " + gh.getVersion()
                + "." + gh.getCompatVersion());
        if (gh.getCompatVersion() != GraphBuilder.COMPAT_VERSION)
            throw new RuntimeException("Incompatible GraphFile. Version x.x.x."
                    + GraphBuilder.COMPAT_VERSION + " required");


        this.containsWrongWays = gh.isContainsWrongWays();

        int nEdges = gh.getNumberOfEdges();
        int nVertices = gh.getNumberOfVertices();
        this.id = gh.getId();
        this.nSegments = gh.getNumberOfSegments();
        this.fptrBaseEof = gh.getFptrBaseEof();
        this.fptrEdgeNameIndex = gh.getFptrEdgeNameIndex();
        this.fptrGeomsIndex = gh.getFptrGeomIndex();

        log.info("Graph ID is " + this.getGraphId());

        log.info("Loading SourceVertex-EntryPoints");
        if (graphInStream.getReadIndex() - offset != gh.getFptrEdgeEntries()) {
            throw new RuntimeException("Corrupt Fileposition: EdgeEntries");
        }
        this.edgeEntries = new int[nVertices +1];
        for (int i = 0; i < nVertices; i++) {
            if (this.halted) return;
            int v = graphInStream.readInt();
            this.edgeEntries[i + 1] = v;
            if (i % 1000000 == 0)
                log.progress(DF(i) + " of " + DF(nVertices) + " Vertices read.");
        }
        log.info(DF(nVertices) + " Vertices loaded. - " + MB());

        this.edgeIds = new int[nEdges];
        this.edgeSourceIds = new int[nEdges];
        this.edgeTargetIds = new int[nEdges];
        this.edgeCostsH = new float[nEdges];
        this.edgeCostsKm = new float[nEdges];
        this.edgeNoTurnBits = new byte[nEdges];
        log.info("Memory for "+ DF(nEdges) + " edges reserved. - " + MB());

        if (graphInStream.getReadIndex() - offset != gh.getFptrEdges()) {
            throw new RuntimeException("Corrupt Fileposition: Edges");
        }

        Edge edge = new Edge();
        if (supportsEdgeFlags) {
            try {
                this.edgeFlags = new int[nEdges];
            } catch (OutOfMemoryError ome) {
                log.warn("Could not reserve Memory for "
                        + DF(nEdges) + " EdgeFlags");
                this.supports &=~ SUPPORT_EDGEFLAGS;
                supportsEdgeFlags = false;
            }
        }
        for (int i = 0; i < nEdges; i++) {
            if (this.halted) return;
            edge.readFromStream(graphInStream);
            this.edgeIds[i] = edge.getId();
            this.edgeSourceIds[i] = edge.getSourceId();
            this.edgeTargetIds[i] = edge.getTargetId();
            this.edgeNoTurnBits[i] = edge.getNoTurnBits();
            this.edgeCostsKm[i] = edge.getKm();
            this.edgeCostsH[i] = edge.getH();
            if (supportsEdgeFlags) this.edgeFlags[i] = edge.getFlags();
            if (i % 1000000 == 0) log.progress(DF(i) + " of " + DF(nEdges) + " Edges read.");
        }
        log.info(DF(nEdges) + " Edges loaded. - " + MB());


        if (graphInStream.getReadIndex() - offset != gh.getFptrVertexClazzes()) {
            throw new RuntimeException("Corrupt Fileposition: VertexClasses");
        }

        this.vertexClazzes = new byte[nVertices +1];
        for (int i = 0; i < nVertices; i++) {
            if (this.halted) return;
            byte vertexClazz = graphInStream.readByte();
            this.vertexClazzes[i + 1] = vertexClazz;
            if (i % 1000000 == 0)
                log.progress(DF(i) + " of " + DF(nVertices) + " VertexClazzes read.");
        }
        log.info(DF(nVertices) + " VertexClazzes loaded. - " + MB());


        if (graphInStream.getReadIndex() - offset != gh.getFptrVertexCoords()) {
            throw new RuntimeException("Corrupt Fileposition: VertexCoords");
        }

        if (supportsLatLons) {
            try {
                this.lats = new float[nVertices +1];
                this.lons = new float[nVertices +1];
                log.info("Memory for "+ DF(nVertices) + " coords reserved. - " + MB());
            } catch (OutOfMemoryError ome) {
                log.warn("Could not reserve Memory for "
                        + DF(nVertices) + " Coordinates");
                this.lats = null;
                this.lons = null;
                this.supports &=~ SUPPORT_LATLON;
                supportsLatLons = false;
            }
        }

        this.minLat = 90; this.minLon = 180;
        this.maxLat = -90; this.maxLon = -180;
        LatLon latLon = new LatLon();
        for (int i = 0; i < nVertices; i++) {
            if (this.halted) return;
            latLon.readFromStream(graphInStream);
            float lat = (float) latLon.getLat();
            float lon = (float) latLon.getLon();
            if (lat < this.minLat) this.minLat = lat;
            if (lon < this.minLon) this.minLon = lon;
            if (lat > this.maxLat) this.maxLat = lat;
            if (lon > this.maxLon) this.maxLon = lon;

            if (supportsLatLons) {
                this.lats[i + 1] = lat;
                this.lons[i + 1] = lon;
            }

            if (i % 1000000 == 0)
                log.progress(DF(i) + " of " + DF(nVertices) + " Coords read.");
        }

        if (graphInStream.getReadIndex() - offset != gh.getFptrGeomIndex())
            throw new RuntimeException("Corrupt Fileposition: Geoms");

        // Segment-Dateien fuer ReverseLookup oeffnen oder cachen.
        if (supportsInMemory) {
            log.info("Caching Segments into Memory");
            RandomAccess ram = new RandomAccessMemory();

            // Fuer RandomAccess-Pointer-Anpassung
            long fptr = graphInStream.getReadIndex() - offset;

            try {
                final int bufSize = 0x10000/*64k*/;
                byte[] buf = new byte[bufSize];
                long nCached = 0;
                for (long nLeft = gh.getFptrBaseEof() - fptr; nLeft > 0;) {
                    if (this.halted) return;
                    if (nLeft < bufSize) {
                        buf = new byte[(int) nLeft];
                    }
                    int nRead = graphInStream.readBytes(buf);
                    ram.writeBytes(buf, nRead);
                    nCached += nRead; nLeft -= nRead;
                    if (nCached % 0x4000000 == 0) // n*bufSize
                        log.progress(DF(nCached) + " Bytes cached");
                }
                log.info(DF(nCached) + " Bytes cached");

                if (graphInStream.getReadIndex() - offset != gh.getFptrBaseEof()) {
                    throw new RuntimeException("Corrupt Fileposition: EOF");
                }

                // RandomAccess-Pointer anpassen.
                this.fptrEdgeNameIndex -= fptr;
                this.fptrGeomsIndex -= fptr;

                this.rrSegments = ram;

            } catch (OutOfMemoryError ome) {
                ram.close();
                System.gc();
                if (null == graphFile) throw ome;
                log.warn("Not enough memory for InMemorySegments.\n"
                        + "Falling back to direct Disk Access");
                log.debug(Log.translateError(ome, 10));
                supportsInMemory = false;
            }
        }

        // kein else!!! Siehe OutOfMemory-Fehlerbehandlung oben.
        if (!supportsInMemory) {
            this.supports &=~ SUPPORT_INMEMORY;
            this.rrSegments = new RandomAccessDisk(graphFile, true);
        }

        try {
            // Edge-Geometrien
            this.rrSegments.setReadIndex(this.fptrGeomsIndex);
            this.geomsIndex = new IndexOpt<LatLons>();
            this.geomsIndex.connect(this.rrSegments);

            // Edge-Namen
            this.rrSegments.setReadIndex(this.fptrEdgeNameIndex);
            this.edgeNameIndex = new IndexSeqOpt<VarString>();
            this.edgeNameIndex.connect(this.rrSegments);

        } catch (Exception e) {throw new RuntimeException(e);}

        final boolean streamMode = (null == graphFile);
        final boolean supportExtensions = this.isSupported(SUPPORT_EXTENSIONS);
        if (supportExtensions || streamMode) {
            graphInStream.skip(this.fptrBaseEof - (graphInStream.getReadIndex() - offset));

            for (ExtensionEntry ee : this.extensionEntries) {
                final byte eeId = ee.getId();
                final byte eeVarType = ee.getVarType();
                final Class<? extends Var<?>> varClass = VarTypeDesk.classOfType(eeVarType);
                final String eeName = varClass.getSimpleName() +  "#" + eeId;

                // Stream immer vollstaendig lesen
                if (!supportExtensions) {
                    log.info("Extension " + eeName + " skipped");
                } else {
                    try {
                        log.progress("Loading extension " + eeName);
                        Var<?> var = VarTypeDesk.newInstanceOf(ee.getVarType());
                        if (var instanceof GraphExtension) {
                            GraphExtension<?> gex = (GraphExtension<?>) var;
                            gex.connect(graphFile, ee.getOffset(), log, props);
                        }
                        var.readFromStream(graphInStream);
                        short key = (short) (eeVarType << 8 | eeId);
                        this.extensionDataMap.put(key, var);
                        log.info("Extension " + eeName + " loaded");

                    } catch (Throwable t) {
                        log.error("Could not load Extension " + eeName
                                + "\n" + t + "\nStacktrace on debug level");
                        log.debug(Log.translateError(t, 10));
                    }
                }
                // egal, was passiert ist, springe zur naechsten Extension
                long nSkip = ee.getOffset() + ee.getSize()
                    - graphInStream.getReadIndex() + offset;
                if (nSkip > 0) {
                    log.debug("Skipping " + DF(nSkip) + " bytes");
                    graphInStream.skip(nSkip);
                }
            }

            if (graphInStream.getReadIndex() - offset != gh.getFptrEof())
                throw new RuntimeException(
                        "FilePointer not EOF after reading extensions");

        }

        // Dynamische Berechnungen / Erweiterungen anstarten.
        this.supportDynamics();

        log.info(this.supportsToString());

        System.gc();
        this.log.info("Graph is in memory - " + MB() + " free");
    }

    private String supportsToString() {
        return "Graph supports: "
            + (this.isSupported(SUPPORT_LATLON) ? "+" : "-") + "C "
            + (this.isSupported(SUPPORT_INMEMORY) ? "+" : "-") + "M "
            + (this.isSupported(SUPPORT_EDGEFLAGS) ? "+" : "-") + "F "
            + (this.isSupported(SUPPORT_EXTENSIONS) ? "+" : "-") + "X "
            + (this.isSupported(SUPPORT_BARRIERS) ? "+" : "-") + "B "
            + (this.isSupported(SUPPORT_REVERSE) ? "+" : "-") + "R "
            + (this.isSupported(SUPPORT_RASTER_VERTEX) ? "+" : "-") + "V "
            + (this.isSupported(SUPPORT_RASTER_EDGE) ? "+" : "-") + "E ";
    }

    /**
     * Findet den naechsten Vertex zu einer Koordinate im Umkreis
     * von etwa 5 km.
     * Ruft intern {@link #findClosestVertexIds(float, float, int, int)} auf.
     * @param lat float Latitude (Breitengrad)
     * @param lon float Longitude (Laengengrad)
     * @return VertexId oder -1
     */
    public int findClosestVertexId(float lat, float lon) {
        int[] ids = this.findClosestVertexIds(lat, lon, 10, 1);
        if (ids.length > 0) return ids[0]; else return -1;
    }

    /**
     * Findet die naechsten Vertices (IDs) zu einer Koordinate im Umkreis
     * und liefert diese sortiert nach Distanz (kuerzeste vorne) zurueck.
     * <br><br><b>Wichtig:</b><br>
     * Diese Methode liefert exakte Werte, ist dafuer aber auch
     * entsprechend langsam.
     * @param lat float Latitude (Breitengrad)
     * @param lon float Longitude (Laengengrad)
     * @param maxDist float Maximaler Suchradius in km.
     * @param n int Vertices moeglichst finden
     * @return Array von VertexIds sortiert nach Distanz NOT NULL.
     * @see #findClosestVertexIds(float, float, int, int)
     */
    public int[] findClosestVertexIdsSphere(
            float lat, float lon, float maxDist, int n) {
        // TODO findClosesVertexIdsSphere beschleunigen,
        //      z.B. mittels GeoUtils.calcDistanceSphereApprox()

        if (!this.isSupported(SUPPORT_LATLON)) {
            log.warn("Graph does not support Coordinates!");
            return new int[0];
        }

        Ranking ranking = new Ranking(n);

        double minVal = Double.MAX_VALUE;
        int minId = -1;
        int nCoords = this.lats.length;
        for (int vertexId = 1; vertexId < nCoords; vertexId++) { // 1-wertig!
            double dist = GeoUtils.calcDistance(
                    lat, lon, lats[vertexId], lons[vertexId]);
            if (dist > maxDist) {
                continue;
            }
            if (1 == n) { // Schneller ohne Ranking
                if (dist < minVal) {
                    minVal = dist;
                    minId = vertexId;
                }
            } else {
                ranking.add(vertexId, dist);
            }
        }
        if (1 == n && minId != -1) { // Speed up if only one requested
            ranking.add(minId, minVal);
        }
        return ranking.flushIds();
    }

    /**
     * Findet die naechsten Vertices (IDs) zu einer Koordinate im Umkreis
     * und liefert diese sortiert nach Distanz (kuerzeste vorne) zurueck.
     * @param lat float Latitude (Breitengrad)
     * @param lon float Longitude (Laengengrad)
     * @param zoneRadius int. Grober Suchradius. 1 entspricht 0.01 Grad.
     * @param n Maximale Anzahl zurueckzuliefernder Vertices.
     * @return Array von VertexIds sortiert nach Distanz NOT NULL.
     */
    public int[] findClosestVertexIds(float lat, float lon, int zoneRadius, int n) {

        if (!this.isSupported(SUPPORT_LATLON)) {
            log.warn("Graph does not support Coordinates!");
            return new int[0];
        }

        Ranking ranking = new Ranking(n);

        if (this.isSupported(SUPPORT_RASTER_VERTEX)) {
            int zoneId = createZoneId(lat, lon, 2);
            // Pseudo-Ringe +1 von innen nach aussen wachsen lassen.
            for (int r = 1; r <= zoneRadius && ranking.isEmpty(); r++) {
                // Betroffene Zonen ermitteln
                int[] zoneIds = expandZone(zoneId, r, 2);
                for (int z = 0; z < zoneIds.length; z++) {
                    // Alle Vertices einer Zone sammeln
                    int[] vertexIds = this.vertexRasterBag.getAll(zoneIds[z]);
                    if (vertexIds != null) {
                        for (int v = 0; v < vertexIds.length; v++) {
                            int vertexId = vertexIds[v];
                            double dist = GeoUtils.calcDistanceEuklid(
                                    lat, lon, lats[vertexId], lons[vertexId]);
                            ranking.add(vertexId, dist);
                        }
                    }
                }
            }

        } else { // Kein Raster, dann sequentiell mittels vorgeschalteter Manhattenbox
            double minVal = Double.MAX_VALUE;
            int minId = -1;
            int nCoords = lats.length;
            // Korrektur: 1.4 ~ 1 / sin(45 Grad)
            double manhattenBox = zoneRadius / 100.0 * 1.4;
            for (int vertexId = 1; vertexId < nCoords; vertexId++) { // 1-wertig!
                double distM = StrictMath.abs(
                        lat - lats[vertexId]) + StrictMath.abs(lon - lons[vertexId]);
                if (distM < manhattenBox) {
                    double dist = GeoUtils.calcDistanceEuklid(
                            lat, lon, lats[vertexId], lons[vertexId]);
                    if (1 == n) { // Schneller ohne Ranking
                        if (dist < minVal) {
                            minVal = dist;
                            minId = vertexId;
                        }
                    } else {
                        ranking.add(vertexId, dist);
                    }
                }
            }
            if (1 == n && minId != -1) {
                ranking.add(minId, minVal);
            }
        }
        return ranking.flushIds();
    }

    /**
     * Findet alle wegfuehrenden Edges zu einem Vertex.
     * @param sourceId int Id des SourceVertex.
     * @return int[] edgeIdx-Array NOT NULL.
     */
    public int[] findOutgoingEdges(int sourceId) {
        int edgeIdx = this.edgeEntries[sourceId] -1;
        int n = 0;
        int[] edgeIdxs = new int[64];
        while (this.edgeSourceIds[++edgeIdx] == sourceId) {
            if (edgeIdxs.length == n) // ArrayOverflow
                edgeIdxs = copyOf(edgeIdxs, n * 2);
            edgeIdxs[n++] = edgeIdx;
        }
        return Arrays.copyOf(edgeIdxs, n);
    }

    /**
     * Findet alle ankommenden Edges zu einem Vertex.
     * Unterstuetzt der Graph kein {@link Graph#SUPPORT_REVERSE}, so
     * wird das Ergebnis mittels {@link #findOutgoingEdges(int)} und
     * {@link #findReverseEdge(int)} ermittelt. Um hier korrekte
     * Ergebnisse zu liefern, muss der Graph auch WrongWays beinhalten,
     * also Einbahnstrassen in Gegenrichtung.
     * @param targetId int Id des TargetVertex
     * @return int[] edgeIdx-Array NOT NULL.
     */
    public int[] findIncomingEdges(int targetId) {
        int n = 0;
        int[] edgeIdxs = new int[64];
        if (this.isSupported(SUPPORT_REVERSE)) {
            int edgeIdxIdx = this.edgeEntriesR[targetId];
            int edgeIdx = this.edgeIdxsR[edgeIdxIdx];
            while (this.edgeTargetIds[edgeIdx] == targetId) {
                if (edgeIdxs.length == n) // ArrayOverflow
                    edgeIdxs = copyOf(edgeIdxs, n * 2);
                edgeIdxs[n++] = edgeIdx;
                edgeIdx = this.edgeIdxsR[++edgeIdxIdx];
            }
        } else {
            if (!this.containsWrongWays) throw new IllegalStateException(
                    "Graph neither supports ReverseRouting nor contains WrongWays.");
            int[] outs = this.findOutgoingEdges(targetId);
            for (int i = 0; i < outs.length; i++) {
                int idx = this.findReverseEdge(outs[i]);
                if (idx > 0) {
                    if (edgeIdxs.length == n) // ArrayOverflow
                        edgeIdxs = copyOf(edgeIdxs, n * 2);
                    edgeIdxs[n++] = idx;
                }
            }
        }
        return Arrays.copyOf(edgeIdxs, n);
    }

    /**
     * Liefert alle Koordinaten zu einer Edge.
     * @param edgeIdx Index der Edge in der Adjazenzliste.
     * @return {@link LatLons}
     *      Koordinaten des korrespondierenden Segments.
     *      <br><b>Wichtig:</b><br>
     *      Handelt es sich um eine Rueckwaertskante,
     *      so wird die Geometrie reversed.
     */
    public final LatLons lookupGeometry(int edgeIdx) {
        int edgeId = this.edgeIds[edgeIdx];
        int segId = toSegmentId(edgeId);
        LatLons latLons = new LatLons();
        if (segId <= 0) return latLons;
        latLons = this.geomsIndex.fetch(segId -1, latLons);
        if (Graph.isReverse(edgeId)) latLons.reverse();
        return latLons;
    }

    /**
     * Liefert den Strassennamen zu einer Edge.
     * @param edgeIdx Index der Edge in der Adjazenzliste.
     * @return Name der Strasse NOT NULL.
     */
    public final VarString lookupEdgeName(int edgeIdx) {
        VarString vs = new VarString();
        int segId = toSegmentId(this.edgeIds[edgeIdx]);
        if (segId <= 0) return vs;
        vs = this.edgeNameIndex.fetch(segId -1, vs);
        return vs;
    }

    /**
     * Findet die gespiegelte Rueckwaers-Kante, wenn vorhanden.
     * @param edgeIdx int Index-Position in der Adjazenzliste.
     * @return int edgeIdx oder 0, wenn nicht gefunden.
     */
    public int findReverseEdge(int edgeIdx) {
        if (edgeIdx < 0 || edgeIdx >= this.edgeIds.length) return 0;
        int edgeId = this.edgeIds[edgeIdx];
        int targetId = this.edgeTargetIds[edgeIdx];
        int edgeIdRev = isReverse(edgeId) ? edgeId - 1 : edgeId + 1;

        edgeIdx = this.edgeEntries[targetId];
        if (edgeIdx >= 0) { // 0 ist DeadEdge
            int sourceId = this.edgeSourceIds[edgeIdx];
            while (sourceId == targetId) {
                edgeId = this.edgeIds[edgeIdx];
                if (edgeId == edgeIdRev) return edgeIdx; // found
                sourceId = this.edgeSourceIds[++edgeIdx];
            }
        }
        return 0; // not found
    }


    /**
     * Ermittelt, ob zwischen zwei Edges einer Kreuzung eine
     * NoTurn-Restriktion vorhanden ist.
     * @param fromEdgeIdx Index der hinfuehrenden Edge.
     * @param toEdgeIdx Index der wegfuehrenden Edge.
     * @return 0, wenn keine oder Fehler, &gt; 0, Turn-Bit-Index (1-wertig).
     */
    public int isTurnRestricted(int fromEdgeIdx, int toEdgeIdx) {
        int noTurnBits = this.edgeNoTurnBits[fromEdgeIdx];
        if (0 == noTurnBits) return 0;
        int bitIdx = -1;
        int toTargetId = this.edgeTargetIds[fromEdgeIdx];
        int edgeIdx = this.edgeEntries[toTargetId] -1;
        while(this.edgeSourceIds[++edgeIdx] == toTargetId) {
            if (++bitIdx > 7) {
                log.bugfix("More than 8 outgoing edges from Segment "
                        + Graph.toSegmentId(this.edgeIds[fromEdgeIdx]));
                return 0;
            }
            int turnBit = 1 << bitIdx;
            if (edgeIdx == toEdgeIdx) {
                return ((noTurnBits & turnBit) != 0) ? bitIdx +1 : 0;
            }
        }
        return 0;
    }

    /**
     * Erzeugt ein Index-Array fuer eine umgekehrte Zugriffssicht,
     * naemlich von einer bekannten EdgeId zum EdgeIdx innerhalb
     * der Adjazenz-Liste.
     * @return <code>int[]</code>-Array von Zeigern auf EdgeIdxs.<br>
     *      Das heisst: EdgeId 1 befindet sich an Index 0 und
     *      der Wert (wenn nicht 0) zeigt auf die Index-Position
     *      in der Adjazenz-Liste (edgeIdx).<br>
     *      0-Werte im zurueckelieferten Array sind nicht vorhandene Edges.
     */
    public int[] createTransformedEdgeIndex() {
        int[] edgeIdxs = new int[this.nSegments * 2];
        int nEdges = this.edgeIds.length -2; // -2 ohne Sentinels
        for (int i = 1; i <= nEdges; i++) {
            int edgeId = this.edgeIds[i] & 0x7FFFFFFF;
            edgeIdxs[edgeId -1] = i;
        }
        return edgeIdxs;
    }

    /**
     * Liefert zu einer edgeId die edgeIdx.
     * Ruft initial einmal {@link #createTransformedEdgeIndex()}
     * und cached das Ergebnis.
     * @param edgeId int
     * @return edgeIdx int
     */
    public int getEdgeIdxForId(int edgeId) {
        if (null == this.edgeIdToIdxCache) {
            this.edgeIdToIdxCache = createTransformedEdgeIndex();
        }
        return this.edgeIdToIdxCache[edgeId -1];
    }

    /**
     * Rekonstruiert alle Segmentinformationen aus einer Edge.
     * <br><b>Wichtig:</b><br>
     * Wenn es sich um eine Reverse-Edge handelt, dann werden
     * die Koordinaten der Geometrie bereits umgekehrt zurueckgeliefert.
     * @param edgeIdx Index (Position) der Edge im Graphen.
     * @return {@link RoutingResultSegment}
     *      Segment oder null, wenn nicht ermittelbar
     *      oder edgeIdx &lt;= 0.
     */
    public RoutingResultSegment lookupSegment(int edgeIdx) {
        if (edgeIdx <= 0) return null;

        RoutingResultSegment rrs = new RoutingResultSegment();

        int edgeId = this.edgeIds[edgeIdx];
        int segId = toSegmentId(edgeId);

        rrs.setEdgeId(edgeId);
        rrs.setKm(this.edgeCostsKm[edgeIdx]);
        rrs.setH(this.edgeCostsH[edgeIdx]);
        rrs.setSourceId(this.edgeSourceIds[edgeIdx]);
        rrs.setTargetId(this.edgeTargetIds[edgeIdx]);
        rrs.setNoTurnBits(this.edgeNoTurnBits[edgeIdx]);

        this.edgeNameIndex.fetch(segId -1, rrs.getName());
        this.geomsIndex.fetch(segId -1, rrs.getCoords());

        if (this.edgeFlags != null)
            rrs.setFlags(this.edgeFlags[edgeIdx]);

        if (rrs.isReverse()) rrs.getCoords().reverse(); // Gerade IDs sind Reverse-Edges.

        return rrs;
    }

    /**
     * Berechnet die Laenge eines Pfades.
     * @param path int[]-Arrray von EdgedIdx.
     * @return double Laenge in km.
     */
    public double calcPathLength(int[] path) {
        double sum = 0d;
        for (int i = 0; i < path.length; i++) {
            sum += this.edgeCostsKm[path[i]];
        }
        return sum;
    }

    /**
     * Berechnet die Fahrzeit eines Pfades.
     * @param path int[]-Arrray von EdgedIdx.
     * @return double Zeite in h.
     */
    public double calcPathTime(int[] path) {
        double sum = 0d;
        for (int i = 0; i < path.length; i++) {
            sum += this.edgeCostsH[path[i]];
        }
        return sum;
    }

    /**
     * Findet die Indizes aller Vertices einer bestimmten Klasse.
     * Da die Indizes der Id entsprechen, findet diese Methode
     * folglich direkt die VertexIds.
     * @param clazz byte Klasse.
     * @return int[]-Array von Indizes auf Vertices im Graphen.
     */
    public int[] findVerticesByClazz(byte clazz) {
        int nVertices = this.vertexClazzes.length;
        int[] pois = new int[1000];
        int nPois = 0;
        for (int i = 0; i < nVertices; i++) {
            if (clazz == this.vertexClazzes[i]) {
                if (pois.length == nPois) {
                    pois = Arrays.copyOf(pois, pois.length * 2);
                }
                pois[nPois++] = i;
            }
        }
        return Arrays.copyOf(pois, nPois);
    }


    /**
     * Findet nahe Edges zu einer Koordinate im Umkreis.
     * Wenn der Graph ein Raster unterstuetzt, so wird dies
     * herangezogen, ansonsten wird auf eine rudimentaere Suche
     * ueber die dichtesten Vertices zurueckgegriffen.
     * @param lat float Latitude (Breitengrad)
     * @param lon float Longitude (Laengengrad)
     * @param minCount grobe minimale Anzahl zurueckgelieferter Edges.
     * @return int[] EdgeIdxs NOT NULL DISTINCT SORTED BY IDX.
     */
    public int[] findCloseEdges(float lat, float lon, int minCount) {
        int[] edgeIdxs = new int[0];

        if (this.isSupported(Graph.SUPPORT_RASTER_EDGE)) {
            for (int x = 0; x < 4; x++) { // ca. 10km Fang.
                int zones = (x * x) + 1;
                int[] zoneIds = expandZone(createZoneId(lat, lon, 2), zones, 2);
                int[][] edgeIdxs2D = new int[zoneIds.length][];
                for (int i = 0; i < edgeIdxs2D.length; i++) {
                    edgeIdxs2D[i] = this.edgeRasterBag.getAll(zoneIds[i]);
                }
                edgeIdxs = intArrayConcat(edgeIdxs2D);
                if (edgeIdxs.length >= minCount) break;
            }

        } else {
            int[] vertexIds = this.findClosestVertexIds(lat, lon, 10, minCount);

            edgeIdxs = new int[0x400]; // 1k
            int nEdgeIdxs = 0;
            for (int i = 0; i < vertexIds.length; i++) {
                int[] tmpIdxs = this.findOutgoingEdges(vertexIds[i]);
                for (int j = 0; j < tmpIdxs.length; j++) {
                    int edgeIdx = tmpIdxs[j];
                    if (isReverse(this.edgeIds[edgeIdx])) {
                        edgeIdx = this.findReverseEdge(edgeIdx);
                    }
                    if (edgeIdx > 0) {
                        if (nEdgeIdxs == edgeIdxs.length) {
                            edgeIdxs = Arrays.copyOf(edgeIdxs, edgeIdxs.length * 2);
                        }
                        edgeIdxs[nEdgeIdxs++] = edgeIdx;
                    }
                }
            }
            edgeIdxs = Arrays.copyOf(edgeIdxs, nEdgeIdxs);
        }
        int[] edgeIdxsDistinct = intArrayDistinct(edgeIdxs);

        log.debug("Number of close Edges: " + edgeIdxs.length
                + " distinct: " + edgeIdxsDistinct.length);

        return edgeIdxsDistinct;
    }

    /**
     * Setzt alle Bits in den NoTurns, wenn der Target
     * einer Kante eine negative Clazz hat.
     */
    private void createCarBarriersAsNoTurnBits() {
        int n = this.edgeTargetIds.length;
        for (int fromEdgeIdx = 1; fromEdgeIdx < n - 1; fromEdgeIdx++) {
            int targetId = this.edgeTargetIds[fromEdgeIdx];
            if (this.vertexClazzes[targetId] < 0)
                this.edgeNoTurnBits[fromEdgeIdx] = -1; // All Bits set
        }
    }

    private void createVertexRaster() {
        if (this.lats == null || this.lons == null) return;

        try {
            this.vertexRasterBag = new IntIntBag();
            int nVertices = this.edgeEntries.length;
            for (int vertexId = 1; vertexId < nVertices; vertexId++) {
                int zoneId = createZoneId(this.lats[vertexId], this.lons[vertexId], 2);
                this.vertexRasterBag.add(zoneId, vertexId);

                if ((vertexId & 0x1FFFFF) == 0x1FFFFF)
                    log.progress(DF(vertexId) + " of " + DF(nVertices - 1) + " vertices rastered.");
            }

            log.progress("Sorting");
            this.vertexRasterBag.sort();

            log.info(DF(nVertices) + " vertices rastered");

        } catch (OutOfMemoryError ome) {
            this.supports &=~ SUPPORT_RASTER_VERTEX;
            this.vertexRasterBag.clear();
            log.warn("Not enough memory for Raster");
            log.debug(Log.translateError(ome, 10));
        }

    }

    /**
     * Iteriert ueber alle Segmente, ordnet diese einem
     * oder mehrere Raster-Zonen zu und erstellt intern einen
     * entsprechenden Index ueber den spaeter Kanten aufgrund von
     * uebergebenen Koordinaten schnell gefunden werden koennen.
     * <br><b>Wichtig:</b><br>
     * Intern wird zwar ueber die Segmente iteriert, jedoch
     * die Forward-EdgeIdx mittels {@link #createTransformedEdgeIndex()}
     * im Bag gespeichert.
     */
    private void createEdgeRaster() {
        if (this.lats == null || this.lons == null) return;

        log.info("Building Raster Index for Edges");

        try {
            this.edgeRasterBag = new IntIntBag();
            LatLons coords = new LatLons();
            int zoneIdsLength = 0x400; // 1k
            int[] zoneIds = new int[zoneIdsLength]; // Kuenstliches Set
            int[] edgeIdxsTmp = this.createTransformedEdgeIndex(); // 0-wertig

            for (int s = 0; s < this.nSegments; s++) {
                int edgeIdx = edgeIdxsTmp[s << 1]; // ((segmentId - 1) << 1) + 1);
                this.geomsIndex.fetch(s, coords);
                long[] lcs = coords.getCoords();
                int nLcs = lcs.length;
                // Normalfall: Nur eine Zone. Extrabehandlung = schneller.
                int nZoneIds = 0;
                int zoneId = createZoneId(latOf(lcs[0]), lonOf(lcs[0]), 2);
                int lastZoneId = zoneId;
                zoneIds[nZoneIds++] = zoneId;
                this.edgeRasterBag.add(zoneId, edgeIdx);
                // Pruefung: Sind noch andere Zonen im Spiel?
                for (int c = 1; c < nLcs; c++) {
                    zoneId = createZoneId(latOf(lcs[c]), lonOf(lcs[c]), 2);
                    if (zoneId != lastZoneId) {
                        lastZoneId = zoneId;
                        for (int z = 0; z < nZoneIds; z++) { // kuenstliches Set.contains()
                            if (zoneIds[z] == zoneId) {
                                continue; // Set.contains() == true.
                            }
                        }
                        if (nZoneIds == zoneIdsLength) { // Ueberlaufschutz
                            zoneIdsLength *= 2;
                            zoneIds = Arrays.copyOf(zoneIds, zoneIdsLength);
                        }
                        zoneIds[nZoneIds++] = zoneId;

                        this.edgeRasterBag.add(zoneId, edgeIdx);
                    }
                }

                if ((s & 0x1FFFF) == 0x1FFFF /*~130.000*/)
                    log.progress(DF(s) + " of " + DF(nSegments) + " segments rastered.");

            }
            log.progress("Sorting");
            this.edgeRasterBag.sort();

            log.info(DF(nSegments) + " segments rastered into "
                    + DF(this.edgeRasterBag.getSize()) + " zones.");

        } catch (OutOfMemoryError ome) {
            this.supports &=~ SUPPORT_RASTER_EDGE;
            this.edgeRasterBag.clear();
            log.warn("Not enough memory for Raster");
            log.debug(Log.translateError(ome, 10));
        }
    }


    /**
     * Initialisiert die Rueckwaers-Infos zum
     * Graphen. Das heisst, sortiert nach TargetId
     * und erstellt die Index-Indirektionen.
     * Der Graph wird jedoch aus Speichergruenden nicht gedoppelt.<br>
     * Im Vorwaertsgraphen (sortiert nach SourceId) wird auf die erste Edge
     * einer Kreuzung mittels {@link #getEdgeEntries()} <code>[sourceId]</code>
     * zugegriffen. Alle weiteren Edges dieser Kreuzung folgen direkt
     * im Array bzw. beziehen weitere Attribute direkt aus dieser IndexPosition.
     * <br><br>
     * Hier wird der Zugriff hingegen auf den Kopf gestellt, da ja nach
     * targetId sortiert werden muss. Auch hier existiert eine Analogie, naemlich
     * {@link #getEdgeEntriesR()}<code>[targetId]</code>. Diese Methode liefert
     * allerdings nicht direkt die Position der Edge im Originalgraphen,
     * sondern wiederum einen Index, der die Edge indiziert.
     * Dieser Indirektions-Index ist allerdings so sortiert, dass indizierte
     * Edges aufsteigend nach TargetId sortiert sind.<br><br>
     * <b>Wichtig</b><br>
     * Die Restriktions-Flags werden entgegen dem obigen Prinzip direkt
     * an die Index-Position der Original-Edge geschrieben. Hier findet
     * keine Indirektion ueber einen zusaetzlichen Index statt!
     */
    private void createReverseGraph() {
        log.info("Building Reverse Graph");

        try {
            log.progress("Reversing edges");
            int nEdges = this.edgeTargetIds.length; // inkl. 2xSentinels vorne u. hinten.
            this.edgeIdxsR = new int[nEdges];
            for (int i = 0; i < nEdges; i++) this.edgeIdxsR[i] = i;

            // Edges nach TargetId sortieren
            log.progress("Sorting by target");
            this.qsortReverseGraph(1, nEdges -2); // Ohne Sentinel-Edges

            // EdgeEntries nach targetId rueckwaers aufbauen.
            log.progress("Building reverse entries");
            int nVertices = this.getEdgeEntries().length;
            this.edgeEntriesR = new int[nVertices];
            int lastTargetId = 0;
            for (int i = 1; i < nEdges -1; i++) {
                if (this.halted) return;
                int edgeIdx = this.edgeIdxsR[i];
                int targetId = this.edgeTargetIds[edgeIdx];
                if (targetId != lastTargetId) {
                    this.edgeEntriesR[targetId] = i;
                    lastTargetId = targetId;
                }
            }

            log.progress("Building reverse restrictions");
            this.edgeNoTurnBitsR = new byte[nEdges];
            // Kreuzungen des Graphen rueckwaerts durchlaufend untersuchen.
            for (int fromEdgeIdxIdx = 1; fromEdgeIdxIdx < nEdges -1; fromEdgeIdxIdx++) {
                if (this.halted) return;
                int fromEdgeIdx = this.edgeIdxsR[fromEdgeIdxIdx];
                int targetId = this.edgeSourceIds[fromEdgeIdx];
                int toEdgeIdxIdx = this.edgeEntriesR[targetId];
                int toEdgeIdx = this.edgeIdxsR[toEdgeIdxIdx];
                int turnBits = 0;
                int bitIdx = -1;
                while (targetId == this.edgeTargetIds[toEdgeIdx]) {
                    if (++bitIdx > 7) {
                        if (turnBits != 0) log.bugfix(
                                "More than 8 outgoing edges from Segment "
                                + Graph.toSegmentId(this.edgeIds[fromEdgeIdx]));
                        break;
                    }
                    if (this.isTurnRestricted(toEdgeIdx, fromEdgeIdx) != 0) {
                        turnBits |= 1 << bitIdx;
                    }

                    toEdgeIdx = this.edgeIdxsR[++toEdgeIdxIdx];
                }
                this.edgeNoTurnBitsR[fromEdgeIdx] = (byte) turnBits;

                if ((fromEdgeIdxIdx & 0xFFFFF) == 0xFFFFF /*1MB-1*/) {
                    log.progress(fromEdgeIdxIdx + " of "
                            + nEdges + " crossings analyzed.");
                }
            }

            log.info("Reverse Graph built - " + MB());

        } catch (OutOfMemoryError ome) {
            this.supports &=~ SUPPORT_REVERSE;
            this.edgeEntriesR = null;
            this.edgeIdxsR = null;
            this.edgeNoTurnBitsR = null;
            log.warn("Not enough memory for ReverseGraph");
            log.debug(Log.translateError(ome, 10));
        }
    }

    private void qsortReverseGraph(int low, int high) {
        int i = low, j = high;
        int pivot =  this.edgeTargetIds[this.edgeIdxsR[low + (high-low)/2]];
        while (i <= j) {
            while (this.edgeTargetIds[this.edgeIdxsR[i]] < pivot) i++;
            while (this.edgeTargetIds[this.edgeIdxsR[j]] > pivot) j--;
            if (i <= j) {
                int x = this.edgeIdxsR[i];
                this.edgeIdxsR[i] = this.edgeIdxsR[j];
                this.edgeIdxsR[j] = x;
                i++; j--;
            }
        }
        if (low < j) qsortReverseGraph(low, j);
        if (i < high) qsortReverseGraph(i, high);
    }

    // ####  Von SDGraphBuilder benoetigte Methoden #####
    public IntIntBag getEdgeRasterBag() {return this.edgeRasterBag;}

} // EOC

