/*
 * $Id: GraphCostsOverrider.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.MB;

import java.io.File;
import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;

/**
 * Bietet komfortable Moeglichkeiten, um die Kosten
 * [km] und [h] eines Graphen zur Laufzeit zu ueberschreiben
 * und abschliessend wieder herzustellen.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class GraphCostsOverrider {

    private Log log;
    private int nEdges;
    private float[] originalCosts; // gesichertes Original.
    private float[] currentCosts; // Verweis auf aktuelle Graph-Kosten. (km oder h)
    private int[] edgeIdxs;

    /**
     * Konstruktor.<br>
     * Sichert die urspruenglichen Kosten und erstellt
     * den EdgeId-To-EdgeIdx-Index.
     * @param graph {@link Graph} Ursprung !NULL.
     * @param hOrKm Ueberschreibt entweder [h] (true) oder [km] (false).
     * @param log {@link Log} !NULL
     *
     */
    public GraphCostsOverrider(Graph graph, boolean hOrKm, Log log) {
        this.log = log;
        this.nEdges = graph.getEdgeIds().length;
        this.edgeIdxs = graph.createTransformedEdgeIndex();
        this.currentCosts = hOrKm ? graph.getEdgeCostsH() : graph.getEdgeCostsKm();
        this.originalCosts = Arrays.copyOf(this.currentCosts, this.nEdges); // sichern
        log.debug("Original Costs are saved - " + MB());
    }

    /**
     * Liefert die gesicherten, urspruenglichen Kosten des Graphen.
     * @return float[] Kosten.
     */
    public float[] getOriginalCosts() {return this.originalCosts;}

    /**
     * Liefert einen Verweis auf die aktuell verwalteten Kosten
     * des Graphen - entweder h oder km.
     * @return float[] Kosten.
     * @see Graph#getEdgeCostsH()
     * @see Graph#getEdgeCostsKm()
     */
    public float[] getCurrentCosts() {return this.currentCosts;}

    /**
     * Schreibt die gesicherten Kosten des Graphen zurueck
     * und stellt so dessen Urzustand wieder her.
     */
    public void resetToOriginalCosts() {
        System.arraycopy(this.originalCosts, 0, currentCosts, 0, this.nEdges);
    }

    public void close() {
        this.resetToOriginalCosts();
        this.originalCosts = null;
        this.currentCosts = null;
    }

    /**
     * Ueberschreibt die Kosten genau einer Edge.
     * @param segmentId int Id des Segments (implizit Edge)
     * @param cost float neue Kosten.
     * @param reverse Original- oder Rueckwaers-Segment (implizit Edge)
     */
    public void overrideCost(int segmentId, float cost, boolean reverse) {
        int edgeId = ((segmentId - 1) << 1); //nicht +1, da Id1 an Index 0
        if (reverse) edgeId++;
        int edgeIdx = this.edgeIdxs[edgeId];
        this.currentCosts[edgeIdx] = cost;
    }

    /**
     * Liest ein Postgres-Binary-Export-File und modifiziert
     * die Kosten des Graphen aufgrund dieser Daten.
     * <p>
     * Die Daten muessen folgendes Spalten-Format einhalten,
     * wobei die Spaltennamen selbst nicht von Bedeutung sind.
     * Es muss lediglich sicher gestellt sein, dass die
     * IDs der Segmente mit denen des Graphen uebereinstimmen.
     * Sie brauchen jedoch weder vollstaendig noch aufsteigend
     * sortiert sein.
     * </p>
     * <ul>
     *      <li>segmentId (integer NOT NULL)</li>
     *      <li>cost (double precision NULLABLE)</li>
     *      <li>reverse_cost (double precision NULLABLE) </li>
     * </ul>
     * Hier ein Beispiel eines pgplSql-Aufrufs:<br><br>
     * <code>
     *      COPY (SELECT id, cost, reverse_cost FROM hh_2po_4pgr)
     *      TO 'c:/test.bin' WITH BINARY;
     * </code>
     * @param inFile {@link File} Postgres-Binary-File.
     */
    public void overrideCostsFromPgBinary(File inFile) {

        InStream inStream = new InStreamDisk(inFile);

        byte[] signature = new byte[11];
        inStream.readBytes(signature);
        int flagsField = inStream.readInt();
        boolean hasOIDS = (flagsField & 0x10000) != 0; // Bit 16 set
        int remainder = inStream.readInt();
        for (int i = 0; i < remainder; i++) inStream.readByte(); // Skip

        log.debug(""
                + "Signature  : " + Arrays.toString(signature) + "\n"
                + "FlagsField : " + Integer.toBinaryString(flagsField) + "\n"
                + "Has OIDS   : " + hasOIDS + "\n"
                + "Remainder  : " + remainder);

        short cols = inStream.readShort(); // 1.Satz.
        if (!(cols == 3 || cols == -1)) throw new RuntimeException(
            "Expected colums: segmentId[integer32], cost[double64], reverse_cost[double64]");

        int n = 0;
        while (cols != -1) { // -1 pgEOF
            n++;
            // OID Laut Doku auch variabel moeglich.
            // Hier Annahme: int4-length-oid + int4-value-oid.
            if (hasOIDS) inStream.readLong();

            int segmentId = 0;
            if (inStream.readInt() != -1) { // !NULL
                segmentId = inStream.readInt();
            } else {
                throw new RuntimeException("SegmentId must not be NULL");
            }

            int edgeId = ((segmentId - 1) << 1); //nicht +1, da Id1 an Index 0
            if (inStream.readInt() != -1) { // !NULL
                double cost = Double.longBitsToDouble(inStream.readLong());
                int edgeIdx = this.edgeIdxs[edgeId];
                this.currentCosts[edgeIdx] = (float) cost;
            }

            edgeId++;
            if (inStream.readInt() != -1) { // !NULL
                double cost = Double.longBitsToDouble(inStream.readLong());
                int edgeIdx = this.edgeIdxs[edgeId];
                this.currentCosts[edgeIdx] = (float) cost;
            }

            cols = inStream.readShort(); // Naechster Satz.

            if (log.isEnabled(Log.LEVEL_DEBUG)) {
                if ((n & 0x1FFFF) == 0x1FFFF) // ~130.000
                    log.progress(DF(n) + " Rows processed - " + MB());
            }
        }
        inStream.close();

        System.gc();
        log.debug(DF(n) + " Rows processed - " + MB());
    }

}
