/*
 * $Id: GraphHeader.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.misc.Utils.DF;

import java.io.File;
import java.util.Arrays;

import de.cm.osm2po.Version;
import de.cm.osm2po.converter.GraphBuilder;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.InStreamDisk;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.primitives.VarTypeDesk;
/**
 * Kapselt den Header einer osm2po-Graph-Datei.
 * @author (c) 2012/2013 - Carsten Moeller - info@osm2po
 */
public class GraphHeader implements Var<GraphHeader> {

    public final static int HEADER_SIZE_DEFAULT = 1024;

    private int compatFileType;
    private int compatVersion;
    private int headerSize;
    private String version;
    private int id;

    private boolean containsWrongWays;

    private int nSegments;
    private int nEdges;
    private int nVertices;
    private long fptrEdgeNameIndex; // fptr=FilePointer
    private long fptrGeomIndex;
    private long fptrVertexCoords;
    private long fptrVertexClazzes;
    private long fptrEdgeEntries;
    private long fptrEdges;
    private long fptrBaseEof;
    private long fptrEof;

    private byte nExtensions;
    private ExtensionEntry[] extensions;

    /**
     * Ruft Konstruktor {@link #GraphHeader(int)} mit {@value #HEADER_SIZE_DEFAULT}
     */
    public GraphHeader() {
        this(HEADER_SIZE_DEFAULT);
    }

    /**
     * Initialisiert den Header mit
     * {@link GraphBuilder#COMPAT_FILE_TYPE} und
     * {@link GraphBuilder#COMPAT_VERSION}, sowie mit
     * {@link #HEADER_SIZE_DEFAULT} und {@link Version#getVersion()}
     * @param headerSize int
     * @throws IllegalArgumentException wenn headerSize &lt; 512
     */
    public GraphHeader(int headerSize) {
        if (headerSize < 512) throw new IllegalArgumentException(
                "Header size must be at least 512 bytes");
        this.compatFileType = GraphBuilder.COMPAT_FILE_TYPE;
        this.compatVersion = GraphBuilder.COMPAT_VERSION;
        this.headerSize = headerSize;
        this.version = Version.getVersion();
    }

    /**
     * Konstruktor.
     * Liest den Header aus einem GraphFile.
     * @param graphFile {@link File}.
     * @throws Exception wenn inkompatibel oder IO-Fehler.
     */
    public GraphHeader(File graphFile) throws Exception {
        InStream inStream = new InStreamDisk(graphFile);
        this.readFromStream(inStream);
        inStream.close();
    }

    public int getCompatFileType() {return this.compatFileType;}
    public int getCompatVersion() {return this.compatVersion;}
    public String getVersion() {return this.version;}

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.compatFileType);
        outStream.writeInt(this.compatVersion);
        new VarString(this.version).writeToStream(outStream);
        outStream.writeInt(this.id);
        outStream.writeInt(this.headerSize);
        outStream.writeByte((byte) (this.containsWrongWays ? 1 : 0));
        outStream.writeInt(this.nSegments);
        outStream.writeInt(this.nEdges);
        outStream.writeInt(this.nVertices);
        outStream.writeLong(this.fptrGeomIndex);
        outStream.writeLong(this.fptrEdgeNameIndex);
        outStream.writeLong(this.fptrVertexCoords);
        outStream.writeLong(this.fptrVertexClazzes);
        outStream.writeLong(this.fptrEdgeEntries);
        outStream.writeLong(this.fptrEdges);
        outStream.writeLong(this.fptrBaseEof);
        outStream.writeLong(this.fptrEof);

        outStream.writeByte(this.nExtensions);
        for (byte i = 0; i < this.nExtensions; i++) {
            int nBytesFree =  this.getHeaderSize() - (int) outStream.getWriteIndex();
            if (ExtensionEntry.SIZE > nBytesFree) {
                throw new IllegalStateException("Too many Extensions in GraphHeader");
            }
            this.extensions[i].writeToStream(outStream);
        }

        // Remainder
        while (outStream.getWriteIndex() < this.getHeaderSize())
            outStream.writeByte((byte) 0);
    }

    @Override
    public GraphHeader readFromStream(InStream inStream) {
        long offset = inStream.getReadIndex();

        this.compatFileType = inStream.readInt();
        this.compatVersion = inStream.readInt();

        // Header Compatibility
        if (this.compatFileType != GraphBuilder.COMPAT_FILE_TYPE)
            throw new RuntimeException("No osm2po-GraphFile");
        if (this.compatVersion != GraphBuilder.COMPAT_VERSION)
            throw new RuntimeException("Incompatible GraphFile. Version x.x.x."
                    + GraphBuilder.COMPAT_VERSION + " required"
                    + "\nPlease rebuild the graph with the current version");

        this.version = new VarString().readFromStream(inStream).toString();
        this.id = inStream.readInt();
        this.headerSize = inStream.readInt();
        this.containsWrongWays = inStream.readByte() != 0;
        this.nSegments = inStream.readInt();
        this.nEdges = inStream.readInt();
        this.nVertices = inStream.readInt();
        this.fptrGeomIndex = inStream.readLong();
        this.fptrEdgeNameIndex = inStream.readLong();
        this.fptrVertexCoords = inStream.readLong();
        this.fptrVertexClazzes = inStream.readLong();
        this.fptrEdgeEntries = inStream.readLong();
        this.fptrEdges = inStream.readLong();
        this.fptrBaseEof = inStream.readLong();
        this.fptrEof = inStream.readLong();

        this.nExtensions = inStream.readByte();
        if (this.nExtensions > 0) this.extensions =
            new ExtensionEntry[this.nExtensions];
        for (byte i = 0; i < this.nExtensions; i++) this.extensions[i] =
            new ExtensionEntry().readFromStream(inStream);

        long nRemainingBytes = this.headerSize - (inStream.getReadIndex() - offset);
        if (nRemainingBytes < 0) throw new RuntimeException(
                "Header too small! Remaining bytes: " + nRemainingBytes);
        // Empty Remainder
        for (long i = 0; i < nRemainingBytes; i++) inStream.readByte();

        return this;
    }

    public int getHeaderSize() {return this.headerSize;}

    public int getId() {return this.id;}
    public void setId(int id) {this.id = id;}

    public boolean isContainsWrongWays() {return this.containsWrongWays;}
    public void setContainsWrongWays(boolean containsWrongWays) {this.containsWrongWays = containsWrongWays;}

    public int getNumberOfSegments() {return this.nSegments;}
    public void setNumberOfSegments(int segments) {this.nSegments = segments;}
    public int getNumberOfEdges() {return this.nEdges;}
    public void setNumberOfEdges(int edges) {this.nEdges = edges;}
    public int getNumberOfVertices() {return this.nVertices;}
    public void setNumberOfVertices(int vertices) {this.nVertices = vertices;}

    public long getFptrEdgeNameIndex() {
        return this.fptrEdgeNameIndex;}
    public void setFptrEdgeNameIndex(long fptrEdgeNameIndex) {
        this.fptrEdgeNameIndex = fptrEdgeNameIndex;}
    public long getFptrGeomIndex() {return this.fptrGeomIndex;}
    public void setFptrGeomIndex(long fptrGeomIndex) {
        this.fptrGeomIndex = fptrGeomIndex;}
    public long getFptrVertexCoords() {return this.fptrVertexCoords;}
    public void setFptrVertexCoords(long fptrVertexCoords) {
        this.fptrVertexCoords = fptrVertexCoords;}
    public long getFptrVertexClazzes() {return this.fptrVertexClazzes;}
    public void setFptrVertexClazzes(long fptrVertexClazzes) {
        this.fptrVertexClazzes = fptrVertexClazzes;}
    public long getFptrEdgeEntries() {return this.fptrEdgeEntries;}
    public void setFptrEdgeEntries(long fptrEdgeEntries) {
        this.fptrEdgeEntries = fptrEdgeEntries;}
    public long getFptrEdges() {return this.fptrEdges;}
    public void setFptrEdges(long fptrEdges) {this.fptrEdges = fptrEdges;}
    public long getFptrBaseEof() {return this.fptrBaseEof;}
    public void setFptrBaseEof(long fptrBaseEof) {this.fptrBaseEof = fptrBaseEof;}
    public long getFptrEof() {return this.fptrEof;}
    public void setFptrEof(long fptrEof) {this.fptrEof = fptrEof;}

    public boolean isExtensionExist(Class<? extends Var<?>> varClass, byte id) {
        byte varType = VarTypeDesk.typeIdOf(varClass);
        for (int i = 0; i < this.nExtensions; i++) {
            if (this.extensions[i].varType == varType && this.extensions[i].id == id)
                return true;
        }
        return false;
    }

    public void addExtension(Class<? extends Var<?>> varClass, byte id, long offset, long size) {
        if (isExtensionExist(varClass, id))
            throw new IllegalStateException(
                    "Extension " + varClass.getSimpleName() + "#" + id + " already exists");

        if (null == this.extensions) this.extensions = new ExtensionEntry[1];

        if (this.nExtensions >= this.extensions.length)
            this.extensions = Arrays.copyOf(this.extensions, this.extensions.length * 2);

        byte varType = VarTypeDesk.typeIdOf(varClass);
        this.extensions[this.nExtensions++] =
            new ExtensionEntry(varType, id, offset, size);
    }

    /**
     * Liefert die Extensions.
     * @return Array of {@link ExtensionEntry} NOT NULL.
     */
    public ExtensionEntry[] getExensions() {
        if (null == this.extensions) return new ExtensionEntry[0];
        else return Arrays.copyOf(this.extensions, nExtensions);
    }

    public static class ExtensionEntry implements Var<ExtensionEntry> {

        public final static int SIZE = 1 + 1 + 8 + 8; // id, varType, ...
        private byte id;
        private byte varType;
        private long offset;
        private long size;

        private ExtensionEntry() {}

        private ExtensionEntry(byte varType, byte id, long offset, long size) {
            this.id = id;
            this.varType = varType;
            this.offset = offset;
            this.size = size;
        }

        public byte getId() {return this.id;}
        public byte getVarType() {return this.varType;}
        public long getOffset() {return this.offset;}
        public long getSize() {return this.size;}

        @Override
        public void writeToStream(OutStream outStream) {
            outStream.writeByte(this.id);
            outStream.writeByte(this.varType);
            outStream.writeLong(this.offset);
            outStream.writeLong(this.size);
        }

        @Override
        public ExtensionEntry readFromStream(InStream inStream) {
            this.id = inStream.readByte();
            this.varType = inStream.readByte();
            this.offset = inStream.readLong();
            this.size = inStream.readLong();
            return this;
        }

        @Override
        public String toString() {
            return "varType=" + varType + ", id=" + id
                    + ", size=" + DF(size) + ", offset=" + DF(offset);
        }

    }
}
