/*
 * $Id: MultilevelVertexRouter.java 5466 2021-12-02 20:45:54Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.postp.MlgExtensionBuilder.ID_MASK_BASE;
import static de.cm.osm2po.postp.MlgExtensionBuilder.ID_MASK_FIND_SHORTEST;
import static de.cm.osm2po.postp.MlgExtensionBuilder.ID_MASK_IGNORE_ONEWAYS;
import static de.cm.osm2po.postp.MlgExtensionBuilder.ID_MASK_IGNORE_RESTRICTIONS;

import java.util.Arrays;

import de.cm.osm2po.extensions.MlgExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.postp.MlgExtensionBuilder;
/**
 * Dijkstra-Implemetation eines Routings auf einem statischen Graphen
 * mit AStar- und MultiLevel-Support.
 * <br><br>
 * Dieser Router beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * <li> {@link RouterParameters#getHeuristicFactor()} </li>
 * <li> {@link RouterParameters#getLocalityFactor()} </li>
 * </ul>
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see MlgExtension
 * @see MlgExtensionBuilder
 * @see Graph
 * @see OverlayVertexRouter
 * @see VisitedVerticesAware
 */
public class MultilevelVertexRouter extends AbstractRouter<Graph>
implements VisitedVerticesAware {

    public MultilevelVertexRouter(Graph graph, Log log) {
        super(graph, log);
    }

    private final static int PQ_INITIAL_CAPACITY = 10000; // reicht fuer Europa und mehr.
    private int pqSize; // Aktuelle ArrayGroesse der PQ
    private int pqCapacity; // Maximale ArrayGroesse der PQ
    private int[] pqEdgeIdxs;
    private double[] pqCostsR; // Reelle Weg- oder Zeit-Kosten.
    private double[] pqCostsH; // Heuristische Gesamtkosten.

    // *1) White-Gray-Black Array. Index ist die TargetId (1-wertig).
    //     Graue Knoten setzen positive PQ-Array-Positionen (pq arbeitet 1-wertig).
    //     Schwarze Knoten setzen EdgeIdx |= 0x80000000. Fuer EdgeIdx == 0 liefert
    //     diese Veroderung Integer.MinValue, EdgeId == 1 also MinValue + 1, usw.
    //     Als Sentinel fuer den ersten Knoten (ohne Vorgaengerweg) wird -1 verwendet,
    //     also 32 gesetzte Bits, was einen maximalen Abstand ergibt.
    //     Dies ist die Abbruchbedingung bei der Erstellung in Rueckwaertsrichtung.
    //     Es werden also niemals 0-Werte gesetzt. Daher ist 0 immer initial.
    private int[] wgbs; // siehe *1)

    // Aktuelle Edge (nach pqRemove)
    private int pqrEdgeIdx;
    private double pqrCostR;

    // Kosten werden entweder mit km oder h belegt.
    private float[] edgeCosts;

    // Members fuer Heuristiken
    private float[] lats, lons; // Alle Koordinaten.
    private float targetLat, targetLon; // Zielkoordinate.
    private double heuristicFactor = 0d; // Bewertungsfaktor 0 Dijkstra, 1 AStar

    // Zus. Members fuer Multilevel-Routing
    private byte[] edgeLevels;
    private double sourceLat, sourceLon; // Startkoordinate.
    private double level1CellSize;
    private double localityFactor = 0d;

    // Modi
    private boolean findShortestPath;
    private boolean ignoreRestrictions;
    private boolean ignoreOneWays;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        if (constraints.targetIds.length > 1) throw new IllegalArgumentException(
                "OneToMany-Routing with more than One Target not supported");

        this.initParams(params);
        this.initAStarAndLevels(sourceId, constraints.targetIds[0]); // Unkritisch, wenn Targets > 1
        this.initPqAndWgb();
        this.settleSourceVertex(sourceId, constraints.maxCost);
        return this.traverseInternal(constraints.targetIds[0], constraints.maxCost);
    }

    private void initPqAndWgb() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Maximale ArrayGroesse der PQ
        this.pqEdgeIdxs = new int[pqCapacity];
        this.pqCostsR = new double[pqCapacity]; // Reelle Weg- oder Zeit-Kosten.
        this.pqCostsH = new double[pqCapacity]; // Heuristische Gesamtkosten.

        int nVertices = this.graph.getEdgeEntries().length;
        this.wgbs = new int[nVertices];

        // Die DeadEdge (alle Member = 0) verweist auf diesen Vertex.
        // Daher markieren wir ihn als bereits besucht.
        this.wgbs[0] = Integer.MAX_VALUE | 0x80000000;
    }

    private void initParams(RouterParameters params) {

        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();
        this.heuristicFactor = params.getHeuristicFactor();
        this.localityFactor = params.getLocalityFactor();

        // Kosten entweder fuer Shortest oder Fastest
        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
            // Wenn Fastest, dann muss die Toleranz vergroessert werden,
            // damit die AStar-Bedingung eingehalten wird.
            // Bedingung: Len(A) + Heuristic(A) <= Len(A) + Len(B) + Heuristic(B)
            // Hier 110 kmh: D.h. zwei aufeinanderfolgende Wege duerfen
            // in ihren Geschwindigkeiten maximal um 110 kmh differieren.
            this.heuristicFactor /= 110;
        }
    }

    private void initAStarAndLevels(int sourceId, int targetId) {
        if (!this.graph.isSupported(Graph.SUPPORT_LATLON)) {
            throw new IllegalStateException("Graph does not support LatLon");
        }

        byte extensionId = ID_MASK_BASE; // 0
        if (this.findShortestPath) extensionId |= ID_MASK_FIND_SHORTEST;
        if (this.ignoreOneWays) extensionId |= ID_MASK_IGNORE_ONEWAYS;
        if (this.ignoreRestrictions) extensionId |= ID_MASK_IGNORE_RESTRICTIONS;

        String idc = MlgExtension.class.getSimpleName() + "#" + extensionId;
        MlgExtension mlx = this.graph.getExtension(MlgExtension.class, extensionId);
        if (null == mlx) {
            log.warn("Graph does not support Extension " + idc + ". Setting all Levels to Zero");
            this.edgeLevels = new byte[graph.getEdgeIds().length];
        } else {
        	this.edgeLevels = mlx.getBytes();
        	this.level1CellSize =  mlx.getMaxCellSize() / Math.pow(2, mlx.getNumberOfLevels() -1);
            log.info("MultiLevelExtension " + idc + " loaded");
            log.debug(mlx.toString());
        }

        this.lats = this.graph.getLats();
        this.lons = this.graph.getLons();

        this.sourceLat = this.lats[sourceId];
        this.sourceLon = this.lons[sourceId];
        this.targetLat = this.lats[targetId];
        this.targetLon = this.lons[targetId];
    }

    private void settleSourceVertex(int fromSourceId, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();

        // Kein Vorgaengerweg. (all Bits Set = -1)
        this.wgbs[fromSourceId] = Integer.MAX_VALUE | 0x80000000;

        int edgeIdx = edgeEntries[fromSourceId];
        while (edgeSourceIds[edgeIdx] == fromSourceId) {
            int targetId = edgeTargetIds[edgeIdx];

            // Schon hier pruefen, da Start auch eine
            // Loop (source=target) sein kann.
            if (this.wgbs[targetId] >= 0) {
                boolean isWrongWay = (edgeIds[edgeIdx] & 0x80000001) == 0x80000000;
                if (!isWrongWay || this.ignoreOneWays) {
                    double cost = this.edgeCosts[edgeIdx];
                    if (cost < maxCost)
                        this.pqInsertOrUpdate(edgeIdx, cost, targetId);
                }
            }
            edgeIdx++;
        }
    }

    private boolean traverseInternal(int toTargetId, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        byte[] edgeNoTurnBits = this.graph.getEdgeNoTurnBits();

        while (this.pqSize > 0) {
            this.pqRemove();
            int targetId = edgeTargetIds[this.pqrEdgeIdx];

            this.wgbs[targetId] = this.pqrEdgeIdx | 0x80000000;

            // Jepp!!! Hier ist der kuerzeste Weg.
            if (targetId == toTargetId) return true;

            // Locality-Filter
            int fromLevel = this.edgeLevels[this.pqrEdgeIdx];
            int maxLevel = this.localityLevel(targetId);
            if (maxLevel < fromLevel) fromLevel = maxLevel;

            // Anschlusssegmente suchen.
            int sourceId = targetId; // Rollenwechsel
            int edgeIdx = edgeEntries[sourceId];

            int noTurnBits = edgeNoTurnBits[this.pqrEdgeIdx];
            int noTurnBit = 1;

            do { // Anschlusssegmente durchlaufen.
                targetId = edgeTargetIds[edgeIdx];

                int toLevel = this.edgeLevels[edgeIdx];
                if (toLevel >= fromLevel) {

                    // Jetzt pruefen, ob TargetVertex noch in der Pipeline ist.
                    if (this.wgbs[targetId] >= 0) {

                        boolean mayTurn = true;
                        if (noTurnBits != 0 && !this.ignoreRestrictions) // Abbiegevorschriften
                            mayTurn &= ((noTurnBits & noTurnBit) == 0);
                        if (!this.ignoreOneWays) // Einbahnstrassen beachten
                            mayTurn &= (!((edgeIds[edgeIdx] & 0x80000001) == 0x80000000));

                        if (mayTurn) {
                            // Reelle Kosten bis hier
                            double cost = this.pqrCostR + this.edgeCosts[edgeIdx];
                            if (cost < maxCost)
                                this.pqInsertOrUpdate(edgeIdx, cost, targetId);
                        }
                    }
                }
                edgeIdx++;
                noTurnBit <<= 1;
            } while (sourceId == edgeSourceIds[edgeIdx]);
        }
        return false;
    }

    private int localityLevel(int vertexId) {
        // Sieht viel aus, ist aber schneller als Math.abs().
        double lat = this.lats[vertexId];
        double lon = this.lons[vertexId];

        double deltaLatSource = this.sourceLat - lat;
        double deltaLonSource = this.sourceLon - lon;
        double deltaLatTarget = this.targetLat - lat;
        double deltaLonTarget = this.targetLon - lon;

        double deltaLatSourceAbs = deltaLatSource <= 0d ? 0d - deltaLatSource : deltaLatSource;
        double deltaLonSourceAbs = deltaLonSource <= 0d ? 0d - deltaLonSource : deltaLonSource;
        double deltaLatTargetAbs = deltaLatTarget <= 0d ? 0d - deltaLatTarget : deltaLatTarget;
        double deltaLonTargetAbs = deltaLonTarget <= 0d ? 0d - deltaLonTarget : deltaLonTarget;

        // Manhattan-Distance
        double distSource = deltaLatSourceAbs + deltaLonSourceAbs;
        double distTarget = deltaLatTargetAbs + deltaLonTargetAbs;

        double dist = distSource < distTarget ? distSource : distTarget;

        if (this.localityFactor != 0d) dist *= this.localityFactor;

        // getExponent() aufgeloest
        final long EXP_BIT_MASK = 0x7FF0000000000000L;
        final int EXP_BIAS = 0x3FF;
        final int SIGNIFICAND_WIDTH_MINUS_ONE = 53 - 1;
        long RAWBITS = Double.doubleToRawLongBits(dist / this.level1CellSize); // native
        int level = 1 + (int) (((RAWBITS & EXP_BIT_MASK) >> SIGNIFICAND_WIDTH_MINUS_ONE) - EXP_BIAS);

        return level;
    }

    @Override
    public final int[] makePath(int toTargetId) {
        if (!this.isVisited(toTargetId)) return null;

        int[] edgeSourceIds = this.graph.getEdgeSourceIds();

        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        int targetId = toTargetId;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY);
            int edgeIdx = this.wgbs[targetId] & 0x7FFFFFFF; // Bit 32 (Visited-Black-Flag).
            if (edgeIdx == Integer.MAX_VALUE) break; // -1 hat keinen Vorgaengerweg. -1 wg. so.
            int sourceId = edgeSourceIds[edgeIdx];
            pathIds[n++] = edgeIdx;
            targetId = sourceId;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        // Reverse
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    @Override
    public final int[] getVisited() {
        int len = this.wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) { // wgbs ist 1-wertig.
            int edgeIdx = this.wgbs[vertexId];
            if (edgeIdx < 0) { // Kante zum ZielVertex wurde besucht.
                if (buffer.length < vertexId) // Ueberlauf korrigieren
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    @Override
    public final boolean isVisited(int vertexId) {
        if (vertexId < 1 || vertexId > this.wgbs.length -1) return false;
        return this.wgbs[vertexId] < 0;
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqEdgeIdxs = Arrays.copyOf(this.pqEdgeIdxs, this.pqCapacity);
        this.pqCostsR = Arrays.copyOf(this.pqCostsR, this.pqCapacity);
        this.pqCostsH = Arrays.copyOf(this.pqCostsH, this.pqCapacity);
        log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsertOrUpdate(int edgeIdx, double costR, int targetId) {
        int pqPos = this.wgbs[targetId];

        if (0 == pqPos) { // neuer Ziel-Knoten
            // AStar
            double h = 0d;
            if (this.heuristicFactor != 0) {
                h = GeoUtils.calcDistanceEuklidCoarse(this.lats[targetId], this.lons[targetId],
                        this.targetLat, this.targetLon);
                h *= this.heuristicFactor;
            }

            pqPos = ++this.pqSize;
            if (this.pqSize >= this.pqCapacity) this.pqGrow();

            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + h;
            this.wgbs[targetId] = pqPos;

        } else if (this.pqCostsR[pqPos] > costR) { // kuerzeren Weg gefunden
            double costH = this.pqCostsH[pqPos] - this.pqCostsR[pqPos];
            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + costH;

        } else {
            return;
        }

        this.pqUpheap(pqPos);
    }

    private void pqRemove() {
        this.pqrEdgeIdx = this.pqEdgeIdxs[1];
        this.pqrCostR = this.pqCostsR[1];
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap(1);
    }

    private void pqSwap(int pos1, int pos2) {
        // TODO change getEdgeTargetIds() into a Member-Variable
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();

        // PQ-Swap
        int edgeIdx1 = this.pqEdgeIdxs[pos1];
        double costR1 = this.pqCostsR[pos1];
        double costH1 = this.pqCostsH[pos1];
        int edgeIdx2 = this.pqEdgeIdxs[pos2];
        double costR2 = this.pqCostsR[pos2];
        double costH2 = this.pqCostsH[pos2];

        this.pqEdgeIdxs[pos1] = edgeIdx2;
        this.pqCostsR[pos1] = costR2;
        this.pqCostsH[pos1] = costH2;
        this.pqEdgeIdxs[pos2] = edgeIdx1;
        this.pqCostsR[pos2] = costR1;
        this.pqCostsH[pos2] = costH1;

        // WGBS-Swap
        int targetId1 = edgeTargetIds[edgeIdx1];
        int targetId2 = edgeTargetIds[edgeIdx2];

        this.wgbs[targetId1] = pos2;
        this.wgbs[targetId2] = pos1;
    }

    private void pqUpheap(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[parentPos] < this.pqCostsH[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos) {
        int childPos = pos << 1; // * 2 = linkes Kind
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqCostsH[childRightPos] < this.pqCostsH[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[pos] < this.pqCostsH[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
