/*
 * $Id: OverlayEdge.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.misc.Utils.DF2;
import static java.lang.Float.floatToIntBits;
import static java.lang.Float.intBitsToFloat;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;

/**
 * Eine Verbindung zwischen zwei Vertices
 * innerhalb eine {@link OverlayGraph}-Instanz.
 * <br><br><b>Wichtig:</b><br>
 * Als Source und Target werden lediglich IDs 
 * verwaltet. Diese duerfen jedoch sowohl auf echte
 * Vertices aus dem Original-Graphen als auch auf
 * virtuelle aus dem Overlay verweisen.
 * 
 * @author (c) 2014 - Carsten Moeller - info@osm2po.de
 */
public class OverlayEdge implements Var<OverlayEdge> {
    private int idx; // Virtuelle idx im OverlayGraph
    private int edgeIdx; // Referenz auf Original-Edge
    private int sourceId; // Virtuelle oder echte Id
    private int targetId; // Virtuelle oder echte Id
    private float km;
    private float h;
    private LatLons geometry;

    protected void setIdx(int idx) {this.idx = idx;}

    public int getIdx() {return this.idx;}
    public int getEdgeIdx() {return this.edgeIdx;}
    public int getSourceId() {return this.sourceId;}
    public int getTargetId() {return this.targetId;}
    public float getKm() {return this.km;}
    public float getH() {return this.h;}
    public LatLons getGeometry() {return this.geometry;}
    
    public OverlayEdge(){} // Public wegen Var.readFromStream()

    /**
     * Erstellt eine kuenstliche Edge. Ruft intern<br>
     * {@link OverlayEdge#OverlayEdge(int, int, int, float, float, LatLons)}
     * <br>mit edgeIdx = 0.
     * @param sourceId int Id des StartVertex
     * @param targetId int Id des ZielVertex
     * @param km Laenge in Kilometern
     * @param h Fahrzeit in Stunden
     * @param geometry {@link LatLons}
     */
    public OverlayEdge(
            int sourceId, int targetId, float km, float h, LatLons geometry) {
        this(0, sourceId, targetId, km, h, geometry);
    }

    /**
     * Erstellt eine kuenstlich OverlayEdge mit einer Referenz
     * auf eine echte Edge als Platzhalter fuer das Routing mit nicht
     * {@link OverlayGraph}-faehigen Routern.
     * @param edgeIdx int Index der Edge im Original-Graphen.
     */
    protected OverlayEdge(int edgeIdx) {
        this.edgeIdx = edgeIdx;
    }
    
    /**
     * Erstellt eine kuenstliche Edge, die eine Original-Edge aus dem
     * statischen Graphen ganz oder teilweise ueberlagert.
     * @param edgeIdx int Index der Edge im Original-Graphen, oder 0 wenn rein virtuell.
     * @param sourceId int Id des StartVertex
     * @param targetId int Id des ZielVertex
     * @param km Laenge in Kilometern
     * @param h Fahrzeit in Stunden
     * @param geometry {@link LatLons}
     */
    protected OverlayEdge(int edgeIdx,
            int sourceId, int targetId, float km, float h, LatLons geometry) {
        this.idx = 0;
        this.edgeIdx = edgeIdx;
        this.sourceId = sourceId;
        this.targetId = targetId;
        this.km = km;
        this.h = h;
        this.geometry = geometry;
    }
    
    @Override
    public String toString() {
        return "Idx:" + this.idx 
            + " EdgeIdx:" + this.edgeIdx
            + " SourceId:" + this.sourceId
            + " TargetId:" + this.targetId
            + " km:" + DF2(this.km)
            + " h:" + DF2(this.h);
    }
    
    /**
     * Erstellt die Rueckwaerts-Variante.
     * @param graph {@link Graph} NULLABLE (dann rein virtuell).
     * @return {@link OverlayEdge} oder NULL, wenn nicht erstellbar.
     */
    public OverlayEdge copyReversed(Graph graph) {
        OverlayEdge s = new OverlayEdge();
        s.edgeIdx = this.edgeIdx;
        s.sourceId = this.targetId;
        s.targetId = this.sourceId;
        s.km = this.km;
        s.h = this.h;
        s.geometry = this.geometry.copy().reverse();
        
        if (graph != null && s.edgeIdx != 0) {
            s.edgeIdx = graph.findReverseEdge(this.edgeIdx);
            if (0 == s.edgeIdx) return null;
        }

        return s;
    }

    @Override
    public OverlayEdge readFromStream(InStream inStream) {
        this.idx = inStream.readInt();
        this.edgeIdx = inStream.readInt();
        this.sourceId = inStream.readInt();
        this.targetId = inStream.readInt();
        this.km = intBitsToFloat(inStream.readInt());
        this.h = intBitsToFloat(inStream.readInt());
        LatLons lls = new LatLons().readFromStream(inStream);
        if (0 == lls.getSize()) lls = null;
        this.geometry = lls;
        return this;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeInt(this.idx);
        outStream.writeInt(this.edgeIdx);
        outStream.writeInt(this.sourceId);
        outStream.writeInt(this.targetId);
        outStream.writeInt(floatToIntBits(this.km));
        outStream.writeInt(floatToIntBits(this.h));
        LatLons lls = this.geometry;
        if (null == lls) lls = new LatLons();
        lls.writeToStream(outStream);
    }

}
