/*
 * $Id: OverlayEdgeRouter.java 5721 2023-01-02 18:54:35Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoUtils;

/**
 * Edge-Based-Implemetation eines OneToOne-Routings auf einem
 * ueberlagerten Graphen mit AStar-Unterstuetzung.<br>
 * Die interne PriorityQueue arbeitet dabei auf Kanten-Basis, so dass
 * jede Kante 1x besucht werden kann. Das heisst, ein Segment wird maximal
 * 2x besucht, 1x vorwaerts und 1x in Rueckrichtung, was genauere Resultate
 * fuer Abbiegevorschriften erzeugt.
 * <br><br>
 * Diese Implementation beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * <li> {@link RouterParameters#getHeuristicFactor()} </li>
 * <li> {@link RouterParameters#getLocalityFactor()} </li>
 * </ul>
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see OverlayGraph
 * @see OverlayVertexRouter
 * @see VertexRouter
 * @see VisitedVerticesAware
 */
public class OverlayEdgeRouter  extends AbstractRouter<OverlayGraph>
implements VisitedVerticesAware {

    public OverlayEdgeRouter(OverlayGraph graph, Log log) {
        super(graph, log);
        this.firstVirtualEdgeIdx = graph.getFirstVirtualEdgeIdx();
        this.firstVirtualVertexId = graph.getFirstVirtualVertexId();
    }

    private final static int PQ_INITIAL_CAPACITY = 100000;
    private int pqSize; // Aktuelle ArrayGroesse der PQ
    private int pqCapacity; // Maximale ArrayGroesse der PQ
    private int[] pqIdxs;
    private double[] pqCostsR; // Reelle Weg- oder Zeit-Kosten.
    private double[] pqCostsH; // Heuristische Weg- oder Zeit-Kosten.

    private int[] ses; // settled Edges, gleichzeitig Hash fuer TargetEdges.
    private int targetEdgeIdx; // letzte Edge zum Ziel, -1 wenn nicht gefunden.

    // Aktuelle Edge (nach pqRemove)
    private int pqrEdgeIdx;
    private double pqrCostsR; // Kosten bis hierher real

    private int firstVirtualVertexId;
    private int firstVirtualEdgeIdx;

    // Kosten werden entweder mit km oder h belegt.
    private float[] edgeCosts;

    // Members fuer Heuristiken
    private float[] lats, lons; // Alle Koordinaten.
    private float targetLat, targetLon; // Zielkoordinate.
    private double heuristicFactor = 0d; // Bewertungsfaktor 0 Dijkstra, 1 AStar

    // Modi
    private boolean findShortestPath;
    private boolean ignoreRestrictions;
    private boolean ignoreOneWays;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        if (constraints.targetIds.length > 1) throw new IllegalArgumentException(
                "OneToMany-Routing with more than One Target not supported");

        this.initParams(params);
        this.initAstar(constraints.targetIds[0]); // Unkritisch, wenn Targets > 1
        this.initPqAndSes();

        if (sourceId < this.firstVirtualVertexId) {
            this.expandStartVertex(sourceId,  this.firstVirtualEdgeIdx -1 /*DeadEdge*/, constraints.maxCost);
        } else {
            this.expandVirtualVertex(sourceId, this.firstVirtualEdgeIdx -1, 0d, constraints.maxCost);
        }

        this.targetEdgeIdx = this.traverseInternal(constraints.targetIds[0], constraints.maxCost);
        return this.targetEdgeIdx >= 0;
    }

    private void initParams(RouterParameters params) {
        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();
        this.heuristicFactor = params.getHeuristicFactor();

        // Kosten entweder fuer Shortest oder Fastest
        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getGraph().getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getGraph().getEdgeCostsH();
            // Wenn Fastest, dann muss die Toleranz vergroessert werden,
            // damit die AStar-Bedingung eingehalten wird.
            // Bedingung: Len(A) + Heuristic(A) <= Len(A) + Len(B) + Heuristic(B)
            // Hier 110 kmh: D.h. zwei aufeinanderfolgende Wege duerfen
            // in ihren Geschwindigkeiten maximal um 110 kmh differieren.
            this.heuristicFactor /= 110;
        }
    }

    private void initAstar(int targetId) {
        Graph staticGraph = graph.getGraph();
        if (!staticGraph.isSupported(Graph.SUPPORT_LATLON)) {
            throw new IllegalStateException("Graph does not support LatLon");
        }

        this.lats = staticGraph.getLats();
        this.lons = staticGraph.getLons();
        if (targetId < this.firstVirtualVertexId) {
            this.targetLat = this.lats[targetId];
            this.targetLon = this.lons[targetId];
        } else {
            OverlayNode nd = this.graph.getOverlayNode(targetId);
            this.targetLat = nd.getLat();
            this.targetLon = nd.getLon();
        }
    }

    private void initPqAndSes() {
        // Init PQ
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Maximale ArrayGroesse der PQ
        this.pqIdxs = new int[pqCapacity];
        this.pqCostsR = new double[pqCapacity]; // Reelle Weg- oder Zeit-Kosten.
        this.pqCostsH = new double[pqCapacity]; // Heuristische Weg- oder Zeit-Kosten.

        // Init White-Black-List Settled Edges
        this.ses = new int[this.graph.getLastVirtualEdgeIdx() +1];
    }

    private void expandStartVertex(int vertexId, int fromEdgeIdx, double maxCost) {
        Graph staticGraph = this.graph.getGraph();
        int[] edgeIds = staticGraph.getEdgeIds();
        int[] edgeTargetIds = staticGraph.getEdgeTargetIds();
        int[] edgeIdxs;

        edgeIdxs = staticGraph.findOutgoingEdges(vertexId);
        for (int i = 0; i < edgeIdxs.length; i++) {
            int edgeIdx = edgeIdxs[i];
            int pqIdx = edgeIdx;
            int targetId = edgeTargetIds[edgeIdx];
            double costR = this.edgeCosts[edgeIdx];

            OverlayEdge s = this.graph.getEdgeRamp(edgeIdx);
            if (s != null) {
                pqIdx = s.getIdx();
                targetId = s.getTargetId();
                costR = this.findShortestPath ? s.getKm() : s.getH();
            }

            boolean isWrongWay = Graph.isWrongWay(edgeIds[edgeIdx]);
            if (!isWrongWay || this.ignoreOneWays) {
                double costH = costR + this.calcVertexCost(targetId);
                if (costR < maxCost) {
                    this.pqInsert(pqIdx, costR, costH);
                    this.ses[pqIdx] = fromEdgeIdx;
                }
            }
            edgeIdx++;
        }
    }

    private void expandVirtualVertex(int virtualVertexId, int fromEdgeIdx, double fromCost, double maxCost) {
        Graph staticGraph = this.graph.getGraph();
        OverlayEdge fromOverlayEdge = this.graph.getOverlayEdge(fromEdgeIdx);
        int uturnId = fromOverlayEdge != null ? fromOverlayEdge.getSourceId() : 0;
        OverlayEdge[] outgoingOverlayEdges = this.graph.getOutgoingOverlayEdges(virtualVertexId);
        for (int i = 0; i < outgoingOverlayEdges.length; i++) {
            OverlayEdge s = outgoingOverlayEdges[i];
            if (this.ses[s.getIdx()] == 0 && s.getTargetId() != uturnId) {
                boolean mayTurn = true;
                if (!this.ignoreOneWays)
                    mayTurn = (!((staticGraph.getEdgeIds()[s.getEdgeIdx()] & 0x80000001) == 0x80000000));

                if (mayTurn) {
                    double costR = fromCost +(this.findShortestPath ? s.getKm() : s.getH());
                    double costH = costR + this.calcVertexCost(s.getTargetId());
                    if (costR < maxCost) {
                        this.pqInsert(s.getIdx(), costR, costH);
                        this.ses[s.getIdx()] = fromEdgeIdx;
                    }
                }
            }
        }
    }

    private int traverseInternal(int toTargetId, double maxCost) {
        Graph staticGraph = this.graph.getGraph();

        int[] edgeEntries = staticGraph.getEdgeEntries();
        int[] edgeSourceIds = staticGraph.getEdgeSourceIds();
        int[] edgeTargetIds = staticGraph.getEdgeTargetIds();
        int[] edgeIds = staticGraph.getEdgeIds();
        byte[] edgeNoTurnBits = staticGraph.getEdgeNoTurnBits();

        while (this.pqSize > 0) {
            this.pqRemove();
            int fromEdgeIdx = this.pqrEdgeIdx;

            int targetId = 0;
            int uturnId = 0;

            if (fromEdgeIdx < this.firstVirtualEdgeIdx) {
                targetId = edgeTargetIds[fromEdgeIdx];
                uturnId = edgeSourceIds[fromEdgeIdx];

            } else {
                OverlayEdge s = this.graph.getOverlayEdge(fromEdgeIdx);
                targetId = s.getTargetId();
                uturnId = s.getSourceId();
            }

            // Jepp!!! Hier ist der kuerzeste Weg.
            if (toTargetId == targetId) return fromEdgeIdx;

            // Anschlusssegmente suchen.
            int sourceId = targetId; // Rollenwechsel
            if (sourceId >= this.firstVirtualVertexId) { // Dynamic Graph
                this.expandVirtualVertex(sourceId, fromEdgeIdx, this.pqrCostsR, maxCost);
            } else {
                int toEdgeIdx = edgeEntries[sourceId];
                int noTurnBits = 0;
                int noTurnBit = 1;
                if (fromEdgeIdx < this.firstVirtualEdgeIdx) {
                    noTurnBits = edgeNoTurnBits[this.pqrEdgeIdx];
                } else { // Dynamic Graph
                    OverlayEdge s = this.graph.getOverlayEdge(this.pqrEdgeIdx);
                    noTurnBits = edgeNoTurnBits[s.getEdgeIdx()];
                }

                do { // Anschlusssegmente durchlaufen.
                    if (this.ses[toEdgeIdx] == 0) { // Edge wurde noch nicht durchfahren

                        int pqIdx = toEdgeIdx;
                        double costR = this.edgeCosts[toEdgeIdx];

                        OverlayEdge s = this.graph.getEdgeRamp(pqIdx);
                        if (s != null) {
                            targetId = s.getTargetId();
                            costR = this.findShortestPath ? s.getKm() : s.getH();
                            pqIdx = s.getIdx();
                        } else {
                            targetId = edgeTargetIds[toEdgeIdx];
                        }

                        if (uturnId != targetId) { // Kein UTurn!
                            // Abbiegevorschriften und Einbahnstrassen beachten.
                            boolean mayTurn = true;
                            if (noTurnBits != 0 && !this.ignoreRestrictions)
                                mayTurn &= ((noTurnBits & noTurnBit) == 0);
                            if ((edgeIds[toEdgeIdx] & 0x80000001) == 0x80000000)  // WrongWay
                                mayTurn &= this.ignoreOneWays;
                            if (mayTurn) {
                                // Reelle Kosten bis hier
                                costR += this.pqrCostsR;
                                double costH = costR + this.calcVertexCost(targetId);

                                if (costR < maxCost) {
                                    this.pqInsert(pqIdx, costR, costH);
                                    // Anders als beim VertexRouting: Hier schon setteln
                                    // Wir sparen damit die FromEdge in der PQ
                                    this.ses[pqIdx] = fromEdgeIdx;
                                }
                            }
                        }
                    }
                    toEdgeIdx++;
                    noTurnBit <<= 1;
                } while (sourceId == edgeSourceIds[toEdgeIdx]);
            }
        }
        return -1;
    }

    @Override
    public boolean isVisited(int targetId) {
        if (this.targetEdgeIdx < 0) return false;
        if (this.targetEdgeIdx < this.firstVirtualEdgeIdx) {
            return this.graph.getGraph().getEdgeTargetIds()[this.targetEdgeIdx] == targetId;
        } else {
            OverlayEdge s = this.graph.getOverlayEdge(this.targetEdgeIdx);
            return s.getTargetId() == targetId;
        }
    }

    @Override
    public int[] makePath(int targetId) {
        if (!this.isVisited(targetId)) return null;
        int edgeIdx = this.targetEdgeIdx;

        final int INITIAL_CAPACITY = 1000;
        int[] pathIdxs = new int[INITIAL_CAPACITY];
        int n = 0;
        while (edgeIdx != firstVirtualEdgeIdx -1 /*DeadEdgeIdx*/) {
            if (pathIdxs.length == n) pathIdxs = Arrays.copyOf(
                    pathIdxs, pathIdxs.length + INITIAL_CAPACITY);
            pathIdxs[n++] = edgeIdx;
            edgeIdx = this.ses[edgeIdx]; // Vorgaenger
        }
        pathIdxs = Arrays.copyOf(pathIdxs, n);
        // Reverse
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIdxs[i]; pathIdxs[i] = pathIdxs[j]; pathIdxs[j] = z;
        }
        return pathIdxs;
    }

    @Override
    public final int[] getVisited() {
        int[] edgeTargetIds = this.graph.getGraph().getEdgeTargetIds();
        int len = edgeTargetIds.length -1; // nur reelle Edges
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int edgeIdx = 0; edgeIdx < len; edgeIdx++) {
            if (this.ses[edgeIdx] > 0) {
                if (buffer.length < edgeIdx) // Ueberlauf korrigieren
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = edgeTargetIds[edgeIdx];
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    /**
     * Liefert die heuristischen Kosten eines Vertex.
     * Diese Implementation liefert entweder 0 fuer Dijkstra
     * oder die heuristische Distanz zum Ziel fuer AStar.
     * Eine abgeleitete Klasse koennte an dieser Stelle z.B.
     * auch noch konstante Kosten einer Vertex-Klasse hinzuaddieren.
     * So z.B. fuer Ampeln, Mautstationen oder Bahnuebergaenge.
     * @param targetId Id des Vertex
     * @return double Kosten.
     */
    protected double calcVertexCost(int targetId) {
        double h = 0d;
        if (this.heuristicFactor != 0d) { // nur rechnen wenn !=0 (schneller).
            float lat, lon;
            if (targetId < this.firstVirtualVertexId) {
                lat = this.lats[targetId];
                lon = this.lons[targetId];
            } else {
                OverlayNode nd = this.graph.getOverlayNode(targetId);
                lat = nd.getLat();
                lon = nd.getLon();
            }

            if (this.heuristicFactor != 0) {
                h = GeoUtils.calcDistanceEuklidCoarse(
                        lat, lon, this.targetLat, this.targetLon);
                h *= this.heuristicFactor;
            }
        }
        return h;
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqIdxs = Arrays.copyOf(this.pqIdxs, this.pqCapacity);
        this.pqCostsR = Arrays.copyOf(this.pqCostsR, this.pqCapacity);
        this.pqCostsH = Arrays.copyOf(this.pqCostsH, this.pqCapacity);
        log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsert(int edgeIdx, double costsR, double costsH) {
        int pqPos = ++this.pqSize;
        if (this.pqSize >= this.pqCapacity) this.pqGrow();

        this.pqIdxs[pqPos] = edgeIdx;
        this.pqCostsR[pqPos] = costsR;
        this.pqCostsH[pqPos] = costsH;

        this.pqUpheap(pqPos);
    }

    private void pqRemove() {
        this.pqrEdgeIdx = this.pqIdxs[1];
        this.pqrCostsR = this.pqCostsR[1];
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap(1);
    }

    private void pqSwap(int pos1, int pos2) {
        // PQ-Swap
        int idx1 = this.pqIdxs[pos1];
        double costsR1 = this.pqCostsR[pos1];
        double costsH1 = this.pqCostsH[pos1];
        int idx2 = this.pqIdxs[pos2];
        double costsR2 = this.pqCostsR[pos2];
        double costsH2 = this.pqCostsH[pos2];

        this.pqIdxs[pos1] = idx2;
        this.pqCostsR[pos1] = costsR2;
        this.pqCostsH[pos1] = costsH2;
        this.pqIdxs[pos2] = idx1;
        this.pqCostsR[pos2] = costsR1;
        this.pqCostsH[pos2] = costsH1;
    }

    private void pqUpheap(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[parentPos] < this.pqCostsH[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos) {
        int childPos = pos << 1; // * 2 = linkes Kind
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqCostsH[childRightPos] < this.pqCostsH[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[pos] < this.pqCostsH[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
