/*
 * $Id: OverlayGraph.java 5721 2023-01-02 18:54:35Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static java.util.Arrays.copyOf;

import java.util.Arrays;
import java.util.Comparator;

import de.cm.osm2po.primitives.IntTHashMap;
import de.cm.osm2po.primitives.VarBooleanArray;

/**
 * Repraesentiert eine Ueberlagerung des statische OSM-Streckennetzes
 * {@link Graph} und bildet die Grundlage fuer das Routing
 * mittels {@link AbstractRouter}-Instanzen. Das heisst:
 * <br> Edges und Vertices duerfen hier sowohl auf dem Graphen selbst
 * (siehe {@link OverlayNode#snap(Graph, int[])}) als auch als reine
 * Koordinaten ergaenzt werden, und dies sogar pro Instanz.
 * Der grundsaetzliche Trick besteht darin, die im Routing benoetigten
 * Indizes bzw. IDs fuer Edges und Vertices beim Hinzufuegen,
 * ausgehend von den bereits vergebenen im statischen Graphen,
 * fortlaufend weiter hochzuzaehlen, wodurch eine einfache Unterscheidung
 * moeglich wird.
 *
 * @author (c) 2014 - Carsten Moeller - info@osm2po
 */
public class OverlayGraph implements AbstractGraph{

    private Graph graph;

    private int vertexIdOffset;
    private int edgeIdxOffset;

    private int nOverlayEdges;
    private int nOverlayNodes;

    private OverlayEdge[] overlayEdges; // Alle OverlayEdges
    private OverlayNode[] overlayNodes; // Alle OverlayNodes

    // Start-Segmente einer OrignalEdge, wenn gesplittet
    private IntTHashMap<OverlayEdge> edgeRampMap;
    // Alle virtuellen Nodes, die auf einer gesplitteten Edge liegen.
    // Wird benoetigt, um nachtraeglich weitere Nodes hinzuzufuegen.
    private IntTHashMap<OverlayNode[]> snappedNodesMap;
    // Alle wegfuehrenden Edges virtueller aber auch statischer VertexIDs
    private IntTHashMap<OverlayEdge[]> outgoingEdgesMap;
    // Indikator-Hash fuer schnelles Lookup von echten Vertices, an denen
    // rein virtuelle Edges haengen.
    private VarBooleanArray vertexRamp;

    /**
     * @param graph {@link Graph} NOT NULL.
     */
    public OverlayGraph(Graph graph) {
        if (null == graph)
            throw new IllegalArgumentException("Graph must not be null");

        this.graph = graph;

        // Interne offsets sind Max(Original) + 1;
        this.vertexIdOffset = graph.getEdgeEntries().length;
        this.edgeIdxOffset = graph.getEdgeIds().length;

        this.nOverlayEdges = 0;
        this.overlayEdges = new OverlayEdge[0x400];
        this.nOverlayNodes = 0;
        this.overlayNodes = new OverlayNode[0x400];
        this.outgoingEdgesMap = new IntTHashMap<OverlayEdge[]>(0x400);
        this.edgeRampMap = new IntTHashMap<OverlayEdge>(0x400);
        this.snappedNodesMap = new IntTHashMap<OverlayNode[]>(0x400);
    }

    /** @return int Zuletzt vergebene, virtuelle VertexId. */
    public int getLastVirtualVertexId() {
        return this.vertexIdOffset + this.nOverlayNodes -1;
    }

    /** @return int Zuletzt vergebene, virtuelle EdgeIdx. */
    public int getLastVirtualEdgeIdx() {
        return this.edgeIdxOffset + this.nOverlayEdges -1;
    }

    /** @return int Erste (zu) vergebene virtuelle Vertex-ID. */
    public int getFirstVirtualVertexId() {
        return this.vertexIdOffset;
    }

    /** @return int Erste (zu) vergebene virtueller Edge-Index. */
    public int getFirstVirtualEdgeIdx() {
        return this.edgeIdxOffset;
    }

    /**
     * Findet einen vergleichbaren Knoten im Graphen, wenn vorhanden.
     * @param overlayNode {@link OverlayNode}
     * @return bereits vorhandenen {@link OverlayNode} mit gleichen
     * Daten oder null, wenn nicht gefunden.
     */
    private OverlayNode findEqualNode(OverlayNode overlayNode) {
        // Self is already inserted
    	if (null == overlayNode || overlayNode.isOnVertex()) return null;
        if (overlayNode.isInserted()) return overlayNode;

        // Another snapped Node equals
        if (overlayNode.isOnEdge()) {
            int edgeIdx = overlayNode.getEdgeIdx();
            OverlayNode[] nds = this.snappedNodesMap.get(edgeIdx);
            if (null == nds) return null;
            for (OverlayNode nd : nds) {
                if (nd.getRatio() == overlayNode.getRatio()) {
                    return nd;
                }
            }
        }
        return null;
    }

    /**
     * Fuegt einen OverlayNode in den virtuellen Graphen ein.
     * @param overlayNode {@link OverlayNode}
     * @return uebergebener {@link OverlayNode} oder, wenn bereits
     * vorhanden, ein bereits gleicher aus dem Graphen.
     * Nodes, die lediglich Vertices aus dem Original-Graphen
     * repraesentieren, werden nicht eingefuegt und lediglich um
     * Koordinaten erweitert zurueckgeliefert.
     */
    public OverlayNode insertNode(OverlayNode overlayNode) {
        if (null == overlayNode) return null;
        if (overlayNode.isOnVertex()) {
        	// OverlayNode(vertexId)-Konstruktor absichern
        	overlayNode.setLat(this.graph.getLats()[overlayNode.getId()]);
        	overlayNode.setLon(this.graph.getLons()[overlayNode.getId()]);
        	return overlayNode;
        }

        OverlayNode nd = this.findEqualNode(overlayNode);

        if (null == nd) {
            nd = overlayNode;
            int nextId = this.vertexIdOffset + this.nOverlayNodes;
            int edgeIdx = nd.getEdgeIdx();

            // ArrayOverflow-Check
            if (this.nOverlayNodes == this.overlayNodes.length)
                this.overlayNodes = Arrays.copyOf(
                        this.overlayNodes, this.overlayNodes.length * 2);

            nd.setId(nextId);
            this.overlayNodes[this.nOverlayNodes++] = nd;

            // overlayNode liegt auf Edge
            if (nd.isOnEdge()) {
                OverlayNode[] nds = this.snappedNodesMap.get(edgeIdx);
                if (null == nds) {
                    nds = new OverlayNode[] {nd};
                } else {
                	int n = nds.length;
                    nds = Arrays.copyOf(nds, n + 1);
                    nds[n] = nd;

                    Arrays.sort(nds, new Comparator<OverlayNode>() {
                        @Override
                        public int compare(OverlayNode o1, OverlayNode o2) {
                            if (o1.getRatio() < o2.getRatio()) return -1;
                            else if (o1.getRatio() > o2.getRatio()) return 1;
                            else return 0;
                        }
                    });
                }
                this.snappedNodesMap.put(edgeIdx, nds);
                if (this.snappedNodesMap.isRehashNeeded()) {
                    this.snappedNodesMap.rehash();
                }

                this.split(nds);
            }
        }
        return nd;
    }

    /**
     * Fuegt eine OverlayEdge in den Graphen ein.
     * @param overlayEdge {@link OverlayEdge}
     * <br><br><b>Wichtig:</b><br>
     * Da die Verbindungen aufgrund der in der OverlayEdge
     * gespeicherten VertexIds erstellt werden, muessen diese
     * vorher bekannt und gueltig sein. Source- und TargetId
     * duerfen sowohl statische als auch virtuelle Ids sein.
     */
    public void insertEdge(OverlayEdge overlayEdge) {
        overlayEdge.setIdx(this.edgeIdxOffset + this.nOverlayEdges);

        // Array-Overflow-Check
        if (this.nOverlayEdges == this.overlayEdges.length)
            this.overlayEdges = Arrays.copyOf(this.overlayEdges, this.nOverlayEdges * 2);

        this.overlayEdges[this.nOverlayEdges++] = overlayEdge;

        // Outgoing Edge an Node anheften
        int sourceId = overlayEdge.getSourceId();
        OverlayEdge[] outEdges = this.outgoingEdgesMap.get(sourceId);
        if (null == outEdges) {
        	outEdges = new OverlayEdge[]{overlayEdge};
        } else {
        	int n = outEdges.length;
        	outEdges = Arrays.copyOf(outEdges, n +1);
        	outEdges[n] = overlayEdge;
        }
        this.outgoingEdgesMap.put(sourceId, outEdges);
        if (this.outgoingEdgesMap.isRehashNeeded()) {
            this.outgoingEdgesMap.rehash();
        }

        // Echte Vertices vermerken, an denen rein virtuelle
        // Kanten haengen
        int edgeIdx = overlayEdge.getEdgeIdx();
        if (0 == edgeIdx && sourceId < this.vertexIdOffset) {
            if (null == this.vertexRamp) // LazyInit, da selten verwendet
                this.vertexRamp = new VarBooleanArray(this.vertexIdOffset);
            this.vertexRamp.set(sourceId, true);
        }

    }

    /**
     * Liefert eine OverlayEdge aufgrund ihrer virtuellen Idx.
     * @param virtualIdx int Idx.
     * @return {@link OverlayEdge} oder null, wenn nicht gefunden.
     */
    public OverlayEdge getOverlayEdge(int virtualIdx) {
        int idx = virtualIdx - this.edgeIdxOffset;
        if (idx < 0 || idx >= this.nOverlayEdges) return null;
        return this.overlayEdges[idx];
    }

    /**
     * Liefert einen OverlayNode aufgrund seiner virtuellen Id.
     * @param virtualId int Id.
     * @return {@link OverlayNode} oder null, wenn nicht gefunden.
     */
    public OverlayNode getOverlayNode(int virtualId) {
        int idx = virtualId - this.vertexIdOffset;
        if (idx < 0 || idx >= this.nOverlayNodes) return null;
        return this.overlayNodes[idx];
    }

    /**
     * Liefert die abgehenden Kanten zu einem Node.
     * <br><br><b>Wichtig:</b><br>
     * Der Node kann ebenso ein Vertex aus dem statischen Graphen sein.
     * In diesem Fall handelt es sich entweder um den Beginn einer
     * gesplitteten realen oder um eine rein virtuelle Edge.
     * @param sourceId int Id des Vertex oder OverlayNodes.
     * @return Array of {@link OverlayEdge} NULLABLE.
     */
    public OverlayEdge[] getOutgoingOverlayEdges(int sourceId) {
        return this.outgoingEdgesMap.get(sourceId);
    }

    /**
     * Liefert abgehende, rein virtuelle Kanten eines Vertex.
     * @param sourceVertexId int Id des Vertex aus statischem Graphen.
     * @return Array of {@link OverlayEdge} NULLABLE.
     * @throws IllegalArgumentException wenn id nicht statisch
     * @see #getOutgoingOverlayEdges(int)
     */
    public OverlayEdge[] getOutgoingVirtualEdges(int sourceVertexId) {
        if (sourceVertexId >= this.vertexIdOffset) throw new IllegalArgumentException(
                "OverlayVertexId not supported");
        if (null == this.vertexRamp) return null;
        if (!this.vertexRamp.get(sourceVertexId)) return null;
        OverlayEdge[] oes = this.outgoingEdgesMap.get(sourceVertexId);
        OverlayEdge[] oesFiltered = new OverlayEdge[oes.length];
        int n = 0;
        for (OverlayEdge oe : oes) {
            if (oe.getEdgeIdx() == 0) oesFiltered[n++] = oe;
        }
        return copyOf(oesFiltered, n);
    }


    /**
     * Findet zu einer statischen Edge die gesplittete virtuelle
     * Variante, wenn vorhanden.
     * @param edgeIdx int Index der Edge im original Graphen.
     * @return {@link OverlayEdge} oder null, wenn nicht gefunden.
     */
    public OverlayEdge getEdgeRamp(int edgeIdx) {
        return this.edgeRampMap.get(edgeIdx);
    }

    /**
     * Generiert die OverlayEdges fuer eine nicht virtuelle Edge
     * aufgrund der uebergebenen OverlayNodes und fuegt diese
     * in den Graphen ein.
     * @param overlayNodes Array of {@link OverlayNode} NOT NULL.
     * <br><b>Wichtig:</b><br>
     * Alle uebergebenen OverlayNodes muessen auf der gleichen Edge liegen
     * und bereits eine Id besitzen. Wenn mehrere Punkte uebergeben werden,
     * dann muessen diese zudem nach Ratio sortiert sein.
     */
    private void split(OverlayNode[] overlayNodes) {
        int n = overlayNodes.length;
        if (0 == n) return;
        int edgeIdx = overlayNodes[0].getEdgeIdx(); // Precondition!

        // Echten Start- und ZielVertex der Edge rekonstruieren
        n += 2;
        int sourceId = this.graph.getEdgeSourceIds()[edgeIdx];
        int targetId = this.graph.getEdgeTargetIds()[edgeIdx];
        OverlayNode[] ndChain = new OverlayNode[n];
        System.arraycopy(overlayNodes, 0, ndChain, 1, n-2);

        // Kuenstliche Start- und Ziel-Vertices einbauen
        ndChain[0] = new OverlayNode(sourceId);
        ndChain[n-1] = new OverlayNode(targetId);

        for (int i = 1; i < n; i++) {
            // Vorwaerts
            OverlayEdge overlayEdge = OverlayNode.createOverlayEdge(
                    this.graph, ndChain[i-1], ndChain[i]);
            this.insertEdge(overlayEdge);
            if (1 == i) this.edgeRampMap.put(overlayEdge.getEdgeIdx(), overlayEdge);

            // Rueckwaerts
            overlayEdge = overlayEdge.copyReversed(this.graph);
            if (overlayEdge != null) {
                this.insertEdge(overlayEdge);
                if (n -1 == i) this.edgeRampMap.put(overlayEdge.getEdgeIdx(), overlayEdge);
            }
        }

        if (this.edgeRampMap.isRehashNeeded()) {
            this.edgeRampMap.rehash();
        }
    }



    /**
     * Wandelt eine Liste von EdgeIdx in eine Liste von {@link OverlayEdge}s.<br>
     * Statische Edges aus dem {@link Graph} werden nicht mit Geometrien
     * angereichert und bleiben somit <code>null</code><br>
     * Ruft dazu {@link #translatePath(OverlayGraph, int[])} mit this.
     * @param path int[] Array von Referenzen auf Edges.
     * @return {@link OverlayEdge}[] Array oder null, wenn null uebergeben.
     */
    public OverlayEdge[] translatePath(int[] path) {
        return translatePath(this, path);
    }

    /**
     * Wandelt eine Liste von EdgeIdx in eine Liste von {@link OverlayEdge}s.<br>
     * Statische Edges aus dem {@link Graph} werden nicht mit Geometrien
     * angereichert und bleiben somit <code>null</code>
     * @param overlayGraph {@link OverlayGraph}
     * @param path int[] Array von Referenzen auf Edges.
     * @return {@link OverlayEdge}[] Array oder null, wenn null uebergeben.
     */
    public static OverlayEdge[] translatePath(
            OverlayGraph overlayGraph, int[] path) {

        if (null == path) return null;
        int[] sourceIds = overlayGraph.graph.getEdgeSourceIds();
        int[] targetIds = overlayGraph.graph.getEdgeTargetIds();
        float[] costsKm = overlayGraph.graph.getEdgeCostsKm();
        float[] costsH = overlayGraph.graph.getEdgeCostsH();

        OverlayEdge[] overlayEdges = new OverlayEdge[path.length];
        for (int i = 0; i < path.length; i++) {
            int idx = path[i];
            if (idx >= overlayGraph.edgeIdxOffset) {
                overlayEdges[i] = overlayGraph.getOverlayEdge(idx);
            } else {
                overlayEdges[i] = new OverlayEdge(
                        idx, sourceIds[idx], targetIds[idx],
                        costsKm[idx], costsH[idx], null);
            }
        }
        return overlayEdges;
    }

    /**
     * Wandelt eine Liste von EdgeIdx in eine Liste von
     * Pseudo-{@link OverlayEdge}s ohne Geometrien.
     * @param graph {@link Graph}
     * @param path int[] Array von Referenzen auf Edges.
     * @return {@link OverlayEdge}[] Array oder null, wenn null uebergeben.
     */
    public static OverlayEdge[] translatePath(Graph graph, int[] path) {
        if (null == path) return null;
        int[] sourceIds = graph.getEdgeSourceIds();
        int[] targetIds = graph.getEdgeTargetIds();
        float[] costsKm = graph.getEdgeCostsKm();
        float[] costsH = graph.getEdgeCostsH();

        OverlayEdge[] overlayEdges = new OverlayEdge[path.length];
        for (int i = 0; i < overlayEdges.length; i++) {
            int idx = path[i];
            overlayEdges[i] = new OverlayEdge(
                    idx, sourceIds[idx], targetIds[idx],
                    costsKm[idx], costsH[idx], null);
        }
        return overlayEdges;
    }

    /**
     * Berechnet die Strecke einer Route.
     * @param path int[] Array von Referenzen auf Edges.
     * @return double Strecke
     */
    public double calcPathLength(int[] path) {
        float[] kms = this.graph.getEdgeCostsKm();
        double sum = 0d;
        for (int i = 0; i < path.length; i++) {
            int x = path[i];
            OverlayEdge oe = this.getOverlayEdge(x);
            sum += null == oe ? kms[x] : oe.getKm();
        }
        return sum;
    }

    /**
     * Berechnet die Fahrzeit einer Route.
     * @param path int[] Array von Referenzen auf Edges.
     * @return double Strecke
     */
    public double calcPathTime(int[] path) {
        float[] hs = this.graph.getEdgeCostsH();
        double sum = 0d;
        for (int i = 0; i < path.length; i++) {
            int x = path[i];
            OverlayEdge oe = this.getOverlayEdge(x);
            sum += null == oe ? hs[x] : oe.getH();
        }
        return sum;
    }

    public Graph getGraph() {
        return this.graph;
    }
}
