/*
 * $Id: OverlayNode.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2012/2014 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static de.cm.osm2po.primitives.VarNum.readUnsigned;
import static de.cm.osm2po.primitives.VarNum.writeUnsigned;
import static java.lang.Double.doubleToLongBits;
import static java.lang.Double.longBitsToDouble;
import static java.lang.Float.floatToIntBits;
import static java.lang.Float.intBitsToFloat;

import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;

/**
 * Diese Klasse repraesentiert einen geografischen Punkt
 * auf dem {@link OverlayGraph}<code>en</code>.<br>
 * Dies kann a) ein Vertex, b) ein beliebiger Punkt auf einer Edge
 * oder c) eine voellig losgeloeste Koordinate sein.
 * <br><br><b>Wichtig:</b><br>
 * Da unterschiedliche Knoten gleiche Koordinaten besitzen koennen,
 * wird die Eindeutigkeit ueber die EdgeIdx (Kante) und der Ratio
 * (Relative Position auf der Kante) hergestellt. Rueckwaertskanten,
 * auf denen Knoten ebenfalls liegen, sind daher unzulaessig und
 * werden falls moeglich durch ihre Vorwaerts-Varianten ersetzt.<br>
 * Dies ist vor allem deshalb notwendig, da bei einer Aufsplittung
 * einer Kante durch mehr als einen Knoten die Reihenfolge definiert
 * sein muss und sich eine solche Aktion nicht auf zwei faktisch
 * identische Kanten beziehen kann, die sich lediglich in der Richtung
 * unterscheiden.<br>
 * Handelt es sich hingegen um einen echten Vertex, also einer
 * Kreuzung, dann wird die originale VertexId fuer die Eindeutigkeit
 * herangezogen.
 *
 * @author (c) 2012/2014 - Carsten Moeller - info@osm2po.de
 */
public class OverlayNode implements Var<OverlayNode> {

    private int id; // Id fuer das Routing. VertexId oder VirtualVertexId
    private int edgeIdx; // Referenz auf Original-EdgeIdx bei Ueberlagerung
    private int pointIdx; // GeometriePunkt-Index einer Edge, 0-Wertig
    private boolean isPoint; // false: GeometriePunkt liegt zwischen pointIdx und PointIdx+1
    private double ratio; // Teilungsverhaeltnis, wenn Original-Edge gesplittet.
    private float lat;
    private float lon;

    // Sollten nur intern, z.B. vom OverlayGraphen gerufen werden
    protected void setId(int id) {this.id = id;}
    protected void setLat(float lat) {this.lat = lat;}
    protected void setLon(float lon) {this.lon = lon;}

    // Klarstellung der Interpretation der Member-Kombis im OverlayGraphen
    public boolean isOnVertex() {return this.edgeIdx < 0;} // entspricht Vertex
    public boolean isOnEdge() {return this.edgeIdx > 0;} // liegt auf Edge
    public boolean isDetached() {return this.edgeIdx == 0;} // freistehend
    public boolean isInserted() {return this.edgeIdx >= 0 && this.id > 0;} // eingefuegt

    public int getId() {return this.id;}
    public float getLat() {return this.lat;}
    public float getLon() {return this.lon;}
    public LatLon toLatLon(){return new LatLon(this.lat, this.lon);};
    public int getEdgeIdx() {return this.edgeIdx;}
    public int getEdgePointIdx() {return this.pointIdx;}
    public boolean isPoint() {return this.isPoint;}
    public double getRatio() {return this.ratio;}

    /**
     * Standard-Konstruktor.
     * <br><br><b>Wichtig:</b><br>
     * Dieser Konstruktor sollte dient fuer interne Zwecke.
     * Daher nicht direkt verwenden.
     */
    public OverlayNode() {} // Public wegen Var.readFromStream()

    /**
     * Erstellt einen Knoten, der direkt einen Vertex im
     * Graphen adressiert.
     * @param vertexId int VertexId im Original-Graphen.
     * @see #isOnVertex()
     * <br><br><b>Wichtig:</b><br>
     * Latitude und Longitude werden <b>nicht</b> gesetzt.
     * Hierzu muss (nochmals) {@link #snap(Graph, int[])}
     * aufgerufen werden.
     */
    public OverlayNode(int vertexId) {
        if (vertexId < 0) throw new IllegalArgumentException(
                "vertexId must be greater or equal 0");
        this.id = vertexId;
        this.edgeIdx = -1;
    }

    /**
     * Erstellt einen virtuellen Knoten ohne Bezug zum Graphen.
     * @param lat float Latitude
     * @param lon float Longitude
     * @see #isDetached()
     */
    public OverlayNode(float lat, float lon) {
        this.lat = lat;
        this.lon = lon;
    }

    /**
     * Erstellt eine Kopie.
     * @return {@link OverlayNode}
     */
    public OverlayNode copy() {
        OverlayNode nd = new OverlayNode();
        nd.id = this.id;
        nd.lat = this.lat;
        nd.lon = this.lon;
        nd.edgeIdx = this.edgeIdx;
        nd.isPoint = this.isPoint;
        nd.ratio = this.ratio;
        return nd;
    }

    /**
     * Findet den dichteste Beruehrungspunkt auf dem Original-Graphen
     * und verschiebt diesen Punkt dorthin.
     * Handelt es sich um einen Vertex-Stellvertreter, der mittels
     * {@link #OverlayNode(int)} erzeugt wurde, so werden lediglich die
     * Koordinaten gesetzt und dabei implizit die Id selbst validiert.
     * @param graph {@link Graph}
     * @param edgeIdxs int[] Array von EdgeIdx, NULLABLE.
     * @return boolean gefunden und verschoben, ja/nein.
     * <br><br><b>Wichtig:</b><br>
     * Diese Methode modifiziert interne Members. Soll das Original
     * beibehalten werden, so sollte vorher {@link #copy()} ausgefuehrt
     * werden.<br>
     * Wird der Parameter <code>edgeIdxs=null</code> gesetzt, so werden ca. 100
     * beliebige Edges mittels {@link Graph#findCloseEdges(float, float, int)}
     * untersucht. Sollten jedoch nur bestimmte Edge-Typen infrage kommen,
     * so kann hier ggf. eine reduzierte Suchmenge gesetzt werden.
     */
    public boolean snap(Graph graph, int[] edgeIdxs) {
        if (this.isOnVertex()) {
            // Koordinaten setzen und implizit Id validieren.
        	this.setLat(graph.getLats()[this.id]);
        	this.setLon(graph.getLons()[this.id]);
        	return true;
        }

        if (null == edgeIdxs) edgeIdxs = graph.findCloseEdges(this.lat, this.lon, 100);
        if (0 == edgeIdxs.length) return false;

        int[] edgeIds = graph.getEdgeIds();
        int[] edgeSourceIds = graph.getEdgeSourceIds();
        int[] edgeTargetIds = graph.getEdgeTargetIds();

        double minDist = Double.MAX_VALUE;
        double dist = minDist;

        LatLons foundGeometry = null;
        double foundLat = 0d, foundLon = 0d; // wg.double!

        // TypeCast und Lesbarkeit. p3 ist der zu projezierende Punkt
        double p3y = GeoUtils.EPSG_4326_TO_900913_Lat(this.lat);
        double p3x = GeoUtils.EPSG_4326_TO_900913_Lon(this.lon);

        // Nur ueber Vorwaerts-Kanten iterieren.
        // Wenn nicht vorwaerts dann vorher umdrehen.
        for (int i = 0; i < edgeIdxs.length; i++) {
            int edgeIdx = edgeIdxs[i];
            if (Graph.isReverse(edgeIds[edgeIdx])) {
                edgeIdx = graph.findReverseEdge(edgeIdx);
                if (edgeIdx < 1) throw new RuntimeException("No Reverse-Edge found");
            }

            LatLons currentGeometry = graph.lookupGeometry(edgeIdx);
            int nCoords = currentGeometry.getSize();

            // Ueber Geometrie-Punkte iterieren.
            for (int j = 0; j < nCoords; j++) {
                long coord = currentGeometry.getCoord(j);
                double lat = latOf(coord);
                double lon = lonOf(coord);

                double p1y = GeoUtils.EPSG_4326_TO_900913_Lat(lat);
                double p1x = GeoUtils.EPSG_4326_TO_900913_Lon(lon);

                // Stuetzpunkte A, B, C, ... betrachten
                dist = StrictMath.hypot(p3x - p1x, p3y - p1y);
                if (dist < minDist) {
                    minDist = dist;
                    foundGeometry = currentGeometry;
                    foundLat = p1y;
                    foundLon = p1x;
                    this.pointIdx = j;
                    this.isPoint = true;
                    this.edgeIdx = edgeIdx;

                    // Wenn Vertex (Start/End), dann Sonderlocke
                    if (j == 0) {
                        this.id = edgeSourceIds[edgeIdx];
                        this.edgeIdx = -1;
                    } else if (j == nCoords -1) {
                        this.id = edgeTargetIds[edgeIdx];
                        this.edgeIdx = -1;
                    } else {
                    	this.id = 0;
                    }
                }

                // Teilstrecken A->B, B->C, ... betrachten
                if (j < nCoords - 1) {
                    coord = currentGeometry.getCoord(j +1);
                    lat = latOf(coord);
                    lon = lonOf(coord);

                    double p2y = GeoUtils.EPSG_4326_TO_900913_Lat(lat);
                    double p2x = GeoUtils.EPSG_4326_TO_900913_Lon(lon);
                    // Vectoren normalisieren
                    double ay = p2y - p1y, ax = p2x - p1x; // a = p2-p1
                    double by = p3y - p1y, bx = p3x - p1x; // b = p3-p1
                    // Lotfusspunkt berechnen
                    double aa = ax * ax + ay * ay; // Skalarprodukt a * a = |a|^2
                    double ab = ax * bx + ay * by; // Skalarprodukt a * b
                    double e = ab / aa; // Stauchungsfaktor des Vektors p1-p2

                    // Innerhalb der Strecke, sonst Gerade.
                    if (!(e < 0 || e > 1)) {
                        // p4 ist der gesuchte Punkt auf der Strecke
                        double p4y = p1y + ay * e, p4x = p1x + ax * e;
                        dist = StrictMath.hypot(p3x - p4x, p3y - p4y);
                        if (dist < minDist) {
                            minDist = dist;
                            foundGeometry = currentGeometry;
                            foundLat = p4y;
                            foundLon = p4x;
                            this.isPoint = false;
                            this.id = 0;
                            this.edgeIdx = edgeIdx;
                            this.pointIdx = j;
                        }
                    }
                }
            }
        }

        // Egal ob Vertex oder auf Kante!
        this.lat = (float) GeoUtils.EPSG_900913_TO_4326_Lat(foundLat);
        this.lon = (float) GeoUtils.EPSG_900913_TO_4326_Lon(foundLon);

        if (this.isOnEdge()) {
            OverlayNode source = new OverlayNode(edgeSourceIds[edgeIdx]);
            OverlayNode target = this;
            target.ratio = 1; // Bedingung fuer createOverlayEdge()
            OverlayEdge overlayEdge = createOverlayEdge(graph, source, target);
            double edgeLen = (null == foundGeometry) ? 0d : foundGeometry.calcLengthKm();
            double overlayEdgeLen = overlayEdge.getGeometry().calcLengthKm();
            this.ratio = overlayEdgeLen / edgeLen;
        }

        return true;
    }

    /**
     * Ermittelt, ob zwischen zwei {@link OverlayNode}s eine direkte Verbindung
     * existiert - beide also auf der gleichen Kante liegen und liefert
     * den ggf. neu entstandenen Abschnitt als {@link OverlayEdge} zurueck.
     * @param graph {@link Graph}
     * @param ndSource {@link OverlayNode} Start
     * @param ndTarget {@link OverlayNode} Ziel
     * @return Verbindung als {@link OverlayEdge} oder null, wenn keine.
     * <br><b>Wichtig:</b><br>
     * Diese Methode liefert nur dann ein Ergebnis, wenn source oder/und target
     * echte virutelle Punkte sind und einer Edge zugeordnet werden koennen
     * (edgeIdx != 0). Ferner muss die Reihenfolge beachtet werden, da
     * diese Klasse nur Vorwaerts-Edges verarbeitet und somit
     * ratio(source) &lt; ratio(target) sein muss.<br>
     * Eine OverlayEdge speichert keine OverlayNodes, sondern nur deren IDs,
     * daher sollten diese vorher gesetzt sein.
     */
    public static OverlayEdge createOverlayEdge(Graph graph,
            OverlayNode ndSource, OverlayNode ndTarget) {

        boolean sourceIsVertex = ndSource.isOnVertex();
        boolean targetIsVertex = ndTarget.isOnVertex();

        if (sourceIsVertex && targetIsVertex) return null;
        if (!sourceIsVertex && !targetIsVertex
                && ndSource.ratio >= ndTarget.ratio) return null;

        int edgeIdx = ndSource.edgeIdx;
        int sourceId = ndSource.id;
        int targetId = ndTarget.id;

        if (sourceIsVertex) {
            edgeIdx = ndTarget.edgeIdx;
            sourceId = graph.getEdgeSourceIds()[edgeIdx];
            if (sourceId != ndSource.id) return null;
        } else if (targetIsVertex) {
            edgeIdx = ndSource.edgeIdx;
            targetId = graph.getEdgeTargetIds()[edgeIdx];
            if (targetId != ndTarget.id) return null;
        }

        LatLons llEdge = graph.lookupGeometry(edgeIdx);
        LatLons llOverlayEdge = new LatLons();

        if (targetIsVertex) { // Sonderlockenkorrektur
            ndTarget.pointIdx = llEdge.getSize() -1;
            ndTarget.ratio = 1;
        }

        int pIdx = ndSource.pointIdx;
        if (!ndSource.isOnVertex()) {
            llOverlayEdge.append(new LatLon(ndSource.lat, ndSource.lon));
            pIdx++;
        }
        for (; pIdx <= ndTarget.pointIdx; pIdx++) {
            llOverlayEdge.append(llEdge.getCoord(pIdx));
        }
        if (!ndTarget.isOnVertex()) {
            llOverlayEdge.append(new LatLon(ndTarget.lat, ndTarget.lon));
        }

        double ratio = ndTarget.ratio - ndSource.ratio;
        float km = (float) (ratio * graph.getEdgeCostsKm()[edgeIdx]);
        float h = (float) (ratio * graph.getEdgeCostsH()[edgeIdx]);

        OverlayEdge overlayEdge = new OverlayEdge(
                edgeIdx, sourceId, targetId, km, h, llOverlayEdge);
        return overlayEdge;
    }

    @Override
    public String toString() {
        String s = "Id:" + this.id
            + " EdgeIdx:" + this.edgeIdx
            + " PointIdx:" + this.pointIdx
            + " Ratio:" + Utils.DF2(this.ratio)
            + " LatLon:" + new LatLon(this.lat, this.lon);
        return s;
    }

    @Override
    public void writeToStream(OutStream outStream) {
        boolean isP = this.isPoint;
        boolean isV = this.edgeIdx == -1;

        int flags = 0;
        // Bit > 5 garantiert Code32.encoded startsWith(A)
        if (isP) flags |= 0x80;
        if (isV) flags |= 0x40;
        outStream.writeByte((byte) flags);

        if (!isV) {
            writeUnsigned(this.edgeIdx, outStream);
            writeUnsigned(this.pointIdx, outStream);
            writeUnsigned(doubleToLongBits(this.ratio), outStream);
        }

        writeUnsigned(this.id, outStream);

        outStream.writeInt(floatToIntBits(this.lat));
        outStream.writeInt(floatToIntBits(this.lon));
    }

    @Override
    public OverlayNode readFromStream(InStream inStream) {
        int flags = inStream.readByte();
        boolean isP = (flags & 0x80) != 0;
        boolean isV = (flags & 0x40) != 0;

        this.isPoint = isP;
        this.edgeIdx = -1;
        if (!isV) {
            this.edgeIdx = (int) readUnsigned(inStream);
            this.pointIdx = (int) readUnsigned(inStream);
            this.ratio = longBitsToDouble(readUnsigned(inStream));
        }

        this.id = (int) readUnsigned(inStream);

        this.lat = intBitsToFloat(inStream.readInt());
        this.lon = intBitsToFloat(inStream.readInt());

        return this;
    }

}