/*
 * $Id: OverlayVertexRouter.java 5721 2023-01-02 18:54:35Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.primitives.IntDoubleHashMap;
import de.cm.osm2po.primitives.IntSet;

/**
 * Dijkstra-Implemetation eines Routings auf einem ueberlagerten Graphen.<br>
 * Per Konfiguration sind sowohl AStar als auch OneToMany abbildbar.
 * Eine nachtraegliche Abfrage der Kosten aller Targets ist ebenfalls moeglich.
 * <br><br>
 * Diese Implementation beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * <li> {@link RouterParameters#getHeuristicFactor()} </li>
 * </ul>
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see OverlayGraph
 * @see VertexRouter
 * @see VisitedVerticesAware
 * @see TargetCostsAware
 */
public class OverlayVertexRouter extends AbstractRouter<OverlayGraph>
implements VisitedVerticesAware, FoundTargetsAware, TargetCostsAware {

    public OverlayVertexRouter(OverlayGraph graph, Log log) {
        super(graph, log);
        this.firstVirtualEdgeIdx = graph.getFirstVirtualEdgeIdx();
        this.firstVirtualVertexId = graph.getFirstVirtualVertexId();
    }

    private final static int PQ_INITIAL_CAPACITY = 10000;
    private int pqSize; // Current PQ size.
    private int pqCapacity; // Current PQ Capacity
    private int[] pqEdgeIdxs;
    private int[] pqTargetIds;
    private double[] pqCostsR;
    private double[] pqCostsH;

    private int[] wgbs; // OpenVisitedClosed-List

    // Current Edge (after pqRemove)
    private int pqrEdgeIdx;
    private int pqrTargetId;
    private double pqrCostR;

    private int firstVirtualVertexId;
    private int firstVirtualEdgeIdx;

    // Wird auf entweder h oder km gesetzt.
    private float[] edgeCosts;

    // Modi
    private boolean findShortestPath;
    private boolean ignoreOneWays;
    private boolean ignoreRestrictions;

    // Members fuer Heuristiken
    private float[] lats, lons; // Alle Koordinaten.
    private float targetLat, targetLon; // Zielkoordinate.
    private double heuristicFactor = 0d; // Bewertungsfaktor 0 Dijkstra, 1 AStar

    // Target-Kosten fuer OneToMany- aber auch OneToOne-Mode
    // TODO Target-Kosten ohne Map loesen
    private IntDoubleHashMap targetCostsMap;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        this.initParams(params);
        this.initAStar(constraints.targetIds[0]); // Unkritisch, wenn Targets > 1
        this.initPqAndWgb();
        if (constraints.targetIds.length > 1) this.heuristicFactor = 0; // kein AStar, da OneToMany
        return this.traverseInternal(sourceId, constraints.targetIds, constraints.maxCost);
    }

    @Override
    public int[] makePath(int toTargetId) {
        if (!this.isVisited(toTargetId)) return null;

        int[] edgeSourceIds = this.graph.getGraph().getEdgeSourceIds();

        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        int targetId = toTargetId;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY);
            int edgeIdx = this.wgbs[targetId] & 0x7FFFFFFF; // Bit 32 (Visited-Black-Flag).
            if (edgeIdx == this.firstVirtualEdgeIdx -1) break;

            int sourceId = 0;
            if (edgeIdx < this.firstVirtualEdgeIdx) { // Static Graph
                sourceId = edgeSourceIds[edgeIdx];
            } else { // Virtual Graph
                sourceId = this.graph.getOverlayEdge(edgeIdx).getSourceId();
            }

            pathIds[n++] = edgeIdx;
            targetId = sourceId;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        // Reverse
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    /**
     * Liefert die besuchten Vertex-Ids des unterlegten statischen
     * {@link Graph}en, nicht aber die des {@link OverlayGraph}en.
     * @return int[] statische VertexIds (not null)
     * @see VisitedVerticesAware#getVisited()
     */
    @Override
    public final int[] getVisited() {
        int len = this.wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) {
            int edgeIdx = this.wgbs[vertexId];
            boolean visited = edgeIdx < 0;
            boolean nonvirtual = vertexId < this.firstVirtualVertexId;
            if (visited && nonvirtual) {
                if (buffer.length == n)
                    buffer = Arrays.copyOf(buffer, (int) (n * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    @Override
    public final boolean isVisited(int vertexId) {
        if (vertexId < 1 || vertexId > this.wgbs.length -1) return false;
        return this.wgbs[vertexId] < 0;
    }

    @Override
    public double getTargetCost(int vertexId) {
        double cost = this.targetCostsMap.get(vertexId);
        if (IntDoubleHashMap.NULLVALUE == cost) return Double.NaN;
        return cost;
    }

    @Override
    public int[] getVisitedTargetIds() {
        return this.targetCostsMap.getKeys();
    }

    private void initParams(RouterParameters params) {
        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();
        this.heuristicFactor = params.getHeuristicFactor();

        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getGraph().getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getGraph().getEdgeCostsH();
            this.heuristicFactor /= 110;
        }
    }

    private void initPqAndWgb() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Max. ArraySize PQ
        this.pqEdgeIdxs = new int[pqCapacity];
        this.pqTargetIds = new int[pqCapacity];
        this.pqCostsR = new double[pqCapacity];
        this.pqCostsH = new double[pqCapacity];
        this.wgbs = new int[this.graph.getLastVirtualVertexId() + 1];
    }

    private void initAStar(int targetId) {
        Graph staticGraph = this.graph.getGraph();
        if (!staticGraph.isSupported(Graph.SUPPORT_LATLON)) {
            this.heuristicFactor = 0; // Kein AStar
        } else {
            this.lats = staticGraph.getLats();
            this.lons = staticGraph.getLons();
            if (targetId < this.firstVirtualVertexId) {
                this.targetLat = this.lats[targetId];
                this.targetLon = this.lons[targetId];
            } else {
                OverlayNode nd = this.graph.getOverlayNode(targetId);
                this.targetLat = nd.getLat();
                this.targetLon = nd.getLon();
            }
        }
    }

    private boolean traverseInternal(int sourceId, int[] targetIds, double maxCost) {
        Graph staticGraph = this.graph.getGraph();
        int[] edgeEntries = staticGraph.getEdgeEntries();
        int[] edgeSourceIds = staticGraph.getEdgeSourceIds();
        int[] edgeTargetIds = staticGraph.getEdgeTargetIds();
        int[] edgeIds = staticGraph.getEdgeIds();
        byte[] edgeNoTurnBits = staticGraph.getEdgeNoTurnBits();

        this.targetCostsMap = new IntDoubleHashMap(targetIds.length);

        int targetId0 = targetIds[0];
        IntSet targetIdSet = null;
        boolean oneToManyMode = targetIds.length > 1;
        if (oneToManyMode) {
            targetIdSet = new IntSet(targetIds.length * 2);
            for (int targetId : targetIds) targetIdSet.add(targetId);
        }

        // DeadEdge references this Vertex.
        this.wgbs[0] = Integer.MAX_VALUE | 0x80000000;
        // Source without incoming edge. High-DeadEdge.
        this.pqInsertOrUpdate(this.firstVirtualEdgeIdx -1, 0, sourceId);

        while (this.pqSize > 0) {
            this.pqRemove();
            int targetId = this.pqrTargetId;
            this.wgbs[targetId] = this.pqrEdgeIdx | 0x80000000;

            if (oneToManyMode) { // if-else, da One-Check schneller ohne Set
                if (targetIdSet.contains(targetId)) {
                    targetIdSet.remove(targetId);
                    this.targetCostsMap.put(targetId, this.pqrCostR);
                }
                // Jepp!!! Kuerzeste Wege zu allen Zielen gefunden
                if (targetIdSet.getSize() == 0) {
                    return true;
                }

            } else { // OneToOne
                // Jepp!!! Hier ist der kuerzeste Weg.
                if (targetId == targetId0) {
                    this.targetCostsMap.put(targetId, this.pqrCostR);
                    return true;
                }
            }

            int nextSourceId = targetId;

            // Is this a virtual node?
            if (nextSourceId >= this.firstVirtualVertexId) {
                this.expandOverlayGraphVertex(nextSourceId, this.pqrCostR);
                continue;
            }

            int edgeIdx = edgeEntries[nextSourceId];
            int noTurnBit = 1;
            int noTurnBits = 0;

            if (this.pqrEdgeIdx < this.firstVirtualEdgeIdx) {
                noTurnBits = edgeNoTurnBits[this.pqrEdgeIdx];
            } else { // Dynamic Graph
                OverlayEdge s = this.graph.getOverlayEdge(this.pqrEdgeIdx);
                noTurnBits = edgeNoTurnBits[s.getEdgeIdx()];
            }

            do { // Outgoing Edges.
                int pqIdx = edgeIdx;
                double costs = this.calcEdgeCost(edgeIdx);

                OverlayEdge s = this.graph.getEdgeRamp(edgeIdx);
                if (s != null) {
                    targetId = s.getTargetId();
                    costs = this.findShortestPath ? s.getKm() : s.getH();
                    pqIdx = s.getIdx();
                } else {
                    targetId = edgeTargetIds[edgeIdx];
                }

                if (this.wgbs[targetId] >= 0) {
                    boolean mayTurn = true;
                    if (noTurnBits != 0 && !this.ignoreRestrictions) // Abbiegevorschriften
                        mayTurn &= ((noTurnBits & noTurnBit) == 0);
                    if (!this.ignoreOneWays)
                        mayTurn &= (!((edgeIds[edgeIdx] & 0x80000001) == 0x80000000));

                    if (mayTurn) {
                        costs += this.pqrCostR;
                        if (costs < maxCost) {
                            this.pqInsertOrUpdate(pqIdx, costs, targetId);
                        }
                    }
                }
                edgeIdx++;
                noTurnBit <<= 1;
            } while (nextSourceId == edgeSourceIds[edgeIdx]);

            // Sonderlocke: Von realem Vertex abgehende, rein virtuelle Kanten
            OverlayEdge[] virtualEdges = this.graph.getOutgoingVirtualEdges(nextSourceId);
            if (virtualEdges != null) for (OverlayEdge oe : virtualEdges) {
                double costs = this.pqrCostR + (this.findShortestPath ? oe.getKm() : oe.getH());
                this.pqInsertOrUpdate(oe.getIdx(), costs, oe.getTargetId());
            }
        }
        return false;
    }

    private void expandOverlayGraphVertex(int sourceId, double fromCost) {
        OverlayEdge[] outgoingOverlayEdges = this.graph.getOutgoingOverlayEdges(sourceId);
        if (null == outgoingOverlayEdges) return;
        Graph staticGraph = this.graph.getGraph();
        for (int i = 0; i < outgoingOverlayEdges.length; i++) {
            OverlayEdge s = outgoingOverlayEdges[i];
            int targetId = s.getTargetId();
            if (this.wgbs[targetId] >= 0) {

                boolean mayTurn = true;
                if (!this.ignoreOneWays)
                    mayTurn = (!((staticGraph.getEdgeIds()[s.getEdgeIdx()] & 0x80000001) == 0x80000000));

                if (mayTurn) {
                    double cost = fromCost + (this.findShortestPath ? s.getKm() : s.getH());
                    this.pqInsertOrUpdate(s.getIdx(), cost, s.getTargetId());
                }
            }
        }
    }

    protected double calcVertexCost(int vertexId) {
        double h = 0d;
        if (this.heuristicFactor != 0d) { // nur rechnen wenn !=0 (schneller).
            float lat, lon;
            if (vertexId < this.firstVirtualVertexId) {
                lat = this.lats[vertexId];
                lon = this.lons[vertexId];
            } else {
                OverlayNode nd = this.graph.getOverlayNode(vertexId);
                lat = nd.getLat();
                lon = nd.getLon();
            }
            h = GeoUtils.calcDistanceEuklidCoarse(
                    lat, lon, this.targetLat, this.targetLon);
            h *= this.heuristicFactor;
        }
        return h;
    }

    /**
     * Liefert die Kosten einer Kante.
     * Diese Implementation liefert entweder die Distanz oder die Zeit.
     * @param edgeIdx int Index in der Adjazenzliste.
     * @return double Kosten.
     */
    protected double calcEdgeCost(int edgeIdx) {
        return this.edgeCosts[edgeIdx];
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqEdgeIdxs = Arrays.copyOf(this.pqEdgeIdxs, this.pqCapacity);
        this.pqCostsR = Arrays.copyOf(this.pqCostsR, this.pqCapacity);
        this.pqCostsH = Arrays.copyOf(this.pqCostsH, this.pqCapacity);
        this.pqTargetIds = Arrays.copyOf(this.pqTargetIds, this.pqCapacity);
        log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsertOrUpdate(int edgeIdx, double costR, int targetId) {
        int pqPos = this.wgbs[targetId];

        if (0 == pqPos) { // new Target-Node
            pqPos = ++this.pqSize;

            if (this.pqSize >= this.pqCapacity) this.pqGrow();

            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + this.calcVertexCost(targetId);
            this.pqTargetIds[pqPos] = targetId;

            this.wgbs[targetId] = pqPos;

        } else if (this.pqCostsR[pqPos] > costR) { // shorter way found
            double costH = this.pqCostsH[pqPos] - this.pqCostsR[pqPos];
            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + costH;

        } else {
            return;
        }

        this.pqUpheap(pqPos);
    }

    private void pqRemove() {
        this.pqrEdgeIdx = this.pqEdgeIdxs[1];
        this.pqrTargetId = this.pqTargetIds[1];
        this.pqrCostR = this.pqCostsR[1];
        this.pqSwap(1, this.pqSize);
        this.pqSize--;
        this.pqDownheap(1);
    }

    private void pqSwap(int pos1, int pos2) {

        // PQ-Swap
        int edgeIdx1 = this.pqEdgeIdxs[pos1];
        int targetId1 = this.pqTargetIds[pos1];
        double costR1 = this.pqCostsR[pos1];
        double costH1 = this.pqCostsH[pos1];
        int edgeIdx2 = this.pqEdgeIdxs[pos2];
        int targetId2 = this.pqTargetIds[pos2];
        double costR2 = this.pqCostsR[pos2];
        double costH2 = this.pqCostsH[pos2];

        this.pqEdgeIdxs[pos1] = edgeIdx2;
        this.pqTargetIds[pos1] = targetId2;
        this.pqCostsR[pos1] = costR2;
        this.pqCostsH[pos1] = costH2;
        this.pqEdgeIdxs[pos2] = edgeIdx1;
        this.pqTargetIds[pos2] = targetId1;
        this.pqCostsR[pos2] = costR1;
        this.pqCostsH[pos2] = costH1;

        // WGBS-Swap
        this.wgbs[targetId1] = pos2;
        this.wgbs[targetId2] = pos1;
    }

    private void pqUpheap(int pos) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            if (this.pqCostsH[parentPos] < this.pqCostsH[pos]) break;
            this.pqSwap(pos, parentPos);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos) {
        int childPos = pos << 1; // * 2 = left child
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // right child exists?
                if (this.pqCostsH[childRightPos] < this.pqCostsH[childPos]) {
                    childPos = childRightPos;
                }
            }
            if (this.pqCostsH[pos] < this.pqCostsH[childPos]) break;
            this.pqSwap(pos, childPos);
            pos = childPos;
            childPos = pos << 1;
        }
    }


} // EOC
