/*
 * $Id: RoutingResultWriter.java 5258 2021-04-03 16:20:36Z carsten $
 *
 * Copyright 2010 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import static de.cm.osm2po.misc.Code32.encode;
import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static de.cm.osm2po.primitives.VarString.toQuotedJsonBytes;
import static de.cm.osm2po.service.ServiceHelpers.createRRS;

import java.io.OutputStream;
import java.util.Arrays;

import de.cm.osm2po.misc.Utils;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.RandomAccessArray;

/**
 * Schreibt ermittelte Routing-Ergebnisse,
 * als z.B. CSV, GeoJsonBuilder, usw.
 * UTF-8-formatiert in OutputStreams.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po
 */
public class RoutingResultWriter {
    private final static char OBJ_CLOSE = '}';
    private final static char ARR_OPEN = '[';
    private final static char ARR_CLOSE = ']';
    private final static char KOMMA = ',';
    private final static char QUOTE = '\"';
    private final static char NEWLINE = '\n';

    private RoutingResultWriter(){/*STATIC*/}

    /**
     * Macht aus id und LatLon ein CSV-Array der Form
     * id,lon,lat und schreibt es in einen OutputStream.
     * @param os {@link OutputStream}
     * @param id int VertexID
     * @param lon double Laengengrad
     * @param lat double Breitengrad
     * @throws Exception beliebiger Natur
     */
    public static void writeVertexAsCsv(OutputStream os,
            int id, double lon, double lat) throws Exception {
        String csv = id + "," + lat + "," + lon;
        os.write(csv.getBytes());
    }

    /**
     * Macht aus einem {@link OverlayNode} ein CSV-Array der Form
     * Code32(Node),lon,lat und schreibt es in einen OutputStream.
     * @param os {@link OutputStream}
     * @param overlayNode {@link OverlayNode}
     * @throws Exception beliebiger Natur
     */
    public static void writeOverlayNodeAsCsv(
            OutputStream os, OverlayNode overlayNode) throws Exception {
        float lat = overlayNode.getLat();
        float lon = overlayNode.getLon();
        RandomAccessArray raa = new RandomAccessArray();
        overlayNode.writeToStream(raa);
        byte[] bytes = raa.toBytes();
        String csv = '\"' + encode(bytes) + "\"," + lat + "," + lon;
        os.write(csv.getBytes());
    }

    /**
     * Erzeugt aus einer Route ein CSV von paarig angeordneten Floats
     * fuer (Lat, Lon) und schreibt sie in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param path int[] Pfad als Array von EdgeIdx NOT NULL.
     * @throws Exception beliebiger Natur
     */
    public static void writePathAsCsv(
            OutputStream os, Graph graph, int[] path) throws Exception {
        OverlayEdge[] rre = OverlayGraph.translatePath(graph, path);
        writePathAsCsv(os, graph, rre);
    }

    /**
     * Erzeugt aus einer Route ein CSV von paarig angeordneten Floats
     * fuer (Lat, Lon) und schreibt sie in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param path {@link OverlayEdge}[] NOT NULL.
     * @throws Exception beliebiger Natur
     */
    public static void writePathAsCsv(OutputStream os, Graph graph, OverlayEdge[] path) throws Exception {
        boolean isNext = false;
        for (int e = 0; e < path.length; e++) {
            OverlayEdge rre = path[e];
            LatLons geometry = rre.getGeometry();
            if (null == geometry) geometry = graph.lookupGeometry(rre.getEdgeIdx());

            for (int j = 0; j < geometry.getSize(); j++) {
                if (isNext) os.write(KOMMA); else isNext = true;
                long coord = geometry.getCoord(j);
                double lat = latOf(coord); // 1e7
                double lon = lonOf(coord); // 1e7

                os.write(String.valueOf(lat).getBytes());
                os.write(KOMMA);
                os.write(String.valueOf(lon).getBytes());
            }
        }
    }

    /**
     * Liefert die Koordinaten einer Sammlung von VertexIds als CSV
     * und schickt diese auf einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param vids int[] vertexIds
     * @throws Exception beliebiger Natur
     */
    public static void writeVerticesAsCsv(
            OutputStream os, Graph graph, int[] vids) throws Exception {

        float[] lats = graph.getLats();
        float[] lons = graph.getLons();

        for (int i = 0; i < vids.length; i++) {
            if (i > 0) os.write(KOMMA);
            int vertexId = vids[i];
            float lat = lats[vertexId];
            float lon = lons[vertexId];

            os.write(Float.toString(lat).getBytes());
            os.write(KOMMA);
            os.write(Float.toString(lon).getBytes());
        }
    }

    /**
     * Macht aus id und LatLon ein GeoJsonBuilder-Point-Feature
     * mit Geometry und Id.
     * @param os {@link OutputStream}
     * @param id int VertexID
     * @param lon double Laengengrad
     * @param lat double Breitengrad
     * @throws Exception beliebiger Natur
     */
    public static void writeVertexAsFeaturePoint(OutputStream os,
            int id, double lon, double lat) throws Exception {
        os.write("{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[".getBytes());
        os.write((lon + "," + lat).getBytes());
        os.write(("]},\"properties\":{\"id\":" +  id + "}}").getBytes());
    }

    /**
     * Erstellt aus einem {@link OverlayNode} ein GeoJsonBuilder-PointFeature.
     * @param os {@link OutputStream}
     * @param overlayNode {@link OverlayNode}
     * @throws Exception beliebiger Natur
     */
    public static void writeOverlayNodeAsFeaturePoint(
            OutputStream os, OverlayNode overlayNode) throws Exception {
        float lat = overlayNode.getLat();
        float lon = overlayNode.getLon();
        RandomAccessArray raa = new RandomAccessArray();
        overlayNode.writeToStream(raa);
        byte[] bytes = raa.toBytes();

        os.write("{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[".getBytes());
        os.write((lon + "," + lat).getBytes());
        os.write(("]},\"properties\":{\"oln\":\"" + encode(bytes) + "\"}}").getBytes());
    }

    /**
     * Erzeugt aus einer Menge von Vertices eine GeoJsonBuilder-FeatureCollection
     * und schickt diese in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param vids Array von VertexIds
     * @param costs Array von Vertex-Kosten NULLABLE.
     * @throws Exception beliebiger Natur
     */
    public static void writeVerticesAsFeatureCollection (OutputStream os,
            Graph graph, int[] vids, float[] costs) throws Exception {
        float[] lats = graph.getLats();
        float[] lons = graph.getLons();

        final byte[] FEATURE_BEGIN = ("{\"type\":\"Feature\",\"geometry\":{"
            + "\"type\":\"Point\",\"coordinates\":[").getBytes();

        os.write("{\"type\":\"FeatureCollection\",\"features\":[".getBytes());
        for (int i = 0; i < vids.length; i++) {
            int v = vids[i];
            if (i != 0) os.write(KOMMA);
            os.write(NEWLINE);
            os.write(FEATURE_BEGIN);
            os.write((lons[v] + "," + lats[v]).getBytes()); // Dezimeter genau
            os.write(("]},\"properties\":{\"id\":" +  v).getBytes());
            if (costs != null)
                os.write((", \"cost\":" +  costs[i]).getBytes());
            os.write("}}".getBytes()); // FEATURE_END
        }
        os.write("\n]}".getBytes()); // COLLECTION_END
    }

    /**
     * Liefert alle vertices als GeoJsonBuilder-MultiPoint
     * und schreibt diese in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param vids int[] vertexIds
     * @throws Exception beliebiger Natur
     */
    public static void writeVerticesAsMultiPoint(
            OutputStream os, Graph graph, int[] vids) throws Exception {
        float[] lats = graph.getLats();
        float[] lons = graph.getLons();

        os.write("{\"type\":\"MultiPoint\",\"coordinates\":[".getBytes());

        for (int i = 0; i < vids.length; i++) {
            if (i > 0) os.write(KOMMA);
            int vertexId = vids[i];
            float lat = lats[vertexId];
            float lon = lons[vertexId];

            os.write(ARR_OPEN);
            os.write(String.valueOf(lon).getBytes());
            os.write(KOMMA);
            os.write(String.valueOf(lat).getBytes());
            os.write(ARR_CLOSE);
        }
        os.write("]}".getBytes());
    }

    /**
     * Liefert alle vertices als GeoJsonBuilder-(Outer)Polygon
     * und schreibt diese in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param vids int[] vertexIds
     * @throws Exception beliebiger Natur
     */
    public static void writeVerticesAsPolygon(
            OutputStream os, Graph graph, int[] vids) throws Exception {
        float[] lats = graph.getLats();
        float[] lons = graph.getLons();

        os.write("{\"type\":\"Polygon\",\"coordinates\":".getBytes());

        int nVids = vids.length;
        if (nVids > 2 && vids[0] != vids[nVids -1]) {
            vids = Arrays.copyOf(vids, ++nVids);
            vids[nVids -1] = vids[0];
        }

        os.write(ARR_OPEN);
        os.write(ARR_OPEN);
        for (int i = 0; i < nVids; i++) {
            if (i > 0) os.write(KOMMA);
            int vertexId = vids[i];
            float lat = lats[vertexId];
            float lon = lons[vertexId];

            os.write(ARR_OPEN);
            os.write(String.valueOf(lon).getBytes());
            os.write(KOMMA);
            os.write(String.valueOf(lat).getBytes());
            os.write(ARR_CLOSE);
        }
        os.write(ARR_CLOSE);
        os.write(ARR_CLOSE);

        os.write("}".getBytes());
    }

    /**
     * Erzeugt aus dem Pfad eine GeoJsonBuilder-FeatureCollection
     * und schickt diese in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param path {@link OverlayEdge}[]
     * @throws Exception beliebiger Natur
     */
    public static void writePathAsFeatureCollection (OutputStream os,
            Graph graph, int[] path) throws Exception {
        writePathAsFeatureCollection(os,
                graph, OverlayGraph.translatePath(graph, path));
    }

    /**
     * Erzeugt aus dem Pfad eine GeoJsonBuilder-FeatureCollection
     * und schickt diese in einen OutputStream.
     * @param os {@link OutputStream}
     * @param graph {@link Graph}
     * @param path {@link OverlayEdge}[]
     * @throws Exception beliebiger Natur
     */
    public static void writePathAsFeatureCollection (OutputStream os,
            Graph graph, OverlayEdge[] path) throws Exception {

        final byte[] ID = "\"id\":".getBytes();
        final byte[] POS = "\"pos\":".getBytes();
        final byte[] NAME = "\"name\":".getBytes();
        final byte[] FLAGS = "\"flags\":".getBytes();
        final byte[] NOTURNS = "\"noturns\":".getBytes();
        final byte[] LENGTH = "\"length\":".getBytes();
        final byte[] TIME = "\"time\":".getBytes();
        final byte[] SOURCE = "\"source\":".getBytes();
        final byte[] TARGET = "\"target\":".getBytes();
        final byte[] ONEWAY = "\"oneway\":".getBytes();
        final byte[] REVERSE = "\"reverse\":".getBytes();
        final byte[] WRONGWAY = "\"wrongway\":".getBytes();

        final byte[] FEATURECOLLECTION_BEGIN =
            "{\"type\":\"FeatureCollection\",\"features\":[".getBytes();
        final byte[] FEATURE_BEGIN =
            "\n{\"type\":\"Feature\"".getBytes();
        final byte[] GEOMETRY_BEGIN =
            ",\"geometry\":{\"type\":\"LineString\"".getBytes();
        final byte[] COORDINATES_BEGIN = ",\"coordinates\":[".getBytes();
        final byte[] PROPERTIES_BEGIN = ",\"properties\":{".getBytes();

        final byte[] CLOSE = "\n]}".getBytes();

        os.write(FEATURECOLLECTION_BEGIN);
        for (int e = 0; e < path.length; e++) {
            if (e > 0) os.write(KOMMA);

            RoutingResultSegment rrs = createRRS(graph, path[e]);

            LatLon[] latLons = rrs.getLatLons();
            String noTurnBits = Integer.toBinaryString(
                    (rrs.getNoTurnBits() | 0x100) & 0x1FF).substring(1);

            os.write(FEATURE_BEGIN);
                os.write(GEOMETRY_BEGIN);
                    os.write(COORDINATES_BEGIN);
                    for (int i = 0; i < latLons.length; i++) {
                        if (i > 0) os.write(KOMMA);
                        os.write(ARR_OPEN);
                        os.write(String.valueOf(latLons[i].getLon()).getBytes());
                        os.write(KOMMA);
                        os.write(String.valueOf(latLons[i].getLat()).getBytes());
                        os.write(ARR_CLOSE);
                    }
                    os.write(ARR_CLOSE); // Coordinates
                os.write(OBJ_CLOSE); // Geometry

                os.write(PROPERTIES_BEGIN);
                    os.write(ID);
                    os.write(Integer.toString(rrs.getId()).getBytes());

                    os.write(KOMMA);
                    os.write(POS);
                    os.write(Integer.toString(e + 1).getBytes());

                    os.write(KOMMA);
                    os.write(NAME);
                    os.write(toQuotedJsonBytes(rrs.getName(), false));

                    os.write(KOMMA);
                    os.write(FLAGS);
                    os.write(QUOTE);
                    os.write(Utils.toOctetString(rrs.getFlags()).getBytes());
                    os.write(QUOTE);

                    os.write(KOMMA);
                    os.write(NOTURNS);
                    os.write(QUOTE);
                    os.write(noTurnBits.getBytes());
                    os.write(QUOTE);

                    os.write(KOMMA);
                    os.write(LENGTH);
                    os.write(String.valueOf(rrs.getKm()).getBytes());

                    os.write(KOMMA);
                    os.write(TIME);
                    os.write(String.valueOf(rrs.getH()).getBytes());

                    os.write(KOMMA);
                    os.write(SOURCE);
                    os.write(Integer.toString(rrs.getSourceId()).getBytes());

                    os.write(KOMMA);
                    os.write(TARGET);
                    os.write(Integer.toString(rrs.getTargetId()).getBytes());

                    os.write(KOMMA);
                    os.write(ONEWAY);
                    os.write(Boolean.toString(rrs.isOneWay()).getBytes());

                    os.write(KOMMA);
                    os.write(REVERSE);
                    os.write(Boolean.toString(rrs.isReverse()).getBytes());

                    os.write(KOMMA);
                    os.write(WRONGWAY);
                    os.write(Boolean.toString(rrs.isWrongWay()).getBytes());

                 os.write(OBJ_CLOSE); // Properties
           os.write(OBJ_CLOSE); // Feature;
        }
        os.write(CLOSE);
    }
}
