/*
 * $Id: SimpleVertexRouter.java 5466 2021-12-02 20:45:54Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;

/**
 * Minimale OneToOne-Dijkstra-Implemetation eines Routings auf einem statischen Graphen.<br>
 * AStar wird nicht unterstuetzt. Die folgenden Parameter werden hingegen beruecksichtigt:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isResetOnZero()} </li>
 * </ul>
 * @since 5.5.2
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see Graph
 * @see VertexRouter
 */
public class SimpleVertexRouter extends AbstractRouter<Graph>
implements VisitedVerticesAware {

    public SimpleVertexRouter(Graph graph, Log log) {
        super(graph, log);
    }

    private final static int PQ_INITIAL_CAPACITY = 10000; // reicht fuer Europa und mehr.
    private int pqSize; // Aktuelle ArrayGroesse der PQ
    private int pqCapacity; // Maximale ArrayGroesse der PQ
    private int[] pqEdgeIdxs;
    private double[] pqCosts; // Heuristische Gesamtkosten.

    // White-Gray-Black Array. Index ist die TargetId (1-wertig).
    private int[] wgbs;

    // Aktuelle Edge (nach pqRemove)
    private int pqrEdgeIdx;
    private double pqrCost;

    // Kosten werden entweder mit km oder h belegt.
    private float[] edgeCosts;

    // Modi
    private boolean findShortestPath;
    private boolean resetOnZero;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        this.initParams(params);
        this.initPqAndWgb();
        this.settleSourceVertex(sourceId, constraints.maxCost);
        return this.traverseInternal(constraints.targetIds, constraints.maxCost);
    }

    @Override
    public final int[] makePath(int toTargetId) {
        if (!this.isVisited(toTargetId)) return null;

        int[] edgeSourceIds = this.graph.getEdgeSourceIds();

        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        int targetId = toTargetId;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY);
            int edgeIdx = this.wgbs[targetId] & 0x7FFFFFFF; // Bit 32 (Visited-Black-Flag).
            if (edgeIdx == Integer.MAX_VALUE) break; // -1 hat keinen Vorgaengerweg. -1 wg. so.
            int sourceId = edgeSourceIds[edgeIdx];
            pathIds[n++] = edgeIdx;
            targetId = sourceId;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        // Reverse Path from last found to first
        for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    @Override
    public final int[] getVisited() {
        int len = this.wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) { // wgbs ist 1-wertig.
            int edgeIdx = this.wgbs[vertexId];
            if (edgeIdx < 0) { // Kante zum ZielVertex wurde besucht.
                if (buffer.length < vertexId) // Ueberlauf korrigieren
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    @Override
    public boolean isVisited(int vertexId) {
        if (vertexId < 1 || vertexId > this.wgbs.length -1) return false;
        return this.wgbs[vertexId] < 0;
    }

    private void initPqAndWgb() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Maximale ArrayGroesse der PQ
        this.pqEdgeIdxs = new int[this.pqCapacity];
        this.pqCosts = new double[this.pqCapacity]; // Reelle Weg- oder Zeit-Kosten.

        int nVertices = this.graph.getEdgeEntries().length;
        this.wgbs = new int[nVertices];

        // Die DeadEdge (alle Member = 0) verweist auf diesen Vertex.
        // Daher markieren wir ihn als bereits besucht.
        this.wgbs[0] = Integer.MAX_VALUE | 0x80000000;
    }


    private void initParams(RouterParameters params) {
        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.resetOnZero = params.isResetOnZero();

        // Kosten entweder fuer Shortest oder Fastest
        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
        }
    }

    private void settleSourceVertex(int sourceId, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();

        // Kein Vorgaengerweg. (all Bits Set = -1)
        this.wgbs[sourceId] = Integer.MAX_VALUE | 0x80000000;

        // Erste abgehende Kanten anders behandeln, wir sparen
        // uns so die Haltung der TargetId in der PQ selbst.
        int edgeIdx = edgeEntries[sourceId];

        while (edgeSourceIds[edgeIdx] == sourceId) {
            int targetId = edgeTargetIds[edgeIdx];

            // Schon hier pruefen, da Start auch eine
            // Loop (source=target) sein kann.
            if (this.wgbs[targetId] >= 0d) {
                boolean isWrongWay = (edgeIds[edgeIdx] & 0x80000001) == 0x80000000;
                double cost = this.calcEdgeCost(edgeIdx, 0, isWrongWay, false);
                if (cost >= 0 && cost < maxCost) {
                    this.pqInsertOrUpdate(edgeIdx, cost, targetId, edgeTargetIds);
                }
            }
            edgeIdx++;
        }
    }

    private boolean traverseInternal(int[] targetIds, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        byte[] edgeNoTurnBits = this.graph.getEdgeNoTurnBits();

        int targetId0 = targetIds[0];

        while (this.pqSize > 0) {
            this.pqRemove(edgeTargetIds);
            int fromEdgeIdx = this.pqrEdgeIdx;
            int targetId = edgeTargetIds[fromEdgeIdx];

            this.wgbs[targetId] = fromEdgeIdx | 0x80000000;

            // Jepp!!! Hier ist der kuerzeste Weg.
            if (targetId == targetId0) {
                return true;
            }

            // Anschlusssegmente suchen.
            int sourceId = targetId; // Rollenwechsel
            int edgeIdx = edgeEntries[sourceId];
            int noTurnBits = edgeNoTurnBits[fromEdgeIdx];
            int noTurnBit = 1;

            do { // Anschlusssegmente durchlaufen.
                targetId = edgeTargetIds[edgeIdx];
                // Jetzt pruefen, ob TargetVertex noch in der Pipeline ist.
                if (this.wgbs[targetId] >= 0) {
                    boolean isTurnRestriced = (noTurnBits & noTurnBit) != 0;
                    boolean isWrongWay = (edgeIds[edgeIdx] & 0x80000001) == 0x80000000;
                    double cost = this.calcEdgeCost(edgeIdx, fromEdgeIdx, isWrongWay, isTurnRestriced);
                    if (cost >= 0d) { // Edge ignorieren bei negativen Kosten
                        if (cost > 0 || !this.resetOnZero) // oder bei 0 Kosten resetten
                            cost += this.pqrCost; // Reelle Kosten bis hier
                        if (cost < maxCost)
                            this.pqInsertOrUpdate(edgeIdx, cost, targetId, edgeTargetIds);
                    }
                }
                edgeIdx++;
                noTurnBit <<= 1;
            } while (sourceId == edgeSourceIds[edgeIdx]);
        }
        return false;
    }


    /**
     * Liefert die Kosten einer Kante.
     * Diese Implementation liefert entweder die Distanz oder die Zeit.
     * Der Verstoss gegen die Einbahnstrasse oder Abbiegevorschrift wird
     * mit -1 bestraft bzw. zurueckgeliefert.
     * @param edgeIdx int Index der Kante ueber die wir fahren.
     * @param fromEdgeIdx int Index der Kante, von der wir kommen.
     * @param isWrongWay true: Versuch, gegen die Einbahnstrasse zu fahren
     * @param isTurnRestriced true: Versuch, gegen Abbiegevorschrift zu fahren
     * @return double Kosten.
     * <br><br><b>Wichtig:</b><br>
     * Liefert diese Methode negative Kosten, so wird die Kante
     * gesperrt, sind diese 0, so werden die bisher aufgelaufenen
     * Kosten ignoriert, was eine Bevorzugung dieser Kante erzwingt.
     */
    protected double calcEdgeCost(
            int edgeIdx, int fromEdgeIdx, boolean isWrongWay, boolean isTurnRestriced) {
        if (isWrongWay || isTurnRestriced) return -1;
        else return this.edgeCosts[edgeIdx];
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqEdgeIdxs = Arrays.copyOf(this.pqEdgeIdxs, this.pqCapacity);
        this.pqCosts = Arrays.copyOf(this.pqCosts, this.pqCapacity);
        this.log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsertOrUpdate(
            int edgeIdx, double cost, int targetId, int[] edgeTargetIds) {
        int pqPos = this.wgbs[targetId];

        if (0 == pqPos) { // neuer Ziel-Knoten
            pqPos = ++this.pqSize;

            if (this.pqSize >= this.pqCapacity) this.pqGrow();

            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCosts[pqPos] = cost;
            this.wgbs[targetId] = pqPos;

        } else if (this.pqCosts[pqPos] > cost) { // kuerzeren Weg gefunden
            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCosts[pqPos] = cost;

        } else {
            return;
        }

        this.pqUpheap(pqPos, edgeTargetIds);
    }

    private void pqRemove(int[] edgeTargetIds) {
        this.pqrEdgeIdx = this.pqEdgeIdxs[1];
        this.pqrCost = this.pqCosts[1];
        this.pqSwap(1, this.pqSize, edgeTargetIds);
        this.pqSize--;
        this.pqDownheap(1, edgeTargetIds);
    }

    private void pqSwap(int pos1, int pos2, int[] edgeTargetIds) {
        // PQ-Swap
        int edgeIdx1 = this.pqEdgeIdxs[pos1];
        double cost1 = this.pqCosts[pos1];
        int edgeIdx2 = this.pqEdgeIdxs[pos2];
        double cost2 = this.pqCosts[pos2];

        this.pqEdgeIdxs[pos1] = edgeIdx2;
        this.pqCosts[pos1] = cost2;
        this.pqEdgeIdxs[pos2] = edgeIdx1;
        this.pqCosts[pos2] = cost1;

        // WGBS-Swap
        int targetId1 = edgeTargetIds[edgeIdx1];
        int targetId2 = edgeTargetIds[edgeIdx2];

        this.wgbs[targetId1] = pos2;
        this.wgbs[targetId2] = pos1;
    }

    private void pqUpheap(int pos, int[] edgeTargetIds) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCosts[parentPos] < this.pqCosts[pos]) break;
            this.pqSwap(pos, parentPos, edgeTargetIds);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos, int[] edgeTargetIds) {
        int childPos = pos << 1; // * 2 = linkes Kind
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqCosts[childRightPos] < this.pqCosts[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCosts[pos] < this.pqCosts[childPos]) break;
            this.pqSwap(pos, childPos, edgeTargetIds);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
