/*
 * $Id: VertexRouter.java 5547 2022-02-25 17:50:27Z carsten $
 *
 * Copyright 2021 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.routing;

import java.util.Arrays;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoUtils;
import de.cm.osm2po.primitives.IntDoubleHashMap;
import de.cm.osm2po.primitives.IntSet;

/**
 * Dijkstra-Implemetation eines Routings auf einem statischen Graphen.<br>
 * Per Konfiguration sind sowohl AStar, ReverseRouting als auch OneToMany abbildbar.
 * Eine nachtraegliche Abfrage der Kosten aller Targets ist ebenfalls moeglich.
 * <br><br><b>Hinweis:</b><br>
 * Diese Implementation vereint beinahe alle Moeglichkeiten, die ein Dijkstra unter
 * osm2po liefern kann und ist daher nicht optimal im Sinne der Performanz.
 * <br><br>
 * Dieser Router beruecksichtigt die folgenden Parameter:
 * <ul>
 * <li> {@link RouterParameters#isFindShortestPath()} </li>
 * <li> {@link RouterParameters#isIgnoreOneWays()} </li>
 * <li> {@link RouterParameters#isIgnoreRestrictions()} </li>
 * <li> {@link RouterParameters#getHeuristicFactor()} </li>
 * <li> {@link RouterParameters#isResetOnZero()} </li>
 * <li> {@link RouterParameters#isFindReverse()} </li>
 * </ul>
 * <br><br><b>Wichtig:</b><br>
 * Dieser Router kann auch in den Reverse-Modus geschaltet werden.
 * Da hier auch ein OneToMany unterstuetzt wird, erhalten wir
 * darueber so automatisch ein ManyToOne.<br>
 * Aber aufgepasst: {@link #makePath(int)} liefert danach
 * die Route in der korrekten Richtung, also vom Target zum Source
 * zurueck, was die Darstellung der Geometrie befriedigt, jedoch
 * genau umgekehrt zur eigentlichen Abfrage, also vom Start zum
 * Ziel ist.
 * @since 5.4.0
 * @author (c) 2012/2021 - Carsten Moeller - info@osm2po
 * @see Graph
 * @see OverlayVertexRouter
 * @see VisitedVerticesAware
 * @see TargetCostsAware
 */
public class VertexRouter extends AbstractRouter<Graph>
implements VisitedVerticesAware, FoundTargetsAware, TargetCostsAware {
    /*
     * Intern wird komplett auf Objekte (verbundene Datentypen) verzichtet.
     * Stattdessen werden Listen von Objekten, z.B. die der Edges,
     * pro Attribut auf einzelnde Arrays aufgeteilt.
     * Das bringt einen enormen Geschwindigkeitsgewinn, gerade unter Java.<br>
     * Unter C++ waere dies unnoetig, da hier Struct-Maskerading moeglich waere.<br>
     * Der {@link Graph} liefert aus Speichergruenden fuer die Kosten pro Segments
     * kleine <code>float</code>s. Dies ist hinreichend genau
     * fuer kurze Segmente (10mm genau auf ~1km), jedoch nicht fuer eine Summe von
     * Segmenten (1m genau auf ~1000km).
     * Aus diesem Grund rechnet die PriorityQueue intern mit <code>double</code>s.
     * Koordinaten fuer die AStar-Heuristik sind noch ungenauer (~10cm).
     * Dies ist jedoch zu vernachlaessigen, da es sich lediglich um eine
     * Heuristik pro Punkt handelt und hier nicht aufsummiert wird.
     * Der Fehler bleibt hier konstant klein. Abschliessend muss jedoch erwaehnt
     * sein, dass <code>float</code> schneller als <code>double</code> zu berechnen ist,
     * Java allerdings kaum <code>float</code>-Funktionen bereitstellt.
     */

    public VertexRouter(Graph graph, Log log) {
        super(graph, log);
    }

    private final static int PQ_INITIAL_CAPACITY = 10000; // reicht fuer Europa und mehr.
    private int pqSize; // Aktuelle ArrayGroesse der PQ
    private int pqCapacity; // Maximale ArrayGroesse der PQ
    private int[] pqEdgeIdxs;
    private double[] pqCostsR; // Reelle Weg- oder Zeit-Kosten.
    private double[] pqCostsH; // Heuristische Gesamtkosten.

    // *1) White-Gray-Black Array. Index ist die TargetId (1-wertig).
    //     Graue Knoten setzen positive PQ-Array-Positionen (pq arbeitet 1-wertig).
    //     Schwarze Knoten setzen EdgeIdx |= 0x80000000. Fuer EdgeIdx == 0 liefert
    //     diese Veroderung Integer.MinValue, EdgeId == 1 also MinValue + 1, usw.
    //     Als Sentinel fuer den ersten Knoten (ohne Vorgaengerweg) wird -1 verwendet,
    //     also 32 gesetzte Bits, was einen maximalen Abstand ergibt.
    //     Dies ist die Abbruchbedingung bei der Erstellung in Rueckwaertsrichtung.
    //     Es werden also niemals 0-Werte gesetzt. Daher ist 0 immer initial.
    private int[] wgbs; // siehe *1)

    // Aktuelle Edge (nach pqRemove)
    private int pqrEdgeIdx;
    private double pqrCostR;

    // Kosten werden entweder mit km oder h belegt.
    private float[] edgeCosts;

    // Members fuer Heuristiken
    private float[] lats, lons; // Alle Koordinaten.
    private float targetLat, targetLon; // Zielkoordinate.
    private double heuristicFactor = 0d; // Bewertungsfaktor 0 Dijkstra, 1 AStar

    // Modi
    private boolean findShortestPath;
    private boolean resetOnZero;
    private boolean findReverse;
    private boolean ignoreRestrictions;
    private boolean ignoreOneWays;

    // Target-Kosten fuer OneToMany- aber auch OneToOne-Mode
    // TODO Target-Kosten ohne Map loesen
    private IntDoubleHashMap targetCostsMap;

    @Override
    public boolean traverse(int sourceId, RouterConstraints constraints, RouterParameters params) {
        if (null == constraints) constraints = new RouterConstraints();
        this.initParams(params);
        this.initAStar(constraints.targetIds);
        this.initPqAndWgb();
        this.settleSourceVertex(sourceId, constraints.maxCost);
        return this.traverseInternal(constraints.targetIds, constraints.maxCost);
    }

    @Override
    public final int[] makePath(int toTargetId) {
        if (!this.isVisited(toTargetId)) return null;

        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        if (this.findReverse) edgeSourceIds = this.graph.getEdgeTargetIds();

        final int INITIAL_CAPACITY = 1000;
        int[] pathIds = new int[INITIAL_CAPACITY];
        int n = 0;
        int targetId = toTargetId;
        for (;;) {
            if (pathIds.length == n) pathIds = Arrays.copyOf(
                    pathIds, pathIds.length + INITIAL_CAPACITY);
            int edgeIdx = this.wgbs[targetId] & 0x7FFFFFFF; // Bit 32 (Visited-Black-Flag).
            if (edgeIdx == Integer.MAX_VALUE) break; // -1 hat keinen Vorgaengerweg. -1 wg. so.
            int sourceId = edgeSourceIds[edgeIdx];
            pathIds[n++] = edgeIdx;
            targetId = sourceId;
        }
        pathIds = Arrays.copyOf(pathIds, n);

        // Reverse Path from last found to first
        if (!this.findReverse) for (int i = 0, j = n - 1; i < j; i++, j--) {
            int z = pathIds[i]; pathIds[i] = pathIds[j]; pathIds[j] = z;
        }
        return pathIds;
    }

    @Override
    public final int[] getVisited() {
        int len = this.wgbs.length;
        int[] buffer = new int[len / 2];
        int n = 0;
        for (int vertexId = 1; vertexId < len; vertexId++) { // wgbs ist 1-wertig.
            int edgeIdx = this.wgbs[vertexId];
            if (edgeIdx < 0) { // Kante zum ZielVertex wurde besucht.
                if (buffer.length < vertexId) // Ueberlauf korrigieren
                    buffer = Arrays.copyOf(buffer, (int) (buffer.length * 1.5));
                buffer[n++] = vertexId;
            }
        }
        return Arrays.copyOf(buffer, n);
    }

    @Override
    public boolean isVisited(int vertexId) {
        if (vertexId < 1 || vertexId > this.wgbs.length -1) return false;
        return this.wgbs[vertexId] < 0;
    }

    @Override
    public double getTargetCost(int vertexId) {
        double cost = this.targetCostsMap.get(vertexId);
        if (IntDoubleHashMap.NULLVALUE == cost) return Double.NaN;
        return cost;
    }

    @Override
    public int[] getVisitedTargetIds() {
        return this.targetCostsMap.getKeys();
    }

    private void initPqAndWgb() {
        this.pqSize = 0;
        this.pqCapacity = PQ_INITIAL_CAPACITY; // Maximale ArrayGroesse der PQ
        this.pqEdgeIdxs = new int[this.pqCapacity];
        this.pqCostsR = new double[this.pqCapacity]; // Reelle Weg- oder Zeit-Kosten.
        this.pqCostsH = new double[this.pqCapacity]; // Heuristische Gesamtkosten.

        int nVertices = this.graph.getEdgeEntries().length;
        this.wgbs = new int[nVertices];

        // Die DeadEdge (alle Member = 0) verweist auf diesen Vertex.
        // Daher markieren wir ihn als bereits besucht.
        this.wgbs[0] = Integer.MAX_VALUE | 0x80000000;
    }


    private void initParams(RouterParameters params) {
        if (null == params) params = new RouterParameters();

        this.findShortestPath = params.isFindShortestPath();
        this.findReverse = params.isFindReverse();
        this.resetOnZero = params.isResetOnZero();
        this.heuristicFactor = params.getHeuristicFactor();
        this.ignoreRestrictions = params.isIgnoreRestrictions();
        this.ignoreOneWays = params.isIgnoreOneWays();

        if (this.findReverse && !this.graph.isSupported(Graph.SUPPORT_REVERSE))
            throw new IllegalStateException("Graph does not support reverse routing");

        // Kosten entweder fuer Shortest oder Fastest
        if (this.findShortestPath) {
            this.edgeCosts = this.graph.getEdgeCostsKm();
        } else {
            this.edgeCosts = this.graph.getEdgeCostsH();
            // Wenn Fastest, dann muss die Toleranz vergroessert werden,
            // damit die AStar-Bedingung eingehalten wird.
            // Bedingung: Len(A) + Heuristic(A) <= Len(A) + Len(B) + Heuristic(B)
            // Hier 110 kmh: D.h. zwei aufeinanderfolgende Wege duerfen
            // in ihren Geschwindigkeiten maximal um 110 kmh differieren.
            this.heuristicFactor /= 110;
        }
    }

    private void initAStar(int[] targetIds) {
        if (this.heuristicFactor != 0) {
            if (!this.graph.isSupported(Graph.SUPPORT_LATLON)) {
                this.heuristicFactor = 0; // Kein AStar
                log.debug("AStar (heuristicFactor) disabled. Graph does not support coordinates");
                return;
            }

            if (targetIds.length > 1) {
                this.heuristicFactor = 0; // kein AStar
                log.debug("AStar (heuristicFactor) disabled in OneToMany-Mode");
                return;
            }

            this.lats = this.graph.getLats();
            this.lons = this.graph.getLons();
            this.targetLat = this.lats[targetIds[0]];
            this.targetLon = this.lons[targetIds[0]];
        }
    }

    private void settleSourceVertex(int sourceId, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        int[] edgeEntriesR = this.graph.getEdgeEntriesR();
        int[] edgeIdxsR = this.graph.getEdgeIdxsR();

        if (this.findReverse) {
            edgeSourceIds = this.graph.getEdgeTargetIds();
            edgeTargetIds = this.graph.getEdgeSourceIds();
        }

        // Kein Vorgaengerweg. (all Bits Set = -1)
        this.wgbs[sourceId] = Integer.MAX_VALUE | 0x80000000;

        // Erste abgehende Kanten anders behandeln, wir sparen
        // uns so die Haltung der TargetId in der PQ selbst.
        int edgeIdx, edgeIdxIdx = 0;
        if (this.findReverse) {
            edgeIdxIdx = edgeEntriesR[sourceId];
            edgeIdx = edgeIdxsR[edgeIdxIdx]; // echte edgeIdx
        } else {
            edgeIdx = edgeEntries[sourceId];
        }

        while (edgeSourceIds[edgeIdx] == sourceId) {
            int targetId = edgeTargetIds[edgeIdx];

            // Schon hier pruefen, da Start auch eine
            // Loop (source=target) sein kann.
            if (this.wgbs[targetId] >= 0d) {
                boolean isWrongWay = (edgeIds[edgeIdx] & 0x80000001) == 0x80000000;
                if (!isWrongWay || this.ignoreOneWays) {
                    double cost = this.calcEdgeCost(edgeIdx, 0);
                    if (cost >= 0 && cost < maxCost) {
                        this.pqInsertOrUpdate(edgeIdx, cost, targetId, edgeTargetIds);
                    }
                }
            }
            if (this.findReverse) edgeIdx = edgeIdxsR[++edgeIdxIdx]; else edgeIdx++;
        }
    }

    private boolean traverseInternal(int[] targetIds, double maxCost) {
        int[] edgeEntries = this.graph.getEdgeEntries();
        int[] edgeSourceIds = this.graph.getEdgeSourceIds();
        int[] edgeTargetIds = this.graph.getEdgeTargetIds();
        int[] edgeIds = this.graph.getEdgeIds();
        byte[] edgeNoTurnBits = this.graph.getEdgeNoTurnBits();

        int[] edgeEntriesR = this.graph.getEdgeEntriesR();
        int[] edgeIdxsR = this.graph.getEdgeIdxsR();
        if (this.findReverse) {
            edgeSourceIds = this.graph.getEdgeTargetIds();
            edgeTargetIds = this.graph.getEdgeSourceIds();
            edgeNoTurnBits = this.graph.getEdgeNoTurnBitsR();
        }

        this.targetCostsMap = new IntDoubleHashMap(targetIds.length);

        int targetId0 = targetIds[0];
        IntSet targetIdSet = null;
        boolean oneToManyMode = targetIds.length > 1;
        if (oneToManyMode) {
            targetIdSet = new IntSet(targetIds.length * 2);
            for (int targetId : targetIds) targetIdSet.add(targetId);
        }

        while (this.pqSize > 0) {
            this.pqRemove(edgeTargetIds);
            int fromEdgeIdx = this.pqrEdgeIdx;
            int targetId = edgeTargetIds[fromEdgeIdx];

            this.wgbs[targetId] = fromEdgeIdx | 0x80000000;

            if (oneToManyMode) { // if-else, da One-Check schneller ohne Set
                if (targetIdSet.contains(targetId)) {
                    targetIdSet.remove(targetId);
                    this.targetCostsMap.put(targetId, this.pqrCostR);
                }
                // Jepp!!! Kuerzeste Wege zu allen Zielen gefunden
                if (targetIdSet.getSize() == 0) {
                    return true;
                }

            } else { // OneToOne
                // Jepp!!! Hier ist der kuerzeste Weg.
                if (targetId == targetId0) {
                    this.targetCostsMap.put(targetId, this.pqrCostR);
                    return true;
                }
            }

            // Anschlusssegmente suchen.
            int sourceId = targetId; // Rollenwechsel

            int edgeIdx, edgeIdxIdx = 0;
            if (this.findReverse) {
                edgeIdxIdx = edgeEntriesR[sourceId];
                edgeIdx = edgeIdxsR[edgeIdxIdx]; // echte edgeIdx
            } else {
                edgeIdx = edgeEntries[sourceId];
            }

            int noTurnBits = edgeNoTurnBits[fromEdgeIdx];
            int noTurnBit = 1;

            do { // Anschlusssegmente durchlaufen.
                targetId = edgeTargetIds[edgeIdx];
                // Jetzt pruefen, ob TargetVertex noch in der Pipeline ist.
                if (this.wgbs[targetId] >= 0) {

                    boolean mayTurn = true;

                    if (noTurnBits != 0 && !this.ignoreRestrictions) // Abbiegevorschriften
                        mayTurn &= ((noTurnBits & noTurnBit) == 0);
                    if (!this.ignoreOneWays) // Einbahnstrassen beachten
                        mayTurn &= (!((edgeIds[edgeIdx] & 0x80000001) == 0x80000000));

                    if (mayTurn) {
                        double cost = this.calcEdgeCost(edgeIdx, fromEdgeIdx);
                        if (cost >= 0d) { // Edge ignorieren bei negativen Kosten
                            if (cost > 0 || !this.resetOnZero) // oder bei 0 Kosten resetten
                                cost += this.pqrCostR; // Reelle Kosten bis hier
                            if (cost < maxCost)
                                this.pqInsertOrUpdate(edgeIdx, cost, targetId, edgeTargetIds);
                        }
                    }
                }
                if (this.findReverse) edgeIdx = edgeIdxsR[++edgeIdxIdx]; else edgeIdx++;
                noTurnBit <<= 1;
            } while (sourceId == edgeSourceIds[edgeIdx]);
        }
        return false;
    }


    /**
     * Diese Implementation liefert entweder 0 fuer Dijkstra
     * oder die heuristische Distanz zum Ziel fuer AStar.
     * Eine abgeleitete Klasse koennte an dieser Stelle z.B.
     * auch noch konstante Kosten einer Vertex-Klasse hinzuaddieren.
     * So z.B. fuer Ampeln, Mautstationen oder Bahnuebergaenge.
     * @param vertexId Id des Vertex.
     * @return double Kosten.
     */
    protected double calcVertexCost(int vertexId) {
        double heuristicCosts = 0d;
        if (this.heuristicFactor != 0d) { // nur rechnen wenn !=0 (schneller).
            heuristicCosts = GeoUtils.calcDistanceEuklidCoarse(
                    this.lats[vertexId], this.lons[vertexId],
                    this.targetLat, this.targetLon);
            heuristicCosts *= this.heuristicFactor;
        }
        return heuristicCosts;
    }

    /**
     * Liefert die Kosten einer Kante.
     * Diese Implementation liefert entweder die Distanz oder die Zeit.
     * @param edgeIdx int Index der Kante ueber die wir fahren.
     * @param fromEdgeIdx int Index der Kante, von der wir kommen.
     * @return double Kosten.
     * <br><br><b>Wichtig:</b><br>
     * Liefert diese Methode negative Kosten, so wird die Kante
     * gesperrt, sind diese 0, so werden die bisher aufgelaufenen
     * Kosten ignoriert, was eine Bevorzugung dieser Kante erzwingt.
     */
    protected double calcEdgeCost(int edgeIdx, int fromEdgeIdx) {
        return this.edgeCosts[edgeIdx];
    }

    /*
     * ==================== PriorityQueue ==============================
     */

    private void pqGrow() {
        this.pqCapacity += PQ_INITIAL_CAPACITY;
        this.pqEdgeIdxs = Arrays.copyOf(this.pqEdgeIdxs, this.pqCapacity);
        this.pqCostsR = Arrays.copyOf(this.pqCostsR, this.pqCapacity);
        this.pqCostsH = Arrays.copyOf(this.pqCostsH, this.pqCapacity);
        this.log.debug("Priority queue capacity increased to " + this.pqCapacity);
    }

    private void pqInsertOrUpdate(int edgeIdx, double costR, int targetId, int[] edgeTargetIds) {
        int pqPos = this.wgbs[targetId];

        if (0 == pqPos) { // neuer Ziel-Knoten
            pqPos = ++this.pqSize;

            if (this.pqSize >= this.pqCapacity) this.pqGrow();

            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + this.calcVertexCost(targetId);

            this.wgbs[targetId] = pqPos;

        } else if (this.pqCostsR[pqPos] > costR) { // kuerzeren Weg gefunden
            double costH = this.pqCostsH[pqPos] - this.pqCostsR[pqPos];
            this.pqEdgeIdxs[pqPos] = edgeIdx;
            this.pqCostsR[pqPos] = costR;
            this.pqCostsH[pqPos] = costR + costH;

        } else {
            return;
        }

        this.pqUpheap(pqPos, edgeTargetIds);
    }

    private void pqRemove(int[] edgeTargetIds) {
        this.pqrEdgeIdx = this.pqEdgeIdxs[1];
        this.pqrCostR = this.pqCostsR[1];
        this.pqSwap(1, this.pqSize, edgeTargetIds);
        this.pqSize--;
        this.pqDownheap(1, edgeTargetIds);
    }

    private void pqSwap(int pos1, int pos2, int[] edgeTargetIds) {
        // PQ-Swap
        int edgeIdx1 = this.pqEdgeIdxs[pos1];
        double costR1 = this.pqCostsR[pos1];
        double costH1 = this.pqCostsH[pos1];
        int edgeIdx2 = this.pqEdgeIdxs[pos2];
        double costR2 = this.pqCostsR[pos2];
        double costH2 = this.pqCostsH[pos2];

        this.pqEdgeIdxs[pos1] = edgeIdx2;
        this.pqCostsR[pos1] = costR2;
        this.pqCostsH[pos1] = costH2;
        this.pqEdgeIdxs[pos2] = edgeIdx1;
        this.pqCostsR[pos2] = costR1;
        this.pqCostsH[pos2] = costH1;

        // WGBS-Swap
        int targetId1 = edgeTargetIds[edgeIdx1];
        int targetId2 = edgeTargetIds[edgeIdx2];

        this.wgbs[targetId1] = pos2;
        this.wgbs[targetId2] = pos1;
    }

    private void pqUpheap(int pos, int[] edgeTargetIds) {
        while (pos != 1) {
            int parentPos = pos >> 1;
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[parentPos] < this.pqCostsH[pos]) break;
            this.pqSwap(pos, parentPos, edgeTargetIds);
            pos = parentPos;
        }
    }

    private void pqDownheap(int pos, int[] edgeTargetIds) {
        int childPos = pos << 1; // * 2 = linkes Kind
        while (childPos <= this.pqSize) {
            int childRightPos = childPos +1;
            if (childRightPos <= this.pqSize) { // rechtes Kind vorhanden?
                // hat rechtes Kind Prioritaet?
                if (this.pqCostsH[childRightPos] < this.pqCostsH[childPos]) {
                    childPos = childRightPos;
                }
            }
            // Ist PQ-Bedingung erfuellt? Dann fertig.
            if (this.pqCostsH[pos] < this.pqCostsH[childPos]) break;
            this.pqSwap(pos, childPos, edgeTargetIds);
            pos = childPos;
            childPos = pos << 1;
        }
    }

} // EOC
