/*
 * $Id: Osm2poService.java 5856 2024-02-24 10:48:09Z carsten $
 *
 * Copyright 2011/2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;
import static de.cm.osm2po.misc.Utils.MB;
import static de.cm.osm2po.misc.Utils.TS;
import static de.cm.osm2po.misc.Utils.UTF8;
import static de.cm.osm2po.misc.Utils.findFileOrDir;
import static de.cm.osm2po.misc.Utils.loadFileIntoByteArray;
import static de.cm.osm2po.misc.Utils.loadStreamIntoByteArray;
import static de.cm.osm2po.misc.Utils.readArgsIntoProps;
import static de.cm.osm2po.service.ServiceHandler.MIME_JSON;
import static de.cm.osm2po.service.ServiceHandler.MIME_TEXT;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_DOWN_MAINTENANCE;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_HTTP_METHOD_NOT_ALLLOWED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_PARAMETER_INVALID;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TOO_MANY_HEAVY_REQUESTS;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNAUTHORIZED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_EXCEPTION;
import static de.cm.osm2po.service.security.SecurityHelpers.createSSLContextTLS;
import static java.util.concurrent.Executors.newFixedThreadPool;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;

import com.sun.net.httpserver.Authenticator;
import com.sun.net.httpserver.Headers;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import com.sun.net.httpserver.HttpServer;
import com.sun.net.httpserver.HttpsConfigurator;
import com.sun.net.httpserver.HttpsServer;

import de.cm.osm2po.Config;
import de.cm.osm2po.Version;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.reflection.JsonBuilder;
import de.cm.osm2po.routing.AbstractRouter;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.service.errors.LangEnum;
import de.cm.osm2po.service.errors.ServiceException;
import de.cm.osm2po.service.handlers.DebugHandler;
import de.cm.osm2po.service.handlers.GeoHandler;
import de.cm.osm2po.service.handlers.PositionHandler;
import de.cm.osm2po.service.handlers.RouteHandler;
import de.cm.osm2po.service.handlers.StateHandler;
import de.cm.osm2po.service.handlers.TspHandler;
import de.cm.osm2po.service.security.SecurityHelpers;

public class Osm2poService implements HttpHandler, ServiceResourcesChecker {
    private HttpServer httpServer;
    private HttpsServer httpsServer;
    private Config config;
    private Graph graph;
    private Log log;
    private File webDir;
    private Map<String, String> mimeTypeMap = new HashMap<String, String>();
    private Map<String, String> cmdMappingMap = new HashMap<String, String>();
    private List<ServiceHandler> handlerList = new ArrayList<ServiceHandler>();
    private List<ServiceFilter> filterList = new ArrayList<ServiceFilter>();
    private Set<Long> heavyThreadIdSet = new HashSet<Long>();
    private int nMaxHeavyThreads;
    private ServiceResponse indexHtmlResponse;
    private ServiceHandlerProxy handlerProxy;
    private boolean offline;
    private boolean errorsAsText;

    private String httpUrl = null; // 4UnitTest

    // Puplics for Tests and more
    public final String getHttpUrl() {return this.httpUrl;}
    public final Graph getGraph() {return this.graph;}
    public final File getWebDir() {return this.webDir;}
    public final Log getLog() {return this.log;}
    public final Config getConfig() {return this.config;}

    public final static String getHeaderValue(
            final Headers headers, final String name) {
        List<String> lh = headers.get(name);
        if (lh != null && lh.size() > 0) return lh.get(0);
        else return null;
    }

    /**
     * Macht aus '//http:///a/////b///c/' 'http://a/b/c' oder aus '/a//b/' 'a/b'.
     * @param s String Pfad
     * @return evtl korrigierter Pfad
     */
    private static String cleanPath(final String s) {
        if (null == s) return null;
        char[] carr = s.toCharArray();
        int x = 0, y = carr.length -1;
        while (x <= y && '/' == carr[x]) x++;
        while (y >= x && '/' == carr[y]) y--;
        int n = 0;
        for (int z = x; z <= y; z++) {
            if (z > x && '/' == carr[z] && '/' == carr[z - 1]) { // //
                // Wenn nicht sowas wie http:// oder ftp:// dann / loeschen
                if (!(z > x + 1 && ':' == carr[z - 2])) continue;
            }
            carr[n++] = carr[z];
        }
        return new String(carr, 0, n);
    }

    // WICHTIG! Https-Geschichten haben Laufzeit-Probleme mit Java 1.6
	public final void startService(Config config) {
	    this.offline = Boolean.valueOf(config.getProperty("service.offline"));
        this.errorsAsText = Boolean.valueOf(config.getProperty("service.errors.astext"));

	    if (!this.offline) {
	        // Wichtig: findGraph ausserhalb try/catch
            File graphFile = config.findGraphFile(true);

            if (null == graphFile) throw new IllegalStateException(
                    "Cannot start " + getClass().getSimpleName() + " without Graph.\n"
                    + "USAGE EXAMPLE:\njava -jar osm2po.jar cmd=r path/to/graphFile.gph");
            this.graph = openGraph(graphFile, config);
            // Wenn hier Fehler, dann Server noch nicht gestartet.
	    }

	    try {
            // Allgemeine Konfiguration
            this.config = config;
            this.log = config.getLog();
            this.nMaxHeavyThreads = Integer.parseInt(
                    config.getProperty("service.threads.heavy", "5"));
            int nThreadPoolThreads = Integer.parseInt(
                    config.getProperty("service.threads", "50"));

            this.webDir = findOrCreateWebDir(config);

            // HTTP-Infos sammeln
            final String strHost = config.getProperty("service.host" /*new*/,
                    config.getProperty("service.domain"/*old*/, "localhost"));
            final String strPort = config.getProperty("service.port", "8888");
            final String strPortSsl = config.getProperty("service.port.ssl");

            final int intPort = Integer.parseInt(strPort);
            final int intPortSsl = (null == strPortSsl) ? (intPort + 1) : Integer.parseInt(strPortSsl);

            Authenticator auth = null;
            final String authClassName = config.getProperty("service.auth.class");
            if (authClassName != null) {
                @SuppressWarnings("unchecked")
                Class<Authenticator> authClass = (Class<Authenticator>) Config.loadClass(authClassName);
                try {
                    auth = authClass.getDeclaredConstructor(Config.class).newInstance(config);
                } catch (NoSuchMethodException e) {
                    auth = authClass.getConstructor().newInstance();
                }
            }

            String webContext = config.getProperty(
                    "service.context", this.getClass().getSimpleName());
            // Muss vorne / haben, wir wollen aber hinten keinen
            webContext = "/" + cleanPath(webContext);

            if (!this.offline) {
                // kein Progress, da Filter/Handler-Logs folgen koennen
                this.log.progress("Initializing Filters and Handlers");
                boolean addDefaultHandlers = Boolean.valueOf(
                        config.getProperty("service.handler.defaults", "true"));
                this.initFiltersAndHandlers(addDefaultHandlers);
                this.initMimeAndMappings();
            }

            this.log.progress("Starting Services");

            {// WebServer fuer HTTPS bauen (fuer port +1)
                String ksPath = this.config.getProperty("service.ssl.keystore.path");
                String ksPass = this.config.getProperty("service.ssl.keystore.pass", "osm2po");
                URL ksUrl = SecurityHelpers.class.getResource("osm2po.int"); // PKCS12-Keystore
                if (ksPath != null) ksUrl = new File(ksPath).toURI().toURL();

                this.httpsServer = HttpsServer.create();
                this.httpsServer.bind(new InetSocketAddress(strHost, intPortSsl), 0);
                this.httpsServer.createContext(webContext, this).setAuthenticator(auth);
                this.httpsServer.setExecutor(newFixedThreadPool(nThreadPoolThreads));
                this.httpsServer.setHttpsConfigurator(
                        new HttpsConfigurator(createSSLContextTLS(ksUrl, ksPass)));
                this.httpsServer.start();

                this.httpUrl = "https://" + strHost + ":" + intPortSsl + webContext;
                if (!this.httpUrl.endsWith("/")) this.httpUrl += "/"; // Wichtig fuer Browser
                this.log.message(this.httpUrl + " STARTED" + " - " + TS() +  " - " + MB());
            }

            boolean sslOnly = Boolean.valueOf(
                    config.getProperty("service.ssl.only", "false"));

            // https://www.programcreek.com/java-api-examples/index.php ...
            // ?class=com.sun.net.httpserver.HttpExchange&method=getRequestBody
            if (!sslOnly) {// WebServer fuer HTTP bauen.
                this.httpServer = HttpServer.create();
                this.httpServer.bind(new InetSocketAddress(strHost, intPort), 0);
                this.httpServer.createContext(webContext, this).setAuthenticator(auth);
                this.httpServer.setExecutor(newFixedThreadPool(nThreadPoolThreads));
                this.httpServer.start();

                this.httpUrl = "http://" + strHost + ":" + intPort + webContext;
                if (!this.httpUrl.endsWith("/")) this.httpUrl += "/"; // Wichtig fuer Browser
                this.log.message(this.httpUrl + " STARTED" + " - " + TS() +  " - " + MB());
            }


            if (this.offline) {
              this.log.message("Services are OFFLINE! - " + TS() +  " - " + MB());
            }

            final String indexHtml = config.getProperty("service.index.html", "osm2po.html");
            this.indexHtmlResponse = this.createResourceResponse(indexHtml);
            if (null == this.indexHtmlResponse) log.warn(
                    "Could not find 'service.index.html' at " + indexHtml);

            System.gc();

            this.log.logWarnCounters();
            this.log.resetCounters();

            Runtime.getRuntime().addShutdownHook(new Thread(){
                @Override public void run() {
                    log.info("TERMSIG received at " + TS());
                    stopService();
                    getConfig().close();
                }
            });

        } catch (Throwable t) {
            stopService();
            throw new RuntimeException(t);
        }
    }

    public void stopService() {
        try {
            this.log.progress("Stopping Services");

            if (httpServer != null) {
                this.httpServer.stop(0);
                this.httpServer = null;
            }
            if (httpsServer != null) {
                this.httpsServer.stop(0);
                this.httpsServer = null;
            }

            if (this.graph != null) {
                this.graph.close();
                this.graph = null;
            }

            if (this.handlerProxy != null) this.handlerProxy.shutDown();
            for (ServiceFilter filter : this.filterList) filter.shutDown();

            this.filterList.clear();
            this.mimeTypeMap.clear();
            this.handlerList.clear();
            this.cmdMappingMap.clear();
            this.heavyThreadIdSet.clear();

            this.log.message("Services stopped at " + TS());
            this.log.logWarnCounters();

        } catch (Throwable t) {
            this.log.error(t);
        }
    }

    private boolean writeResponse(ServiceResponse response, HttpExchange htx) {
        boolean written = false;
        try {
            Headers headers = htx.getResponseHeaders();

            if (null == response.mimeType) {
                headers.set("Content-Disposition", "attachment");
            } else {
                headers.set("Content-Type", response.mimeType);
            }

            int responseBodySize = null == response.responseBody ? 0
                    : response.responseBody.length;

            htx.sendResponseHeaders(response.httpCode, responseBodySize);

            if (responseBodySize > 0) {
                OutputStream body = htx.getResponseBody();

                // Wenn Browser AJAX abbricht, kann genau hier die folgende IOException kommen.
                // "Eine bestehende Verbindung wurde softwaregesteuert durch den Hostcomputer abgebrochen"
                body.write(response.responseBody);

                body.flush();
                body.close();
            }

            htx.close();

            written = true;

        } catch (IOException e) {
            log.debug("IOException caught: " + e.getMessage());
        } catch (Exception e) {
            log.error(e.toString());
        }
        return written;
    }

    /**
     * Erstellt eine {@link ServiceResponse}, aus einer Datei
     * aus dem osm2po-web (dir) oder aus dem Resource-Pfad "webdir".
     * @param path String.
     * @return {@link ServiceResponse} oder NULL, wenn nicht gefunden
     * @throws Exception beliebiger Natur
     */
    public final ServiceResponse createResourceResponse(String path)
    throws Exception {
        if (null == path) return null;
        if (!path.startsWith("/")) path = "/" + path;
        byte[] bytes = null;

        // if external webDir is configured, check this first
        if (this.webDir != null) { // WebDir-File-Resource
    		String filePath = this.webDir.getAbsolutePath() + path;
    		File extFile = new File(filePath);
    		if (extFile.isFile()) bytes = loadFileIntoByteArray(extFile);
		}

        // fallback is internal resource package "/webroot".
        if (null == bytes) { // ClassPath-Resource, must reside in "/webroot"
            String resourcePath = "webroot" + path;
            // PITFALL: Wenn keine Datei, dann wird hier DIR ausgelesen!
            InputStream is = Config.getResourceAsStream(resourcePath);
            if (is != null) {
                bytes = loadStreamIntoByteArray(is);
                is.close();
            }
        }

        if (bytes != null) {
            String fileExt = null;
            int extDotIdx = path.lastIndexOf('.');
            if (extDotIdx != -1) fileExt = path.substring(extDotIdx);
            String mimeType = this.mimeTypeMap.get(fileExt); // NULLABLE

            return new ServiceResponse(200, mimeType, bytes);
        }

		return null;
    }

    private String resolveLanguage(HttpExchange hx) {
        String lang = getHeaderValue(hx.getRequestHeaders(), "Accept-language");
        if (lang != null) {
            lang = lang.split(",")[0];
            int q = lang.indexOf(";q="); // quality
            if (q >= 0) lang = lang.substring(0, q);
            lang = lang.split("-")[0]; // z.B. de o. de-de
        }
        return lang;
    }

    @Override
    public void checkResources() throws Exception {
        long threadId = Thread.currentThread().getId();
        if (this.heavyThreadIdSet.size() >= this.nMaxHeavyThreads) {
            throw new ServiceException(ERR_TOO_MANY_HEAVY_REQUESTS);
        }
        this.heavyThreadIdSet.add(threadId);
    }

    @Override
    public void resourcesReleased() {
        long threadId = Thread.currentThread().getId();
        this.heavyThreadIdSet.remove(threadId);
    }

    @Override
    public void handle(HttpExchange htx) throws IOException {
        String i18n = this.resolveLanguage(htx);

        ServiceRequest serviceRequest = null;
        ServiceResponse serviceResponse = null;
        ServiceFilter preFilter = null;

        try {
            if (this.offline) throw new ServiceException(ERR_DOWN_MAINTENANCE);

            String ctx = htx.getHttpContext().getPath();
            URI uri = htx.getRequestURI();
            String query = uri.getQuery();

            // Query-Parameter in Properties einlesen
            if (null == query) query = "";
            String[] strParams = query.split("&");
            Properties request = readArgsIntoProps(strParams);

            // Big Ding: Wenn Resource-Namen mit Context-Namen beginnen, dann kommen sie hier durch
            // System.err.println(uri);
            String path = "/" + cleanPath(uri.getPath());
            if (path.startsWith(ctx)) {
                if (path.length() == ctx.length()) { // context-path ohne Resource
                    if (!uri.getPath().endsWith("/")) {
                        String redirectUrl = uri.getPath() + "/";
                        if (!query.isEmpty()) redirectUrl += '?' + query;

                        // ohne / am Ende vergessen <script>-tags den Context-Pfad vorne einzubauen
                        htx.getResponseHeaders().set("Location", redirectUrl);
                        htx.sendResponseHeaders(302, 0L); // Redirect to '/context/'
                        htx.close();
                        return;
                    }
                    path = "/";

                } else if ('/' == path.charAt(ctx.length())) { // '/' + Context + '/' + Resource
                    path = path.substring(ctx.length()); // jetzt nur noch '/' + Resource
                }
            }

            request.setProperty("urlContextPath", ctx);
            request.setProperty("urlResourcePath", path);
            request.setProperty("urlQueryString", query);

            boolean cmdIsQueryParam = true;
            String cmd = request.getProperty("cmd");

            // Fallback via Path-Mapping der Config-Form
            // service.mapping.mycmd=/hello/world,/hi/there
            if (null == cmd && !this.cmdMappingMap.isEmpty()) {
                cmdIsQueryParam = false;
                cmd = this.findCmdMapping(path, request);
                if (cmd != null) request.setProperty("cmd", cmd);
            }

            serviceRequest = new ServiceRequest(cmd, request, htx);
            for (ServiceFilter filter : this.filterList) {
                preFilter = filter;
                if (filter.preHandle(serviceRequest)) break;
                preFilter = null; // Keine Exception, kein handled=true
            }

            if (null == preFilter) { // Wenn kein Filter was dagegen hat

                if ("init".equals(cmd)) { // Fuer Selbstauskunft keinen Handler rufen
                    serviceResponse = createInitResponse(this.config, this.graph);
                }

                if (null == serviceResponse && cmd != null) {
                    // Neue Handler-Logik via Reflection
                    serviceResponse = this.handlerProxy.invokeHandler(serviceRequest);

                    // Alte Handler-Logik via Handler-Chain
                    if (null == serviceResponse) {
                        final String requestMethod = htx.getRequestMethod();
                        if (!"GET".equals(requestMethod)) throw new ServiceException(
                                ERR_HTTP_METHOD_NOT_ALLLOWED, requestMethod);
                        for (ServiceHandler handler : this.handlerList) {
                            serviceResponse = handler.handle(cmd,
                                    this.config, this.graph, request, this);
                            if (serviceResponse != null) break;
                        }
                    }

                    if (null == serviceResponse && cmdIsQueryParam)
                        throw new ServiceException(ERR_PARAMETER_INVALID, "cmd", cmd);
                }

                if (null == serviceResponse && "/".equals(path)) {
                    serviceResponse = this.indexHtmlResponse;
                }

                if (null == serviceResponse) {
                    serviceResponse = this.createResourceResponse(path);
                }

                if (null == serviceResponse) {
                    serviceResponse = new ServiceResponse(404,
                            MIME_JSON, "{\"http\":404}".getBytes());
                }
            }

        } catch (Throwable t) { // Globaler Alles-Catcher
            ServiceException e = new ServiceException(t); // Wrap
            if (ERR_UNEXPECTED_EXCEPTION == e.getError()) log.error(t);

            LangEnum lang = LangEnum.findByCode(i18n);

            if (this.errorsAsText) {
                String err = "#Error: " + e.getNumber() + "\n" + e.getLocalizedMessage(lang);
                serviceResponse = new ServiceResponse(
                        e.getHttpCode(), MIME_TEXT, err.getBytes(UTF8));
            } else {
                String err = "{\"err\":" + e.getNumber() + ",\"msg\":"
                    + JsonBuilder.build(e.getLocalizedMessage(lang)) + "}";
                serviceResponse = new ServiceResponse(
                        e.getHttpCode(), MIME_JSON, err.getBytes());
            }


        } finally {
            this.resourcesReleased();
        }

        // Achtung. Response auch durch Throwable s.o. setzbar!
        boolean written = serviceResponse != null // wenn PreFilter true
            && this.writeResponse(serviceResponse, htx);
        for (ServiceFilter filter : this.filterList) {
            filter.postHandle(serviceRequest, serviceResponse, written);
            if (filter == preFilter) break; // Nur die Pre's kriegen Post's
        }
    }

    private void initMimeAndMappings() throws Exception {

        final String mimeKeyPrefix = "service.mime";
        final int mimeKeyPrefixLen = mimeKeyPrefix.length();
        for (String key : this.config) {
            if (key.startsWith(mimeKeyPrefix)) {
                String val = this.config.getProperty(key);
                key = key.substring(mimeKeyPrefixLen);
                this.mimeTypeMap.put(key, val);
            }
        }

        final String cmdKeyPrefix = "service.mapping.";
        final int cmdKeyPrefixLen = cmdKeyPrefix.length();
        for (String key : this.config) {
            if (key.startsWith(cmdKeyPrefix)) {
                String[] mappings = config.getProperty(key).split(",");
                key = key.substring(cmdKeyPrefixLen);
                for (String mapping : mappings) {
                    // TODO Mapping-Klasse fuer z.B. ids/4711/find
                    mapping = mapping.trim();
                    if (!"".equals(mapping)) {
                        this.cmdMappingMap.put(mapping, key);
                    }
                }
            }
        }
    }

    private String findCmdMapping(String path, Properties request) {
        String cmd = null;
        String[] params = new String[0x80];
        int nParams = 0;

        while (null == (cmd = this.cmdMappingMap.get(path))) {
            int idx = path.lastIndexOf('/');
            if (idx <= 0) break;
            params[nParams++] = path.substring(idx +1);
            path = path.substring(0, idx);
        }

        for (int i = 0; i < nParams; i++) {
            request.setProperty("urlPathParam" + (nParams - i), params[i]);
        }

        return cmd;
    }

    private void initFiltersAndHandlers(boolean addDefaultHandlers) throws Exception {

        for (int i = 0; i <= 127; i++) {
            String key = "service.filter." + i + ".class";
            String val = this.config.getProperty(key);
            if (val != null) {
                try {
                    Class<?> clazz = Config.loadClass(val);
                    Object obj = clazz.getDeclaredConstructor().newInstance();
                    ServiceFilter filter = (ServiceFilter) obj;
                    this.filterList.add(filter);
                    filter.init(this.config, i);

                } catch (Exception e) {
                    this.log.error("Could not load service filter: " + val);
                    this.log.debug(Log.translateError(e, 5));
                }
            }
        }

        this.handlerProxy = new ServiceHandlerProxy(this, this.config, this.graph);
        for (int i = 0; i <= 127; i++) {
            String key = "service.handler." + i + ".class";
            String val = this.config.getProperty(key);
            if (val != null) {
                try {
                    Class<?> clazz = Config.loadClass(val);
                    if (ServiceHandler.class.isAssignableFrom(clazz)) {
                        Object obj = clazz.getDeclaredConstructor().newInstance();
                        this.handlerList.add((ServiceHandler) obj);
                    } else {
                        this.handlerProxy.registerHandlerClass(clazz);
                    }

                } catch (Exception e) {
                    this.log.error("Could not load service handler: " + val);
                    this.log.debug(Log.translateError(e, 5));
                }
            }
        }
        this.cmdMappingMap.putAll(this.handlerProxy.getCmdMapping());

        // Fallbacks:
        if (addDefaultHandlers) {
            this.handlerList.add(new PositionHandler());
            this.handlerList.add(new RouteHandler());
            this.handlerList.add(new GeoHandler());
            this.handlerList.add(new TspHandler());
            this.handlerList.add(new StateHandler());
            this.handlerList.add(new DebugHandler());
        }

        if (log.isEnabled(Log.LEVEL_DEBUG)) {
            if (!this.handlerList.isEmpty()) {
                String s = "Registered legacy handlers:";
                for (ServiceHandler sh : this.handlerList) {
                    s += "\n" + sh.getClass().getName();
                }
                log.debug(s);
            }

            final String ps = this.handlerProxy.toString();
            if (ps != null) {
                log.debug("Registered proxy handlers:\n" + ps);
            }

        }
    }

    private static Graph openGraph(File graphFile, Config cfg) {
        int graphSupports = Graph.SUPPORT_LATLON;
        if (Boolean.valueOf(cfg.getProperty("graph.support.inmemory", "false")))
            graphSupports |= Graph.SUPPORT_INMEMORY;
        if (Boolean.valueOf(cfg.getProperty("graph.support.edgeflags", "false")))
            graphSupports |= Graph.SUPPORT_EDGEFLAGS;
        if (Boolean.valueOf(cfg.getProperty("graph.support.reverse", "false")))
            graphSupports |= Graph.SUPPORT_REVERSE;
        if (Boolean.valueOf(cfg.getProperty("graph.support.raster", "false")))
            graphSupports |= Graph.SUPPORT_RASTER_EDGE | Graph.SUPPORT_RASTER_VERTEX;
        if (Boolean.valueOf(cfg.getProperty("graph.support.raster.e", "false")))
            graphSupports |= Graph.SUPPORT_RASTER_EDGE;
        if (Boolean.valueOf(cfg.getProperty("graph.support.raster.v", "false")))
            graphSupports |= Graph.SUPPORT_RASTER_VERTEX;
        if (Boolean.valueOf(cfg.getProperty("graph.support.barriers", "false")))
            graphSupports |= Graph.SUPPORT_BARRIERS;
        if (Boolean.valueOf(cfg.getProperty("graph.support.extensions", "false")))
            graphSupports |= Graph.SUPPORT_EXTENSIONS;

        return new Graph(graphFile, cfg.getLog(), graphSupports, false, cfg.getProps());
    }

    /**
     * Erstellt den Initialisierungs-Json.
     * @param config {@link Config}
     * @param graph {@link Graph}
     * @return {@link ServiceResponse}
     */
    private static ServiceResponse createInitResponse(Config config, Graph graph) {

        String poiClazzes = "[";
        for (int i = -128; i <= 127; i++) {
            String key = "nodeClazz." + i + ".name";
            String val = config.getProperty(key);
            if (val != null) {
                if (poiClazzes.length() > 1) poiClazzes += ",";
                poiClazzes += "{\"clazz\":" + i + ",\"name\":\"" + val + "\"}";
            }
        }
        poiClazzes += "]";

        String tmsLayers = "[";
        for (int i = 1; i < 128; i++) {
            String key = "service.tmsLayer." + i;
            String val = config.getProperty(key);
            if (val != null) {
                String[] kv = val.split(",");
                String[] params = new String[] {
                        kv[0].trim(), "tmsLayer" + i, "20", "false"};
                for (int p = 1; p < params.length && p < kv.length; p++)
                    params[p] = kv[p].trim();

                if (tmsLayers.length() > 1) tmsLayers += ",";
                tmsLayers += "{\"url\":\"" + params[0];
                tmsLayers += "\",\"name\":\"" + params[1];
                tmsLayers += "\",\"levels\":" + params[2];
                tmsLayers += ",\"base\":" + params[3] + "}";
            }
        }
        tmsLayers += "]";

        List<Entry<Integer, Class<? extends AbstractRouter<?>>>> routerClassList =
            new ArrayList<Entry<Integer,Class<? extends AbstractRouter<?>>>>(
                    config.getRouterClasses().entrySet());
        Collections.sort(routerClassList, new Comparator<Entry<Integer, Class<? extends AbstractRouter<?>>>>() {
            @Override public int compare(Entry<Integer, Class<? extends AbstractRouter<?>>> o1,
                    Entry<Integer, Class<? extends AbstractRouter<?>>> o2) {
                return o1.getKey().compareTo(o2.getKey());}});
        String routers = "[";
        for (Entry<Integer, Class<? extends AbstractRouter<?>>> entry : routerClassList) {
            if (routers.length() > 1) routers += ",";
            routers += "{\"id\":" + entry.getKey()
                + ",\"name\":\"" + entry.getValue().getSimpleName() + "\"}";
        }
        routers += "]";

        String json = "{"
            + "\"version\":\"" + Version.getVersion() + "\","
            + "\"bounds\":["
            + graph.getMinLat() + "," + graph.getMinLon() + ","
            + graph.getMaxLat() + "," + graph.getMaxLon() + "],"
            + "\"routers\":" + routers + ","
            + "\"poiClazzes\":" + poiClazzes + ","
            + "\"tmsLayers\":" + tmsLayers + ""
            + "}";

        return new ServiceResponse(200, MIME_JSON, json.getBytes());
    }

    public static void requestBasicAuth(HttpExchange htx) throws Exception {
        Headers headers = htx.getResponseHeaders();
        headers.set("WWW-Authenticate", "BASIC realm=\"OSM2PO\"");
        byte[] msg = "".getBytes();

        htx.sendResponseHeaders(ERR_UNAUTHORIZED.getHttpCode(), msg.length);

        htx.getResponseBody().write(msg);
        htx.getResponseBody().flush();
        htx.getResponseBody().close();

        htx.close();
    }

    private static File findOrCreateWebDir(Config config) {
        // WebDir - ensure NOT NULL
        Log log = config.getLog();
        File webDir = null;
        String webDirPath = config.getProperty("service.webdir");
        if (webDirPath != null) webDir = new File(webDirPath);
        else webDir = findFileOrDir("osm2po-web", false);
        if (webDir != null && !webDir.isDirectory()) webDir = null;
        if (null == webDir) { // Still not found, then use default
            File graphFile = config.findGraphFile(true);
            File parentDir = graphFile == null ? new File(".") : graphFile.getParentFile();
            webDir = new File(parentDir, "osm2po-web");
            if (!webDir.exists()) {
                log.warn("WebDir does not exist! Creating default.");
                webDir.mkdirs();
            }
        }
        log.info("WebDir is " + webDir.getAbsolutePath());
        return webDir;
    }

}