/*
 * $Id: ServiceHelpers.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2012 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;

import static de.cm.osm2po.misc.Code32.decode;
import static de.cm.osm2po.misc.Utils.getJarDir;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_PARAMETER_INVALID;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_SOURCE_INVALID;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_SOURCE_NOT_FOUND;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TARGET_INVALID;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TARGET_NOT_FOUND;
import static java.lang.Double.parseDouble;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;

import de.cm.osm2po.Version;
import de.cm.osm2po.model.Bounding;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.primitives.RandomAccessArray;
import de.cm.osm2po.routing.AbstractRouter;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayEdge;
import de.cm.osm2po.routing.OverlayNode;
import de.cm.osm2po.routing.RoutingResultSegment;
import de.cm.osm2po.service.errors.ErrorEnum;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * Bildet eine Kompatibilitaets-Aufruf-Klammer um den {@link AbstractRouter}
 * fuer die Verwendung im {@link Osm2poService}.
 *
 * @author (c) 2012 - Carsten Moeller - info@osm2po.de
 */
public class ServiceHelpers {

    public final static String SERVICESTATE_FILENAME =
        ".osm2po." + Version.getVersion() + ".server.states";

    private ServiceHelpers() {/*static*/}

    /**
     * Erstellt aus einem gegebenen Pfad eine Liste von {@link RoutingResultSegment}
     * angereichert und korrigiert um die Koordinaten pro Segment. Letzteres
     * unter der Beruecksichtigung von Teil-Segmenten aufgrund des OverlayRoutings.
     * @param graph {@link Graph}
     * @param overlayEdges Array[] of {@link OverlayEdge}
     * @return Array[] or {@link RoutingResultSegment}
     */
    public static RoutingResultSegment[] createResult (Graph graph, OverlayEdge[] overlayEdges) {
        RoutingResultSegment[] rrs = new RoutingResultSegment[overlayEdges.length];
        for (int i = 0; i < overlayEdges.length; i++) {
            rrs[i] = createRRS(graph, overlayEdges[i]);
        }
        return rrs;
    }

    /**
     * Wandelt eine OverlayEdge in ein vollstaendiges RoutingResultSegment.
     * Zuerst wird dabei {@link Graph#lookupSegment(int)} aufgerufen und,
     * wenn nicht vorhanden, die Geometrien angereichert und mittels der
     * Ratio die Kosten neu berechnet.
     * @param graph {@link Graph} not null
     * @param overlayEdge {@link OverlayEdge} nullable
     * @return {@link RoutingResultSegment} oder null
     */
    public static RoutingResultSegment createRRS(Graph graph, OverlayEdge overlayEdge) {
        if (null == overlayEdge) return null;
        RoutingResultSegment rrs = graph.lookupSegment(overlayEdge.getEdgeIdx());
        if (null == rrs) return null;
        LatLons coords = overlayEdge.getGeometry();
        if (coords != null) {
            rrs.setCoords(coords);
            rrs.setH(overlayEdge.getH());
            rrs.setKm(overlayEdge.getKm());
        }
        return rrs;
    }

    /**
     * Parsed einen String, der einen Punkt beschreibt.
     * Dies kann entweder eine VertexID, eine Koordinate der
     * Form "lat,Lon" oder ein {@link OverlayNode} als Code32
     * mit fuehrendem "A" sein.
     * @param graph {@link Graph}
     * @param strNode ID oder Koordinate als String
     * @param forOverlayRouting true: Edges werden untersucht.
     * @param forSource true: source, false target.
     * @return {@link OverlayNode}
     * @throws ServiceException beliebiger Fehler
     */
    public static OverlayNode parseNode(Graph graph, String strNode,
            boolean forOverlayRouting, boolean forSource)
    throws ServiceException {

        ErrorEnum errNotFound =
            forSource ? ERR_SOURCE_NOT_FOUND : ERR_TARGET_NOT_FOUND;
        ErrorEnum errInvalid =
            forSource ? ERR_SOURCE_INVALID : ERR_TARGET_INVALID;

        if (strNode.startsWith("A")) { //Konvention! OverlayNode als Parameter
            byte[] olnBytes = decode(strNode);
            RandomAccessArray raa = new RandomAccessArray();
            raa.writeBytes(olnBytes, olnBytes.length);
            OverlayNode oln = new OverlayNode().readFromStream(raa);
            if (forOverlayRouting) return oln;
            strNode = oln.getLat() + "," + oln.getLon(); // Fallback to static Graph
        }

        String[] arrStrLatLon = strNode.split(",");
        if (arrStrLatLon.length == 2) { // LatLon als Parameter
            float lat;
            float lon;
            try {
                lat = Float.parseFloat(arrStrLatLon[0]);
                lon = Float.parseFloat(arrStrLatLon[1]);
            } catch (NumberFormatException e) {
                throw new ServiceException(errInvalid);
            }
            if (forOverlayRouting) {
                OverlayNode tp = new OverlayNode(lat, lon);
                if (!tp.snap(graph, null)) throw new ServiceException(errNotFound);
                return tp;
            } else {
                int vertexId = graph.findClosestVertexId(lat, lon);
                if (vertexId < 0) throw new ServiceException(errNotFound);
                return new OverlayNode(vertexId);
            }
        } else { // VertexId als Parameter
            int vertexId = 0;
            try {
                vertexId = Integer.parseInt(strNode);
            } catch (NumberFormatException e) {
                throw new ServiceException(errInvalid);
            }
            int maxVertexId = graph.getEdgeEntries().length -1;
            // ID-Range-Check. Jedoch 0 zulassen fuer DrivingCircleDistance
            if (vertexId < 0 || (vertexId > maxVertexId))
                throw new ServiceException(errInvalid);
            return new OverlayNode(vertexId);
        }
    }

    /**
     * Sucht in den Properties nach einem Schluessel und liefert diesen
     * als double zurueck. Wirft ansonsten eine {@link ServiceException}.
     * Auch fuer ints, etc. verwendbar.
     * @param props {@link Properties}
     * @param key String Schluessel
     * @param defaultValue Rueckgabewert, wenn Schluessel nicht gefunden wurde.
     * @return double, Wert.
     * @throws ServiceException beliebiger Fehler
     */
    public static double parseDoubleProperty(
            Properties props, String key, double defaultValue)
    throws ServiceException {
        String strNum = props.getProperty(key);
        if (null == strNum) return defaultValue;
        try {
            return Double.parseDouble(strNum);
        } catch (Exception e) {
            throw new ServiceException(ERR_PARAMETER_INVALID, key, strNum);
        }
    }

    public static Bounding parseBounding(Properties props) {
        try {
            String[] uiBounds = props.getProperty("bounds").split(",");
            return new Bounding(
                    parseDouble(uiBounds[1].trim()), parseDouble(uiBounds[0].trim()),
                    parseDouble(uiBounds[3].trim()), parseDouble(uiBounds[2].trim()));

        } catch (Exception e) {
            return new Bounding(-90, -180, 90, 180);
        }
    }

    public static void saveUiState(String key, String value) throws Exception {
        Properties props = new Properties();
        File jarDir = getJarDir(); // wenn null, dann current
        File propsFile = new File(jarDir, SERVICESTATE_FILENAME);
        if (propsFile.exists()) {
            InputStream is = new FileInputStream(propsFile);
            props.load(is);
            is.close();
        }
        props.setProperty(key, value);
        OutputStream os = new FileOutputStream(propsFile);
        props.store(os, "Osm2poService saved states");
        os.close();
    }

    public static String loadUiState(String key) throws Exception {
        Properties props = new Properties();
        File jarDir = getJarDir(); // wenn null, dann current
        File propsFile = new File(jarDir, SERVICESTATE_FILENAME);
        if (propsFile.exists()) {
            InputStream is = new FileInputStream(propsFile);
            props.load(is);
            is.close();
        }
        return props.getProperty(key);
    }

}
