/*
 * $Id: Smtp.java 5211 2020-11-04 20:28:21Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service;
import static de.cm.osm2po.misc.Utils.UTF8;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;

import de.cm.osm2po.logging.Log;
import de.cm.osm2po.logging.LogMailWriter;
import de.cm.osm2po.misc.Base64;

/**
 * Ein einfacher SMTP-Mail-Sender.<br>
 * Unterstuetzt Standard (Port 25) und SSL/TLS (normalerweise Port 465).<br>
 * STARTTLS wird nicht unterstuetzt.
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 * @see LogMailWriter
 */
public class Smtp {

    private OutputStream out;
    private SmtpResponse in;

    private final Log log;
    private final String from;
    private final String host;
    private final int port;
    private final boolean ssltls;
    private final String userpass;

    private boolean sending;

    /**
     * CTOR.
     * @param log {@link Log}
     * @param from EMail-Adresse des Absenders
     * @param host String SMTP-Host z.B. mail.gmx.net
     * @param port int SMTP-Port, z.B. 25 (standard) oder 465 (ssltls)
     * @param ssltls true: SSL-Socket mit TLS, false: Unverschluesselt
     * @param user String SMTP-AnmeldeName fuer SMTP-Host
     * @param pass String SMTP-AnmeldePassWort fuer SMTP-Host
     */
    public Smtp(Log log, String from, String host,
            int port, boolean ssltls, String user, String pass) {
        if (null == log) log = Log.empty();
        this.sending = false;
        this.log = log;
        this.from = from;
        this.host = host;
        this.port= port;
        this.ssltls = ssltls;
        String userpass = '\0' + user + '\0' + pass;
        this.userpass = Base64.encodeMime(userpass.getBytes());
    }

    public boolean isSending() {return this.sending;}

    /**
     * Sendet eine Mail.
     * @param subject String Betreff
     * @param message String Text-Botschaft
     * @param to String Empfaenger EMail-Adresse
     * @param bccs String-Array NULLABLE, weitere Empfaenger als BCC
     * @param contentType String NULLABLE. Wenn NULL, dann text/plain UTF-8
     * @return true: Mail wurde versendet, false: Fehler
     * <br><br><b>Wichtig:</b><br>
     * Hier werden keine Exceptions geworfen, {@link Throwable}
     * wird aufgefangen.
     */
    public synchronized boolean send( // nicht final wg. moegl. UnitTest
            String subject, String message, String to,
            String[] bccs, String contentType) {
        if (null == contentType) contentType = "text/plain; charset=\"utf-8\"";

        this.sending = true;
        boolean sent = false;
        try {
            Socket smtpSocket = null;
            if (this.ssltls) {
                SSLContext sslContext = SSLContext.getInstance("TLS");
                sslContext.init(null, null, null);
                SSLSocketFactory sf = sslContext.getSocketFactory();
                smtpSocket = sf.createSocket(this.host, this.port);
            } else {
                smtpSocket = new Socket(this.host, this.port);
            }

            this.out = smtpSocket.getOutputStream();
            this.in = new SmtpResponse(log, smtpSocket.getInputStream());
            this.in.start();

            // https://cr.yp.to/smtp/mail.html
            // https://www.w3.org/Protocols/rfc1341/5_Content-Transfer-Encoding.html

            // 1x Server ready
            if ("220".equals(this.in.waitFor("220"))) {

                // 1x Server greets and/or Nx enhanced
                this.sendAndWait("EHLO", "250");

                // 1x Auth 235 ok, 535 failed
                if ("235".equals(this.sendAndWait(
                        "AUTH PLAIN " + this.userpass, "235", "535"))) {

                    this.sendAndWait("MAIL From: " + this.from, "250");
                    this.sendAndWait("RCPT To: " + to, "250");

                    if (bccs != null) for (String bcc : bccs)
                        this.sendAndWait("RCPT To: " + bcc, "250");

                    this.sendAndWait("DATA");
                    this.sendAndWait("X-Mailer:osm2po.de");

                    this.sendAndWait("Content-Type: " + contentType);

                    this.sendAndWait("From: " + this.from);
                    this.sendAndWait("To: " + to);

                    this.sendAndWait("Subject:" + subject);
                    this.sendAndWait(""); // Muss laut Doku, geht aber ohne
                    this.sendAndWait(message);
                    this.sendAndWait("");

                    sent = "250".equals(this.sendAndWait(".", "250", "500"));
                }
            }

            this.sendAndWait("QUIT", "221");

            this.in.interrupt();
            this.in.join();

            for (String s : this.in.flush()) {
                if (s.startsWith("5")) log.warn(s);
                else log.debug(s);
            }

            smtpSocket.close();


        } catch(InterruptedException ie) { //Wait interrupted
            log.warn(Smtp.class.getSimpleName() + " interrupted");

        } catch(Throwable t) {
            log.error(t);
        }

        this.sending = false;
        return sent;
    }

    private String sendAndWait(
            String smtpMessage, String... prefix) throws Exception {
        this.out.write(smtpMessage.getBytes(UTF8));
        this.out.write("\r\n".getBytes());
        return this.in.waitFor(prefix);
    }

    private static class SmtpResponse extends Thread {
        private Log log;
        private BufferedReader br;
        private Deque<String> protocol;

        private SmtpResponse(Log log, InputStream is) {
            this.log = log;
            this.br = new BufferedReader(new InputStreamReader(is));
            this.protocol = new LinkedList<String>();
        }

        @Override public void run() {
            try {
                String s;
                while ((s = this.br.readLine()) != null) {
                    this.protocol.add(s);
                }
            } catch (Exception e) {log.error(e);}
        }

        private String waitFor(String... prefix) throws Exception {
            return this.waitFor(3000, prefix);
        }

        private String waitFor(
                long maxWait, String... prefix) throws Exception {
            if (prefix != null && prefix.length > 0) {
                while (maxWait > 0) {
                    if (!this.protocol.isEmpty()) for (String p : prefix) {
                        if (this.protocol.peekLast().startsWith(p)) {
                            return p;
                        }
                    }
                    Thread.sleep(10);
                    maxWait -= 10;
                }
            }
            return null;
        }

        /**
         * Liefert das aktuell aufgelaufene Antwort-Protokoll
         * des SMTP-Servers.
         * @return {@link List} of String (Protokoll-Zeilen)
         * @throws Exception beliebiger Natur
         */
        public List<String> flush() throws Exception {
            List<String> list = new ArrayList<String>();
            list.addAll(this.protocol);
            this.protocol.clear();
            return list;
        }
    }

}

