/*
 * $Id: ServiceException.java 4734 2018-03-26 11:56:24Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.errors;

import static de.cm.osm2po.service.errors.ErrorEnum.ERR_OUT_OF_MEMORY;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_EXCEPTION;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_ILLEGAL_ARGUMENT;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_ILLEGAL_STATE;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_NUMBER_FORMAT;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNEXPECTED_UNSUPPORTED_OPERATION;

public class ServiceException extends Exception {

    private static final long serialVersionUID = 190407051142499093L;

    private ErrorEnum error;
    private Object[] args;

    public ServiceException(ErrorEnum error, Object... args) {
        this.error = error;
        this.args = args;
    }

    public ServiceException(Throwable t) {
        super(t);

        if (t instanceof ServiceException) {
            ServiceException t2 = (ServiceException) t;
            this.args = t2.args;
            this.error = t2.error;
            return;
        }

        StackTraceElement[] st = t.getStackTrace();
        this.setStackTrace(st);

        if (t instanceof OutOfMemoryError) {
            this.error = ERR_OUT_OF_MEMORY;
            this.args = new Object[]{t.getClass().getSimpleName()};
        } else if (t instanceof NumberFormatException) {
            this.error = ERR_UNEXPECTED_NUMBER_FORMAT;
            this.args = new Object[]{t.getMessage()};
        } else if (t instanceof UnsupportedOperationException) {
            this.error = ERR_UNEXPECTED_UNSUPPORTED_OPERATION;
            String method = "unknown()";
            if (st != null && st.length > 0) {
                method = "[" + st[0].getClassName() + "." + st[0].getMethodName() + "]";
            }
            this.args = new Object[]{method};
        } else if (t instanceof IllegalArgumentException) {
            this.error = ERR_UNEXPECTED_ILLEGAL_ARGUMENT;
            this.args = new Object[]{t.getMessage()};
        } else if (t instanceof IllegalStateException) {
            this.error = ERR_UNEXPECTED_ILLEGAL_STATE;
            this.args = new Object[]{t.getMessage()};
        } else {
            this.error = ERR_UNEXPECTED_EXCEPTION;
            this.args = new Object[]{t.toString()};
        }
    }

    @Override
    public String getMessage() {
        return this.error.getMessage(null, args);
    }

    public String getLocalizedMessage(LangEnum lang) {
        return this.error.getMessage(lang, args);
    }

    public int getNumber() {
        return this.error.getNumber();
    }

    public ErrorEnum getError() {
        return this.error;
    }

    public int getHttpCode() {
        return this.error.getHttpCode();
    }

}
