/*
 * $Id: OncePerIpFilter.java 4974 2018-10-25 10:42:01Z carsten $
 *
 * Copyright 2018 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.filters;

import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TOO_MANY_HEAVY_REQUESTS;

import java.util.HashSet;
import java.util.Set;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.service.ServiceFilter;
import de.cm.osm2po.service.ServiceRequest;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * @author (c) 2018 - Carsten Moeller, info@osm2po.de
 */
public class OncePerIpFilter implements ServiceFilter {
    private final static char ENDPOINT_SEP = '$';
    private final static String HTX_ATTR = "HTX_ONCE_PER_IP_KEY";

    private Set<String> endpointSet;
    private Set<String> activeEndpointSet;
    private Log log;

    @Override
    public void init(Config config, int index) {
        this.log = config.getLog();
        if (null == this.log) this.log = Log.stderr();

        final String prefix = "service.filter." + index + ".";

        this.activeEndpointSet = new HashSet<String>();
        this.endpointSet = new HashSet<String>();

        String endpointsCsv = config.getProperty(prefix + "endpoints");
        if (endpointsCsv != null) for (String endpoint : endpointsCsv.split(",")) {
            this.endpointSet.add(endpoint.trim());
        }
    }

    @Override
    public void shutDown() {
        this.endpointSet.clear();
        this.activeEndpointSet.clear();
    }

    @Override
    public boolean preHandle(ServiceRequest sr) throws Exception {
        if (this.endpointSet.contains(sr.cmd)) {
            synchronized (this.activeEndpointSet) {
                String key = sr.ip + ENDPOINT_SEP + sr.cmd;
                if (this.activeEndpointSet.contains(key)) {
                    log.warn("OIP [" + key + "] denied: " + sr.htx.getRequestURI());
                    throw new ServiceException(ERR_TOO_MANY_HEAVY_REQUESTS);
                }
                this.activeEndpointSet.add(key);
                sr.htx.setAttribute(HTX_ATTR, key);
            }
        }
        return false;
    }


    @Override public void postHandle(
            ServiceRequest req, ServiceResponse resp, boolean sent) {
        // TODO postHandle garantiert? Ggf. MaxIdle einbauen und abraeumen, falls verwaist.
        String key = (String) req.htx.getAttribute(HTX_ATTR);
        if (key != null) synchronized (this.activeEndpointSet) {
            this.activeEndpointSet.remove(key);
        }
    }

}
