/*
 * $Id: RateLimitFilter.java 5828 2023-07-19 10:49:15Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.filters;

import static de.cm.osm2po.service.Osm2poService.getHeaderValue;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_ACCOUNT_EXPIRED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_RATELIMIT_EXCEEDED;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_UNAUTHORIZED;
import static de.cm.osm2po.service.security.RateLimit.KEY_DDOS;
import static de.cm.osm2po.service.security.RateLimit.KEY_EXPIRED;
import static de.cm.osm2po.service.security.RateLimit.KEY_OK;
import static de.cm.osm2po.service.security.RateLimit.KEY_RATELIMIT;

import java.io.File;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import com.sun.net.httpserver.Authenticator;
import com.sun.net.httpserver.HttpPrincipal;

import de.cm.osm2po.Config;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.ServiceFilter;
import de.cm.osm2po.service.ServiceRequest;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ServiceException;
import de.cm.osm2po.service.security.RateLimit;
import de.cm.osm2po.service.security.RateLimiter;

/**
 * Ein {@link ServiceFilter}, der Aufrufe mittels {@link RateLimiter}
 * ueberwacht. Sonderrechte koennen per ApiKey (x-api-key) aber auch
 * {@link HttpPrincipal} (via {@link Authenticator} im {@link Osm2poService})
 * getriggert werden.<br>
 * <strong>Config Beispiel:</strong>
 * <pre>
 * service.filter.1.class=de.cm.osm2po.service.filters.RateLimitFilter
 * service.filter.1.keystore=[user.home]/Desktop/apikeys.txt
 * service.filter.1.backup=[user.home]/Desktop/apikeys.backup
 * service.filter.1.refresh=1 (Minuten - Intervall fuer Keystore-Abgleich)
 * service.filter.1.ttl=5 (Minuten - MaxIdle fuer transientes {@link RateLimit})
 * service.filter.1.ignoreCmds=NULL,route (Diese Kommandos nicht validieren. NULL=Resourcen)
 * </pre>
 * @see RateLimit
 * @see RateLimiter
 * @author (c) 2017/2018 - Carsten Moeller, info@osm2po.de
 */
public class RateLimitFilter implements ServiceFilter {
    private final static String HTX_ATTR = "HTX_RATELIMIT_KEY";

    private RateLimiter rateLimiter;
    private Set<String> ignoreCmdSet;
    private Log log;

    @Override
    public void init(Config config, int index) {
        this.log = config.getLog();
        if (null == this.log) this.log = Log.stderr();

        final String prefix = "service.filter." + index + ".";

        this.ignoreCmdSet = new HashSet<String>(0x20);
        String ignoreCmdCsv = config.getProperty(prefix + "ignoreCmds");
        if (ignoreCmdCsv != null) for (String cmd : ignoreCmdCsv.split(",")) {
            this.ignoreCmdSet.add(cmd.trim());
        }

        File keystoreFile = null;
        String keystorePath = config.getProperty(prefix + "keystore");
        if (keystorePath != null) keystoreFile = new File(keystorePath);

        File backupFile = null;
        String backupPath = config.getProperty(prefix + "backup");
        if (backupPath != null) backupFile = new File(backupPath);

        // RefreshInterval in Minuten. Default = 1
        final long refreshInterval = 60000 * Long.parseLong(
                config.getProperty(prefix + "refresh", "1"));

        // MaxIdle von RateLimit-Elementen in Minuten. Default = 5
        final long maxIdle = 60000 * Long.parseLong(
                config.getProperty(prefix + "ttl", "5"));

        this.rateLimiter = new RateLimiter(config.getLog(), backupFile,
                refreshInterval, 10, maxIdle, keystoreFile).start();
    }

    @Override
    public boolean preHandle(ServiceRequest sr) throws Exception {
        String cmd = sr.cmd;
        if (null == cmd) {
            cmd = "NULL";
        }
        if (this.ignoreCmdSet.contains(cmd)) {
            return false;
        }

        String authKey = Optional.ofNullable(sr.htx.getPrincipal()).map(HttpPrincipal::getName).orElse(null);
        if (null == authKey) { // Wenn kein BasicAuth
            authKey = getHeaderValue(sr.htx.getRequestHeaders(), "x-api-key");
            if (null == authKey) authKey = sr.params.getProperty("xak");
        }

        RateLimit rateLimit = this.rateLimiter.validate(sr.ip, authKey, sr.cmd);
        int status = rateLimit.getStatus();

        if (status != KEY_OK) {
            if (KEY_DDOS == status) Thread.sleep(1000);

            switch (status) {
            case KEY_RATELIMIT: throw new ServiceException(ERR_RATELIMIT_EXCEEDED);
            case KEY_EXPIRED: throw new ServiceException(ERR_ACCOUNT_EXPIRED);
            default: throw new ServiceException(ERR_UNAUTHORIZED); // auch bei DDOS
            }
        } else if (authKey != null) {
            // Fuer Fehler von ausserhalb, RateLimit fuer PostHandle merken.
            sr.htx.setAttribute(HTX_ATTR, rateLimit);
        }

        return false;
    }

    @Override
    public void shutDown() {
        this.rateLimiter.shutdown();
    }

    @Override public void postHandle(
            ServiceRequest req, ServiceResponse resp, boolean sent) {
        if (!sent || null == resp || resp.httpCode != 200) { // Wenn externer Fehler
            // Ein Benutzer bezahlt nur fuer erfolgreiche Berechnungen, daher
            // muss das RateLimit bei externem Fehler rehabilitiert werden.
            // Sicherheitsluecke: Wir wissen hier nicht, ob es ein valider User ist,
            // sondern nur, dass ein AuthKey uebergeben wurde, wenn RateLimit != null.
            RateLimit rateLimit = (RateLimit) req.htx.getAttribute(HTX_ATTR);
            if (rateLimit != null) rateLimit.decreaseLimitCounter();
        }
    }

}
