/*
 * $Id: DebugHandler.java 5712 2022-09-23 07:52:26Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.handlers;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.routing.RoutingResultWriter.writePathAsFeatureCollection;
import static de.cm.osm2po.routing.RoutingResultWriter.writeVerticesAsCsv;
import static de.cm.osm2po.service.ServiceHelpers.parseDoubleProperty;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_CLOSEST_OVERLAYNODE_NOT_FOUND;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_CLOSEST_VERTEX_NOT_FOUND;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_NO_DATA_FOUND;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.MlgExtension;
import de.cm.osm2po.hull.ConcaveHull;
import de.cm.osm2po.hull.ConvexHull;
import de.cm.osm2po.hull.HullSegment;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoJsonBuilder;
import de.cm.osm2po.misc.GeoJsonBuilder.FeatureCollection;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayEdge;
import de.cm.osm2po.routing.OverlayGraph;
import de.cm.osm2po.routing.OverlayNode;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.ServiceHandler;
import de.cm.osm2po.service.ServiceResourcesChecker;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * Behandelt Requests des {@link Osm2poService}.
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class DebugHandler implements ServiceHandler {
    @Override
    public ServiceResponse handle(final String cmd, Config config, Graph graph,
            Properties request, ServiceResourcesChecker lttc) throws Exception {

        byte[] response = null;
        Log log = config.getLog();

        // Konvention:
        // Ctrl und Alt (ohne Kombi) fragen zusaetzlichen DebugValue ab.
        if ("click".equals(cmd)) {
            response = new byte[0];

        } else if ("clickCS".equals(cmd)) { // Ctrl+Shift = Debug Vertex
            float lat = (float) parseDoubleProperty(request, "lat", 0);
            float lon = (float) parseDoubleProperty(request, "lon", 0);
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x400);
            this.debugVertex(baos, graph, lat, lon);
            response = baos.toByteArray();

        } else if ("clickAS".equals(cmd)) { // Alt+Shift = Debug Edge
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x400);
            float lat = (float) parseDoubleProperty(request, "lat", 0);
            float lon = (float) parseDoubleProperty(request, "lon", 0);
            this.debugEdge(baos, graph, log, lat, lon);
            response = baos.toByteArray();

        } else if ("clickA".equals(cmd)) { // Alt = Debug Expansion-Hull
            int maxSteps = (int) parseDoubleProperty(request, "debugValue", 0);
            response = this.debugExpansion(maxSteps, graph, log);

        } else if ("clickC".equals(cmd)) { // Ctrl = Debug level
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x400);
            byte level = (byte) parseDoubleProperty(request, "debugValue", 0);
            float lat = (float) parseDoubleProperty(request, "lat", 0);
            float lon = (float) parseDoubleProperty(request, "lon", 0);
            this.debugLevel(baos, graph, log, lat, lon, level);
            response = baos.toByteArray();

        } else if ("gp".equals(cmd)) { // Get POIs of one clazz
            ByteArrayOutputStream baos = new ByteArrayOutputStream(0x400);
            byte clazz = Byte.parseByte(request.getProperty(
                    "clazz", String.valueOf(Byte.MIN_VALUE)));
            int[] vids = graph.findVerticesByClazz(clazz);
            baos = new ByteArrayOutputStream(0x10000/*64k*/);
            writeVerticesAsCsv(baos, graph, vids);
            response = baos.toByteArray();
        }

        if (null == response) return null;
        else return new ServiceResponse(200, MIME_TEXT, response);
    }

    private void debugEdge(OutputStream os,
            Graph graph, Log log, float lat, float lon) throws Exception {
        OverlayNode tp = new OverlayNode(lat, lon);
        if (!tp.snap(graph, null))
            throw new ServiceException(ERR_CLOSEST_OVERLAYNODE_NOT_FOUND);
        log.debug(tp.toString());

        OverlayEdge[] overlayEdges = new OverlayEdge[0];
        if (tp.isOnVertex()) {
            int[] edgeIdxs = graph.findOutgoingEdges(tp.getId());
            overlayEdges = OverlayGraph.translatePath(graph, edgeIdxs);
        } else {
            int edgeIdx = tp.getEdgeIdx();
            int sourceId = graph.getEdgeSourceIds()[edgeIdx];
            int targetId = graph.getEdgeTargetIds()[edgeIdx];
            OverlayNode ndSource = new OverlayNode(sourceId);
            OverlayNode ndTarget = new OverlayNode(targetId);
            OverlayEdge overlayEdgeIn = OverlayNode.createOverlayEdge(graph, ndSource, tp);
            OverlayEdge overlayEdgeOut = OverlayNode.createOverlayEdge(graph, tp, ndTarget);
            if (overlayEdgeIn != null && overlayEdgeOut != null)
                overlayEdges = new OverlayEdge[]{overlayEdgeIn, overlayEdgeOut};
        }

        writePathAsFeatureCollection(os, graph, overlayEdges);
    }

    private void debugVertex(OutputStream os,
            Graph graph, float lat, float lon) throws Exception {
        int fromVertexIdx = graph.findClosestVertexId(lat, lon);
        if (fromVertexIdx < 1) throw new ServiceException(ERR_CLOSEST_VERTEX_NOT_FOUND);
        int[] edges = graph.findOutgoingEdges(fromVertexIdx);
        writePathAsFeatureCollection(os, graph, edges);
    }

    private void debugLevel(OutputStream os, Graph graph, Log log,
            float lat, float lon, byte level) throws Exception {

        MlgExtension vba = graph.getExtension(MlgExtension.class, (byte) 0);
        if (null == vba) {
            throw new UnsupportedOperationException("Graph does not contain MultiLevels");
        }
        byte[] edgeLevels = vba.getBytes();
        int[] edges = new int[0x1000];
        int n = 0;

        for (int edgeIdx : graph.findCloseEdges(lat, lon, (level + 1) * 0x400)) {
            if (edgeLevels[edgeIdx] == level) {
                if (n == edges.length) edges = Arrays.copyOf(edges, n * 2);
                edges[n++] = edgeIdx;
            }
        }
        log.debug(DF(n) + " Streets on level " + level + " found");
        edges = Arrays.copyOf(edges, n);
        if (n > 0) writePathAsFeatureCollection(os, graph, edges);
        else throw new ServiceException(
                ERR_NO_DATA_FOUND, "No street on level " + level);
    }


    private byte[] debugExpansion(int maxSteps, Graph graph, Log log)
    throws Exception {
        float[] lats = graph.getLats();
        float[] lons = graph.getLons();
        int[] vids = new int[lats.length -1];
        for (int i = 1; i < lats.length; i++) vids[i-1] = i;

        float minLat = graph.getMinLat(), minLon = graph.getMinLon();
        float maxLat = graph.getMaxLat(), maxLon = graph.getMaxLon();
        LatLons boundingRect = new LatLons()
            .append(new LatLon(minLat, minLon))
            .append(new LatLon(minLat, maxLon))
            .append(new LatLon(maxLat, maxLon))
            .append(new LatLon(maxLat, minLon))
            .append(new LatLon(minLat, minLon));

        ConvexHull convexHull = new ConvexHull(graph, log);
        int[] convexPoints = convexHull.build(vids);
        LatLons convexLatLons = this.vertexIdsToRing(convexPoints, lats, lons);

        ConcaveHull concaveHull = new ConcaveHull(graph, log);
        int[] concavePoints = concaveHull.build(vids, 0.05f, maxSteps);
        LatLons concaveLatLons = this.vertexIdsToRing(concavePoints, lats, lons);

        LatLons nextSegment = new LatLons();

        int[] nextRasterCells = new int[0];
        HullSegment hsNext = concaveHull.peekNext();
        if (hsNext != null) {
            nextSegment = hsNext.toLine(lats, lons);
            nextRasterCells = concaveHull.getHullRaster().rasterLine(
                    hsNext.getVertexId(), hsNext.getNext().getVertexId(), 0.01f);
        }

        Properties blue = new Properties();
        blue.setProperty("color", "blue");
        Properties black = new Properties();
        black.setProperty("color", "black");
        Properties gray = new Properties();
        gray.setProperty("color", "gray");
        Properties red = new Properties();
        red.setProperty("color", "red");
        Properties green = new Properties();
        green.setProperty("color", "green");

        GeoJsonBuilder geoJson = new GeoJsonBuilder();
        FeatureCollection fc = geoJson.featureCollection()
            .addLine(concaveLatLons).with("style", blue).with("name", "ConcaveHull").endWith()
            .addLine(convexLatLons).with("style", black).with("name", "ConvexHull").endWith()
            .addLine(boundingRect).with("style", gray).with("name", "BoundingBox").endWith()
            .addLine(nextSegment).with("style", red).with("name", "NextSegment").endWith();

        for (int i = 0; i < nextRasterCells.length; i++) {
            LatLons rect = concaveHull.getHullRaster().toRect(nextRasterCells[i]);
            fc.addLine(rect).with("style", green);
        }

        return geoJson.build().getBytes();
    }

    private LatLons vertexIdsToRing(int[] vids, float[] lats, float[] lons) {
        LatLons ring = new LatLons();
        for (int i = 0; i <= vids.length; i++) {
            int vid = vids[i % vids.length];
            ring.append(new LatLon(lats[vid], lons[vid]));
        }
        return ring;
    }

}