/*
 * $Id: GeoHandler.java 5712 2022-09-23 07:52:26Z carsten $
 *
 * Copyright 2015 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.handlers;

import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_COUNTRY;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_DISTRICT;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_METROPOLE;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_STATE;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_SUBURB;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_ADMIN_TOWN;
import static de.cm.osm2po.converter.tiler.RelationTagResolver.CLAZZ_BOUNDARY_POSTAL;
import static de.cm.osm2po.misc.GeoUtils.calcDistance;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.misc.Utils.roundE5;
import static de.cm.osm2po.model.LatLon.latOf;
import static de.cm.osm2po.model.LatLon.lonOf;
import static de.cm.osm2po.service.ServiceHelpers.parseBounding;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_NO_DATA_FOUND;
import static java.lang.Double.parseDouble;
import static java.lang.Integer.parseInt;

import java.util.Arrays;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.extensions.BndExtension;
import de.cm.osm2po.extensions.PoiExtension;
import de.cm.osm2po.extensions.SndExtension;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.misc.GeoJsonBuilder;
import de.cm.osm2po.misc.GeoJsonBuilder.FeatureCollection;
import de.cm.osm2po.model.Bounding;
import de.cm.osm2po.model.LatLon;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.model.OsmNode;
import de.cm.osm2po.model.OsmPolygon;
import de.cm.osm2po.primitives.VarString;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.ServiceHandler;
import de.cm.osm2po.service.ServiceResourcesChecker;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ErrorEnum;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * Behandelt Requests des {@link Osm2poService}.
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class GeoHandler implements ServiceHandler {

    private Log log;

    @Override
    public ServiceResponse handle(final String cmd, Config config, Graph graph,
            Properties request, ServiceResourcesChecker lttc) throws Exception {

        this.log = config.getLog();

        String subCmd = cmd;
        boolean debugUi = false;
        if ("clickCA".equals(cmd) || "clickCAS".equals(cmd)) {
            debugUi = true;
            // CA=Point CAS=Boundary
            subCmd = "clickCA".equals(cmd) ? "fgpc" : "fgbc";
            String lat = request.getProperty("lat");
            String lon = request.getProperty("lon");
            request.setProperty("geocode", lat + "," + lon);
        }

        if (subCmd != null && subCmd.startsWith("fg")) { // Find GeoCode aus UI

            String geocodeParam = request.getProperty("geocode", "?");
            if (0 == geocodeParam.length()) geocodeParam = "?";
            if ("fg".equals(subCmd)) {
                switch (geocodeParam.charAt(0)) { // Convention: NameSearch must endWith('n')
                case '*': subCmd = "fgpn"; break; // Find Geo by Poi by Name
                case '@': subCmd = "fgbn"; break; // Find Geo by Bounary by Name
                case '~': subCmd = "fgsn"; break; // Find Geo by Street by Name
                case ':': subCmd = "fgbx"; break; // Find Geo by Boundary by Index
                case ';': subCmd = "fgbc"; break; // Find Geo by Boundary by Coordinate
                case ',': subCmd = "fgpc"; break; // Find Geo by Poi by Coordinate
                default: subCmd = "fgpn"; break;} // Find Geo by Poi by Name - Default
                if (!"fgpn".equals(subCmd)) geocodeParam = geocodeParam.substring(1);
            }

            String [] geocodes = geocodeParam.split(",");
            boolean[] exacts = new boolean[geocodes.length];
            for (int i = 0; i < geocodes.length; i++) {
                String geocode = geocodes[i].trim();
                if (geocode.endsWith("$")) {
                    geocode = geocode.substring(0, geocode.length() -1);
                    exacts[i] = true;
                }
                int minLen = subCmd.endsWith("n") ? 3 : 1;
                if (geocode.length() < minLen) throw new IllegalArgumentException(
                        "At least 3 characters expected: '" + geocode + "'");
                geocodes[i] = geocode;
            }

            boolean bndOptional = false;

            SndExtension sex = null;
            if (subCmd.equals("fgsn")) {
                bndOptional = true;
                sex = graph.getFirstExtension(SndExtension.class);
                if (null == sex) throw new ServiceException(
                        ErrorEnum.ERR_UNEXPECTED_UNSUPPORTED_OPERATION,
                        "Graph does not contain " + SndExtension.class.getSimpleName());
            }

            BndExtension bex = null;
            if (bndOptional || subCmd.startsWith("fgb")) {
                bex = graph.getFirstExtension(BndExtension.class);
                if (null == bex && !bndOptional) throw new ServiceException(
                        ErrorEnum.ERR_UNEXPECTED_UNSUPPORTED_OPERATION,
                        "Graph does not contain " + BndExtension.class.getSimpleName());
            }

            PoiExtension pex = null;
            if (subCmd.startsWith("fgp")) {
                pex = graph.getFirstExtension(PoiExtension.class);
                if (null == pex) throw new ServiceException(
                        ErrorEnum.ERR_UNEXPECTED_UNSUPPORTED_OPERATION,
                        "Graph does not contain " + PoiExtension.class.getSimpleName());
            }

            Bounding bbox = parseBounding(request);

            byte[] response = null;
            if (subCmd.equals("fgbn")) {
                response = this.findBoundsByName(bex, exacts, bbox, geocodes);
            } else if (subCmd.equals("fgbc")) {
                response = this.findBoundsByCoordinate(bex, bbox, debugUi, geocodes);
            } else if (subCmd.equals("fgbx")) {
                response = this.findBoundsByIndex(bex, bbox, geocodes);
            } else if (subCmd.equals("fgsn")) {
                response = this.findStreetsByName(graph, bex, sex, exacts, bbox, geocodes);
            } else if (subCmd.equals("fgpn")) {
                response = this.findPoisByName(pex, exacts, geocodes);
            } else if (subCmd.equals("fgpc")) {
                response = this.findPoiByCoordinate(pex, exacts, geocodes);
            }

            if (null == response) {
                throw new ServiceException(ERR_NO_DATA_FOUND, "cmd=" + subCmd);
            } else {
                return new ServiceResponse(HTTP_200, MIME_JSON, response);
            }
        }

        return null;
    }

    protected byte[] findStreetsByName(Graph graph, BndExtension bex,
            SndExtension sex, boolean[] exacts, Bounding bbox, String...geocodes) {

        int nFound = 0;
        FeatureCollection fc = new GeoJsonBuilder().featureCollection();

        if (geocodes != null && geocodes.length > 0) {
            int[] edgeIdxs = sex.findByName(geocodes[0], exacts[0]);
            VarString[] edgeBoundsNames = null;
            if (edgeIdxs != null) {
                edgeBoundsNames = new VarString[edgeIdxs.length];
                if (geocodes.length == 1) {
                    this.log.warn("StreetFinder uses viewport " + bbox);
                    for (int i = 0; i < edgeIdxs.length; i++) {
                        LatLons coords = graph.lookupGeometry(edgeIdxs[i]);
                        if (bbox.contains(coords.center()))
                            edgeIdxs[nFound++] = edgeIdxs[i];
                    }
                } else {
                    int[] ids = bex.findByName(geocodes[1], exacts[1]);
                    int nBounds = null == ids ? 0 : ids.length;
                    OsmPolygon bounds[] = new OsmPolygon[nBounds];
                    for (int i = 0; i < nBounds; i++)
                        bounds[i] = bex.findByIndex(ids[i], new OsmPolygon());
                    for (int i = 0; i < edgeIdxs.length; i++) {
                        LatLon center = graph.lookupGeometry(edgeIdxs[i]).center();
                        for (int j = 0; j < nBounds; j++) {
                            if (bounds[j].contains(center.getLat(), center.getLon())) {
                                edgeIdxs[nFound] = edgeIdxs[i];
                                edgeBoundsNames[nFound] = bounds[j].getName();
                                nFound++;
                                // Eine Edge kann in mehreren Polygonen liegen, daher:
                                break;
                            }
                        }
                    }
                }
            }

            if (nFound > 10000) {
                this.log.warn("Too many edges found. Showing the first 10000");
                nFound = 10000;
            }

            for (int i = 0; i < nFound; i++) {
                int edgeIdx = edgeIdxs[i];
                fc.addLine(graph.lookupGeometry(edgeIdx))
                .with("name", graph.lookupEdgeName(edgeIdx))
                .with("index", edgeIdx)
                .with("in", edgeBoundsNames[i]);
            }
        }

        return (nFound > 0) ? fc.build().getBytes() : null;
    }

    protected byte[] findBoundsByName(
            BndExtension bex, boolean[] exacts, Bounding bbox, String... geocodes) {

        boolean foundAny = false;
        FeatureCollection fc = new GeoJsonBuilder().featureCollection();

        if (geocodes != null && geocodes.length > 0) {
            int[] ids = bex.findByName(geocodes[0], exacts[0]);
            int nBounds = null == ids ? 0 : ids.length;
            OsmPolygon bounds[] = new OsmPolygon[nBounds];
            for (int i = 0; i < nBounds; i++)
                bounds[i] = bex.findByIndex(ids[i], new OsmPolygon());
            if (bounds.length > 1 && geocodes.length > 1) {
                ids = bex.findByName(geocodes[1], exacts[1]);
                nBounds = null == ids ? 0 : ids.length;
                OsmPolygon bounds2[] = new OsmPolygon[nBounds];
                for (int i = 0; i < nBounds; i++)
                    bounds2[i] = bex.findByIndex(ids[i], new OsmPolygon());
                bounds = this.matchPolygons(bounds, bounds2);
            }

            Arrays.sort(bounds); // Laender oben, Kaeffer unten.
            Properties style = new Properties();
            for (int i = 0; i < bounds.length; i++) {
                foundAny = true;
                OsmPolygon plgs = bounds[i];
                for (LatLons[] plg : plgs.getPolygons()) {
                    style.setProperty("color", colorForBoundsClazz(plgs.getClazz()));
                    fc.addPoly(plg)
                    .with("name", plgs.getName())
                    .with("clazz", plgs.getClazz())
                    .with("lat", latOf(plgs.getCentroid()))
                    .with("lon", lonOf(plgs.getCentroid()))
                    .with("osmid", plgs.getId())
                    .with("index", plgs.getIndex())
                    .with("style", style);
                }
            }
        }
        return foundAny ? fc.build().getBytes() : null;
    }

    protected byte[] findBoundsByIndex(
            BndExtension gex, Bounding bbox, String... geocodes) {
        FeatureCollection fc = new GeoJsonBuilder().featureCollection();
        OsmPolygon plgs = gex.findByIndex(parseInt(geocodes[0].trim()), new OsmPolygon());
        if (null == plgs) return null;
        for (LatLons[] plg : plgs.getPolygons()) fc.addPoly(plg)
//        .with("color", "gray")
        .with("name", plgs.getName())
        .with("clazz", plgs.getClazz())
        .with("lat", latOf(plgs.getCentroid()))
        .with("lon", lonOf(plgs.getCentroid()))
        .with("size", plgs.getAreaSize())
        .with("osmid", plgs.getId())
        .with("index", plgs.getIndex());
        return fc.build().getBytes();
    }

    protected byte[] findBoundsByCoordinate(
            BndExtension gex, Bounding bbox, boolean debugUi, String... geocodes) {

        boolean foundAny = false;
        FeatureCollection fc = new GeoJsonBuilder().featureCollection();

        if (geocodes != null && geocodes.length == 2) {
            double lat = parseDouble(geocodes[0].trim());
            double lon = parseDouble(geocodes[1].trim());
            OsmPolygon[] polygons = gex.findByCoordinate(lat, lon, null);

            Properties props = new Properties();
            Properties style = new Properties();
            Arrays.sort(polygons); // Laender oben, Kaeffer unten.
            for (int i = 0; i < polygons.length; i++) {
                foundAny = true;
                OsmPolygon plgs = polygons[i];

                if (debugUi) props.setProperty(
                        "Clazz:" + plgs.getClazz(), plgs.getName().toString());

                for (LatLons[] plg : plgs.getPolygons()) {
                    style.setProperty("color", colorForBoundsClazz(plgs.getClazz()));
                    (debugUi ? fc.addLines(plg) : fc.addPoly(plg))
                        .with("name", plgs.getName())
                        .with("clazz", plgs.getClazz())
                        .with("lat", latOf(plgs.getCentroid()))
                        .with("lon", lonOf(plgs.getCentroid()))
                        .with("size", plgs.getAreaSize())
                        .with("osmid", plgs.getId())
                        .with("index", plgs.getIndex())
                        .with("style", style);
                }
            }

            style.setProperty("color", "red");
            if (debugUi) fc.addPoint(new LatLon(lat, lon))
                .with("name", roundE5(lat) + "," + roundE5(lon))
                .with(props)
                .with("style", style);
        }

        return foundAny ? fc.build().getBytes() : null;
    }


    protected static String colorForBoundsClazz(byte clazz) {
        switch (clazz) {
        case CLAZZ_BOUNDARY_ADMIN_COUNTRY: return "gray"; // Deutschland
        case CLAZZ_BOUNDARY_ADMIN_STATE: return "blue"; // Hamburg
        case CLAZZ_BOUNDARY_ADMIN_DISTRICT: return "gray"; // Bezirk
        case CLAZZ_BOUNDARY_ADMIN_METROPOLE: return "red"; // Muenchen
        case CLAZZ_BOUNDARY_ADMIN_TOWN: return "green"; // Pinneberg
        case CLAZZ_BOUNDARY_POSTAL: return "orange"; // 25421
        case CLAZZ_BOUNDARY_ADMIN_SUBURB: return "blue"; // Barmbek
        default: return "black";}
    }

    /**
     * Uebeprueft, ob Boundaries einer Menge in eine der
     * zweiten Menge passen. Dies ist keine polygone Berechnung
     * der Ueberscheidungen, sondern hier wird mittels Centroid
     * und Groesse der Flaechen lediglich Annahmen getroffen.
     * @param plgs1 Array von {@link OsmPolygon}
     * @param plgs2 Array von {@link OsmPolygon}
     * @return Array von {@link OsmPolygon}
     */
    protected OsmPolygon[] matchPolygons(OsmPolygon[] plgs1, OsmPolygon[] plgs2) {
        OsmPolygon[] plgs = new OsmPolygon[plgs1.length];
        int n = 0;
        for (int i = 0; i < plgs1.length; i++) {
            LatLon centroid1 = new LatLon().setCoord(plgs1[i].getCentroid());
            for (int j = 0; j < plgs2.length; j++) {
                if (plgs2[j].contains(centroid1.getLat(), centroid1.getLon())) {
                    if (plgs2[j].getAreaSize() > plgs1[i].getAreaSize()) {
                        plgs[n++] = plgs1[i];
                        break;
                    }
                }
            }
        }
        return Arrays.copyOf(plgs, n);
    }


    protected byte[] findPoisByName(
            PoiExtension pex, boolean[] exacts, String... geocodes) {
        boolean foundAny = false;
        FeatureCollection fc = new GeoJsonBuilder().featureCollection();

        if (geocodes != null && geocodes.length > 0) {
            int[] ids = pex.findByName(geocodes[0], exacts[0]);
            int nPois = null == ids ? 0 : ids.length;
            for (int i = 0; i < nPois; i++) {
                foundAny = true;
                OsmNode poi = pex.findByIndex(ids[i], new OsmNode());
                fc.addPoint(poi)
                .with("name", poi.getName())
                .with("clazz", poi.getClazz())
                .with("lat", poi.getLat())
                .with("lon", poi.getLon())
                .with("osmid", poi.getId());
            }

        }
        return foundAny ? fc.build().getBytes() : null;
    }

    protected byte[] findPoiByCoordinate(
            PoiExtension pex, boolean[] exacts, String... geocodes) {
        if (geocodes != null && geocodes.length == 2) {
            double lat = parseDouble(geocodes[0].trim());
            double lon = parseDouble(geocodes[1].trim());
            // TODO findByCoordinate unterstuetzt jetzt auch NodeClazz-Filter
            OsmNode poi = pex.findClosestPoiByCoordinate(lat, lon, null);
            if (poi != null) {
                double dist = calcDistance(lat, lon, poi.getLat(), poi.getLon());
                return new GeoJsonBuilder()
                    .pointFeature(new LatLon(poi.getLat(), poi.getLon()))
                    .with("name", poi.getName())
                    .with("dist", DF2(dist) + " km")
                    .with("clazz", poi.getClazz())
                    .with("lat", poi.getLat())
                    .with("lon", poi.getLon())
                    .with("osmid", poi.getId())
                    .build().getBytes();
            }
        }
        return null;
    }

}