/*
 * $Id: RouteHandler.java 5502 2022-02-06 15:25:33Z carsten $
 *
 * Copyright 2021,2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.handlers;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.DF2;
import static de.cm.osm2po.routing.RoutingResultWriter.writePathAsCsv;
import static de.cm.osm2po.routing.RoutingResultWriter.writePathAsFeatureCollection;
import static de.cm.osm2po.routing.RoutingResultWriter.writeVerticesAsCsv;
import static de.cm.osm2po.routing.RoutingResultWriter.writeVerticesAsMultiPoint;
import static de.cm.osm2po.routing.RoutingResultWriter.writeVerticesAsPolygon;
import static de.cm.osm2po.service.ServiceHelpers.parseDoubleProperty;
import static de.cm.osm2po.service.ServiceHelpers.parseNode;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_ROUTER_NOT_FOUND;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_ROUTE_NOT_FOUND;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_SOURCE_TARGET_EQUAL;
import static de.cm.osm2po.service.errors.ErrorEnum.ERR_TOO_FEW_POINTS;

import java.io.ByteArrayOutputStream;
import java.util.Date;
import java.util.Properties;

import de.cm.osm2po.Config;
import de.cm.osm2po.hull.ConcaveHull;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.routing.AbstractRouter;
import de.cm.osm2po.routing.Graph;
import de.cm.osm2po.routing.OverlayEdge;
import de.cm.osm2po.routing.OverlayGraph;
import de.cm.osm2po.routing.OverlayNode;
import de.cm.osm2po.routing.RouterParameters;
import de.cm.osm2po.routing.VisitedVerticesAware;
import de.cm.osm2po.service.Osm2poService;
import de.cm.osm2po.service.ServiceHandler;
import de.cm.osm2po.service.ServiceResourcesChecker;
import de.cm.osm2po.service.ServiceResponse;
import de.cm.osm2po.service.errors.ServiceException;

/**
 * Behandelt Requests des {@link Osm2poService}.
 * @author (c) 2015 - Carsten Moeller - info@osm2po
 */
public class RouteHandler implements ServiceHandler {
    @Override
    public ServiceResponse handle(final String cmd, Config config, Graph staticGraph,
            Properties request, ServiceResourcesChecker lttc)
    throws Exception {

        boolean isGeoJson = "geojson".equals((request.getProperty("format")));
        Log log = config.getLog();

        // FindRoute / Vertices / Hull
        if ("fr".equals(cmd) || "fx".equals(cmd) || "fh".equals(cmd)) {
            lttc.checkResources();

            final int baosSize = isGeoJson ? 0x200000/*2M*/ : 0x20000/*128k*/;
            ByteArrayOutputStream baos = new ByteArrayOutputStream(baosSize);

            // Allgemeine Parameter abgreifen
            double maxCost = parseDoubleProperty(request, "maxCost", 0);
            parseDoubleProperty(request, "heuristicFactor", 0d); // nur fuer Validation
            parseDoubleProperty(request, "localityFactor", 0d); // nur fuer Validation
            boolean showVisited = "fx".equals(cmd);
            boolean showHull = "fh".equals(cmd);
            String strSource = request.getProperty("source", "0");
            String strTarget = request.getProperty("target", "0");

            // Router ermitteln
            int routerId = (int) parseDoubleProperty(request, "routerId", 0d);
            Class<? extends AbstractRouter<?>> routerClass =
                config.getRouterClasses().get(routerId);
            if (null == routerClass)
                throw new ServiceException(ERR_ROUTER_NOT_FOUND, routerId);

            Class<?> graphClass = AbstractRouter.graphClassOf(routerClass);
            boolean isOverlayRouter = OverlayGraph.class == graphClass;

            // OverlayNodes einfuegen - wenn kein Overlay, dann dichtesten Vertex in Graph
            OverlayNode ndSource = parseNode(staticGraph, strSource, isOverlayRouter, true);
            OverlayNode ndTarget = parseNode(staticGraph, strTarget, isOverlayRouter, false);

            OverlayGraph overlayGraph = new OverlayGraph(staticGraph); // Egal ob Overlay oder nicht
            int sourceId = overlayGraph.insertNode(ndSource).getId(); // neue Id nach Einfuegen
            int targetId = overlayGraph.insertNode(ndTarget).getId(); // neue Id nach Einfuegen
            if (sourceId == targetId) throw new ServiceException(ERR_SOURCE_TARGET_EQUAL);

            // Router mit dem korrekten Graphen bauen
            AbstractRouter<?> router = AbstractRouter.newAbstractRouter(
                    routerClass, isOverlayRouter ? overlayGraph : staticGraph, log);

            // Jetzt Route berechnen
            long t1 = System.currentTimeMillis();
            log.note("Routing from " + sourceId + " to " + targetId + " using "
                    + router.getClass().getSimpleName() + " " + DF(new Date(t1)));

            int[] path = router.findPath(sourceId, targetId, maxCost, new RouterParameters(request));
            long t2 = System.currentTimeMillis();
            log.note("Route calculated in: " + ((t2 - t1) / 1000d) + "s");

            if (!showHull && !showVisited) { // Pfad-Geometrie erzeugen
                if (null == path) {
                    log.warn("Route not found");
                    throw new ServiceException(ERR_ROUTE_NOT_FOUND);
                }

                OverlayEdge[] route = overlayGraph.translatePath(path);

                if (isGeoJson) writePathAsFeatureCollection(baos, staticGraph, route);
                else writePathAsCsv(baos, staticGraph, route);

                int nSegs = path.length;
                double length = overlayGraph.calcPathLength(path);
                double time = overlayGraph.calcPathTime(path);
                long t3 = System.currentTimeMillis();
                log.debug("Geometry created in: " + ((t3 - t2) / 1000d) + "s");

                log.note("Segments:" + DF(nSegs) + " Dist:"
                        + DF2(length) + "km" + " Time:" + DF2(time) + "h");

            } else { // Hull or VisitedVertices
                if (!(router instanceof VisitedVerticesAware)) {
                    throw new IllegalStateException(router.getClass().getName()
                            + " does not implement " + VisitedVerticesAware.class.getSimpleName());
                }
                int[] visited = ((VisitedVerticesAware)router).getVisited();
                log.note("Number of visited Vertices: " + DF(visited.length));

                if (showHull) { // Punkte auf Konvexe Huelle reduzieren
                    if (visited.length < 3) throw new ServiceException(ERR_TOO_FEW_POINTS);
                    float alpha = (float) parseDoubleProperty(request, "hullAlpha", 0d);
                    visited = new ConcaveHull(staticGraph, log).build(visited, alpha);
                    log.note("Number of Hull Points: " + DF(visited.length));
                }

                if (!isGeoJson) writeVerticesAsCsv(baos, staticGraph, visited);
                else if (showHull) writeVerticesAsPolygon(baos, staticGraph, visited);
                else writeVerticesAsMultiPoint(baos, staticGraph, visited);

                long t3 = System.currentTimeMillis();
                log.debug("Geometry created in: " + ((t3 - t2) / 1000d) + "s");
            }

            byte[] response = baos.toByteArray();
            return new ServiceResponse(HTTP_200, isGeoJson ? MIME_JSON : MIME_TEXT, response);
        }

        return null;
    }

}