/*
 * $Id: RateLimit.java 5420 2021-11-16 20:31:42Z carsten $
 *
 * Copyright 2017 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.security;

import static de.cm.osm2po.primitives.VarNum.readUnsigned;
import static de.cm.osm2po.primitives.VarNum.writeUnsigned;
import static java.lang.Integer.parseInt;
import static java.lang.Long.parseLong;
import static java.lang.System.currentTimeMillis;

import java.sql.Date;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import de.cm.osm2po.primitives.InStream;
import de.cm.osm2po.primitives.OutStream;
import de.cm.osm2po.primitives.Var;

/**
 * Merkt sich Frequenz und Zeitstempel von Aufrufen und kann
 * diese validieren.<br>
 * Ueberwacht hierzu die Anfragen pro Zeitintervall und wendet
 * hierauf eine einfache Regel an.
 * @author (c) 2017 - Carsten Moeller, info@osm2po.de
 * <br><br><b>Wichtig:</b><br>
 * Objekte dieser Klasse sollten direkt beim ersten Aufruf instanziiert
 * werden, da hier die interne Stoppuhr zu ticken beginnt.
 * @see #RateLimit(boolean, int, long, long)
 * @see #validate()
 * @see RateLimiter
 */
public class RateLimit implements Var<RateLimit> {
    private final static DateFormat DF = new SimpleDateFormat("yyyy-MM-dd");
    private final static long VT_2100() {
        long vt2100 = 4102441200000L; // je nach Zone 2100-01-01 oder 2099-12-31
        try { return DF.parse("2100-01-01").getTime();
        } catch (ParseException ignore){}
        return vt2100;
    }

    public final static long VALIDTHRU_1970 = 0;
    public final static long VALIDTHRU_2100 = VT_2100();

    public final static int KEY_INITIAL = 1;
    public final static int KEY_OK = 0;
    public final static int KEY_RATELIMIT = -1;
    public final static int KEY_EXPIRED = -2;
    public final static int KEY_UNAUTHORIZED = -3;
    public final static int KEY_DDOS = -4;

    // Konstruktor ARGS
    private int limit; // Max number of Request
    private long interval; // within this interval in MilliSeconds
    private long expires; // Date in Millis
    private boolean declared; // Allows detection of source/creator

    // Runtime Variables
    private long created; // TimeStamp MilliSeconds
    private long lastCall; // TimeStamp MilliSeconds, currentTime
    private long intervalStart; // TimeStamp MilliSeconds
    private long intervalEnd; // TimeStamp MilliSeconds
    private long limitCounter; // Increment
    private long malCounter; // Anzahl verweigerter Aufrufe
    private int status; // Ergebnis nach letzten validate()

    /**
     * Nur fuer Serialisierung gedacht.<br>
     * Loest KEIN {@link #resetMonitoring()} aus und
     * setzt auch keine internen Variablen!
     */
    public RateLimit() {
        this.created = currentTimeMillis();
        this.status = KEY_INITIAL;
    }

    /**
     * Parameter-Konstruktor.
     * @param declared boolean true: Aus z.B. KeyStore
     * @param limit int Maximale Anzahl von Aufrufen
     * @param interval long (Millisekunden) In diesem Zeitinterval
     * @param expires long Absolutes Ablaufdatum in Millis.
     * @see RateLimit
     * @see #RateLimit(boolean, String)
     */
    public RateLimit(boolean declared, int limit, long interval, long expires) {
        this();
        this.updateConstraints(declared, limit, interval, expires);
        this.resetMonitoring();
    }

    /**
     * CSV-Konstruktor, der einen Strimg-Parameter parst und an
     * den eigentlichen Konstruktor deligiert.
     * @param declared boolean true: Aus z.B. KeyStore
     * @param csv Parameter fuer {@link #RateLimit(boolean, int, long, long)}
     *      <br>Format:<br><code>Anzahl, ZeitInterval, ZeitStrafe, JJJJ-MM-TT</code>.
     *      <br>Die Zeitangaben sind hier in Sekunden, statt in Millisekunden.
     *      <br>Parameter 2-4 sind optional und werden genullt, bzw. fuer
     *      expires in die Zukunft gelegt.
     * @see #updateConstraintsCsv(boolean, String)
     */
    public RateLimit(boolean declared, String csv) {
        this();
        this.updateConstraintsCsv(declared, csv);
        this.resetMonitoring();
    }

    /**
     * Setzt die Ueberwachungswerte und die Stoppuhr
     * auf Initial und Jetzt zurueck.
     * Ueberschreibt NICHT den letzten {@link #getStatus()}
     * @see #updateConstraints(boolean, int, long, long)
     */
    public final void resetMonitoring() {
        long now = currentTimeMillis();
        this.moveInterval(now);
        this.lastCall = now;
        this.limitCounter = 0;
        this.malCounter = 0;
    }

    /**
     * Ueberschreibt die Ueberwachungsregel zur Laufzeit.
     * @param declared boolean true: Aus z.B. KeyStore
     * @param limit int Maximale Anzahl von Aufrufen
     * @param interval long (Millisekunden) In diesem Zeitinterval
     * @param expires Absolutes Ablaufdatum in Millis.
     * @return true: Regel hat sich geaendert, false: Nicht modifiziert.
     * @see #resetMonitoring()
     * @see #RateLimit(boolean, int, long, long)
     * @see #updateConstraintsCsv(boolean, String)
     */
    public final boolean updateConstraints(boolean declared,
            int limit, long interval, long expires) {

        boolean modified = declared != this.declared
        || limit != this.limit
        || interval != this.interval
        || expires != this.expires;

        if (modified) {
            // PITFALL wenn vorher nicht OK, dann wirkt neues interval als Zusatzzeit
            if (KEY_INITIAL != this.status) this.status = KEY_OK;
            this.declared = declared;
            this.limit = limit;
            this.interval = interval;
            this.expires = expires;
        }

        return modified;
    }

    /**
     * Ueberschreibt die Ueberwachungsregel zur Laufzeit und benutzt
     * die Grenzwerte aus einer Vorlage.
     * @param rl {@link RateLimit} Vorlage, NULLABLE
     * @param declared boolean true: Aus z.B. KeyStore
     * @see #resetMonitoring()
     * @see #RateLimit(boolean, int, long, long)
     * @return true: Regel hat sich geaendert, false: Nicht modifiziert.
     */
    public final boolean updateConstraints(boolean declared, RateLimit rl) {
        if (null == rl) return false;
        return this.updateConstraints(declared, rl.limit, rl.interval, rl.expires);
    }

    /**
     * Ueberschreibt die CSV-Ueberwachungsregel zur Laufzeit.
     * @param declared boolean true: Aus z.B. KeyStore
     * @param csv Parameter fuer {@link #RateLimit(boolean, int, long, long)}
     *      <br>Format:<br><code>Anzahl, ZeitInterval, GueltigBis(JJJJ-MM-TT)</code>.
     *      <br>Die Zeitangaben sind hier in Sekunden, statt in Millisekunden.
     *      <br>Parameter 2-3 sind optional und werden genullt, bzw. fuer
     *      expires in die Zukunft gelegt.
     * @return true: Regel hat sich geaendert, false: Nicht modifiziert.
     * @see #resetMonitoring()
     * @see #RateLimit(boolean, String)
     */
    public final boolean updateConstraintsCsv(boolean declared, String csv) {
        try {
            int limit = 0;
            long interval = 0;
            long expires = 0;

            String[] params = csv.split(",");
            for (int i = 0; i < params.length; i++) {
                String param = params[i].trim();
                switch (i) {
                case 0: limit = parseInt(param); break;
                case 1: interval = parseLong(param) * 1000; break;
                case 2: expires = DF.parse(param).getTime(); break;
                }
            }

            if (VALIDTHRU_1970 == expires && params.length < 4) {
                expires = VALIDTHRU_2100;
            }

            return this.updateConstraints(declared, limit, interval, expires);

        } catch (Exception e) {
            throw new IllegalArgumentException(csv);
        }
    }

    public final boolean isDeclared() {return this.declared;}
    public final long getAge() {return currentTimeMillis() - this.created;}
    public final long getIdle() {return currentTimeMillis() - this.lastCall;}
    public final long getMalCounter() {return this.malCounter;}
    public final long getIntervalEnd() {return this.intervalEnd;}
    public final long getExpires() {return this.expires;}
    public final long getLimitCounter() {return this.limitCounter;}
    public final int getStatus() {return this.status;}

    public final void decreaseLimitCounter() {
        if (this.limitCounter > 0) this.limitCounter--;
    }

    private final void moveInterval(long now) {
        this.intervalStart = now;
        this.intervalEnd = now + this.interval;
    }

    /**
     * Validiert gegen die hier geltende Regel und aktualisiert
     * den Status, die internen Zaehler und Zeitstempel.
     * @see #RateLimit(boolean, int, long, long)
     */
    public final void validate() {
        long now = currentTimeMillis();

        this.lastCall = now; // Fuer idle-check

        if (KEY_INITIAL == this.status) {
            this.status = KEY_OK;
            this.moveInterval(now); // Kein Zeitbonus fuer 1. Aufruf
        }

        // ShortCutCheck - Aufruf komplett untersagt
        if (this.limit <= 0) {
            this.malCounter++;
            this.status = KEY_UNAUTHORIZED;
            return;
        }

        // ShortCutCheck - Konto ist abgelaufen
        if (this.lastCall > this.expires) {
            this.malCounter++;
            this.status = KEY_EXPIRED;
            return;
        }

        // ShortCutCheck - Keine Begrenzung, Intervall unendlich
        if (this.interval <= 0) {
            this.status = KEY_OK;
            return;
        }

        // Anzahl Aufrufe im Intervall bereits ueberschritten?
        if (!(KEY_OK == this.status)) {
            if (now > this.intervalEnd) { // Lange genug gewartet?
                this.status = KEY_OK; // Rehabilitieren
                this.limitCounter = 1; // Erster Aufruf im Intervall
                this.malCounter = 0; // Nicht (mehr) boese
                this.moveInterval(now);

            } else { // Sonst Boese-Zaehler hochsetzen
                // status nicht setzen, da auch DDOS & Co moeglich
                this.malCounter++;
            }

        // Sonst hochzaehlen und maximaler Aufrufe validieren
        } else if (++this.limitCounter > this.limit) {
            if (now < this.intervalEnd) { // Intervall noch nicht abgelaufen
                this.status = KEY_RATELIMIT;
                this.malCounter++;
            } else { // alles ok
                this.status = KEY_OK; // Rehabilitieren
                this.limitCounter = 1; // Erster Aufruf im Intervall
                this.malCounter = 0; // Nicht (mehr) boese
                this.moveInterval(now);
            }
        }
    }

    public final void penalizeDdos(long penalty) {
        this.status = KEY_DDOS;
        this.intervalEnd += penalty;
    }

    @Override
    public String toString() {
        return "[" + this.limit + '|' + (this.interval / 1000) + '|'
        + DF.format(new Date(this.expires)) + ']';
    }

    @Override
    public void writeToStream(OutStream outStream) {
        outStream.writeByte(this.declared ? (byte)1 : (byte)0);
        writeUnsigned(this.created, outStream);
        writeUnsigned(this.limit, outStream);
        writeUnsigned(this.interval, outStream);
        writeUnsigned(this.expires, outStream);
        writeUnsigned(this.intervalStart, outStream);
        writeUnsigned(this.intervalEnd, outStream);
        writeUnsigned(this.lastCall, outStream);
        writeUnsigned(this.limitCounter, outStream);
        writeUnsigned(this.malCounter, outStream);
        writeUnsigned(this.status, outStream);
    }

    @Override
    public RateLimit readFromStream(InStream inStream) {
        this.declared = inStream.readByte() != 0;
        this.created = readUnsigned(inStream);
        this.limit = (int) readUnsigned(inStream);
        this.interval = readUnsigned(inStream);
        this.expires = readUnsigned(inStream);
        this.intervalStart = readUnsigned(inStream);
        this.intervalEnd = readUnsigned(inStream);
        this.lastCall = readUnsigned(inStream);
        this.limitCounter = readUnsigned(inStream);
        this.malCounter = readUnsigned(inStream);
        this.status = (int) readUnsigned(inStream);
        return this;
    }

}
