/*
 * $Id: SecurityHelpers.java 5818 2023-06-15 11:41:07Z carsten $
 *
 * Copyright 2011 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.service.security;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Signature;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.KeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import java.util.Base64.Encoder;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESedeKeySpec;
import javax.crypto.spec.PBEKeySpec;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

/**
 * @author (c) 2023 - Carsten Moeller - info@osm2po
 */
public class SecurityHelpers {
    private SecurityHelpers() {/*STATIC*/}

    static {
        TrustManager[] trustAllCerts = new TrustManager[] {
            new X509TrustManager() { // One of many
                @Override public X509Certificate[] getAcceptedIssuers() {return null;}
                @Override public void checkClientTrusted(X509Certificate[] certs, String authType) {}
                @Override public void checkServerTrusted(X509Certificate[] certs, String authType) {}
            }
        };

        try {
            // Fuer HTTPS-CrossSiteResourcen, die von hier geholt werden, z.B. Bilder
            SSLContext sc = SSLContext.getInstance("SSL");
            sc.init(null, trustAllCerts, new SecureRandom());
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
            HttpURLConnection.setFollowRedirects(true);

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static SSLContext createSSLContextTLS(
            URL p12KsUrl, String ksPass) throws Exception {

        KeyStore keystore = loadKeyStorePKCS12(p12KsUrl, ksPass);

        KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
        kmf.init(keystore, ksPass.toCharArray());

        SSLContext sslContext = SSLContext.getInstance("TLS");
        sslContext.init(kmf.getKeyManagers(), null, null);

        return sslContext;
    }

    public static byte[] createDigestMD5(byte[] bytes) throws Exception {
        MessageDigest md = MessageDigest.getInstance("MD5");
        return md.digest(bytes);
    }

    public static KeyPair createKeyPairRSA() throws Exception {
        SecureRandom ran = new SecureRandom();
        KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
        keyGen.initialize(1024, ran);
        return keyGen.generateKeyPair();
    }

    public static byte[] createSignatureMD5RSA(
            byte[] bytes, PrivateKey privateKey) throws Exception {
        Signature sig = Signature.getInstance("MD5withRSA");
        sig.initSign(privateKey);
        sig.update(bytes);
        return sig.sign();
    }

    public static boolean verifySignatureMD5RSA(byte[] bytes,
            byte[] signature, PublicKey publicKey) throws Exception {
        Signature sig = Signature.getInstance("MD5withRSA");
        sig.initVerify(publicKey);
        sig.update(bytes);
        return sig.verify(signature);
    }

    public static RSAPublicKey pem64ToPublicKeyRSA(String pubKeyPem64) {
        try {
            byte[] pubKeyPem = Base64.getDecoder().decode(pubKeyPem64);
            KeyFactory kf = KeyFactory.getInstance("RSA");
            return (RSAPublicKey) kf.generatePublic(new X509EncodedKeySpec(pubKeyPem));
        } catch (Exception e) {
            return null;
        }
    }

    public static X509Certificate pem64ToCertX509(String certPem64) {
        try {
            byte[] certPem = Base64.getDecoder().decode(certPem64);
            CertificateFactory cf = CertificateFactory.getInstance("X509");
            Certificate cert = cf.generateCertificate(new ByteArrayInputStream(certPem));
            return (X509Certificate) cert;
        } catch (Exception e) {
            return null;
        }
    }

    public static boolean verifySignatureSHA256RSA(byte[] bytes,
            byte[] signature, PublicKey publicKey) {
        try {
            Signature sig = Signature.getInstance("SHA256withRSA");
            sig.initVerify(publicKey);
            sig.update(bytes);
            return sig.verify(signature);
        } catch (Exception e) {
            return false;
        }
    }


    public static byte[] encryptDESede(
            SecretKey key, byte[] bytes) throws Exception {
        Cipher cipher = Cipher.getInstance("DESede");
        cipher.init(Cipher.ENCRYPT_MODE, key);
        return cipher.doFinal(bytes);
    }

    public static byte[] decryptDESede(
            SecretKey key, byte[] bytes) throws Exception {
        Cipher cipher = Cipher.getInstance("DESede");
        cipher.init(Cipher.DECRYPT_MODE, key);
        return cipher.doFinal(bytes);
    }

    public static byte[] cryptRSA(byte[] bytes, Key key) throws Exception {

        BigInteger source = new BigInteger(bytes);
        BigInteger exponent = null;
        BigInteger modulus = null;
        BigInteger crypted = null;

        if (key instanceof PublicKey) {
            RSAPublicKey pubKey = (RSAPublicKey) key;
            exponent = pubKey.getPublicExponent();
            modulus = pubKey.getModulus();

        } else {
            RSAPrivateKey privKey = (RSAPrivateKey) key;
            exponent = privKey.getPrivateExponent();
            modulus = privKey.getModulus();
        }

        crypted = source.modPow(exponent, modulus);
        return crypted.toByteArray();
    }

    public static SecretKey createSecretKeyDESede() throws Exception {
        return KeyGenerator.getInstance("DESede").generateKey();
    }

    public static byte[] secretKey2BytesDESede(SecretKey key) throws Exception {
        SecretKeyFactory keyfactory = SecretKeyFactory.getInstance("DESede");
        DESedeKeySpec keyspec =
            (DESedeKeySpec) keyfactory.getKeySpec(key, DESedeKeySpec.class);
        return keyspec.getKey();
    }

    public static SecretKey bytes2SecretKeyDESede(byte[] bytes) throws Exception {
        DESedeKeySpec keyspec = new DESedeKeySpec(bytes);
        SecretKeyFactory keyfactory = SecretKeyFactory.getInstance("DESede");
        return keyfactory.generateSecret(keyspec);
    }

    public static Certificate loadCertX509(URL url) throws Exception {
        CertificateFactory cf = CertificateFactory.getInstance("X509");
        InputStream is = url.openStream();
        Certificate cert = cf.generateCertificate(is);
        is.close();
        return cert;
    }

    public static void extractKeystoreCert(
            KeyStore ks, String alias, File to) throws Exception {
        Certificate cert = getKeystoreCert(ks, alias);
        byte[] certBytes = cert.getEncoded();
        OutputStream os = new FileOutputStream(to);
        os.write(certBytes);
        os.close();
    }

    private static KeyStore loadKeyStore(String ksSpec,
            URL url, String ksPass) throws Exception {
        InputStream is = url.openStream();
        KeyStore ks = KeyStore.getInstance(ksSpec);
        ks.load(is, ksPass.toCharArray());
        is.close();
        return ks;
    }

    public static KeyStore loadKeyStorePKCS12(
            URL url, String ksPass) throws Exception {
        return loadKeyStore("PKCS12", url, ksPass);
    }

    public static KeyStore loadKeyStoreJKS(
            URL url, String ksPass) throws Exception {
        return loadKeyStore("JKS", url, ksPass);
    }

    public static Certificate getKeystoreCert(
            KeyStore ks, String alias) throws Exception {
        Certificate cert = ks.getCertificate(alias);
        return cert;
    }

    public static PrivateKey getKeystoreKey(
            KeyStore ks, String alias, String keyPass) throws Exception {
        Key key = ks.getKey(alias, keyPass.toCharArray());
        return (PrivateKey) key;
    }

    public static String encryptPasswordPbkdf2(final String password) throws Exception {
        final int keyLength = 512;
        final int iterations = 25000; // instead of 65536;

        byte[] salt = new byte[16];
        SecureRandom random = new SecureRandom();
        random.nextBytes(salt);

        // Check against tested Odoo Pwd "kaputt" which is stored in table as
        // $pbkdf2-sha512$25000$n9O611orJSRkTMm51xoDQA$vyxY/4zm3DcV1BETcuvZfNQUEDCogDawJIuMH0K5Z4DtMMNNyU8O0prMroobXqqj1FSzMJ0/E5K1B3B96Jsvpw
        // Therefore, uncomment the following Salt
        // salt = Base64.getDecoder().decode("n9O611orJSRkTMm51xoDQA"); // => Length=16

        KeySpec spec = new PBEKeySpec(password.toCharArray(), salt, iterations, keyLength);
        SecretKeyFactory factory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA512");
        byte[] hash = factory.generateSecret(spec).getEncoded();

        Encoder encoder64 = Base64.getEncoder().withoutPadding();
        String hash64 = encoder64.encodeToString(hash);
        String salt64 = encoder64.encodeToString(salt);

        String pwdPass = "$pbkdf2-sha512$" + iterations + '$' + salt64 + '$' + hash64;
        return pwdPass;
    }

}