/*
 * $Id: Tsp.java 5566 2022-02-27 18:38:18Z carsten $
 *
 * Copyright 2012/2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.tsp;

import static de.cm.osm2po.tsp.TspUtils.FINITY;
import static de.cm.osm2po.tsp.TspUtils.FINITY_BITS;
import static de.cm.osm2po.tsp.TspUtils.INFINITY;
import static de.cm.osm2po.tsp.TspUtils.calcTourCosts;
import static de.cm.osm2po.tsp.TspUtils.createGreedyRingTour;
import static de.cm.osm2po.tsp.TspUtils.improveInsertion;
import static de.cm.osm2po.tsp.TspUtils.improveTwoOpt;
import static de.cm.osm2po.tsp.TspUtils.isLast5TourOptimal;
import static java.lang.System.arraycopy;

import java.util.Arrays;

import de.cm.osm2po.tsp.TspResultLogger.EventListener;

/**
 * Statische Sammlung von Funktionen zum loesen eines optimalen
 * asymmetrischen und folglich auch symmetrischen TSPs.
 * @author (c) 2022 - Carsten Moeller - info@osm2po
 */
public class TspA {

    private TspA() {/*STATIC*/}

    public static TspResultLogger createPermuteAndPruneTour(final long[][] mx, EventListener eventListener) {
        int[] initialTour = createGreedyRingTour(mx);
        return improvePermuteAndPrune(mx, initialTour, 0d, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, final double tolerance, EventListener eventListener) {
        return improvePermuteAndPrune(mx, initialTour, 0, tolerance, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, long minCost, double tolerance, EventListener eventListener) {

        if (minCost <= 0) minCost = calcTourCosts(initialTour, mx);
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        int[] minTour = Arrays.copyOf(initialTour, initialTour.length);

        final long initialBounds = calcInitialBounds(mx, minTour, 0)[mx.length];

        TspResultLogger tspResultLogger = new TspResultLogger(eventListener);
        tspResultLogger.onStart(minTour, minCost, initialBounds);
        permuteAndPrune(1, tour.length-2, mx, tour, 0, null, minTour, minCost, tolerance, tspResultLogger);
        tspResultLogger.onStop();
        return tspResultLogger;
    }

    /**
     * Permutiert ueber alle Kombinationen und berechnet die beste Besuchsreihenfolge.<br>
     * Zudem werden hier, einem BranchAndBound sehr aehnlich, einige Grenzen betrachtet und
     * wenn bekannt, ein frueher Abbruch (Pruning) ausgeloest.<br>
     * Kurz gesagt: Dies ist ein PnP und kein BnB ;-).
     * @param x1 int Permutation Start-Index (Rekursions-Tiefe),
     *        normalerweise die zweite Tour-Station (Index = 1)
     * @param x2 int Permutation End-Index (inklusive),
     *        normalerweise die vorletzte Tour-Station
     * @param mx long[][] Adjazenz-Matrix (Diagonale: 0,
     *        Nicht erreichbar: {@link TspUtils#INFINITY})
     * @param tour int[] aktuelle Tour-Reihenfolge (Reference-Var).<br>
     *        Initial z.B. <code>{0,1,2,3...,8,0}</code>,<br>
     *        kann aber auch mittels z.B. {@link TspUtils#createGreedyRingTour(long[][])}
     *        vorbelegt werden.
     * @param cost long Aufsummierte Kosten bei Rekursions-Abstieg
     * @param mintour int[] aktuell beste Tour (Reference-Var)<br>
     *        Darf zwangslaeufig nicht das gleiche Array wie <code>tour</code> sein,
     *        sollte aber sehr wohl initial die gleichen Werte enthalten.
     * @param mincost long aktuell beste Kosten (Upper-Bound)<br>
     *        Duerfen initial auch unabhaengig von den eigentlichen Tour-Kosten gesetzt
     *        werden.<br>
     *        Durchaus nuetzlich fuer den Fall, bei dem die Anfangs-Reihenfolge gut aber
     *        die Kosten hoch und bereits niedrigere bekannt sind.
     * @param bps long[] Kosten der kuerzesten Kanten pro Station. (Bounds per Station)<br>
     *        Wird intern zur Weitergabe der aktuell berechneten Bounds verwendet
     *        und sollte beim Initialaufruf <code>null</code> sein.<br>
     *        Per Trick wird hier in die Bytes ab {@link TspUtils#INFINITY} der betroffene
     *        Vertex hineinkodiert.
     * @param tolerance double &gt;0 ueberschaetzt die minimalen Restkosten und findet
     *        so schneller eine gute aber auch nicht immer optimale Loesung.<br>
     *        <code>0.01</code> bedeutet eine Toleranz von etwa 1%. Oder anders
     *        ausgedrueckt: Schummelt die Lower Bound ein wenig nach oben.<br>
     *        Dieser Faktor wirkt lediglich auf die Heuristik, nicht auf die bereits
     *        ermittelten realen Kosten und geht mit jedem Rekursionsabstieg
     *        (dichter am Ziel) gegen 0.
     * @param trl {@link TspResultLogger}
     * @return long Kosten der Tour von <code><b>x1</b>-1 nach <b>x2</b>+1</code>
     * <p>
     * <b>Einige Tipps:</b>
     * <ul><li>
     * Mittels <code>x1 und x2</code> lassen sich auch partielle Permutationen realisieren.
     * So waere es moeglich, ABC stabil zu lassen, DEF zu permutieren und GH ebenfalls nicht
     * anzufassen. Als Tour wuerde diese Funktion dann C-DEF-G betrachten.
     * </li><li>
     * Wird dieser Funktion bereits eine verbesserte Anfangs-Konfiguration,
     * z.B. mittels {@link TspUtils#createGreedyRingTour(long[][])} inklusive Startkosten uebergeben,
     * so zeigen Analysen, dass das Ergebnis haeufig deutlich schneller gefunden wird..
     * </li></ul>
     */
    private static long permuteAndPrune(int x1, int x2,
            final long[][] mx, int[] tour, long cost, long[] bps, int[] mintour, long mincost,
            final double tolerance, TspResultLogger trl) {

        trl.onPermutation(tour);
        if (2 == x1) trl.onTopLevelProgress(tour);

        if (x1 == x2) { // Waere dies ein Baum, so waere dies ein Blatt
            cost += mx[tour[x1-1]][tour[x1]]; // Kosten vom vorigen zu diesem Knoten addieren
            cost += mx[tour[x1]][tour[x1+1]]; // und natuerlich plus Kosten zum Ziel
            if (cost <= mincost) {
                mincost = cost;
                arraycopy(tour, 0, mintour, 0, tour.length); // Dies ist die bisher beste Tour
                trl.onAlternativeOrBetterTour(tour, cost);

                /* Hilft ab und zu aus dem Tal der Traenen (Lokales Minimun) herauszukommen.
                Beispielsweise braucht ein ftv33 mit Greedy unter einer Minute, mit einer simplen
                Ring-Tour (0,1,2,3,...,0) hingegen ewig. Baut man allerdings dies hier ein, so
                pressen wir auch eine schlechte Anfangs-Tour in einen ertraeglichen Zeitrahmen.
                Wir verbessern so frueh die Upper-Bounds ohne dabei den eigentlichen Algo zu
                unterbrechen. Das ist erlaubt.
                 */
                int[] randomTour = improveTwoOpt(mx, improveInsertion(mx, tour));
                long randomCost = calcTourCosts(randomTour, mx);
                if (randomCost < mincost) {
                    mincost = randomCost;
                    trl.onUpperBoundsReduced(randomTour, randomCost);
                }
            }

        } else if (isLast5TourOptimal(mx, tour, x1-1)) {

            if (x1 < 2) bps = calcInitialBounds(mx, tour, x1);
            else bps = calcNextBounds(mx, bps, tour, x1);
            long bounds = bps[mx.length]; // Konvention: Summe aller Bounds in [last]

            // Mit der Toleranz laesst sich "mal eben schnell" schonmal etwas "Gutes", manchmal gar "Optimales" finden!
            if (tolerance > 0) {
                // Die Formel bedeutet, dass die Toleranz am Anfang voll und zum Ziel hin linear gegen 0 wirkt.
                double overbound = 1+(x2-x1-1)*tolerance/x2; // DevisionByZero-Safe if x2 != 0
                bounds *= overbound;
            }

            for (int k = x1; k <= x2; k++) { // Schleife: Position k gegen x durchtauschen (Standard-Permutation)

                // Vorhersage: Kosten vom vorigen zu diesem Knoten (Achtung! k veraendert sich!)
                final long cc = cost + mx[tour[x1-1]][tour[k]]; // Reale Tourkosten bis hier. cc = cumulated costs

                // Tourkosten bis hier plus minimale Kosten von hier zum Ziel (outgoing edges = lower-bounds)
                // noch kleiner als bisher kuerzeste Strecke? Also noch unterhalb der aktuellen UpperBounds?
                if (cc + bounds <= mincost) {
                    // Oben: "<=" anstatt "<" verwendet! Test: Zusaetzliche 0 in modifizierten ftv33-Tour eingebaut.
                    // MinCost ebenfalls korrekt. Proof! Aber eben mit "<=" mehr (alle?) Loesungen gefunden.
                    int c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                    mincost = permuteAndPrune(x1 +1, x2, mx, tour, cc, bps, mintour, mincost, tolerance, trl);
                    c = tour[k]; tour[k] = tour[x1]; tour[x1] = c; // zuruecktauschen
                }
            }
        }
        return mincost;
    }

    /**
     * Ermittelt fuer jede Station die guenstigsten ausgehenden Kanten.<br>
     * In Summe sind dies die Lower Bounds bzw. minimalen Restkosten der aktuellen Tour.<br>
     * Diese Funktion arbeitet ebenfalls fuer Nicht-TSP-Touren korrekt,
     * also fuer den Fall, dass Start ungleich Ziel ist.
     * @param mx long[][] Adjazenz-Matrix
     * @param tour int[] aktuelle Tour
     * @param x1 int Start-Index der Berechnung.<br>
     * @return long[] Minimal-Kosten pro Station NOT NULL.
     * <p>
     * <strong>Anmerkungen:</strong><br>
     * Die letzte Station wird dabei nicht beruecksicht, da sie keine ausgehenden
     * Kanten mehr hat.<br>
     * Die eingehenden Kanten bis Station <code>x1</code> (inklusive) sind ja bereits
     * real berechenbar, wohingegen die Restkosten, also die ausgehenden Kanten, wiederum
     * ab <code>x1</code> (inklusive) noch in der Reihenfolge permutieren. Dabei ist
     * die Station an <code>x1</code> selber noch nicht ermittelt und ein Heranziehen
     * der eingehenden Kante hier nicht stabil.<br>
     * Die Abschaetzung ueber den Mittelwert der ein- und ausgehenden Kanten, so
     * wie sie <a href="https://gtl.csa.iisc.ac.in/dsa/node187.html">hier</a>
     * beschrieben wird, ist irrefuehrend und Analysen zeigen, dass sie fuer
     * ATSPs oder fehlende Nebenbedingungen wie Symmetrie oder gegebener Dreiecksungleichung
     * nicht herangezogen werden kann. Zudem basiert sie auf der Annahme, dass
     * die Bounds einen Ring bilden, was ja fuer die restlichen Stationen nicht zutrifft.
     * </p>
     */
    private static long[] calcInitialBounds(final long[][] mx, final int[] tour, final int x1) {
        final long[] bps = new long[mx.length +1];
        long bounds = 0;
        for (int x = x1; x < tour.length -1; x++) {
            final int vx = tour[x]; // Vertex von X
            final long bound = calcOneBound(mx, bps, tour, x1, vx);
            bounds += bound;
        }
        bps[mx.length] = bounds; // Konvention: Summe aller Bounds in [last]
        return bps;
    }

    private static long[] calcNextBounds(final long[][] mx,
            final long[] bps, final int[] tour, final int x1) {
        long[] bpsNew = bps.clone();

        final int v0 = tour[x1-1]; // Vorgaenger-Vertex
        final long bps0 = bps[v0]; // Info vom Vorgaenger - jetzt nicht mehr Teil der Bounds
        final long bnd0 = bps0 & FINITY; // Bit-Trick: Bounds dieser Station herausholen

        final long oldBounds = bps[mx.length]; // Konvention: Summe aller Bounds in [last]
        long newBounds = oldBounds - bnd0;

        // Bounds der uebrigen Vertices betrachten und falls abhaengig, neu berechnen
        for (int x = x1; x < tour.length-1; x++) {
            final int vx = tour[x]; // Vertex an x
            final long bpsX = bps[vx]; // Bounds und verursachende Station herausholen

            final int vmin = (int) (bpsX >>> FINITY_BITS); // Bit-Trick: Index herausholen

            if (vmin == v0) { // Vertex ist vom soeben entfernten betroffen
                long bndOld = bpsNew[vx] & FINITY; // Bit-Trick: Alte Bounds des Vertex
                long bndNew = calcOneBound(mx, bpsNew, tour, x1, vx); // diesen neu berechnen
                newBounds += bndNew - bndOld; // Gesamt-Bound-Summe korrigieren
            }
        }

        bpsNew[mx.length] = newBounds; // Konvention: Summe aller Bounds in [last]
        return bpsNew;
    }

    private static long calcOneBound(final long[][] mx,
            final long[] bps, final int[] tour, final int x1, final int v) {
        long minOut = INFINITY; // Minimum ausgehender Kanten
        long minvOut = 0;

        final int x2 = tour.length -1;
        for (int x = x1; x <= x2; x++) { // Iteration ueber Ziel-Stationen inklusive Target
            int vx = tour[x]; // Vertex(Station) an Position x
            if (v == vx) continue; // Selbstbesuch ignorieren
            final long out = mx[v][vx]; // Ausgehende Kante
            if (out <= minOut) {
                minOut = out; // Ausgehende Kante
                minvOut = vx; // betroffene Zielstation merken
            }
        }

        if (INFINITY == minOut) minOut = 0; // Nix gefunden, dann 0?

        // Kombi aus Bound-Pro-Station und Verursacher-Vertex im Array speichern
        bps[v] = (minvOut << FINITY_BITS) | minOut;

        return minOut; // Die Bounds separat zurueckliefern
    }

} // EOC
