package de.cm.osm2po.tsp;

import static de.cm.osm2po.tsp.TspResultLogger.Event.ALIVE;
import static de.cm.osm2po.tsp.TspResultLogger.Event.ALTERNATIVE;
import static de.cm.osm2po.tsp.TspResultLogger.Event.IMPROVED;
import static de.cm.osm2po.tsp.TspResultLogger.Event.INIT;
import static de.cm.osm2po.tsp.TspResultLogger.Event.TOPLEVELCHANGED;
import static de.cm.osm2po.tsp.TspResultLogger.Event.UPPERBOUNDSREDUCED;

import java.util.ArrayList;
import java.util.List;

import de.cm.osm2po.misc.Utils;

public class TspResultLogger {

    private static final long CARRYON = 1000000000;

    public final EventListener eventListener;

    private int[] initialTour;
    private long initialCosts;
    private long initialBounds;
    private List<int[]> tours = new ArrayList<>(); // Anzahl begrenzt
    private long nSolutions; // Echte Anzahl gefundener Alternativen
    private long costs;
    private long permutCounter;
    private long permutCarryCounter;
    private long stopwatchStart;
    private long stopwatchStop;

    public static interface EventListener {
        void on(Event event, final int[] tour, long cost, String counter);
    }

    public TspResultLogger(EventListener eventListener) {
        this.eventListener = eventListener;
    }

    public enum Event {
        INIT('V'), // trigger start config
        IMPROVED('O'), // better tour found
        ALTERNATIVE('o'), // Found another tour with same costs
        UPPERBOUNDSREDUCED('-'), // lower costs for tour known
        TOPLEVELCHANGED('|'), // ping when first level changed
        ALIVE('/'); // ping when long-max overflow reached

        public final char symbol;

        Event(char symbol) {
            this.symbol = symbol;
        }
    }

    public long getCosts() {return this.costs;}
    public long getInitialCosts() {return this.initialCosts;}
    public long getInitialBounds() {return this.initialBounds;}
    public int[] getInitialTour() {return this.initialTour;}
    public int[] getAnyTour() {return this.isEqualOrBetterFound() ? this.tours.get(0) : this.initialTour;}
    public List<int[]> getSomeTours() {return this.tours;}
    public long getNumberOfSolutions() {return this.isEqualOrBetterFound() ? this.nSolutions : 1;}
    public String getRunTime() {return Utils.msToDhms(this.stopwatchStop - this.stopwatchStart);}
    public boolean isEqualOrBetterFound() {return !this.tours.isEmpty();}

    public final String getPermutationCounter() {
        return this.permutCarryCounter + ":" + this.permutCounter;
    }

    public final void onStart(final int[] initialTour, long initialCosts, long initialBounds) {
        this.initialTour = initialTour;
        this.initialBounds = initialBounds;
        this.initialCosts = initialCosts;
        this.costs = initialCosts;
        this.tours.clear();
        this.stopwatchStart = System.currentTimeMillis();

        this.eventListener.on(INIT, initialTour, initialCosts, this.getPermutationCounter());
    }

    public final void onStop() {
        this.stopwatchStop = System.currentTimeMillis();
    }

    public final void onAlternativeOrBetterTour(final int[] tour, long costs) {
        if (costs < this.costs) {
            this.nSolutions = 0;
            this.tours.clear();
            this.eventListener.on(IMPROVED, tour, costs, this.getPermutationCounter());
        } else {
            this.eventListener.on(ALTERNATIVE, tour, costs, this.getPermutationCounter());
        }
        this.costs = costs;
        if (++this.nSolutions <= 10) { // WorstCase: Millionen, daher abschneiden
            this.tours.add(tour.clone());
        }
    }

    public final void onPermutation(final int[] tour) {
        if (++this.permutCounter == CARRYON) {
            this.permutCounter = 0;
            this.permutCarryCounter++;
            this.eventListener.on(ALIVE, tour, 0, this.getPermutationCounter());
        }
    }

    public final void onUpperBoundsReduced(final int[] tour, long cost) {
        this.eventListener.on(UPPERBOUNDSREDUCED, tour, cost, this.getPermutationCounter());
    }

    public final void onTopLevelProgress(final int[] tour) {
        this.eventListener.on(TOPLEVELCHANGED, tour, 0, this.getPermutationCounter());
    }

}
