/*
 * $Id: TspS.java 5771 2023-03-30 18:14:59Z carsten $
 *
 * Copyright 2012/2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.tsp;

import static de.cm.osm2po.tsp.TspUtils.INFINITY;
import static de.cm.osm2po.tsp.TspUtils.calcTourCosts;
import static de.cm.osm2po.tsp.TspUtils.createGreedyRingTour;
import static de.cm.osm2po.tsp.TspUtils.improveInsertion;
import static de.cm.osm2po.tsp.TspUtils.improveTwoOpt;
import static java.lang.System.arraycopy;

import java.util.Arrays;

import de.cm.osm2po.tsp.TspResultLogger.EventListener;

/**
 * Loest einen symmetrischen TSP auf Basis der Two-Tour-Theorie.<br>
 * Ist eine Adjazenzmatrix symmetrisch und die Dreiecksungleichung gegeben,
 * so lassen sich die Bounds mittels <code>(Min1+Min2)/2</code> genauer
 * berechnen, was die Rechenzeit enorm verkuerzt.<br>
 * Die Formel beruht dabei auf
 * <a href="https://www.geeksforgeeks.org/traveling-salesman-problem-using-branch-and-bound-2/">
 * diesem Code</a> und dieser wiederum auf
 * <a href="https://gtl.csa.iisc.ac.in/dsa/node187.html">diesem Beweis</a>.<br>
 * Ferner werden hier, anders als im Beispiel, die Bounds nur 1x berechnet anstatt diese
 * jedes Mal neu per Schleife zu ermitteln, was die Rechenzeit um Faktor 10 verkuerzt.<br>
 * Ferner mussten einige moegliche Fehlerquellen umgeschrieben werden.
 * <ul>
 * <li> Original:
 * <pre>
 * if (level==1) curr_bound -= ((firstMin(adj, curr_path[level - 1]) + firstMin(adj, i))/2);
 * else curr_bound -= ((secondMin(adj, curr_path[level - 1]) + firstMin(adj, i))/2);
 * </pre>
 * Ein Kommentator bemerkt ebenfalls falsche Ergebnisse und schlaegt die folgende
 * Korrektur vor, die auch in dieser Klasse verwendet wird und eher richtig erscheint:
 * <pre>
 * if (level==1) curr_bound -= (secondMin(adj, curr_path[level-1]) + secondMin(adj, i)) /2;
 * else curr_bound -= (firstMin(adj, curr_path[level-1]) + secondMin(adj, i)) /2;
 * </pre>
 * </li><li>
 * Eine falsche Uebersetzung von C nach Java duerfte ebenfalls in Grenzfaellen falsch laufen:<br>
 * <code>curr_bound = (curr_bound==1)? curr_bound/2 + 1 : curr_bound/2;</code><br>
 * In C ist dies <code>curr_bound&amp;1</code>, also eine Even-Odd-Pruefung.
 * Dies wurde fuer Java offensichtlich falsch uebersetzt.<br>
 * Ueberhaupt: Integer-Rundungsprobleme sind ein entscheidendender Grund, warum manchmal zwar die
 * optimale Tour gefunden wird, die Rueckwaerts-Variante jedoch nicht. Ein symmetrisches System
 * hat IMMER eine gerade Anzahl von Loesungen.
 * </li>
 * </ul>
 * @author (c) 2022 - Carsten Moeller - info@osm2po
 * @see TspX
 * @see TspA
 */
public class TspS {

    private TspS() {/*STATIC*/}

    public static TspResultLogger createPermuteAndPruneTour(final long[][] mx, EventListener eventListener) {
        int[] initialTour = createGreedyRingTour(mx);
        return improvePermuteAndPrune(mx, initialTour, 0d, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, final double tolerance, EventListener eventListener) {
        return improvePermuteAndPrune(mx, initialTour, 0, tolerance, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, long minCost, double tolerance, EventListener eventListener) {

        if (minCost <= 0) minCost = calcTourCosts(initialTour, mx);
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        int[] minTour = Arrays.copyOf(initialTour, initialTour.length);
        long[][] bps = calcInOutMins12(mx);
        long bounds = calcInitialBounds(bps);

        TspResultLogger tspResultLogger = new TspResultLogger(eventListener);
        tspResultLogger.onStart(minTour, minCost, bounds/2);
        permuteAndPrune(1, tour.length-2, mx, tour, 0,
                bps, bounds, minTour, minCost, tolerance, tspResultLogger);
        tspResultLogger.onStop();
        return tspResultLogger;
    }

    private static long permuteAndPrune(int x1, int x2, final long[][] mx,
            int[] tour, long cost, final long[][] bps, final long bounds, int[] mintour, long mincost,
            final double tolerance, TspResultLogger trl) {

        trl.onPermutation(tour);
        if (2 == x1) trl.onTopLevelProgress(tour);

        if (x1 == x2) {
            cost += mx[tour[x1-1]][tour[x1]];
            cost += mx[tour[x1]][tour[x1+1]];
            if (cost <= mincost) {
                mincost = cost;
                arraycopy(tour, 0, mintour, 0, tour.length);
                trl.onAlternativeOrBetterTour(tour, cost);

                /* Hilft ab und zu aus dem Tal der Traenen (Lokales Minimun) herauszukommen.
                Beispielsweise braucht ein ftv33 mit Greedy unter einer Minute, mit einer simplen
                Ring-Tour (0,1,2,3,...,0) hingegen ewig. Baut man allerdings dies hier ein, so
                pressen wir auch eine schlechte Anfangs-Tour in einen ertraeglichen Zeitrahmen.
                Wir verbessern so frueh die Upper-Bounds ohne dabei den eigentlichen Algo zu
                unterbrechen. Das ist erlaubt.
                 */
                int[] randomTour = improveTwoOpt(mx, improveInsertion(mx, tour));
                long randomCost = calcTourCosts(randomTour, mx);
                if (randomCost < mincost) {
                    mincost = randomCost;
                    trl.onUpperBoundsReduced(randomTour, randomCost);
                }

            }

        } else if (TspUtils.isLast5TourOptimal(mx, tour, x1-1)) {
            final int v0 = tour[x1-1]; // Vorgaenger-Vertex
            final long bps0 = bps[x1==1 ? MINOUT2 : MINOUT1][v0];

            for (int k = x1; k <= x2; k++) {
                final int vk = tour[k];
                final long newCost = cost + mx[v0][vk]; // Reale Kosten bis hierher

                long newBounds = bounds - (bps0 + bps[MINOUT2][vk]);

                // if (newCost + mx[vk][0] >= mincost) { // Was ist hier falsch?
                //     continue;
                // }

                // Mit der Toleranz laesst sich "mal eben schnell" schonmal etwas "Gutes", manchmal gar "Optimales" finden!
                if (tolerance > 0) {
                    // Die Formel bedeutet, dass die Toleranz am Anfang voll und zum Ziel hin linear gegen 0 wirkt.
                    final double overbound = 1+(x2-x1-1)*tolerance/x2; // DevisionByZero-Safe if x2 != 0
                    newBounds *= overbound;
                }

                if (newCost + (newBounds>>>1) <= mincost) {
                    int c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                    mincost = permuteAndPrune(x1 +1, x2, mx, tour, newCost,
                            bps, newBounds, mintour, mincost, tolerance, trl);
                    c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                }
            }
        }
        return mincost;
    }

    private static long calcInitialBounds(final long[][] bps) {
        long bounds = 0;
        for (int i = 0; i < bps[0].length; i++) {
            bounds += bps[MINOUT1][i] + bps[MINOUT2][i];
        }
        return bounds;
    }


    // named indexes for calcInOutMins12:
    private final static int MINOUT1 = 0;
    private final static int MINOUT2 = 1;
    private final static int MININ1 = 2;
    private final static int MININ2 = 3;

    private static long[][] calcInOutMins12(final long[][] mx) {
        // only correct for symmetric TSP !
        final long[][] bps = new long[4][mx.length];

        for (int i = 0; i < mx.length; i++) {
            long minOut1 = INFINITY;
            long minOut2 = INFINITY;
            long minIn1 = INFINITY;
            long minIn2 = INFINITY;

            for (int j = 0; j < mx.length; j++) {

                if (j == i) continue;

                final long out = mx[i][j];
                if (out <= minOut1) {
                    minOut2 = minOut1;
                    minOut1 = out;
                } else if (out < minOut2) {
                    minOut2 = out;
                }

                final long in = mx[j][i];
                if (in <= minIn1) {
                    minIn2 = minIn1;
                    minIn1 = in;
                } else if (in < minIn2) {
                    minIn2 = in;
                }
            }

            bps[MINOUT1][i] = minOut1; // [0]
            bps[MINOUT2][i] = minOut2; // [1]
            bps[MININ1][i] = minIn1; // [2]
            bps[MININ2][i] = minIn2; // [3]
        }

        return bps;
    }


} // EOC
