/*
 * $Id: TspX.java 5800 2023-04-25 10:41:47Z carsten $
 *
 * Copyright 2012/2022 - Carsten Moeller Software
 * All rights reserved.
 * Use is subject to license terms.
 */

package de.cm.osm2po.tsp;

import static de.cm.osm2po.tsp.TspUtils.INFINITY;
import static de.cm.osm2po.tsp.TspUtils.calcTourCosts;
import static de.cm.osm2po.tsp.TspUtils.createGreedyRingTour;
import static de.cm.osm2po.tsp.TspUtils.improveInsertion;
import static de.cm.osm2po.tsp.TspUtils.improveTwoOpt;
import static java.lang.System.arraycopy;

import java.util.Arrays;

import de.cm.osm2po.tsp.TspResultLogger.EventListener;

/**
 * Loest einen (a)symmetrischen TSP.<br>
 * Dies ist eine Eigenkomposition, die sowohl auf dem {@link TspS},
 * als auch auf dem {@link TspA} beruht. Die Kernidee ist, die Bounds
 * der restlichen Kanten selektiv neu zu berechnen, anstatt einfach
 * nur den Vorgaenger und Nachfolger nach Herausnahme der Zwischenstation,
 * wenn Teil der Anfangs-Tour, zu ueberbruecken.<br>
 * Bei letzterem sinken die Bounds pro Station stetig. Das Wegnehmen von
 * Kanten sollte einzelne Bounds aber auch wieder steigen lassen duerfen.<br>
 * Zudem werden die initialen Bounds analog des {@link TspS} berechnet.
 * Diese sind hoeher als die des {@link TspA}, da ja TwoTour-Theorie.<br>
 * Diese Variante konnten bereits viele Beispielprobleme
 * (darunter einige aus der TspLib) erfolgreich loesen. Das hier zusaetzlich
 * erforderliche O(n) fuer die Bounds-Berechnung ist in kleineren
 * Problemstellungen dem O(1) des {@link TspS} unterlegen,
 * zeigt aber mit steigender Groesse des Problems seine Ueberlegenheit.<br>
 * Neue Tests zeigen, dass diese Variante haeufig bis sehr dicht an eine
 * optimale Loesung herankommt.
 * @author (c) 2022 - Carsten Moeller - info@osm2po
 */
public class TspX {

    private TspX() {/*STATIC*/}

    public static TspResultLogger createPermuteAndPruneTour(final long[][] mx, EventListener eventListener) {
        int[] initialTour = createGreedyRingTour(mx);
        return improvePermuteAndPrune(mx, initialTour, 0d, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, final double tolerance, EventListener eventListener) {
        return improvePermuteAndPrune(mx, initialTour, 0, tolerance, eventListener);
    }

    public static TspResultLogger improvePermuteAndPrune(final long[][] mx,
            final int[] initialTour, long minCost, double tolerance, EventListener eventListener) {

        if (minCost <= 0) minCost = calcTourCosts(initialTour, mx);
        int[] tour = Arrays.copyOf(initialTour, initialTour.length);
        int[] minTour = Arrays.copyOf(initialTour, initialTour.length);
        long[] bps = calcInitialBounds(mx, tour);

        TspResultLogger tspResultLogger = new TspResultLogger(eventListener);
        tspResultLogger.onStart(minTour, minCost, bps[mx.length << 1]/2);
        permuteAndPrune(1, tour.length-2, mx, tour, 0, bps, minTour, minCost, tolerance, tspResultLogger);
        tspResultLogger.onStop();
        return tspResultLogger;
    }

    private static long permuteAndPrune(int x1, int x2, final long[][] mx, int[] tour, long cost,
            long[] bps, int[] mintour, long mincost, final double tolerance, TspResultLogger trl) {

        trl.onPermutation(tour);
        if (2 == x1) trl.onTopLevelProgress(tour);

        if (x1 == x2) { // Waere dies ein Baum, so waere dies ein Blatt
            cost += mx[tour[x1-1]][tour[x1]]; // Kosten vom vorigen zu diesem Knoten addieren
            cost += mx[tour[x1]][tour[x1+1]]; // und natuerlich plus Kosten zum Ziel
            if (cost <= mincost) {
                mincost = cost;
                arraycopy(tour, 0, mintour, 0, tour.length); // Dies ist die bisher beste Tour
                trl.onAlternativeOrBetterTour(tour, cost);

                /* Hilft ab und zu aus dem Tal der Traenen (Lokales Minimun) herauszukommen.
                Beispielsweise braucht ein ftv33 mit Greedy unter einer Minute, mit einer simplen
                Ring-Tour (0,1,2,3,...,0) hingegen ewig. Baut man allerdings dies hier ein, so
                pressen wir auch eine schlechte Anfangs-Tour in einen ertraeglichen Zeitrahmen.
                Wir verbessern so frueh die Upper-Bounds ohne dabei den eigentlichen Algo zu
                unterbrechen. Das ist erlaubt.
                 */
                int[] randomTour = improveTwoOpt(mx, improveInsertion(mx, tour));
                long randomCost = calcTourCosts(randomTour, mx);
                if (randomCost < mincost) {
                    mincost = randomCost;
                    trl.onUpperBoundsReduced(randomTour, randomCost);
                }

            }

        } else if (TspUtils.isLast5TourOptimal(mx, tour, x1-1)) {
            long newBounds = calcNextBounds(mx, bps, tour, x1);
            newBounds >>= 1; // Halbieren, da TwoTourBounds

            // Mit der Toleranz laesst sich "mal eben schnell" schonmal etwas "Gutes", manchmal gar "Optimales" finden!
            if (tolerance > 0) {
                // Die Formel bedeutet, dass die Toleranz am Anfang voll und zum Ziel hin linear gegen 0 wirkt.
                final double overbound = 1+(x2-x1-1)*tolerance/x2; // DevisionByZero-Safe if x2 != 0
                newBounds *= overbound;
            }

            for (int k = x1; k <= x2; k++) {
                final long newCost = cost + mx[tour[x1-1]][tour[k]]; // Vorhersage [k] Tourkosten bis hier

                /* ... komisch: haeufig sogar langsamer???
                final long costToTarget = mx[tour[x1]][tour[x2+1]]; // Vorhersage [k] Kosten von hier zum Ziel
                if (costToTarget > newBounds) {
                    newBounds = costToTarget;
                }
                */

                if (newCost + newBounds <= mincost) {
                    int c = tour[k]; tour[k] = tour[x1]; tour[x1] = c;
                    mincost = permuteAndPrune(x1 +1, x2, mx, tour, newCost, bps, mintour, mincost, tolerance, trl);
                    c = tour[k]; tour[k] = tour[x1]; tour[x1] = c; // zuruecktauschen
                }
            }

            resetPreviousBounds(mx.length, bps);
        }
        return mincost;
    }

    private static long[] calcInitialBounds(final long[][] mx, final int[] tour) {
        final long[] bps = new long[mx.length * 10];
        long bounds = 0;

        for (int x = 0; x < tour.length-1; x++) {

            long minOut1 = INFINITY;
            long minOut2 = INFINITY;
            long minvOut1 = -1;
            long minvOut2 = -1;

            final int vx = tour[x]; // Vertex von X

            for (int y = 1; y < tour.length; y++) { // Iteration ueber moegliche Ziele inklusive Target
                int vy = tour[y]; // Vertex an Y
                if (vx == vy) continue; // Selbstbesuch (Diagonale) bzw. Start=Ziel verhindern

                final long out = mx[vx][vy]; // Ausgehende Kante
                if (out <= minOut1) {
                    minOut2 = minOut1; // Voriger Min1 wird Min2
                    minvOut2 = minvOut1;
                    minOut1 = out; // Neues Min1 : Ausgehende Kante
                    minvOut1 = vy; // Neues Min1 : betroffene Zielstation merken
                } else if (out < minOut2) { // Unabhaengig von Min1 kann Min2 noch kleiner werden
                    minOut2 = out;
                    minvOut2 = vy;
                }
            }

            final long bound = minOut1 + minOut2;
            bounds += bound; // Aufsummieren
            bps[vx] = x==0 ? minOut2 << 1 : bound; // Ausnahme bei Level 0

            // Verursacher-Vertices im 2.Array-Abschnitt merken (Bit-Trick)
            final long minv = minvOut2 << 32 | minvOut1;
            bps[vx + mx.length] = minv;
        }

        final int bndIdx = mx.length << 1;
        bps[bndIdx] = bounds; // TwoTour-Bounds

        // Push an empty pseudo stack on top
        int stackIdx = bndIdx +1;
        bps[stackIdx] = stackIdx; // Stack empty = [x]=x
        return bps;
    }

    private static long calcOneBound(final long[][] mx, final long[] bps, final int[] tour, final int x1, final int v) {
        long minOut = INFINITY;
        long minvOut = -1;

        final long[] mxv = mx[v]; // 1. Array-Dimension 1x holen -> JavaPerformance!
        for (int x = x1; x < tour.length; x++) { // Iteration ueber moegliche restliche Ziele inklusive Target
            int vx = tour[x]; // Vertex(Station) an Position x
            if (v == vx) continue; // Selbstbesuch (Diagonale) bzw. Start=Ziel verhindern
            final long out = mxv[vx]; // Ausgehende Kante, kein minOut1 und 2 mehr hier!
            if (out <= minOut) {
                minOut = out; // Neues Min1 : Ausgehende Kante
                minvOut = vx; // Neues Min1 : betroffene Zielstation merken
            }
        }

        // Bounds dieses Vertex im 1.Array-Abschnitt merken
        minOut <<= 1; // minOut doppeln, da immer noch TwoTour-Bounds
        bps[v] = minOut;

        // Verursacher-Vertices im 2.Array-Abschnitt merken (Bit-Trick)
        final long minv = minvOut << 32 | minvOut; // Beide gleich jetzt
        bps[v + mx.length] = minv;

        return minOut;
    }


    private static long calcNextBounds(final long[][] mx, final long[] bps, final int[] tour, final int x1) {
        final int bndIdx = mx.length << 1; // [index] der aktuellen Bounds
        int stackIdx = (int) bps[bndIdx+1]; // Dahinter kommt der aktuelle Index-Zeiger auf letztes Stack-Element

        final int v0 = tour[x1-1]; // Vorgaenger-Vertex
        final long bnd0 = bps[v0]; // TwoTourBounds der Vorgaenger-Station herausholen
        final long oldBounds = bps[bndIdx]; // Die Summe der Bounds ebenso
        long newBounds = oldBounds - bnd0; // Bounds der vorigen Station jetzt abziehen

        bps[++stackIdx] = oldBounds; // Push

        long stackCounter = 0;
        // Bounds der uebrigen Vertices betrachten und falls abhaengig vom soeben entfernten, neu berechnen
        for (int x = x1; x < tour.length -1; x++) { // -1, weil Target keine ausgehenden Kanten hat.
            final int vx = tour[x]; // Vertex an x
            final long v12 = bps[vx + mx.length]; // Vertices der betroffenen Stationen

            final int v1 = (int) (v12); // Erster betroffener Vertex
            final int v2 = (int) (v12 >>> 32); // Zweiter betroffener Vertex

            if (v1 == v0 || v2 == v0) { // Vertex ist vom soeben entfernten betroffen
                final long bpsOld = bps[vx];

                // Push previous bps entries to stack
                bps[++stackIdx] = v12;
                bps[++stackIdx] = bpsOld;
                bps[++stackIdx] = vx;
                stackCounter++;

                final long bndNew = calcOneBound(mx, bps, tour, x1, vx); // Jetzt neue
                newBounds += bndNew - bpsOld;
            }
        }

        bps[++stackIdx] = stackCounter; // Anzahl ebenfalls auf Stack - beim rueckwaerts Einlesen dann vorne.
        bps[bndIdx+1] = stackIdx;

        // Workaround aufgrund diverser Tests mit FTV44.
        // Bounds um gerade einmal 5 zu hoch innerhalb der Pruning-Schleife.
        // Aber jetzt ist der Dantzig42 kaum noch loesbar, da sehr kleine mx-Werte.
        // TODO Genauer pruefen, wann und wo diese Korrektur vorgenommen werden darf oder muss.
//        newBounds -= 1; // Auf Rundungsprobleme deutet aktuell nichts hin!

        bps[bndIdx] = newBounds; // Konvention: Echte Summe aller Bounds in [last]
        return newBounds;
    }

    private static boolean resetPreviousBounds(final int mxLen, final long[] bps) {

        final int bndIdx = mxLen << 1; // Konvention: [index] der aktuellen Bounds ist hier
        int stackIdx = (int) bps[bndIdx+1]; // Dahinter kommt der aktuelle Index-Zeiger auf letztes Stack-Element
        if (stackIdx == bndIdx +1) { // Wenn der Index auf sich selbst zeigt [x]=x
            return false; // dann ist Stack empty
        }

        // Pop-Stack
        int n = (int) bps[stackIdx--]; // Anzahl gepushter Vertices
        for (int i = 0; i < n; i++) {
            final int vx = (int) bps[stackIdx--]; // Betroffener Vertex
            final long vb = bps[stackIdx--]; // Vorige Bounds des Vertex
            final long vt = bps[stackIdx--]; // Vorige Zeiger auf betroffene andere Vertices

            // und resetten
            bps[vx] = vb;
            bps[vx + mxLen] = vt;
        }
        final long oldBounds = bps[stackIdx--]; // Vorige Gesamt-Bounds
        bps[bndIdx] = oldBounds; // Bounds ebenfalls resetten

        bps[bndIdx+1] = stackIdx; // Stack selbst wieder verkleinern

        return true;
    }

} // EOC
